#include <limits.h>
#include <sys/types.h>
#include <sys/scheduler.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/mman.h>
#include <dirent.h>
#include <sys/keyboard.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <pci/pci.h>
#include <byteorder.h>

#include <ns/ns.h>
#include <devio/devio.h>
#include <sys/utsname.h>

static void __uname(int argc, char **argv);
static void __whoami(int argc, char **argv);
static void __lshw(int argc, char **argv);
static void __lspci(int argc, char **argv);
static void __hw_mkdir(int argc, char **argv);
static void __hw_mknode(int argc, char **argv);
static void __echo(int argc, char **argv);
static void __date(int argc, char **argv);
static void __lsd(int argc, char **argv);
static void __pwd(int argc, char **argv);
static void __help(int argc, char **argv);
static void __cat_test(int argc, char **argv);
static void __ls_test(int argc, char **argv);
static void __setkbrate(int argc, char **argv);
static void __ps(int argc, char **argv);
static void __pcitest(int argc, char **argv);
static void __cd_test(int argc, char **argv);
static void __touch(int argc,char **argv);
static void __fcat(int argc, char **argv);
static void __cp(int argc,char **argv);
static void __mkdir_test(int argc, char **argv);
static void __rm(int argc,char **argv);
static void __rmdir(int argc,char **argv);
static void __log(int argc, char **argv);

typedef struct _cntsh_commands cntsh_commands_t;
struct _cntsh_commands
{
  const char * cmd;
  const char * descr;
  void (*cnt_cmd_clbk)(int argc, char **argv);
};

static const cntsh_commands_t sh_commands[] = {
  {"uname", "get uname", __uname},
  {"whoami", "emm... realy, hwo am I?", __whoami},
  {"lshw", "list hardware", __lshw},
  {"hw_mkdir", "make hw dir", __hw_mkdir},
  {"hw_mknode", "make hw node", __hw_mknode},
  {"echo", "display some text", __echo},
  {"lsd", "Nope, we don't use it, have our own bugs in the head. :)", __lsd},
  {"lspci", "list PCI devices", __lspci},
  {"date", "get date", __date},
  {"pwd", "current dir", __pwd},
  {"cat", "concatenate file content", __cat_test},
  {"ls", "list directory content", __ls_test},
  {"cd", "change current working directory", __cd_test},
  {"pcitest", "pci test", __pcitest},
  {"ps", "process list", __ps},
  {"help", "show this help", __help},
  {"setkbrate", "set keyboard typematic rate and delay", __setkbrate},
  {"fcat", "cat via stream io", __fcat},
  {"touch", "touch in terms of file creation",__touch},
  {"cp", "copying file",__cp},
  {"mkdir", "make directory", __mkdir_test},
  {"rm", "remove file", __rm},
  {"rmdir","remove directory", __rmdir},
  {"log", "log arguments to /tmp/system.log", __log},
  {NULL, NULL, NULL}
};

static void __setkbrate(int argc, char **argv);

void ascii_printf(const char *fmt,... );

#define MAXARGS    16
#define MAX_ARGLEN 255

static char cwdbuf[512];

void builtin(char *in)
{
  char *args[MAXARGS];
  int len=strlen(in);
  int nargs = 1;
  int quoted = 0; //0 = no, 1 = single, 2 = double
  int escaped = 0;
  int next = 0;

  int i = 0;

  args[0] = in;

  while(1) {
    if(!in[i]) break;

    if(next && !isspace(in[i])) {
      args[nargs++] = &in[i];
      if(nargs == MAXARGS) break;
      next = 0;
    }

    switch(in[i]) {
      case '\\':
        if(escaped) escaped = 0;
        else escaped = 1;
      case ' ':
      case '\n':
      case '\t':
        if(!quoted && !escaped){
          in[i] = 0;
          next = 1;
        }
        break;
      case '\'':
        if(quoted == 1 && !escaped) quoted = 0;
        break;
      case '"':
        if(quoted == 2 && !escaped) quoted = 0;
        break;
      default:
        break;
    }
    i++;
  }

  if(len<=0) /* nil */
    return;

  const cntsh_commands_t *cmdptr = sh_commands;
  while (cmdptr->cmd != NULL) {
    if (!strcmp(args[0],cmdptr->cmd)) {
      cmdptr->cnt_cmd_clbk(nargs, args);
      break;
    }
    ++ cmdptr;
  }

  if (!cmdptr->cmd)
    ascii_printf("Illegal command: \"%s\".\n", args[0]);

  return;
}

static void __pwd(int argc, char **argv)
{
  char *pwd=getcwd(cwdbuf,1024);

  ascii_printf("%s\n",pwd);

  return;
}

static void __hw_mknode(int argc, char **argv)
{
  int r;

  if(argc < 2) {
    ascii_printf("No node name specified.\n");
    return;
  }
 
  r=hw_create(argv[1]);

  if(r != 0) {
    ascii_printf("Failed to create node.\n");
  }

  return;
}

static void __hw_mkdir(int argc, char **argv)
{
  int r;

  if(argc < 2) {
    ascii_printf("No directory name specified.\n");
    return;
  }
 
  r=hw_mkdir(argv[1]);

  if(r != 0) {
    ascii_printf("Failed to create directory.\n");
  }


  return;
}

static void __lshw(int argc, char **argv)
{
  char attrs[32];
  int r;
  int fd;
  hw_dirent_t dentry;
  char *dir;

  if(argc < 2) dir = "/";
  else dir = argv[1];
  
  fd=hw_opendir(dir);
  if(fd<0) {
    ascii_printf("{%s} doesn't exist (%d).\n",argv[1],fd);
    return;
  }
  
  while(1) {
    r=hw_readdir(fd,&dentry);
    if(dentry.ntype & HWDIR)
      sprintf(attrs,"<DIR>");
    else if((dentry.ntype & HWNODE) && (dentry.ntype & HWREGISTERED))
      sprintf(attrs,"n--r");
    else if(dentry.ntype & HWNODE)
      sprintf(attrs,"n---");
    if(r==HWEOD) {
      ascii_printf("\\->%s %s\n",dentry.ename,attrs);
      break;
    } else
      ascii_printf("|->%s %s\n",dentry.ename,attrs);
    memset(attrs,'\0',32);
  }

  return;
}

static void __ls_test(int argc, char **argv)
{
  DIR *ldir;
  char name[PATH_MAX];
  char *end;
  struct dirent *ent;
  struct stat stbuf;

  if(argc<2) {
    if (getcwd(name, PATH_MAX) == NULL) {
      ascii_printf("getcwd() failed: %d\n", errno);
      return;
    }
  } else {
    strncpy(name, argv[1], PATH_MAX);
  }
  end = name + strlen(name);

  ldir=opendir(name);
  if(ldir==NULL) {
    ascii_printf("Failed to read directory '%s' (%d)",name,errno);
    return;
  } else {
    while((ent=readdir(ldir))!=NULL) {
      if(!strcmp(ent->d_name,".") || !strcmp(ent->d_name,".."))
	continue;
      *end = '/';
      *(end+1) = 0;
      strcat(name, ent->d_name);
      if (stat(name,&stbuf) == -1) {
        ascii_printf("stat() failed on `%s': %d", ent->d_name, errno);
	break;
      }
      if(S_ISDIR(stbuf.st_mode)) 
        ascii_printf(" \e[31m%s/\e[0m ",ent->d_name);
      else if(S_ISCHR(stbuf.st_mode)) 
        ascii_printf(" \e[33m%s\e[0m ",ent->d_name);
      else if(S_ISBLK(stbuf.st_mode)) 
        ascii_printf(" \e[35m%s\e[0m ",ent->d_name);
      else
        ascii_printf(" (%s %db)",ent->d_name,stbuf.st_size);
    }
    ascii_printf("\n");
  }

  closedir(ldir);

  return;
}

static void __mkdir_test(int argc, char **argv)
{
  int r;

  if(argc<2) {
    ascii_printf("USAGE: mkdir <DIR_NAME>\n");
    return;
  }

  r=mkdir(argv[1],S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
  if(r<0) {
    ascii_printf("Error creation directory %d\n",errno);
    return;
  }

  return;
}

static void __cd_test(int argc, char **argv)
{
  char *__cwd=malloc(512);
  char *__chdir=malloc(256);
  int i;
  char *ee;

  if(argc<2) {
    i=chdir("/");
  } else {
    ee=argv[1];
    if(ee[0]=='/')
      i=chdir(ee);
    else {
      memset(__cwd,'\0',512);
      getcwd(__cwd,500);
      sprintf(__chdir,"%s/%s",__cwd,argv[1]);
      i=chdir(__chdir);
    }
  }

  if(i!=0) 
    ascii_printf("Failed to change directory %d error.\n",errno);
  
  free(__cwd);
  free(__chdir);

  return;
}

static void __cat_test(int argc, char **argv)
{
  int fd;
  char buf[513];
  ssize_t size;

  if(argc < 2 ) {
    ascii_printf("check argument!\n");
    return;
  } else {
    fd=open(argv[1],O_RDONLY);
    if(fd==-1) {
      ascii_printf("Failed to open '%s' with (%d)\n",argv[1],errno);
    } else {
//      memset(buf,'\0',513);
      while((size=read(fd,buf,512)) >= 0) {
        buf[size] = 0;
        ascii_printf("%s",buf);
/*	tsize+=size;
	ascii_printf("size=%d;ridden=%d\n",size,tsize);*/
        //memset(buf,'\0',512);
      }
      close(fd);
    }
  }

  return;
}

static void __whoami(int argc, char **argv)
{
  pid_t mypid=getpid();

  ascii_printf("PID: %d\n",mypid);

  return;
}

static void __uname(int argc, char **argv)
{
  struct utsname un;
  int r;

  r = uname(&un);
  if(r != 0) {
    ascii_printf("uname() failed: %d\n", errno);
    return;
  }
  ascii_printf("%s %s %s %s %s\n", un.sysname, un.nodename, un.version, un.release, un.machine);
  return;
}

static void __lspci(int argc, char **argv)
{
  char *pcibuf;
  int r;
  int fd;

  fd = open("/dev/pci", O_RDONLY);
  if(fd < 0) {
    ascii_printf("Failed to open pci device, error code %d\n", errno);
    return;
  }

  pcibuf = malloc(16384);
  if(!pcibuf) {
    ascii_printf("Not enough memory\n");
    return;
  }

  r = read(fd, pcibuf, 16384);

  if(r < 0) {
    ascii_printf("Failed to read pci device list\n");
  } else {
    ascii_printf("PCI devices:\n%s", pcibuf);
  }

  close(fd);

  free(pcibuf);
}



static void __pcitest(int argc, char **argv)
{
  int r, i, nd, k;
  pci_addr_t dev[20];
  pci_dev_t pdev;

  uint16_t vend = 0x10EC;
  uint16_t prod = 0x8139;

  if(argc != 3) {
    ascii_printf("wrong args, need two hex nums.\n");
    return;
  }

  errno = 0;
  vend = strtoul(argv[1], NULL, 16);
  if(errno) {
    ascii_printf("cannot parse vendor or it is out of range\n");
    return;
  }
  errno = 0;
  prod = strtoul(argv[2], NULL, 16);
  if(errno) {
    ascii_printf("cannot parse product or out it is of range\n");
    return;
  }

  if(r != 0) {
    ascii_printf("Failed to init pci access, errno = %d\n", errno);
    goto out;
  }
  
  ascii_printf("Looking for 0x%x:0x%x...\n", vend, prod);

  nd = pci_lookup_prod_vend(vend, prod, dev, 20);

  if(nd <= 0) {
    ascii_printf("Failed to lookup device 2 (%d)\n", errno);
    goto out;
  }

  ascii_printf("Found %d devices:\n", nd);

  for(k = 0; k < nd; k++) {

    r = pci_get_dev(&dev[k], &pdev);

    ascii_printf("Bus %hhu, device %hhu, function %hhu:\n", dev[k].bus, dev[k].device, dev[k].function);
    
    if(r < 0) {
      ascii_printf("Failed to get device, errno = %d", errno);
      goto out;
    }
    
    for(i = 0; i < pdev.num_resources; i++) {
      ascii_printf("Resource %d (", i);
      if(pdev.resources[i].flags & PCI_RESOURCE_IO) {
        ascii_printf("IO");
      } else if(pdev.resources[i].flags & PCI_RESOURCE_MEM64) {
        ascii_printf("MEM64");
      } else if(pdev.resources[i].flags & PCI_RESOURCE_MEM) {
        ascii_printf("MEM");
      }
      ascii_printf("):\n");
      ascii_printf("  start: %x\n", pdev.resources[i].start);
      ascii_printf("    end: %x\n", pdev.resources[i].end);
      ascii_printf(" length: %d %s\n", pdev.resources[i].end - pdev.resources[i].start + 1,
                   (pdev.resources[i].flags & PCI_RESOURCE_IO) ? "1 byte ports" : "bytes");
    }
    
    ascii_printf("Interrupt line is %d\n", pdev.irq_line);

    ascii_printf("------------------------\n\n");
  }
  
  out:
  pci_close();  
}

static void __echo(int argc, char **argv)
{
  int i;
  int written = 0;
  int fd;
  int tofile = 0;

  char tmp[2048];

  if(argc > 2 && argv[argc-1][0] == '>') {
    fd = open(&argv[argc-1][1], O_WRONLY);
    if(fd < 0) {
      ascii_printf("Cannot open: \"%s\"\n", &argv[argc-1][1]);
    }
    argc--;
    tofile = 1;
  }
  
  for(i = 1; i < argc; i++) {
    int l;
    if(i < argc - 1) l = sprintf(&tmp[written],"%s ", argv[i]);
    else l = sprintf(&tmp[written], "%s", argv[i]);
    written += l;
  }

  sprintf(&tmp[written], "\n");
  written++;
  
  if(!tofile) {
    ascii_printf("%s", tmp);
  } else {
    write(fd, tmp, written);
    close(fd);
  }
}

static void __date(int argc, char **argv)
{
  time_t now;
  int mon, mday, hour, min;
  int year = -1, sec = -1;
  char str[24], *p;
  struct tm tnow;

  static const int mon_days[12] = {
    31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
  };

  if (argc == 1) {
    now = time(NULL);
    if (now == (time_t)-1)
      ascii_printf("time call failed(1): %d\n", errno);
    else if (now == (time_t)-2)
      ascii_printf("time call failed(2): %d\n", errno);
    else
      ascii_printf("%s", ctime(&now));
    return;
  }

  if (argc > 2) {
    ascii_printf("date: extra operand %s\n"
      "Try `date --help' for more information.\n", argv[2]);
    return;
  }

  if (strcmp(argv[1], "--help") == 0) {
    ascii_printf("Usage: date [MMDDhhmm[[CC]YY][.ss]]\n"
      "Display the current date/time, or set the system date.\n\n"
      "       --help     display this help and exit\n");
    return;
  }

  if (strlen(argv[1]) < 8)
    goto error;
  p = argv[1];

  str[0] = *p++;
  str[1] = *p++;
  str[2] = '\0';
  mon = atoi(str);
  if ((errno != 0) || (mon < 1) || (mon > 12))
    goto error;

  str[0] = *p++;
  str[1] = *p++;
  mday = atoi(str);
  if ((errno != 0) || (mday < 1) || (mday > mon_days[mon - 1]))
    goto error;

  str[0] = *p++;
  str[1] = *p++;
  hour = atoi(str);
  if ((errno != 0) || (hour < 0) || (hour > 23))
    goto error;

  str[0] = *p++;
  str[1] = *p++;
  min = atoi(str);
  if ((errno != 0) || (min < 0) || (min > 59))
    goto error;

  if (*p != '\0') {
    if (*p != '.') {
      if ((strlen(p) >= 4) && (p[2] != '.')) {
        /* CCYY */
	str[0] = *p++;
	str[1] = *p++;
	str[2] = *p++;
	str[3] = *p++;
	str[4] = '\0';
	year = atoi(str);
	if ((errno != 0) || (year < 1900) || (year > 2037))
          goto error;
	year -= 1900;
      } else if (strlen(p) >= 2) {
        /* YY */
        str[0] = *p++;
	str[1] = *p++;
	str[2] = '\0';
	year = atoi(str);
	if ((errno != 0) || (year < 0))
          goto error;
	if (year < 70)
          year += 100;
      } else
        goto error;
    }
    if ((*p != '\0') && (*p != '.'))
      goto error;
    if (*p == '.') {
      /* .ss */
      p++;
      if (strlen(p) != 2)
        goto error;
      str[0] = *p++;
      str[1] = *p++;
      str[2] = '\0';
      sec = atoi(str);
      if ((errno != 0) || (sec < 0) || (sec > 59))
        goto error;
    }
  }

  now = time(NULL);
  if (now == (time_t)-1) {
    ascii_printf("time call failed: %d\n", errno);
    return;
  }
  tnow = *localtime(&now);

  tnow.tm_mon = mon - 1;
  tnow.tm_mday = mday;
  tnow.tm_hour = hour;
  tnow.tm_min = min;
  if (year != -1)
    tnow.tm_year = year;
  if (sec != -1)
    tnow.tm_sec = sec;
  now = mktime(&tnow);

  if (settime(now) != 0)
    ascii_printf("settime call failed: %d\n", errno);

  return;

error:
  ascii_printf("date: invalid date `%s'\n", argv[1]);
}

void clrscr(void);
void paint_border(void);

static void __lsd(int argc, char **argv)
{
  clrscr();
//  paint_border();
  ascii_printf("\e[20;20Haaaa\e[30;30Hbbbb");
}

static void __help(int argc, char **argv)
{
  const cntsh_commands_t *cmdptr = sh_commands;
  while (cmdptr->cmd != NULL) {
    ascii_printf("%s - %s\n", cmdptr->cmd, cmdptr->descr);
    ++ cmdptr;
  }
}

static void __setkbrate(int argc, char **argv)
{
  int r;
  uni_rpc_t rpc;
  static int opened = 0;
  static hw_direct_io_t kbd;

  ulong_t rt, d;

  if(argc != 3) {
    ascii_printf("usage: setkbrate <rate, 0-31> <delay, 0-3>\nLess means faster.\n");
    return;
  }

  errno = 0;
  rt = strtoul(argv[1], NULL, 10);
  if(errno || rt > 31) {
    ascii_printf("cannot parse rate or it is out of range\n");
    return;
  }
  errno = 0;
  d = strtoul(argv[2], NULL, 10);
  if(errno || r > 3) {
    ascii_printf("cannot parse delay or out it is of range\n");
    return;
  }

  
  if(!opened) {
    r = hw_dio_open("/kbd", HWRDWR, &kbd);
    if(r < 0) {
      ascii_printf("Failed to open kbd device, error code %d\n", r);
      return;
    }
    opened = 1;
  }

  memset(&rpc, 0, sizeof(rpc));
  rpc.arg3 = SET_KB_TYPEMATIC;
  rpc.arg4 = TM_RATE(rt) | TM_DELAY(d);

  r = hw_dio_ioctl(&kbd, &rpc, NULL, 0, 0);
  
  if(r < 0) {
    ascii_printf("ioctl failed, error code is %d\n", r);
  }
}

static void __ps(int argc, char **argv)
{
  int i,r;

  ascii_printf("   PID  CPU  VSize \n");
  ascii_printf("---------------------\n");

  for(i=1;i<=30;i++) {
    r=task_scheduler_control(i,0,SYS_SCHED_CTL_GET_CPU,0);
    if( r >= 0 ) {
      ascii_printf("   %-2d    %d    N/A\n",i,r);
    }
  }
}

static void __fcat(int argc, char **argv)
{
  FILE *fd;
  char buf[128];

  if (argc < 2) {
    ascii_printf("USAGE: fcat <filename>\n");
    return;
  }

  fd = fopen(argv[1], "r");
  if (fd == NULL) {
    ascii_printf("fopen() failed: %d\n", errno);
    return;
  }

  memset(buf, 0, 128);
  while (fgets(buf, 128, fd) != NULL) {
    ascii_printf("%s", buf);
    memset(buf, 0, 128);
  }

  fclose(fd);
  return;
}

static void __touch(int argc,char **argv)
{
  int fd;

  if(argc<2) {
    ascii_printf("USAGE: touch <filename>\n");
    return;
  } 

  fd=creat(argv[1],(S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));
  if(fd<0) {
    ascii_printf("Failed with %d\n",errno);
    return;
  } else {
    close(fd);
  }

  return;
}

static void __rm(int argc,char **argv)
{
  int r;

  if(argc<2) {
    ascii_printf("USAGE: rm <filename>\n");
    return;
  } 

  r=unlink(argv[1]);
  if(r<0)
    ascii_printf("Error removing file with %d\n",errno);

  return;
}

static void __rmdir(int argc,char **argv)
{
  int r;

  if(argc<2) {
    ascii_printf("USAGE: rmdir <directory name>\n");
    return;
  } 

  r=rmdir(argv[1]);
  if(r<0)
    ascii_printf("Error removing directory with %d\n",errno);

  return;
}

static void __cp(int argc,char **argv)
{
  int fd_r,fd_w;
  char *rbuf=malloc(2048);
  ssize_t size,tsize=0;

  if(argc<3) {
    ascii_printf("USAGE: cp <source> <destination>\n");
    return;
  } 


  fd_r=open(argv[1],O_RDONLY);
  if(fd_r<0) {
    ascii_printf("Read open failed with %d\n",errno);
    return;
  }
  fd_w=creat(argv[2],(S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));
  if(fd_w<0) {
    ascii_printf("Write open failed with %d\n",errno);
    close(fd_r);
    return;
  }

    ascii_printf("fd_r=%d,fd_w=%d\n",fd_r,fd_w);

  while((size=read(fd_r,rbuf,2048))!=0) {
    size=write(fd_w,rbuf,size);
    if(size<0) {
      ascii_printf("error %d!\n",errno);
      break;
    } else {
      tsize+=size;
      ascii_printf("written %d bytes\n",tsize);
    }
  }

  close(fd_r);
  close(fd_w);

  return;
}

static void __log(int argc, char **argv)
{
  FILE *fd;
  int i;

  fd = fopen("/tmp/system.log", "a");
  if (fd == NULL) {
    ascii_printf("fopen() failed: %d\n", errno);
    return;
  }

  for (i = 1; i < argc; i++) {
    if (fprintf(fd, "%s", argv[i]) < 0) {
      ascii_printf("fprintf() failed: %d\n", errno);
      fclose(fd);
      return;
    }
    if (i != argc - 1)
      fprintf(fd, " ");
  }
  fprintf(fd, "\n");

  fclose(fd);
}
