#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ipc/rpc.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <wchar.h>
#include <errno.h>
#include <ctype.h>
#include <pthread.h>
#include <sys/uevent.h>
#include <sys/scheduler.h>
#include <sys/proc.h>
#include <sys/thread.h>
#include <sys/uevent.h>
#include <sys/syscall.h>
#include <alloc/imcache.h>
#include <sys/ptd.h>
#include <dircons.h>
#include <datastructs.h>
#include <sys/keyboard.h>
#include <datalib/circular.h>
#include <termios.h>

#include "builtin.h"

#define TBUFLEN 32768

#define TERM_SLOT 6
#define BUFF_SLOT 5

#define STARTULEN 7
#define MAXULEN 200
#define WIDTH 80
#define HEIGHT 24

#define START_DELAY 600000;

int delay;
int ulen;
bool paused;

circular_t keypresses;

int xs[MAXULEN];
int ys[MAXULEN];

int field[WIDTH][HEIGHT];

pthread_mutex_t umutex;
event_t uevent;


int vx, vy, lvx, lvy;

int term;

status_t open_term(ulong_t termno);
void ascii_printf(const char *fmt,... );

static void printxy(int x, int y, char c)
{
  ascii_printf("\e[%d;%dH%c", y+1, x+1, c);
}

void clrscr(void)
{
  ascii_printf("\e[2J");
}

int set_raw(void)
{
  int r;
  struct termios t;
  uint8_t kbmode = 0;
  
  int termfd = (int)PTD()->custom_slots[TERM_SLOT];

  r = ioctl(termfd, TIOC_GETATTR, &t);

  if(r < 0) {
    ascii_printf("cannot get term attrs %d %d\n", r, errno);
    return -1;
  }

  t.c_lflag = 0;

  r = ioctl(termfd, TIOC_SETATTR, &t);

  if(r < 0) {
    ascii_printf("cannot set term attrs %d %d\n", r, errno);
    return -1;
  }
  
  r = ioctl(termfd, TIOC_SETKBMODE, &kbmode);

  if(r < 0) {
    ascii_printf("cannot set kbmode %d %d\n", r, errno);
    return -1;
  }

  return 0;
}

char get_char(void)
{
  char c;
  
  read(term, &c, 1);
  
  return c;
}

int is_killed(int nx, int ny)
{
  int i;
  if(nx >= WIDTH-1) return 1;
  if(nx < 1) return 1;
  if(ny >= HEIGHT-1) return 1;
  if(ny < 1) return 1;
  for(i = 0; i < ulen; i++) {
    if(xs[i] == nx && ys[i] == ny) return 1;
  }
  return 0;
}

uint16_t random(uint16_t y)
{
  static ulong_t a = 3, b = 11, c = 17;
  uint16_t x = (1 + a + 23 * b + 17 * c);
  
  a = b;
  b = c;
  c = x;
  
  return x % y;
}

void put_food(void)
{
  int i;
  int flag;
  uint16_t fx, fy;
  while(1) {
    fx = 1 + random(WIDTH-2);
    fy = 1 + random(HEIGHT-2);
    flag = 0;
    for(i = 0; i < ulen; i++) {
      if(fx == xs[i] || fy == ys[i] || field[fx][fy] != 0) flag = 1;
    }
    if(!flag) break;
  }
  printxy(fx, fy, 'r');
  field[fx][fy] = 1;
}
                                                
void paint_border(void)
{
  char tmp[WIDTH+1];
  int i;

  memset(tmp, '-', sizeof(tmp));
  tmp[0] = '+';
  tmp[WIDTH-1] = '+';
  tmp[WIDTH] = 0;
    
  ascii_printf("\e[%d;1H%s\e[1;1H%s", HEIGHT, tmp, tmp);
  for(i = 2; i < HEIGHT; i++) {
    ascii_printf("\e[%d;1H|\e[%d;%dH|", i, i, WIDTH);
  }
}

void print_status(void)
{
  ascii_printf("\e[%d;5HLength: %d", HEIGHT + 1, ulen);
  if(paused) ascii_printf("\e[%d;35H[PAUSED, press SPACE]", HEIGHT + 1);
  else ascii_printf("\e[%d;35H                                            ", HEIGHT + 1);
}

void udav_init(void)
{
  int i;

  clrscr();
  paint_border();

  delay = START_DELAY;

  circular_flush(&keypresses);
  
  vx = 1;
  vy = 0;

  paused = true;

  ulen = STARTULEN;

  memset(field, 0, sizeof(field));
  print_status();

  for(i = 0; i < STARTULEN; i++) {
    xs[i] = STARTULEN-i+4;
    ys[i] = HEIGHT/2;
    printxy(xs[i], ys[i], 'o');
  }

  put_food();
  put_food();
  put_food();
  put_food();
  put_food();
  put_food();
  put_food();
}

void udav_move(void)
{
  int nx, ny, i;
  int enl = 0;
  char code;
  int got = 0;
  
  while(got == 0 && circular_getc(&keypresses, &code)){
    switch(code) {
      case KC_LEFT:
        if(vx != 1) {
          vx = -1;
          vy = 0;
          got = 1;
        }
        break;
      case KC_DOWN:
        if(vy != -1) {
          vy = 1;
          vx = 0;
          got = 1;
        }
        break;
      case KC_RIGHT:
        if(vx != -1) {
          vx = 1;
          vy = 0;
          got = 1;
        }
        break;
      case KC_UP:
        if(vy != 1) {
          vy = -1;
          vx = 0;
          got = 1;
        }
        break;
      case KC_SPACE:
      case ' ':
        paused = !paused;
        print_status();
        got = 1;
        break;
    }
  }

  if(paused) return;
  
  nx = xs[0] + vx;
  ny = ys[0] + vy;

  if(is_killed(nx, ny)){
    udav_init();
    return;
  }

  pthread_mutex_lock(&umutex);

  lvx = vx;
  lvy = vy;
  
  if(field[nx][ny]) {
    field[nx][ny] = 0;
    put_food();
    xs[ulen] = xs[ulen-1];
    ys[ulen] = ys[ulen-1];
    enl = 1;
    delay -= delay/50;
    print_status();
  }

  printxy(nx, ny, 'o');
  if(!enl)printxy(xs[ulen-1], ys[ulen-1], ' ');
  
  for(i = ulen-1; i > 0; i--) {
    xs[i] = xs[i-1];
    ys[i] = ys[i-1];
  }

  if(enl) ulen++;

  xs[0] = nx;
  ys[0] = ny;
  pthread_mutex_unlock(&umutex);
}

void *control(void *arg)
{
  char c;

  while(1) {
    c = get_char();
    circular_putc(&keypresses, c);
  }
}

void *udav(void *arg)
{
  pthread_t ct;
  
  if(open_term(9) != 0) abort();
  term = (int)PTD()->custom_slots[TERM_SLOT];
  if(set_raw() != 0) for(;;) sleep(1000);

  pthread_mutex_init(&umutex, NULL);
  event_init(&uevent);

  if(circular_init(&keypresses, 64) != 0) return NULL;

  ascii_printf("\e[?25l");
  
  udav_init();

  if(pthread_create(&ct, NULL, control, NULL) != 0) {
    ascii_printf("Failed to create thread.\n");
    abort();
  }
  
  while(1) {
    usleep(delay);
    udav_move();
  }
}
