/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Copyright (C) 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2008 Paul Graphov <graphov@jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 * Copyright (C) 2009 Alfeiks Kaanoken <madtirra@jarios.org> (*_GRUB opt added)
 *
 * Real time clock driver
 *
 */

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include <ipc/rpc.h>
#include <devio/devio.h>
#include <sys/ioport.h>
#include <chardev.h>
#include <dircons.h>

#include "rtclock.h"

#define DRVNAME "rtc.d"
#define DEVNAME "/dev/rtc"

#define BCD2BIN(x) ((x) & 0x0f) + ((x) >> 4) * 10
#define BIN2BCD(x) (((x) / 10) << 4) + (x) % 10

static uint8_t rtc_read_register(uint8_t reg)
{
  outb(RTC_ADDRESS, reg);
  return inb(RTC_DATA);
}

static void rtc_write_register(uint8_t reg, uint8_t val)
{
  outb(RTC_ADDRESS, reg);
  outb(RTC_DATA, val);
}

static int get_rtc_time(struct tm *time)
{
  if (alloc_ioport_range(RTC_ADDRESS, 2))
    return 0;

  /* Overflow of seconds or minutes may happen,
   * so we check it to return consistent values.
   */
  for (;;) {
    time->tm_sec = rtc_read_register(RTC_SECONDS);
    time->tm_min = rtc_read_register(RTC_MINUTES);
    time->tm_hour = rtc_read_register(RTC_HOURS);

    if ((time->tm_min == rtc_read_register(RTC_MINUTES)) &&
        (time->tm_hour == rtc_read_register(RTC_HOURS)))
      break;
  }
  time->tm_mday = rtc_read_register(RTC_DAY_OF_MONTH);
  time->tm_mon = rtc_read_register(RTC_MONTH);
  time->tm_year = rtc_read_register(RTC_YEAR);

  free_ioport_range(RTC_ADDRESS, 2);

  time->tm_sec = BCD2BIN(time->tm_sec);
  time->tm_min = BCD2BIN(time->tm_min);
  time->tm_hour = BCD2BIN(time->tm_hour);
  time->tm_mday = BCD2BIN(time->tm_mday);
  time->tm_mon = BCD2BIN(time->tm_mon);
  time->tm_year = BCD2BIN(time->tm_year);
  time->tm_wday = BCD2BIN(time->tm_wday);

  if (time->tm_year <= 69)
    time->tm_year += 100;
  time->tm_mon--;

  time->tm_wday = 0;
  time->tm_yday = 0;
  time->tm_isdst = 0;

  return 1;
}

static int set_rtc_time(const struct tm *time)
{
  unsigned int year;
  unsigned char mon, mday, hour, min, sec;

  year = time->tm_year;
  mon = time->tm_mon + 1;
  mday = time->tm_mday;
  hour = time->tm_hour;
  min = time->tm_min;
  sec = time->tm_sec;

  if (year > 169)
    return 0;
  if (year >= 100)
    year -= 100;

  sec = BIN2BCD(sec);
  min = BIN2BCD(min);
  hour = BIN2BCD(hour);
  mday = BIN2BCD(mday);
  mon = BIN2BCD(mon);
  year = BIN2BCD(year);

  if (alloc_ioport_range(RTC_ADDRESS, 2))
    return 0;

  rtc_write_register(RTC_YEAR, year);
  rtc_write_register(RTC_MONTH, mon);
  rtc_write_register(RTC_DAY_OF_MONTH, mday);
  rtc_write_register(RTC_HOURS, hour);
  rtc_write_register(RTC_MINUTES, min);
  rtc_write_register(RTC_SECONDS, sec);

  free_ioport_range(RTC_ADDRESS, 2);

  return 1;
}

static status_t rtc_read(int devno, void *buf, size_t size, size_t offset, void **interim)
{
  struct tm time;
  int len;
  int am;
  char str[32];

  if(!get_rtc_time(&time)) {
    return -EIO;
  }

  len = snprintf(str, size, "%lu", mktime(&time)) + 1;

  if(offset >= len) {
    return 0;
  }

  am = MIN(size, len - offset);
  
  memcpy(buf, str + offset, am);
  
  return am;
}

static status_t rtc_write(int devno, void *buf, size_t size, size_t offset, void **interim)
{
  time_t time;
  struct tm *local;

  time = strtoul(buf, NULL, 10);
  if((time == 0) && (errno != 0))
    return -errno;

  local = localtime(&time);
  return (set_rtc_time(local) ? 0 : -EIO);
}

static struct device_ops rtc_ops = {
  .read  = rtc_read,
  .write = rtc_write,
  .ioctl = NULL,
  .log = NULL
};

int main(int argc, char **argv)
{
  char *mesg;
  int r;

#ifndef CONFIG_RTC_GRUB
  start_service();
#endif

  if((mesg=malloc(1024))==NULL) return -ENOMEM;

  open_console();

  r = init_simple_device_server(512, 512, DRVNAME, DEVNAME, &rtc_ops);
  if (r != 0) {
    console_printf(mesg, "rtc.d: failed to initialize server (%d)\n", r);
    abort();
  }

  console_output("[rtc.d]\n");
  device_server_start();

  return 0;
}
