/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * ATA block level device driver
 */

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <ata/ata.h>
#include <ata/libata.h>
#include <block/blkdev.h>
#include <block/bdev_drv.h>
#include <block/iosched.h>
#include <block/sigaction.h>
#include <ata/host_drv.h>
#include <debug.h>
#include <errno.h>
#include <err.h>
#include <timeout.h>
#include <limits.h>
#include "ata_blk.h"
#include "drv.h"

#define NR_ATA_TRANSPORT_DRIVERS (sizeof(ata_transport_drivers) / sizeof(ata_driver_t*))

static blkdev_operations_t ata_bdev_ops;

#ifdef CONFIG_BDEV_STATLINK

extern ata_host_driver_t *piix_init_ide_driver(void);
extern ata_host_driver_t *amd74_init_ide_driver(void);

ata_host_driver_t *init_host_drv_fn(void)
{
  if (!strcmp(CONFIG_IDE_HOST_DRIVER, "piix"))
    return piix_init_ide_driver();
  else if (!strcmp(CONFIG_IDE_HOST_DRIVER, "amd74_nv_nforce"))
    return amd74_init_ide_driver();
  else
    return NULL;
}

#else

ata_host_driver_t *init_host_drv_fn(void)
{
  return NULL;
}

#endif /* CONFIG_BDEV_STATLINK */

static int ata_scan_ifaces(void *host_lib_handle);
static int ata_init_iface(void *host_lib_handle, struct bd_iface*, const char *arg_line);
static void ata_shutdown_iface(struct bd_iface*);
static int ata_probe_device(block_device_t *bdev);
static int ata_prepare_rq_proc(blkreq_queue_t *q);
static int ata_suspend_device(block_device_t *bdev);
static int ata_resume_device(block_device_t *bdev);
static void ata_shutdown_device(block_device_t*);
static bool ata_host_probe_port(ata_port_t *port, int num);

static blkdev_driver_t generic_ata_driver = {
  .name = "ata",
  .version = "0.0.1",
  .author = "JariOS core team",
  .iface_type = BDEV_IFACE_ATA,

  .scan_ifaces = ata_scan_ifaces,
  .init_iface = ata_init_iface,
  .shutdown_iface = ata_shutdown_iface,
  .probe = ata_probe_device,
  .prepare_rq_proc = ata_prepare_rq_proc,
  .suspend_device = ata_suspend_device,
  .resume_device = ata_resume_device,
  .shutdown_device = ata_shutdown_device
};

static ata_driver_t *ata_transport_drivers[] = {
#ifdef CONFIG_IDE_DISK
  &idedisk_driver,
#endif
#ifdef CONFIG_IDE_CDROM
  &idecd_driver
#endif
};

static ata_device_t *choose_best_ata_device(ata_port_t *port)
{
  return NULL;
}

static ata_driver_t *lookup_driver(ata_device_t *dev)
{
  ata_driver_t *drv = NULL;
  ata_host_t *host = dev->host_port->host;
  int i;

  for (i = 0; i < NR_ATA_TRANSPORT_DRIVERS; i++) {
    if (ata_transport_drivers[i]->match(host, dev)) {
      drv = ata_transport_drivers[i];
      break;
    }
  }

  return drv;
}

/* [dg]: now, temporary static linking with a PCI host driver is used */
static int ata_scan_ifaces(void *host_lib_handle)
{
  ata_host_driver_t *drv;

  drv = init_host_drv_fn();
  ASSERT(drv != NULL, "ATA host driver is not choosed or choosed incorrectly!\n");
  if (drv->phys_iface != ATA_PHYIF_IDE &&
      drv->phys_iface != ATA_PHYIF_SATA) {
    return -EINVAL;
  }

  return drv->scan_hosts();
}

static int add_block_device(struct bd_iface *iface, ata_device_t *ata_dev)
{
  block_device_t *bdev;
  ata_port_t *port = ata_dev->host_port;

  bdev = malloc(sizeof(block_device_t));
  if (bdev == NULL)
    return -ENOMEM;

  memset(bdev, 0, sizeof(block_device_t));
  bdev->media.capacity = ata_dev->param_info.total_addr_sects;
  bdev->media.sect_shift = ata_dev->param_info.sect_shift;
  bdev->num = (iface->max_devs / port->host->nr_ports) * port->num + ata_dev->id;
  bdev->private = ata_dev;
  bdev->iface = iface;
  bdev->ops = &ata_bdev_ops;
  if (device_is_atapi(ata_dev)) {
    if (ata_drv_support_media(ata_dev->drv, ATA_MEDIA_DVD))
      bdev->type = BDEV_DVDROM;
    else if (ata_drv_support_media(ata_dev->drv, ATA_MEDIA_CD))
      bdev->type = BDEV_CDROM;
    else
      bdev->type = BDEV_FLOPPY;
  } else {
    bdev->type = BDEV_HDD;
  }
  if (ata_dev->param_info.removable)
    bdev->removable = true;

  ata_dev->bdev = bdev;
  list_add2tail(&iface->devices, &bdev->node);

  return 0;
}

static int ata_init_iface(void *host_lib_handle, struct bd_iface *iface,
                          const char *arg_line)
{
  ata_host_t *host;
  int ret;
  ata_host_driver_t *drv;
  ata_port_t *port;
  ata_device_t *ata_dev;
  ata_driver_t *ata_drv;
  int i, j, n;
  uint64_t size;
  const char *size_pref[] = {"bytes", "KiB", "MiB", "GiB", "TiB"};
  char model_num[41], s[32];

  model_num[41] = '\0';
  drv = init_host_drv_fn();
  host = malloc(sizeof(ata_host_t));
  if (host == NULL)
    return -ENOMEM;

  /* Fixme [dg]: replace later by absolute bus ID */
  iface->bus = drv->phys_iface;
  memset(host, 0, sizeof(ata_host_t));
  host->drv = drv;
  ret = drv->init_host(host, iface->id);
  if (ret) {
    free(host);
    /* TO DO [dg]: print also iface type and bus id */
    printf("[ata drv]: Host initialization failed\n");
  } else {
    iface->bus = drv->phys_iface;
    i = (drv->phys_iface == ATA_PHYIF_IDE) ? 2 : 1;  /* supprorted devices per one port */
    iface->max_devs = host->nr_ports * i;
  }

  if (host->port_ops->probe == NULL) {
    /* detecting by reading device id */
    host->port_ops->probe = ata_host_probe_port;
  }

  if (iface->name)
    printf("[ata_drv]: probing devices on %s interface\n", iface->name);

  ret = ata_probe_init_ports(host, NULL);
  if (ret)
    return ret;

  for (i = 0; (i < host->nr_ports) && !ret; i++) {
    port = host->ports[i];
    for (j = 0; j < 2; j++) {
      ata_dev = port->devices[j];
      if (!ata_dev)
        continue;

      ata_drv = lookup_driver(ata_dev);
      if (ata_drv == NULL) {
        /* TO DO [dg]: output warning to log */
        continue;
      }
      ata_dev->drv = ata_drv;

      if (ata_drv->init(ata_dev)) {
        /* TO DO [dg]: output message to log */
        continue;
      }
      if (add_block_device(iface , ata_dev)) {
        /* TO DO [dg]: output message to log and put the device into sleep */
      }
      if (iface->name && ata_dev->param_info.model_num[0]) {
        s[0] = '\0';
        if (!ata_dev->media.removable) {
          size = (uint64_t)ata_dev->param_info.total_addr_sects <<
                           ata_dev->param_info.sect_shift;
          n = -1;
          while (size > 1024) {
            size >>= 10;
            n++;
          }
          sprintf(s, "of %d %s capacity ", size, size_pref[n + 1]);
        }
        ata_normalize_id_str(model_num, ata_dev->param_info.model_num, 40);

        printf("%s block device %sis detected on port %d\n",
               model_num, s, port->num);
      }
    }
  }

  return ret;
}

static void ata_shutdown_iface(struct bd_iface *iface)
{

}

static int ata_probe_device(block_device_t *bdev)
{
  return 0;
}

static int ata_prepare_rq_proc(blkreq_queue_t *q)
{
  int ret = 0;
  ata_device_t *dev;
  ata_host_t *host;

  q->bdev->queue = q;
  dev = q->bdev->private;
  host = dev->host_port->host;
  if (q->group->private == NULL) {  /* need resources are not allocated and initialized? */
    ret = start_rq_proc_threads(q->group, 1);
    if (ret)
      goto out;

    ret = init_ata_timeouts(q->group);
    if (ret)
      goto out_threads;

    ret = init_ata_interrupts(host, NULL);
    if (ret)
      goto out_timeouts;

    ret = ata_set_blk_iosched(q);
    if (ret)
      goto out_timeouts;
  } else {
    ret = ata_set_blk_iosched(q);
    if (ret)
      goto out;
  }

  return 0;

out_timeouts:
  destroy_ata_timeouts(q->group);
out_threads:
  stop_rq_proc_threads(q->group);
out:
  printf("[ata drv]: Out of resources for block request processing\n");

  return ret;
}

static int ata_suspend_device(block_device_t *bdev)
{
  return 0;
}

static int ata_resume_device(block_device_t *bdev)
{
  return 0;
}

static void ata_shutdown_device(block_device_t *bdev)
{

}

static bool ata_host_probe_port(ata_port_t *port, int num)
{
  ata_device_t *dev;
  bool ret = false;
  int i;
  ata_host_driver_t *drv = port->host->drv;

  dev = malloc(sizeof(ata_device_t));
  if (dev == NULL) {
    printf("[ata.d]: Out of memory at ports scanning\n");
    return false;
  }

  memset(dev, 0, sizeof(ata_device_t));
  dev->host_port = port;
  dev->id = num;
  dev->io_prot = ATA_PROT_PIO;
  dev->ctl_cmd |= DEVCTL_NIEN;  /* disable intrq asserting */

  /*
   * attach temporary the device to the first found driver
   * supporting the physical interface of the host
   */
  for (i = 0; i < NR_ATA_TRANSPORT_DRIVERS; i++) {
    if (ata_transport_drivers[i]->phys_ifaces & (1 << drv->phys_iface)) {
      dev->drv = ata_transport_drivers[i];
      break;
    }
  }

  if (i == NR_ATA_TRANSPORT_DRIVERS) {
    /* TO DO [dg]: output error */
    free(dev);
    return false;
  }

  if (!ata_get_param_info(dev)) {
    ret = true;
    port->devices[num] = dev;
  } else {
    free(dev);
  }

  return ret;
}

blkdev_driver_t *init_ata_driver(void)
{
  return &generic_ata_driver;
}
