/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * Interrupt initialization and processing
 */

#include <ata/libata.h>
#include <pthread.h>
#include <debug.h>
#include <sys/scheduler.h>
#include "ata_blk.h"
#include "drv.h"

enum {
  INTR_THREAD_PRIORITY = 30
};

struct irq_data {
  ata_host_t *host;
  irq_handler_t handler;
  struct completion compl;
};

static void *irq_handler_thread_fn(void *data)
{
  struct irq_data *idata = (struct irq_data*)data;
  ata_host_t *host = idata->host;
  int r = 0, i, n = 0;
  irq_t irqs[MAX_ATA_PORTS];

  if (!host->irq_shared) {
    for (i = 0; i < host->nr_ports; i++) {
      if (host->ports[i])
        irqs[n++] = host->ports[i]->irq;
    }
  } else {
    irqs[n++] = host->ports[0]->irq;
  }

  r = interrupt_event_create(&host->ievent, irqs, n);
  for (i = 0; (i < n) && !r; i++) {
    r = interrupt_event_add_handler(host->ievent, irqs[i], idata->handler,
                                    host->ports[i]);
  }

  complete(&idata->compl, r);
  if (r) {
    interrupt_event_destroy(host->ievent);
    host->ievent = NULL;
    return NULL;
  }

  task_scheduler_control(0, 0, SYS_SCHED_CTL_SET_PRIORITY, INTR_THREAD_PRIORITY);
  while (1) {
    r = interrupt_event_wait(host->ievent);
    ASSERT(!r, "Interrupt processing error!\n");
  }

  return NULL;
}

int init_ata_interrupts(ata_host_t *host, irq_handler_t handler)
{
  int ret;
  ata_port_t *port;
  struct irq_data idata;
  int i;

  if (!ata_host_support_irq(host))
    return 0; /* interrupts are not supported */

  if (host->ievent)
    return -EBUSY;

  idata.host = host;
  idata.handler = handler ? handler : ata_default_irq_handler;
  ret = init_completion(&idata.compl);
  if (ret)
    return ret;

  ret = -pthread_create(&host->irq_thread, NULL, irq_handler_thread_fn, &idata);
  if (ret)
    goto out;

  wait_for_completion(&idata.compl);
  if (idata.compl.result) {
    pthread_join(host->irq_thread, NULL);
    host->irq_thread.__tid = -1;
    goto out;
  }

  /*
   * setup a respective ctl command enabling INTRQ asserting
   * for all devices driven by the host
   */
  for (i = 0; i < host->nr_ports; i++) {
    port = host->ports[i];
    if (!port)
      continue;

    if (port->devices[0])
      port->devices[0]->ctl_cmd &= ~DEVCTL_NIEN;
    if (port->devices[1])
      port->devices[1]->ctl_cmd &= ~DEVCTL_NIEN;
  }

out:
  destroy_completion(&idata.compl);

  return ret;
}

void release_ata_interrupts(ata_host_t *host)
{

}

/*
 * Parallel ata specific irq handler
 *
 * It forces a respetive cmd phase.
 * In case of error it tries to restart the running command if the
 * error is not caused by a host fail, and the number of retries is
 * not exhausted.  In case of successful data transfer the handler
 * does not check, whether all data was really transferred, this
 * must do a specific device driver.
 */
void pata_irq_handler(ata_device_t *dev, port_status_t port_stat)
{
  ata_status_t dev_stat;
  ata_taskfile_t tf;
  bool issue_cmd = false, done = false;
  ata_cmd_t *cmd = dev->run_cmd;
  uint32_t e;
  blk_request_t *rq;

  if (cmd == NULL)  /* command could be canceled concurrently */
    return;

  tf.flags = cmd->tf.flags;
  dev_stat.val = ata_read_device_status(dev);
  if (dev_stat.err) {
    cmd->err = ATA_ERROR_DEVICE;
  } else if (port_stat & PORT_STAT_ERROR) {
    cmd->err = ATA_ERROR_HOST;
  } else {
    if (dev->cmd_phase == PHASE_WAIT_READY)
      dev->cmd_phase = PHASE_INTRQ_ASSERTED;

    if (ata_dma_protocol(cmd->protocol) && dev->cmd_phase == PHASE_TRANSFER) {
      done = true;
    } else if (cmd->protocol == ATA_PROT_NODATA ||
               (cmd->rq && cmd->rq->complete_sectors == cmd->rq->nr_sectors)) {
      done = true;
    }
  }

  while (1) {
    if (cmd->err) {
      e = (cmd->err == ATA_ERROR_HOST) ? port_stat : dev_stat.val;
      dev->drv->error_handler(dev, cmd->err, e);
      if (try_prepare_cmd_restart(dev, cmd))
        issue_cmd = true;
    }

    if (done || cmd->err) {
      if (done && cmd->read_tf_end)
        ata_read_taskfile(dev, &cmd->tf);

      ata_done_cmd(dev, cmd);
      ata_free_cmd(cmd);
      cmd = NULL;
    }

    if (!issue_cmd && !ata_cmd_queue_empty(dev) && dev_stat.serv) {
      /* device is ready to process another queued command */
      cmd = find_ready_queued_cmd(dev);
    }

    if (cmd == NULL) {
      if (!ata_cmd_queue_empty(dev)) {
        dev->run_cmd = container_of(list_node_last(&dev->cmd_queue.cmds),
                                    ata_cmd_t, node);
      }
      break;
    }

    if (issue_cmd)
      ata_start_command(dev, cmd);
    else
      ata_start_transfer(dev, cmd);

    if (!cmd->err) {
      rq = cmd->rq;
      if (rq && !blkreq_rw(rq) && (rq->complete_sectors == rq->nr_sectors))
        done = true;
      else
        break;
    }
  }
}
