/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * Processing requests from the block layer
 */

#include <block/request.h>
#include <block/sigaction.h>
#include <block/iosched.h>
#include <ata/ata.h>
#include <ata/libata.h>
#include <pthread.h>
#include <stdlib.h>
#include <sys/scheduler.h>
#include <string.h>
#include <datalib/list.h>
#include <errno.h>
#include <debug.h>
#include "ata_blk.h"
#include "drv.h"

enum {
  INTRQ_TIMEOUT_SIGNO = SIGRTMIN,
  DMA_TIMEOUT_SIGNO = SIGRTMIN + 1
};

enum {
  PROC_THREAD_PRIORITY = 60,
  TM_THREAD_PRIORITY = 59
};

enum {
  ATA_BLKQUEUE_MULTITHREAD = BLKQUEUE_NOT_PIPELINED + 1
};

struct ata_threads {
  pthread_t *proc_th;  /* processing threads */
  int nr_threads;
  pthread_t tm_th;
};

static void timeout_sigaction(int signo, siginfo_t *info, void *context)
{

}

#ifdef CONFIG_IOSCHED

int ata_set_blk_iosched(blkreq_queue_t *q)
{
  int ret;
  struct ata_threads *threads = q->group->private;
  bool block;
  pthread_t th;

  ret = set_blk_default_iosched(q);
  if (ret)
    return ret;

  if (threads->nr_threads > 1) {
    block = true;
    th = threads->tm_th;
  } else {
    th = threads->proc_th[0];
  }
  ret = init_timeout_tmr(&q->scheduler->tmr, th, IOSCHED_SIGNO, q, block);
  if (ret) {
    disable_blk_iosched(q);
    ret = -errno;
  }

  return ret;
}

#endif

static void default_cmd_done(ata_cmd_t *cmd)
{
  ata_device_t *dev = cmd->dev;
  blk_request_t *rq = cmd->rq;
  blkio_t *b;

  if (!cmd->err) {
    if (dev->drv->cmd_done)
      dev->drv->cmd_done(cmd);

    rq->complete(rq);
    put_blk_request(rq);
  } else {
    dev->nr_failures++;
    if (!blk_drv_request(rq) && rq->type == REQ_TYPE_IO) {
      if (rq->current_blkio)
        b = rq->current_blkio;
      else
        b = container_of(list_node_first(&rq->blkios), blkio_t, node);

      b->status.err = EIO;
      rq->current_blkio = b;
      printf("[ata drv]: EIO error, %ld sector\n",
             rq->start_sector + rq->complete_sectors);
    }
  }
}

static void *timeout_thread_fn(void *data)
{
  /* yet stub */
  sleep(1000000);

  return NULL;
}

static void *proc_rq_thread_fn(void *data)
{
  struct blk_group *group = (struct blk_group*)data;
  blk_request_t *rq;
  blkreq_queue_t *q;
  ata_device_t *dev;
  ata_port_t *port;
  ata_host_t *host;
  ulong_t qmask = ~0L, prev_wports = 0, m, m1;
  int i, j;

  task_scheduler_control(0, 0, SYS_SCHED_CTL_SET_PRIORITY, PROC_THREAD_PRIORITY);
  q = container_of(list_node_first(&group->queues), blkreq_queue_t, node);
  dev = q->bdev->private;
  host = dev->host_port->host;

  /* init done event of ports */
  for (i = 0; i < host->nr_ports; i++)
    host->ports[i]->io_done = &group->not_empty;

  while (1) {
    /* TO DO [dg]: command nesting and multithreading */
    blk_group_wait_event(group);
    /* Fixme [dg]: choose best device here and check whether
     * the port is busy */

    /*
     * update mask of queues
     * TO DO [dg]: it's ugly and not multithread, remake
     */
    m = host->waited_ports ^ prev_wports;
    while ((i = bit_find_lsf(m)) != -1) {
      m1 = 0;
      port = host->ports[i];
      for (j = 0; j < 2; j++) {
        dev = port->devices[j];
        if (dev && dev->bdev->queue)
          m1 |= (1 << dev->bdev->queue->index);
      }
      if (bit_test(&prev_wports, i)) {
        bit_clear(&prev_wports, i);
        qmask |= m1;
      } else {
        bit_set(&prev_wports, i);
        qmask &= ~m1;
      }
      bit_clear(&m, i);
    }

    rq = flush_get_group_blkrq(group, qmask);
    if (rq == NULL)
      continue;

    dev = rq->queue->bdev->private;
    port = dev->host_port;
    ata_port_lock(port);
    if (dev->drv->process_rq(dev, rq)) {
      rq->complete(rq);
      put_blk_request(rq);
    } else if (dev->cmd_phase != PHASE_INACTIVE) {
      ata_set_port_waited(port);
    }
    ata_port_unlock(port);
  }
  return NULL;
}

int start_rq_proc_threads(struct blk_group *group, int nr_threads)
{
  int ret = 0;
  struct ata_threads *threads;
  int i = 0;

  if (group->private)
    return 0;

  threads = malloc(sizeof(struct ata_threads));
  if (threads == NULL)
    return -ENOMEM;

  group->private = threads;
  threads->proc_th = malloc(nr_threads * sizeof(pthread_t));
  if (threads->proc_th == NULL)
    goto out;

  for (i = 0; i < nr_threads; i++)
    ret = -pthread_create(threads->proc_th + i, NULL, proc_rq_thread_fn, group);

  if (!ret) {
    if (nr_threads > 1)
      ret = set_blk_sigaction(IOSCHED_SIGNO, default_iosched_sigaction);
    else
      ret = set_blk_sigaction(IOSCHED_SIGNO, default_iosched_sigaction);
  }

out:
  threads->nr_threads = i;
  if (ret)
    stop_rq_proc_threads(group);

  return ret;
}

void stop_rq_proc_threads(struct blk_group *group)
{
  struct ata_threads *threads = group->private;
  int i;

  if (threads == NULL)
    return;

  if (threads->proc_th) {
   for (i = 0; i < threads->nr_threads; i++) {
     pthread_kill(threads->proc_th[i], SIGKILL);
     pthread_join(threads->proc_th[i], NULL);
   }
   free(threads->proc_th);
  }
  if (threads->tm_th.__tid) {
    pthread_kill(threads->tm_th, SIGKILL);
    pthread_join(threads->tm_th, NULL);
    threads->tm_th.__tid = 0;
  }

  free(threads);
  group->private = NULL;
}

int init_ata_timeouts(struct blk_group *group)
{
  ata_device_t *dev;
  blkreq_queue_t *q;
  ata_host_t *host;
  struct ata_threads *threads = group->private;
  int ret = 0;
  bool irq, dma;
  pthread_t th;

  if (threads == NULL)
    return -EINVAL;

  q = container_of(list_node_first(&group->queues), blkreq_queue_t, node);
  dev = q->bdev->private;
  host = dev->host_port->host;
  irq = ata_host_support_irq(host);
  dma = irq && ata_host_support_dma(host);
  list_for_each_entry(&group->queues, q, node) {
    dev = q->bdev->private;
    dev->intrq_tmr.id = dev->dma_tmr.id = -1;
    th = (threads->nr_threads > 1) ? threads->tm_th : threads->proc_th[0];
    /* TO DO [dg]: not block timeout signals */
    if (irq)
      ret = init_timeout_tmr(&dev->intrq_tmr, th, INTRQ_TIMEOUT_SIGNO, dev, true);
    if (dma && !ret)
      ret = init_timeout_tmr(&dev->dma_tmr, th, DMA_TIMEOUT_SIGNO, dev, true);
    if (ret)
      break;
  }
  if (!ret) {
    ret = set_blk_sigaction(INTRQ_TIMEOUT_SIGNO, timeout_sigaction);
    ret += set_blk_sigaction(DMA_TIMEOUT_SIGNO, timeout_sigaction);
  }

  if (ret) {
    destroy_ata_timeouts(group);
    ret = -1;
  }

  return ret;
}

void destroy_ata_timeouts(struct blk_group *group)
{
  ata_device_t *dev;
  blkreq_queue_t *q;
  ata_host_t *host;
  bool b, irq, dma;

  q = container_of(list_node_first(&group->queues), blkreq_queue_t, node);
  dev = q->bdev->private;
  host = dev->host_port->host;
  irq = ata_host_support_irq(host);
  dma = irq && ata_host_support_dma(host);
  list_for_each_entry(&group->queues, q, node) {
    b = false;
    dev = q->bdev->private;
    if (irq && dev->intrq_tmr.id >= 0) {
      destroy_timeout_tmr(&dev->intrq_tmr);
      b = true;
    }
    if (dma && dev->dma_tmr.id >= 0) {
      destroy_timeout_tmr(&dev->dma_tmr);
      b = true;
    }
    if (!b)
      break;
  }
}

int ata_process_rq(ata_device_t *dev, blk_request_t *rq)
{
  ata_cmd_t *cmd;
  int ret;
  uint32_t status;
  ata_port_t *port;

  cmd = ata_alloc_init_cmd(dev, dev->io_prot);
  if (cmd == NULL)
    return -ENOMEM;

  cmd->rq = rq;
  cmd->done = default_cmd_done;
  if (rq->type == REQ_TYPE_IO)
    ata_iocmd_from_request(rq, cmd);

  ret = ata_start_command(dev, cmd);
  if (ret) {
    if (cmd->err == ATA_ERROR_HOST) {
      port = dev->host_port;
      status = port->host->port_ops->read_status(port);
    } else {
      status = ata_read_device_status(dev);
    }
    dev->drv->error_handler(dev, cmd->err, status);
  }
  if (ret || !ata_intrq_enabled(dev)) {
    ata_done_cmd(dev, cmd);
    ata_free_cmd(cmd);
  }

  return ret;
}

int atapi_process_rq(ata_device_t *dev, blk_request_t *rq)
{
  return 0;
}
