/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 */

#include "generic_ide.h"

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <errno.h>
#include <sys/uevent.h>
#include <sys/ioport.h>
#include <pthread.h>
#include <byteorder.h>
#include <debug.h>
#include <sys/dma.h>
#include "generic.h"

/* dma operations */
static bool ide_build_prdt(dma_channel_t* chan, struct sg_entry *sg_entry,
                           uint16_t nr_entries);
static bool ide_start_dma(dma_channel_t *chan);
static bool ide_stop_dma(dma_channel_t *chan);
static void ide_dma_set_direction(dma_channel_t *chan, int dir);

static uint8_t ide_pci_inb(ulong_t port);
static void ide_pci_outb(ulong_t port, uint8_t val);
static uint16_t ide_pci_inw(ulong_t port);
static void ide_pci_outw(ulong_t port, uint16_t val);

static dma_channel_ops_t ide_dma_ops = {
  .load_sg_list = ide_build_prdt,
  .start = ide_start_dma,
  .stop = ide_stop_dma,
  .set_direction = ide_dma_set_direction
};

static ata_hwif_t ide_pci_port_hwif = {
  .inb = ide_pci_inb,
  .outb = ide_pci_outb,
  .inw = ide_pci_inw,
  .outw = ide_pci_outw
};

static uint8_t ide_pci_inb(ulong_t port)
{
  return inb(port);
}

static void ide_pci_outb(ulong_t port, uint8_t val)
{
  outb(port, val);
}

static uint16_t ide_pci_inw(ulong_t port)
{
  return inw(port);
}

static void ide_pci_outw(ulong_t port, uint16_t val)
{
  outw(port, val);
}

static inline uint16_t __get_pci_port(ata_port_t *port, int addr)
{
  pci_dev_t *dev = port->host->bus_private;

  return (uint16_t)dev->resources[dev->num_resources - 1].start;
}

static bool ide_build_prdt(dma_channel_t* chan, struct sg_entry *sg_list,
                           uint16_t nr_entries)
{
  long i, ndesc = 0;
  ide_prd_t *prd = chan->desc_tbl;
  uint32_t curr, len, count;
  bool ret = true;

  for (i = 0; (i < nr_entries) && ret; i++) {
    curr = (uint32_t)sg_list[i].start;
    if (curr & chan->align_mask) {
      /* TO DO [dg]: write message to log about data are not dword aligned */
      ret = false;
      break;
    }

    len = sg_list[i].len;
    if (curr + len < curr) {
      /* TO DO [dg]: write message to log about overflow */
      ret = false;
      break;
    }

    while (!len) {
      if (ndesc == chan->max_descs) {
        /* TO DO [dg]: write error message to log */
        ret = false;
        break;
      }
      len = MIN(len, chan->max_single_tsize);
      count = ((curr + len) & chan->boundary) - curr;
      count = MIN(count, len);

      prd[ndesc].base_addr = cpu_to_le32(curr);
      /* note: value of 0 in the 'count' field means 64 KB transfer */
      prd[ndesc++].count = cpu_to_le16((uint16_t)count);
      len -= count;
    }
  }

  if (ret && (chan->desc_tbl_len != ndesc)) {
    prd_highw_t w;

    prd[chan->desc_tbl_len].highw.value = 0;
    w.eot = 1;
    prd[ndesc - 1].highw.value = cpu_to_le16(w.value);
    chan->desc_tbl_len = ndesc;
  }

  return ret;
}

static int ide_start_dma(dma_channel_t *chan)
{
  ide_pci_set_bits(chan->private, IDE_BMICX, BMICX_SSBM);

  return 0;
}

static int ide_stop_dma(dma_channel_t *chan)
{
  ide_pci_clear_bits((ata_port_t*)chan->private, IDE_BMICX, BMICX_SSBM);

  return 0;
}

static void ide_dma_set_direction(dma_channel_t *chan, int dir)
{
  if (chan->dir == dir)
    return;

  chan->dir = dir;
  if (dir)
    ide_pci_set_bits((ata_port_t*)chan->private, IDE_BMICX, BMICX_RWCON);
  else
    ide_pci_clear_bits((ata_port_t*)chan->private, IDE_BMICX, BMICX_RWCON);
}

static void init_ide_registers(ulong_t base, ulong_t ctl, ata_registers_t *regs)
{
  regs->data = base;
  regs->feature = base + 1;
  regs->sect_count = base + 2;
  regs->lba_low = base + 3;
  regs->lba_mid = base + 4;
  regs->lba_high = base + 5;
  regs->device = base + 6;
  regs->command = base + 7;
  regs->dev_ctl = ctl;
}

int ide_pci_alloc_bmus_ports(ata_host_t *host)
{
  pci_dev_t *dev = host->bus_private;

  return alloc_ioport_range(dev->resources[dev->num_resources - 1].start,
                            IDE_IO_AREA_SIZE);
}

void ide_pci_free_bmus_ports(ata_host_t *host)
{
  pci_dev_t *dev = host->bus_private;

  free_ioport_range(dev->resources[dev->num_resources - 1].start,
                    IDE_IO_AREA_SIZE);
}

int ide_pci_init_port(ata_port_t *port, bool legacy)
{
  int i, ret;
  pthread_mutexattr_t attr = {PTHREAD_PROCESS_PRIVATE};
  pci_dev_t *dev = port->host->bus_private;
  ulong_t base, cntl;     /* base I/O and control IDE ports */

  port->hwif = &ide_pci_port_hwif;
  /*
   * allocate IDE ports, four PCI bars contains base
   * I/O and address ports for the primary and secondary
   * channel
   */
  if (legacy) {
    /* legacy mode */
    if (!port->num) {
      base = 0x1F0;
      cntl = 0x3F6;
      port->irq = 14;
    } else {
      base = 0x170;
      cntl = 0x376;
      port->irq = 15;
    }
  } else {
    /*
     * Native mode
     *
     * control block constists of 4 bytes
     * 0-1: reserved
     * 2: control register
     * 3: forward to ISA
     */
    i = port->num * 2;
    base = dev->resources[i].start;
    cntl = dev->resources[i + 1].start + 2;
    port->irq = dev->irq_line;
  }

  ret = alloc_ioport_range(base, NR_ATA_IOREGS);
  if (ret)
    return ret;

  ret = alloc_ioport_range(cntl, 1);
  if (ret)
    goto out;

  init_ide_registers(base, cntl, &port->regs);
  ret = pthread_mutex_init(&port->mutex, &attr);
  if (!ret)
    return 0;

  free_ioport_range(cntl, 1);
out:
  free_ioport_range(base, NR_ATA_IOREGS);

  return ret;
}

int ide_pci_init_dma(ata_port_t *port, uint32_t min_chunk_size, uint32_t max_tsize)
{
  uint16_t n;

  if (ata_pci_alloc_dma_res(port, min_chunk_size, max_tsize))
    return -ENOMEM;

  n = max_tsize / min_chunk_size + 1;

  return ide_pci_init_dma_chan(port, n, max_tsize, 0x10000);
}

int ide_pci_init_dma_chan(ata_port_t *port, uint16_t max_descs,
                          uint32_t res_area_size, uint32_t max_single_tsize)
{
  dma_channel_t *chan;
  uint32_t bound;

  memset(chan, 0, sizeof(dma_channel_t));
  chan->align_mask = 0xFFFC;          /* dword aligned */
  chan->boundary = ~(uint32_t)0xFFFF; /* 64 KB limit */
  chan->ops = &ide_dma_ops;
  chan->addr_width = 32;
  chan->desc_size = sizeof(ide_prd_t);
  chan->max_descs = max_descs;
  bound = ~(uint32_t)0xFFFF;          /* prd table must not accross 64 KB boundary */

  return ata_pci_init_dma_chan(port, bound, res_area_size,
                               max_single_tsize);
}

void ide_pci_release_port(ata_port_t *port)
{
  pci_dev_t *dev = port->host->bus_private;
  int i;

  ASSERT(port->num == 0 || port->num == 1, "");

  if (port->dma_chan) {
    ata_pci_destroy_dma_chan(port->dma_chan);
    free(port->dma_chan);
  }

  /* free IDE ports */
  i = port->num * 2;
  free_ioport_range(dev->resources[i].start, NR_ATA_IOREGS);
  free_ioport_range(dev->resources[i + 1].start, NR_ATA_IOREGS);
  pthread_mutex_destroy(&port->mutex);
}

port_status_t ide_pci_read_status(ata_port_t *port)
{
  port_status_t stat = 0;
  uint8_t s;

  s = (uint8_t)ide_pci_read_reg(port, IDE_BMISX);
  if (s & BMISX_IDEINTS) {
    stat |= PORT_STAT_IRQ;
    if (!(s & BMISX_BMIDEA))
      stat |= PORT_STAT_IO_DONE;
  }
  if (s & BMISX_DMA_ERR)
    stat |= PORT_STAT_ERROR;

  return s;
}

void ide_pci_set_irq(ata_port_t *port, uint16_t irq)
{
  /* yet stub */
}

void ide_pci_clear_irq(ata_port_t *port)
{
  /*
   * clear interrupt and error bits in the status register
   * by writing '1' to them
   */
  ide_pci_set_bits(port, IDE_BMISX, BMISX_IDEINTS | BMISX_DMA_ERR);
}

void ide_pci_enable_dma(ata_device_t *dev, bool enable)
{
  /*
   * yet stub, cannot by implemented because the
   * pci driver is not still supporting access
   * to the extended configuration space
   */
}

void ide_pci_init_bmdma(ata_host_t *host)
{
  ide_pci_clear_bits(host->ports[0], IDE_BMISX, ~(BMISX_DMA0CAP | BMISX_DMA1CAP));
  if (host->nr_ports > 1)
    ide_pci_clear_bits(host->ports[1], IDE_BMISX, ~(BMISX_DMA0CAP | BMISX_DMA1CAP));
}

uint32_t ide_pci_read_reg(ata_port_t *port, int addr)
{
  uint32_t ret;
  uint16_t p;

  p = __get_pci_port(port, addr);
  if (addr == IDE_BMIDTPX)  /* 32-bit register */
    ret = inl(p);
  else
    ret = inb(p);

  return ret;
}

void ide_pci_write_reg(ata_port_t *port, int addr, uint32_t val)
{
  uint16_t p;

  p = __get_pci_port(port, addr);
  if (addr == IDE_BMIDTPX)
    outl(p, val);
  else
    outb(p, (uint8_t)val);
}

void ide_pci_set_bits(ata_port_t *port, int reg_addr, uint8_t mask)
{
  uint16_t p;
  uint8_t val;

  if (reg_addr != IDE_BMIDTPX) {
    p = __get_pci_port(port, reg_addr);
    val = inb(p);
    val |= mask;
    outb(p, val);
  }
}

void ide_pci_clear_bits(ata_port_t *port, int reg_addr, uint8_t mask)
{
  uint16_t p;
  uint8_t val;

  if (reg_addr != IDE_BMIDTPX) {
    p = __get_pci_port(port, reg_addr);
    val = inb(p);
    val &= ~mask;
    outb(p, val);
  }
}
