/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * ATA PCI controller init and release routines
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/dma.h>
#include <sys/mman.h>
#include <debug.h>
#include <pci/names.h>
#include "generic.h"

#define DMA_ALLOC_FAILED (uintptr_t)MAP_FAILED

void ata_print_pci_devices(pci_dev_t *devs, int nr_devs)
{
  vendor_desc_t *vdesc;
  product_desc_t *pdesc;
  int i;

  if (nr_devs <= 0)
    return;

  for (i = 0; i < nr_devs; i++) {
    vdesc = lookup_vendor(devs[i].ident.vendor_id);
    pdesc = lookup_product(vdesc, devs[i].ident.device_id);
    printf("[hw] %s by %s is detected\n", pdesc->name, vdesc->name);
  }
}

int ata_pci_alloc_dma_res(ata_port_t *port, uint32_t min_chunk_size,
                          uint32_t max_tsize)
{
  int ret = 0;

  /* TO DO [dg]: alloc scatter/gather slab */

  port->dma_chan = malloc(sizeof(dma_channel_t));
  if (port->dma_chan == NULL) {
    ret = -ENOMEM;
  }

  return ret;
}

int ata_pci_init_dma_chan(ata_port_t *port, uint32_t desc_tbl_boundary,
                          uint32_t res_area_size, uint32_t max_single_tsize)
{
  int ret = -ENOMEM;
  uint32_t n, npages;
  dma_channel_t *chan = port->dma_chan;

  chan->max_single_tsize = max_single_tsize;
  chan->flags |= DMA_CHAN_SG;
  chan->private = port;
  if (res_area_size) {
    chan->phy_region_start = (uintptr_t)dma_alloc_pages(res_area_size >> PAGE_WIDTH);
    if (chan->phy_region_start == DMA_ALLOC_FAILED)
      goto out;

    chan->phy_region_len = res_area_size & (~0L << PAGE_WIDTH);
    chan->vdma_buf = mmap(NULL, chan->phy_region_len, PROT_READ | PROT_WRITE,
                          MAP_PRIVATE | MAP_ANONYMOUS | MAP_PHYS, 0, chan->phy_region_start);
    if (chan->vdma_buf == MAP_FAILED)
      goto out;
  }

  /* allocate physical region descriptor table */
  n = chan->max_descs * chan->desc_size;
  npages = n >> PAGE_WIDTH;

  if ((chan->max_descs * chan->desc_size) & PAGE_MASK)
    npages++;

  chan->phy_desc_tbl_addr = (uintptr_t)dma_alloc_pages(npages);
  if (chan->phy_desc_tbl_addr == DMA_ALLOC_FAILED)
    goto out;

  /* the table must not across 64 KB boundary */
  n = ((chan->phy_desc_tbl_addr + n) & ~0xFFFFL) - chan->phy_desc_tbl_addr;
  if (n >> PAGE_WIDTH != npages) {
    /* unmap the tail acrossing 64 KB boundary */
    dma_free_pages((void*)chan->phy_desc_tbl_addr + n, npages - (n >> PAGE_WIDTH));
  }

  chan->max_descs = MIN(chan->max_descs, n / chan->desc_size);
  chan->max_total_tsize = max_single_tsize * chan->max_descs;
  chan->desc_tbl = mmap(NULL, n, PROT_READ | PROT_WRITE,
         MAP_PRIVATE | MAP_ANONYMOUS | MAP_PHYS, 0, chan->phy_desc_tbl_addr);

  if (chan->desc_tbl != MAP_FAILED)
    return 0;

out:
  if (chan->phy_desc_tbl_addr && chan->phy_desc_tbl_addr != DMA_ALLOC_FAILED)
    dma_free_pages((void*)chan->phy_desc_tbl_addr, npages);
  if (chan->vdma_buf && chan->vdma_buf != MAP_FAILED)
    munmap(chan->vdma_buf, chan->phy_region_len);
  if (chan->phy_region_start && chan->phy_region_start != DMA_ALLOC_FAILED)
    dma_free_pages((void*)chan->phy_region_start, res_area_size >> PAGE_WIDTH);

  return ret;
}

void ata_pci_destroy_dma_chan(dma_channel_t *chan)
{
  uint32_t n;

  if (chan->vdma_buf) {
    munmap(chan->vdma_buf, chan->phy_region_len);
    dma_free_pages((void*)chan->phy_region_start, chan->phy_region_len >> PAGE_WIDTH);
  }
  n = chan->max_descs * chan->desc_size;
  n += PAGE_SIZE - (n & PAGE_MASK);
  munmap(chan->desc_tbl, n);
  dma_free_pages((void*)chan->phy_desc_tbl_addr, n >> PAGE_WIDTH);
}
