/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (C) 2005-2006 Red Hat Inc
 *	Alan Cox <alan@redhat.com>
 *
 *  Ported from Linux by Dmitry Gromada
 */

#include <stdbool.h>
#include <errno.h>
#include <ata/ata.h>
#include <ata/libata.h>
#include <block/blkdev.h>
#include <ata/host_drv.h>
#include <pci/pci.h>
#include <pci_ids.h>
#include <pci_regs.h>
#include <debug.h>
#include "generic_ide.h"

/* PROGIG bits */
enum {
  PROGIF_PRIM_NATIVE = 0x10,    /* primary port works in the native mode flag */
  PROGIF_SEC_NATIVE = 0x40      /* secondary port works in the native mode flag */
};

struct chipset_desc {
  uint16_t vend_id;
  uint16_t dev_id;
  uint8_t dma_mode;
  /* timing related values */
  uint8_t timing_offset;  /* register offset in the configuration space */
  uint8_t timing_clock;
  ata_port_ops_t *ops;
};

static pci_dev_t *devices = NULL;
static int nr_devices;

static int amd74_scan_hosts(void);

static int amd74_init_host(ata_host_t *host, int host_num);
static void amd74_shutdown_host(ata_host_t *host);

static int amd74_init_port(ata_port_t *port);
static void amd74_shutdown_port(ata_port_t *port);
static void amd74_enable_dma(ata_port_t *port, int dev_id, bool enable);
static void amd74_set_pio_mode(ata_port_t *port, int dev_id, uint8_t mode);
static void amd74_set_dma_mode(ata_port_t *port, int dev_id, uint8_t mode);
static void amd74_filter_io_modes(ata_port_t *port, int dev_id,
                                  struct ata_io_modes *modes);

static ata_host_driver_t amd74_ide_driver = {
  .name = "AMD74_IDE",
  .phys_iface = ATA_PHYIF_IDE,
  .scan_hosts = amd74_scan_hosts,
  .init_host = amd74_init_host,
  .shutdown_host = amd74_shutdown_host
};

static ata_port_ops_t amd74_ide_port_ops = {
  .init = amd74_init_port,
  .shutdown = amd74_shutdown_port,
  .set_pio_mode = amd74_set_pio_mode,
  .set_dma_mode = amd74_set_dma_mode,
  .init_dma = ide_pci_init_dma,
  .enable_dma = amd74_enable_dma,
  .read_status = ide_pci_read_status,
  .set_irq = ide_pci_set_irq,
  .clear_irq = ide_pci_clear_irq,
  .filter_io_modes = amd74_filter_io_modes
};

static struct chipset_desc amd74_chipsets[] = {
  { PCI_VENDOR_ID_AMD,  PCI_DEVICE_ID_AMD_COBRA_7401, ATA_UDMA3, 0x40, 1 },
  { PCI_VENDOR_ID_AMD,  PCI_DEVICE_ID_AMD_VIPER_7409, ATA_UDMA4, 0x40, 2 },
  { PCI_VENDOR_ID_AMD,  PCI_DEVICE_ID_AMD_VIPER_7411, ATA_UDMA5, 0x40, 3 },
  { PCI_VENDOR_ID_AMD,  PCI_DEVICE_ID_AMD_OPUS_7441, ATA_UDMA5, 0x40, 3 },
  { PCI_VENDOR_ID_AMD,  PCI_DEVICE_ID_AMD_8111_IDE, ATA_UDMA6, 0x40, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_IDE, ATA_UDMA5, 0x50, 3 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE2_IDE,  ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE2S_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE3_IDE,  ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE3S_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_CK804_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_MCP04_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_MCP51_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_MCP55_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_MCP61_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_MCP65_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_MCP67_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_MCP73_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_NVIDIA, PCI_DEVICE_ID_NVIDIA_NFORCE_MCP77_IDE, ATA_UDMA6, 0x50, 4 },
  { PCI_VENDOR_ID_AMD,  PCI_DEVICE_ID_AMD_CS5536_IDE, ATA_UDMA5, 0x40, 3 },
};

DECLARE_FIND_DESC_FN(amd74_chipset_desc, amd74_chipsets)
DECLARE_SCAN_HOSTS_FN(amd74_scan_hosts, amd74_chipsets, devices, nr_devices)

/**
 *	timing_setup		-	shared timing computation and load
 *	@ap: ATA port being set up
 *	@adev: drive being configured
 *	@offset: port offset
 *	@speed: target speed
 *	@clock: clock multiplier (number of times 33MHz for this part)
 *
 *	Perform the actual timing set up for Nvidia or AMD PATA devices.
 *	The actual devices vary so they all call into this helper function
 *	providing the clock multipler and offset (because AMD and Nvidia put
 *	the ports at different locations).
 */

static void timing_setup(ata_port_t *port, int dev_id, int offset,
                         int clock, uint8_t mode, bool dma)
{
	static const unsigned char amd74_cyc2udma[] = {
		6, 6, 5, 4, 0, 1, 1, 2, 2, 3, 3, 3, 3, 3, 3, 7
	};

	pci_dev_t *dev = port->host->bus_private;
	ata_device_t *peer = port->devices[1 - dev_id];
	int dn = port->num * 2 + dev_id;
	struct ata_timing at, apeer;
	int T, UT;
	const int amd74_clock = 33333;	/* KHz. */
	uint8_t t;
	char buf[8];

	/* debug */
	uint32_t reg_addr;
	uint8_t reg_val;

	T = 1000000000 / amd74_clock;
	UT = T;
	if (clock >= 2)
		UT = T / 2;

	if (dma)
	  sprintf(buf, "dma");
	else
	  sprintf(buf, "pio");

	if (ata_timing_compute(port->devices[dev_id], mode, dma, &at, T, UT) < 0) {
		printf("[ide_amd]: unknown %s mode %d\n", buf, mode);
		return;
	}

	if (peer) {
		/* This may be over conservative */
		if (peer->dma_mode) {
			ata_timing_compute(peer, peer->dma_mode, dma, &apeer, T, UT);
			ata_timing_merge(&apeer, &at, &at, ATA_TIMING_8BIT);
		}
		ata_timing_compute(peer, peer->pio_mode, false, &apeer, T, UT);
		ata_timing_merge(&apeer, &at, &at, ATA_TIMING_8BIT);
	}

	if (dma) {
	  if (mode == ATA_UDMA5 && amd74_clock <= 33333) at.udma = 1;
	  if (mode == ATA_UDMA6 && amd74_clock <= 33333) at.udma = 15;
	}

	/*
	 *	Now do the setup work
	 */
	/* Configure the address set up timing */
	pci_read_config_byte(dev, offset + 0x0C, &t);
	t = (t & ~(3 << ((3 - dn) << 1))) | ((BOUNDED_VAL(at.setup, 1, 4) - 1) << ((3 - dn) << 1));
	pci_write_config_byte(dev, offset + 0x0C , t);

	/* Configure the 8bit I/O timing */
	pci_write_config_byte(dev, offset + 0x0E + (1 - (dn >> 1)),
		((BOUNDED_VAL(at.act8b, 1, 16) - 1) << 4) | (BOUNDED_VAL(at.rec8b, 1, 16) - 1));

	/* Drive timing */
	pci_read_config_byte(dev, reg_addr, &reg_val);
	pci_write_config_byte(dev, offset + 0x08 + (3 - dn),
		((BOUNDED_VAL(at.active, 1, 16) - 1) << 4) | (BOUNDED_VAL(at.recover, 1, 16) - 1));

	switch (clock) {
		case 1:
		t = at.udma ? (0xc0 | (BOUNDED_VAL(at.udma, 2, 5) - 2)) : 0x03;
		break;

		case 2:
		t = at.udma ? (0xc0 | amd74_cyc2udma[BOUNDED_VAL(at.udma, 2, 10)]) : 0x03;
		break;

		case 3:
		t = at.udma ? (0xc0 | amd74_cyc2udma[BOUNDED_VAL(at.udma, 1, 10)]) : 0x03;
		break;

		case 4:
		t = at.udma ? (0xc0 | amd74_cyc2udma[BOUNDED_VAL(at.udma, 1, 15)]) : 0x03;
		break;

		default:
			return;
	}

	/* UDMA timing */
	if (at.udma)
		pci_write_config_byte(dev, offset + 0x10 + (3 - dn), t);
}

#if 0
/**
 *	amd74_pre_reset		-	perform reset handling
 *	@link: ATA link
 *	@deadline: deadline jiffies for the operation
 *
 *	Reset sequence checking enable bits to see which ports are
 *	active.
 */

static int amd74_pre_reset(struct ata_link *link, unsigned long deadline)
{
	static const struct pci_bits amd74_enable_bits[] = {
		{ 0x40, 1, 0x02, 0x02 },
		{ 0x40, 1, 0x01, 0x01 }
	};

	struct ata_port *ap = link->ap;
	struct pci_dev *pdev = to_pci_dev(ap->host->dev);

	if (!pci_test_config_bits(pdev, &amd74_enable_bits[ap->port_no]))
		return -ENOENT;

	return ata_sff_prereset(link, deadline);
}

/**
 *	nv_probe_init	-	cable detection
 *	@lin: ATA link
 *
 *	Perform cable detection. The BIOS stores this in PCI config
 *	space for us.
 */

static int nv_pre_reset(struct ata_link *link, unsigned long deadline)
{
	static const struct pci_bits nv_enable_bits[] = {
		{ 0x50, 1, 0x02, 0x02 },
		{ 0x50, 1, 0x01, 0x01 }
	};

	struct ata_port *ap = link->ap;
	struct pci_dev *pdev = to_pci_dev(ap->host->dev);

	if (!pci_test_config_bits(pdev, &nv_enable_bits[ap->port_no]))
		return -ENOENT;

	return ata_sff_prereset(link, deadline);
}

static void nv_host_stop(struct ata_host *host)
{
	u32 udma = (unsigned long)host->bus_private_data;

	/* restore PCI config register 0x60 */
	pci_write_config_dword(to_pci_dev(host->dev), 0x60, udma);
}
#endif

static int amd74_init_host(ata_host_t *host, int host_num)
{
	pci_dev_t *dev;
	uint8_t fifo;
	int udma = 0;
	struct chipset_desc *desc;

	if (host_num > nr_devices)
	  return -EINVAL;

	dev = devices + host_num;
	ata_pci_init_host(host, dev);
	pci_read_config_byte(dev, 0x41, &fifo);
	host->drv = &amd74_ide_driver;
	host->nr_ports = 2;
	host->port_ops = &amd74_ide_port_ops;

	/* Serenade ? */
	if (dev->ident.device_id == PCI_DEVICE_ID_AMD_8111_IDE &&
	    dev->ident.subsys_vendor_id == PCI_VENDOR_ID_AMD &&
			dev->ident.subsys_id == PCI_DEVICE_ID_AMD_SERENADE) {
		udma = ATA_UDMA6;	/* UDMA 100 only */
	}

	desc = amd74_chipset_desc(dev);
	if (!udma)
	  udma = desc->dma_mode;

	/* init bmdma for amd 7401, 7409 chipsets */
	if ((dev->ident.vendor_id == PCI_VENDOR_ID_AMD) && (udma < ATA_UDMA5))
		ide_pci_init_bmdma(host);

	/* Check for AMD7411 */
	if (dev->ident.device_id == PCI_DEVICE_ID_AMD_VIPER_7411)
		/* FIFO is broken */
		pci_write_config_byte(dev, 0x41, fifo & 0x0F);
	else
		pci_write_config_byte(dev, 0x41, fifo | 0xF0);

	/* Cable detection on Nvidia chips doesn't work too well,
	 * cache BIOS programmed UDMA mode.
	 */
	if (dev->ident.vendor_id == PCI_VENDOR_ID_NVIDIA) {
		uint32_t udma;

		pci_read_config_dword(dev, 0x60, &udma);
		host->private = (void *)(unsigned long)udma;
	}

	host->io_modes.pio = 0x1F;       /* PIO modes from 0 upto 4 */
	host->io_modes.mwdma = (1 << ATA_MWDMA0) |
                         (1 << ATA_MWDMA1) |
                         (1 << ATA_MWDMA2);
	host->io_modes.udma = 1 << udma;

	return ide_pci_alloc_bmus_ports(host);
}

static void amd74_shutdown_host(ata_host_t *host)
{

}

static int amd74_init_port(ata_port_t *port)
{
  bool legacy;
  pci_dev_t *dev = port->host->bus_private;

  if (!port->num)
    legacy = !(dev->ident.prog_if_id & PROGIF_PRIM_NATIVE);
  else
    legacy = !(dev->ident.prog_if_id & PROGIF_SEC_NATIVE);

  return ide_pci_init_port(port, legacy);
}

static void amd74_shutdown_port(ata_port_t *port)
{

}

static void amd74_enable_dma(ata_port_t *port, int dev_id, bool enable)
{

}

static void amd74_set_pio_mode(ata_port_t *port, int dev_id, uint8_t mode)
{
  struct chipset_desc *desc;

  desc = amd74_chipset_desc(port->host->bus_private);
  timing_setup(port, dev_id, desc->timing_offset, desc->timing_clock, mode, false);
}

static void amd74_set_dma_mode(ata_port_t *port, int dev_id, uint8_t mode)
{
  struct chipset_desc *desc;

  desc = amd74_chipset_desc(port->host->bus_private);
  timing_setup(port, dev_id, desc->timing_offset, desc->timing_clock, mode, true);
}

static void amd74_filter_io_modes(ata_port_t *port, int dev_id,
                                struct ata_io_modes *modes)
{
  static const unsigned int udma_mask_map[] =
    { ATA_UDMA2, ATA_UDMA1, ATA_UDMA0, 0,
      ATA_UDMA3, ATA_UDMA4, ATA_UDMA5, ATA_UDMA6 };
  uint32_t udma;

  /* find out what BIOS configured */
  udma = (unsigned long)port->host->private;

  if (port->num == 0)
    udma >>= 16;
  if (dev_id == 0)
    udma >>= 8;

  if ((udma & 0xc0) == 0xc0)
    modes->udma &= 1 << udma_mask_map[udma & 0x7];
}

/*
 * TO DO [dg]: replace the name of the function by the by standard one
 *   after the dynamic elf will be available
 */
ata_host_driver_t *amd74_init_ide_driver(void)
{
  return &amd74_ide_driver;
}
