/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Copyright (C) 1998-1999 Andrzej Krzysztofowicz, Author and Maintainer
 * Copyright (C) 1998-2000 Andre Hedrick <andre@linux-ide.org>
 * Copyright (C) 2003 Red Hat Inc <alan@redhat.com>
 * Copyright (C) 2006-2007 MontaVista Software, Inc. <source@mvista.com>
 *
 *
 * Documentation:
 *
 *  Publically available from Intel web site. Errata documentation
 * is also publically available. As an aide to anyone hacking on this
 * driver the list of errata that are relevant is below.going back to
 * PIIX4. Older device documentation is now a bit tricky to find.
 *
 * Errata of note:
 *
 * Unfixable
 *  PIIX4    errata #9  - Only on ultra obscure hw
 *  ICH3   errata #13     - Not observed to affect real hw
 *          by Intel
 *
 * Things we must deal with
 *  PIIX4 errata #10  - BM IDE hang with non UDMA
 *          (must stop/start dma to recover)
 *  440MX   errata #15  - As PIIX4 errata #10
 *  PIIX4 errata #15  - Must not read control registers
 *          during a PIO transfer
 *  440MX   errata #13  - As PIIX4 errata #15
 *  ICH2  errata #21  - DMA mode 0 doesn't work right
 *  ICH0/1  errata #55  - As ICH2 errata #21
 *  ICH2  spec c #9 - Extra operations needed to handle
 *          drive hotswap [NOT YET SUPPORTED]
 *  ICH2    spec c #20  - IDE PRD must not cross a 64K boundary
 *          and must be dword aligned
 *  ICH2    spec c #24  - UDMA mode 4,5 t85/86 should be 6ns not 3.3
 *
 * Should have been BIOS fixed:
 *  450NX:  errata #19  - DMA hangs on old 450NX
 *  450NX:  errata #20  - DMA hangs on old 450NX
 *  450NX:  errata #25  - Corruption with DMA on old 450NX
 *  ICH3    errata #15      - IDE deadlock under high load
 *          (BIOS must set dev 31 fn 0 bit 23)
 *  ICH3  errata #18  - Don't use native mode
 *
 *  Ported by Dmitry Gromada <gromada@jarios.org>
 */

/*
 * NOTE:
 *  PIO and ultra dma mode settings are duplicated from the original
 *  implementation. Some comments are added there.
 *
 *  Differences from the original implementation:
 *
 *  1) Supporting of 82371MX and 82801EB_1 chips is deprecated
 *  2) Supporting of SWDMA modes is deprecated
 *  3) IDE mode detection is added
 *  4) Need control logic to integrate the driver into the system is added
 */

#include <stdbool.h>
#include <errno.h>
#include <ata/ata.h>
#include <block/blkdev.h>
#include <ata/host_drv.h>
#include <pci/pci.h>
#include <pci_ids.h>
#include <pci_regs.h>
#include <debug.h>
#include "generic_ide.h"

#define HOST_CHIPSET_ICH 0x01

/* chipset enumeration */
enum {
  PIIXA,
  PIIXB,
  PIIX3,
  PIIX4,
  ICH0,
  ICH,
  ICH2,
  ICH2M,
  ICH3,
  ICH3M,
  ICH4,
  ICH5,
  C_ICH,
  ICH6,
  ICH7,
  ESB2,
  ICH8M
};

/* PCI register address map */
enum {
  PIIX_PRIM_IDE_TIMING = 0x40,
  PIIX_SEC_IDE_TIMING = 0x42,
  PIIX_SLAVE_IDE_TIMING = 0x44,
  PIIX_UDMA_CTL = 0x48,       /* ultra DMA control */
  PIIX_UDMA_TIMING = 0x4A,    /* ultra DMA timings */
  /* ich chipsets registers */
  PIIX_IDE_IO_CONFIG = 0x54,
  PIIX_ATC = 0xC0,            /* APM trapping control */
  PIIX_ATS = 0xC4             /* APM trapping status */
};

/* ich chipsets programming interface bits */
enum {
  POP_MODE_SEL = 0x01,  /* primary channel native mode selected */
  POP_MODE_CAP = 0x02,  /* primary channel both legacy and native modes supporting */
  SOP_MODE_SEL = 0x04,  /* secondary channel native mode selected */
  SOP_MODE_CAP = 0x08,  /* secondary channel both legacy and native modes supporting */
  BM_SUPP = 0x80        /* bus master operations suporting */
};

struct chipset_desc {
  uint16_t vend_id;
  uint16_t dev_id;
  /* chipset family */
  int family;
  uint8_t dma_mode;
};

/* supported chipsets */
static struct chipset_desc piix_chipsets[] = {
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371FB_0, PIIXA, ATA_DMA_NOT},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371FB_1, PIIXB, ATA_DMA_NOT},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371SB_1, PIIX3, ATA_DMA_NOT},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371AB, PIIX4, ATA_UDMA2},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801AB_1, ICH0, ATA_UDMA2},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82443MX_1, PIIX4, ATA_UDMA2},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801AA_1, ICH, ATA_UDMA4},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82372FB_1, PIIX4, ATA_UDMA4},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82451NX, PIIX4, ATA_UDMA2},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801BA_9, ICH2, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801BA_8, ICH2M, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801CA_10, ICH3M, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801CA_11, ICH3, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801DB_11, ICH4, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801EB_11, ICH5, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801E_11, C_ICH, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801DB_10, ICH4, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ESB_2, ICH5, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH6_19, ICH6, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH7_21, ICH7, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82801DB_1, ICH4, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ESB2_18, ESB2, ATA_UDMA5},
  {PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_ICH8_6, ICH8M, ATA_UDMA5}
};

static pci_dev_t *devices = NULL;
static int nr_devices;
static int no_piix_dma;

static int piix_scan_hosts(void);

static int piix_init_host(ata_host_t *host, int host_num);
static void piix_shutdown_host(ata_host_t *host);

static int piix_init_port(ata_port_t *port);
static void piix_shutdown_port(ata_port_t *port);
static void piix_set_pio_mode(ata_port_t *port, int dev_id, uint8_t pio);
static void piix_set_dma_mode(ata_port_t *port, int dev_id, uint8_t speed);
static void piix_enable_dma(ata_port_t *port, int dev_id, bool enable);

static void piix_check_450nx(void);

static ata_host_driver_t piix_driver = {
  .name = "PIIX_IDE",
  .phys_iface = ATA_PHYIF_IDE,
  .scan_hosts = piix_scan_hosts,
  .init_host = piix_init_host,
  .shutdown_host = piix_shutdown_host
};

static ata_port_ops_t piix_port_ops = {
  .init = piix_init_port,
  .shutdown = piix_shutdown_port,
  .set_pio_mode = piix_set_pio_mode,
  .set_dma_mode = piix_set_dma_mode,
  .init_dma = ide_pci_init_dma,
  .enable_dma = piix_enable_dma,
  .read_status = ide_pci_read_status,
  .set_irq = ide_pci_set_irq,
  .clear_irq = ide_pci_clear_irq
};

DECLARE_FIND_DESC_FN(piix_chipset_desc, piix_chipsets)
DECLARE_SCAN_HOSTS_FN(__piix_scan_hosts, piix_chipsets, devices, nr_devices)

static inline bool ata_media_disk(ata_device_t *dev)
{
  return (dev && dev->media.type == ATA_MEDIA_DISK);
}

static int piix_nr_ports(pci_dev_t *dev, int family)
{
  int n;

  if (family < ICH6) {
    n = 2;
  } else {
    /*
     * IDE controller has 1 port; the first SATA controller (function 2)
     * emulates 2 IDE ports; the second SATA controller on ICH8M
     * chipsets emulates 1 port.
     */
    n = (dev->addr.function == 2) ? 2 : 1;
  }

  return n;
}

/*
 * TO DO [dg]: replace the name of the function by the by standard one
 *   after the dynamic elf will be available
 */
ata_host_driver_t *piix_init_ide_driver(void)
{
  return &piix_driver;
}

int piix_scan_hosts(void)
{
  int r;

  r = __piix_scan_hosts();
  if (!r)
    piix_check_450nx();

  return r;
}

static int piix_init_host(ata_host_t *host, int host_num)
{
  struct chipset_desc *desc;
  pci_dev_t *dev;

  if (host_num >= nr_devices)
    return -EINVAL;

  ata_pci_init_host(host, devices + host_num);

  dev = host->bus_private;
  host->drv = &piix_driver;
  host->port_ops = &piix_port_ops;
  desc = piix_chipset_desc(host->bus_private);

  if (desc->family >= ICH0) {
    uint32_t config;

    /* enable the write buffer to be used in a split manner */
    pci_read_config_dword(host->bus_private, PIIX_IDE_IO_CONFIG, &config);
    pci_write_config_dword(host->bus_private, PIIX_IDE_IO_CONFIG, config | 0x4000);
    host->flags |= HOST_CHIPSET_ICH;
  }

  host->nr_ports = piix_nr_ports(host->bus_private, desc->family);
  host->io_modes.pio = 0x1F;  /* PIO modes UP to 4 are supported */
  if (no_piix_dma)
    return 0;

  if (desc->dma_mode > 2) {
    /* modes from udma 3 (ATA/33) up to desc->dma_mode are supported */
    host->io_modes.udma = ((1 << (desc->dma_mode + 1)) - 1) & 0xF8;
  } else {
    host->io_modes.udma = 1 << desc->dma_mode;
  }
  host->io_modes.mwdma = (1 << ATA_MWDMA0) | (1 << ATA_MWDMA1) | (1 << ATA_MWDMA2);

  return ide_pci_alloc_bmus_ports(host);
}

static void piix_shutdown_host(ata_host_t *host)
{
  /* yet stub */
}


/* functions servicing the piix ide controller */

static int piix_init_port(ata_port_t *port)
{
  int ret = 0;
  struct chipset_desc *desc;
  bool legacy = false;
  pci_dev_t *dev = port->host->bus_private;

  /* check set IDE mode */
  desc = piix_chipset_desc(dev);
  if ((desc->family < ICH3) || (desc->family == C_ICH)) {
    legacy = true;
  } else {
    uint8_t mask = port->num ? SOP_MODE_SEL : POP_MODE_SEL;

    legacy = !(dev->ident.prog_if_id & mask);
  }

  ret = ide_pci_init_port(port, legacy);
  if (ret)
    return ret;

  /* set pio mode 2 for both drives */
  piix_set_pio_mode(port, 0, 2);
  piix_set_pio_mode(port, 1, 2);

  return ret;
}

static void piix_shutdown_port(ata_port_t *port)
{
  piix_enable_dma(port, 0, false);
  piix_enable_dma(port, 1, false);
  ide_pci_release_port(port);
}

static void piix_set_pio_mode(ata_port_t *port, int dev_id,  uint8_t pio)
{
  pci_dev_t *pci_dev = port->host->bus_private;
  ata_device_t *ata_dev = port->devices[dev_id];
  int master_port = port->num ? PIIX_SEC_IDE_TIMING : PIIX_PRIM_IDE_TIMING;
  int slave_port = PIIX_SLAVE_IDE_TIMING;
  uint16_t master_data;
  uint8_t slave_data;
  int control = 0;

  /* ISP  RTC */
  static const uint8_t timings[][2]= {
    { 0, 0 }, { 0, 0 },
    { 1, 0 }, { 2, 1 },
    { 2, 3 },
  };

  pci_read_config_word(pci_dev, master_port, &master_data);

  if (pio > 1)
    control |= 1; /* Programmable timing on */
  if (ata_media_disk(ata_dev))
    control |= 4; /* Prefetch, post write */
  if (pio > 2)
    control |= 2; /* IORDY */
  if (dev_id) {
    master_data |=  0x4000;   /* enable ISP and RTC programming uniquely for each drive */
    master_data &= ~0x0070;   /* clear selects for drive 1 */
    if (pio > 1) {
      /* Set PPE, IE and TIME */
      master_data |= control << 4;
    }
    pci_read_config_byte(pci_dev, slave_port, &slave_data);
    slave_data &= port->num ? 0x0f : 0xf0;  /* clear timings for an appropriate port */
    /* set timings for the appropriate channel */
    slave_data |= ((timings[pio][0] << 2) | timings[pio][1]) <<
             (port->num ? 4 : 0);
  } else {
    master_data &= ~0x3307;   /* clear ISP, RTC and selects for drive 0 */
    if (pio > 1) {
      /* enable PPE, IE and TIME */
      master_data |= control;
    }
    master_data |= (timings[pio][0] << 12) | (timings[pio][1] << 8);
  }
  pci_write_config_word(pci_dev, master_port, master_data);
  if (dev_id)
    pci_write_config_byte(pci_dev, slave_port, slave_data);
}

static void piix_set_dma_mode(ata_port_t *port, int dev_id, uint8_t speed)
{
  pci_dev_t *dev = port->host->bus_private;
  uint8_t maslave    = port->num ? PIIX_SEC_IDE_TIMING : PIIX_PRIM_IDE_TIMING;
  int dn        = port->num * 2 + dev_id;
  /* udma timings for the appropriate device */
  int a_speed   = 3 << (dn * 4);
  /* udma enable mask for the appropriate drive */
  int u_flag    = 1 << dn;
  /* base clock mask for the appropriate device */
  int v_flag    = 0x01 << dn;
  /* fast base clock mask for the appropriate device */
  int w_flag    = 0x10 << dn;
  int u_speed   = 0;
  int sitre;
  uint16_t reg4042, reg4a;
  uint8_t reg48, reg54, reg55;

  pci_read_config_word(dev, maslave, &reg4042);
  sitre = (reg4042 & 0x4000) ? 1 : 0;
  pci_read_config_byte(dev, PIIX_UDMA_CTL, &reg48);
  pci_read_config_word(dev, PIIX_UDMA_TIMING, &reg4a);
  pci_read_config_byte(dev, PIIX_IDE_IO_CONFIG, &reg54);
  pci_read_config_byte(dev, PIIX_IDE_IO_CONFIG + 1, &reg55);

  if (speed >= ATA_UDMA0) {
    uint8_t udma = speed - ATA_UDMA0;

    u_speed = MIN(2 - (udma & 1), udma) << (dn * 4);

    if (!(reg48 & u_flag))
      pci_write_config_byte(dev, PIIX_UDMA_CTL,
                            reg48 | u_flag);
    if (speed == ATA_UDMA5) {
      pci_write_config_byte(dev, PIIX_IDE_IO_CONFIG + 1,
                            (uint8_t) reg55|w_flag);
    } else {
      pci_write_config_byte(dev, PIIX_IDE_IO_CONFIG + 1,
                            (uint8_t) reg55 & ~w_flag);
    }
    if ((reg4a & a_speed) != u_speed)
      pci_write_config_word(dev, PIIX_UDMA_TIMING, (reg4a & ~a_speed) | u_speed);
    if (speed > ATA_UDMA2) {
      if (!(reg54 & v_flag))
        pci_write_config_byte(dev, PIIX_IDE_IO_CONFIG, reg54 | v_flag);
    } else
        pci_write_config_byte(dev, PIIX_IDE_IO_CONFIG, reg54 & ~v_flag);
  } else {
    const uint8_t mwdma_to_pio[] = { 0, 3, 4 };
    uint8_t pio;

    if (reg48 & u_flag)
      pci_write_config_byte(dev, PIIX_UDMA_CTL, reg48 & ~u_flag);
    if (reg4a & a_speed)
      pci_write_config_word(dev, PIIX_UDMA_TIMING, reg4a & ~a_speed);
    if (reg54 & v_flag)
      pci_write_config_byte(dev, PIIX_IDE_IO_CONFIG, reg54 & ~v_flag);
    if (reg55 & w_flag)
      pci_write_config_byte(dev, PIIX_IDE_IO_CONFIG + 1, (uint8_t) reg55 & ~w_flag);

    pio = mwdma_to_pio[speed - ATA_MWDMA0];
    piix_set_pio_mode(port, dev_id, pio);
    port->devices[dev_id]->pio_mode = pio;
  }
}

static void piix_enable_dma(ata_port_t *port, int dev_id, bool enable)
{
  uint8_t val;
  pci_dev_t *dev = port->host->bus_private;

  pci_read_config_byte(dev, PIIX_UDMA_CTL, &val);
  /*
   * the first four bits enable/disable dma for drives on the primary
   * and the secondary channel
   */
  val |= 1 << (port->num * 2 + 1);
  pci_write_config_byte(dev, PIIX_UDMA_CTL, val);
}


/**
 *  piix_check_450nx  - Check for problem 450NX setup
 *
 *  Check for the present of 450NX errata #19 and errata #25. If
 *  they are found, disable use of DMA IDE
 */
static void piix_check_450nx(void)
{
#if 0
  pci_dev_t dev;
  uint16_t cfg;
  pci_addr_t addr;
  int r;

  r = pci_lookup_prod_vend(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82454NX, &addr, 1);
  if (r <= 0)
    return;

  r = pci_get_dev(&addr, &dev);
  if (r == -1)
    return;

  /* Look for 450NX PXB. Check for problem configurations
     A PCI quirk checks bit 6 already */
  pci_read_config_word(&dev, 0x41, &cfg);
  /* Only on the original revision: IDE DMA can hang */
  if (dev.revision == 0x00)
    no_piix_dma = 1;
  /* On all revisions below 5 PXB bus lock must be disabled for IDE */
  else if (cfg & (1<<14) && dev.revision < 5)
    no_piix_dma = 2;

  if(no_piix_dma)
    printf("[piix]: 450NX errata present, disabling IDE DMA.\n");
  if(no_piix_dma == 2)
    printf("[piix]: A BIOS update may resolve this.\n");
#endif
}
