
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Paul Graphov <graphov@jarios.org>
 * (c) Copyright 2009 Alfeiks Kaanoken <madtirra@jarios.org> (*_GRUB option)
 *
 * PCI bus driver
 *
 */

#include <sys/types.h>
#include <sys/ioport.h>
#include <errno.h>
#include <devio/devio.h>
#include <ns/ns.h>
#include <datalib/usrtc.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <chardev.h>
#include <dircons.h>

#include <pci/pci.h>
#include <pci/names.h>

#include "pci.h"
#include "pci_io.h"
#include "pci_header.h"

#define DRVNAME "pci.d"
#define DEVNAME "/dev/pci"

static int pci_errno = 0;

static pci_bus_t *buses[256];

static status_t _pcid_read(int devno, void *buf, size_t size, size_t offset, void **interim);
static status_t _pcid_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize);

static status_t _pci_read_text(char *msg, size_t size, size_t offset);

static void console_puts(const char *msg) {
  console_printf(msg);
}

static struct device_ops pcid_ops = {
  .read = _pcid_read,
  .write = NULL,
  .ioctl = _pcid_ioctl,
  .log = console_puts,
};

static inline pci_dev_descr_t *_device(pci_addr_t *addr)
{
  pci_bus_t *bus;
  pci_bus_dev_t *bdev;
    
  if(addr->bus > 255 || addr->device > 31 || addr->function > 5) {
    return NULL;
  }
  bus = buses[addr->bus];
  if(!bus) {
    return NULL;
  }
  bdev = bus->devices[addr->device];
  if(!bdev) {
    return NULL;
  }
  return bdev->functions[addr->function];
}

static int _get_cap_list_start(pci_dev_t *dev)
{
  if(!(dev->status & PCI_STATUS_CAP_LIST)) {
    return 0;
  }
  switch(dev->header_type & 0x7f) {
    case PCI_HEADER_TYPE_NORMAL:
    case PCI_HEADER_TYPE_BRIDGE:
      return PCI_CAPABILITY_LIST;
    case PCI_HEADER_TYPE_CARDBUS:
      return PCI_CB_CAPABILITY_LIST;
  }
  return 0;
}

static int _read_capabilities(pci_dev_t *dev)
{
  int pos, ret;
  int cap_num = 0;
  int cap_id;

  dev->num_caps = 0;
  
  pos = _get_cap_list_start(dev);

  while(1) {
    if(cap_num >= PCI_MAX_CAP) {
      break;
    }
    
    ret = pci_read(&dev->addr, pos, 1, &pos);
    
    if(pos < 0x40) {
      break;
    }
    
    ret += pci_read(&dev->addr, pos + PCI_CAP_ID_OFFSET, 1, &cap_id);
    if(cap_id == 0xff) {
      break;
    }
    
    dev->capabilities[cap_num].cap_id = cap_id;
    ret += pci_read(&dev->addr, pos + PCI_CAP_FLAGS_OFFSET, 2,
                    &dev->capabilities[cap_num].cap_flags);

    if(ret != 0) {
      ret = -1;
      break;
    }
    
    pos += PCI_CAP_NEXT_OFFSET;
    cap_num++;
  }

  dev->num_caps = cap_num;
  return ret;
}

                               
static int _get_bar(pci_addr_t *addr, int barno, uint32_t *start, uint32_t *len)
{
  int res;
  uint16_t offset;
  uint32_t mask = ~0U;

  // We write all ones to the register and read it.

  offset = PCI_BAR_0 + 4 * barno;
    
  res = pci_read(addr, offset, 4, start);

  if(res != 0) {
    return -1;
  }
  
  res = pci_write(addr, offset, 4, &mask);

  if(res != 0) {
    return -1;
  }

  res = pci_read(addr, offset, 4, len);

  if(res != 0) {
    return -1;
  }

  res = pci_write(addr, offset, 4, start);

  if(res != 0) {
    return -1;
  }

  return 0;
}

static int _read_dev_conf(pci_addr_t *addr, pci_dev_t *dev)
{
  pci_dev_header_t header;
  int ret;
  int consumed = 0; //number of read BARs
  int produced = 0; //number of filled resources;
  int max_bar;

  ret = pci_generic_read(addr, 0, sizeof(header), &header);

  if(ret != 0) {
    return -1;
  }

  memcpy(&dev->addr, addr, sizeof(*addr));

  dev->ident.device_id = header.device_id;
  dev->ident.vendor_id = header.vendor_id;
  
  dev->ident.subsys_vendor_id = header.subsys_vendor_id;
  dev->ident.subsys_id = header.subsys_id;

  dev->ident.revision_id = header.revision_id;
  
  dev->ident.class_id = header.class_id;
  dev->ident.subclass_id = header.subclass_id;
  dev->ident.prog_if_id = header.prog_if_id;

  dev->header_type = header.header_type;
  dev->status = header.status;

  if((dev->header_type & 0x7f) == PCI_HEADER_TYPE_NORMAL) {
    max_bar = 6;
  } else if((dev->header_type & 0x7f) == PCI_HEADER_TYPE_CARDBUS) {
    max_bar = 1;
  } else if((dev->header_type & 0x7f) == PCI_HEADER_TYPE_BRIDGE) {
    max_bar = 2;
  } else {
    pci_errno = PCI_ERR_UNKNOWN_HEADER;
    goto error;
  }
  //TODO: add sanity check if somehow ones were written.
  
  // Now the sweetest part - resources.
  for(consumed = 0; consumed < max_bar; consumed++) {
    uint32_t s32, l32;
    if(header.base_address_registers[consumed] == 0) {
      continue;
    }

    ret = _get_bar(addr, consumed, &s32, &l32);
    if(ret != 0) {
      pci_errno = PCI_ERR_BAR_READ;
      goto error;
    }
    
    if(s32 & PCI_BAR_IO) {
      // it's io port resource
      s32 &= PCI_BASE_ADDRESS_IO_MASK;
      l32 &= PCI_BASE_ADDRESS_IO_MASK;
      l32 = ~l32;
      dev->resources[produced].start = s32;
      dev->resources[produced].end = s32 + l32;
      dev->resources[produced].flags = PCI_RESOURCE_IO;
    } else if(s32 & PCI_BAR_64) {
      // it's 64 bit mem address
      uint64_t s64, l64;

      //Sanity check;
      if(consumed + 1 >= max_bar) {
        pci_errno = PCI_ERR_WRONG_BAR;
        goto error;
      }

      s64 = s32 & PCI_BASE_ADDRESS_MEM_MASK;
      l64 = l32 & PCI_BASE_ADDRESS_MEM_MASK;

      ret = _get_bar(addr, consumed + 1, &s32, &l32);

      if(ret != 0) {
        pci_errno = PCI_ERR_BAR_READ;
        goto error;
      }

      s64 |= (uint64_t)s32 << 32;
      l64 |= (uint64_t)l32 << 32;
      l64 = ~l64;
      dev->resources[produced].start = s64;
      dev->resources[produced].end = s64 + l64;
      dev->resources[produced].flags = PCI_RESOURCE_MEM | PCI_RESOURCE_MEM64;
      consumed++;
    } else {
      // it's 32 bit mem address
      s32 &= PCI_BASE_ADDRESS_MEM_MASK;
      l32 &= PCI_BASE_ADDRESS_MEM_MASK;
      l32 = ~l32;
      dev->resources[produced].start = s32;
      dev->resources[produced].end = s32 + l32;
      dev->resources[produced].flags = PCI_RESOURCE_MEM;
    }

    produced++;
  }
  
  dev->irq_line = header.interrupt_line;
  dev->num_resources = produced;

  return 0;

  error:

  return -1;
}

void scan_bus(int bn)
{  
  pci_addr_t addr;
  pci_bus_t *bus;
  uint8_t dn, fn;
    
  if(buses[bn]) {
    console_printf("Bus already added\n");
    return;
  }

  memset(&addr, 0, sizeof(addr));
  
  bus = malloc(sizeof(*bus));
  if(!bus) {
    console_printf("Not enough memory.\n");
    abort();
  }

  memset(bus, 0, sizeof(*bus));
  
  buses[bn] = bus;

  for(dn = 0; dn < 32; dn ++) {
    uint8_t multifun = 0;
    for(fn = 0; (fn == 0) || (fn < 8 && multifun ); fn++ ) {
      int ret;
      pci_dev_descr_t *dev;
      uint16_t vendor;

      addr.bus = bn;
      addr.device = dn;
      addr.function = fn;

      ret = pci_read(&addr, PCI_VENDOR_ID, 2, &vendor);
      if(ret < 0 || vendor == 0xffff) {
        // no device
        continue;
      }

      dev = malloc(sizeof(*dev));
      if(!dev) {
        console_printf("Not enough memory, aborting.\n");
        abort();
      }

      ret = _read_dev_conf(&addr, &dev->device);
      if(ret != 0) {
        console_printf("failed to get device, aborting, error code = %d\n", pci_errno);
        abort();
      }

      ret = _read_capabilities(&dev->device);
      if(ret != 0) {
        console_printf("failed to get device capabilities, aborting.\n");
        abort();
      }      

      if(fn == 0) {
        pci_bus_dev_t *pbd;
        pbd = malloc(sizeof(*pbd));
        if(!pbd) {
          console_printf("Not enough memory.\n");
          abort();
        }
        memset(pbd, 0, sizeof(*pbd));
        buses[bn]->devices[dn]=pbd;
      }

      buses[bn]->devices[dn]->functions[fn] = dev;

      if(fn == 0) {
        multifun = dev->device.header_type & 0x80;
      }
      
      if((dev->device.header_type & 0x7f) == PCI_HEADER_TYPE_BRIDGE ||
         (dev->device.header_type & 0x7f) == PCI_HEADER_TYPE_CARDBUS) {
        // we have discovered another bus.
        uint8_t sec_bus;
        pci_read(&dev->device.addr, PCI_SECONDARY_BUS, 1, &sec_bus);
        scan_bus(sec_bus);
      }
    }
  }
}

static int _device_presents(pci_addr_t *dev)
{
  if(!buses[dev->bus] ||
     !buses[dev->bus]->devices[dev->device] ||
     !buses[dev->bus]->devices[dev->device]->functions[dev->function]) {
    return false;
  }
  return true;
}

static status_t _pcid_read(int devno, void *buf, size_t size, size_t offset, void **interim)
{
  return _pci_read_text((char*)buf, size, offset);
}

static status_t _ioctl_reg_op(pci_reg_op_t *r, int op)
{
  status_t ret;
  
  if(r->offset % r->size) {
    return -EINVAL;
  }
  if(r->size != 1 && r->size != 2 && r->size != 4) {
    return -EINVAL;
  }

  if(!_device_presents(&r->addr)) {
    return -ENODEV;
  }
  
  if(op == _PCI_READ) {
    ret = pci_read(&r->addr, r->offset, r->size, &r->value);
  } else {
    ret = pci_write(&r->addr, r->offset, r->size, &r->value);
  }
  
  if(ret != 0) {
    return -EIO;
  } else {
    return 0;
  }
}

static status_t _ioctl_lookup(char *buf, size_t size, pci_ident_t *ident)
{
  int bn, dn, fn;
  int nd = 0;
  pci_addr_t *devs = (pci_addr_t *)buf;
  int maxd = size / sizeof(*devs);

  for(bn = 0; bn < 256; bn++ ) {
    if(!buses[bn]) continue;
    for(dn = 0; dn < 32; dn++ ) {
      if(!buses[bn]->devices[dn]) continue;
      for(fn = 0; fn < 8; fn++ ) {
        pci_dev_descr_t *dev = buses[bn]->devices[dn]->functions[fn];

        if(!dev) continue;

        if((!(ident->lookup_opt & PCI_LOOKUP_V) || (ident->vendor_id == dev->device.ident.vendor_id)) &&
           (!(ident->lookup_opt & PCI_LOOKUP_P) || (ident->device_id == dev->device.ident.device_id))) {
          if(buf) {
            devs[nd].bus = bn;
            devs[nd].device = dn;
            devs[nd].function = fn;
            nd++;
            if(nd >= maxd) {
              goto out;
            }
          } else {
            nd++;
          }
        }
      }
    }
  }
  out:
  if(buf) {
    return nd * sizeof(*devs);
  } else {
    return nd;
  }
}

static status_t _ioctl_read_conf(void *buf, size_t size, pci_addr_t *addr)
{
  int ret;
  if(size > 256) return -ERANGE;

  if(!_device(addr)) {
    return -ENODEV;
  }

  ret = pci_generic_read(addr, (uint16_t)0, (uint16_t)size, buf);
  
  if(ret == 0) {
    return size;
  } else {
    return -EIO;
  }
}

static status_t _ioctl_get_dev(pci_addr_t *addr, pci_dev_t *buf)
{
  pci_dev_descr_t *dev = _device(addr);
  if(!dev) {
    return -ENODEV;
  }
  memcpy(buf, dev, sizeof(*dev));
  return sizeof(*dev);
}

static status_t _pcid_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize)
{
  if(type != PCI_IOCTL) {
    return -ENOSYS;
  }

  switch(nr) {
    case PCI_READ_REG:
    case PCI_WRITE_REG:
    {
      pci_reg_op_t *op = (pci_reg_op_t *)buf;
      
      if(size < sizeof(pci_reg_op_t)) {
        return -EINVAL;
      }
      
      if(nr == PCI_READ_REG) {
        if(osize < sizeof(pci_reg_op_t *)) {
          return -EINVAL;
        }
        return _ioctl_reg_op(op, _PCI_READ);
        
      } else if(nr == PCI_WRITE_REG) {
        return _ioctl_reg_op(op, _PCI_WRITE);
        
      } else {
        return -ENOSYS;
      }
      break;
    }
    case PCI_LOOKUP:
    case PCI_LOOKUP_CNT:
    {
      pci_ident_t *ident = (pci_ident_t *) buf;
      
      if(size < sizeof(pci_ident_t)) {
        return -EINVAL;
      }
      
      switch(nr) {
        case PCI_LOOKUP:
          return _ioctl_lookup(obuf, osize, ident);
          
        case PCI_LOOKUP_CNT:
          return _ioctl_lookup(NULL, 0, ident);
          
        default:
          return -ENOSYS;
      }
      break;
    }
    case PCI_READ_CONF:
    {
      pci_addr_t *dev = (pci_addr_t *)buf;
      if(size < sizeof(pci_addr_t)) {
        return -EINVAL;
      }
      return _ioctl_read_conf(obuf, osize, dev);
      break;
    }
    case PCI_GET_DEV:
    {
      return _ioctl_get_dev((pci_addr_t *)buf, (pci_dev_t *)obuf);
    }
    default:
      return -ENOSYS;
  }
}

#define START_ALLOC 4096

#define addline(...)                                                    \
  {                                                                     \
    n = snprintf(&buf[written], allocated - written, __VA_ARGS__);      \
    if(allocated - written - n <= 1) {                                  \
      allocated *= 2;                                                   \
      new_buf = realloc(buf, allocated);                                \
      if(!new_buf) {                                                    \
        goto out;                                                       \
      }                                                                 \
      buf = new_buf;                                                    \
      n = snprintf(&buf[written], allocated - written, __VA_ARGS__);    \
    }                                                                   \
    if(allocated - written - n <= 1) {                                  \
      goto out;                                                         \
    }                                                                   \
    written += n;                                                       \
  }

static char *buf = NULL;
static int written = 0;

static status_t _pci_gen_text(void);

static status_t _pci_read_text(char *msg, size_t size, size_t offset)
{
  int len;
  
  if(!buf) {
    _pci_gen_text();
  }

  if(!buf) {
    return -EIO;
  }

  if(offset == written) {
    return 0;
  } else if(offset > written) {
    return -EINVAL;
  }
  
  len = MIN(size, written - offset);

  memcpy(msg, buf + offset, len);
  return len;
}

static status_t _pci_gen_text(void)
{
  static int allocated;
  int bn, dn, fn;
  int n;
  char *new_buf;
  
  vendor_desc_t *vendor;
  product_desc_t *product;
  subprod_desc_t *subprod;

  class_desc_t *class;
  subclass_desc_t *subclass;
  prog_if_desc_t *prog_if;

  if(!buf) {
    buf = malloc(START_ALLOC);
    if(!buf) {
      goto out;
    }
    allocated = START_ALLOC;
  }

  written = 0;
 
  for(bn = 0; bn < 256; bn++ ) {
    if(!buses[bn]) continue;
    for(dn = 0; dn < 32; dn++ ) {
      if(!buses[bn]->devices[dn]) continue;
      for(fn = 0; fn < 8; fn++ ) {
        pci_dev_descr_t *dev = buses[bn]->devices[dn]->functions[fn];
        if(!dev) continue;
        addline("bus %3hhd device %2hhd function %hhd (0x%04X:0x%04X)\n", bn, dn, fn,
                dev->device.ident.vendor_id, dev->device.ident.device_id);
        vendor = lookup_vendor(dev->device.ident.vendor_id);
        if(!vendor) goto classes;
        addline("\tVendor: %s\n", vendor->name);
        product = lookup_product(vendor, dev->device.ident.device_id);
        if(!product) {
          addline("\tUnknown product.\n");
          goto classes;
        }
        addline("\tProduct: %s\n", product->name);
        subprod = lookup_subprod(product, dev->device.ident.subsys_id, dev->device.ident.subsys_vendor_id);
        if(!subprod) goto classes;
        addline("\tSubproduct: %s\n", subprod->name);
        classes:
        class = lookup_class(dev->device.ident.class_id);
        if(!class) {
          addline("\tUnknown device class\n");
          continue;
        }
        addline("\tClass: %s\n", class->name);
        subclass = lookup_subclass(class, dev->device.ident.subclass_id);
        if(!subclass) {
          addline("\tUnknown device subclass\n");
          continue;
        }
        addline("\tSubclass: %s\n", subclass->name);
        prog_if = lookup_prog_if(subclass, dev->device.ident.prog_if_id);
        if(!prog_if) continue;
        addline("\tProgramming interface: %s\n", prog_if->name);
      }
    }
  }
  return written+1;

  out:
  free(buf);
  buf = NULL;

  return 0;
}

int main(int argc, char **argv)
{
  int r;

#ifndef CONFIG_PCID_GRUB
  start_service();
#endif

  open_console();

  console_printf("[pci.d] Jari OS PCI bus driver.\n");

  r = init_pci_access();

  if(r != 0) {
    console_printf("pci: failed to initialize pci access, %s, aborting.\n",
		     (r == 1)?"failed to find appropriage access method":
		     "unable to get pci port access from kernel");
    abort();
  }

  scan_bus(0);

  init_simple_device_server(16384, 16384, DRVNAME, DEVNAME, &pcid_ops);
  if(r != 0) {
    console_printf("Failed to init server.\n");
    abort();
  }

  set_loglevel(1);

  device_server_start();

  abort();
  return -1;
}
