/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Paul Graphov <graphov@jarios.org> 
 *
 * PCI direct access through IO ports
 *
 */

#include <sys/types.h>
#include <sys/ioport.h>
#include <stdio.h>
#include <pci/pci.h>

#include "pci_io.h"

static pci_methods_t *methods;

static int pci_conf1_detect( void ) {
  uint32_t tmp;
  int res = -1;
  
  outb(PCI_CONFIG_ADDRESS, 0x01);
  tmp = inl(PCI_CONFIG_ADDRESS);
  outl(PCI_CONFIG_ADDRESS, 0x80000000);
  if (inl(PCI_CONFIG_ADDRESS) == 0x80000000) {
    res = 0;
  }
  outl(PCI_CONFIG_ADDRESS, tmp);

  return res;
}

static int pci_conf1_read(pci_addr_t *d, uint16_t pos, uint8_t len, void *buf){
  
  uint16_t port;
  uint32_t addr;
  
  if(d->bus > 255 || d->device > 31 || d->function > 5 || pos > 4096) {
    return -1;
  }
  

  port = PCI_CONFIG_DATA + (pos & 3);
  addr = PCI_CONF1_MAKE_ADDRESS(d->bus, d->device, d->function, pos); 

  outl(PCI_CONFIG_ADDRESS, addr);

  switch(len) {
  case 1:
    ((uint8_t *)buf)[0] = inb( port );
    break;
  case 2:
    ((uint16_t *)buf)[0] = inw( port );
    break;
  case 4:
    ((uint32_t *)buf)[0] = inl( port );
    break;
  default:
    return -1;
  }
  return 0;
}

static int pci_conf1_write(pci_addr_t *d, uint16_t pos, uint8_t len, void *buf)
{
	uint16_t port;
  uint32_t addr;

  if(d->bus > 255 || d->device > 31 || d->function > 5 || pos > 4096) {
    return -1;
  }

  
  port = PCI_CONFIG_DATA + (pos & 3);
  addr = PCI_CONF1_MAKE_ADDRESS(d->bus, d->device, d->function, pos);
  
  outl(PCI_CONFIG_ADDRESS, addr);
  
  
	switch (len) {
    case 1:
      outb(port, ((uint8_t *)buf)[0]);
      break;
    case 2:
      outw(port, ((uint16_t *)buf)[0]);
      break;
    case 4:
      outl(port, ((uint32_t *)buf)[0]);
      break;
    default:
      return -1;
	}
  
	return 0;
}

static int pci_conf2_detect(void)
{
	outb(PCI_CONFIG_ADDRESS+3, 0);
	outb(PCI_CONFIG_ADDRESS, 0);
	outb(PCI_CONFIG_ADDRESS+2, 0);
	if(inb(PCI_CONFIG_ADDRESS) == 0x00 && inb(PCI_CONFIG_ADDRESS) == 0x00)
		return 0;
	else
		return -1;
}

static int pci_conf2_read(pci_addr_t *d, uint16_t pos, uint8_t len, void *buf)
{
	uint32_t addr = PCI_CONF2_MAKE_ADDRESS(d->device, pos);

	if (pos >= 255)
		return -1;

	if (d->device >= 16)
		/* conf2 supports only 16 devices per bus */
		return -1;
  
	outb(PCI_CONFIG_ADDRESS, (d->function << 1) | 0xF0);
	outb(PCI_CONFIG_ADDRESS+2, d->bus);
  
	switch (len) {
    case 1:
      ((uint8_t *)buf)[0] = inb(addr);
      break;
    case 2:
      ((uint16_t *)buf)[0] = inw(addr);
      break;
    case 4:
      ((uint32_t *)buf)[0] = inl(addr);
      break;
    default:
      outb(PCI_CONFIG_ADDRESS, 0);
      return -1;
	}
	outb(PCI_CONFIG_ADDRESS,0);
	return 0;
}

static int pci_conf2_write(pci_addr_t *d, uint16_t pos, uint8_t len, void *buf)
{
	uint32_t addr = PCI_CONF2_MAKE_ADDRESS(d->device, pos);

	if (pos > 255)
		return -1;

	if (d->device >= 16)
    return -1;
  
	outb(PCI_CONFIG_ADDRESS, (d->function << 1) | 0xf0);
	outb(PCI_CONFIG_ADDRESS+2, d->bus);
	switch (len) {
	case 1:
		outb(addr, ((uint8_t *)buf)[0]);
		break;
	case 2:
		outw(addr, ((uint16_t *)buf)[0]);
		break;
	case 4:
		outl(addr, ((uint32_t *)buf)[0]);
		break;
	default:
		outb(PCI_CONFIG_ADDRESS, 0);
    return -1;
	}
	outb(PCI_CONFIG_ADDRESS, 0);
	return 0;
}

static pci_methods_t conf1_methods = {pci_conf1_read, pci_conf1_write, pci_conf1_detect};
static pci_methods_t conf2_methods = {pci_conf2_read, pci_conf2_write, pci_conf2_detect};

/*
 * Check if bus 0 contains any valid devices. Not 100% access correctness proof.
 * 0x0000 is a valid vendor (Gammagraphx Inc.) but we will not trust it.
 */

static int sanity_check(pci_methods_t *m, uint8_t nd)
{
  int k, res = 0;
  uint16_t vend;

  pci_addr_t d;

  d.bus = 0;
  d.function = 0;
  
  if(m->detect() == 0) {
    for(k = 0; k < nd; k++ )
    {
      d.device = k;
      if(m->read( &d, 0, 2, &vend ) == 0 && vend != 0 && vend != 0xffff)
      {
        res = 1;
        break;
      }
    }
  }
  if(res) methods = m;
  return res;
}

int init_pci_access(void)
{
  if(alloc_ioport_range( 0xCF8, 8 ) != 0) return 2;
  if(sanity_check(&conf1_methods, 32) == 0) return 0;
  if(sanity_check(&conf2_methods, 16) == 0) return 0;
  return 1;
}

int pci_write(pci_addr_t *d, uint16_t pos, uint8_t len, void *buf)
{
  return methods->write(d, pos, len, buf);
}

int pci_read(pci_addr_t *d, uint16_t pos, uint8_t len, void *buf)
{
  return methods->read(d, pos, len, buf);
}

int pci_generic_read(pci_addr_t *d, uint16_t pos, uint16_t len, void *buf)
{
  int c;
  int step;

  while(len > 0) {
    if((pos & 1) && len >= 1) {
      step = 1;
    } else if((pos & 3) && len >= 2) {
      step = 2;
    } else if(len < 2) {
      step = 1;
    } else if(len < 4) {
      step = 2;
    } else {
      step = 4;
    }

    c = pci_read(d, pos, step, buf);
    
    if( c != 0) {
      return c;
    }
    buf += step;
    pos += step;
    len -= step;
  }
  
  return 0;
}
