/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008,2009 MadTirra <madtirra@jarios.org>
 *
 * vesacons.d: Generic VESA graphical console
 * console.c: main entry and IPC internals
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <wchar.h>
#include <errno.h>
#include <time.h>
#include <datalib/usrtc.h>
#include <ns/ns.h>
#include <sys/scheduler.h>
#include <ipc/rpc.h>
#include <devio/devio.h>
#include <stdio.h>
#include <int10/int10.h>
#include <dircons.h>

#include <sys/ioctl.h>
#include <sys/console.h>

#include "vga.h"
#include "hwops.h"
#include "debug.h"
#include "fb_types.h"
#include "fbcon.h"

#include <chardev.h>

#define DRVNAME  "vesacons.d"
#define DEVNAME  "/dev/console"

static status_t fb_write(int devno, void *buf, size_t size, size_t offset, void **data);
static status_t fb_read(int devno, void *buf, size_t size, size_t offset, void **data);
static status_t fb_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize);

uint16_t VESA_XRES, VESA_YRES, VESA_YRES_CLIENT, VESA_BPL;
uint8_t  VESA_BITS, VESA_RES;
uint8_t  VESA_RED_OFFS, VESA_GREEN_OFFS, VESA_BLUE_OFFS;
uint8_t  VESA_RED_SIZE, VESA_GREEN_SIZE, VESA_BLUE_SIZE;

struct device_ops dev_op = {
  .read = fb_read,
  .write = fb_write,  
  .ioctl = fb_ioctl,
  .log = (void (*)(const char *))fbcon_puts,
};

int main(int argc,char **argv)
{
  int r=0;
  char buff[256];
  vbe_ctrl_t *ctrl_info=NULL;
  vbe_mode_t *mode_info=NULL;
  unsigned vx=0,vy=0;
  int g=0,bpp,i,btype;
  void *vmem;
  void *vbuf;
  uint16_t *pix;
  uintptr_t next_super_offs;

#ifndef CONFIG_VESACONS_GRUB
  start_service();
#endif

  g = int10_init(); /* init x86emu for int10 */
  if(g) {
    abort();
  }

  g = int10_set_vbe_mode(0x114, &ctrl_info, &mode_info);

  if(mode_info != NULL && g == 0) {
    /* fill out VESA parameters */
    VESA_YRES        = mode_info->y_resolution;
    VESA_XRES        = mode_info->x_resolution;
    VESA_BITS        = mode_info->bits_per_pixel;
    VESA_BPL         = mode_info->bytes_per_scanline;
    VESA_RES         = 0;
    VESA_RED_OFFS    = mode_info->red_field_position;
    VESA_RED_SIZE    = mode_info->red_mask_size;
    VESA_GREEN_OFFS  = mode_info->green_field_position;
    VESA_GREEN_SIZE  = mode_info->green_mask_size;
    VESA_BLUE_OFFS   = mode_info->blue_field_position;
    VESA_BLUE_SIZE   = mode_info->blue_mask_size;
    /* try to map */
    bpp=(VESA_BITS + 1)/8;

    if(ctrl_info->total_memory > 0) {
      vbuf = mmap(NULL, ctrl_info->total_memory << 16, PROT_READ|PROT_WRITE,
                  MAP_ANON|MAP_PHYS|MAP_PRIVATE, NOFD, mode_info->phys_base);


      if(vbuf == MAP_FAILED) {
        abort();
      }
      /* try to pan */
      next_super_offs=VESA_YRES * VESA_BPL;
      vx = (next_super_offs % VESA_BPL) / bpp;
      vy = (next_super_offs / VESA_BPL) - VESA_YRES_CLIENT;
      if(int10_pan(&vx, &vy)==0) {
        vmem = vbuf + (vy * VESA_BPL + vx*bpp);
      } else {
        vmem = vbuf;
      }
    } else {
      vbuf = mmap(NULL, VESA_YRES * VESA_XRES * bpp, PROT_READ|PROT_WRITE,
                  MAP_ANON|MAP_PHYS|MAP_PRIVATE, NOFD, mode_info->phys_base);
      if(vbuf == MAP_FAILED) {
        abort();
      }
      vmem=vbuf;
    }

    /* to ensure on it fill */
    pix=(uint16_t*)vmem; i=0;
    /* initialize framebuffer console */
    btype=fbcon_bpptype(bpp,VESA_RED_OFFS,VESA_GREEN_OFFS,VESA_BLUE_OFFS,
                        VESA_RED_SIZE,VESA_GREEN_SIZE,VESA_BLUE_SIZE);

    i=fbcon_init(vmem,VESA_XRES,VESA_YRES,bpp,btype);

  } else {
    abort();
  }

  int10_done(); /* finished up x86emu */

  task_scheduler_control(0,0, SYS_SCHED_CTL_SET_PRIORITY, 30);


  r=init_simple_device_server(65536, 65536, DRVNAME, DEVNAME, &dev_op);

  set_loglevel(1);

  if(r!=0) {
    sprintf(buff, "[vesacons] device service init failed (%d)\n",r);
    fbcon_puts(buff);
    abort();
  }

  fbcon_puts("[vesacons] Generic VBE framebuffer console.\n");

  device_server_start();

  fbcon_puts("[vesacons] :((\n");

  return 0;
}

static status_t fb_write(int devno, void *buf, size_t size, size_t offset, void **data)
{
  fbcon_print((cschar_t*)buf, size);
  return size;
}

void ascii_puts(const char *msg)
{
  return;
}

static status_t fb_read(int devno, void *buf, size_t size, size_t offset, void **data)
{
  return -ENOSYS;
}

static status_t fb_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize)
{
  
  if(type != CON_IOCTL) return -ENOSYS;
  
//  if(size < get_ioctl_size(CON_IOCTL, nr)) return -EINVAL;
  //TODO: add sane size checking

  switch(nr) {
    case PHYT_DISABLEAUTOSCROLL:
      fbcon_scrolldisable();
      break;
    case PHYT_ENABLEAUTOSCROLL:
      fbcon_scrollenable();
      break;
    case PHYT_GETSIZE:
      ((coord_t*)obuf)->x=fbcon_width();
      ((coord_t*)obuf)->y=fbcon_height();
      break;
    case PHYT_CLEAR:
      fbcon_clear_rect(((clear_desc_t *)buf)->rect.x1, ((clear_desc_t *)buf)->rect.y1,
                       ((clear_desc_t *)buf)->rect.x2, ((clear_desc_t *)buf)->rect.y2,
                       ((clear_desc_t *)buf)->color);
      break;
    case PHYT_SCROLL:
      fbcon_scroll_area(((scroll_desc_t *)buf)->startline, ((scroll_desc_t *)buf)->stopline,
                      ((scroll_desc_t *)buf)->lines);
      break;
    case PHYT_SETCURPOSITION:
      fbcon_setcursor(((coord_t *)buf)->x, ((coord_t *)buf)->y);
      break;
    case PHYT_SHOWCURSOR:
      fbcon_setcursorvisibility(1);
      break;
    case PHYT_HIDECURSOR:
      fbcon_setcursorvisibility(0);
      break;
    default:
      return -EINVAL;
      break;
  }

  return 0;
}
