/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008,2009 MadTirra <madtirra@jarios.org>
 *
 * vesacons.d: Generic VESA graphical console
 * fb_primitives.c: framebuffer primitives
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <wchar.h>
#include <errno.h>
#include <time.h>
#include <datalib/usrtc.h>
#include <ns/ns.h>
#include <sys/scheduler.h>
#include <ipc/rpc.h>
#include <devio/devio.h>
#include <stdio.h>
#include <stdbool.h>
#include <int10/int10.h>
#include <pthread.h>
#include <arch/atomic.h>

#include <sys/kcontrol.h>

#include "vga.h"
#include "hwops.h"
#include "fb_types.h"

#include "default_font.h"

#include "pcf.h"

#include "debug.h"

#include <chardev.h>

/* local functions - color representation */
/* 32-bit bpp */
static void col_rgb0888(void *dst,int col);
static int rgb0888_col(void *src);
static void bgr_col0888(void *dst,int rgb);
static int col0888_bgr(void *src);

/* 24-bit bpp */
static void rgb_col888(void *dst, int rgb);
static int col888_rgb(void *src);

/* 16-bit bpp (5:5:5) */
static int col555_rgb(void *src);
/* 16-bit bpp (5:6:5) */
static void rgb_col565(void *dst, int rgb);
static int col565_rgb(void *src);

/* general pixel output */
static inline void putpixel(viewport_t *vp,int x,int y,int color);
static inline int fgcc2col(viewport_t *vp,uint8_t col);
static inline int bgcc2col(viewport_t *vp,uint8_t col);
static inline void draw_rect(viewport_t *vp,int x,int y,int width,int height,
			     int col);
static inline void sync_rect(viewport_t *vp,int sx,int sy,int width,int height);

/* tricks */
static inline void invertpixel(viewport_t *vp,int x,int y);
static inline int getpixel(viewport_t *vp,int x,int y);

static void __vp_scroll_area(viewport_t *vp,int lines, int startline, int stopline);

void *pending_repainter(void *vp);

int fbfont_init(fbfont_t *p,int width,int sc_line)
{
  p->width=width;
  p->scan_lines=sc_line;
  p->font=malloc(0xff*sizeof(uintptr_t));
  memset(p->font, 0, 0xff*sizeof(uintptr_t));
  if(!p->font)
    return ENOMEM;
  
  return 0;
}

int fbfont_settablefont(fbfont_t *p,int tbl,void *data)
{
  if(!p || !p->font)
    return EINVAL;
  if(tbl>255)
    return EINVAL;
  
  p->font[tbl]=data;
  
  return 0;
}

void schedule_repaint(viewport_t *vp)
{
  if(vp->pending_repaint) return;
  vp->pending_repaint = 1;
}

#define PAD(a,b) (((a) + (b) - 1) & (~((b) - 1)))

static void fbfont_draw_bitmap(viewport_t *vp, uint8_t *bitmap_data, int width, int height,
                               int padding, int x, int y, int opaque, int color, int bgcolor, int rbgc)
{
  int dx, dy;
  uint8_t *scanline;
  int pwidth = PAD((width + 7)/8, padding);
  
  for(dy = 0; dy < height; dy++) {
    scanline = &bitmap_data[dy * pwidth];
    for(dx = 0; dx < width; dx++) {
      if(scanline[dx / 8] & (1 << (7 - (dx % 8)))) {
        putpixel(vp, x + dx, y + dy, color);
      } else if(bgcolor != 0) {
        putpixel(vp, x + dx, y + dy, bgcolor);
      } else if(opaque) {
        putpixel(vp, x + dx, y + dy, rbgc);
      }
    }
  }
}

int fbfont_put(viewport_t *vp,fbfont_t *fnt,int sx,
	       int sy,cschar_t c,int opaque)
{
  int color, bgcolor, rbgc;
  unsigned char *font;
  gly_char_t gc;
  uint8_t *bitmap;
  
  if(!vp || !fnt)
    return EINVAL;
  
  /* check for overflow */
  if(sx>vp->xres-fnt->width)
    return ERANGE;
  if(sy>vp->yres-fnt->scan_lines)
    return ERANGE;
  
  gc.s = c.ch;
  color = fgcc2col(vp, c.cattr);
  bgcolor = bgcc2col(vp, c.cattr);
  rbgc = bgcc2col(vp, 0);

  font=fnt->font[gc.p1];
  if(!font) { /* no font found */
    gc.s = L'?';
    font=fnt->font[0];
  }
  
  if(gc.s == L' ') {
    draw_rect(vp,sx,sy,fnt->width,fnt->scan_lines,rbgc);
    goto __fnt_putexit;
  }

  if(gc.p1 > 0) {
    goto __fnt_putexit;
  }
  
  bitmap = &font[gc.p0*fnt->scan_lines];

  fbfont_draw_bitmap(vp, bitmap, fnt->width, fnt->scan_lines, 1, sx, sy, opaque, color, bgcolor, rbgc);

  __fnt_putexit:
  
  if(!vp->pending_repaint) {
    sync_rect(vp,sx,sy,fnt->width,fnt->scan_lines);
  }

  return 0;
}

int pcf_put(viewport_t *vp,pcf_font_t *fnt,int sx,
	       int sy,cschar_t c,int opaque)
{
  int color, bgcolor;
  int rbgc;
  int index;
  uint8_t *bitmap;
  
  if(!vp || !fnt)
    return EINVAL;
  
  if(sx>vp->xres-vp->font_width)
    return ERANGE;
  
  if(sy>vp->yres-vp->font_height)
    return ERANGE;
  
  color = fgcc2col(vp, c.cattr);
  bgcolor = bgcc2col(vp, c.cattr);
  rbgc = bgcc2col(vp, 0);

  index = pcf_get_char_index(fnt, (uint16_t) c.ch);

  bitmap = pcf_get_bitmap(fnt, index);

  if(!bitmap) {
    goto out;
  }  
  
  fbfont_draw_bitmap(vp, bitmap, vp->font_width, vp->font_height, 4, sx, sy, opaque, color, bgcolor, rbgc);

  out:
  
  if(!vp->pending_repaint) {
    sync_rect(vp, sx, sy, vp->font_width, vp->font_height);
  }

  return 0;
}


void fbfont_putcursor(viewport_t *vp,fbfont_t *fnt,int sx,int sy)
{
  int y,x;

  for(y=0;y<vp->font_height;y++) {
    for(x=0;x<vp->font_width;x++) {
      invertpixel(vp,sx+x,sy+y);
    }
  }

  sync_rect(vp,sx,sy,vp->font_width,vp->font_height);
}

/* view port functions */

/* local functions */
/* terminal color converting */
static int __convfg_color16(viewport_t *vp,uint8_t col);
static int __convbg_color16(viewport_t *vp,uint8_t col);

/* scrolling */
static void __vp_scroll(viewport_t *vp,int lines);

/* pixels related */
static void __putpix16(viewport_t *vp,int x,int y,int color);

static inline void putpixel(viewport_t *vp,int x,int y,int color)
{
  vp->vc_rgb2col(&vp->bg[(y*vp->xres)*vp->bpp+x*vp->bpp],color);
}

static inline void invertpixel(viewport_t *vp,int x,int y)
{
  putpixel(vp,x,y,~getpixel(vp,x,y));
}

static inline int getpixel(viewport_t *vp,int x,int y)
{
  return vp->vc_col2rgb(&vp->bg[(y*vp->xres)*vp->bpp+x*vp->bpp]);
}

static inline void sync_all(viewport_t *vp)
{
  memcpy(vp->marea,vp->bg,vp->xres*vp->yres*vp->bpp);
}

static inline void sync_rect(viewport_t *vp,int sx,int sy,int width,int height)
{
  int y;
  int scan=vp->bpp*vp->xres;

  for(y=0;y<height;y++) 
    memcpy(vp->marea+(sy+y)*scan+(sx)*vp->bpp,vp->bg+(sy+y)*scan+(sx)*vp->bpp,width*vp->bpp);
}

int viewport_init(viewport_t *vp,void *marea,int xres,int yres,int bpp,
		  int bpptype)
{
  pthread_t rep_thr;

  if(!vp)
    return EINVAL;

  /* setup all entries */
  vp->marea=marea;
  vp->xres=xres;
  vp->yres=yres;
  vp->bpp=bpp;
  /* font related */
  vp->curx=0;
  vp->cury=0;
  vp->general_font=NULL;
  vp->bg=NULL;
  vp->disable_scroll=false;
  vp->cur_show=true;
  vp->pending_repaint = 0;

//  if(pthread_mutex_init(&vp->vp_lock, NULL) != 0) return EINVAL;
  
  vp->vc_scroll=__vp_scroll;
  vp->vc_scroll_area=__vp_scroll_area;
  
  switch(bpptype) {
    case RGB_555: /* 16 high color */
      vp->vc_rgb2col=rgb_col565;
      vp->vc_col2rgb=col555_rgb;
      vp->ppixel_fg=__putpix16;
      vp->vc_bcconvert=__convbg_color16;
      vp->vc_fcconvert=__convfg_color16;
      vp->vc_rgb2col(&vp->bgcol,DEFAULT_BGCOLOR);
      vp->vc_rgb2col(&vp->fgcol,DEFAULT_FGCOLOR);
      break;
    case RGB_565:
      vp->vc_rgb2col=rgb_col565;
      vp->vc_col2rgb=col565_rgb;
      vp->ppixel_fg=__putpix16;
      vp->vc_bcconvert=__convbg_color16;
      vp->vc_fcconvert=__convfg_color16;
      vp->vc_rgb2col(&vp->bgcol,DEFAULT_BGCOLOR);
      vp->vc_rgb2col(&vp->fgcol,DEFAULT_FGCOLOR);
      break;
    case RGB_888:
      vp->vc_rgb2col=rgb_col888;
      vp->vc_col2rgb=col888_rgb;
      vp->ppixel_fg=__putpix16;
      vp->vc_bcconvert=__convbg_color16;
      vp->vc_fcconvert=__convfg_color16;
      vp->vc_rgb2col(&vp->bgcol,DEFAULT_BGCOLOR);
      vp->vc_rgb2col(&vp->fgcol,DEFAULT_FGCOLOR);
      break;
    case RGB_8880:
      break;
    case RGB_0888:
      break;
    case BGR_0888:
      break;
    default:
      return ERANGE;
  }

  vp->bg=malloc(vp->xres*vp->yres*vp->bpp);

  if(vp->bg == NULL)
    return ENOMEM;

  if(pthread_create(&rep_thr, NULL, pending_repainter, vp)) {
    return EFAULT;
  }
  
  return 0;
}

int viewport_setfont(viewport_t *vp,fbfont_t *font)
{
  if(!font)
    return EINVAL;

  vp->general_font=font;
//  vp->xxf=vp->xres/font->width;
//  vp->yyf=vp->yres/font->scan_lines;

  return 0;
}

void viewport_set_pcf_font(viewport_t *vp,pcf_font_t *font)
{
  vp->pcf_font = font;
  // vp->xxf=vp->xres/font->width;
  //vp->yyf=vp->yres/font->scan_lines;
}

void viewport_set_hw(viewport_t *vp)
{
  vp->xxf=vp->xres/vp->font_width;
  vp->yyf=vp->yres/vp->font_height;
}


void viewport_put(viewport_t *vport,int sx,int sy,cschar_t gly,int opaque)
{
  int x,y;

  if(sx>vport->xxf-1)    return;
  if(sy>vport->yyf-1)    return;

  x=sx*vport->font_width;
  y=sy*vport->font_height;

  if(vport->font_type == BUILTIN) {
    fbfont_put(vport,vport->general_font,x,y,gly,opaque);
  } else {
    pcf_put(vport, vport->pcf_font, x, y, gly, opaque);
  }

//  sync_rect(vport, x, y, vport->general_font->width, vport->general_font->scan_lines);

  return;
}

void viewport_showcursor(viewport_t *vp,int sx,int sy)
{
  int x,y;

  if(sx>vp->xxf-1)    return;
  if(sy>vp->yyf-1)    return;

  x=sx*vp->font_width;
  y=sy*vp->font_height;

  fbfont_putcursor(vp,vp->general_font,x,y);
}

void viewport_putc(viewport_t *vp,cschar_t gly)
{
  cschar_t spc;

  switch(gly.ch) {
  case '\n':
    vp->curx=0;
    vp->cury++;
    break;
  case BACKSPACE:
    vp->curx--;
    if(vp->curx<0 && vp->cury>0) {
      vp->curx=vp->xxf-1;
      vp->cury--;
    } 
    spc.ch=L' ';
    spc.cattr=0;
    spc.xattr=0;
    viewport_put(vp,vp->curx,vp->cury,spc,1);
    return;
    break;
  case TAB:
    vp->curx+=8;
    break;
  case L'\r':
    vp->curx=0;
    break;
  case L' ':
    spc.ch=L' ';
    spc.cattr=0;
    spc.xattr=0;
    viewport_put(vp,vp->curx,vp->cury,spc,1);
    vp->curx++;
    break;
  default:
    viewport_put(vp,vp->curx,vp->cury,gly,0);
    vp->curx++;
    break;
  }
  if(vp->disable_scroll==false && vp->curx>vp->xxf-1) {
    vp->curx=0;
    vp->cury++;
  } else if(vp->curx>vp->xxf-1 && vp->disable_scroll==true)
    vp->curx=vp->xxf;

  if(vp->cury>vp->yyf-1) {
    vp->cury--;
    /* scroll */
    if(vp->disable_scroll==false)
      vp->vc_scroll(vp,vp->font_height);
  }
//  sync_all(vp);
  return;
}

void *pending_repainter(void *data)
{
  viewport_t *vp = (viewport_t *)data;
  while(1) {
    usleep(30000);
    if(vp->pending_repaint){
      sync_all(vp);
      vp->pending_repaint = 0;
    }
  }
}

void viewport_putcxy(viewport_t *vp,cschar_t gly,int *x,int *y)
{
  cschar_t spc;
  int curx=*x,cury=*y;

  switch(gly.ch) {
  case L'\n':
    curx=0;
    cury++;
    break;
  case BACKSPACE:
    curx--;
    if(curx<0 && cury>0) { /* back */
      curx=vp->xxf-1;
      cury--;
    } else
      curx--;
    spc.ch=L' ';
    spc.cattr=0;
    spc.xattr=0;
    viewport_put(vp,curx,cury,spc,1);
    *x=curx; *y=cury;
    return;
    break;
  case L' ':
    spc.ch=L' ';
    spc.cattr=0;
    spc.xattr=0;
    viewport_put(vp,curx,cury,spc,1);
    curx++;
    break;
  default:
    viewport_put(vp,curx,cury,gly,0);
    curx++;
    break;
  }

  if(vp->disable_scroll==false && curx>vp->xxf-1) {
    curx=0;
    cury++;
  } else if(curx>vp->xxf-1 && vp->disable_scroll==true)
    curx=vp->xxf;

  if(cury>vp->yyf-1) {
    cury--;
    /* scroll */
    if(vp->disable_scroll==false)
      vp->vc_scroll(vp,vp->font_height);
  }

  *x=curx; *y=cury;

  return;
}

void viewport_drawpixmap(viewport_t *vp,uint32_t *pixbuf,
			 int sx,int sy,int width,int height)
{
  int i,j;

  for(i=0;i<height;i++) {
    for(j=0;j<width;j++) {
      putpixel(vp,sx+j,sy+i,pixbuf[i*width+j]);
    }
  }

  sync_rect(vp,sx,sy,width,height);
}

void viewport_clear_rect(viewport_t *vp, int x1, int y1, int x2, int y2, uint8_t col)
{
  int color;
  color=bgcc2col(vp,col);

  //sanity check; we don't fix coordinates misplacement, only fatal errors prevented.
  if(x1 < 0 || x2 < 0 || y1 < 0 || y2 < 0) return;
  if(x1 > vp->xxf-1 || x2 > vp->xxf-1) return;
  if(y1 > vp->yyf-1 || y2 > vp->yyf-1) return;
  if(y1 > y2 || x1 > x2) return;

  draw_rect(vp, x1 * vp->font_width, y1 * vp->font_height,
            (x2-x1+1) * vp->font_width, (y2-y1+1) * vp->font_height, color);
  sync_rect(vp, x1 * vp->font_width, y1 * vp->font_height,
            (x2-x1+1) * vp->font_width, (y2-y1+1) * vp->font_height);
}

void viewport_clear_all(viewport_t *vp)
{
  int color;
  color=bgcc2col(vp,0);

  draw_rect(vp, 0, 0, vp->xres, vp->yres, color);
  sync_all(vp);
}


/* local functions - implementation */
typedef union {
  uint16_t val;
  struct {
    uint16_t b:5;
    uint16_t g:6;
    uint16_t r:5;
  } __attribute__ ((packed));
} p16_t;

static void __vp_scroll(viewport_t *vp,int lines)
{
/*  uint8_t *m = (uint8_t*)vp->bg;
  
  if(!vp)    return;
  
  if(lines>0) {
    memcpy(m,(void *)(m+(lines*vp->xres)*vp->bpp),
           (vp->xres*vp->yres*vp->bpp)-(lines*vp->xres*vp->bpp));
    draw_rect(vp,0,vp->yres-lines,vp->xres,lines,DEFAULT_BGCOLOR);
  } else {
    lines = -lines;
    memmove((void *)(m+(lines*vp->xres)*vp->bpp), m,
           (vp->xres*vp->yres*vp->bpp)-(lines*vp->xres*vp->bpp));
    draw_rect(vp,0,0,vp->xres,lines,DEFAULT_BGCOLOR);
  }

  vp->pending_repaint = 1;
  
  return;*/
  __vp_scroll_area(vp, lines, 0, vp->yres);
}

static void __vp_scroll_area(viewport_t *vp,int lines, int startline, int stopline)
{
  uint8_t *m = &((uint8_t*)vp->bg)[startline * vp->xres];
  int len = stopline - startline;
  
  if(!vp)    return;
  
  if(lines>0) {
    memcpy(m,(void *)(m+(lines*vp->xres)*vp->bpp),
           vp->xres * vp->bpp * (len - lines));
    draw_rect(vp,0,stopline-lines,vp->xres,lines,DEFAULT_BGCOLOR);
  } else {
    lines = -lines;
    memmove((void *)(m+(lines*vp->xres)*vp->bpp), m,
            vp->xres * vp->bpp * (len - lines));
    draw_rect(vp,0,startline,vp->xres,lines,DEFAULT_BGCOLOR);
  }

  vp->pending_repaint = 1;
  
  return;
}


static inline int fgcc2col(viewport_t *vp,uint8_t col)
{
  gly_color_t f;

  f.comb_color=col;
  switch(f.fg) {
  case TC_BLACK: /* it's a default */
    return DEFAULT_FGCOLOR;
    break;
  case TC_BLUE:
    return 0x000ec3;
    break;
  case TC_GREEN:
    return 0x1aa000;
    break;
  case TC_CYAN:
    return 0x9c00c3;
    break;
  case TC_RED:
    return 0xc30000;
    break;
  case TC_MAGENTA:
    return 0xc30085;
    break;
  case TC_BROWN:
    return 0xab5d00;
    break;
  case TC_WHITE:
    return 0xe3e3e3;
    break;
  case TC_LIGHT_BLUE:
    return 0x5c64ff;
    break;
  case TC_LIGHT_GREEN:
    return 0x00e005;
    break;
  case TC_LIGHT_RED:
    return 0xff1d1d;
    break;
  case TC_LIGHT_BROWN:
    return 0xfd8f00;
    break;
  case TC_LIGHT_GREY:
    return 0xb3b3b3;
    break;
  }

  return 0;
}

static inline void draw_rect(viewport_t *vp,int x,int y,int width,int height,int col)
{
  int i,j;
  
  for(i=0;i<height;i++) {
    for(j=0;j<width;j++) {
      vp->vc_rgb2col(&vp->bg[((y+i)*vp->xres)*vp->bpp+(x+j)*vp->bpp],col);
    }
  }  
}
static inline int bgcc2col(viewport_t *vp,uint8_t col)
{
  gly_color_t f;

  f.comb_color=col;
  switch(f.bg) {
  case TC_BLACK: /* it's a default */
    return DEFAULT_BGCOLOR;
    break;
  case TC_BLUE:
    return 0x000ec3;
    break;
  case TC_GREEN:
    return 0x1aa000;
    break;
  case TC_CYAN:
    return 0x9c00c3;
    break;
  case TC_RED:
    return 0xc30000;
    break;
  case TC_MAGENTA:
    return 0xc30085;
    break;
  case TC_BROWN:
    return 0xab5d00;
    break;
  case TC_WHITE:
    return 0xe3e3e3;
    break;
  case TC_LIGHT_BLUE:
    return 0x5c64ff;
    break;
  case TC_LIGHT_GREEN:
    return 0x00e005;
    break;
  case TC_LIGHT_RED:
    return 0xff1d1d;
    break;
  case TC_LIGHT_BROWN:
    return 0xfd8f00;
    break;
  case TC_LIGHT_GREY:
    return 0xb3b3b3;
    break;
  }

  return 0;
}

static int __convbg_color16(viewport_t *vp,uint8_t col)
{
  gly_color_t f;
  p16_t cc;

  f.comb_color=col;
  switch(f.bg) {
  case TC_BLACK: /* it's a default */
    return 0;
    break;
  case TC_BLUE:
    cc.b=24;
    cc.g=0;
    cc.r=0;
    return (int)cc.val;
    break;
  case TC_GREEN:
    cc.b=0;
    cc.g=24;
    cc.r=0;
    return (int)cc.val;
    break;
  case TC_CYAN:
    cc.b=8;
    cc.g=5;
    cc.r=23;
    return (int)cc.val;
    break;
  case TC_RED:
    cc.b=0;
    cc.g=0;
    cc.r=24;
    return (int)cc.val;
    break;
  case TC_MAGENTA:
    cc.b=8;
    cc.g=20;
    cc.r=20;
    return (int)cc.val;
    break;
  case TC_BROWN:
    cc.b=16;
    cc.g=20;
    cc.r=20;
    return (int)cc.val;
    break;
  case TC_WHITE:
    cc.b=24;
    cc.g=24;
    cc.r=24;
    return (int)cc.val;
    break;
  case TC_LIGHT_BLUE:
    cc.b=31;
    cc.g=0;
    cc.r=0;
    return (int)cc.val;
    break;
  case TC_LIGHT_GREEN:
    cc.b=0;
    cc.g=31;
    cc.r=0;
    return (int)cc.val;
    break;
  case TC_LIGHT_RED:
    cc.b=0;
    cc.g=0;
    cc.r=31;
    return (int)cc.val;
    break;
  case TC_LIGHT_BROWN:
    return vp->bgcol;
    break;
  case TC_LIGHT_GREY:
    return vp->bgcol;
    break;
  }

  return vp->bgcol;
}

static int __convfg_color16(viewport_t *vp,uint8_t col)
{
  gly_color_t f;
  p16_t cc;

  f.comb_color=col;
  switch(f.fg) {
  case TC_BLACK: /* it's a default */
    return DEFAULT_FGCOLOR;
    break;
  case TC_BLUE:
    cc.b=24;
    cc.g=0;
    cc.r=0;
    return (int)cc.val;
    break;
  case TC_GREEN:
    cc.b=0;
    cc.g=24;
    cc.r=0;
    return (int)cc.val;
    break;
  case TC_CYAN:
    cc.b=8;
    cc.g=5;
    cc.r=23;
    return (int)cc.val;
    break;
  case TC_RED:
    cc.b=0;
    cc.g=0;
    cc.r=24;
    return (int)cc.val;
    break;
  case TC_MAGENTA:
    cc.b=8;
    cc.g=20;
    cc.r=20;
    return (int)cc.val;
    break;
  case TC_BROWN:
    cc.b=16;
    cc.g=20;
    cc.r=20;
    return (int)cc.val;
    break;
  case TC_WHITE:
    cc.b=24;
    cc.g=24;
    cc.r=24;
    return (int)cc.val;
    break;
  case TC_LIGHT_BLUE:
    cc.b=31;
    cc.g=0;
    cc.r=0;
    return (int)cc.val;
    break;
  case TC_LIGHT_GREEN:
    cc.b=0;
    cc.g=32;
    cc.r=0;
    return (int)cc.val;
    break;
  case TC_LIGHT_RED:
    cc.b=0;
    cc.g=0;
    cc.r=31;
    return (int)cc.val;
    break;
  case TC_LIGHT_BROWN:
    return vp->fgcol;
    break;
  case TC_LIGHT_GREY:
    return vp->fgcol;
    break;
  }

  return vp->fgcol;
}

/* TODO: make a difference between 565 555 556 */

static void __putpix16(viewport_t *vp,int x,int y,int color)
{
  uint16_t *pp=(uint16_t*)vp->marea;
  uint16_t *bb;

  pp[y*vp->xres+x]=(uint16_t)color;

  bb=(uint16_t*)vp->bg;
  bb[y*vp->xres+x]=(uint16_t)color;
  
  return;
}

/* color representation */
/* 32bit bpp */
static void col_rgb0888(void *dst,int col)
{
  *(int*)dst=col;
}

static int rgb0888_col(void *src)
{
  return (*(uint32_t *)src) & 0xffffff;
}

static void bgr_col0888(void *dst,int rgb)
{
  *((uint32_t *) dst) = BLUE(rgb, 8) << 16 | GREEN(rgb, 8) << 8 | RED(rgb, 8);
}

static int col0888_bgr(void *src)
{
  int color = *(uint32_t *)(src);

  return ((color & 0xff) << 16) | (((color >> 8) & 0xff) << 8) |
    ((color >> 16) & 0xff);
}

/* 24-bit bpp */
static void rgb_col888(void *dst, int rgb)
{
  uint8_t *scr=dst;

#ifdef INVERT_ENDIAN
  scr[0]=RED(rgb,8);
  scr[1]=GREEN(rgb,8);
  scr[2]=BLUE(rgb,8);
#else
  scr[2]=RED(rgb,8);
  scr[1]=GREEN(rgb,8);
  scr[0]=BLUE(rgb,8);
#endif
}

static int col888_rgb(void *src)
{
  uint8_t *scr=src;
  
#ifdef INVERT_ENDIAN
  return scr[0] << 16 | scr[1] << 8 | scr[2];
#else
  return scr[2] << 16 | scr[1] << 8 | scr[0];
#endif
}

/* 16-bit bpp (5:5:5) */
static int col555_rgb(void *src)
{
  int color = *(uint16_t *)(src);
  
  return (((color >> 10) & 0x1f) << (16 + 3)) |
    (((color >> 5) & 0x1f) << (8 + 3)) | ((color & 0x1f) << 3);
}

/* 16-bit bpp (5:6:5) */
static void rgb_col565(void *dst, int rgb)
{
  *((uint16_t *)(dst)) = RED(rgb, 5) << 11 | GREEN(rgb, 6) << 5 | BLUE(rgb, 5);
}

static int col565_rgb(void *src)
{
  int color=*(uint16_t *)(src);
  
  return (((color >> 11) & 0x1f) << (16 + 3)) |
    (((color >> 5) & 0x3f) << (8 + 2)) | ((color & 0x1f) << 3);
}

