/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008,2009 MadTirra <madtirra@jarios.org>
 *
 * vesacons.d: Generic VESA graphical console
 * fb_types.h: framebuffer related types
 *
 */

#ifndef __FB_TYPES_H__
#define __FB_TYPES_H__

#include <sys/types.h>
#include <pthread.h>
#include <wchar.h>

#include "pcf.h"

/* color defines */
#define RED(x, bits)    ((x >> (16 + 8 - bits)) & ((1 << bits) - 1))
#define GREEN(x, bits)  ((x >> (8 + 8 - bits)) & ((1 << bits) - 1))
#define BLUE(x, bits)   ((x >> (8 - bits)) & ((1 << bits) - 1))

/* framebuffer bpp types */
#define RGB_555   0
#define RGB_565   1
#define RGB_888   2
#define RGB_8880  3
#define RGB_0888  4
#define BGR_0888  5

/* default colors */
#define DEFAULT_BGCOLOR  0x1e5cdb
#define DEFAULT_FGCOLOR  0xeaeaea

typedef union {
  uint8_t comb_color;
  struct {
    uint8_t bg : 4;
    uint8_t fg : 4;
  } __attribute__ ((packed));
} gly_color_t;

typedef union {
  wchar_t s;
  struct {
    uint16_t p0 : 8;
    uint16_t p1 : 8;
    uint16_t p2 : 8;
    uint16_t p3 : 8;
  } __attribute__ ((packed));
} gly_char_t;

typedef struct __fbfont_t {
  int width;
  int scan_lines;
  unsigned char **font;
} fbfont_t;

typedef enum {
  PCF,
  BUILTIN,
} font_type_t;

typedef struct __viewport_t {
  /* general structure */
  uint8_t *marea;
  pthread_mutex_t vp_lock;
  int xres;
  int yres;
  int bpp;
  int bgcol;
  int fgcol;
  int font_type;
  int font_width;
  int font_height;
  /* layers */
  /* bg */
  uint8_t *bg;
  pthread_mutex_t bg_lock;
  /* fg */
  void *fg;
  pthread_mutex_t fg_lock;
  /* root text area */
  fbfont_t *general_font;
  pcf_font_t *pcf_font;
  int xxf; /* x pointplace */
  int yyf; /* y pointplace */
  int curx,cury; /* current position */
  int disable_scroll;
  int cur_show;
  /* functions */
  void (*ppixel_fg)(struct __viewport_t *vp,int x,int y,int color);
  int (*vc_bcconvert)(struct __viewport_t *vp,uint8_t c);
  int (*vc_fcconvert)(struct __viewport_t *vp,uint8_t c);
  /* color conversion */
  void (*vc_rgb2col)(void *dst,int col);
  int (*vc_col2rgb)(void *src);
  /* double buffer functions */
  void (*vc_scroll_area)(struct __viewport_t *vp, int lines, int startline, int stopline);
  void (*vc_scroll)(struct __viewport_t *vp, int lines);
  ulong_t pending_repaint;
  ulong_t dirty_area;
} viewport_t;

/* framebuffer font operations */
int fbfont_init(fbfont_t *p,int width,int sc_line);
int fbfont_settablefont(fbfont_t *p,int tbl,void *data);
int fbfont_put(viewport_t *vp,fbfont_t *fnt,int sx,int sy,
	       cschar_t c,int opaque);

/* viewport */
int viewport_init(viewport_t *vp,void *marea,int xres,
		  int yres,int bpp,int bpptype);
int viewport_setfont(viewport_t *vp,fbfont_t *font);
void viewport_set_pcf_font(viewport_t *vp,pcf_font_t *font);
void viewport_put(viewport_t *vport,int sx,int sy,
		  cschar_t gly,int opaque);
void viewport_putc(viewport_t *vp,cschar_t gly);
void viewport_putcxy(viewport_t *vp,cschar_t gly,int *x,int *y);
/* extended */
void viewport_drawpixmap(viewport_t *vp,uint32_t *pixbuf,
			 int sx,int sy,int width,int height);
void viewport_clear_rect(viewport_t *vp, int x1, int y1, int x2, int y2, uint8_t col);
void viewport_clear_all(viewport_t *vp);
void viewport_showcursor(viewport_t *vp,int sx,int sy);
void viewport_set_hw(viewport_t *vp);


#endif /* __FB_TYPES_H__ */
