/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008,2009 MadTirra <madtirra@jarios.org>
 * (c) Copyright 2009 Paul Graphov <graphov@jarios.org>
 *
 * vesacons.d: Generic VESA graphical console
 * fbcon.c: framebuffer console high-level operations
 *
 */

#define FONTPATH "/etc/vesacons/font.pcf"


#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <wchar.h>
#include <errno.h>
#include <time.h>
#include <datalib/usrtc.h>
#include <ns/ns.h>
#include <sys/scheduler.h>
#include <ipc/rpc.h>
#include <devio/devio.h>
#include <stdio.h>
#include <int10/int10.h>
#include <pthread.h>

#include "vga.h"
#include "hwops.h"
#include "fb_types.h"
#include "default_font.h"
#include "fbcon.h"
#include "fb_logo.h"
#include "pcf.h"


#define DEFW_FONT  8  /* default font is 8x16 */
#define DEFH_FONT  16

static viewport_t vop;
static fbfont_t fot;



int fbcon_bpptype(int bpp,uint8_t red_off,uint8_t green_off,uint8_t blue_off,
		  uint8_t red_size,uint8_t green_size,uint8_t blue_size)
{
  switch(bpp) {
  case 2:
    if(red_size==green_size)
      return RGB_555;
    else if(green_size==6)
      return RGB_565;
    return RGB_565; /* default */
    break;
  case 3:
    return RGB_888; /* default for true color */
    break;
  case 4:
    return RGB_0888;
    break;
  default:
    return -1;
    break;
  }

  return -1;
}

int fbcon_init(void *area,int xres,int yres,int bpp,int bpptype)
{
  int r;
  pcf_font_t *font;

  
  r=viewport_init(&vop,area,xres,yres,bpp,bpptype);
  if(r!=0)
    return r;

  /* initialize font, first try pcf */
  font = pcf_read_font(FONTPATH);
  if(!font) {
    r=fbfont_init(&fot,DEFW_FONT,DEFH_FONT);
    if(r!=0) {
      //no font, total fail.
      return r;
    }
    fbfont_settablefont(&fot,0,sfnt8x16_table0); /* ASCII table */
    r=viewport_setfont(&vop,&fot);
    if(r!=0)
      return r;

    vop.font_type = BUILTIN;
    vop.font_width = 8;
    vop.font_height = 16;
  } else {
    pcf_metrics_t *m;
    viewport_set_pcf_font(&vop, font);
    vop.font_type = PCF;
    m = pcf_get_metrics(font, 0);
    vop.font_width = m->character_width;
    vop.font_height = m->character_ascent + m->character_descent;
  }
  viewport_set_hw(&vop);

  fbcon_clear_all();

  /* show Jari OS logo */
  viewport_drawpixmap(&vop,logo82x82_data,0,0,82,82);

  vop.cury=5;

  return 0;
}

void fbcon_printxy(cschar_t *buf,size_t len,int x,int y)
{
  int i,ulen;
  int xs=x,ys=y;

  if(x>vop.xxf) /* check for overflow */
    return;
  if(y>vop.yyf)
    return;

  ulen=len/sizeof(cschar_t);
  for(i=0;i<ulen;i++) 
    viewport_putcxy(&vop,buf[i],&xs,&ys);

  return;
}

void fbcon_print(cschar_t *msg,size_t len)
{
  int i,ulen;

  if(vop.cur_show)
    viewport_showcursor(&vop,vop.curx,vop.cury);

  ulen=len/sizeof(cschar_t);
  for(i=0;i<ulen;i++) 
    viewport_putc(&vop,msg[i]);
  
  if(vop.cur_show)
    viewport_showcursor(&vop,vop.curx,vop.cury);

  return;
}

void fbcon_puts(char *msg)
{
  cschar_t t;
  gly_color_t co;
  int ulen=strlen(msg);
  int i;

  co.bg=0;
  co.fg=TC_RED;
  t.cattr=co.comb_color;

  for(i=0;i<ulen;i++) {
    t.ch=(wchar_t)msg[i];
    viewport_putc(&vop,t);
  }

  if(vop.cur_show) viewport_showcursor(&vop,vop.curx,vop.cury);

  return;
}

void fbcon_setcursorvisibility(int to_show)
{
  if(vop.cur_show == to_show) return;
  
  viewport_showcursor(&vop,vop.curx,vop.cury);
  vop.cur_show = to_show;
}

void fbcon_setcursor(int x, int y)
{
  if(vop.cur_show)
    viewport_showcursor(&vop,vop.curx,vop.cury);  
  if(x < 0) x = 0;
  else if(x >= vop.xxf) x = vop.xxf - 1;
  if(y < 0) y = 0;
  else if(y >= vop.yyf) y = vop.yyf - 1;
  vop.curx = x;
  vop.cury = y;
  if(vop.cur_show)
    viewport_showcursor(&vop,vop.curx,vop.cury);  
}

void fbcon_scroll(int lines)
{
  vop.vc_scroll(&vop, lines*vop.font_height);
}

void fbcon_scroll_area(int startline, int stopline, int lines)
{
  vop.vc_scroll_area(&vop, lines * vop.font_height,
                     startline * vop.font_height,
                     (stopline + 1) * vop.font_height);
}

int fbcon_width(void)
{
  return vop.xxf;
}

int fbcon_height(void)
{
  return vop.yyf;
}

void fbcon_scrolldisable(void)
{
  vop.disable_scroll=true;
}

void fbcon_scrollenable(void)
{
  vop.disable_scroll=false;
}

void fbcon_clear_rect(int x1, int y1, int x2, int y2, uint8_t col)
{
  viewport_clear_rect(&vop, x1, y1, x2, y2, col);
  if(vop.cur_show && vop.curx >= x1 && vop.curx <=x2 && vop.cury >= y1 && vop.cury <= y2) viewport_showcursor(&vop,vop.curx,vop.cury);
}

void fbcon_clear_all(void)
{
  viewport_clear_all(&vop);
}
