#ifndef __PCF_H__
#define __PCF_H__

typedef uint32_t lsbint32;

typedef struct __pcf_toc_entry {
  lsbint32 type;      // See below, indicates which table 
  lsbint32 format;    // See below, indicates how the data are formatted in the table 
  lsbint32 size;      // In bytes
  lsbint32 offset;    // from start of file 
  void *table;
} pcf_toc_entry_t;

//The type field may be one of:

#define PCF_PROPERTIES            (1<<0)
#define PCF_ACCELERATORS          (1<<1)
#define PCF_METRICS               (1<<2)
#define PCF_BITMAPS               (1<<3)
#define PCF_INK_METRICS           (1<<4)
#define PCF_BDF_ENCODINGS         (1<<5)
#define PCF_SWIDTHS               (1<<6)
#define PCF_GLYPH_NAMES           (1<<7)
#define PCF_BDF_ACCELERATORS      (1<<8)

//The format field may be one of:

#define PCF_DEFAULT_FORMAT  0x00000000
#define PCF_INKBOUNDS    0x00000200
#define PCF_ACCEL_W_INKBOUNDS  0x00000100
#define PCF_COMPRESSED_METRICS  0x00000100

//The format field may be modified by:

#define PCF_GLYPH_PAD_MASK  (3<<0)    // See the bitmap table for explanation 
#define PCF_BYTE_MASK    (1<<2)    // If set then Most Sig Byte First 
#define PCF_BIT_MASK    (1<<3)    // If set then Most Sig Bit First 
#define PCF_SCAN_UNIT_MASK  (3<<4)    // See the bitmap table for explanation 

#define PCF_BYTE_ORDER(f)       (((f) & PCF_BYTE_MASK) ? PCF_BE : PCF_LE)
#define PCF_BIT_ORDER(f)        (((f) & PCF_BIT_MASK)  ? PCF_BE : PCF_LE)
#define PCF_GLYPH_PAD_INDEX(f)  ((f) & PCF_GLYPH_PAD_MASK)
#define PCF_GLYPH_PAD(f)        (1<<PCF_GLYPH_PAD_INDEX(f))
#define PCF_SCAN_UNIT_INDEX(f)  (((f) & PCF_SCAN_UNIT_MASK) >> 4)
#define PCF_SCAN_UNIT(f)        (1<<PCF_SCAN_UNIT_INDEX(f))
#define PCF_FORMAT_BITS(f)      ((f) & (PCF_GLYPH_PAD_MASK|PCF_BYTE_MASK|PCF_BIT_MASK|PCF_SCAN_UNIT_MASK))

#define PCF_SIZE_TO_INDEX(s)    ((s) == 4 ? 2 : (s) == 2 ? 1 : 0)
#define PCF_INDEX_TO_SIZE(b)    (1<<b)


// Property record
typedef struct __pcf_prop_t {
  uint32_t name_offset;    // Offset into the following string table 
  uint8_t is_string_prop;
  uint32_t value;            // The value for integer props, the offset for string props 
} pcf_prop_t;

// Properties Table
typedef struct __pcf_prop_table {
  uint32_t nprops;      // Stored in whatever byte order is specified in the format 
  pcf_prop_t *props;
  //here goes padding to 4 byte boundary
  uint32_t string_size;    // total size of all strings (including their terminating NULs) 
  char *strings;
  //here goes padding to 4 byte boundary
} pcf_prop_table_t;


// Get name of nth property in table t
#define PROP_NAME(t,n)                          \
  (&(t)->strings[(t)->props[n].name_offset])

// Get value of nth property in table t
#define PROP_VALUE(t,n)                         \
  (&(t)->strings[(t)->props[n].value])

// Metrics data

//Uncompressed
typedef struct __pcf_metrics {
  int16_t left_side_bearing;
  int16_t right_side_bearing;
  int16_t character_width;
  int16_t character_ascent;
  int16_t character_descent;
  uint16_t character_attributes;
} pcf_metrics_t;

//The accelerator tables look like:

typedef struct __pcf_accel_table {
  uint8_t no_overlap;    // if for all i, max(metrics[i].rightSideBearing - metrics[i].characterWidth) 
   //   <= minbounds.leftSideBearing 
  uint8_t constant_metrics;    // Means the perchar field of the XFontStruct can be NULL 
  uint8_t terminal_font;    // constantMetrics true and forall characters: 
  //   the left side bearing==0 
  //  the right side bearing== the character's width 
  //  the character's ascent==the font's ascent 
  //  the character's descent==the font's descent 
  uint8_t constant_width;    // monospace font like courier 
  uint8_t ink_inside;    // Means that all inked bits are within the rectangle with x between [0,charwidth] 
  //  and y between [-descent,ascent]. So no ink overlaps another char when drawing 
  uint8_t ink_metrics;    // true if the ink metrics differ from the metrics somewhere 
  uint8_t draw_direction;    // 0=>left to right, 1=>right to left 
  int32_t font_ascent;    // byte order as specified in format 
  int32_t font_descent;
  int32_t max_overlap;    // ??? 
  pcf_metrics_t *minbounds;
  pcf_metrics_t *maxbounds;
  // If format is PCF_ACCEL_W_INKBOUNDS then include the following fields 
  // Otherwise those fields are not in the file and should be filled by duplicating min/maxbounds above 
  pcf_metrics_t *ink_minbounds;
  pcf_metrics_t *ink_maxbounds;
} pcf_accel_table_t;


// Metrics table

typedef struct __pcf_metrics_table {
  // if the format is compressed 
  // int16 metrics_count;
  // Compressed_Metrics metrics[metrics_count];
  // else if format is default (uncompressed) 
  int32_t metrics_count;
  pcf_metrics_t **metrics;
  // endif 
} pcf_metrics_table_t;


// The Bitmap Table
// The bitmap table has type PCF_BITMAPS. Its format must be PCF_DEFAULT.

typedef struct __pcf_bitmap_table {
  uint32_t glyph_count;     // byte ordering depends on format, should be the same as the metrics count 
  uint32_t *offsets;         //[glyph_count];  byte offsets to bitmap data 
//  int32_t bitmapSizes[4];    // the size the bitmap data will take up depending on various padding options 
  //  which one is actually used in the file is given by (format&3) 
  uint8_t *bitmap_data;        //[bitmapsizes[format&3]] the bitmap data. format contains flags that indicate: 
  // the byte order (format&4 => LSByte first)
  // the bit order (format&8 => LSBit first) 
  // how each row in each glyph's bitmap is padded (format&3) 
  //  0=>bytes, 1=>shorts, 2=>ints 
  // what the bits are stored in (bytes, shorts, ints) (format>>4)&3 
  //  0=>bytes, 1=>shorts, 2=>ints 
} pcf_bitmap_table_t;


// The Encoding Table
//The encoding table has type PCF_BDF_ENCODINGS. Its format must be PCF_DEFAULT.

typedef struct __pcf_enc_table {
  int16_t glyphindeces[256 * 256];
} pcf_enc_table_t;
  
typedef struct __pcf_swidth_table {  
  int32_t glyph_count;    // byte ordering depends on format, should be the same as the metrics count
  int32_t *swidths;      // byte offsets to bitmap data
} pcf_swidth_table_t;

// The encoding table has type PCF_GLYPH_NAMES. Its format must be PCF_DEFAULT.

typedef struct __pcf_glyph_table {
  int32_t glyph_count;    // byte ordering depends on format, should be the same as the metrics count 
  int32_t *offsets;  // byte offsets to string data 
  int32_t string_size;
  char *string;
} pcf_glyph_table_t;

#define EMAGIC 123


typedef struct __pcf_font {
  pcf_prop_table_t *prop_table;
  pcf_metrics_table_t *metrics_table;
  pcf_accel_table_t *accel_table;
  pcf_bitmap_table_t *bitmap_table;
  pcf_enc_table_t *enc_table;
  pcf_swidth_table_t *swidth_table;
  pcf_glyph_table_t *glyph_table;
} pcf_font_t;

int open_font(const char *path);
pcf_font_t *pcf_read_font(const char *path);
uint16_t pcf_get_char_index(pcf_font_t *fnt, uint16_t code);
pcf_metrics_t *pcf_get_metrics(pcf_font_t *fnt, uint16_t index);
uint8_t *pcf_get_bitmap(pcf_font_t *fnt, uint16_t index);


#endif
