#include "internal.h"
#include <stdio.h>
#include <stdlib.h>
#include <png/png.h>
#include <sys/kcontrol.h>
#include <byteorder.h>

// Length of file signature to check, 8 at most
#define SIG_LEN 8

// TODO: add error codes or logging.

int load_background(const char *path, viewport_ops_t *viewport)
{
  int ret, i;
  FILE *png_file;
  uint8_t header[SIG_LEN];
  png_structp png_ptr;
  png_infop info_ptr;
  png_infop end_info;
  uint32_t width, height;
  int bit_depth, color_type, interlace_type,
    compression_type, filter_type;
  uint32_t mxcsr_reg = 4096;
  uint8_t *row = NULL;
  uint8_t *bg = NULL;
  

  // Hack: mask some SIMD exceptions.
  // TODO: do something with it.
  asm volatile (
    "ldmxcsr %0"
    :
    : "m" (mxcsr_reg));

  // we will convert all pngs to 0RGB or RGB0 so we
  // need 4 bytes per pixel here
  row = malloc(4 * viewport->width);
  if(!row) {
    goto out;
  }

  bg = malloc(viewport->bytes_per_pixel * viewport->width * viewport->height);
  if(!bg) {
    goto out;
  }
  
  png_file = fopen(path, "r");
  if(!png_file) {
    klog("Failed to open\n");
    return -1;
  }

  ret = fread(header, 1, SIG_LEN, png_file);
  if(ret != SIG_LEN) {
    klog("Signature read fail\n");
    goto out;
  }

  // Make sure we have png file.
  ret = png_sig_cmp(header, 0, 8);
  if(ret != 0) {
    klog("Signature check fail\n");
    goto out;
  }

  png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, (png_voidp)NULL,
                                   NULL, NULL);
  if(!png_ptr) {
    klog("Failed to alloc.\n");
    goto out;
  }
  
  info_ptr = png_create_info_struct(png_ptr);
  if(!info_ptr) {
    klog("Failed to alloc 2.\n");
    png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
    goto out;
  }
  
  end_info = png_create_info_struct(png_ptr);
  if(!end_info)
  {
    klog("Failed to alloc 3.\n");
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    goto out;
  }

  // Errors detected by libpng will jump here.
  // so no return code checking is performed.
  if(setjmp(png_jmpbuf(png_ptr))) {
    error:
    klog("setjmp error handler called.\n");
    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
    goto out;
  }

  // Set stream to read png from.
  png_init_io(png_ptr, png_file);

  // Inform libpng that we already read some bytes.
  png_set_sig_bytes(png_ptr, SIG_LEN);

  // Width and height.
  png_set_user_limits(png_ptr, 5000, 4000);

  // Ignore unknown chunks.
  png_set_keep_unknown_chunks(png_ptr, PNG_HANDLE_CHUNK_NEVER, NULL, 0);

  // Read file info chunks but not the image data
  png_read_info(png_ptr, info_ptr);

  // Check image properties:
  png_get_IHDR(png_ptr, info_ptr, &width, &height,
               &bit_depth, &color_type, &interlace_type,
               &compression_type, &filter_type);

  klog("width: %d; height: %d; bit_depth: %d, color_type: %d, "
       "interlace_type: %d; compression_type: %d; filter_type: %d\n",
       width, height, bit_depth, color_type, interlace_type,
       compression_type, filter_type);

  if(width != viewport->width || height != viewport->height) {
    klog("Wrong image size\n");
    goto error;
  }
    
  if(color_type == PNG_COLOR_TYPE_PALETTE)
    png_set_palette_to_rgb(png_ptr);
  
  if(color_type & PNG_COLOR_MASK_ALPHA)
    png_set_strip_alpha(png_ptr);

  if(color_type == PNG_COLOR_TYPE_GRAY ||
     color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
    png_set_gray_to_rgb(png_ptr);

  if(bit_depth < 8)
    png_set_packing(png_ptr);

  if(bit_depth == 16)
    png_set_strip_16(png_ptr);

#if BYTE_ORDER == LITTLE_ENDIAN
  png_set_filler(png_ptr, 0, PNG_FILLER_AFTER);
  png_set_bgr(png_ptr);
#elif BYTE_ORDER == BIG_ENDIAN
  png_set_filler(png_ptr, 0, PNG_FILLER_BEFORE);
#else
#error "Unknown endianness"
#endif

  //TODO: maybe it's good to add support for interlaced pngs.
  
  png_read_update_info(png_ptr, info_ptr);

  // after that transormations (above) we hope
  // to have 4 bytes per pixel, 0RGB or RGB0
  // depending on endianness;

  // Now it's time to read it line by line and load to

  for(i = 0; i < height; i++) {
    png_read_row(png_ptr, row, NULL);
    viewport->palette_transform(row, &bg[width * i * viewport->bytes_per_pixel], width);
  }

  viewport->set_bg(bg);
  
  fclose(png_file);
  free(row);

  return 0; 
  
  out:
  fclose(png_file);
  free(row);
  free(bg);
  return -1;
}
