#include <sys/types.h>
#include <stdbool.h>
#include <string.h>
#include <pthread.h>
#include <stdlib.h>
#include <arch/atomic.h>

#include <sys/kcontrol.h>

#include "palette.h"
#include "internal.h"

static int cur_x;
static int cur_y;

static bool cur_printed;

static font_ops_t *font = NULL;
static viewport_ops_t *viewport = NULL;

extern viewport_ops_t vesa_ops;

extern font_ops_t bf_ops;
extern font_ops_t pcf_ops;

cschar_t *user_text;
cschar_t *buffer_text;
cschar_t *video_text;

// OOPs... 
#define this gc_ops

// TODO: maybe it's better to move
// repainting routines to viewport somehow...
pthread_t repainter;
bool need_repaint;
bool need_repaint_all;
ulong_t *repaint_mask = NULL;

static int gc_init(void);
static void gc_set_cursor(int x, int y);
static void gc_hide_cursor(void);
static void gc_put_char(int x, int y, cschar_t ch);
static void gc_scroll_area(int startline, int stopline, int lines, int color);
static void gc_fill_rect(int x, int y, int width, int height, int color);

static inline void mark(int x, int y);
static inline void mark_all(void);
static inline void clear_marks(void);
static void *repainter_routine(void *arg);

console_ops_t gc_ops = {
  .init = gc_init,
  .set_cursor = gc_set_cursor,
  .hide_cursor = gc_hide_cursor,
  .put_char = gc_put_char,
  .scroll_area = gc_scroll_area,
  .fill_rect = gc_fill_rect
};

#define BLOCK_SIZE (sizeof(ulong_t))
#define BLOCK_BITS (BLOCK_SIZE * 8)
#define NUM_BLOCKS ((this.width * this.height + BLOCK_BITS - 1) / BLOCK_BITS)
#define MASK_BYTES (NUM_BLOCKS * BLOCK_SIZE)

// Init font, then viewport
static int gc_init(void)
{
  int ret;

  //try fonts:
  if(pcf_ops.init("/etc/vesacons/font.pcf") == 0) {
    font = &pcf_ops;
  } else if(bf_ops.init(NULL) == 0) {
    font = &bf_ops;
  } else {
    goto out;
  }

  //try viewports
  if(vesa_ops.init() == 0) {
    viewport = &vesa_ops;
  } else {
    goto out;
  }

  this.width = viewport->width / font->width;
  this.height = viewport->height / font->height;

  repaint_mask = malloc(MASK_BYTES);
  if(!repaint_mask) {
    goto out;
  }

  memset(repaint_mask, 0, MASK_BYTES);

  ret = load_background("/etc/vesacons/wallpaper.png", viewport);
  
  ret = pthread_create(&repainter, NULL, repainter_routine, NULL);
  if(ret != 0) {
    goto out;
  }
  
  return 0;

  out:

  free(repaint_mask);
  repaint_mask = NULL;
  
  if(font && font->destroy) {
    font->destroy();
  }
  if(viewport && viewport->destroy) {
    viewport->destroy();
  }
  return -1;
}

static inline void invert_cursor(int x, int y)
{
  viewport->invert_rect(font->width * x, font->height * y, font->width, font->height);
  mark(x, y);
}

static void gc_set_cursor(int x, int y)
{
  if(cur_printed) {
    if(x == cur_x && y == cur_y) {
      return;
    }
    invert_cursor(cur_x, cur_y);
  }
  
  cur_x = x;
  cur_y = y;
  
  invert_cursor(cur_x, cur_y);
  cur_printed = true;
}

static void gc_hide_cursor(void)
{
  if(cur_printed) {
    invert_cursor(cur_x, cur_y);
    cur_printed = false;
  }
}

static void gc_put_char(int x, int y, cschar_t ch)
{
  int px, py;
  uint8_t *bitmap;
  
  px = font->width * x;
  py = font->height * y;

  bitmap = font->get_bitmap(ch.ch);
  if(bitmap) {
    viewport->draw_bitmap(px, py, FG_COLOR(ch.cattr), BG_COLOR(ch.cattr), bitmap, font->width, font->height, font->padding);
    mark(x, y);
  }
}

static void gc_scroll_area(int startline, int stopline, int lines, int color)
{
  viewport->scroll_area(startline * font->height, (stopline + 1) * font->height,
                       lines * font->height, color);

  mark_all();
}

static void gc_fill_rect(int x, int y, int width, int height, int color)
{
  viewport->fill_rect(x * font->width, y * font->height,
                     width * font->width, height * font->height, color);
  if(cur_x >= x && cur_x <= x + width &&
     cur_y >= y && cur_y <= y + height) {
    cur_printed = false;
  }
  
  if(x == 0 && y == 0 && width == this.width && this.height) {
    mark_all();
  } else {
    int i, j;
    for(i = 0; i < height; i++) {
      for(j = 0; j < width; j++) {
        mark(i, j);
      }
    }
  }
}

static inline void mark_all(void)
{
  need_repaint_all = true;
}

static inline void clear_marks(void)
{
  if(need_repaint) {
    memset(repaint_mask, 0, MASK_BYTES);
  }
  need_repaint = false;
  need_repaint_all = false;
}

static inline void mark(int x, int y)
{
  int pos = y * this.width + x;
  if(!need_repaint_all) {
    repaint_mask[pos / BLOCK_BITS] |= 1L << (pos % BLOCK_BITS);
    need_repaint = true;
  }
}

static inline void do_repaint(void)
{
  int i, j;
  int offs, x, y;
  
  if(need_repaint_all) {
    clear_marks();
    viewport->sync_rect(0, 0, viewport->width, viewport->height);
  } else if(need_repaint) {
    need_repaint = false;
    for(i = 0; i < NUM_BLOCKS; i++) {
      ulong_t local;
      if(!repaint_mask[i]) {
        continue;
      }
      local = atomic_swap_long(&repaint_mask[i], 0);
      for(j = 0; j < BLOCK_BITS; j++) {
        if(local & (1L << j)) {
          offs = BLOCK_BITS * i + j;
          x = offs % this.width;
          y = offs / this.width;
          if(y < this.height) {
            viewport->sync_rect(x * font->width, y * font->height, font->width, font->height);
          }
        }
      }
    }
  }
}

static void *repainter_routine(void *arg)
{
  while(1) {
    do_repaint();
    usleep(10000);
  }
}
