#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <byteorder.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/kcontrol.h>

#include "pcf.h"
#include "pcfio.h"

const char pcf_magic[] = "\1fcp";

// TODO: make some error reporting

static int __read_toc_entry(pcfio_t *pio, pcf_toc_entry_t *te)
{
  te->type = pcfio_readlsb32(pio);
  te->format = pcfio_readlsb32(pio);
  te->size = pcfio_readlsb32(pio);
  te->offset = pcfio_readlsb32(pio);
  if(pcfio_haserror(pio)) {
    return -1;
  }
  return 0;
}

static pcf_prop_table_t *_read_prop_table(pcfio_t *pio, pcf_toc_entry_t *te)
{
  pcf_prop_table_t *tbl;
  uint32_t format;
  pcf_prop_t *props;
  char *strings;
  int i;

  tbl = malloc(sizeof(*tbl));

  if(!tbl) {
    errno = ENOMEM;
    return NULL;
  }

  pcfio_seekto(pio, te->offset);
  
  format = pcfio_readlsb32(pio);
  pcfio_set_format(pio, format);
  tbl->nprops = pcfio_read32(pio);

  props = malloc(tbl->nprops * sizeof(*props));
  if(!props) {
    errno = ENOMEM;
    goto rpt_cleanup;
  }

  // Read property entries;
  for(i = 0; i < tbl->nprops; i++) {
    props[i].name_offset = pcfio_read32(pio);
    props[i].is_string_prop = pcfio_read_char(pio);
    props[i].value = pcfio_read32(pio);
    if(pcfio_haserror(pio)) {
      errno = EIO;
      goto rpt_cleanup2;
    }
  }

  pcfio_pad4(pio);

  tbl->string_size = pcfio_read32(pio);
  if(pcfio_haserror(pio)) {
    errno = EIO;
    goto rpt_cleanup2;
  }

  strings = malloc(tbl->string_size);
  if(!strings) {
    errno = ENOMEM;
    goto rpt_cleanup2;
  }

  pcfio_read_block_rel(pio, strings, tbl->string_size);
  
  if(pcfio_haserror(pio)) {
    errno = EIO;
    goto rpt_cleanup3;
  }
  
  //This should already be zero, just for safety.
  strings[tbl->string_size - 1] = 0;

  tbl->strings = strings;
  tbl->props = props;

  return tbl;

  rpt_cleanup3:
  free(strings);
  rpt_cleanup2:
  free(props);
  rpt_cleanup:
  free(tbl);
  return NULL;
}

static pcf_metrics_t *_read_compr_metrics(pcfio_t *pio)
{
  pcf_metrics_t *met = malloc(sizeof(*met));

  if(!met) {
    return NULL;
  }
  
  met->left_side_bearing = pcfio_read_char(pio) - 0x80;
  met->right_side_bearing = pcfio_read_char(pio) - 0x80;
  met->character_width = pcfio_read_char(pio) - 0x80;
  met->character_ascent = pcfio_read_char(pio) - 0x80;
  met->character_descent = pcfio_read_char(pio) - 0x80;
  met->character_attributes = 0;

  if(pcfio_haserror(pio)) {
    free(met);
    return NULL;
  }
  return met;
}

static pcf_metrics_t *_read_uncompr_metrics(pcfio_t *pio)
{
  pcf_metrics_t *met = malloc(sizeof(*met));

  if(!met) {
    return NULL;
  }

  met->left_side_bearing = pcfio_read16(pio);
  met->right_side_bearing = pcfio_read16(pio);
  met->character_width = pcfio_read16(pio);
  met->character_ascent = pcfio_read16(pio);
  met->character_descent = pcfio_read16(pio);
  met->character_attributes = pcfio_read16(pio);
  
  if(pcfio_haserror(pio)) {
    free(met);
    return NULL;
  }
  return met;
}

static pcf_accel_table_t *_read_accel_table(pcfio_t *pio, pcf_toc_entry_t *te)
{
  pcf_accel_table_t *tbl;
  uint32_t format;

  tbl = malloc(sizeof(*tbl));

  if(!tbl) {
    errno = ENOMEM;
    return NULL;
  }

  pcfio_seekto(pio, te->offset);
  if(pcfio_haserror(pio)) {
    goto rat_cleanup;
  }
  
  format = pcfio_readlsb32(pio);
  pcfio_set_format(pio, format);

  tbl->no_overlap = pcfio_read_char(pio);
  tbl->constant_metrics = pcfio_read_char(pio);
  tbl->terminal_font = pcfio_read_char(pio);
  tbl->constant_width = pcfio_read_char(pio);
  tbl->ink_inside = pcfio_read_char(pio);
  
  tbl->ink_metrics = pcfio_read_char(pio);
  tbl->draw_direction = pcfio_read_char(pio);
// padding
  pcfio_read_char(pio);
  
  tbl->minbounds = _read_uncompr_metrics(pio);
  if(!tbl->minbounds) {
    goto rat_cleanup;
  }
  tbl->maxbounds = _read_uncompr_metrics(pio);
  if(!tbl->maxbounds) {
    goto rat_cleanup2;
  }

  if(format & PCF_ACCEL_W_INKBOUNDS) {
    tbl->minbounds = _read_uncompr_metrics(pio);
    if(!tbl->minbounds) {
      goto rat_cleanup3;
    }
    tbl->maxbounds = _read_uncompr_metrics(pio);
    if(!tbl->maxbounds) {
      goto rat_cleanup4;
    }
  }
  
  return tbl;

  rat_cleanup4:
  free(tbl->ink_minbounds);
  
  rat_cleanup3:
  free(tbl->maxbounds);
  
  rat_cleanup2:
  free(tbl->minbounds);
  
  rat_cleanup:
  free(tbl);
  return NULL;
}

static pcf_metrics_table_t *_read_metrics_table(pcfio_t *pio, pcf_toc_entry_t *te)
{
  pcf_metrics_table_t *tbl;
  uint32_t format;
  int i;

  tbl = malloc(sizeof(*tbl));

  if(!tbl) {
    errno = ENOMEM;
    return NULL;
  }

  pcfio_seekto(pio, te->offset);
  if(pcfio_haserror(pio)) {
    goto rmt_cleanup;
  }
  
  format = pcfio_readlsb32(pio);
  pcfio_set_format(pio, format);

  if(format & PCF_COMPRESSED_METRICS) {
    tbl->metrics_count = pcfio_read16(pio);
  } else {
    tbl->metrics_count = pcfio_read32(pio);
  }

  if(pcfio_haserror(pio)) {
    goto rmt_cleanup;
  }

  tbl->metrics = malloc(sizeof(*tbl->metrics) * tbl->metrics_count);

  for(i = 0; i < tbl->metrics_count; i++) {
    if(format & PCF_COMPRESSED_METRICS) {
      tbl->metrics[i] = _read_compr_metrics(pio);
    } else {
      tbl->metrics[i] = _read_uncompr_metrics(pio);
    }
    if(!tbl->metrics[i]) {
      goto rmt_cleanup2;
    }
  }

  return tbl;

  rmt_cleanup2:
  for(i--;i >= 0;i--) {
    free(tbl->metrics[i]);
  }

  free(tbl->metrics);
  
  rmt_cleanup:
  free(tbl);

  return NULL;
}

static pcf_bitmap_table_t *_read_bitmap_table(pcfio_t *pio, pcf_toc_entry_t *te)
{
  pcf_bitmap_table_t *tbl;
  uint32_t format;
  uint32_t *offsets = NULL;
  int bitmaps_size;
  uint8_t *bitmaps = NULL;
  int i;

  tbl = malloc(sizeof(*tbl));

  if(!tbl) {
    errno = ENOMEM;
    return NULL;
  }

  pcfio_seekto(pio, te->offset);
  if(pcfio_haserror(pio)) {
    goto cleanup;
  }
  
  format = pcfio_readlsb32(pio);
  pcfio_set_format(pio, format);

  tbl->glyph_count = pcfio_read32(pio);
  if(pcfio_haserror(pio)) {
    goto cleanup;
  }

  offsets = malloc(sizeof(*offsets) * tbl->glyph_count);
  if(!offsets) {
    goto cleanup;
  }

  for(i = 0; i < tbl->glyph_count; i++) {
    offsets[i] = pcfio_read32(pio);
  }

  if(pcfio_haserror(pio)) {
    goto cleanup;
  }

  for(i = 0; i < 4; i++) {
    if(i == PCF_GLYPH_PAD_INDEX(format)) {
      bitmaps_size = pcfio_read32(pio);
    } else {
      pcfio_read32(pio);
    }
  }

  // Empty fonts go to hell
  if(pcfio_haserror(pio) || bitmaps_size <= 0) {
    goto cleanup;
  }

  bitmaps = malloc(bitmaps_size);

  if(!bitmaps) {
    goto cleanup;
  }

  pcfio_read_block_rel(pio, (char *)bitmaps, bitmaps_size);

  // Byte order transformation if needed.
  // transformation converts to big endian because it is
  // possible to work with it as with just bytes
  switch(PCF_SCAN_UNIT(format)) {
    case 2:
      pcfio_swap_to_be2((uint16_t*)bitmaps, bitmaps_size, PCF_BYTE_ORDER(format));
      break;
    case 4:
      pcfio_swap_to_be4((uint32_t*)bitmaps, bitmaps_size, PCF_BYTE_ORDER(format));
      break;
  }

  tbl->offsets = offsets;
  tbl->bitmap_data = bitmaps;

  return tbl;
  
  cleanup:
  free(offsets);
  free(tbl);
  free(bitmaps);
  return NULL;
}

static pcf_enc_table_t *_read_enc_table(pcfio_t *pio, pcf_toc_entry_t *te)
{
  pcf_enc_table_t *tbl = NULL;
  uint32_t format;
  int i, j;
  uint16_t min1, max1, min2, max2, default_char;

  tbl = malloc(sizeof(*tbl));

  if(!tbl) {
    errno = ENOMEM;
    return NULL;
  }

  pcfio_seekto(pio, te->offset);
  if(pcfio_haserror(pio)) {
    goto cleanup;
  }
  
  format = pcfio_readlsb32(pio);
  pcfio_set_format(pio, format);

  min2 = pcfio_read16(pio);
  max2 = pcfio_read16(pio);
  min1 = pcfio_read16(pio);
  max1 = pcfio_read16(pio);
  default_char = pcfio_read16(pio);

  for(i = 0; i < 256 * 256; i++) {
    tbl->glyphindeces[i] = default_char;
  }

  //flat case
  if(min1 == 0 && max1 == 0) {
    for(i = min2; i <= max2; i++) {
      tbl->glyphindeces[i] = pcfio_read16(pio);
    }
  } else {
    for(i = min1; i <= max1; i++) {
      for(j = min2; j <= max2; j++) {
        tbl->glyphindeces[i * 256 + j] = pcfio_read16(pio);
      }
    }
  }

  if(pcfio_haserror(pio)) {
    goto cleanup;
  }

  return tbl;

  cleanup:
  
  free(tbl);
  
  return NULL;
}


pcf_font_t *pcf_read_font(const char *path)
{
  int r, i;
  pcf_font_t *res = NULL;
  pcfio_t *pio = NULL;
  int table_count;
  pcf_toc_entry_t *entries = NULL;

  pio = malloc(sizeof(*pio));
  if(!pio) {
    return NULL;
  }
  
  if(pcfio_init(path, pio) != 0) {
    free(pio);
    return NULL;
  }

  res = malloc(sizeof(*res));

  if(!res) return NULL;

  memset(res, 0, sizeof(*res));
  
  for(i = 0; i < 4; i++) {
    char a, b;
    a = pcf_magic[i];
    b = pcfio_read_char(pio);
    if(a != b) {
      errno = EMAGIC;
      return NULL;
    }
  }

  table_count = pcfio_readlsb32(pio);

  if(pcfio_haserror(pio)) {
    errno = 234;
    goto cleanup;
  }

  entries = malloc(table_count * sizeof(*entries));
  if(!entries) {
    errno = ENOMEM;
    goto cleanup;
  }

  for(i = 0; i < table_count; i++) {
    r = __read_toc_entry(pio, &entries[i]);
    if(r != 0) {
      errno = EIO;
      goto cleanup;
    }
  }

  // Read some tables  
  for(i = 0; i < table_count; i++) {
    if(pcfio_haserror(pio)) {
      goto cleanup;
    }
    switch(entries[i].type) {
      case PCF_PROPERTIES:
//        res->prop_table = _read_prop_table(pio, &entries[i]);
        break;
      case PCF_ACCELERATORS:
//        res->accel_table = _read_accel_table(pio, &entries[i]);
        break;        
      case PCF_METRICS:
        res->metrics_table = _read_metrics_table(pio, &entries[i]);
        break;
      case PCF_BITMAPS:
        res->bitmap_table = _read_bitmap_table(pio, &entries[i]);
        break;
      case PCF_INK_METRICS:
        break;
      case PCF_BDF_ENCODINGS:
        res->enc_table = _read_enc_table(pio, &entries[i]);
        break;
      case PCF_SWIDTHS:
        break;
      case PCF_GLYPH_NAMES:
        break;
      case PCF_BDF_ACCELERATORS:
        break;
    }
  }

  return res;
  
  cleanup:
  free(res);
  pcfio_close(pio);
  free(pio);
  return NULL;
}

uint16_t pcf_get_char_index(pcf_font_t *fnt, uint16_t code)
{
  return fnt->enc_table->glyphindeces[code];
}

pcf_metrics_t *pcf_get_metrics(pcf_font_t *fnt, uint16_t index)
{
  if(index >= fnt->metrics_table->metrics_count) {
    return NULL;
  } else {
    return fnt->metrics_table->metrics[index];
  }
}

uint8_t *pcf_get_bitmap(pcf_font_t *fnt, uint16_t index)
{
  if(index >= fnt->bitmap_table->glyph_count) {

    return NULL;
  } else {
    int offset = fnt->bitmap_table->offsets[index];
    return &fnt->bitmap_table->bitmap_data[offset];
  }
}
