#include <sys/types.h>
#include <errno.h>
#include <stdlib.h>

#include <sys/kcontrol.h>

#include "internal.h"

static uint8_t *pcf_get_bitmap(wchar_t ch);
static int pcf_init(const char *path);
static void pcf_destroy(void);


font_ops_t pcf_ops = {
  .init = pcf_init,
  .destroy = pcf_destroy,
  .get_bitmap = pcf_get_bitmap,
};

#include "pcf.h"
#include "pcfio.h"

uint16_t encoding[256 * 256];

int       glyph_count;
uint32_t *offsets = NULL;
uint8_t  *bitmaps = NULL;

const char pcf_magic[] = "\1fcp";

static int read_toc_entry(pcfio_t *pio, pcf_toc_entry_t *te)
{
  te->type = pcfio_readlsb32(pio);
  te->format = pcfio_readlsb32(pio);
  te->size = pcfio_readlsb32(pio);
  te->offset = pcfio_readlsb32(pio);
  if(pcfio_haserror(pio)) {
    return -1;
  }
  return 0;
}

static int read_compr_metrics(pcfio_t *pio, pcf_metrics_t *met)
{
  met->left_side_bearing = pcfio_read_char(pio) - 0x80;
  met->right_side_bearing = pcfio_read_char(pio) - 0x80;
  met->character_width = pcfio_read_char(pio) - 0x80;
  met->character_ascent = pcfio_read_char(pio) - 0x80;
  met->character_descent = pcfio_read_char(pio) - 0x80;
  met->character_attributes = 0;

  if(pcfio_haserror(pio)) {
    return -1;
  }
  return 0;
}

static int read_uncompr_metrics(pcfio_t *pio, pcf_metrics_t *met)
{
  met->left_side_bearing = pcfio_read16(pio);
  met->right_side_bearing = pcfio_read16(pio);
  met->character_width = pcfio_read16(pio);
  met->character_ascent = pcfio_read16(pio);
  met->character_descent = pcfio_read16(pio);
  met->character_attributes = pcfio_read16(pio);
  
  if(pcfio_haserror(pio)) {
    return -1;
  }
  return 0;
}

static int check_accels(pcfio_t *pio, pcf_toc_entry_t *te)
{
  uint32_t format;
  int no_overlap;
  int constant_metrics;
  int terminal_font;
  int constant_width;
  int ink_inside;

  pcfio_seekto(pio, te->offset);
  if(pcfio_haserror(pio)) {
    return -1;
  }
  
  format = pcfio_readlsb32(pio);
  pcfio_set_format(pio, format);

  no_overlap = pcfio_read_char(pio);
  constant_metrics = pcfio_read_char(pio);
  terminal_font = pcfio_read_char(pio);
  constant_width = pcfio_read_char(pio);
  ink_inside = pcfio_read_char(pio);

  // We need only monospace fonts.
  if(no_overlap && constant_metrics && terminal_font &&
     constant_width && ink_inside) {
    return 0;
  } else {
    return -1;
  }
}

// read only one metrics record as font is monospace and have constant metrics
static int read_metrics(pcfio_t *pio, pcf_toc_entry_t *te)
{
  uint32_t format;
  pcf_metrics_t met;
  int ret, metrics_count;

  pcfio_seekto(pio, te->offset);
  if(pcfio_haserror(pio)) {
    goto out;
  }
  
  format = pcfio_readlsb32(pio);
  pcfio_set_format(pio, format);

  if(format & PCF_COMPRESSED_METRICS) {
    metrics_count = pcfio_read16(pio);
  } else {
    metrics_count = pcfio_read32(pio);
  }

  if(metrics_count < 1) {
    goto out;
  }

  if(pcfio_haserror(pio)) {
    goto out;
  }

  if(format & PCF_COMPRESSED_METRICS) {
    ret = read_compr_metrics(pio, &met);
  } else {
    ret = read_uncompr_metrics(pio, &met);
  }

  if(ret != 0) {
    goto out;
  }
  
  pcf_ops.width = met.character_width;
  pcf_ops.height = met.character_ascent + met.character_descent;

  return 0;

  out:

  return -1;
}

static int read_bitmaps(pcfio_t *pio, pcf_toc_entry_t *te)
{
  uint32_t format;
  int bitmaps_size;
  int i;

  pcfio_seekto(pio, te->offset);
  if(pcfio_haserror(pio)) {
    goto cleanup;
  }
  
  format = pcfio_readlsb32(pio);
  pcfio_set_format(pio, format);

  glyph_count = pcfio_read32(pio);
  if(pcfio_haserror(pio)) {
    goto cleanup;
  }

  offsets = malloc(sizeof(*offsets) * glyph_count);
  if(!offsets) {
    goto cleanup;
  }

  for(i = 0; i < glyph_count; i++) {
    offsets[i] = pcfio_read32(pio);
  }

  if(pcfio_haserror(pio)) {
    goto cleanup;
  }

  pcf_ops.padding = PCF_GLYPH_PAD(format);
  
  for(i = 0; i < 4; i++) {
    if(i == PCF_GLYPH_PAD_INDEX(format)) {
      bitmaps_size = pcfio_read32(pio);
    } else {
      pcfio_read32(pio);
    }
  }

  // Empty fonts go to hell
  if(pcfio_haserror(pio) || bitmaps_size <= 0) {
    goto cleanup;
  }

  bitmaps = malloc(bitmaps_size);

  if(!bitmaps) {
    goto cleanup;
  }

  pcfio_read_block_rel(pio, (char *)bitmaps, bitmaps_size);

  // Byte order transformation if needed.
  // transformation converts to big endian because it is
  // possible to work with it as with just bytes
  switch(PCF_SCAN_UNIT(format)) {
    case 2:
      pcfio_swap_to_be2((uint16_t*)bitmaps, bitmaps_size, PCF_BYTE_ORDER(format));
      break;
    case 4:
      pcfio_swap_to_be4((uint32_t*)bitmaps, bitmaps_size, PCF_BYTE_ORDER(format));
      break;
  }

  return 0;
  
  cleanup:
  free(offsets);
  free(bitmaps);
  return -1;
}

static int read_encodings(pcfio_t *pio, pcf_toc_entry_t *te)
{
  uint32_t format;
  int i, j;
  uint16_t min1, max1, min2, max2, default_char;


  pcfio_seekto(pio, te->offset);
  if(pcfio_haserror(pio)) {
    goto cleanup;
  }
  
  format = pcfio_readlsb32(pio);
  pcfio_set_format(pio, format);

  min2 = pcfio_read16(pio);
  max2 = pcfio_read16(pio);
  min1 = pcfio_read16(pio);
  max1 = pcfio_read16(pio);
  default_char = pcfio_read16(pio);

  for(i = 0; i < 256 * 256; i++) {
    encoding[i] = default_char;
  }

  //flat case
  if(min1 == 0 && max1 == 0) {
    for(i = min2; i <= max2; i++) {
      encoding[i] = pcfio_read16(pio);
    }
  } else {
    for(i = min1; i <= max1; i++) {
      for(j = min2; j <= max2; j++) {
        encoding[i * 256 + j] = pcfio_read16(pio);
      }
    }
  }

  if(pcfio_haserror(pio)) {
    goto cleanup;
  }

  return 0;

  cleanup:
  
  return -1;
}

#define ONCE(x)                                 \
  if((x)) {                                     \
    goto cleanup;                               \
  } else {                                      \
    (x) = 1;                                    \
  }

#define CHECK(x)                                \
  if((x) != 0) {                                \
    goto cleanup;                               \
  }


// TODO: add some sanity checks to forbid char indices
// above number of glyphs etc...
static int pcf_read_font(const char *path)
{
  int r, i;
  pcfio_t *pio = NULL;
  int table_count;
  pcf_toc_entry_t *entries = NULL;
  int has_bitmaps = 0, has_metrics = 0,
    has_encodings = 0, has_accels = 0;

  pio = malloc(sizeof(*pio));
  if(!pio) {
    return -1;
  }
  
  if(pcfio_init(path, pio) != 0) {
    free(pio);
    return -1;
  }

  for(i = 0; i < 4; i++) {
    char a, b;
    a = pcf_magic[i];
    b = pcfio_read_char(pio);
    if(a != b) {
      return -1;
    }
  }

  table_count = pcfio_readlsb32(pio);

  if(pcfio_haserror(pio)) {
    errno = 234;
    goto cleanup;
  }

  entries = malloc(table_count * sizeof(*entries));
  if(!entries) {
    errno = ENOMEM;
    goto cleanup;
  }

  for(i = 0; i < table_count; i++) {
    r = read_toc_entry(pio, &entries[i]);
    if(r != 0) {
      errno = EIO;
      goto cleanup;
    }
  }

  // Read some tables  
  for(i = 0; i < table_count; i++) {
    if(pcfio_haserror(pio)) {
      goto cleanup;
    }
    switch(entries[i].type) {
      case PCF_ACCELERATORS:
        ONCE(has_accels);
        CHECK(check_accels(pio, &entries[i]))
        break;        
      case PCF_METRICS:
        ONCE(has_metrics);
        CHECK(read_metrics(pio, &entries[i]));
        break;
      case PCF_BITMAPS:
        ONCE(has_bitmaps);
        CHECK(read_bitmaps(pio, &entries[i]))
        break;
      case PCF_BDF_ENCODINGS:
        ONCE(has_encodings);
        CHECK(read_encodings(pio, &entries[i]));
        break;
    }
  }

  pcfio_close(pio);
  free(entries);
  
  return 0;
  
  cleanup:
  pcfio_close(pio);
  free(entries);
  pcf_destroy();
  
  return -1;
}


static int pcf_init(const char *path)
{
  return pcf_read_font(path);
}

static uint8_t *pcf_get_bitmap(wchar_t ch)
{
  uint32_t index;

  //no martian characters yet...
  if(ch & 0xFFFF0000L) {
    return NULL;
  }
  
  index = encoding[ch & 0xFFFF];

  return &bitmaps[offsets[index]];
}

static void pcf_destroy(void)
{
  free(bitmaps);
  free(offsets);
  bitmaps = NULL;
  offsets = NULL;
}
