#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <int10/int10.h>
#include <sys/kcontrol.h>
#include <stdbool.h>

#include "internal.h"
#include "palette.h"

static uint8_t *screen = NULL;
static uint8_t *screen_buffer = NULL;
static uint8_t *palette;
static uint8_t *background = NULL;
//static int vesa_ops.bytes_per_pixel;

static bool large_mem;

ulong_t video_mem_size;

static int vesa_init(void);
static void vesa_draw_bitmap(int x, int y, int fgcolor, int bgcolor, uint8_t *bitmap,
                             int width, int height, int padding);
static void vesa_invert_rect(int x, int y, int width, int height);
static void vesa_draw_pixmap(int x, int y, pixmap_t *pixmap);
static void vesa_scroll_area(int startline, int stopline, int lines, int color);
static void vesa_fill_rect(int x, int y, int width, int height, int color);
static void vesa_sync_rect(int x, int y, int width, int height);
static int vesa_set_bg(uint8_t *bg);


viewport_ops_t vesa_ops = {
  .init = vesa_init,
  .destroy = NULL,
  .draw_bitmap = vesa_draw_bitmap,
  .invert_rect = vesa_invert_rect,
  .draw_pixmap = vesa_draw_pixmap,
  .scroll_area = vesa_scroll_area,
  .fill_rect = vesa_fill_rect,
  .sync_rect = vesa_sync_rect,
  .set_bg = vesa_set_bg
};

//Common macros:

#define SCREEN_BYTES                                    \
  (vesa_ops.width * vesa_ops.height * vesa_ops.bytes_per_pixel)

//Take palette color
#define COLOR(n)                                \
  (&palette[(n) * vesa_ops.bytes_per_pixel])

#define PIXEL(buf, x, y)                                            \
  (&buf[((x) + (y) * vesa_ops.width) * vesa_ops.bytes_per_pixel])

#define BUF_PIXEL(x, y)                         \
  PIXEL(screen_buffer, x, y)

#define SCREEN_PIXEL(x, y)                      \
  PIXEL(screen, x, y)

#define BG_PIXEL(x, y)                          \
  PIXEL(background, x, y)

// TODO: make it clearer.
#define PUT_PIXEL(dest,src)                              \
  {                                                      \
    if(vesa_ops.bytes_per_pixel == 2) {                  \
      *(uint16_t *)(dest) = *(uint16_t *)(src);          \
    } else if(vesa_ops.bytes_per_pixel == 4) {           \
      *(uint32_t *)(dest) = *(uint32_t *)(src);          \
    } else if(vesa_ops.bytes_per_pixel == 1) {           \
      *(uint8_t *)(dest) = *(uint8_t *)(src);            \
    } else if(vesa_ops.bytes_per_pixel == 3) {           \
      *(uint16_t *)(dest) = *(uint16_t *)(src);          \
      *(((uint8_t *)dest) + 2) =                         \
        *(((uint8_t *)src) + 2);                        \
    }                                                    \
  }

// I tried to optimize it as possible so it calls no functions.
// there also is no bound checking, we hope upper level checks this
// and we don't want to draw shedded bitmaps.
static void vesa_draw_bitmap(int x, int y, int fgcolor, int bgcolor, uint8_t *bitmap,
                        int width, int height, int padding)
{
  int dx, dy;
  uint8_t *scanline;
  uint8_t *dest;
  uint8_t *fgpix, *bgpix;
  int pwidth = PAD((width + 7)/8, padding);

  fgpix = COLOR(fgcolor);
  if(bgcolor >= 0) {
    bgpix = COLOR(bgcolor);
  }
  
  for(dy = 0; dy < height; dy++) {
    scanline = &bitmap[dy * pwidth];
    dest = BUF_PIXEL(x, y + dy);
    for(dx = 0; dx < width; dx++) {
      if(scanline[dx / 8] & (1 << (7 - (dx % 8)))) {
        PUT_PIXEL(dest, fgpix);
      } else if(bgcolor >= 0) {
        PUT_PIXEL(dest, bgpix);
      }
      dest += vesa_ops.bytes_per_pixel;
    }
  }
}

static void vesa_fill_rect(int x, int y, int width, int height, int color)
{
  int dx, dy;
  uint8_t *dest;
  uint8_t *pix;

  pix = COLOR(color);
  
  for(dy = 0; dy < height; dy++) {
    dest = BUF_PIXEL(x, dy + y);
    for(dx = 0; dx < width; dx++) {
      PUT_PIXEL(dest, pix);
      dest += vesa_ops.bytes_per_pixel;
    }
  }
}


#define CHECK_COLOR(rs,gs,bs,ro,go,bo)                 \
  (((rs) == mode_info->red_mask_size)        &&        \
   ((gs) == mode_info->green_mask_size)      &&        \
   ((bs) == mode_info->blue_mask_size)       &&        \
   ((ro) == mode_info->red_field_position)   &&        \
   ((go) == mode_info->green_field_position) &&        \
   ((bo) == mode_info->blue_field_position))

static int vesa_init(void)
{
  vbe_ctrl_t *ctrl_info = NULL;
  vbe_mode_t *mode_info = NULL;
  unsigned int px, py;

  int r;

  r = int10_init();
  if(r) {
    klog("Failed to init int10 lib\n");
    return -1;
  }
  
  r = int10_set_vbe_mode(VESA_MODE_800x600x16, &ctrl_info, &mode_info);
    
  if(mode_info == NULL || r != 0) {
    klog("Failed to set vbe mode\n");
    goto out;
  }

  vesa_ops.height   = mode_info->y_resolution;
  vesa_ops.width    = mode_info->x_resolution;
  vesa_ops.bytes_per_pixel   = PAD(mode_info->bits_per_pixel, 8) / 8;

  // Determine palette type:

  switch(vesa_ops.bytes_per_pixel) {
    case 1:
      // TODO: add support;
      goto out;
    case 2:
      if(CHECK_COLOR(5, 6, 5, 11, 5, 0)) {
        palette = (uint8_t *)vesa_565_palette;
        vesa_ops.palette_transform = rgb_to_565_transform;
        break;
      } else if(CHECK_COLOR(5, 5, 5, 10, 5, 0)) {
        palette = (uint8_t *)vesa_555_palette;
        vesa_ops.palette_transform = rgb_to_555_transform;
        break;
      }
      // we have no such palette
      goto out;
    case 3:
      // TODO: add support;
      goto out;
    case 4:
      if(CHECK_COLOR(8, 8, 8, 16, 8, 0)) {
        palette = (uint8_t *)vesa_0888_palette;
        vesa_ops.palette_transform = rgb_to_0888_transform;
      }
      goto out;
  }

  video_mem_size = ctrl_info->total_memory << 16;
  
  if(video_mem_size > SCREEN_BYTES) {
    large_mem = true;
  } else {
    video_mem_size = SCREEN_BYTES;
    large_mem = false;
  }

  if(video_mem_size < SCREEN_BYTES) {
    goto out;
  }

  screen = mmap(NULL, video_mem_size, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PHYS | MAP_PRIVATE,
                NOFD, mode_info->phys_base);

  if(!screen) {
    klog("Cannot map screen buffer\n");
    goto out;
  }

  //TODO: revome this black magic...

  px = 0;
  py = 0;

  if(large_mem) {
    int10_pan(&px, &py);
  }
 
  screen_buffer = malloc(SCREEN_BYTES);

  if(!screen_buffer) {
    goto out;
  }

  //
  
  int10_done();
  return 0;
  
  out:
  free(screen_buffer);
  if(screen) {
    munmap(screen, video_mem_size);
  }
  int10_done();

  return -1;
}

// TODO: optimize
static void vesa_invert_rect(int x, int y, int width, int height)
{
  int dx, dy;
  for(dx = x; dx < x + width; dx++) {
    for(dy = y; dy < y + height; dy++) {
      *BUF_PIXEL(dx,dy) = ~(*BUF_PIXEL(dx,dy));
    }
  }
}

static void vesa_draw_pixmap(int x, int y, pixmap_t *pixmap)
{
// TODO: doo-bee-doo-bee-doo
}

static void vesa_scroll_area(int startline, int stopline, int lines, int color)
{
  uint8_t *m = &screen_buffer[startline * vesa_ops.width * vesa_ops.bytes_per_pixel];
  int dir = (lines > 0) ? 1 : -1;
  int len = stopline - startline;
  int amount;
  uint8_t *block;

  if(lines < 0) lines = -lines;
  
  amount = vesa_ops.width * vesa_ops.bytes_per_pixel * (len - lines);
  block = &m[lines * vesa_ops.width * vesa_ops.bytes_per_pixel];

  if(dir > 0) {
    memcpy(m, block, amount);
    vesa_fill_rect(0, stopline-lines, vesa_ops.width, lines, color);
  } else {
    memmove(block, m, amount);
    vesa_fill_rect(0, startline, vesa_ops.width, lines, color);
  }

  return;
}

static void vesa_sync_rect(int x, int y, int width, int height)
{
  int dy;
  for(dy = y; dy < y + height; dy++) {
    memcpy(SCREEN_PIXEL(x, dy), BUF_PIXEL(x, dy), width * vesa_ops.bytes_per_pixel);
  }
}

static void vesa_sync_rect_bg(int x, int y, int width, int height);

static int vesa_set_bg(uint8_t *bg)
{
  background = bg;
  vesa_ops.sync_rect = vesa_sync_rect_bg;
  return 0;
}

// TODO: make it ten times clearer.
#define MERGE_PIXEL(bg,dest,src)                         \
  {                                                      \
    if(vesa_ops.bytes_per_pixel == 2) {                  \
      *(uint16_t *)(dest) = (*(uint16_t *)(src)) ?       \
        *(uint16_t *)(src) : *(uint16_t *)(bg);          \
    } else if(vesa_ops.bytes_per_pixel == 4) {           \
      *(uint32_t *)(dest) = (*(uint32_t *)(src)) ?       \
        *(uint32_t *)(src) : *(uint32_t *)(bg);          \
    } else if(vesa_ops.bytes_per_pixel == 1) {           \
      *(uint8_t *)(dest) = (*(uint8_t *)(src)) ?         \
        *(uint8_t *)(src) : *(uint8_t *)(bg);            \
    } else if(vesa_ops.bytes_per_pixel == 3) {           \
      if(*(uint16_t *)src || *(((uint8_t *)src) + 2)) {  \
        *(uint16_t *)(dest) = *(uint16_t *)(src);        \
        *(((uint8_t *)dest) + 2) =                       \
          *(((uint8_t *)src) + 2);                       \
      } else {                                           \
        *(uint16_t *)(dest) = *(uint16_t *)(bg);         \
        *(((uint8_t *)dest) + 2) =                       \
          *(((uint8_t *)bg) + 2);                        \
      }                                                  \
    }                                                    \
  }


// background-aware sync variant;
// we tread black (zero) as transparent color
static void vesa_sync_rect_bg(int x, int y, int width, int height)
{
  int dy, dx;
  uint8_t *dest;
  uint8_t *src, *bg;

  for(dy = y; dy < y + height; dy++) {
    bg = BG_PIXEL(x, dy);
    src = BUF_PIXEL(x, dy);
    dest = SCREEN_PIXEL(x, dy);
    for(dx = 0; dx < width; dx++) {
      MERGE_PIXEL(bg, dest, src);
      src += vesa_ops.bytes_per_pixel;
      dest += vesa_ops.bytes_per_pixel;
      bg += vesa_ops.bytes_per_pixel;
    }
  }
}
