/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 *
 * vgacons.d: Generic VGA console driver
 * hwops.c: hardware ops
 *
 */

#include <stdbool.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ioport.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <ipc/port.h>
#include <ns/ns.h>
#include <ipc/rpc.h>
#include <devio/devio.h>

#include "vga.h"
#include "hwops.h"

union colb {
  uint8_t val;
  struct {
    uint8_t bg:4;
    uint8_t fg:4;
  };
};

static vgadev_t __vga;
static vgaregs_t __vregs;

/* function prototypes */
static void vga_scroll(void);
static void vga_putch(uint16_t c);
static void vga_set_cursor_attrs(const uint8_t flags);
static void vga_update_cursor(int x,int y);

static void vga_set_x(const uint16_t x) {  __vga.x = x;   }
static void vga_set_y(const uint16_t y) {  __vga.y = y;   }
/*static void vga_set_fg(vga_color_t fg)  {  __vga.fg = fg; }
  static void vga_set_bg(vga_color_t bg)  {  __vga.bg = bg; }*/
static void vga_blinking(const bool action);

static inline void vga_set_attr(uint16_t* out, uint8_t col)
{
  union __vga_catr b;

  b.val=*out;
  b.bg=col >> 4;
  b.fg=col & 0xF;

  *out=b.val;
}

static void *scr_buf=NULL;

void vga_init(void *buf)
{
  int __attr=0;
  status_t ret;  

  __vga.base = (uint16_t *)buf;
  __vga.cols = 80;
  __vga.rows = 25;
  __vga.x = 0;
  __vga.y = 25;
  __vga.bg = VC_BLACK;
  __vga.fg = VC_WHITE;
  __vga.flags = 0;
  __vga.flags |= USCROLL;
  memset(&__vregs, 0, sizeof(__vregs));
  __vga.es_char = -1;
  /* TODO: take a check */
  ret = alloc_ioport_range(0x3d4,2);
  ret |= alloc_ioport_range(0x3da,1);
  ret |= alloc_ioport_range(0x3c0,1);
  if (ret) {
    abort();
  }
             
  vga_set_cursor_attrs(VGA_CRSR_HIGH);
  if(__attr!=0) {
    vga_cursor(true);
    vga_blinking(true);
  }
  scr_buf = malloc(4096);
  if(!scr_buf) {
    abort();
  }
  //vga_cls();
}

static void vga_update_cursor(int x,int y)
{
  uint16_t where, crt_dp, crt_ap;
  uint16_t crt_ar;

  where=y * __vga.cols + x ;
  crt_dp=vga_set_mode(VDP_CRT, VGA_COLOR); /* CRT data port */
  crt_ap=vga_set_mode(VAP_CRT, VGA_COLOR); /* CRT address port */

  /* i/o all */
  crt_ar = inb(crt_ap);
  outb(crt_ap, CLHR);
  __vregs.crtc[CLHR] = where >> 8;
  outb(crt_dp,__vregs.crtc[CLHR]);
  outb(crt_ap,CLLR);
  __vregs.crtc[CLLR]=where & 0xFFFFFFFF;
  outb(crt_dp,__vregs.crtc[CLLR]);
  outb(crt_ap,crt_ar);
}

void vga_cursor(const bool action)
{
  uint16_t crt_dp,crt_ap,crt_ar;

  crt_dp=vga_set_mode(VDP_CRT,VGA_COLOR);
  crt_ap=vga_set_mode(VAP_CRT,VGA_COLOR);

  crt_ar=inb(crt_ap);
  outb(crt_ap,CSR);
  __vregs.crtc[CSR]=inb(crt_dp);

  if(!action && !(__vregs.crtc[CSR] & VGA_CDISABLE))
    outb(crt_dp, __vregs.crtc[CSR] | VGA_CDISABLE);
  if(action && (__vregs.crtc[CSR] & VGA_CDISABLE))
    outb(crt_dp, __vregs.crtc[CSR] & VGA_CENABLE);
  outb(crt_ap, crt_ar);

  return;
}

static void vga_set_cursor_attrs(const uint8_t flags)
{
  uint16_t crt_dp, crt_ap, crt_ar;

  crt_dp=vga_set_mode(VDP_CRT, VGA_COLOR);
  crt_ap=vga_set_mode(VAP_CRT, VGA_COLOR);  
  crt_ar=inb(crt_ap);
  
  if(flags & VGA_CRSR_LOW) {
    vga_putch('C');
    outb(crt_ap, CSR);
    __vregs.crtc[CSR]=inb(crt_dp);
    outb(crt_dp, 0x0A);
  }
  
  if(flags & VGA_CRSR_HIGH){
    vga_putch('G');
    outb(crt_ap, CER);
    __vregs.crtc[CER]=inb(crt_dp);
    outb(crt_dp, 0x0B);
  }
  outb(crt_ap, crt_ar);

  return;
}

static void vga_blinking(const bool action)
{
  uint16_t isr1_dp, isr1_save, atr_save;
  
  isr1_dp=vga_set_mode(VDP_ISR1,VGA_COLOR);
  isr1_save=inb(isr1_dp);
  atr_save=inb(VDP_ATR_RD);
  outb(VAP_ATR_WR,AMCR);
  __vregs.atr[AMCR]=inb(VDP_ATR_RD);

  if(action && !(__vregs.atr[AMCR] & VGA_ABLINK_ENABLE))
    outb(VAP_ATR_WR,__vregs.atr[AMCR] | VGA_ABLINK_ENABLE);
  if(!action && (__vregs.atr[AMCR] & VGA_ABLINK_ENABLE))
    outb(VAP_ATR_WR,__vregs.atr[AMCR] & VGA_ABLINK_DISABLE);

  outb(VAP_ATR_WR, atr_save);
  outb(isr1_dp, isr1_save);

  return;
}

static void vga_putch(uint16_t c)
{
  uint16_t *where;

  switch(c) {
  case BACKSPACE:
    if(__vga.x!=0) {
      __vga.x--;
      c=SPACE;
      where=(uint16_t *)(__vga.base+__vga.cols*__vga.y+__vga.x);
      *where = c;
    }
    break;
  case TAB:
    __vga.x+=8 & ~(8 - 1);
    break;
  case '\r':
    __vga.x=0;
    break;
  case '\n':
    __vga.x=0; /* increment to 0 later */
    __vga.y++;
    break;
  default:
    if(c < ' ')
      break;
    where=(uint16_t *)(__vga.base+__vga.cols*__vga.y+__vga.x);
    *where = c;
    __vga.x++;
    break;
  }

  if(__vga.flags & USCROLL)
    vga_scroll();
  else if(__vga.y > __vga.rows)
    __vga.y=__vga.rows;
  else if(__vga.x > __vga.cols)
    __vga.x=__vga.cols;

  return;
}

static void uniput(uint16_t cc)
{
  uint16_t *where;
  char c=cc;

  switch(c) {
  case BACKSPACE:
    if(__vga.x!=0) {
      __vga.x--;
      cc=SPACE;
      where=(uint16_t *)(__vga.base+__vga.cols*__vga.y+__vga.x);
      *where = cc;
    }
    break;
  case TAB:
    __vga.x+=8 & ~(8 - 1);
    break;
  case '\r':
    __vga.x=0;
    break;
  case '\n':
    __vga.x=0; /* increment to 0 later */
    __vga.y++;
    break;
  default:
    where=(uint16_t *)(__vga.base+__vga.cols*__vga.y+__vga.x);
    *where = cc;
    __vga.x++;
    break;
  }

  if(__vga.flags & USCROLL)
    vga_scroll();
  else if(__vga.y > __vga.rows)
    __vga.y=__vga.rows;
  else if(__vga.x > __vga.cols)
    __vga.x=__vga.cols;

  return;
}

static void vga_clear_rect(int x1, int y1, int x2, int y2, uint8_t col)
{
  uint16_t out;
  int i, j;
  
  // sanity check; we don't fix coordinates misplacement, only fatal errors prevented.
  if(x1 < 0 || x2 < 0 || y1 < 0 || y2 < 0) return;
  if(x1 > __vga.cols-1 || x2 > __vga.cols-1) return;
  if(y1 > __vga.rows-1 || y2 > __vga.rows-1) return;
  if(y1 > y2 || x1 > x2) return;
  
  out=SPACE;
  vga_set_attr(&out, col);
  
  for(i = x1; i <= x2; i++) {
    for(j = y1; j <= y2; j++) {
      __vga.base[j * __vga.cols + i] = out;
    }
  }
}

void vga_attr_init(void)
{
  int i, j;
  
  for(i = 0; i < __vga.cols; i++) {
    for(j = 0; j < __vga.rows; j++) {
      __vga.base[j * __vga.cols + i] &= 0xFF00;
      __vga.base[j * __vga.cols + i] |= 0x0070;
    }
  }
}

static void vga_scroll(void)
{
  uint16_t out,*vgam;
  int i=0;

  if(__vga.x>=__vga.cols) {
    __vga.y++;
    __vga.x=0;
  }

  if(__vga.y>=__vga.rows) {
    out=SPACE;
    memcpy((void *)__vga.base, (void *) (__vga.base + (__vga.cols)),2*((__vga.cols * __vga.rows)-__vga.cols));
    vgam=(uint16_t *)(__vga.base+__vga.cols*(__vga.rows-1));
    for(i=0;i<__vga.cols;i++)
      vgam[i]=out;
    __vga.y=__vga.rows-1;
  }

  return;
}

static void _puts(const char *msg,int cur)
{
  int i,s;

  if(!msg) return;

  s=strlen(msg);
  for(i=0;i<s;i++)    vga_putch((uint16_t)msg[i]);

  if(cur!=0)
    vga_update_cursor(__vga.x,__vga.y);

  return;
}

static void _puts_xy(char *msg,int x,int y)
{
  int xo=__vga.x,yo=__vga.y;

  if(!msg) return;

  vga_set_x(x);
  vga_set_y(y);
  _puts(msg,0);
  /* return old values */
  vga_set_x(xo);
  vga_set_y(yo);

  return;
}

static void _uniputs(cschar_t *msg,size_t len,int updcur)
{
  int i=0;
  union __vga_catr vc;
  union colb a;
  int ulen=len/sizeof(cschar_t);

  /* this driver doesn't support unicode */
  for(i=0;i<ulen;i++) {
    vc.s=msg[i].ch;
    a.val=msg[i].cattr;
    vc.fg=a.fg;
    vc.bg=a.bg;
    uniput(vc.val);
  }

//  if(updcur)
    vga_update_cursor(__vga.x,__vga.y);

  return;
}

static void _uniputsxy(cschar_t *buf,size_t len,int x,int y)
{
  int xo=__vga.x,yo=__vga.y;

  vga_set_x(x);
  vga_set_y(y);
  _uniputs(buf,len,0);
  /* return old values */
  vga_set_x(xo);
  vga_set_y(yo);

  return;
}

void vga_print(const char *msg)
{
  _puts(msg,1);
}

void vga_setcursor(int x, int y)
{
  __vga.x = x;
  __vga.y = y;
  if(x < 0) x = 0;
  if(y < 0) y = 0;
  if(__vga.y > __vga.rows)
    __vga.y=__vga.rows;
  if(__vga.x > __vga.cols)
    __vga.x=__vga.cols;
  vga_update_cursor(x, y);
}

void vga_printpos(char *msg,int x,int y)
{
  _puts_xy(msg,x,y);
}

void uni_print(cschar_t *buf,size_t len)
{
  _uniputs(buf,len,1);
}

void uni_printxy(cschar_t *buf,size_t len,int x,int y)
{
  _uniputsxy(buf,len,x,y);
}

/*  __vga.cols = 80;
    __vga.rows = 25;*/

size_t uni_width(void)
{
  return __vga.cols;
}

size_t uni_height(void)
{
  return __vga.rows;
}

void uni_scrollenable(void)
{
  __vga.flags |= USCROLL;

  return;
}

void uni_scrolldisable(void)
{
  __vga.flags &= ~USCROLL;

  return;
}

#define abs(x) ((x>0)?x:-x)

int uni_scroll_area(int startline, int stopline, int lineno)
{
  int start, stop, len, amount;
  int i;

  if(startline < 0 || stopline > __vga.rows-1) return -EINVAL;
  if(startline > stopline) return -EINVAL;
  amount = abs(lineno) * __vga.cols;
  start = startline * __vga.cols;
  stop = (stopline * __vga.cols + __vga.cols);
  len = stop - start - amount;
  if(len <= 0) return -EINVAL;

  if(lineno > 0) {
    memcpy(&((char*)__vga.base)[start*2], &((char*)__vga.base)[(start + amount)*2], len * 2);
    for(i = start + len; i < stop; i++) {
      ((uint16_t*)__vga.base)[i] = SPACE;
    }
  } else {
    memmove(&((char*)__vga.base)[(start+amount)*2], &((char*)__vga.base)[start*2], len * 2);
    for(i = start; i < start + amount; i++) {
      ((uint16_t*)__vga.base)[i] = SPACE;
    }
  }
  return 0;
}

void uni_scrolldown(int lineno)
{
  uint16_t out,*vgam;
  int i=0;

  memcpy((void *)__vga.base,(void*) (__vga.base + (__vga.cols*lineno)),2*(__vga.cols * (__vga.rows-lineno)));

  out=SPACE;
  vgam=(uint16_t *)(__vga.base+((__vga.rows-lineno)*__vga.cols));
  for(i=0;i<__vga.cols * lineno;i++) 
    vgam[i]=out;
  
  return;
}

void uni_scrollup(int lineno)
{
  uint16_t out,*vgam;
  int i=0;

  memcpy(scr_buf,(void*)__vga.base,(__vga.cols*(__vga.rows-lineno))*sizeof(uint16_t)); /* put to buffer */

  memcpy((void*) (__vga.base + (__vga.cols*lineno)),scr_buf,(__vga.cols*(__vga.rows-lineno))*sizeof(uint16_t));

  out=SPACE;
  vgam=(uint16_t *)__vga.base;
  for(i=0;i<__vga.cols * lineno;i++) 
    vgam[i]=out;

  return;
}

void uni_clear(void)
{
  vga_clear_rect(0, 0, __vga.cols-1, __vga.rows-1, 0);

  return;
}

void uni_clear_rect(int x1, int y1, int x2, int y2, uint8_t col)
{
  vga_clear_rect(x1, y1, x2, y2, col);

  return;
}
