/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 * (c) Copyright 2009 Graphov Paul <graphov@jarios.org>
 *
 * PC keyboard driver
 *
 * AK (madtirra@jarios.org) added *_GRUB option support [1-apr-2009]
 *
 */

#include <stdbool.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/ioport.h>
#include <sys/ievent.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <ipc/rpc.h>
#include <devio/devio.h>
#include <ns/ns.h>
#include <sys/scheduler.h>
#include <pthread.h>
#include <sys/uevent.h>
#include <time.h>
#include <dircons.h>
#include <lock/lock.h>

#include <sys/keyboard.h>
#include <chardev.h>
#include <datalib/circular.h>

#include "i8042.h"

#define KBD_IRQ  0x1

#define SKEY_RELEASE  0x80

#define KBDBUF_SIZE  256
#define CMDBUF_SIZE  128

#define DRVNAME  "pckeyb.d"
#define DEVNAME  "/dev/kbd"

volatile static ulong_t buf_sem = 0;

int i;
uint8_t kcode[2];
uint8_t kmask;
char *ipc_buf;

event_t kbd_event;
pthread_mutex_t i8042_lock;

circular_t kbdbuffer;

static status_t kbd_read(int devno, void *buf, size_t size, size_t offset, void **interim);
static status_t kbd_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize);

struct device_ops kbd_ops = {
  .read = kbd_read,
  .write = NULL,
  .ioctl = kbd_ioctl,
  .log = console_output,
};

//                 0   1   2   3   4   5   6   7   8   9   a   b   c   d   e   f
uint8_t ext1_table[128] = {0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   //0
                           0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 96, 97,  0,  0,   //1
                           0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 99,  0,  0,  0,  0,  0,   //2
                           0,  0,  0,  0,  0, 98,  0,  0,100,  0,  0,  0,  0,  0,  0,  0,   //3
                           0,  0,  0,  0,  0,  0,  0,102,103,104,  0,105,  0,106,  0,107,   //4
                         108,109,110,101,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0};   //5

typedef enum {
  NOEXT = 0,
  JUST_E0,
  PREV_E0,
  JUST_E1,
  PREV_E1,
  PPREV_E1,
} ext_state_t;

int ext_state = NOEXT;
  
/* local functions prototypes */
static bool kbd_irq_handler(void *data);
static void *kbd_iothread(void *data);
static void __init_hw_resources(void);
static void __init_kbd(void);

device_server_t *srv;

status_t device_work(void)
{
  int r;

  r=init_simple_mt_device_server(512, 512, DRVNAME, DEVNAME, &kbd_ops, 2);

  if(r!=0) {
    console_output("[kbd] failed to init device server\n");
    abort();
  }

  device_server_start();

  return 0;
}

int main(int argc,char **argv)
{
  int r;
  pthread_t thr;

#ifndef CONFIG_PCKEYB_GRUB
  start_service();
#endif

  __init_hw_resources();

  open_console();

  console_output("[pckeyb.d] Jari OS generic PC keyboard driver.\n");

  event_init(&kbd_event);
  pthread_mutex_init(&i8042_lock, NULL);

  r = init_i8042();
  if(r != 0) {
    console_output("pckeyb.d: failed to init i8042 controller, aborting.\n");
    abort();
  }

  r=pthread_create(&thr, NULL, kbd_iothread, NULL);
  if(r!=0) {
    abort();
  }

  device_work();

  return 0;
}

static status_t set_leds(uint8_t mask)
{
  status_t ret = 0;

  mask &= 7;
  
  pthread_mutex_lock(&i8042_lock);

  wait_i8042_out();
  send_to_kbd(0xED);
  wait_i8042_out();
  send_to_kbd(mask);

  pthread_mutex_unlock(&i8042_lock);

  return ret;
}

static status_t set_typematic(uint8_t mask)
{
  status_t ret = 0;

  mask &= 0x7F;
  
  pthread_mutex_lock(&i8042_lock);

  wait_i8042_out();
  send_to_kbd(0xF3);
  wait_i8042_out();
  send_to_kbd(mask);

  pthread_mutex_unlock(&i8042_lock);

  return ret;
}

/* implementation */

static status_t kbd_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize)
{
  if(type != KBD_IOCTL) {
    return -ENOSYS;
  }

  if(size < 1) {
    return -EINVAL;
  }

  switch(nr) {
    case SET_KB_LEDS:
      if(set_leds(*(uint8_t*)buf) == 0) {
        return 0;
      } else {
        return -EINVAL;
      }
    case SET_KB_TYPEMATIC:
      if(set_typematic(*(uint8_t*)buf) == 0) {
        return 0;
      } else {
        return -EINVAL;
      }
  }
  return -ENOSYS;
}

static status_t kbd_read(int devno, void *buf, size_t size, size_t offset, void **interim)
{
  size_t ret;
  
  while((ret = circular_read(&kbdbuffer, (char *)buf, size)) <= 0) {
    event_wait(&kbd_event);
  }
  
  return ret;
}

//TODO: No way to read scan codes now, should be implemented like KB_RAW
static bool kbd_irq_handler(void *data)
{
  uint8_t scode;
  uint8_t prev;

  while(i8042_in_ready()) {
    scode = read_data_i8042();
    switch(scode) {
        // Ext keys  
      case EXT1_VAL:
        ext_state = JUST_E0;
        break;
      case EXT2_VAL:
        ext_state = JUST_E1;
        break;
      case KB_ACK:
        continue;
      default:
        if(ext_state == NOEXT) {
          circular_putc(&kbdbuffer, scode);
          event_signal(&kbd_event);
        } else if(ext_state == PREV_E0) {
          if(scode & RELEASE_MASK) {
            circular_putc(&kbdbuffer, RELEASE_MASK|ext1_table[scode & ~RELEASE_MASK]);
            event_signal(&kbd_event);
          } else {
            circular_putc(&kbdbuffer, ext1_table[scode]);
            event_signal(&kbd_event);
          }
        } else if(ext_state == PPREV_E1) {
          if(prev == 0x1d && scode == 0x45) {
            circular_putc(&kbdbuffer, KC_PAUSE);
            event_signal(&kbd_event);
          } else if(prev == 0x9d && scode == 0xc5) {
            circular_putc(&kbdbuffer, KC_PAUSE|RELEASE_MASK);
            event_signal(&kbd_event);
          }
        }
    }
    switch(ext_state) {
      case JUST_E0:
        ext_state = PREV_E0;
        break;
      case JUST_E1:
        ext_state = PREV_E1;
        break;
      case PREV_E1:
        ext_state = PPREV_E1;
        prev = scode;
        break;
      default:
        ext_state = NOEXT;
        break;
    }
  }

  return true;
}

static void __init_kbd(void)
{
  uint8_t rr;

  /* flush all io */
  while(read_status_i8042() & INPT_BUF_FULL);

  /* test keyboard */
  write_command_i8042(0xaa);
  while(!(read_status_i8042() & OUT_BUF_FULL));
  
  rr=read_data_i8042();
  
  /* test interface */
  write_command_i8042(0xab);
  while(!(read_status_i8042() & OUT_BUF_FULL));
  rr=read_data_i8042();

  write_command_i8042(0x60); /* keyb cmd */
  while(read_status_i8042() & INPT_BUF_FULL);

  write_data_i8042(0x69); /* keyb cmd */
  while(read_status_i8042() & INPT_BUF_FULL);

  /* enable keyboard */
  write_data_i8042(0xf4);
  while(!(read_status_i8042() & 0x1));

  rr=read_data_i8042();

  return;
}

static void __init_hw_resources(void)
{
  circular_init(&kbdbuffer, KBDBUF_SIZE);

  return;
}

static void *kbd_iothread(void *data)
{
  irq_t irqs[]={KBD_IRQ};
  interrupt_event_t *e;

  __init_kbd();

  task_scheduler_control(0,0,SYS_SCHED_CTL_SET_PRIORITY,40);
  task_scheduler_control(0,0, SYS_SCHED_CTL_SET_POLICY, SCHED_FIFO);
  
  /* assign irq */
  if( interrupt_event_create(&e,irqs,1) ) {
    console_output("[kbd] failed to create interrupt handler. aborting.\n");
  __failed:
    abort();
  }

  if(interrupt_event_add_handler(e,KBD_IRQ,kbd_irq_handler,NULL) ) {
    console_output("[kbd] failed to add interrupt handler. aborting.\n");
    goto __failed;
  }

  while(true) {
    interrupt_event_wait(e);
  };

  return NULL;
}
