/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Tirra <madtirra@jarios.org>
 * (c) Copyright 2008 Dmitry Gromada <gromada82@gmail.com>
 * (c) Copyright 2009 Paul Graphov <graphov@jarios.org>
 *
 * Implements IO APIC support driver.
 *
 */

/*
 * TO DO: Implemets irq statisic update, get and set irq balancer
 *
 * replace logical destination to the physical destination
 */
/*
 * AK (madtirra@jarios.org): added *_GRUB option support [1-apr-2009]
 */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <bitwise.h>
#include <ipc/port.h>
#include <unistd.h>
#include <ns/ns.h>
#include <errno.h>
#include <lock/lock.h>
#include <devio/devio.h>
#include <devio/irq.h>
#include <devio/irqioctl.h>
#include <chardev.h>
#include <irqcon/irqcon.h>
#include <ipc/rpc.h>

#include <sys/console.h>
#include "ioapic.h"

#define DEVNAME "/dev/ioapic"
#define DRVNAME "ioapic.d"

#define BUFFER_SIZE 32
#define MAX_CLIENTS 32

static uint8_t *apic_phys_ids;
static ioapic_desc_t *ioapics;
static generic_irq_controller_t gencon;

static hw_direct_io_t *super;
static cschar_t *termbuffer;
static char msg_buf[512];

static status_t ioapic_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize);
static status_t ioapic_dio_read(int devno, void *buf, size_t size, size_t offset, void **interim);
void ascii_puts(const char *msg);

static struct device_ops ioapic_ops = {
  .read = ioapic_dio_read,
  .ioctl = ioapic_ioctl,
  .log = ascii_puts
};

static int ioapic_mask_irq(int irq, int mask);
static int ioapic_route_irq(int irq, int cpu);

static irqcon_hw_operations_t ioapic_operations = {
  .mask_irq = ioapic_mask_irq,
  .route_irq = ioapic_route_irq
};

void open_console(void)
{
  super=malloc(sizeof(*super));
  memset(super, 0, sizeof(*super));
  termbuffer = malloc(512);
  memset(termbuffer, 0, 512);
  hw_dio_open("/console",HWRDWR,super);
}

void ascii_puts(const char *msg)
{
  int len = strlen(msg);
  int i;

  for(i=0;i<len;i++) {
    termbuffer[i].ch=(wchar_t)((unsigned)msg[i]);
    termbuffer[i].cattr=0;
    termbuffer[i].xattr=0;
  }

  hw_dio_write(super,NULL,termbuffer,len*sizeof(cschar_t));
  return;
}

#define fmtout_console(msg, fmt...) \
{ \
  sprintf(msg_buf, msg, fmt);\
  ascii_puts(msg_buf);\
}

static int always_inline imin(int x, int y)
{
  return (x < y) ? x : y;
}


/*
 * NOTE: there are too black magic,
 * my attemption was to make full support 
 * in one day
 * Also, I didn't define a structure of io apic
 * in case of it's operational specific
 */

/* basic functions for read/write from/to IO APIC */
static  uint32_t ioapic_read(int chip, uint8_t addr)
{
  ioapic_regsel_t  rg;
  uint32_t *base = ioapics[chip].base;

  rg.value = base[IOAPICRG]; /* save parameter and init selector*/
  rg.addr = addr; /* set address need to be ridden */
  base[IOAPICRG] = rg.value; /* restore set */

  return base[IOAPICWIN]; /* return data */
}

static  void ioapic_write(int chip, uint8_t addr,uint32_t v)
{
  ioapic_regsel_t  rg;
  uint32_t *base = ioapics[chip].base;

  /* the logic the same like in read, but we're write our value */
  rg.value = base[IOAPICRG]; /* save parameter and init selector*/
  rg.addr = addr; /* set address need to be ridden */
  base[IOAPICRG] = rg.value; /* restore set */
  base[IOAPICWIN] = v;
}

/* set IO redirection table item */

static  void ioapic_set_ioredir(irq_desc_t *idesc)
{
  ioapic_redir_t  rd;
  int chip = idesc->line.chip;
  int pin = idesc->line.pin;

  /* read low bytes from redirection table for this virq */  
  rd.low=ioapic_read(chip, (uint8_t) (IOAPICRED + pin*2)); 
  /* read high bytes from redirection table for this virq */
  rd.high=ioapic_read(chip, (uint8_t) (IOAPICRED + pin*2 +1));
  
  if (idesc->targ_cpu >= 0)
    rd.dest = 1 << idesc->targ_cpu;
  else
    rd.dest = 0;
  rd.dmod = DMODE_LOGIC;
  rd.trig_mode= idesc->mode.trig;      /* set trigger mode to edge mode */
  rd.polarity= idesc->mode.polarity;  /* set high polarity */
  rd.txmod = idesc->mode.deliv;        /* set delivery mode (TX) */
  rd.vector = idesc->vector;          /* set destination vector */

  /* write low bytes from redirection table for this virq */  
  ioapic_write(chip, (uint8_t) (IOAPICRED + pin*2),rd.low); 
  /* write high bytes from redirection table for this virq */
  ioapic_write(chip, (uint8_t) (IOAPICRED + pin*2 +1),rd.high);
}

/* mask interrupt on io apic, disable interrupt */
static  void ioapic_disable_pin(int chip, uint8_t pin)
{
  ioapic_redir_t  rd;

  /* read low bytes from redirection table for this virq */  
  rd.low=ioapic_read(chip, (uint8_t) (IOAPICRED + pin*2));
  rd.mask=0x1;
  /* write low bytes from redirection table for this virq */
  ioapic_write(chip, (uint8_t) (IOAPICRED + pin*2),rd.low); 
}

/* enable interrupt */
static  void ioapic_enable_pin(int chip, uint8_t pin)
{
  ioapic_redir_t  rd;

  /* read low bytes from redirection table for this virq */  
  rd.low=ioapic_read(chip, (uint8_t) (IOAPICRED + pin*2)); 
  rd.mask=0x0;
  /* write low bytes from redirection table for this virq */
  ioapic_write(chip, (uint8_t) (IOAPICRED + pin*2),rd.low);
}

/* if number of chip is -1 disable all pins on all chips */
static  void ioapic_disable_all(int chip)
{
  int i;

  for (i = 0; i < ioapics[chip].npins; i++)
    ioapic_disable_pin(chip, i);
}

static  void ioapic_init(int chip)
{
  int i;
  ioapic_id_t  ioapic_id;
  ioapic_version_t  ioapic_ver;
  irq_desc_t *idesc;

  /* read version info
   * the register contains max entries minus 1 
   */ 
  ioapic_ver.value = ioapic_read(chip, IOAPIC_VERREG);
  ioapics[chip].npins = ioapic_ver.redir_entries + 1; 
  
  ioapic_id.id = chip;
  ioapic_write(chip, IOAPIC_IDREG, ioapic_id.value);

//  fmtout_console("[HW] init I/O APIC version %d, %d pins ...\n", ioapic_ver.version, ioapics[chip].npins);

  ioapic_disable_all(chip);   

  /* set default redirection table */
  for (i = 0, idesc = gencon.irqs; i < gencon.nr_irqs; i++, idesc++)
    ioapic_set_ioredir(idesc);
}


static void init_vector_map(void)
{
  init_bitmap(&gencon.vecmap, TOTAL_IOAPIC_VECTORS);
  bitrange_set_multi(&gencon.vecmap, BASE_IRQ_VECTOR, MAX_APIC_IRQ_VECTOR);
}


static int ioapic_mask_irq(int irq, int mask)
{
  irq_desc_t *idesc = gencon.irqs + irq;  

  if (mask) {
    ioapic_disable_pin(idesc->line.chip, idesc->line.pin);
    set_irq_flags(idesc, IRQ_MASKED);
  } else {
    ioapic_enable_pin(idesc->line.chip, idesc->line.pin);
    clear_irq_flags(idesc, IRQ_MASKED);
  }

  return 0;
}


static int ioapic_route_irq(int irq, int cpu)
{
  irq_desc_t *idesc = gencon.irqs + irq;

  if (idesc->targ_cpu == cpu)
     return 0;

  if (idesc->targ_cpu != cpu) {
    gencon.caps[idesc->targ_cpu].curr_num--;
    gencon.caps[cpu].curr_num++;
  }

  ioapic_set_ioredir(idesc);
  
  return 0;
}

static void init_driver(void)
{
  int i = 0, ncon = 1;
  pthread_mutexattr_t attr = {PTHREAD_PROCESS_PRIVATE};

  /* TO DO: get found system information, stored on the sysinfo server */

  /* fexme: get number of io apics from the mptable */
  ioapics = (ioapic_desc_t*)malloc(sizeof(ioapic_desc_t));
  if (ioapics == NULL)
     abort();

  /* fill based on the information from the acpi tables */
  ioapics[0].base = mmap(NULL, PAGE_SIZE, PROT_READ | PROT_WRITE,
                         MAP_PHYS | MAP_PRIVATE | MAP_ANON, NOFD, DEFAULT_IOAPIC_BASE);

  if (ioapics[0].base == MAP_FAILED)
    fmtout_console("Cannot map ioapic page, %d error!\n", errno);

  ioapics[0].base_gsi = 0;
  ioapics[0].npins = 24;

  gencon.name = "ioapic";

  /* fexme: get cpus number and apic physical ids from the mptable */
  gencon.ncpus = 2;
  gencon.caps = (irq_capacity_t*)malloc(gencon.ncpus * sizeof(irq_capacity_t));
  if (gencon.caps == NULL)
     abort();

  if (pthread_mutex_init(&gencon.mutex, &attr))
    abort();

  for (i = 0; i < gencon.ncpus; i++) {
    gencon.caps[i].limit = 24;
    gencon.caps[i].curr_num = 0;
  }  

  /*
   * Fixme: get physical apic ids
   */
  
  /* fexme: get number of irqs from mptable */
  gencon.nr_irqs = 24;
  gencon.irqs = (irq_desc_t*)malloc(gencon.nr_irqs * sizeof(irq_desc_t));
  if (gencon.irqs == NULL)
     abort();

  /*  fixme: fill based on the information taken from the acpi server */
  for (i = 0; i < gencon.nr_irqs; i++) {
    irq_desc_t *idesc = gencon.irqs + i;      
     
    idesc->vector = i + BASE_IRQ_VECTOR;
    idesc->targ_cpu = 0/*-1*/;
    set_irq_flags(idesc, IRQ_NOT_MIGRATE | IRQ_MASKED);
    idesc->line.chip = 0;
    idesc->line.pin = i;
    idesc->mode.trig = IRQ_TRIG_EDGE;
    idesc->mode.polarity = IRQ_POLAR_HIGH;
    idesc->mode.deliv = TXMODE_FIXED;
  }
  gencon.irqs[0].line.pin = 2;

  gencon.ops = &ioapic_operations;

  init_vector_map();
  
  for (i = 0; i < ncon; i++)
     ioapic_init(i);

  list_init_head(&gencon.balancers);
}


static int __balance_ctl(irq_ctl_t *ctl)
{
  int ret = 0;
  irq_desc_t *idesc;

  switch (ctl->cmd) {
  case BALCTL_SET_IRQ_FLAGS:
  case BALCTL_GET_IRQ_FLAGS:
    if (ctl->irq > gencon.nr_irqs)
      ret = -EINVAL;
    else {
      idesc = gencon.irqs + ctl->irq;
      if (ctl->cmd == BALCTL_SET_IRQ_FLAGS)
        *(uint16_t*)(&idesc->flags) = ctl->flags;
      else
        ctl->flags = idesc->flags;
    }
    break;
  case BALCTL_SET_CPU_LIMIT:
    if (ctl->cpu > gencon.ncpus)
      ret = -EINVAL;
    else {
      irq_balancer_t *bal;

      irqcon_lock(&gencon);
      gencon.caps[ctl->cpu].limit = ctl->limit;
      if (gencon.caps[ctl->cpu].curr_num > gencon.caps[ctl->cpu].limit) {
        if (!list_is_empty(&gencon.balancers)) {
          bal = container_of(&gencon.balancers, irq_balancer_t, l);
          if (bal->balance)
            bal->balance(&gencon);
        }  
      }
      irqcon_unlock(&gencon);
    }
    break;
  case BALCTL_GET_CPU_LIMIT:
    if (ctl->cpu > gencon.ncpus)
      ret = -EINVAL;
    else
      ctl->limit = gencon.caps[ctl->cpu].limit;
    break;
  default:
    ret = -EINVAL;
  }

  return ret;
}

static status_t ioapic_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize)
{
	int ret = 0;
  irq_ctl_t *ctl = (irq_ctl_t *)buf;
	irq_desc_t *idesc;
	irq_listener_t *lis;
	tid_t tid;
	irq_balancer_t *bal = NULL;

	fmtout_console("[%s] line %d, cmd = %d\n", __FUNCTION__, __LINE__, nr);

	if (nr != IRQCON_BALANCE_CONTROL) {
		if (ctl->irq > gencon.nr_irqs)
			return -EINVAL;
	}

	idesc = gencon.irqs + ctl->irq;

	switch (nr) {
    case IRQCON_ROUTE:
      if (ctl->cpu >= gencon.ncpus)
        ret = -EINVAL;
      else if (gencon.caps[ctl->cpu].curr_num == gencon.caps[ctl->cpu].limit)
        ret = -EPERM;
      else if (idesc->targ_cpu != ctl->cpu) {
        irqcon_lock(&gencon);
        ioapic_mask_irq(ctl->irq, 1);
        ioapic_route_irq(ctl->irq, ctl->cpu);
        /* TO DO: move threads to another cpu */
        ioapic_mask_irq(ctl->irq, 0);
        
        if (!list_is_empty(&gencon.balancers) && idesc->targ_cpu >= 0 && ctl->vec >= 0) {
          bal = container_of(&gencon.balancers, irq_balancer_t, l);
          if (bal->balance)
            bal->balance(&gencon);
        }
        irqcon_unlock(&gencon);
      }
      break;
    case IRQCON_MASK:
      /* the fifth argument is the mask value */
      ioapic_mask_irq(ctl->irq, ctl->mask);
      break;
    case IRQCON_SET_PRIO:	
      if (ctl->vec > (MAX_APIC_IRQ_VECTOR - BASE_IRQ_VECTOR))
        return -EINVAL;
      
      ctl->vec += BASE_IRQ_VECTOR;
      /* find an apropriate vector */
      if (!bit_test_multi(&gencon.vecmap, ctl->vec)) {
        int i;
        
        i = bit_find_lsf_multi(&gencon.vecmap, ctl->vec);
        if (i == -1)
          i = bit_find_msf_multi(&gencon.vecmap, ctl->vec);
        
        ctl->vec = i;
      }
      irqcon_lock(&gencon);	
      idesc->vector = ctl->vec;
      /* fixme: write the irq number to the kernel virtual interrupt table */
      if (idesc->targ_cpu >= 0)
        ioapic_set_ioredir(idesc);
      
      irqcon_unlock(&gencon);
      break;
      /* TO DO: implement balance control */
    case IRQCON_BALANCE_CONTROL:
      ret = __balance_ctl(ctl);
      
//      rpc->arg0 = data;
      break;
    case IRQCON_ADD_LISTENER:
      tid = get_sender_tid();
      
      list_for_each_entry(&idesc->listeners, lis, l) {
        if (lis->tid == tid)
          break;
      }
      irqcon_lock(&gencon);	
      if (&lis->l == list_head(&idesc->listeners)) {
        if (idesc->nr_lis) {
          if (!list_is_empty(&gencon.balancers))
            bal = container_of(&gencon.balancers, irq_balancer_t, l);
          
          if (bal && bal->choose_dest_cpu)
            idesc->targ_cpu = bal->choose_dest_cpu(ctl->irq, &gencon);
          else
            idesc->targ_cpu = 0;
          
          ioapic_route_irq(ctl->irq, idesc->targ_cpu);
        }
        lis = (irq_listener_t*)malloc(sizeof(irq_listener_t));
        lis->tid = tid;
        list_add2tail(&idesc->listeners, &lis->l);
        /* Fexme dg: move the thread to the target cpu */
        ioapic_mask_irq(ctl->irq, 0);
      }
      idesc->nr_lis++;
      irqcon_unlock(&gencon);
      break;
    case IRQCON_RM_LISTENER:
      tid = get_sender_tid();
      
      list_for_each_entry(&idesc->listeners, lis, l) {
        if (lis->tid == tid)
          break;
      }
      if (&lis->l != list_head(&idesc->listeners)) {
        irqcon_lock(&gencon);
        list_del(&lis->l);
        free(lis);
        idesc->nr_lis--;	
        if (!idesc->nr_lis) {
          idesc->targ_cpu = -1;
          ioapic_mask_irq(ctl->irq, 0);
          bal = container_of(&gencon.balancers, irq_balancer_t, l);
          if (bal->balance)
            bal->balance(&gencon);
        }
        irqcon_unlock(&gencon);
      }
      break;
    default:
      ret = -EINVAL;
      break;
	}

	return ret;
}

static status_t ioapic_dio_read(int devno, void *buf, size_t size, size_t offset, void **interim)
{
	/* dummy */
	return -ENOSYS;
}

int main(int argc, char **argv)
{ 
  int  r;
  device_server_t serv;

#ifndef CONFIG_IOAPIC_GRUB
  start_service();
#endif

  memset(&serv, 0, sizeof(serv));

  open_console();

  ascii_puts("[ioapic.d] Jari OS I/O APIC driver ...");
  init_driver();
  //ascii_puts("OK\n");

  r = init_simple_device_server(BUFFER_SIZE, BUFFER_SIZE, DRVNAME, DEVNAME, &ioapic_ops); 

  if(r) {
    ascii_puts("/dev/ioapic registration failed\n");
    abort();
  } else
    ascii_puts("OK\n");

  device_server_start();
  ascii_puts("[ioapicd] Server start error!\n");

  return 0;
}
