#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <datalib/usrtc.h>
#include <errno.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ns/ns.h>
#include <errno.h>
#include <devio/devio.h>
#include <sys/poll.h>
#include <sys/thread.h>
#include <debug.h>
#include <ns/ns.h>
#include <sys/procmgr.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <pci/pci.h>
#include <sys/ievent.h>
#include <arch/atomic.h>
#include <arch/bitwise.h>

#include "e1000.h"
#include "e1000_hw.h"
#include "e1000_osdep.h"

#define DEBUGOUT(x)  console_printf((x))
#define DEBUGFUNC(x)  console_printf((x))

static int __e1000_read_ich8_data(struct e1000_hw *hw, uint32_t index,
                                  uint32_t size, uint16_t* data);
static void __e1000_shift_out_ee_bits(struct e1000_hw *hw,
                                      uint16_t data,uint16_t count);
static uint16_t __e1000_shift_in_ee_bits(struct e1000_hw *hw,uint16_t count);
static void __e1000_standby_eeprom(struct e1000_hw *hw);
static void __e1000_raise_ee_clk(struct e1000_hw *hw,uint32_t *eecd);
static void __e1000_lower_ee_clk(struct e1000_hw *hw,uint32_t *eecd);
static int32_t __e1000_spi_eeprom_ready(struct e1000_hw *hw);

static int32_t __e1000_poll_eerd_eewr_done(struct e1000_hw *hw, int eerd)
{
  uint32_t attempts = 100000;
  uint32_t i, reg = 0;
  int32_t done = E1000_ERR_EEPROM;

  for (i = 0; i < attempts; i++) {
    if (eerd == E1000_EEPROM_POLL_READ)
      reg = E1000_READ_REG(hw, EERD);
    else
      reg = E1000_READ_REG(hw, EEWR);

    if (reg & E1000_EEPROM_RW_REG_DONE) {
      done = E1000_SUCCESS;
      break;
    }
    udelay(5);
  }

  return done;
}

static void __e1000_put_hw_eeprom_semaphore(struct e1000_hw *hw)
{
  uint32_t swsm;

  if (!hw->eeprom_semaphore_present)
    return;

  swsm = E1000_READ_REG(hw, SWSM);
  if (hw->mac_type == e1000_80003es2lan) {
    /* Release both semaphores. */
    swsm &= ~(E1000_SWSM_SMBI | E1000_SWSM_SWESMBI);
  } else
    swsm &= ~(E1000_SWSM_SWESMBI);
  E1000_WRITE_REG(hw, SWSM, swsm);
}

static int32_t __e1000_get_software_semaphore(struct e1000_hw *hw)
{
  int32_t timeout = hw->eeprom.word_size + 1;
  uint32_t swsm;

    if (hw->mac_type != e1000_80003es2lan) {
        return E1000_SUCCESS;
    }

    while (timeout) {
        swsm = E1000_READ_REG(hw, SWSM);
        /* If SMBI bit cleared, it is now set and we hold the semaphore */
        if (!(swsm & E1000_SWSM_SMBI))
            break;
        mdelay(1);
        timeout--;
    }

    if (!timeout) {
        DEBUGOUT("Driver can't access device - SMBI bit is set.\n");
        return -E1000_ERR_RESET;
    }

    return E1000_SUCCESS;
}

static void __e1000_lower_ee_clk(struct e1000_hw *hw,uint32_t *eecd)
{
  /* Lower the clock input to the EEPROM (by clearing the SK bit), and then
   * wait 50 microseconds.
   */
  *eecd = *eecd & ~E1000_EECD_SK;
  E1000_WRITE_REG(hw, EECD, *eecd);
  E1000_WRITE_FLUSH(hw);
  udelay(hw->eeprom.delay_usec);
}

static void __e1000_raise_ee_clk(struct e1000_hw *hw,uint32_t *eecd)
{
    /* Raise the clock input to the EEPROM (by setting the SK bit), and then
     * wait <delay> microseconds.
     */
    *eecd = *eecd | E1000_EECD_SK;
    E1000_WRITE_REG(hw, EECD, *eecd);
    E1000_WRITE_FLUSH(hw);
    udelay(hw->eeprom.delay_usec);
}

static int __e1000_get_hw_eeprom_semaphore(struct e1000_hw *hw)
{
  int32_t timeout;
  uint32_t swsm;

  if (!hw->eeprom_semaphore_present)
    return E1000_SUCCESS;

  if (hw->mac_type == e1000_80003es2lan) {
    /* Get the SW semaphore. */
    if (__e1000_get_software_semaphore(hw) != E1000_SUCCESS)
      return -E1000_ERR_EEPROM;
  }

  /* Get the FW semaphore. */
  timeout = hw->eeprom.word_size + 1;
  while (timeout) {
    swsm = E1000_READ_REG(hw, SWSM);
    swsm |= E1000_SWSM_SWESMBI;
    E1000_WRITE_REG(hw, SWSM, swsm);
    /* if we managed to set the bit we got the semaphore. */
    swsm = E1000_READ_REG(hw, SWSM);
    if (swsm & E1000_SWSM_SWESMBI)
      break;

    udelay(50);
    timeout--;
  }

  if (!timeout) {
    /* Release semaphores */
    __e1000_put_hw_eeprom_semaphore(hw);
    DEBUGOUT("Driver can't access the Eeprom - SWESMBI bit is set.\n");
    return -E1000_ERR_EEPROM;
  }

  return E1000_SUCCESS;
}

static int __e1000_ich8_flash_cycle(struct e1000_hw *hw, uint32_t timeout)
{
  union ich8_hws_flash_ctrl hsflctl;
  union ich8_hws_flash_status hsfsts;
  int32_t error = E1000_ERR_EEPROM;
  uint32_t i = 0;

  /* Start a cycle by writing 1 in Flash Cycle Go in Hw Flash Control */
  hsflctl.regval = E1000_READ_ICH_FLASH_REG16(hw, ICH_FLASH_HSFCTL);
  hsflctl.hsf_ctrl.flcgo = 1;
  E1000_WRITE_ICH_FLASH_REG16(hw, ICH_FLASH_HSFCTL, hsflctl.regval);

  /* wait till FDONE bit is set to 1 */
  do {
    hsfsts.regval = E1000_READ_ICH_FLASH_REG16(hw, ICH_FLASH_HSFSTS);
    if (hsfsts.hsf_status.flcdone == 1)
      break;
    udelay(1);
    i++;
  } while (i < timeout);
  if (hsfsts.hsf_status.flcdone == 1 && hsfsts.hsf_status.flcerr == 0) {
    error = E1000_SUCCESS;
  }
  return error;
}

static int __e1000_ich8_cycle_init(struct e1000_hw *hw)
{
  union ich8_hws_flash_status hsfsts;
  int32_t error = E1000_ERR_EEPROM;
  int32_t i     = 0;

  hsfsts.regval = E1000_READ_ICH_FLASH_REG16(hw, ICH_FLASH_HSFSTS);

    /* May be check the Flash Des Valid bit in Hw status */
    if (hsfsts.hsf_status.fldesvalid == 0) {
        DEBUGOUT("Flash descriptor invalid.  SW Sequencing must be used.");
        return error;
    }

    /* Clear FCERR in Hw status by writing 1 */
    /* Clear DAEL in Hw status by writing a 1 */
    hsfsts.hsf_status.flcerr = 1;
    hsfsts.hsf_status.dael = 1;

    E1000_WRITE_ICH_FLASH_REG16(hw, ICH_FLASH_HSFSTS, hsfsts.regval);

    /* Either we should have a hardware SPI cycle in progress bit to check
     * against, in order to start a new cycle or FDONE bit should be changed
     * in the hardware so that it is 1 after harware reset, which can then be
     * used as an indication whether a cycle is in progress or has been
     * completed .. we should also have some software semaphore mechanism to
     * guard FDONE or the cycle in progress bit so that two threads access to
     * those bits can be sequentiallized or a way so that 2 threads dont
     * start the cycle at the same time */

    if (hsfsts.hsf_status.flcinprog == 0) {
        /* There is no cycle running at present, so we can start a cycle */
        /* Begin by setting Flash Cycle Done. */
        hsfsts.hsf_status.flcdone = 1;
        E1000_WRITE_ICH_FLASH_REG16(hw, ICH_FLASH_HSFSTS, hsfsts.regval);
        error = E1000_SUCCESS;
    } else {
        /* otherwise poll for sometime so the current cycle has a chance
         * to end before giving up. */
        for (i = 0; i < ICH_FLASH_COMMAND_TIMEOUT; i++) {
            hsfsts.regval = E1000_READ_ICH_FLASH_REG16(hw, ICH_FLASH_HSFSTS);
            if (hsfsts.hsf_status.flcinprog == 0) {
                error = E1000_SUCCESS;
                break;
            }
            udelay(1);
        }
        if (error == E1000_SUCCESS) {
            /* Successful in waiting for previous cycle to timeout,
             * now set the Flash Cycle Done. */
            hsfsts.hsf_status.flcdone = 1;
            E1000_WRITE_ICH_FLASH_REG16(hw, ICH_FLASH_HSFSTS, hsfsts.regval);
        } else {
            DEBUGOUT("Flash controller busy, cannot get access");
        }
    }
    return error;
}

static void __e1000_release_eeprom(struct e1000_hw *hw)
{
  uint32_t eecd;

  eecd = E1000_READ_REG(hw, EECD);

  if (hw->eeprom.type == e1000_eeprom_spi) {
    eecd |= E1000_EECD_CS;  /* Pull CS high */
    eecd &= ~E1000_EECD_SK; /* Lower SCK */

    E1000_WRITE_REG(hw, EECD, eecd);

    udelay(hw->eeprom.delay_usec);
  } else if (hw->eeprom.type == e1000_eeprom_microwire) {
    /* cleanup eeprom */

    /* CS on Microwire is active-high */
    eecd &= ~(E1000_EECD_CS | E1000_EECD_DI);

    E1000_WRITE_REG(hw, EECD, eecd);

    /* Rising edge of clock */
    eecd |= E1000_EECD_SK;
    E1000_WRITE_REG(hw, EECD, eecd);
    E1000_WRITE_FLUSH(hw);
    udelay(hw->eeprom.delay_usec);

    /* Falling edge of clock */
    eecd &= ~E1000_EECD_SK;
    E1000_WRITE_REG(hw, EECD, eecd);
    E1000_WRITE_FLUSH(hw);
    udelay(hw->eeprom.delay_usec);
  }

  /* Stop requesting EEPROM access */
  if (hw->mac_type > e1000_82544) {
    eecd &= ~E1000_EECD_REQ;
    E1000_WRITE_REG(hw, EECD, eecd);
  }

  e1000_swfw_sync_release(hw, E1000_SWFW_EEP_SM);
}

void e1000_swfw_sync_release(struct e1000_hw *hw, uint16_t mask)
{
  uint32_t swfw_sync;
  uint32_t swmask = mask;

  if (hw->swfwhw_semaphore_present) {
    e1000_release_software_flag(hw);
    return;
  }

  if (!hw->swfw_sync_present) {
    __e1000_put_hw_eeprom_semaphore(hw);
    return;
  }

  /* if (e1000_get_hw_eeprom_semaphore(hw))
   *    return -E1000_ERR_SWFW_SYNC; */
  while (__e1000_get_hw_eeprom_semaphore(hw) != E1000_SUCCESS);
  /* empty */

  swfw_sync = E1000_READ_REG(hw, SW_FW_SYNC);
  swfw_sync &= ~swmask;
  E1000_WRITE_REG(hw, SW_FW_SYNC, swfw_sync);

  __e1000_put_hw_eeprom_semaphore(hw);
}

int32_t e1000_swfw_sync_acquire(struct e1000_hw *hw, uint16_t mask)
{
  uint32_t swfw_sync = 0;
  uint32_t swmask = mask;
  uint32_t fwmask = mask << 16;
  int32_t timeout = 200;

  if (hw->swfwhw_semaphore_present)
    return e1000_get_software_flag(hw);

  if (!hw->swfw_sync_present)
    return __e1000_get_hw_eeprom_semaphore(hw);

  while (timeout) {
    if (__e1000_get_hw_eeprom_semaphore(hw))
      return -E1000_ERR_SWFW_SYNC;

    swfw_sync = E1000_READ_REG(hw, SW_FW_SYNC);
    if (!(swfw_sync & (fwmask | swmask))) {
      break;
    }

    /* firmware currently using resource (fwmask) */
    /* or other software thread currently using resource (swmask) */
    __e1000_put_hw_eeprom_semaphore(hw);
    mdelay(5);
    timeout--;
  }

  if (!timeout) {
    DEBUGOUT("Driver can't access resource, SW_FW_SYNC timeout.\n");
    return -E1000_ERR_SWFW_SYNC;
  }

  swfw_sync |= swmask;
  E1000_WRITE_REG(hw, SW_FW_SYNC, swfw_sync);

  __e1000_put_hw_eeprom_semaphore(hw);
  return E1000_SUCCESS;
}

static int __e1000_read_ich8_data(struct e1000_hw *hw, uint32_t index,
                                  uint32_t size, uint16_t* data)
{
    union ich8_hws_flash_status hsfsts;
    union ich8_hws_flash_ctrl hsflctl;
    uint32_t flash_linear_address;
    uint32_t flash_data = 0;
    int32_t error = -E1000_ERR_EEPROM;
    int32_t count = 0;

    if (size < 1  || size > 2 || data == NULL ||
        index > ICH_FLASH_LINEAR_ADDR_MASK)
      return error;

    flash_linear_address = (ICH_FLASH_LINEAR_ADDR_MASK & index) +
                           hw->flash_base_addr;

    do {
        udelay(1);
        /* Steps */
        error = __e1000_ich8_cycle_init(hw);
        if (error)
            break;

        hsflctl.regval = E1000_READ_ICH_FLASH_REG16(hw, ICH_FLASH_HSFCTL);
        /* 0b/1b corresponds to 1 or 2 byte size, respectively. */
        hsflctl.hsf_ctrl.fldbcount = size - 1;
        hsflctl.hsf_ctrl.flcycle = ICH_CYCLE_READ;
        E1000_WRITE_ICH_FLASH_REG16(hw, ICH_FLASH_HSFCTL, hsflctl.regval);

        /* Write the last 24 bits of index into Flash Linear address field in
         * Flash Address */
        /* TODO: TBD maybe check the index against the size of flash */

        E1000_WRITE_ICH_FLASH_REG(hw, ICH_FLASH_FADDR, flash_linear_address);

        error = __e1000_ich8_flash_cycle(hw, ICH_FLASH_COMMAND_TIMEOUT);

        /* Check if FCERR is set to 1, if set to 1, clear it and try the whole
         * sequence a few more times, else read in (shift in) the Flash Data0,
         * the order is least significant byte first msb to lsb */
        if (error == 0) {
            flash_data = E1000_READ_ICH_FLASH_REG(hw, ICH_FLASH_FDATA0);
            if (size == 1) {
                *data = (uint8_t)(flash_data & 0x000000FF);
            } else if (size == 2) {
                *data = (uint16_t)(flash_data & 0x0000FFFF);
            }
            break;
        } else {
            /* If we've gotten here, then things are probably completely hosed,
             * but if the error condition is detected, it won't hurt to give
             * it another try...ICH_FLASH_CYCLE_REPEAT_COUNT times.
             */
            hsfsts.regval = E1000_READ_ICH_FLASH_REG16(hw, ICH_FLASH_HSFSTS);
            if (hsfsts.hsf_status.flcerr == 1) {
                /* Repeat for some time before giving up. */
                continue;
            } else if (hsfsts.hsf_status.flcdone == 0) {
                console_printf("[e1000]: Timeout error - flash cycle did not complete.");
                break;
            }
        }
    } while (count++ < ICH_FLASH_CYCLE_REPEAT_COUNT);

    return error;
}

static int __e1000_read_ich8_word(struct e1000_hw *hw, uint32_t index, uint16_t *data)
{
  int status = 0;
  status = __e1000_read_ich8_data(hw, index, 2, data);
  return status;
}

static int __e1000_read_eeprom_ich8(struct e1000_hw *hw, uint16_t offset,
                                    uint16_t words,uint16_t *data)
{
  int32_t  error = 0;
  uint32_t flash_bank = 0;
  uint32_t act_offset = 0;
  uint32_t bank_offset = 0;
  uint16_t word = 0;
  uint16_t i = 0;

  /* We need to know which is the valid flash bank.  In the event
   * that we didn't allocate eeprom_shadow_ram, we may not be
   * managing flash_bank.  So it cannot be trusted and needs
   * to be updated with each read.
   */
  /* Value of bit 22 corresponds to the flash bank we're on. */
  flash_bank = (E1000_READ_REG(hw, EECD) & E1000_EECD_SEC1VAL) ? 1 : 0;

  /* Adjust offset appropriately if we're on bank 1 - adjust for word size */
  bank_offset = flash_bank * (hw->flash_bank_size * 2);

  error = e1000_get_software_flag(hw);
  if (error)
    return error;

  for (i = 0; i < words; i++) {
    if (hw->eeprom_shadow_ram != NULL &&
        hw->eeprom_shadow_ram[offset+i].modified == true) {
      data[i] = hw->eeprom_shadow_ram[offset+i].eeprom_word;
    } else {
      /* The NVM part needs a byte offset, hence * 2 */
      act_offset = bank_offset + ((offset + i) * 2);
      error = __e1000_read_ich8_word(hw, act_offset, &word);
      if (error)
        break;
      data[i] = word;
    }
  }

  e1000_release_software_flag(hw);
  return error;
}

static int32_t __e1000_spi_eeprom_ready(struct e1000_hw *hw)
{
  uint16_t retry_count = 0;
  uint8_t spi_stat_reg;

  /* Read "Status Register" repeatedly until the LSB is cleared.  The
   * EEPROM will signal that the command has been completed by clearing
   * bit 0 of the internal status register.  If it's not cleared within
   * 5 milliseconds, then error out.
   */
  retry_count = 0;
  do {
    __e1000_shift_out_ee_bits(hw, EEPROM_RDSR_OPCODE_SPI,
                              hw->eeprom.opcode_bits);
    spi_stat_reg = (uint8_t)__e1000_shift_in_ee_bits(hw, 8);
    if (!(spi_stat_reg & EEPROM_STATUS_RDY_SPI))
      break;

    udelay(5);
    retry_count += 5;

    __e1000_standby_eeprom(hw);
  } while (retry_count < EEPROM_MAX_RETRY_SPI);

  /* ATMEL SPI write time could vary from 0-20mSec on 3.3V devices (and
   * only 0-5mSec on 5V devices)
   */
  if (retry_count >= EEPROM_MAX_RETRY_SPI) {
    DEBUGOUT("SPI EEPROM Status error\n");
    return -E1000_ERR_EEPROM;
  }

  return E1000_SUCCESS;
}

static uint16_t __e1000_shift_in_ee_bits(struct e1000_hw *hw,uint16_t count)
{
  uint32_t eecd;
  uint32_t i;
  uint16_t data;

  /* In order to read a register from the EEPROM, we need to shift 'count'
   * bits in from the EEPROM. Bits are "shifted in" by raising the clock
   * input to the EEPROM (setting the SK bit), and then reading the value of
   * the "DO" bit.  During this "shifting in" process the "DI" bit should
   * always be clear.
   */

  eecd = E1000_READ_REG(hw, EECD);

  eecd &= ~(E1000_EECD_DO | E1000_EECD_DI);
  data = 0;

  for (i = 0; i < count; i++) {
    data = data << 1;
    __e1000_raise_ee_clk(hw, &eecd);

    eecd = E1000_READ_REG(hw, EECD);

    eecd &= ~(E1000_EECD_DI);
    if (eecd & E1000_EECD_DO)
      data |= 1;

    __e1000_lower_ee_clk(hw, &eecd);
  }

  return data;
}

static void __e1000_shift_out_ee_bits(struct e1000_hw *hw,
                                      uint16_t data,uint16_t count)
{
  struct e1000_eeprom_info *eeprom = &hw->eeprom;
  uint32_t eecd;
  uint32_t mask;

  /* We need to shift "count" bits out to the EEPROM. So, value in the
   * "data" parameter will be shifted out to the EEPROM one bit at a time.
   * In order to do this, "data" must be broken down into bits.
   */
  mask = 0x01 << (count - 1);
  eecd = E1000_READ_REG(hw, EECD);
  if (eeprom->type == e1000_eeprom_microwire) {
    eecd &= ~E1000_EECD_DO;
  } else if (eeprom->type == e1000_eeprom_spi) {
    eecd |= E1000_EECD_DO;
  }
  do {
    /* A "1" is shifted out to the EEPROM by setting bit "DI" to a "1",
     * and then raising and then lowering the clock (the SK bit controls
     * the clock input to the EEPROM).  A "0" is shifted out to the EEPROM
     * by setting "DI" to "0" and then raising and then lowering the clock.
     */
    eecd &= ~E1000_EECD_DI;

    if (data & mask)
      eecd |= E1000_EECD_DI;

    E1000_WRITE_REG(hw, EECD, eecd);
    E1000_WRITE_FLUSH(hw);

    udelay(eeprom->delay_usec);

    __e1000_raise_ee_clk(hw, &eecd);
    __e1000_lower_ee_clk(hw, &eecd);

    mask = mask >> 1;
  } while (mask);

  /* We leave the "DI" bit set to "0" when we leave this routine. */
  eecd &= ~E1000_EECD_DI;
  E1000_WRITE_REG(hw, EECD, eecd);
}

static int __e1000_read_eeprom_eerd(struct e1000_hw *hw,
                                    uint16_t offset,uint16_t words,
                                    uint16_t *data)
{
  uint32_t i, eerd = 0;
  int32_t error = 0;

  for (i = 0; i < words; i++) {
    eerd = ((offset+i) << E1000_EEPROM_RW_ADDR_SHIFT) +
      E1000_EEPROM_RW_REG_START;

    E1000_WRITE_REG(hw, EERD, eerd);
    error = __e1000_poll_eerd_eewr_done(hw, E1000_EEPROM_POLL_READ);

    if (error) {
      break;
    }
    data[i] = (E1000_READ_REG(hw, EERD) >> E1000_EEPROM_RW_REG_DATA);
  }

  return error;
}

static void __e1000_standby_eeprom(struct e1000_hw *hw)
{
  struct e1000_eeprom_info *eeprom = &hw->eeprom;
  uint32_t eecd;

  eecd = E1000_READ_REG(hw, EECD);

  if (eeprom->type == e1000_eeprom_microwire) {
    eecd &= ~(E1000_EECD_CS | E1000_EECD_SK);
    E1000_WRITE_REG(hw, EECD, eecd);
    E1000_WRITE_FLUSH(hw);
    udelay(eeprom->delay_usec);

    /* Clock high */
    eecd |= E1000_EECD_SK;
    E1000_WRITE_REG(hw, EECD, eecd);
    E1000_WRITE_FLUSH(hw);
    udelay(eeprom->delay_usec);

    /* Select EEPROM */
    eecd |= E1000_EECD_CS;
    E1000_WRITE_REG(hw, EECD, eecd);
    E1000_WRITE_FLUSH(hw);
    udelay(eeprom->delay_usec);

    /* Clock low */
    eecd &= ~E1000_EECD_SK;
    E1000_WRITE_REG(hw, EECD, eecd);
    E1000_WRITE_FLUSH(hw);
    udelay(eeprom->delay_usec);
  } else if (eeprom->type == e1000_eeprom_spi) {
    /* Toggle CS to flush commands */
    eecd |= E1000_EECD_CS;
    E1000_WRITE_REG(hw, EECD, eecd);
    E1000_WRITE_FLUSH(hw);
    udelay(eeprom->delay_usec);
    eecd &= ~E1000_EECD_CS;
    E1000_WRITE_REG(hw, EECD, eecd);
    E1000_WRITE_FLUSH(hw);
    udelay(eeprom->delay_usec);
  }
}

static int __e1000_acquire_eeprom(struct e1000_hw *hw)
{
  struct e1000_eeprom_info *eeprom = &hw->eeprom;
  uint32_t eecd, i=0;

  if (e1000_swfw_sync_acquire(hw, E1000_SWFW_EEP_SM))
    return -E1000_ERR_SWFW_SYNC;
  eecd = E1000_READ_REG(hw, EECD);

  if (hw->mac_type != e1000_82573) {
    /* Request EEPROM Access */
    if (hw->mac_type > e1000_82544) {
      eecd |= E1000_EECD_REQ;
      E1000_WRITE_REG(hw, EECD, eecd);
      eecd = E1000_READ_REG(hw, EECD);
      while ((!(eecd & E1000_EECD_GNT)) &&
             (i < E1000_EEPROM_GRANT_ATTEMPTS)) {
        i++;
        udelay(5);
        eecd = E1000_READ_REG(hw, EECD);
      }
      if (!(eecd & E1000_EECD_GNT)) {
        eecd &= ~E1000_EECD_REQ;
        E1000_WRITE_REG(hw, EECD, eecd);
        console_printf("[e1000]: Could not acquire EEPROM grant\n");
        e1000_swfw_sync_release(hw, E1000_SWFW_EEP_SM);
        return -E1000_ERR_EEPROM;
      }
    }
  }

  /* Setup EEPROM for Read/Write */
  if (eeprom->type == e1000_eeprom_microwire) {
    /* Clear SK and DI */
    eecd &= ~(E1000_EECD_DI | E1000_EECD_SK);
    E1000_WRITE_REG(hw, EECD, eecd);

    /* Set CS */
    eecd |= E1000_EECD_CS;
    E1000_WRITE_REG(hw, EECD, eecd);
  } else if (eeprom->type == e1000_eeprom_spi) {
    /* Clear SK and CS */
    eecd &= ~(E1000_EECD_CS | E1000_EECD_SK);
    E1000_WRITE_REG(hw, EECD, eecd);
    udelay(1);
  }

  return 0;
}

int e1000_init_eeprom_params(struct e1000_hw *hw)
{
  struct e1000_eeprom_info *eeprom = &hw->eeprom;
  uint32_t eecd = E1000_READ_REG(hw, EECD);
  int32_t ret_val = 0;
  uint16_t eeprom_size;

  switch (hw->mac_type) {
    case e1000_82571:
    case e1000_82572:
      eeprom->type = e1000_eeprom_spi;
      eeprom->opcode_bits = 8;
      eeprom->delay_usec = 1;
      if (eecd & E1000_EECD_ADDR_BITS) {
        eeprom->page_size = 32;
        eeprom->address_bits = 16;
      } else {
        eeprom->page_size = 8;
        eeprom->address_bits = 8;
      }
      eeprom->use_eerd = false;
      eeprom->use_eewr = false;
      break;
    default:
      console_printf("[e1000]: e1000_init_eeprom_params(): insufficient MAC type: %d\n",
                     hw->mac_type);
      return -1;
  }

  if (eeprom->type == e1000_eeprom_spi) {
    /* eeprom_size will be an enum [0..8] that maps to eeprom sizes 128B to
     * 32KB (incremented by powers of 2).
     */
    if (hw->mac_type <= e1000_82547_rev_2) {
      /* Set to default value for initial eeprom read. */
      eeprom->word_size = 64;
      ret_val = e1000_read_eeprom(hw, EEPROM_CFG, 1, &eeprom_size);
      if (ret_val)
        return ret_val;
      eeprom_size = (eeprom_size & EEPROM_SIZE_MASK) >> EEPROM_SIZE_SHIFT;
      /* 256B eeprom size was not supported in earlier hardware, so we
       * bump eeprom_size up one to ensure that "1" (which maps to 256B)
       * is never the result used in the shifting logic below. */
      if (eeprom_size)
        eeprom_size++;
    } else {
      eeprom_size = (uint16_t)((eecd & E1000_EECD_SIZE_EX_MASK) >>
                               E1000_EECD_SIZE_EX_SHIFT);
    }

    eeprom->word_size = 1 << (eeprom_size + EEPROM_WORD_SIZE_SHIFT);
    console_printf("[e1000] e1000_init_eeprom_params(): EEPROM size = %d\n",
                   eeprom->word_size);
  }
  return ret_val;
}


static bool __e1000_is_onboard_nvm_eeprom(struct e1000_hw *hw)
{
  uint32_t eecd = 0;

  if (hw->mac_type == e1000_ich8lan) {
    return false;
  }

  if (hw->mac_type == e1000_82573) {
    eecd = E1000_READ_REG(hw, EECD);

    /* Isolate bits 15 & 16 */
    eecd = ((eecd >> 15) & 0x03);

    /* If both bits are set, device is Flash type */
    if (eecd == 0x03) {
      return false;
    }
  }
  return true;
}

int e1000_read_eeprom(struct e1000_hw *hw,uint16_t offset,
                      uint16_t words,uint16_t *data)
{
    struct e1000_eeprom_info *eeprom = &hw->eeprom;
    uint32_t i = 0;

    /* If eeprom is not yet detected, do so now */
    if (eeprom->word_size == 0)
        e1000_init_eeprom_params(hw);

    /* A check for invalid values:  offset too large, too many words, and not
     * enough words.
     */
    if ((offset >= eeprom->word_size) || (words > eeprom->word_size - offset) ||
       (words == 0)) {
      console_printf("[e1000] e1000_read_eeprom(): \"words\" parameter out of bounds. Words = %d, size = %d\n",
                     offset, eeprom->word_size);
      return -E1000_ERR_EEPROM;
    }

    /* EEPROM's that don't use EERD to read require us to bit-bang the SPI
     * directly. In this case, we need to acquire the EEPROM so that
     * FW or other port software does not interrupt.
     */
    if (__e1000_is_onboard_nvm_eeprom(hw) == true &&
        hw->eeprom.use_eerd == false) {
        /* Prepare the EEPROM for bit-bang reading */
        if (__e1000_acquire_eeprom(hw) != 0)
            return -E1000_ERR_EEPROM;
    }

    /* Eerd register EEPROM access requires no eeprom aquire/release */
    if (eeprom->use_eerd == true)
        return __e1000_read_eeprom_eerd(hw, offset, words, data);

    /* ICH EEPROM access is done via the ICH flash controller */
    if (eeprom->type == e1000_eeprom_ich8)
        return __e1000_read_eeprom_ich8(hw, offset, words, data);

    /* Set up the SPI or Microwire EEPROM for bit-bang reading.  We have
     * acquired the EEPROM at this point, so any returns should relase it */
    if (eeprom->type == e1000_eeprom_spi) {
        uint16_t word_in;
        uint8_t read_opcode = EEPROM_READ_OPCODE_SPI;

        if (__e1000_spi_eeprom_ready(hw) ) {
          __e1000_release_eeprom(hw);
          return -E1000_ERR_EEPROM;
        }

        __e1000_standby_eeprom(hw);

        /* Some SPI eeproms use the 8th address bit embedded in the opcode */
        if ((eeprom->address_bits == 8) && (offset >= 128))
            read_opcode |= EEPROM_A8_OPCODE_SPI;

        /* Send the READ command (opcode + addr)  */
        __e1000_shift_out_ee_bits(hw, read_opcode, eeprom->opcode_bits);
        __e1000_shift_out_ee_bits(hw, (uint16_t)(offset*2), eeprom->address_bits);

        /* Read the data.  The address of the eeprom internally increments with
         * each byte (spi) being read, saving on the overhead of eeprom setup
         * and tear-down.  The address counter will roll over if reading beyond
         * the size of the eeprom, thus allowing the entire memory to be read
         * starting from any offset. */
        for (i = 0; i < words; i++) {
            word_in = __e1000_shift_in_ee_bits(hw, 16);
            data[i] = (word_in >> 8) | (word_in << 8);
        }
    } else if (eeprom->type == e1000_eeprom_microwire) {
        for (i = 0; i < words; i++) {
            /* Send the READ command (opcode + addr)  */
          __e1000_shift_out_ee_bits(hw, EEPROM_READ_OPCODE_MICROWIRE,
                                    eeprom->opcode_bits);
          __e1000_shift_out_ee_bits(hw, (uint16_t)(offset + i),
                                    eeprom->address_bits);

            /* Read the data.  For microwire, each word requires the overhead
             * of eeprom setup and tear-down. */
            data[i] = __e1000_shift_in_ee_bits(hw, 16);
            __e1000_standby_eeprom(hw);
        }
    }

    /* End this read operation */
    __e1000_release_eeprom(hw);

    return 0;
}

int e1000_validate_eeprom_checksum(struct e1000_hw *hw)
{
  uint16_t checksum = 0;
  uint16_t i, eeprom_data;

  for( i = 0; i < (EEPROM_CHECKSUM_REG + 1); i++ ) {
    if( e1000_read_eeprom(hw, i, 1, &eeprom_data) < 0) {
      console_printf("[e1000]: EEPROM Read Error\n");
      return -EIO;
    }
    checksum += eeprom_data;
  }

  if (checksum == (uint16_t) EEPROM_SUM) {
    return 0;
  } else {
    DEBUGOUT("[e1000]: EEPROM Checksum Invalid\n");
    return -EIO;
  }
}
