/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008,2009 Dmitry Gromada <gromada@jarios.org>
 *
 * ram disk implementation
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <devio/devio.h>
#include <limits.h>
#include <ipc/rpc.h>
#include <sys/mman.h>
#include <sys/scheduler.h>
#include <arch/atomic.h>
#include <bitwise.h>
#include <pthread.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <block/genstor.h>
#include <block/blkdev.h>
#include <block/request.h>
#include <block/server.h>
#include <block/iosched.h>
#include <block/sigaction.h>
#include <timeout.h>
#include <completion.h>
#include <dircons.h>
#include <debug.h>
#include <sys/kcontrol.h>

#define MAX_NR_RAM_DISKS 16
#define MAX_RAM_DISK_SECT_SHIFT PAGE_WIDTH
#define MAX_RAM_DISK_REQUESTS 1024
#define MAX_REQUEST_SECTORS 16
#define RAM_DISK_IOBUF_SIZE 1024

static genstor_t *ram_storages[MAX_NR_RAM_DISKS];
static blkdev_operations_t ramd_bd_ops;
static struct blk_group blk_group;
static pthread_t io_thread = {-1};

static int ramdisk_init_schedulers(void);
static int ramdisk_set_scheduler(blkreq_queue_t *q);
static blk_request_t *wait_get_rq(void);
static sigset_t sigset = 0;

static storage_operations_t ramd_stor_ops = {
  generic_open_storage,
  generic_close_storage,
  generic_scan_partitions,
  generic_read_storage,
  generic_write_storage,
  generic_ioctl_storage,
  generic_sync_storage,
  generic_make_request
};

#ifdef CONFIG_RAMDISK_IOSCHED

static int ramdisk_init_schedulers(void)
{
  return initialize_blk_schedulers();
}

static int ramdisk_set_scheduler(blkreq_queue_t *q)
{
  int ret;

  ret = set_blk_default_iosched(q);
  if (!ret) {
    ret = init_timeout_tmr(&q->scheduler->tmr, io_thread, IOSCHED_SIGNO, q, false);
    if (!ret)
      ret = set_blk_sigaction(IOSCHED_SIGNO, default_iosched_sigaction);

    if (ret)
      ret = -errno;
  }

  if (ret)
    disable_blk_iosched(q);

  return ret;
}

#else

static int ramdisk_init_schedulers(void)
{
  return 0;
}

static int ramdisk_set_scheduler(blkreq_queue_t *q)
{
  return 0;
}

#endif /* CONFIG_RAM_DISK_IOSCHED */

static blk_request_t *wait_get_rq(void)
{
  blk_request_t *rq;
  ulong_t qmask = ~0;
  ulong_t rec_mask;

  do {
    if (blk_group.last_serv_queue)
      rec_mask = ~0L & ~(1 << blk_group.last_serv_queue->index);
    else
      rec_mask = ~0L;

    rq = get_group_blkrq_wait(&blk_group, qmask, rec_mask);
  } while (rq == NULL);

  return rq;
}

static void *io_thread_fn(void *data)
{
  blk_request_t *rq;
  blkio_t *b;
  biorq_t *biorq;
  sector_t sector, capacity;
  void *addr, *disk_data;
  struct completion *compl = (struct completion*)data;
  int sect_shift;

  wait_for_completion(compl);
  destroy_completion(compl);
  /* grow up thread priority */
  task_scheduler_control(0, 0, SYS_SCHED_CTL_SET_PRIORITY, 1);

  capacity = ram_storages[0]->capacity;
  sect_shift = ram_storages[0]->sect_shift;

  while (1) {
    rq = wait_get_rq();
    if (rq == NULL)
      continue;

    disk_data = rq->queue->bdev->private;
    list_for_each_entry(&rq->blkios, b, node) {
      sector = b->start_sector;

      if (sector + b->nr_sectors > capacity) {
        b->status.err = EINVAL;
        rq->current_blkio = b;
        break;
      }
      addr = (void*)((ulong_t)disk_data + (sector << sect_shift));
      biorq = b->biorq;

      unsigned char byte, byte1;

      if (biorq_rw(biorq)) {
        if (biorq_data_available(biorq)) {
          memcpy(addr, b->buf, (size_t)b->nr_sectors << sect_shift);
        }
        else {
          copy_biorq_data_sectors(biorq, sector, b->nr_sectors,
                                  sect_shift, addr);
        }
        byte = ((unsigned char*)b->buf)[1];
        byte1 = ((unsigned char*)addr)[1];
      } else {
        memcpy(b->buf, addr, (size_t)b->nr_sectors << sect_shift);
      }
    }

    if (rq->complete) {
      rq->complete(rq);
    }
    put_blk_request(rq);
  }

  return NULL;
}

static int init_ramd_storage(int index, ulong_t data_phys_addr, sector_t capacity, int sect_shift)
{
  int ret;
  blkreq_queue_t *q;
  block_device_t *bdev;
  char name[STOR_NAME_MAXLEN];
  int flags;

  ret = -ENOMEM;
  sprintf(name, "%s%c", "rd", 'a' + index);
  ram_storages[index] = malloc(sizeof(genstor_t));
  if (ram_storages[index] == NULL)
    return ret;

  bdev = malloc(sizeof(block_device_t));
  if (bdev == NULL)
    return ret;

  memset(ram_storages[index], 0, sizeof(genstor_t));
  q = (blkreq_queue_t*)malloc(sizeof(blkreq_queue_t));
  if (q != NULL)
    ret = init_blk_queue(q, ram_storages[index], bdev, MAX_RAM_DISK_REQUESTS,
                         MAX_REQUEST_SECTORS, &blk_group);
  if (ret)
    return ret;

  ret = -ENOMEM;
  bdev->ops = &ramd_bd_ops;
  bdev->media.sect_shift = sect_shift;
  bdev->media.capacity = capacity;
  /* allocate disk data */
  flags = MAP_ANON | MAP_PRIVATE;
  if (data_phys_addr)
    flags |= MAP_PHYS;

  bdev->private = mmap(NULL, capacity << sect_shift, PROT_READ | PROT_WRITE,
                       flags, 0, data_phys_addr);
  if (bdev->private != NULL) {
    ret = init_generic_storage(ram_storages[index], name,
                               &ramd_stor_ops, q, false);
    ram_storages[index]->flags |= STORAGE_SUPP_CMD_ABORT;
  }

  return ret;
}

int main(int argc, char *argv[])
{
  int ret;
  int i;
  blkreq_queue_t *q;
  struct serv_reg_data rd;
  struct completion init_compl;
  ulong_t disk_data;
  ulong_t nr_disks, shift, capacity;

  start_service();
  /* check configuration parameters */
  nr_disks = strtoul(CONFIG_NR_RAMDISKS, NULL, 0);
  shift = strtoul(CONFIG_RAMDISK_SECTOR_SHIFT, NULL, 0);
  capacity = strtoul(CONFIG_RAMDISK_CAPACITY, NULL, 0);
  if ((shift > MAX_RAM_DISK_SECT_SHIFT) || (nr_disks > MAX_NR_RAM_DISKS) ||
      !nr_disks || (capacity == ULONG_MAX)) {
    abort();
  }

  rd.iobuf_size = RAM_DISK_IOBUF_SIZE;
  rd.nr_cmd_threads = 1;
  rd.nr_io_threads = 1;
  if (init_completion(&rd.compl))
    abort();

  start_service();

  /* open terminal */
  i = (open("/dev/ptm0", O_RDONLY) == -1);
  i += (open("/dev/ptm0", O_WRONLY) == -1);
  i += (open("/dev/ptm0", O_WRONLY) == -1);
  if (i)
    goto fail;

  if (init_completion(&init_compl)) {
    printf("[ram_disk.d]: Failed to initialize a synchronization object\n");
    goto fail;
  }

  memset(ram_storages, 0, sizeof(ram_storages));
  sigaddset(&sigset, IOSCHED_SIGNO);

  /* only for testing schedulers purpose */
  if (ramdisk_init_schedulers()) {
    printf("[ram_disk.d]: Block schedulers initialization failed\n");
    goto fail;
  }

  if (init_blk_group(&blk_group)) {
    printf("[ram_disk.d]: Out of resource at block queues group "
           "initialization\n");
    goto fail;
  }

  if (!ret) {
    ret = pthread_create(&io_thread, NULL, io_thread_fn, &init_compl);
    if (ret) {
      printf("[ram_disk.d] I/O thread creation failure\n");
      goto fail;
    }
  }

  for (i = 0; i < nr_disks; i++) {
    if (i < argc - 1) {
      disk_data = strtoul(argv[i + 1], NULL, 0);
      if (disk_data == ULONG_MAX)
        disk_data = 0;
    } else {
      disk_data = 0;
    }
    ret = init_ramd_storage(i, disk_data, capacity, shift);
    if (ret) {
      printf("[ram_disk.d]: Out of memory or synchronization resources at "
                     "storage initialization\n");
      break;
    }

    q = (blkreq_queue_t*)ram_storages[i]->private;
    ret = ramdisk_set_scheduler(q);
    if (ret) {
      printf("[ram_disk.d]: I/O scheduler initialization failed\n");
      break;
    }

    if (disk_data) {
      ret = ram_storages[i]->ops->scan_partitions(ram_storages[i]);
      if (ret && ret != -EINVAL)
        break;
    }
    if (i == nr_disks - 1) {
      complete(&init_compl, 0);
      ret = start_storage_server(ram_storages[i], RAM_DISK_IOBUF_SIZE, 1, 1, NULL);
    } else {
      rd.stor = ram_storages[i];
      ret = start_storage_serv_thread(&rd);
    }
    if (ret) {
      printf("[ram_disk.d] Server start failure, %s storage, %d error\n",
             ram_storages[i]->name, -ret);
      break;
    }
  }

  fail:
    close_console();
    abort();

    return -1;
}
