/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Paul Graphov <graphov@jarios.org> 
 *
 * Serial port driver
 *
 */

#include <stdbool.h>
#include <sys/types.h>
#include <sys/ioport.h>
#include <stdlib.h>
#include <ipc/rpc.h>
#include <pthread.h>
#include <sys/ievent.h>
#include <sys/uevent.h>
#include <sys/scheduler.h>
#include <dircons.h>
#include <chardev.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>

#include <datastructs.h>

#include "serial.h"

#define DRV_NAME "serial.d"

/* I use default values for irqs and ports
   if they are not valid, you can extract it from BIOS:
   0000:0400 COM1's Base Address
   0000:0402 COM2's Base Address
   0000:0404 COM3's Base Address
   0000:0406 COM4's Base Address
*/
typedef struct __inbuf_t {
  size_t size;
  size_t pos;
  char *data;
} inbuf_t;

typedef struct __serial_port {
  uint8_t irq;
  uint16_t ioport;
  char *name;
  uint16_t br_divisor;
  uint8_t data_bits;
  bool parity_enable;
  bool sticky_parity;
  bool even_parity;
  bool break_enable;
  bool stop_bits;
  bool initialized;
  inbuf_t *inbuf;
  // circular_t in_mini_buffer;
  //circular_t out_buffer;
} serial_port_t;

serial_port_t comports[] = {{.irq = 4, .ioport = 0x3F8, .name = "/dev/uart0"},
                            {.irq = 3, .ioport = 0x2F8, .name = "/dev/uart1"},
                            {.irq = 4, .ioport = 0x3E8, .name = "/dev/uart2"},
                            {.irq = 3, .ioport = 0x2E8, .name = "/dev/uart3"}};

static status_t serial_write(int devno, void *buf, size_t size, size_t offset, void **interim);
static status_t serial_read(int devno, void *buf, size_t size, size_t offset, void **interim);
static status_t serial_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize);

static bool serial_irq_hardler(void *data);
static void *irq_thread(void *data);

struct device_ops serial_ops = {
  .read = serial_read,
  .write = serial_write,
  .ioctl = serial_ioctl,
  .log = console_output,
};

device_server_t server;

void do_init_serial(int port)
{
  int portno = comports[port].ioport;
  int reg;

  //circular_init(&comports[port].in_buffer, SERIAL_BUF_SIZE);
  //circular_init(&comports[port].out_buffer, SERIAL_BUF_SIZE);

  outb(portno + 1, 0x00);              // Disable all interrupts
  outb(portno + 3, 0x80);              // Enable DLAB (set baud rate divisor)

  outb(portno + 0, comports[port].br_divisor & 0xff);    // Set divisor to (lo byte)
  outb(portno + 1, comports[port].br_divisor >> 8);      //                (hi byte)

  // calculate line control register
  reg =
    (comports[port].data_bits & 3) |
    (comports[port].stop_bits ? 4 : 0) |
    (comports[port].parity_enable ? 8 : 0) | 
    (comports[port].even_parity ? 16 : 0) |
    (comports[port].sticky_parity ? 32 : 0) |
    (comports[port].break_enable ? 64 : 0);
  
  outb(portno + 3, reg);

  outb(portno + 2, 0xC7);             // Enable FIFO, clear them, with 14-byte threshold (TODO: make tunable)
  outb(portno + 4, 0x0B);             // No modem support
  outb(portno + 1, 0x03);             // Issue an interrupt when output buffer empty or input full.

  comports[port].initialized = true;
}

static inline int can_receive(int port)
{
  return inb(comports[port].ioport + 5) & 1;
}

static inline char read_serial(int port)
{
  return inb(comports[port].ioport);
}

static inline int can_transmit(int port)
{
  return inb(comports[port].ioport + 5) & 0x20;
}

static inline void write_serial(int port, char a)
{
  outb(comports[port].ioport, a);
}

int main(void)
{
  ulong_t i, ret;
  pthread_t irq_thr;

#ifndef CONFIG_SERIAL_GRUB
  start_service();
#endif

  open_console();

  console_printf("[serial.d]\n");

  for(i = 0; i < 4; i++) {
    ret = alloc_ioport_range(comports[i].ioport, 8);
    if(ret != 0) {
      console_printf("Failed to get io ports.\n");
      abort();
    }
  }

  ret = init_multi_device_server(&server, 65536, 65536, DRV_NAME, &serial_ops, 2);

  if(ret != 0) {
    console_printf("Failed to init device server.\n");
    abort();
  }

  for(i = 0; i < 4; i++) {
    comports[i].initialized = false;
    comports[i].br_divisor = 3; //38kbps
    comports[i].data_bits = DATA_BITS_8;
    comports[i].stop_bits = STOP_BITS_1;
    comports[i].parity_enable = 0;

    ret = add_device(i, comports[i].name);

    if(ret < 0) {
      console_printf("Not enough memory.\n");
      abort();
    }

  }

  outb(0x3f9, 5);

  pthread_create(&irq_thr, NULL, irq_thread, NULL);

  device_server_start();

  abort();

  //Unreachable
  return 0;
}

static void do_pending_writes(void) {
  int i;
  inbuf_t *inbuf;

  for(i = 0; i < 4; i++) {
    if((inbuf = comports[i].inbuf)) {
      while(inbuf->pos < inbuf->size && can_transmit(i)) {
        write_serial(i, inbuf->data[inbuf->pos++]);
      }
      if(inbuf->pos == inbuf->size) {
        free(inbuf);
        comports[i].inbuf = NULL;
        wake_device(i, IN);
      }
    }
  }
}

static status_t serial_write(int devno, void *buf, size_t size, size_t offset, void **interim)
{
  /*if interim == NULL - this is the first entry*/
  status_t ret;

  if(devno > 3) return -ENODEV;

  if(!*interim) {
    inbuf_t *inbuf;
    if(size > 1) {
      inbuf = malloc(sizeof(*inbuf));
      if(!inbuf) return -ENOMEM;
      do_init_serial(devno);
      inbuf->data = buf;
      inbuf->size = size;
      inbuf->pos = 1;
      comports[devno].inbuf = inbuf;
      ret = SUSPEND;
    } else {
      ret = 1;
    }
    // transmit one character to initiate interrupt driven transfer
    while(!can_transmit(devno));
    write_serial(devno, *(char *)buf);
    *interim = (void *)1;
  } else {
    ret = size;
  }
  
  return ret;
}

static status_t serial_read(int devno, void *buf, size_t size,size_t offset, void **interim)
{
  return -ENOSYS;
}

static status_t serial_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize)
{
  // TODO: Add termios compliance.
  return -ENOSYS;
}

static void *irq_thread(void *data)
{
  long i, ret;
  status_t st;
  irq_t interrupts[] = {3, 4};
  interrupt_event_t *serial_irq;

  ret = interrupt_event_create(&serial_irq, interrupts, 2);

  if(ret != 0) {
    console_printf("Failed to create interrupt event, return value = %d\n", ret);
    abort();
  }

  for(i = 4; i < 5; i++) {
    ret = interrupt_event_add_handler(serial_irq, i, serial_irq_hardler, (void *)i);
    if(ret != 0) {
      console_printf("Failed to add interrupt %d handler.\n", i);
      abort();
    }
  }

  task_scheduler_control(0,0, SYS_SCHED_CTL_SET_PRIORITY, 50);
  task_scheduler_control(0,0, SYS_SCHED_CTL_SET_POLICY, SCHED_FIFO);

  while(true) {
    st = interrupt_event_wait(serial_irq);
    if(st < 0) {
      console_printf("Failed to wait for interrupt event.\n");
      abort();
    }
  }
}

static bool serial_irq_hardler(void *data)
{
  do_pending_writes();
  return true;
}

