/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS devices i/o filesystem,
 * internal implementation 
 */

#include <sys/types.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <pthread.h>
#include <unistd.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <sys/v2.h>
#include <sys/kcontrol.h>
#include <sys/mman.h>
#include <ns/ns.h>
#include <errno.h>
#include <datalib/usrtc.h>

#include <devio/devio.h>

#include <vfs/vfsdef.h>
#include <vfs/super.h>
#include <vfs/vnode.h>

#include "deviofs.h"

#ifdef DEVIOFS_DEBUG
#include <dircons.h>
#endif

static int __deviofscmp_hash(const void *a,const void *b);
static inline void __entries_clear(struct dirent **ent);

int deviofs_mount(vnode_t *node,char *path,int *pps,int pp,int flags)
{
  deviofs_super_t *diosuper;
  deviofs_node_t *root,*self,*parent;
  time_t now;

  now=time(NULL);
  if(now==(time_t)-1)    now=0;

  /* allocate all */
  if((diosuper=malloc(sizeof(*diosuper)))==NULL) {
    return ENOMEM;
  }
  if((root=malloc(sizeof(*root)))==NULL) {
    free(diosuper);
    return ENOMEM;
  }
  if((self=malloc(sizeof(*self)))==NULL) {
    free(root);
    free(diosuper);
    return ENOMEM;
  }
  if((parent=malloc(sizeof(*parent)))==NULL) {
  __enomem:
    free(self);
    free(root);
    free(diosuper);
    return ENOMEM;
  }
  /* allocate nodes */
  root->node=malloc(sizeof(*root->node));
  self->node=malloc(sizeof(*self->node));
  if(!root->node || !self->node) {
    free(parent);
    goto __enomem;
  }
  if((parent->node=malloc(sizeof(*parent->node)))==NULL) {
    free(parent);
    free(root->node);
    free(self->node);
    goto __enomem;
  }
  /* assign names */
  root->name=strdup("/");
  self->name=strdup(".");
  parent->name=strdup("..");
  /* assign hashes */
  root->hash=vnode_hash_ch(root->name);
  self->hash=vnode_hash_ch(self->name);
  parent->hash=vnode_hash_ch(parent->name);
  /* init dir data on faked entries */
  usrtc_node_init(self->node,self);
  usrtc_node_init(parent->node,parent);
  self->type=DFS_DIR;
  self->entries=NULL;
  parent->type=DFS_DIR;
  parent->entries=NULL;
  /* init structure */
  usrtc_init(&root->childs,USRTC_BST,USRTC_COUNT_T_MAX,__deviofscmp_hash);
  /* insert self and faked parent */
  usrtc_insert(&root->childs,self->node,&self->hash);
  usrtc_insert(&root->childs,parent->node,&parent->hash);
  /* well it's time to init all and assign pointers */
  root->type=DFS_DIR;
  root->ctime=root->mtime=root->atime=now;
  /* device rel */
  root->drv_pid=0; /* pid of device driver */
  root->drv_port=0;  /* driver port */
  root->dev_id=0;    /* driver id number */
  root->dev_ch=0; /* channel to the device */
  /* access structure */
  root->uid=0; /* owner */
  root->gid=0;
  root->mask=0;
  root->mask=(TFS_UREAD|TFS_UWRITE|TFS_UEXEC) | (TFS_GREAD | TFS_GEXEC) | (TFS_OREAD|TFS_OEXEC);
  /* namespace structure */
  root->parent=NULL;
  /* link rel */
  root->hard_link=NULL;
  root->num_links=0;
  root->symlink_path=NULL; /* path to the real file */
  /* directory related */
  root->num_childs=2;
  /* cached dir entries */
  root->entries=NULL;
  node->super->spec_data=diosuper;
  diosuper->root=root;

  return 0;
}

int deviofs_umount(vnode_t *node)
{
  return 0; /* TODO: umount should unregister all devices and clear all nodes */
}

int deviofs_lookup(vnode_t *node,char *path,int *pps,int pp)
{
  deviofs_super_t *super;
  deviofs_node_t *tfsnode,*tfsroot;
  usrtc_node_t *cnode;
  uint64_t srch_hash=0;
  char tname[256];
  int i=0;
  mode_t tfmode=0;

  if(strlen(path)==1 && path[0]=='/') { /* root directory */
    super=(deviofs_super_t*)node->super->spec_data;
    if(!super) return ENOENT;
    tfsnode=super->root;
    if(!tfsnode) return ENOENT;
    /* fill out values */
    node->spec_data=tfsnode;
    /* fill out stat */
    node->vstat.st_dev=0;
    node->vstat.st_ino=0;
    node->vstat.st_mode=0;
    node->vstat.st_mode|=(00000755);
    node->vstat.st_mode|=S_IFDIR;
    node->vstat.st_uid=0;
    node->vstat.st_gid=0;
    node->vstat.st_rdev=0;
    node->vstat.st_blksize=node->super->block_size;
    node->vstat.st_size=usrtc_count(&tfsnode->childs);
    node->vstat.st_blocks=0;
    node->vstat.st_nlink=0;
    node->vstat.st_mtime=tfsnode->mtime;
    node->vstat.st_atime=tfsnode->atime;
    node->vstat.st_ctime=tfsnode->ctime;
    return 0;
  } else {
    super=(deviofs_super_t*)node->super->spec_data;
    if(!super) return ENOENT;
    tfsroot=super->root;
    if(!tfsroot) return ENOENT;
    for(i=0;i<pp;i++) {
      /**
       * *NOTE:* lookup algorythm ->
       * take hash - lookup in root, if found -> if it end - fill up,
       * if not - make a new hash and change root -> if not found -> ENOENT anyway
       */
      memset(tname,'\0',256);

      if((i+1)>=pp) {
	strncpy(tname,path+pps[i]+1,strlen(path)-pps[i]);
      } else
	strncpy(tname,path+pps[i]+1,pps[i+1]-pps[i]-1);
#ifdef DEVIOFS_DEBUG
      console_printf("[deviofs] {deviofs_lookup} tname = %s\n",tname);
#endif
      /* make new hash */
      srch_hash=vnode_hash_ch(tname);
      /* NOTE: for flamefreaks - i've check it at all */
      cnode=usrtc_lookup(&tfsroot->childs,&srch_hash);
      if(!cnode) 
	return ENOENT; /* ENOENT anyway */
      else {
	if((i+1)>=pp) { /* found ! */
	  tfsnode=usrtc_node_getdata(cnode);
	  /* fill out values */
	  node->spec_data=tfsnode;
	  /* fill out stat */
	  node->vstat.st_dev=0;
	  node->vstat.st_ino=i;
	  node->vstat.st_mode=0;
	  node->vstat.st_blocks=0;
	  node->vstat.st_size=0;
	  /* check it out */
	  if(tfsnode->mask & TFS_UREAD) tfmode|=S_IRUSR;
	  if(tfsnode->mask & TFS_UWRITE) tfmode|=S_IWUSR;
	  if(tfsnode->mask & TFS_UEXEC) tfmode|=S_IXUSR;
	  if(tfsnode->mask & TFS_GREAD) tfmode|=S_IRGRP;
	  if(tfsnode->mask & TFS_GWRITE) tfmode|=S_IWGRP;
	  if(tfsnode->mask & TFS_GEXEC) tfmode|=S_IXGRP;
	  if(tfsnode->mask & TFS_OREAD) tfmode|=S_IROTH;
	  if(tfsnode->mask & TFS_OWRITE) tfmode|=S_IWOTH;
	  if(tfsnode->mask & TFS_OEXEC) tfmode|=S_IXOTH;
	  node->vstat.st_rdev=0;
	  /* set mode */
	  node->vstat.st_mode|=tfmode;
	  switch(tfsnode->type) {
	  case DFS_BLKDEV:
	    node->vstat.st_mode|=S_IFBLK;
	    node->vstat.st_rdev=tfsnode->drv_pid;
	    node->vstat.st_dev=tfsnode->drv_port;
	    node->vstat.st_ino=tfsnode->dev_id;
	    break;
	  case DFS_CHARDEV:
	    node->vstat.st_mode|=S_IFCHR;
	    node->vstat.st_rdev=tfsnode->drv_pid;
	    node->vstat.st_dev=tfsnode->drv_port;
	    node->vstat.st_ino=tfsnode->dev_id;
	    break;
	  case DFS_LINK:
	    node->vstat.st_mode|=S_IFLNK;
	    node->vstat.st_rdev=tfsnode->hard_link->drv_pid;
	    node->vstat.st_dev=tfsnode->hard_link->drv_port;
	    node->vstat.st_ino=tfsnode->hard_link->dev_id;
	    break;
	  case DFS_SYMLINK:
	    node->vstat.st_mode|=S_IFLNK;
	    node->vstat.st_rdev=0;
	    node->vstat.st_dev=0;
	    node->vstat.st_ino=0;
	    break;
	  case DFS_DIR:
	    node->vstat.st_mode|=S_IFDIR;
	    break;
	  }
	  /* fill correct values */
	  node->vstat.st_uid=tfsnode->uid;
	  node->vstat.st_gid=tfsnode->gid;
	  node->vstat.st_blksize=node->super->block_size;
	  node->vstat.st_nlink=tfsnode->num_links;
	  node->vstat.st_mtime=tfsnode->mtime;
	  node->vstat.st_atime=tfsnode->atime;
	  node->vstat.st_ctime=tfsnode->ctime;
	  return 0;
	} else { /* we will continue */
	  /* change root */
	  tfsroot=usrtc_node_getdata(cnode);
	  /* this is directory ? */
	  if(tfsroot->type!=DFS_DIR)
	    return ENOENT;
	}
      }
    }
  }

  return ENOENT;
}

int deviofs_create(vnode_t *node,char *path,int *pps,int pp,int mode,uid_t uid,gid_t gid)
{
  usrtc_node_t *cnode,*newnode;
  deviofs_node_t *tfsroot;
  deviofs_node_t *newn;
  deviofs_super_t *super;
  char tname[256];
  time_t now;
  uint64_t srch_hash;
  int i;

  super=(deviofs_super_t*)node->super->spec_data;
  tfsroot=super->root;

  /* *NOTE:* 0)find root; 1) check permission 2) insert */
  for(i=0;i<pp;i++) {
    memset(tname,'\0',256);
    
    if((i+1)>=pp) {
      strncpy(tname,path+pps[i]+1,strlen(path)-pps[i]);
    } else
      strncpy(tname,path+pps[i]+1,pps[i+1]-pps[i]-1);
#ifdef DEVIOFS_DEBUG
    console_printf("[deviofs] {deviofs_create(intr. lookup)} tname = %s\n",tname);
#endif
    /* make new hash */
    srch_hash=vnode_hash_ch(tname);
    /* NOTE: for flamefreaks - i've check it at all */
    cnode=usrtc_lookup(&tfsroot->childs,&srch_hash);
    if((i+1)>=pp && !cnode) { /* our client */
      /* user should have write and exec access to directory */
      if(uid!=0 || gid!=0) { /* it's not a supervisor - will check */
	if(uid==tfsroot->uid && (!(tfsroot->mask && TFS_UWRITE) && !(tfsroot->mask && TFS_UEXEC))) 
	  return EACCES;
	else if(gid==tfsroot->gid && (!(tfsroot->mask && TFS_GWRITE) && !(tfsroot->mask && TFS_GEXEC)))
	  return EACCES;
	else if(!(tfsroot->mask && TFS_OWRITE) && !(tfsroot->mask && TFS_OEXEC))
	  return EACCES;
      }
      /* well, now we're can insert this */
      if((newn=malloc(sizeof(*newn)))==NULL) return ENOMEM;
      if((newnode=malloc(sizeof(*newnode)))==NULL) return ENOMEM;
      /* associate it with node */
      usrtc_node_init(newnode,newn);
      newn->node=newnode;
      /* give name */
      if((newn->name=strdup(tname))==NULL) {
	free(newn);
	free(newnode);
	return ENOSPC;
      }
      /* hash */
      newn->hash=vnode_hash_ch(newn->name);
#ifdef DEVIOFS_DEBUG
      console_printf("[deviofs] create-> (%s)\n",newn->name);
#endif
      newn->parent=tfsroot; /* mark our parent */
      newn->entries=NULL; /* no cached anything */
      newn->type=DFS_CHARDEV;
      newn->num_links=0;
      now=time(NULL);
      if(now!=(time_t)-1)
	newn->mtime=newn->atime=newn->ctime=now;
      else
	newn->mtime=newn->atime=newn->ctime=0;
      /* owner */
      newn->uid=uid;
      newn->gid=gid;
      /* mode */
      newn->mask=0;
      if(mode & S_IRUSR) newn->mask|=TFS_UREAD;
      if(mode & S_IWUSR) newn->mask|=TFS_UWRITE;
      if(mode & S_IXUSR) newn->mask|=TFS_UEXEC;
      if(mode & S_IRGRP) newn->mask|=TFS_GREAD;
      if(mode & S_IWGRP) newn->mask|=TFS_GWRITE;
      if(mode & S_IXGRP) newn->mask|=TFS_GEXEC;
      if(mode & S_IROTH) newn->mask|=TFS_OREAD;
      if(mode & S_IWOTH) newn->mask|=TFS_OWRITE;
      if(mode & S_IXOTH) newn->mask|=TFS_OEXEC;
      /* well, let's fill vnode stats */
      node->vstat.st_size=0;
      node->vstat.st_mode=0;
      node->vstat.st_mode|=mode;
      node->vstat.st_mode|=S_IFCHR;
      node->vstat.st_uid=newn->uid;
      node->vstat.st_gid=newn->gid;
      node->vstat.st_rdev=0;
      node->vstat.st_dev=0;
      node->vstat.st_ino=0;
      node->vstat.st_blksize=node->super->block_size;
      node->vstat.st_nlink=newn->num_links;
      node->vstat.st_mtime=newn->mtime;
      node->vstat.st_atime=newn->atime;
      node->vstat.st_ctime=newn->ctime;
      node->vstat.st_blocks=0;
      /* big shit deals are made */
      usrtc_insert(&tfsroot->childs,newnode,&newn->hash); /* insert to the parent */
      node->spec_data=newn;
      return 0;
    } else {
      if(!cnode) 
	return ENOENT; /* ENOENT anyway */
      else {
	/* change root */
	tfsroot=usrtc_node_getdata(cnode);
	/* this is directory ? */
	if(tfsroot->type!=DFS_DIR)
	  return ENOENT;
      }
    }
  }
  
  return ENOENT;
}

int deviofs_opendir(vnode_t *node,char *path,int *pps,int pp)
{
  deviofs_node_t *tnode=(deviofs_node_t*)node->spec_data;
  deviofs_node_t *cnode;
  usrtc_node_t *acnode;
  int i=0;
  int ecnt;
  time_t now;

  now=time(NULL);
  tnode->atime=now;
  node->vstat.st_atime=now;

  if(tnode->entries) /* already cached */
    return 0;

  ecnt=usrtc_count(&tnode->childs);
  tnode->entries=malloc(sizeof(uintptr_t)*(ecnt+1));
  if(!tnode->entries) return ENOMEM;
  memset(tnode->entries,0,sizeof(uintptr_t)*(ecnt+1));
  acnode=usrtc_first(&tnode->childs);
#ifdef DEVIOFS_DEBUG
  console_printf("[deviofs] opendir-> ecnt=%d\n",ecnt);
#endif
  for(i=0;i<ecnt;i++) {
    tnode->entries[i]=malloc(sizeof(struct dirent));
    if(!tnode->entries[i]) {
      free(tnode->entries);
      tnode->entries=NULL;
      return ENOMEM;
    }
    cnode=(deviofs_node_t*)usrtc_node_getdata(acnode);
    strcpy(tnode->entries[i]->d_name,cnode->name);
#ifdef DEVIOFS_DEBUG
    console_printf("[deviofs] lu.name='%s'\n",cnode->name);
#endif
    tnode->entries[i]->d_ino=cnode->dev_id;
    tnode->entries[i]->d_off=i;
    tnode->entries[i]->d_namlen=strlen(cnode->name);
    tnode->entries[i]->d_rec_len=sizeof(*tnode->entries[i]);
#ifdef DEVIOFS_DEBUG
    console_printf("[deviofs] acnode->next entry\n");
#endif
    acnode=usrtc_next(&tnode->childs,acnode);
  }

  return 0;
}

int deviofs_ioctl(vnode_t *node,uni_rpc_t *rpc)
{
  deviofs_node_t *tnode=(deviofs_node_t*)node->spec_data;

  if(!tnode) return ENOENT;
  if(tnode->type!=DFS_BLKDEV && tnode->type!=DFS_CHARDEV) 
    return EINVAL;
  if(rpc->arg0!=DEV_ASSIGN) return EINVAL;
  if(rpc->arg5!=DFS_BLKDEV && rpc->arg5!=DFS_CHARDEV)
    return EINVAL;
  if(tnode->drv_pid!=0) return EBUSY;

  tnode->type=rpc->arg5;
  tnode->drv_pid=rpc->arg2;
  tnode->drv_port=rpc->arg3;
  tnode->dev_id=rpc->arg4;
  tnode->dev_ch=ipc_channel_open(tnode->drv_pid,rpc->arg6,IPC_BLOCKED_ACCESS);
  if(tnode->dev_ch<0) {
    tnode->drv_pid=0;
    tnode->drv_port=0;
    tnode->dev_id=0;
    rpc->rpc_call=tnode->dev_ch;
    return EIO;
  }
  /* update stat */
  rpc->rpc_call=0;
  if(tnode->type==DFS_BLKDEV)
    node->vstat.st_mode|=S_IFBLK;
  if(tnode->type==DFS_CHARDEV)
    node->vstat.st_mode|=S_IFCHR;
  node->vstat.st_rdev=tnode->drv_pid;
  node->vstat.st_dev=tnode->drv_port;
  node->vstat.st_ino=tnode->dev_id;

  return 0;
}

int deviofs_close(vnode_t *node)
{
  deviofs_node_t *tnode=(deviofs_node_t*)node->spec_data;

  if(tnode->entries) {
    __entries_clear(tnode->entries);
    free(tnode->entries);
    tnode->entries=NULL;
  }

  node->spec_data=NULL; /* clean */

  return 0;
}

int deviofs_readdir(vnode_t *node,size_t size,off_t off,struct dirent *dentry)
{
  deviofs_node_t *tnode=(deviofs_node_t*)node->spec_data;
  ulong_t ecnt=usrtc_count(&tnode->childs);

  if(off>ecnt) {
    errno=EOVERFLOW;
    return -1;
  } else if(tnode->entries[off]==NULL) {
    return DIRENT_EL;
  } else if(tnode->atime<tnode->mtime) {
    __entries_clear(tnode->entries);
    free(tnode->entries);
    tnode->entries=NULL;
    /* reread all */
    errno=deviofs_opendir(node,NULL,NULL,0);
    return DIRENT_RR;
  }

  strcpy(dentry->d_name,tnode->entries[off]->d_name);
  dentry->d_ino=tnode->entries[off]->d_ino;
  dentry->d_off=tnode->entries[off]->d_off;
  dentry->d_namlen=tnode->entries[off]->d_namlen;
  dentry->d_rec_len=tnode->entries[off]->d_rec_len;

  return 0;
}

int deviofs_createdir(vnode_t *node,char *path,int *pps,int pp,int mode,uid_t uid,gid_t gid)
{
  usrtc_node_t *cnode,*newnode,*self,*parent;
  deviofs_node_t *tfsroot,*parentnode,*selfnode;
  deviofs_node_t *newn;
  deviofs_super_t *super;
  char tname[256];
  time_t now;
  uint64_t srch_hash;
  int i;

  super=(deviofs_super_t*)node->super->spec_data;
  tfsroot=super->root;

  /* *NOTE:* 0)find root; 1) check permission 2) insert */
  for(i=0;i<pp;i++) {
    memset(tname,'\0',256);
    
    if((i+1)>=pp) {
      strncpy(tname,path+pps[i]+1,strlen(path)-pps[i]);
    } else
      strncpy(tname,path+pps[i]+1,pps[i+1]-pps[i]-1);
#ifdef TMPFS_DEBUG
    console_printf("[deviofs] {deviofs_createdir(intr. lookup)} tname = %s\n",tname);
#endif
    /* make new hash */
    srch_hash=vnode_hash_ch(tname);
    /* NOTE: for flamefreaks - i've check it at all */
    cnode=usrtc_lookup(&tfsroot->childs,&srch_hash);
    if((i+1)>=pp && !cnode) { /* our client */
      /* user should have write and exec access to directory */
      if(uid!=0 || gid!=0) { /* it's not a supervisor - will check */
	if(uid==tfsroot->uid && (!(tfsroot->mask && TFS_UWRITE) && !(tfsroot->mask && TFS_UEXEC))) 
	  return EACCES;
	else if(gid==tfsroot->gid && (!(tfsroot->mask && TFS_GWRITE) && !(tfsroot->mask && TFS_GEXEC)))
	  return EACCES;
	else if(!(tfsroot->mask && TFS_OWRITE) && !(tfsroot->mask && TFS_OEXEC))
	  return EACCES;
      }
      /* well, now we're can insert this */
      if((newn=malloc(sizeof(*newn)))==NULL) return ENOMEM;
      if((newnode=malloc(sizeof(*newnode)))==NULL) return ENOMEM;
      /* associate it with node */
      usrtc_node_init(newnode,newn);
      newn->node=newnode;
      /* give name */
      if((newn->name=strdup(tname))==NULL) {
      __crdir_enospc:
	free(newn);
	free(newnode);
	return ENOSPC;
      }
      /* hash */
      newn->hash=vnode_hash_ch(newn->name);
#ifdef DEVIOFS_DEBUG
      console_printf("[deviofs] createdir-> (%s)\n",newn->name);
#endif
      usrtc_init(&newn->childs,USRTC_BST,USRTC_COUNT_T_MAX,__deviofscmp_hash);
      if((parent=malloc(sizeof(*self)))==NULL) {
	free(newn->name);
	goto __crdir_enospc;
      }

      /* insert self with other node */
      if((parentnode=malloc(sizeof(*parentnode)))==NULL) {
	free(parent);
	free(newn->name);
	goto __crdir_enospc;
      }
      if((selfnode=malloc(sizeof(*selfnode)))==NULL) {
	free(parentnode);
	free(parent);
	free(newn->name);
	goto __crdir_enospc;
      }
      if((self=malloc(sizeof(*self)))==NULL) {
	free(parentnode);
	free(selfnode);
	free(parent);
	free(newn->name);
	goto __crdir_enospc;
      } else {
	selfnode->name=strdup(".");
	selfnode->hash=vnode_hash_ch(selfnode->name);
	selfnode->parent=NULL;
	selfnode->type=DFS_DIR;
	parentnode->name=strdup("..");
	parentnode->hash=vnode_hash_ch(parentnode->name);
	parentnode->parent=NULL;
	parentnode->type=DFS_DIR;
	/* insert self */
	usrtc_node_init(self,selfnode);
	usrtc_insert(&newn->childs,self,&selfnode->hash);
	/* insert parent */
	usrtc_node_init(parent,parentnode);
	usrtc_insert(&newn->childs,parent,&parentnode->hash);
      }
      newn->parent=tfsroot; /* mark our parent */
      newn->entries=NULL; /* no cached anything */
      newn->type=DFS_DIR;
      newn->num_links=0;
      now=time(NULL);
      if(now!=(time_t)-1)
	newn->mtime=newn->atime=newn->ctime=now;
      else
	newn->mtime=newn->atime=newn->ctime=0;
      /* owner */
      newn->uid=uid;
      newn->gid=gid;
      /* mode */
      newn->mask=0;
      if(mode & S_IRUSR) newn->mask|=TFS_UREAD;
      if(mode & S_IWUSR) newn->mask|=TFS_UWRITE;
      if(mode & S_IXUSR) newn->mask|=TFS_UEXEC;
      if(mode & S_IRGRP) newn->mask|=TFS_GREAD;
      if(mode & S_IWGRP) newn->mask|=TFS_GWRITE;
      if(mode & S_IXGRP) newn->mask|=TFS_GEXEC;
      if(mode & S_IROTH) newn->mask|=TFS_OREAD;
      if(mode & S_IWOTH) newn->mask|=TFS_OWRITE;
      if(mode & S_IXOTH) newn->mask|=TFS_OEXEC;
      /* big shit deals are made */
      usrtc_insert(&tfsroot->childs,newnode,&newn->hash); /* insert to the parent */
      return 0;
    } else {
      if(!cnode) 
	return ENOENT; /* ENOENT anyway */
      else {
	/* change root */
	tfsroot=usrtc_node_getdata(cnode);
	/* this is directory ? */
	if(tfsroot->type!=DFS_DIR)
	  return ENOENT;
      }
    }
  }
  
  return ENOENT;
}

int deviofs_unlink(vnode_t *node,uid_t uid,gid_t gid)
{
  deviofs_node_t *tfsnode;
  deviofs_node_t *parent;
  int ac=0;

  tfsnode=(deviofs_node_t *)(node->spec_data);
  if(!tfsnode) return ENOENT; /* paranoic */
  else parent=tfsnode->parent;

  if(tfsnode->drv_pid!=0) return EBUSY;

  /* check rights */
  if(uid!=0) { /* parent *always* is directory  */
    if(parent->uid==uid && 
       ((parent->mask & TFS_UEXEC) && (parent->mask & TFS_UWRITE))) ac++;
    else if(parent->gid==gid && 
	    ((parent->mask & TFS_GEXEC) && (parent->mask & TFS_GWRITE))) ac++;
    else if((parent->mask & TFS_OEXEC) && (parent->mask & TFS_OWRITE)) ac++;
    if(!ac) return EACCES;
    else ac=0;
    /* check entry rights */
    if(tfsnode->uid==uid && (tfsnode->mask & TFS_UWRITE)) ac++;
    else if(tfsnode->gid==gid && (tfsnode->mask & TFS_GWRITE)) ac++;
    else if(tfsnode->mask & TFS_OWRITE) ac++;
    if(!ac) return EACCES;
  }

  /* free other stuff */
  free(tfsnode->name);
  /* remove this friend from parent */
  usrtc_delete(&parent->childs,tfsnode->node);
  free(tfsnode->node);
  free(tfsnode);
  node->spec_data=NULL; /* shootout */

  return 0;
}

int deviofs_unlinkdir(vnode_t *node,uid_t uid,gid_t gid)
{
  deviofs_node_t *tfsnode,*child;
  deviofs_node_t *parent;
  usrtc_node_t *cnode;
  int ac=0;

  tfsnode=(deviofs_node_t *)(node->spec_data);
  if(!tfsnode) return ENOENT; /* paranoic */
  else parent=tfsnode->parent;

  /* check rights */
  if(uid!=0) { /* parent *always* is directory  */
    if(parent->uid==uid && 
       ((parent->mask & TFS_UEXEC) && (parent->mask & TFS_UWRITE))) ac++;
    else if(parent->gid==gid && 
	    ((parent->mask & TFS_GEXEC) && (parent->mask & TFS_GWRITE))) ac++;
    else if((parent->mask & TFS_OEXEC) && (parent->mask & TFS_OWRITE)) ac++;
    if(!ac) return EACCES;
    else ac=0;
    /* check entry rights */
    if(tfsnode->uid==uid && (tfsnode->mask & TFS_UWRITE)) ac++;
    else if(tfsnode->gid==gid && (tfsnode->mask & TFS_GWRITE)) ac++;
    else if(tfsnode->mask & TFS_OWRITE) ac++;
    if(!ac) return EACCES;
  }

  /* well let's see - first remove all data from a directory */
  if(usrtc_count(&tfsnode->childs)>2)
    return ENOTEMPTY;
  else {
    while((cnode=usrtc_first(&tfsnode->childs))!=NULL) {
      child=(deviofs_node_t *)usrtc_node_getdata(cnode);
      usrtc_delete(&tfsnode->childs,cnode);
      free(child->name);
      free(child);
      free(cnode);
    }
  }
  /* if we're have a cached directory context */
  if(tfsnode->entries) {
    __entries_clear(tfsnode->entries);
    free(tfsnode->entries);
  }
  /* free other stuff */
  free(tfsnode->name);
  /* remove this friend from parent */
  usrtc_delete(&parent->childs,tfsnode->node);
  free(tfsnode->node);
  free(tfsnode);
  node->spec_data=NULL; /* shootout */

  return 0;
}

/* hooks implementation */
/* hooks */
int deviofs_addhook(pidsession_t *sess,fessid_t *nid)
{
  pid_t apid=sess->pid;
  deviofs_node_t *dnode;
  uni_rpc_t rpc;
  uint16_t unique;
  int r;

  if(!nid->vnode) return 0;
  dnode=(deviofs_node_t *)nid->vnode->spec_data;
  if(!dnode) return 0;

  if(dnode->drv_pid==0) return 0;

  unique=nid->id;

  /* complete message */
  rpc.rpc_call=RPC_REGISTER;
  rpc.arg0=apid; /* affected pid */
  rpc.arg1=unique; /* uniqual id of session */
  rpc.arg2=dnode->dev_id; /* id of device */
  rpc.arg3=nid->flags; /* flags */
  rpc.arg4=nid->mode;

  r=ipc_channel_msgsnd(dnode->dev_ch,sizeof(rpc),&rpc,sizeof(rpc),&rpc);
  if(r!=0) return -r;
  else 
    return rpc.rpc_call;
}

int deviofs_closehook(pidsession_t *sess,fessid_t *nid)
{
  pid_t apid=sess->pid;
  deviofs_node_t *dnode;
  uni_rpc_t rpc;
  uint16_t unique;
  int r;

  if(!nid->vnode) return 0;
  dnode=(deviofs_node_t *)nid->vnode->spec_data;
  if(!dnode) return 0;

  if(dnode->drv_pid==0) return 0;

  unique=nid->id;

  /* complete message */
  rpc.rpc_call=RPC_UNREGISTER;
  rpc.arg0=apid; /* affected pid */
  rpc.arg1=unique; /* uniqual id of session */
  rpc.arg2=dnode->dev_id; /* id of device */
  rpc.arg3=nid->flags; /* flags */
  rpc.arg4=nid->mode;

  r=ipc_channel_msgsnd(dnode->dev_ch,sizeof(rpc),&rpc,sizeof(rpc),&rpc);
  if(r!=0) return -r;
  else 
    return rpc.rpc_call;
}

int deviofs_killhook(pidsession_t *sess) 
{
  pid_t apid=sess->pid;
  deviofs_node_t *dnode;
  usrtc_node_t *node;
  uni_rpc_t rpc;
  fessid_t *nid;
  pid_t apids[128];
  chid_t achs[128];
  int cnt=0;
  int r,i;

  node=usrtc_first(&sess->tree);
  while(node!=NULL) {
    nid=usrtc_node_getdata(node);
    if(!nid->vnode) continue;
    dnode=(deviofs_node_t *)nid->vnode->spec_data;
    if(!dnode) continue;
    if(dnode->drv_pid==0) continue;
    for(i=0;i<cnt;i++) {
      if(apids[i]==dnode->drv_pid) continue;
      else {
	apids[cnt]=dnode->drv_pid;
	achs[cnt]=dnode->dev_ch;
	cnt++;
      }
    }
    if(!cnt) {
      apids[cnt]=dnode->drv_pid;
      achs[cnt]=dnode->dev_ch;
      cnt++;
    }
    node=usrtc_next(&sess->tree,node);
  }

  for(i=0;i<cnt;i++) {
    /* complete message */
    rpc.rpc_call=RPC_PIDFDKILL;
    rpc.arg0=apid; /* affected pid */
    
    r=ipc_channel_msgsnd(achs[i],sizeof(rpc),&rpc,sizeof(rpc),&rpc);
  }

  return 0;
}

int deviofs_clonehook(pidsession_t *sess,pid_t pid) 
{
  pid_t apid=sess->pid;
  deviofs_node_t *dnode;
  usrtc_node_t *node;
  uni_rpc_t rpc;
  fessid_t *nid;
  pid_t apids[128];
  chid_t achs[128];
  int cnt=0;
  int r,i;

  node=usrtc_first(&sess->tree);
  while(node!=NULL) {
    nid=usrtc_node_getdata(node);
    if(!nid->vnode) continue;
    dnode=(deviofs_node_t *)nid->vnode->spec_data;
    if(!dnode) continue;
    if(dnode->drv_pid==0) continue;
    for(i=0;i<cnt;i++) {
      if(apids[i]==dnode->drv_pid) continue;
      else {
	apids[cnt]=dnode->drv_pid;
	achs[cnt]=dnode->dev_ch;
	cnt++;
      }
    }
    if(!cnt) {
      apids[cnt]=dnode->drv_pid;
      achs[cnt]=dnode->dev_ch;
      cnt++;
    }
    node=usrtc_next(&sess->tree,node);
  }

  for(i=0;i<cnt;i++) {
    /* complete message */
    rpc.rpc_call=RPC_PIDFDCLONE;
    rpc.arg0=apid; /* affected pid */
    rpc.arg1=pid;  /* new pid */
    r=ipc_channel_msgsnd(achs[i],sizeof(rpc),&rpc,sizeof(rpc),&rpc);
  }

  return 0;
}

/* local functions */
static int __deviofscmp_hash(const void *a,const void *b)
{
  return *(uint64_t*)a-*(uint64_t*)b;
}

static inline void __entries_clear(struct dirent **ent)
{
  while(*ent!=NULL)    free(*ent++);

  return;
}
