/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008,2009 Dmitriy Balakshin <balakshin@jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS init file system,
 * internal implementation 
 */

#include <sys/types.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <pthread.h>
#include <unistd.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <sys/v2.h>
#include <sys/kcontrol.h>
#include <sys/mman.h>
#include <ns/ns.h>
#include <errno.h>
#include <datalib/usrtc.h>

#include <vfs/vfsdef.h>
#include <vfs/super.h>
#include <vfs/vnode.h>

#include "initfs.h"

#ifdef INITFS_DEBUG
#include <dircons.h>
#endif

/* local variables */
static struct initfs_shead_t super_head;

/* init fs specific funcs local prototypes */
static int handle_ramdisk(uintptr_t *start,uintptr_t *size);
static size_t oct2int(const char *str, size_t size);
static const void* get_next_entry(const void *pos);
static size_t get_entry_name(const void *hdr,char *name);

int initfs_mount(vnode_t *node,char *path,int *pps,int pp,int flags)
{
  uintptr_t ssst,sssize;
  int r=handle_ramdisk(&ssst,&sssize);

  if(r!=0)    return r;

  super_head.start=mmap(NULL,sssize,PROT_READ|PROT_WRITE,MAP_PHYS|MAP_PRIVATE,NOFD,(off_t)ssst);
  if(!super_head.start) {
    return errno;
  }
  super_head.size=sssize;
  node->super->spec_data=(void *)&super_head;
  /* fill super block info */
  node->super->block_size=BLOCK_SIZE;
  node->super->slice_size=sssize/node->super->block_size;

#ifdef INITFS_DEBUG
  //console_printf("[initfs] mapped image at %p with size %dKb.\n",super_head.start,sssize/1024);
  console_printf("[initfs] image with %d blocks (%dKb).\n",node->super->slice_size,
		 (node->super->slice_size*BLOCK_SIZE)/1024);
#endif

  return 0;
}

int initfs_umount(vnode_t *node)
{
  struct initfs_shead_t *shead=(struct initfs_shead_t *)(node->super->spec_data);

  munmap(shead->start,shead->size);

  return 0;
}

int initfs_lookup(vnode_t *node,char *path,int *pps,int pp)
{
  struct tar_header *header;
  struct initfs_node_t *ifsnode;
  char pname[256];
  char fname[256];
  char *tpath=path+sizeof(char);
  int r=ENOENT;

  if(strlen(path)==1 && path[0]=='/') { /* root node <- catch up */
    ifsnode=malloc(sizeof(*ifsnode));
    if(!ifsnode) {
      r=ENOMEM;
      goto __exit;
    }
    node->vstat.st_dev=0;
    node->vstat.st_ino=0;
    node->vstat.st_mode=0;
    node->vstat.st_mode|=(00000755);
    node->vstat.st_mode|=S_IFDIR;
    node->vstat.st_uid=0;
    node->vstat.st_gid=0;
    node->vstat.st_rdev=0;
    node->vstat.st_blksize=BLOCK_SIZE;
    node->vstat.st_size=0;
    node->vstat.st_blocks=node->vstat.st_size/node->vstat.st_blksize;
    node->vstat.st_nlink=0;
    node->vstat.st_mtime=0;
    node->vstat.st_atime=node->vstat.st_mtime=node->vstat.st_ctime;
    ifsnode->data=NULL;
    ifsnode->dentries=NULL;
    ifsnode->dir_entries=0;
    strcpy(ifsnode->head.name,"/");
    node->spec_data=ifsnode;
    r=0;
  } else {
    header=NULL;
    while ((header = (struct tar_header*)get_next_entry(header)) != NULL) {
      memset(pname,'\0',256);
      memset(fname,'\0',256);
      get_entry_name(header,pname);
      /* the fix for name */
      strcpy((char *)fname,tpath);
      fname[strlen(path)-1]='/';
      fname[strlen(path)]='\0';
      if (!strcmp(pname,tpath) || !strcmp(pname,fname)) { /* found ! */
	ifsnode=malloc(sizeof(*ifsnode));
	if(!ifsnode) {
	  r=ENOMEM;
	  goto __exit;
	}
	/* fill structure stat */
	memcpy(&ifsnode->head,header,sizeof(*header));
	node->vstat.st_dev=0;
	node->vstat.st_ino=0;
	node->vstat.st_mode=0;
	node->vstat.st_mode|=oct2int(header->mode,8);
	switch(header->typeflag) {
        case LNKTYPE:     /* Hard link */
	case REGTYPE:     /* Regular file (preferred code) */
	case AREGTYPE:    /* Regular file (alternate code) */
	  node->vstat.st_mode|=S_IFREG;
	  break;
	case SYMTYPE:     /* Symbolic link (hard is internally supported) */
	  node->vstat.st_mode|=S_IFLNK;
	  break;
	case CHRTYPE:     /* Character special */
	  node->vstat.st_mode|=S_IFCHR;
	  break;
	case BLKTYPE:     /* Block special */
	  node->vstat.st_mode|=S_IFBLK;
	  break;
	case DIRTYPE:     /* Directory */
	  node->vstat.st_mode|=S_IFDIR;
	  break;
	case FIFOTYPE:     /* Named pipe */
	  node->vstat.st_mode|=S_IFIFO;
	  break;
	case CONTTYPE:      /* Contiguous file */
	default:
	  node->vstat.st_mode|=S_IFREG;
	  break;
	}
	node->vstat.st_uid=oct2int(header->uid,8);
	node->vstat.st_gid=oct2int(header->gid,8);
	node->vstat.st_rdev=0;
	node->vstat.st_blksize=BLOCK_SIZE;
	node->vstat.st_size=oct2int(header->size,12);
	node->vstat.st_blocks=node->vstat.st_size/node->vstat.st_blksize;
	node->vstat.st_nlink=0;
	node->vstat.st_mtime=oct2int(header->mtime,12);
	node->vstat.st_atime=node->vstat.st_mtime=node->vstat.st_ctime;
	if(!S_ISDIR(node->vstat.st_mode)) {
	  ifsnode->data=(char*)header+BLOCK_SIZE;
	  ifsnode->dentries=NULL;
	  ifsnode->dir_entries=0;
	} else {
	  ifsnode->data=NULL;
	  ifsnode->dentries=NULL;
	  ifsnode->dir_entries=0;
	}
	node->spec_data=ifsnode;
	r=0;
	break;
      }
    }
  }

 __exit:
  return r;
}

void *initfs_read(vnode_t *node,void *buf,size_t *size,off_t offset)
{
  size_t nsize=*size,rs;
  struct initfs_node_t *p;

  if(!node)    return NULL;
  if(!nsize)    return NULL;

  if(S_ISDIR(node->vstat.st_mode))    return NULL;

  p=(struct initfs_node_t*)node->spec_data;
  if((offset+nsize)>node->vstat.st_size) {
    rs=node->vstat.st_size-(offset+nsize);
    if(rs<0) return NULL;
    *size=rs;
  }

  return (void *)((char *)(p->data)+offset);
}

int initfs_readlink(vnode_t *node,char *linkname,size_t *lnksize)
{
  struct initfs_node_t *p;

  p=(struct initfs_node_t*)node->spec_data;
  strcpy(linkname,p->head.linkname);
  *lnksize=strlen(p->head.linkname);

  return 0;
}

int initfs_opendir(vnode_t *node,char *path,int *pps,int pp)
{
  int fcount=0,i,ra,plen;
  int sick_val=100;
  char pname[256];
  char *tpath=path+sizeof(char);
  struct dirent *fdirent;
  struct initfs_node_t *ifsnode;
  struct tar_header *header;
  struct dirent **ndirs;
  int op[2];

  ifsnode=(struct initfs_node_t *)node->spec_data;
  ifsnode->dentries=malloc(sizeof(uintptr_t)*sick_val);
  if(!ifsnode->dentries)    return ENOMEM;

  header=NULL;

  while ((header = (struct tar_header*)get_next_entry(header)) != NULL) {
    memset(pname,'\0',256);
    get_entry_name(header,pname);
    if(!fcount) {
      if((fdirent=malloc(sizeof(*fdirent)))==NULL) 
	return ENOMEM;
      strcpy(fdirent->d_name,"..");
      fdirent->d_namlen=2;
      fdirent->d_rec_len=sizeof(*fdirent);
      fdirent->d_ino=0;
      fdirent->d_off=0;
      ifsnode->dentries[fcount]=fdirent;
      fcount++;
    }
    if(strcmp(ifsnode->head.name,"/")) /* check for root */ {
      plen=strlen(tpath);
      if(!strncmp(tpath,pname,plen)) { /* can be our entries */
	plen=strlen(pname); ra=0;
	op[0]=op[1]=0;
	for(i=0;i<plen;i++) {
	  if(pname[i]=='/') {
	    op[0]=op[1]; /* save old value */
	    op[1]=i;
	    ra++;
	  }
	}
	if(ra==pp) {
	  if((fdirent=malloc(sizeof(*fdirent)))==NULL) 	  return ENOMEM;
	  strcpy(fdirent->d_name,pname+op[1]+1);
	  goto __adddirentry;
	} else if((ra==pp+1) && ((pname[plen]=='\0' && pname[plen-1]=='/'))) {
	  if((fdirent=malloc(sizeof(*fdirent)))==NULL) 	  return ENOMEM;
	  strcpy(fdirent->d_name,pname+op[0]+1);
	  fdirent->d_name[strlen(fdirent->d_name)-1]='\0'; /* cut off trailing separator */
	__adddirentry:
	  fdirent->d_namlen=strlen(fdirent->d_name);
	  fdirent->d_rec_len=sizeof(*fdirent);
	  fdirent->d_ino=0;
	  fdirent->d_off=0;
	  if(strlen(fdirent->d_name)==0) {
	    fdirent->d_namlen=1;
	    fdirent->d_name[0]='.';
	    fdirent->d_name[1]='\0';
	  }
	  ifsnode->dentries[fcount]=fdirent;
	  fcount++;
	  if(fcount>sick_val) {
	    ndirs=malloc(sizeof(uintptr_t)*(sick_val+100));
	    memcpy(ndirs,ifsnode->dentries,sizeof(uintptr_t)*sick_val);
	    ifsnode->dentries=realloc(ifsnode->dentries,sizeof(uintptr_t)*(sick_val+100));
	    memcpy(ifsnode->dentries,ndirs,sizeof(uintptr_t)*sick_val);
	    sick_val+=100;
	    free(ndirs);
	  }
	}
      }
      ifsnode->dir_entries=fcount;
    } else { /* it's a root and not usual entry reads */
      if(fcount==1) {
	if((fdirent=malloc(sizeof(*fdirent)))==NULL) 
	  return ENOMEM;
	fdirent->d_name[0]='.';
	fdirent->d_name[1]='\0';
	fdirent->d_namlen=1;
	fdirent->d_rec_len=sizeof(*fdirent);
	fdirent->d_ino=0;
	fdirent->d_off=0;
	ifsnode->dentries[fcount]=fdirent;
	fcount++;
      }
      plen=strlen(pname); ra=0;
      for(i=0;i<plen;i++) {
	if(pname[i]=='/') ra++;
      }
      if(!ra) { /* found ! */
	if((fdirent=malloc(sizeof(*fdirent)))==NULL) 	  return ENOMEM;
	strcpy(fdirent->d_name,pname);
	goto __addrootentry;
      } else if((pname[plen]=='\0' && pname[plen-1]=='/') && ra==1) { /* taken */
	if((fdirent=malloc(sizeof(*fdirent)))==NULL) 	  return ENOMEM;
	strcpy(fdirent->d_name,pname);
	fdirent->d_name[plen-1]='\0'; /* cut off trailing separator */
      __addrootentry:
	fdirent->d_namlen=strlen(fdirent->d_name);
	fdirent->d_rec_len=sizeof(*fdirent);
	fdirent->d_ino=0;
	fdirent->d_off=0;
	ifsnode->dentries[fcount]=fdirent;
	fcount++;
	if(fcount>sick_val) {
	  ndirs=malloc(sizeof(uintptr_t)*(sick_val+100));
	  memcpy(ndirs,ifsnode->dentries,sizeof(uintptr_t)*sick_val);
	  ifsnode->dentries=realloc(ifsnode->dentries,sizeof(uintptr_t)*(sick_val+100));
	  memcpy(ifsnode->dentries,ndirs,sizeof(uintptr_t)*sick_val);
	  sick_val+=100;
	  free(ndirs);
	}
      }
      ifsnode->dir_entries=fcount;
    }
  }

  ifsnode->dentries[fcount]=NULL;

  return 0;
}

int initfs_close(vnode_t *node)
{
  struct initfs_node_t *ifsnode;
  int i;

  ifsnode=(struct initfs_node_t*)node->spec_data;
  if(!ifsnode->data) {
    for(i=0;i<ifsnode->dir_entries-1;i++) {
      free(ifsnode->dentries[i]);
    }
  }

  free(ifsnode);

  return 0;
}

int initfs_readdir(vnode_t *node,size_t size,off_t off,struct dirent *dentry)
{
  struct initfs_node_t *ifsnode=(struct initfs_node_t *)node->spec_data;

  if(off>ifsnode->dir_entries) {
    errno=EOVERFLOW;
    return -1;
  } else if(ifsnode->dentries[off]==NULL) {
    return DIRENT_EL;
  } else {
    strcpy(dentry->d_name,ifsnode->dentries[off]->d_name);
    dentry->d_ino=ifsnode->dentries[off]->d_ino;
    dentry->d_off=ifsnode->dentries[off]->d_off;
    dentry->d_namlen=ifsnode->dentries[off]->d_namlen;
    dentry->d_rec_len=ifsnode->dentries[off]->d_rec_len;
    return 0;
  }

  return 0;
}

/* local functions implementation */
static int handle_ramdisk(uintptr_t *start,uintptr_t *size)
{
  kcontrol_args_t ka;
  int n1[]={KCTRL_KERNEL,KCTRL_BOOT_INFO,KCTRL_INITRD_START_PAGE};
  int n2[]={KCTRL_KERNEL,KCTRL_BOOT_INFO,KCTRL_INITRD_SIZE};
  long rd_start,rd_size;
  int r;
 
  memset(&ka,0,sizeof(ka));
  ka.name_len=3;
  ka.name=n1;
  ka.old_data=&rd_start;
 
  r=kcontrol(&ka);
  if( r ) {
    return r;
  }
 
  ka.name=n2;
  ka.old_data=&rd_size;
  r=kcontrol(&ka);
  if( r ) {
    return r;
  }
 
  if(!rd_size)
    return ENOENT;

  *start=rd_start << 12;
  *size=rd_size << 12;

  return 0;
}

static size_t oct2int(const char *str, size_t size)
{
  size_t i;
  size_t val, m;
  
  if ((str == NULL) || (size == 0))
    return 0;
  
  for (i = 0; i < size; i++) {
    if (str[i] == '\0')
      break;
    if ((str[i] < '0') || (str[i] > '7'))
      return 0;
  }
  size = i;
  
  val = 0;
  for (i = size, m = 1; i > 0; i--, m *= 8)
    val += (str[i-1] - '0') * m;
  
  return val;
}

static const void* get_next_entry(const void *pos)
{
  size_t size, n;
  const struct tar_header *header;
    
  if (pos == NULL)
    /* first call, take the first entry */
    header = super_head.start;
  else {
    header = pos;
    size = oct2int(header->size, 12);
    n = size / BLOCK_SIZE;
    if (size % BLOCK_SIZE != 0)
      n++;
    n++;
    header = (const struct tar_header*) (((const char*) pos) + n * BLOCK_SIZE);
  }
  
  if (strlen(header->name) == 0)
    return NULL;
  
  return header;
}

static size_t get_entry_name(const void *hdr,char *name)
{
  const struct tar_header *header;
  size_t len;

  if (hdr == NULL) {
    errno = EINVAL;
    return 0;
  }
  header = hdr;

  /* If the first character of prefix is '\0', the file name is
   * tar_header.name; othrewise, it is tar_header.prefix/tar_header.name.
   * Files whose pathname don't fir in that length can not be stored in a tar
   * archive.
   */

  len = strlen(header->name);
  if (len == 0)
    return len;
  if (header->prefix[0] != '\0')
    len += strlen(header->prefix) + 1;
  len++;

  if (header->prefix[0] != '\0') {
    strcpy(name, header->prefix);
    strcat(name, "/");
  }
  strcat(name, header->name);

  return len;
}
