#include <stdlib.h>
#include <errno.h>

#include "log.h"
#include "block_cache.h"

static inline bcache_entry_t *__block_cache_alloc_entry(void)
{
  bcache_entry_t *bentry;

  bentry = malloc(sizeof(*bentry));
  if (!bentry)
    errno = ENOMEM;
  else
    bentry->buffer = NULL;

  return bentry;
}

static inline void __block_cache_dealloc_entry(bcache_entry_t *bentry)
{
  if (bentry->buffer)
    free(bentry->buffer);

  free(bentry);
}

static inline void __block_cache_clear_entry(bcache_entry_t *bentry)
{
  bentry->block_nbr = -1;
  bentry->refcount  = 0;
  bentry->type      = BLOCK_NEW;
  bentry->dirty     = 0;
}

static inline int __block_cache_init_entry(bcache_entry_t *bentry,
                                            unsigned int buffer_size)
{
  if (!bentry)
    return -1;

  bentry->buffer = malloc(buffer_size);
  if (!bentry->buffer)
    return ENOMEM;

  __block_cache_clear_entry(bentry);

  return 0;
}

static bcache_entry_t *__block_cache_find_block(bcache_t *bcache,
                                                 unsigned int block_nbr,
                                                 bentry_type_t type)
{
  bcache_entry_t *bentry;
  list_head_t *block_list;

  if (type == BLOCK_USED)
    block_list = &bcache->used_blocks;
  else if (type == BLOCK_UNUSED)
    block_list = &bcache->unused_blocks;

  list_for_each_entry(block_list, bentry, bentry_node) {
    if (bentry->block_nbr == block_nbr)
      return bentry;
  }

  return NULL;
}

static inline void __block_cache_drop_list(bcache_t *bcache, bentry_type_t type)
{
  list_node_t *node, *save;
  bcache_entry_t *bentry;
  list_head_t *list;

  if (type == BLOCK_USED)
    list = &bcache->used_blocks;
  else if (type == BLOCK_UNUSED)
    list = &bcache->unused_blocks;

  list_for_each_safe(list, node, save) {
    bentry = container_of(node, bcache_entry_t, bentry_node);
    __block_cache_dealloc_entry(bentry);
  }
}

static void __block_cache_pop_lru(bcache_t *bcache)
{
  bcache_entry_t *bentry, *bentry_old;

  if (bcache->unused == bcache->unused_watermark) {
    bentry_old = container_of(list_node_first(&bcache->unused_blocks),
                              bcache_entry_t, bentry_node);
    list_delfromhead(&bcache->unused_blocks);
    __block_cache_dealloc_entry(bentry_old);
  } else
    bcache->unused ++;

  bentry = container_of(list_node_first(&bcache->used_blocks),
                        bcache_entry_t, bentry_node);
  list_delfromhead(&bcache->used_blocks);
  list_add2tail(&bcache->unused_blocks, &bentry->bentry_node);
}

static void __block_cache_push_ru(bcache_t *bcache, bcache_entry_t *bentry)
{
  list_del(&bentry->bentry_node);

  bcache->unused --;

  if (bcache->used == bcache->used_watermark)
    __block_cache_pop_lru(bcache);
  else
    bcache->used ++;

  list_add2tail(&bcache->used_blocks, &bentry->bentry_node);
}

void block_cache_flush(bcache_t *bcache)
{
  __block_cache_drop_list(bcache, BLOCK_USED);
  __block_cache_drop_list(bcache, BLOCK_UNUSED);
}

void block_cache_init(bcache_t *bcache,
                      unsigned int block_size,
                      unsigned int used_watermark,
                      unsigned int unused_watermark)
{
  list_init_head(&bcache->used_blocks);
  list_init_head(&bcache->unused_blocks);

  bcache->block_size       = block_size;

  bcache->used_watermark   = used_watermark;
  bcache->unused_watermark = unused_watermark;
  bcache->block_nbr_last   = -1;
  bcache->used             = 0;
  bcache->unused           = 0;
}

void *block_cache_add_block(bcache_t *bcache,
                             unsigned int block_nbr)
{
  bcache_entry_t *bentry;

  bentry = __block_cache_alloc_entry();
  if (!bentry) {
    errno = ENOMEM;
    return NULL;
  }
  if (__block_cache_init_entry(bentry, bcache->block_size)) {
    free(bentry);
    errno = ENOMEM;
    return NULL;
  }

  bentry->block_nbr = block_nbr;

  if (bcache->used == bcache->used_watermark)
    __block_cache_pop_lru(bcache);
  else
    bcache->used ++;

  list_add2tail(&bcache->used_blocks, &bentry->bentry_node);
  bcache->block_nbr_last = block_nbr;

  return bentry->buffer;
}

bcache_entry_t *block_cache_get_entry(bcache_t *bcache, unsigned int block_nbr)
{
  bcache_entry_t *bentry;

  bentry = __block_cache_find_block(bcache, block_nbr, BLOCK_USED);
  if (!bentry) {
    bentry = __block_cache_find_block(bcache, block_nbr, BLOCK_UNUSED);
    if (!bentry) {
      /* Nope, block was not found in cache */
      return NULL;
    }

    /* Block was unused some time ago */
    __block_cache_push_ru(bcache, bentry);
  }

  /* Block was *recently* used */
  bentry->refcount ++;
  bcache->block_nbr_last = block_nbr;

  list_del(&bentry->bentry_node);
  list_add2tail(&bcache->used_blocks, &bentry->bentry_node);

  return bentry;
}
