#include <sys/types.h>
#include <limits.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <stddef.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <pthread.h>
#include <errno.h>
#include <time.h>

#include <vfs/vfsdef.h>
#include <vfs/super.h>
#include <vfs/vnode.h>
#include <sys/v2.h>

#include "ext2fs.h"
#include "ext2_fs.h"
#include "ext2fsP.h"
#include "v2ext2.h"

/* TODO: I'm tierd reimplementing all this.. And again. D'oh! */

ext2_filsys fs;

static int __cmp_dentry(const void *a, const void *b)
{
  ino_t a_ino = ((struct dirent *)a)->d_ino;
  ino_t b_ino = ((struct dirent *)b)->d_ino;

  return a_ino - b_ino;
}

int ext2_mount(vnode_t *node, char *path, int *pps, int pp, int flags)
{
  ext2_debug("Called.");

  errcode_t ret;

  ret = ext2fs_init_default(path, EXT2_FLAG_RW | EXT2_FLAG_JOURNAL_DEV_OK,
			    0, 0, unix_io_manager, &fs);
  if (ret)
    ext2_error("(initializing defaults)",
               "Error while initializing operations. (ret=%d)\n", ret);

  ret = ext2fs_read_inode_bitmap(fs);
  if (ret)
    ext2_error("(reading inode bitmap)",
               "Error while reading inode bitmap. (ret=%d)\n", ret);

  //  ext2fs_pprint_bitmap("Inode", fs->inode_map->bitmap, 16, 0, 2);

  ret = ext2fs_read_block_bitmap(fs);
  if (ret)
    ext2_error("(reading bitmap)", "Can't read block bitmap.\n");

  //  ext2fs_pprint_bitmap("Block", fs->block_map->bitmap,
  //		       (((fs->block_map->real_end -
  //			  fs->block_map->start) / 8) + 1), 800, 200);

  node->super->spec_data  = fs->super;
  node->super->block_size = fs->blocksize;
  node->super->slice_size = fs->super->s_blocks_count * fs->blocksize;
  node->super->flags      = node->super->flags | flags;
  node->super->backend    = NULL;
  node->super->pagecache  = NULL;

  ext2_debug("Mount request for device '%s' succeded.", fs->device_name);
  ext2_debug("FS blocksize = %d; FS blockcount = %d",
             fs->blocksize, fs->super->s_blocks_count);

  return 0;
}

static int do_lookup(ext2_ino_t parent, const char *name,
		     ext2_ino_t *ino, struct ext2_inode *inode)
{
  errcode_t rc;

  rc = ext2fs_lookup(fs, parent, name, strlen(name), NULL, ino);
  if(rc) {
    ext2_debug("do_lookup()", "while looking up '%s'\n", name);
    return ENOENT;
  }

  rc = ext2fs_read_inode(fs, *ino, inode);
  if(rc) {
    ext2_debug("do_lookup()", "while reading inode %u\n", *ino);
    return EIO;
  }

  return 0;
}

int ext2_lookup(vnode_t *node, char *path, int *pps, int pp)
{
  ext2_debug("Called.");

  char ppart[PATH_MAX];
  icache_entry_t *ientry;
  struct ext2_inode inode;
  ext2_ino_t parent_ino, ino;
  int rc, i;

  if (strlen(path) == 1 && *path == '/') {
    rc = ext2fs_read_inode(fs, EXT2_ROOT_INO, &inode);
    ino = EXT2_ROOT_INO;
    if (rc) return rc;
  } else {
    parent_ino = EXT2_ROOT_INO;

    for (i = 0; i < pp; i++) {
      memset(ppart, '\0', PATH_MAX);
      if ((i+1) >= pp)
	strncpy(ppart, path + pps[i] + 1, strlen(path) - pps[i]);
      else
	strncpy(ppart, path + pps[i] + 1, pps[i + 1] - pps[i] - 1);

      ext2_debug("Path part = '%s'", ppart);
      rc = do_lookup(parent_ino, ppart, &ino, &inode);
      if (rc) {
        errno = rc;
	return -1;
      }
      parent_ino = ino;
    }
  }

  node->vstat.st_rdev    = 0;
  node->vstat.st_ino     = ino;
  node->vstat.st_mode    = inode.i_mode;
  node->vstat.st_blksize = node->super->block_size;
  node->vstat.st_blocks  = inode.i_size / node->vstat.st_blksize;
  node->vstat.st_uid     = inode.i_uid;
  node->vstat.st_gid     = inode.i_gid;
  node->vstat.st_size    = inode.i_size;
  node->vstat.st_nlink   = inode.i_links_count;
  node->vstat.st_mtime   = inode.i_mtime;
  node->vstat.st_atime   = inode.i_atime;
  node->vstat.st_ctime   = inode.i_ctime;

  ientry = icache_get_inode(ino, fs->inode_cache);
  if (!ientry) {
    ientry = icache_inode_alloc();
    if (!ientry)
      ext2_error("Mount stage", "Out of memory!\n");
    icache_fill_inode(ino, &inode, sizeof(struct ext2_inode),
                      ientry, fs->inode_cache);
  }

  node->spec_data = ientry;

  return 0;
}

int ext2_opendir(vnode_t *node, char *path, int *pps, int pp)
{
  ext2_debug("Called.");

  icache_entry_t *ientry = (icache_entry_t *)node->spec_data;
  struct ext2_inode inode;
  time_t now = time(NULL);
  void *buf, *p, *end;
  int block_n, ret = 0, records_n = 0;
  struct ext2_dir_entry *ext2_dirent;
  unsigned int name_len, rec_len;
  off_t off_in = 0;
  usrtc_node_t *list_node;

  if ((ret = ext2fs_read_inode(fs, ientry->ino, &inode)))
    ext2_error("ext2_opendir()", "Can't read inode %d, ret=%d\n",
               ientry->ino, ret);

  if (!S_ISDIR(inode.i_mode))
    ext2_error("Paranoid mode ON",
               "Passed inode is not a directory! Can't be.\n");

  node->vstat.st_atime = inode.i_atime = now;

  /* Already cached */
  if (ientry->dentries != NULL)
    return 0;

  ientry->dentries = malloc(sizeof(*ientry->dentries));
  usrtc_init(ientry->dentries, USRTC_LIST, USRTC_COUNT_T_MAX, __cmp_dentry);

  buf = malloc(fs->blocksize);
  if (!buf) {
    errno = ENOMEM;
    return -1;
  }

  p = (char *)buf;
  end = (char *) buf + fs->blocksize;

  for (block_n = 0; block_n < EXT2_NDIR_BLOCKS; block_n ++) {
    if (inode.i_block[block_n] == 0)
      continue;

    ext2_debug("Reading block %d of directory block %d.",
               inode.i_block[block_n], block_n);
    ret = io_channel_read_blk(fs->io, inode.i_block[block_n], 1, buf);
    if (ret) goto __opendir_fail;

    while (p < end-8) {
      ext2_dirent = malloc(sizeof(*ext2_dirent));
      memcpy(ext2_dirent, (struct ext2_dir_entry *) p, sizeof(*ext2_dirent));

      name_len = ext2_dirent->name_len;
      rec_len = (ext2_dirent->rec_len || fs->blocksize < 65536) ?
        ext2_dirent->rec_len : 65536;
      if ((rec_len < 8) || (rec_len % 4) ||
          ((name_len & 0xFF) + 8) > rec_len) {
        rec_len = 8;
        ext2_error("ext2_readdir()", "Directory block corrupted.\n");
        ret = EXT2_ET_DIR_CORRUPTED;
        goto __opendir_fail;
      }

        p += rec_len;
        off_in += rec_len;

        ext2_debug("Dentry found.\nName: '%s'\nRec len: %d\n"
                   "Offset: %d\nName len: %d\nInode: %d",
                   ext2_dirent->name, ext2_dirent->rec_len,
                   off_in, ext2_dirent->name_len, ext2_dirent->inode);

        list_node = malloc(sizeof(*list_node));
        usrtc_node_init(list_node, ext2_dirent);
        usrtc_insert(ientry->dentries, list_node, ext2_dirent);
        records_n ++;
    }
  }

  free(buf);
  ext2_debug("Directory children number: %d\n", records_n);
  ientry->dentries_n = records_n;

  return 0;

 __opendir_fail:

  free(buf);
  ext2_error("ext2_opendir()", "Failed opening directory. Retcode = %d.\n", ret);

  return ret;
}

int ext2_readdir(vnode_t *node, size_t size, off_t off, struct dirent *dentry)
{
  ext2_debug("Called. Offset = %d", off);

  icache_entry_t *ientry = (icache_entry_t *)node->spec_data;
  usrtc_t *dentries = ientry->dentries;
  usrtc_node_t *list_node;
  struct ext2_dir_entry *ext2_dirent = NULL;
  int n;
  off_t off_in = 0;

  if (off > ientry->dentries_n) {
    ext2_error("ext2_readdir()", "Overflow!!!!");
    errno = EOVERFLOW;
    return -1;
  } else if (off == ientry->dentries_n)
    return DIRENT_EL;

  for (n = 0, list_node = usrtc_first(dentries); list_node != NULL;
       list_node = usrtc_next(dentries, list_node), n ++) {
    ext2_dirent = usrtc_node_getdata(list_node);
    if (n == off)
      break;
    off_in += ext2_dirent->rec_len;
  }

  ext2_debug("!!!Dentry found.\nName: '%s'\nRec len: %d\n"
             "!!!Offset: %d\nName len: %d\nInode: %d",
             ext2_dirent->name, ext2_dirent->rec_len,
             off_in, ext2_dirent->name_len, ext2_dirent->inode);

  strcpy(dentry->d_name, ext2_dirent->name);
  dentry->d_ino     = ext2_dirent->inode;
  dentry->d_off     = off_in;
  dentry->d_namlen  = ext2_dirent->name_len;
  dentry->d_rec_len = ext2_dirent->rec_len;

  return 0;
}

int ext2_createdir(vnode_t *node,char *path,int *pps,
                   int pp,int mode,uid_t uid,gid_t gid)
{
  ext2_debug("Broken.");
  return 0;
}

void *ext2_read(vnode_t *node,void *buf,size_t *size,off_t offset)
{
  ext2_debug("Called. Offset = %lu. Size: %lu", offset, *size);

  int ret;
  size_t sz = *size;
  uint32_t page_n;
  off_t page_off;
  icache_entry_t *ientry = (icache_entry_t *)node->spec_data;
  struct ext2_inode inode;
  void *bbuf;

  if (!ientry || sz == 0 || S_ISDIR(node->vstat.st_mode))
    return NULL;

  page_n   = PAGE_NUMBER(fs,offset);
  page_off = PAGE_OFFSET(fs,offset);

  ext2_debug("Page number %d with offset %d.", page_n, page_off);

  if(page_off + sz > fs->blocksize)
    sz = fs->blocksize - page_off;
  ret = ext2fs_read_inode(fs, ientry->ino, &inode);
  if (ret) {
    abort();
  }

  bbuf = ext2fs_get_mem(fs->blocksize * 3, 1);
  memset(bbuf, '\0', fs->blocksize * 3);
  ext2_debug("Reading block (%d)...", inode.i_block[page_n]);
  //  ret = io_channel_read_blk(fs->io, inode.i_block[page_n], 1, bbuf);
  ret = ext2fs_read_ind_block(fs, inode.i_block[page_n], bbuf);
  if (ret) {
    ext2_debug("Check!!!");
    abort();
  }

  bbuf = (void *)(bbuf + page_off);
  *size = sz;

  return bbuf;
}

int ext2_write(vnode_t *node,void *buf,size_t *size,off_t offset)
{
  ext2_debug("Called. Offset = %lu. Size: %lu", offset, *size);

  int ret;
  size_t nsz = *size, rsz = *size;
  uint32_t page_n;
  off_t page_off;
  icache_entry_t *ientry = (icache_entry_t *)node->spec_data;
  struct ext2_inode inode;
  void *bbuf;

  page_n   = PAGE_NUMBER(fs,offset);
  page_off = PAGE_OFFSET(fs,offset);

  if((page_off + nsz) > fs->blocksize) {
    nsz = fs->blocksize - page_off;
  }

  /* Do write */

  nsz = rsz - nsz;

  ret = ext2fs_read_inode(fs, ientry->ino, &inode);
  if (ret) {
    abort();
  }

  ext2fs_write_ind_block(fs, inode.i_block[0], buf);

  return 0;
}
