#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "log.h"
#include "inode_cache.h"

static icache_entry_t *__icache_hash_search(ino_t ino,
                                            const icache_t *icache)
{
  int hashval = icache->hasher(ino);
  list_head_t *hash_list = &icache->hash[hashval];
  icache_entry_t *entry;

  list_for_each_entry(hash_list, entry, inode_hash) {
    if (ino == entry->ino)
      return entry;
  }
  return NULL;
}

static list_node_t *__icache_hash_add(icache_entry_t *entry,
                                      icache_t *icache)
{
  int hashval = icache->hasher(entry->ino);

  list_add2tail(&icache->hash[hashval], &entry->inode_hash);
  return &entry->inode_hash;
}

static void __icache_dealloc_inode(icache_entry_t *entry)
{
  list_del(&entry->inode_list);
  list_del(&entry->inode_hash);
  free(entry);
}

uint32_t icache_hash_inode(ino_t key)
{
  return hash_long(key, INODE_HASH_BITS);
}

int icache_init(icache_t *icache,
                unsigned int hashbits,
                uint32_t (*hasher)(ino_t ino))
{
  int i;

  icache->hashsize = 1 << hashbits;
  icache->hashbits = hashbits;

  icache->hash = malloc(sizeof(*icache->hash) * icache->hashsize);
  if (!icache->hash)
    return ENOMEM;

  for (i = 0; i < icache->hashsize; i++)
    list_init_head(&icache->hash[i]);

  list_init_head(&icache->used_list);
  list_init_head(&icache->unused_list);
  list_init_head(&icache->dirty_list);

  icache->n_unused = 0;
  icache->n_used   = 0;
  icache->n_dirty  = 0;

  icache->max_ino   = 0;
  icache->min_ino   = 0;
  icache->n_entries = 0;

  icache->hasher = hasher ? hasher : &icache_hash_inode;

  return 0;
}

icache_entry_t *icache_inode_alloc(void)
{
  icache_entry_t *entry;

  entry = calloc(1, sizeof(*entry));
  if (!entry)
    return NULL;

  entry->state    = CACHE_INODE_NEW;
  entry->dentries = NULL;

  return entry;
}

void icache_drop(icache_t *icache)
{
  // TODO:
}

void icache_inode_move(icache_entry_t *entry, icache_t *icache,
                       icache_list_t list_type)
{
  list_head_t *list;

  switch (list_type) {
  case ICACHE_LIST_UNUSED:
    list = &icache->unused_list;
    break;
  case ICACHE_LIST_USED:
    list = &icache->used_list;
    break;
  case ICACHE_LIST_DIRTY:
    list = &icache->dirty_list;
    break;
  }

  entry->list_type = list_type;
  list_del(&entry->inode_list);
  list_add2tail(list, &entry->inode_list);
}

icache_entry_t *icache_get_inode(ino_t ino, icache_t *icache)
{
  icache_entry_t *entry;

  entry = __icache_hash_search(ino, icache);
  if (!entry)
    return NULL;

  entry->counter ++;
  if (entry->list_type == ICACHE_LIST_UNUSED) {
    icache_inode_move(entry, icache, ICACHE_LIST_USED);
  }

  ext2_debug("Inode (%d) used (%d) times.", ino, entry->counter);

  return entry;
}

icache_entry_t *icache_fill_inode(ino_t ino, void *fs_data, size_t sz,
                                  icache_entry_t *entry, icache_t *icache)
{
  entry->list_type = ICACHE_LIST_USED;
  entry->ino       = ino;
  entry->counter   = 1;

  entry->fs_data = malloc(sz);
  if (!entry->fs_data) {
    errno = ENOMEM;
    return NULL;
  }

  memcpy(entry->fs_data, fs_data, sz);

  __icache_hash_add(entry, icache);
  list_add2tail(&icache->used_list, &entry->inode_list);

  ++ icache->n_entries;
  ++ icache->n_used;

  entry->state &= ~CACHE_INODE_NEW;

  return entry;
}

void icache_del_inode(icache_entry_t *entry, icache_t *icache)
{
  if (icache->inode_dealloc)
    icache->inode_dealloc(entry);

  __icache_dealloc_inode(entry);

  -- icache->n_entries;
}

void icache_unuse_inode(icache_entry_t *entry, icache_t *icache)
{
  -- entry->counter;

  if (!entry->counter)
    icache_inode_move(entry, icache, ICACHE_LIST_UNUSED);
}

void icache_dump_unused(icache_t *icache)
{
  list_node_t *node;
  list_node_t *save;
  icache_entry_t *entry;

  list_for_each_safe(&icache->unused_list, node, save) {
    entry = container_of(node, icache_entry_t, inode_list);
    icache_del_inode(entry, icache);
  }
}
