#ifndef ICACHE_H
#define ICACHE_H

#include <sys/types.h>
#include <stdint.h>
#include <hash.h>
#include <math.h>
#include <datalib/list.h>
#include <datalib/usrtc.h>
#include <time.h>

#define INODE_HASH_BITS    16

#define CACHE_INODE_NEW   1 << 0
#define CACHE_INODE_DIRTY 1 << 1

typedef enum {
  ICACHE_LIST_UNUSED = 1 << 0,
  ICACHE_LIST_USED   = 1 << 1,
  ICACHE_LIST_DIRTY  = 1 << 2
} icache_list_t;

typedef struct __icache_entry icache_entry_t;
struct __icache_entry
{
  ino_t ino;
  /* Number of uses of inode */
  uint32_t counter;
  list_node_t inode_hash;
  list_node_t inode_list;
  list_head_t dentry_list;
  icache_list_t list_type;
  int state;

  /* Meta data */
  unsigned int i_mode;
  uid_t i_uid;
  gid_t i_gid;
  unsigned int i_nlink;
  long long i_size;
  timeval_t i_atime;
  timeval_t i_mtime;
  timeval_t i_ctime;
  blkcnt_t i_blocks;
  uint32_t i_generation;
  unsigned int i_flags;

  usrtc_t *dentries;
  uint32_t dentries_n;

  void *fs_data;
};

typedef struct __icache icache_t;
struct __icache
{
  uint32_t hashsize;
  uint32_t hashbits;
  list_head_t *hash;

  list_head_t used_list;
  list_head_t unused_list;
  list_head_t dirty_list;

  uint32_t n_unused;
  uint32_t n_used;
  uint32_t n_dirty;

  /* Statistics related fields */
  ino_t max_ino;
  ino_t min_ino;
  uint32_t n_entries;

  /* Lock all cache elements */
  pthread_mutex_t glock;

  /* User deined callbacks */
  uint32_t (*hasher)(ino_t ino);
  void (*inode_dealloc)(icache_entry_t *cinode);
};

uint32_t icache_hash_inode(ino_t key);
int icache_init(icache_t *icache, unsigned int hashbits, uint32_t (*hasher)(ino_t ino));
void icache_drop(icache_t *icache);
void icache_inode_move(icache_entry_t *entry, icache_t *icache, icache_list_t list_type);
icache_entry_t *icache_get_inode(ino_t ino, icache_t *icache);
icache_entry_t *icache_inode_alloc(void);
icache_entry_t *icache_fill_inode(ino_t ino, void *fs_data, size_t sz, icache_entry_t *entry, icache_t *icache);
void icache_del_inode(icache_entry_t *entry, icache_t *icache);
void icache_unuse_inode(icache_entry_t *entry, icache_t *icache);
void icache_dump_unused(icache_t *icache);

#define ICACHE_INODE_CREATFILL(ino,data,cache)          \
  do {                                                  \
    icache_entry_t *entry = icache_inode_alloc();       \
    if (!entry) abort();                                \
    icache_fill_inode(ino,data,entry,cache);            \
  } while(0)

#endif
