/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 *
 * ISO 9660 CDROM file system
 * main entry
 */

#include <stdint.h>
#include <sys/types.h>
#include <errno.h>

#include <sys/v2.h>

#include <vfs/vfsdef.h>
#include <vfs/super.h>
#include <vfs/vnode.h>

#ifdef CONFIG_ISOFS_DEBUG
#include <dircons.h>
#endif

#include "isofs.h"

static vfs_info_t iso_info = {
  .fstype = "iso9660",
  .author = "Dmitry Balakshin",
  .info   = "ISO 9660 CDROM file system"
};

static struct fs_funct iso_ops = {
  .mount     = iso_mount,
  .umount    = iso_umount,
  .lookup    = iso_lookup,
  .read      = NULL,
  .ioctl     = NULL,
  .write     = NULL,
  .opendir   = NULL,
  .close     = NULL,
  .create    = NULL,
  .setattr   = NULL,
  .getattr   = NULL,
  .createdir = NULL,
  .unlink    = NULL,
  .unlinkdir = NULL,
  .link      = NULL,
  .symlink   = NULL,
  .readdir   = NULL,
  .truncate  = NULL,
  .mmap      = NULL,
  .lockf     = NULL,
  .readlink  = NULL,
};

static super_block_t iso_super = {
  .info       = &iso_info,
  .block_size = 0x2bad,			/* FIXME: this number is stub only */
  .slice_size = 0,
  .spec_data  = NULL,
  .flags      = V2FS_RO,
  .fs_tbl     = &iso_ops,
  .path       = NULL,
  .pps        = NULL,
  .pp         = 0,
  .backend    = NULL,
  .pagecache  = NULL
};

static filesystem_proc_t iso_proc = {
  .super       = &iso_super,
  .listen_port = 0,
  .f_pid       = 0,
  .listeners   = 8
};

int main(int argc, char *argv[])
{
  int err;
  prtid_t port;

#ifndef CONFIG_ISOFS_GRUB
  start_service();
#endif

#ifdef CONFIG_ISOFS_DEBUG
  open_console();
#endif

  port = ipc_port_open(IPC_BLOCKED_ACCESS, 512);
  if (port < 0) {
    err = -port;
#ifdef CONFIG_ISOFS_DEBUG
    console_printf("[iso.fs] ERROR: failed to open port: %d\n", err);
#endif
    return 1;
  }

  iso_proc.listen_port = port;

  err = vfs_register(port, iso_super.flags, &iso_info);
  if (err != 0) {
    err = errno;
    ipc_port_close(port);
#ifdef CONFIG_ISOFS_DEBUG
    console_printf("[iso.fs] ERROR: failed to register: %d\n", err);
#endif
    return 1;
  }
#ifdef CONFIG_ISOFS_DEBUG
  console_printf("[iso.fs] Registered on port %d\n", port);
#endif

  err = fsproc_init(&iso_proc);
  if (err != 0) {
    /* TODO: call vfs_unregister() or similar */
    ipc_port_close(port);
#ifdef CONFIG_ISOFS_DEBUG
    console_printf("[iso.fs] ERROR: mounting failed: %d\n"
		     "[iso.fs] Exting...", err);
#endif
    return 1;
  }

  v2fs_thread(&iso_proc);

  return 0;
}
