/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS temp file system,
 * internal implementation 
 */

#include <sys/types.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <pthread.h>
#include <unistd.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <sys/v2.h>
#include <sys/kcontrol.h>
#include <sys/mman.h>
#include <ns/ns.h>
#include <errno.h>
#include <datalib/usrtc.h>

#include <vfs/vfsdef.h>
#include <vfs/super.h>
#include <vfs/vnode.h>

#include "tmpfs.h"

#ifdef TMPFS_DEBUG
#include <dircons.h>
#endif

/* local variables */
static tmpfs_super_t super_head;

/* local functions */
static inline void __entries_clear(struct dirent **ent);
static int __tmpfscmp_hash(const void *a,const void *b);

int tmpfs_mount(vnode_t *node,char *path,int *pps,int pp,int flags)
{
  tmpfs_node_t *vroot,*froot,*sroot;
  usrtc_node_t *fparent,*sparent;
  time_t now;

  /* init faked parent node */
  if((fparent=malloc(sizeof(*fparent)))==NULL)
    return ENOMEM;
  if((froot=malloc(sizeof(*froot)))==NULL)
    return ENOMEM;
  if((froot->name=strdup(".."))==NULL)
    return ENOMEM;
  froot->hash=vnode_hash_ch(froot->name);
  froot->type=DIRTYPE;
  froot->pages=NULL;
  froot->parent=NULL;
  /* init faked self */
  if((sparent=malloc(sizeof(*sparent)))==NULL)
    return ENOMEM;
  if((sroot=malloc(sizeof(*sroot)))==NULL)
    return ENOMEM;
  if((sroot->name=strdup("."))==NULL)
    return ENOMEM;
  sroot->hash=vnode_hash_ch(sroot->name);
  sroot->type=DIRTYPE;
  sroot->pages=NULL;
  sroot->parent=NULL;
  /* init tmpfs superblock */
  super_head.max_pages=512;
  super_head.alloc_pages=0;
  super_head.root=malloc(sizeof(*super_head.root));
  if(!super_head.root)
    return ENOMEM;
  /* push root */
  now=time(NULL);
  if(now==(time_t)-1) now=0; /* fucking deal */
  /* ok, let's init all */
  vroot=super_head.root;
  vroot->name=strdup("/");
  if(!vroot->name)
    return ENOMEM;
  vroot->hash=vnode_hash_ch(vroot->name);
  vroot->node=malloc(sizeof(*vroot->node));
  vroot->childs=malloc(sizeof(*vroot->childs));
  if(!vroot->node || !vroot->childs)
    return ENOMEM;
  /* init structure */
  usrtc_init(vroot->childs,USRTC_BST,USRTC_COUNT_T_MAX,__tmpfscmp_hash);
  /* insert self */
  usrtc_node_init(vroot->node,vroot);
  usrtc_node_init(sparent,sroot);
  usrtc_insert(vroot->childs,sparent,&sroot->hash);
  /* insert faked parent i.e. itself */
  usrtc_node_init(fparent,froot);
  usrtc_insert(vroot->childs,fparent,&froot->hash);
  /* it's a root dir and it hasn't parent */
  vroot->pages=NULL;
  vroot->parent=NULL;
  /* setup other values */
  vroot->owner=vroot->gowner=0;
  vroot->mask=0; vroot->mask|=(TFS_UREAD | TFS_UWRITE | TFS_UEXEC) | 
		   (TFS_GREAD | TFS_GWRITE | TFS_GEXEC) | 
		   (TFS_OREAD | TFS_OWRITE | TFS_OEXEC);
  vroot->type=DIRTYPE;
  vroot->entries=NULL;
  vroot->atime=vroot->ctime=vroot->mtime=now;
  vroot->num_links=0;

  /* lets init generic superblock */
  node->super->spec_data=&super_head;
  node->super->block_size=4096;
  node->super->slice_size=super_head.max_pages*node->super->block_size;
  node->super->flags|=flags;
  node->super->backend=NULL;
  node->super->pagecache=NULL;

  return 0;
}

int tmpfs_umount(vnode_t *node)
{
  return 0;
}

int tmpfs_lookup(vnode_t *node,char *path,int *pps,int pp)
{
  tmpfs_super_t *super;
  tmpfs_node_t *tfsnode,*tfsroot;
  usrtc_node_t *cnode;
  uint64_t srch_hash=0;
  char tname[256];
  int i=0;
  mode_t tfmode=0;

  if(strlen(path)==1 && path[0]=='/') { /* root directory */
    super=(tmpfs_super_t*)node->super->spec_data;
    if(!super) return ENOENT;
    tfsnode=super->root;
    if(!tfsnode) return ENOENT;
    /* fill out values */
    node->spec_data=tfsnode;
    /* fill out stat */
    node->vstat.st_dev=0;
    node->vstat.st_ino=0;
    node->vstat.st_mode=0;
    node->vstat.st_mode|=(00000777);
    node->vstat.st_mode|=S_IFDIR;
    node->vstat.st_uid=0;
    node->vstat.st_gid=0;
    node->vstat.st_rdev=0;
    node->vstat.st_blksize=node->super->block_size;
    node->vstat.st_size=usrtc_count(tfsnode->childs);
    node->vstat.st_blocks=0;
    node->vstat.st_nlink=0;
    node->vstat.st_mtime=tfsnode->mtime;
    node->vstat.st_atime=tfsnode->atime;
    node->vstat.st_ctime=tfsnode->ctime;
    return 0;
  } else {
    super=(tmpfs_super_t*)node->super->spec_data;
    if(!super) return ENOENT;
    tfsroot=super->root;
    if(!tfsroot) return ENOENT;
#if 0
#ifdef TMPFS_DEBUG
    console_printf("[tmpfs] catch! pp=%d,path=%s pps={",pp,path);
    for(i=0;i<pp;i++) 
      console_printf("%d;",pps[i]);
    console_printf("}\n");
#endif
#endif
    for(i=0;i<pp;i++) {
      /**
       * *NOTE:* lookup algorythm ->
       * take hash - lookup in root, if found -> if it end - fill up,
       * if not - make a new hash and change root -> if not found -> ENOENT anyway
       */
      memset(tname,'\0',256);

      if((i+1)>=pp) {
	strncpy(tname,path+pps[i]+1,strlen(path)-pps[i]);
      } else
	strncpy(tname,path+pps[i]+1,pps[i+1]-pps[i]-1);
#ifdef TMPFS_DEBUG
      console_printf("[tmpfs] {tmpfs_lookup} tname = %s\n",tname);
#endif
      /* make new hash */
      srch_hash=vnode_hash_ch(tname);
      /* NOTE: for flamefreaks - i've check it at all */
      cnode=usrtc_lookup(tfsroot->childs,&srch_hash);
      if(!cnode) 
	return ENOENT; /* ENOENT anyway */
      else {
	if((i+1)>=pp) { /* found ! */
	  tfsnode=usrtc_node_getdata(cnode);
	  /* fill out values */
	  node->spec_data=tfsnode;
	  /* fill out stat */
	  node->vstat.st_dev=0;
	  node->vstat.st_ino=i;
	  node->vstat.st_mode=0;
	  node->vstat.st_blocks=0;
	  node->vstat.st_size=0;
	  /* check it out */
	  if(tfsnode->mask & TFS_UREAD) tfmode|=S_IRUSR;
	  if(tfsnode->mask & TFS_UWRITE) tfmode|=S_IWUSR;
	  if(tfsnode->mask & TFS_UEXEC) tfmode|=S_IXUSR;
	  if(tfsnode->mask & TFS_GREAD) tfmode|=S_IRGRP;
	  if(tfsnode->mask & TFS_GWRITE) tfmode|=S_IWGRP;
	  if(tfsnode->mask & TFS_GEXEC) tfmode|=S_IXGRP;
	  if(tfsnode->mask & TFS_OREAD) tfmode|=S_IROTH;
	  if(tfsnode->mask & TFS_OWRITE) tfmode|=S_IWOTH;
	  if(tfsnode->mask & TFS_OEXEC) tfmode|=S_IXOTH;
	  /* set mode */
	  node->vstat.st_mode|=tfmode;
	  switch(tfsnode->type) {
	  case REGTYPE:     /* Regular file */
	    node->vstat.st_mode|=S_IFREG;
	    node->vstat.st_size=tfsnode->size;
	    node->vstat.st_blocks=tfsnode->size/node->super->block_size; /* TODO: normal calculation */
	    break;
	  case LNKTYPE:     /* Hard link */ /* FIXME: add normal links support */
	    node->vstat.st_mode|=S_IFLNK;
	    break;
	  case SYMTYPE:     /* Symbolic link */ /* FIXME: add normal links support */
	    node->vstat.st_mode|=S_IFLNK;
	    break;
	  case CHRTYPE:     /* Character special */
	    node->vstat.st_mode|=S_IFCHR;
	    break;
	  case BLKTYPE:     /* Block special */
	    node->vstat.st_mode|=S_IFBLK;
	    break;
	  case DIRTYPE:     /* Directory */
	    node->vstat.st_mode|=S_IFDIR;
	    node->vstat.st_size=usrtc_count(tfsnode->childs);
	    break;
	  case FIFOTYPE:      /* Named pipe */
	    node->vstat.st_mode|=S_IFIFO;
	    break;
	  }
	  /* fill correct values */
	  node->vstat.st_uid=tfsnode->owner;
	  node->vstat.st_gid=tfsnode->gowner;
	  node->vstat.st_rdev=0;
	  node->vstat.st_blksize=node->super->block_size;
	  node->vstat.st_nlink=tfsnode->num_links;
	  node->vstat.st_mtime=tfsnode->mtime;
	  node->vstat.st_atime=tfsnode->atime;
	  node->vstat.st_ctime=tfsnode->ctime;
	  return 0;
	} else { /* we will continue */
	  /* change root */
	  tfsroot=usrtc_node_getdata(cnode);
	  /* this is directory ? */
	  if(tfsroot->type!=DIRTYPE)  /* hey, you are wrong */
	    return ENOENT;
	}
      }
    }
  }

  return ENOENT;
}

int tmpfs_truncate(vnode_t *node,off_t offset)
{
  tmpfs_node_t *tfsnode=(tmpfs_node_t*)node->spec_data;
  tmpfs_page_t *pg;
  int pgs,i;
  int pgsrs;

  if(!tfsnode) return ENOENT; /* file was deleted */

  if(tfsnode->size==offset)    return 0;

  /* calculate */
  pgs=tfsnode->size/4096;
  if((tfsnode->size%4096)) pgs++;

  pgsrs=offset/4096;
  if((offset%4096)) pgsrs++;

  if(pgs==pgsrs) {
    tfsnode->size=offset;
    node->vstat.st_size=tfsnode->size;
  } else if(pgs<pgsrs) {
    if(!tfsnode->pages) {
      if((tfsnode->pages=malloc(sizeof(*tfsnode->pages)))==NULL) 
	return ENOSPC;
      tfsnode->pages->next=NULL;
      tfsnode->pages->prev=NULL;
      tfsnode->pages->page=NULL;
      tfsnode->pages->state=PG_HOLE;
    }
    pg=tfsnode->pages;
    for(i=0;i<pgsrs;i++) {
      if(pg->next==NULL) {
	if((pg->next=malloc(sizeof(*pg->next)))==NULL) {
	  tfsnode->size+=4096*i;
	  node->vstat.st_size=tfsnode->size;
	  return ENOSPC;
	}
	pg->next->next=NULL;
	pg->next->prev=pg;
	pg->next->page=NULL;
	pg->next->state=PG_HOLE;
      }
    }
    tfsnode->size=offset;
    node->vstat.st_size=tfsnode->size;
  } else if(pgs>pgsrs) { /* it's possible *only* when page already exist */
    pg=tfsnode->pages;
    if(pgsrs==0) { /* remove all */
      while(pg->next!=NULL) {
	if(pg->state!=PG_HOLE && pg->page) free(pg->page);
	pg=pg->next;
	free(pg->prev);
      }
      if(pg) {
	if(pg->state!=PG_HOLE && pg->page) free(pg->page);
	free(pg);
      }
      tfsnode->pages=NULL;
    } else {
      /* slide to the end */
      while(pg->next!=NULL) pg=pg->next;
      for(i=0;i<pgsrs;i++) {
	pg=pg->prev;
	if(pg->next->state!=PG_HOLE && pg->next->page) free(pg->next->page);
	free(pg->next);
	pg->next=NULL;
      }
    }
    tfsnode->size=offset;
    node->vstat.st_size=tfsnode->size;
  }

  return 0;
}

void *tmpfs_read(vnode_t *node,void *buf,size_t *size,off_t offset)
{
  size_t nsize=*size;
  tmpfs_node_t *tnode=(tmpfs_node_t*)node->spec_data;
  ulong_t pg_num=0;
  ulong_t pg_off=0;
  tmpfs_page_t *rpg;
  char nil_pg[4096]; /* FIXME: change to constant ! */
  void *rb=NULL;
  time_t now;
  int i;

  if(!node) return NULL; /* paranoic */

  if(!tnode) return NULL; /* file was removed */
  if(!nsize) {
    errno=0;
    return NULL;
  }

  if(S_ISDIR(node->vstat.st_mode)) return NULL; /* wtf are you doing ? */

  /* seems all things was checked */
  pg_num=offset/node->super->block_size;
  pg_off=offset%node->super->block_size;
#ifdef TMPFS_DEBUG
  console_printf("[tmpfs] R sgtp #%d:%d for %d:%d\n",pg_num,pg_off,nsize,offset);
#endif
  rpg=tnode->pages; 

  for(i=0;i<pg_num;i++) 
    rpg=rpg->next;

  if(pg_off+nsize>4096)
    nsize=node->super->block_size-pg_off;
  
  if(!rpg) { /* overrun */
    *size=0;
    goto __end_read;
  }

  if(rpg->state==PG_HOLE) {
    memset(nil_pg,'\0',4096);
    rb=(void *)(nil_pg+pg_off);
  } else {
    rb=(void *)(rpg->page+pg_off);
  }

  *size=nsize; /* fix up */

 __end_read:

  now=time(NULL);
  tnode->atime=now;
  node->vstat.st_atime=now;

  return rb;
}

int tmpfs_write(vnode_t *node,void *buf,size_t *size,off_t offset)
{
  tmpfs_node_t *tfsnode=(tmpfs_node_t *)(node->spec_data);
  tmpfs_page_t *pg;
  int pgpass,pgipss,i;
  size_t nsize=*size,rsize=*size;

  if(!tfsnode) return ENOENT; /* file was removed */

  if(tfsnode->type!=REGTYPE)    return EINVAL; /* paranoic */
  /* *NOTE* we're should calculate how much pages do we need to pass */
  pgpass=offset/4096;
  pgipss=offset%4096;
#ifdef TMPFS_DEBUG
  console_printf("[tmpfs] W sgtp #%d:%d for %d:%d\n",pgpass,pgipss,nsize,offset);
#endif
  if(tfsnode->pages==NULL) { /* well create this one */
    if((tfsnode->pages=malloc(sizeof(*tfsnode->pages)))==NULL) 
      return ENOSPC;
    tfsnode->pages->state=PG_HOLE;
    tfsnode->pages->page=NULL;
    tfsnode->pages->next=NULL;
    tfsnode->pages->prev=NULL;
  }

  pg=tfsnode->pages;

  for(i=0;i<pgpass;i++) {
    if(!pg->next) {
      if((pg->next=malloc(sizeof(*pg)))==NULL) 
	return ENOSPC;
      pg->next->state=PG_HOLE;
      pg->next->page=NULL;
      pg->next->next=NULL;
      pg->next->prev=pg;
      if((i+1)<pgpass) { /* in this case we should make a hole */
	tfsnode->size+=4096;
      }
    }
    pg=pg->next;
  }

  if(pg->state==PG_HOLE) {
    if((pg->page=malloc(4096))==NULL) {
      node->vstat.st_size=tfsnode->size;
      return ENOSPC;
    } 
    pg->state=PG_NORM;
  }
  if((pgipss+nsize)>4096) {
    nsize=4096-pgipss;
  }

  memcpy(pg->page+pgipss,buf,nsize);
  tfsnode->size+=nsize;
  /* well - now it's a good to revert page */
  if(nsize<rsize) {
    nsize=rsize-nsize;
    if(!pg->next) {
      if((pg->next=malloc(sizeof(*pg)))==NULL) 
	return ENOSPC;
      pg->next->state=PG_HOLE;
      pg->next->page=NULL;
      pg->next->next=NULL;
      pg->next->prev=pg;
    }
    pg=pg->next;
    if((pg->page=malloc(4096))==NULL) {
      node->vstat.st_size=tfsnode->size;
      return ENOSPC;
    }
    memcpy((char*)pg->page,(char *)buf+nsize,nsize);
    tfsnode->size+=nsize;
  }

  node->vstat.st_size=tfsnode->size;

  return 0;
}

int tmpfs_unlink(vnode_t *node,uid_t uid,gid_t gid)
{
  tmpfs_node_t *tfsnode;
  tmpfs_node_t *parent;
  tmpfs_page_t *pg;
  int ac=0;

  tfsnode=(tmpfs_node_t *)(node->spec_data);
  if(!tfsnode) return ENOENT; /* paranoic */
  else parent=tfsnode->parent;

  /* check rights */
  if(uid!=0) { /* parent *always* is directory  */
    if(parent->owner==uid && 
       ((parent->mask & TFS_UEXEC) && (parent->mask & TFS_UWRITE))) ac++;
    else if(parent->gowner==gid && 
	    ((parent->mask & TFS_GEXEC) && (parent->mask & TFS_GWRITE))) ac++;
    else if((parent->mask & TFS_OEXEC) && (parent->mask & TFS_OWRITE)) ac++;
    if(!ac) return EACCES;
    else ac=0;
    /* check entry rights */
    if(tfsnode->owner==uid && (tfsnode->mask & TFS_UWRITE)) ac++;
    else if(tfsnode->gowner==gid && (tfsnode->mask & TFS_GWRITE)) ac++;
    else if(tfsnode->mask & TFS_OWRITE) ac++;
    if(!ac) return EACCES;
  }

  /* well let's see - first remove all data from a file */
  pg=tfsnode->pages;
  if(pg!=NULL) {
    while(pg->next!=NULL) {
      if(pg->state!=PG_HOLE && pg->page) free(pg->page);
      pg=pg->next;
      free(pg->prev);
    }
    if(pg) {
      if(pg->state!=PG_HOLE && pg->page) free(pg->page);
      free(pg);
    }
  }
  /* free other stuff */
  free(tfsnode->name);
  /* remove this friend from parent */
  usrtc_delete(parent->childs,tfsnode->node);
  free(tfsnode->node);
  free(tfsnode);
  node->spec_data=NULL; /* shootout */

  return 0;
}

int tmpfs_unlinkdir(vnode_t *node,uid_t uid,gid_t gid)
{
  tmpfs_node_t *tfsnode,*child;
  tmpfs_node_t *parent;
  usrtc_node_t *cnode;
  int ac=0;

  tfsnode=(tmpfs_node_t *)(node->spec_data);
  if(!tfsnode) return ENOENT; /* paranoic */
  else parent=tfsnode->parent;

  /* check rights */
  if(uid!=0) { /* parent *always* is directory  */
    if(parent->owner==uid && 
       ((parent->mask & TFS_UEXEC) && (parent->mask & TFS_UWRITE))) ac++;
    else if(parent->gowner==gid && 
	    ((parent->mask & TFS_GEXEC) && (parent->mask & TFS_GWRITE))) ac++;
    else if((parent->mask & TFS_OEXEC) && (parent->mask & TFS_OWRITE)) ac++;
    if(!ac) return EACCES;
    else ac=0;
    /* check entry rights */
    if(tfsnode->owner==uid && (tfsnode->mask & TFS_UWRITE)) ac++;
    else if(tfsnode->gowner==gid && (tfsnode->mask & TFS_GWRITE)) ac++;
    else if(tfsnode->mask & TFS_OWRITE) ac++;
    if(!ac) return EACCES;
  }

  /* well let's see - first remove all data from a directory */
  if(usrtc_count(tfsnode->childs)>2)
    return ENOTEMPTY;
  else {
    while((cnode=usrtc_first(tfsnode->childs))!=NULL) {
      child=(tmpfs_node_t *)usrtc_node_getdata(cnode);
      usrtc_delete(tfsnode->childs,cnode);
      free(child->name);
      free(child);
      free(cnode);
    }
    free(tfsnode->childs);
  }
  /* if we're have a cached directory context */
  if(tfsnode->entries) {
    __entries_clear(tfsnode->entries);
    free(tfsnode->entries);
  }
  /* free other stuff */
  free(tfsnode->name);
  /* remove this friend from parent */
  usrtc_delete(parent->childs,tfsnode->node);
  free(tfsnode->node);
  free(tfsnode);
  node->spec_data=NULL; /* shootout */

  return 0;
}

int tmpfs_create(vnode_t *node,char *path,int *pps,
		 int pp,int mode,uid_t uid,gid_t gid)
{
  usrtc_node_t *cnode,*newnode;
  tmpfs_node_t *tfsroot;
  tmpfs_node_t *newn;
  tmpfs_super_t *super;
  char tname[256];
  time_t now;
  uint64_t srch_hash;
  int i;

  super=(tmpfs_super_t*)node->super->spec_data;
  tfsroot=super->root;

  /* *NOTE:* 0)find root; 1) check permission 2) insert */
  for(i=0;i<pp;i++) {
    memset(tname,'\0',256);
    
    if((i+1)>=pp) {
      strncpy(tname,path+pps[i]+1,strlen(path)-pps[i]);
    } else
      strncpy(tname,path+pps[i]+1,pps[i+1]-pps[i]-1);
#ifdef TMPFS_DEBUG
    console_printf("[tmpfs] {tmpfs_create(intr. lookup)} tname = %s\n",tname);
#endif
    /* make new hash */
    srch_hash=vnode_hash_ch(tname);
    /* NOTE: for flamefreaks - i've check it at all */
    cnode=usrtc_lookup(tfsroot->childs,&srch_hash);
    if((i+1)>=pp && !cnode) { /* our client */
      /* user should have write and exec access to directory */
      if(uid!=0 || gid!=0) { /* it's not a supervisor - will check */
	if(uid==tfsroot->owner && (!(tfsroot->mask && TFS_UWRITE) && !(tfsroot->mask && TFS_UEXEC))) 
	  return EACCES;
	else if(gid==tfsroot->gowner && (!(tfsroot->mask && TFS_GWRITE) && !(tfsroot->mask && TFS_GEXEC)))
	  return EACCES;
	else if(!(tfsroot->mask && TFS_OWRITE) && !(tfsroot->mask && TFS_OEXEC))
	  return EACCES;
      }
      /* well, now we're can insert this */
      if((newn=malloc(sizeof(*newn)))==NULL) return ENOMEM;
      if((newnode=malloc(sizeof(*newnode)))==NULL) return ENOMEM;
      /* associate it with node */
      usrtc_node_init(newnode,newn);
      newn->node=newnode;
      /* give name */
      if((newn->name=strdup(tname))==NULL) {
	free(newn);
	free(newnode);
	return ENOSPC;
      }
      /* hash */
      newn->hash=vnode_hash_ch(newn->name);
#ifdef TMPFS_DEBUG
      console_printf("[tmpfs] create-> (%s)\n",newn->name);
#endif
      newn->childs=NULL; /* it's not a directory */
      newn->pages=NULL; /* yep, will it later */
      newn->parent=tfsroot; /* mark our parent */
      newn->entries=NULL; /* no cached anything */
      newn->type=REGTYPE;
      newn->num_links=0;
      newn->size=0;
      now=time(NULL);
      if(now!=(time_t)-1)
	newn->mtime=newn->atime=newn->ctime=now;
      else
	newn->mtime=newn->atime=newn->ctime=0;
      /* owner */
      newn->owner=uid;
      newn->gowner=gid;
      /* mode */
      newn->mask=0;
      if(mode & S_IRUSR) newn->mask|=TFS_UREAD;
      if(mode & S_IWUSR) newn->mask|=TFS_UWRITE;
      if(mode & S_IXUSR) newn->mask|=TFS_UEXEC;
      if(mode & S_IRGRP) newn->mask|=TFS_GREAD;
      if(mode & S_IWGRP) newn->mask|=TFS_GWRITE;
      if(mode & S_IXGRP) newn->mask|=TFS_GEXEC;
      if(mode & S_IROTH) newn->mask|=TFS_OREAD;
      if(mode & S_IWOTH) newn->mask|=TFS_OWRITE;
      if(mode & S_IXOTH) newn->mask|=TFS_OEXEC;
      /* well, let's fill vnode stats */
      node->vstat.st_size=0;
      node->vstat.st_mode=0;
      node->vstat.st_mode|=mode;
      node->vstat.st_mode|=S_IFREG;
      node->vstat.st_uid=newn->owner;
      node->vstat.st_gid=newn->gowner;
      node->vstat.st_rdev=0;
      node->vstat.st_blksize=node->super->block_size;
      node->vstat.st_nlink=newn->num_links;
      node->vstat.st_mtime=newn->mtime;
      node->vstat.st_atime=newn->atime;
      node->vstat.st_ctime=newn->ctime;
      node->vstat.st_blocks=0;
      /* big shit deals are made */
      usrtc_insert(tfsroot->childs,newnode,&newn->hash); /* insert to the parent */
      node->spec_data=newn;
      return 0;
    } else {
      if(!cnode) 
	return ENOENT; /* ENOENT anyway */
      else {
	/* change root */
	tfsroot=usrtc_node_getdata(cnode);
	/* this is directory ? */
	if(tfsroot->type!=DIRTYPE)  /* hey, you are wrong */
	  return ENOENT;
      }
    }
  }
  
  return ENOENT;
}

int tmpfs_createdir(vnode_t *node,char *path,int *pps,
		    int pp,int mode,uid_t uid,gid_t gid)
{
  usrtc_node_t *cnode,*newnode,*self,*parent;
  tmpfs_node_t *tfsroot,*parentnode,*selfnode;
  tmpfs_node_t *newn;
  tmpfs_super_t *super;
  char tname[256];
  time_t now;
  uint64_t srch_hash;
  int i;

  super=(tmpfs_super_t*)node->super->spec_data;
  tfsroot=super->root;

  /* *NOTE:* 0)find root; 1) check permission 2) insert */
  for(i=0;i<pp;i++) {
    memset(tname,'\0',256);
    
    if((i+1)>=pp) {
      strncpy(tname,path+pps[i]+1,strlen(path)-pps[i]);
    } else
      strncpy(tname,path+pps[i]+1,pps[i+1]-pps[i]-1);
#ifdef TMPFS_DEBUG
    console_printf("[tmpfs] {tmpfs_createdir(intr. lookup)} tname = %s\n",tname);
#endif
    /* make new hash */
    srch_hash=vnode_hash_ch(tname);
    /* NOTE: for flamefreaks - i've check it at all */
    cnode=usrtc_lookup(tfsroot->childs,&srch_hash);
    if((i+1)>=pp && !cnode) { /* our client */
      /* user should have write and exec access to directory */
      if(uid!=0 || gid!=0) { /* it's not a supervisor - will check */
	if(uid==tfsroot->owner && (!(tfsroot->mask && TFS_UWRITE) && !(tfsroot->mask && TFS_UEXEC))) 
	  return EACCES;
	else if(gid==tfsroot->gowner && (!(tfsroot->mask && TFS_GWRITE) && !(tfsroot->mask && TFS_GEXEC)))
	  return EACCES;
	else if(!(tfsroot->mask && TFS_OWRITE) && !(tfsroot->mask && TFS_OEXEC))
	  return EACCES;
      }
      /* well, now we're can insert this */
      if((newn=malloc(sizeof(*newn)))==NULL) return ENOMEM;
      if((newnode=malloc(sizeof(*newnode)))==NULL) return ENOMEM;
      /* associate it with node */
      usrtc_node_init(newnode,newn);
      newn->node=newnode;
      /* give name */
      if((newn->name=strdup(tname))==NULL) {
      __crdir_enospc:
	free(newn);
	free(newnode);
	return ENOSPC;
      }
      /* hash */
      newn->hash=vnode_hash_ch(newn->name);
#ifdef TMPFS_DEBUG
      console_printf("[tmpfs] createdir-> (%s)\n",newn->name);
#endif
      if((newn->childs=malloc(sizeof(*newn->childs)))==NULL) {
	free(newn->name);
	goto __crdir_enospc;
      } else
	usrtc_init(newn->childs,USRTC_BST,USRTC_COUNT_T_MAX,__tmpfscmp_hash);
      if((parent=malloc(sizeof(*self)))==NULL) {
	free(newn->name);
	free(newn->childs);
	goto __crdir_enospc;
      }

      /* insert self with other node */
      if((parentnode=malloc(sizeof(*parentnode)))==NULL) {
	free(parent);
	free(newn->name);
	free(newn->childs);
	goto __crdir_enospc;
      }
      if((selfnode=malloc(sizeof(*selfnode)))==NULL) {
	free(parentnode);
	free(parent);
	free(newn->name);
	free(newn->childs);
	goto __crdir_enospc;
      }
      if((self=malloc(sizeof(*self)))==NULL) {
	free(parentnode);
	free(selfnode);
	free(parent);
	free(newn->name);
	free(newn->childs);
	goto __crdir_enospc;
      } else {
	selfnode->name=strdup(".");
	selfnode->hash=vnode_hash_ch(selfnode->name);
	selfnode->parent=NULL;
	selfnode->type=DIRTYPE;
	parentnode->name=strdup("..");
	parentnode->hash=vnode_hash_ch(parentnode->name);
	parentnode->parent=NULL;
	parentnode->type=DIRTYPE;
	/* insert self */
	usrtc_node_init(self,selfnode);
	usrtc_insert(newn->childs,self,&selfnode->hash);
	/* insert parent */
	usrtc_node_init(parent,parentnode);
	usrtc_insert(newn->childs,parent,&parentnode->hash);
      }
      newn->pages=NULL; /* it's not a regular file */
      newn->parent=tfsroot; /* mark our parent */
      newn->entries=NULL; /* no cached anything */
      newn->type=DIRTYPE;
      newn->num_links=0;
      newn->size=2;
      now=time(NULL);
      if(now!=(time_t)-1)
	newn->mtime=newn->atime=newn->ctime=now;
      else
	newn->mtime=newn->atime=newn->ctime=0;
      /* owner */
      newn->owner=uid;
      newn->gowner=gid;
      /* mode */
      newn->mask=0;
      if(mode & S_IRUSR) newn->mask|=TFS_UREAD;
      if(mode & S_IWUSR) newn->mask|=TFS_UWRITE;
      if(mode & S_IXUSR) newn->mask|=TFS_UEXEC;
      if(mode & S_IRGRP) newn->mask|=TFS_GREAD;
      if(mode & S_IWGRP) newn->mask|=TFS_GWRITE;
      if(mode & S_IXGRP) newn->mask|=TFS_GEXEC;
      if(mode & S_IROTH) newn->mask|=TFS_OREAD;
      if(mode & S_IWOTH) newn->mask|=TFS_OWRITE;
      if(mode & S_IXOTH) newn->mask|=TFS_OEXEC;
      /* big shit deals are made */
      usrtc_insert(tfsroot->childs,newnode,&newn->hash); /* insert to the parent */
      return 0;
    } else {
      if(!cnode) 
	return ENOENT; /* ENOENT anyway */
      else {
	/* change root */
	tfsroot=usrtc_node_getdata(cnode);
	/* this is directory ? */
	if(tfsroot->type!=DIRTYPE)  /* hey, you are wrong */
	  return ENOENT;
      }
    }
  }
  
  return ENOENT;
}

int tmpfs_opendir(vnode_t *node,char *path,int *pps,int pp)
{
  tmpfs_node_t *tnode=(tmpfs_node_t*)node->spec_data;
  tmpfs_node_t *cnode;
  usrtc_node_t *acnode;
  int i=0;
  int ecnt;
  time_t now;

  now=time(NULL);
  tnode->atime=now;
  node->vstat.st_atime=now;

  if(tnode->entries) /* already cached */
    return 0;

  ecnt=usrtc_count(tnode->childs);
  tnode->entries=malloc(sizeof(uintptr_t)*(ecnt+1));
  if(!tnode->entries) return ENOMEM;
  memset(tnode->entries,0,sizeof(uintptr_t)*(ecnt+1));
  acnode=usrtc_first(tnode->childs);
#ifdef TMPFS_DEBUG
  console_printf("[tmpfs] opendir-> ecnt=%d\n",ecnt);
#endif
  for(i=0;i<ecnt;i++) {
    tnode->entries[i]=malloc(sizeof(struct dirent));
    if(!tnode->entries[i]) {
      free(tnode->entries);
      tnode->entries=NULL;
      return ENOMEM;
    }
    cnode=(tmpfs_node_t*)usrtc_node_getdata(acnode);
    strcpy(tnode->entries[i]->d_name,cnode->name);
#ifdef TMPFS_DEBUG
    console_printf("[tmpfs] lu.name='%s'\n",cnode->name);
#endif
    tnode->entries[i]->d_ino=i;
    tnode->entries[i]->d_off=i;
    tnode->entries[i]->d_namlen=strlen(cnode->name);
    tnode->entries[i]->d_rec_len=sizeof(*tnode->entries[i]);
#ifdef TMPFS_DEBUG
    console_printf("[tmpfs] acnode->next entry\n");
#endif
    acnode=usrtc_next(tnode->childs,acnode);
  }

  return 0;
}

int tmpfs_close(vnode_t *node)
{
  tmpfs_node_t *tnode=(tmpfs_node_t*)node->spec_data;

  if(tnode->entries) {
    __entries_clear(tnode->entries);
    free(tnode->entries);
    tnode->entries=NULL;
  }

  node->spec_data=NULL; /* clear it up */

  return 0;
}

int tmpfs_readdir(vnode_t *node,size_t size,off_t off,struct dirent *dentry)
{
  tmpfs_node_t *tnode=(tmpfs_node_t*)node->spec_data;
  ulong_t ecnt=usrtc_count(tnode->childs);

  if(off>ecnt+2) {
    errno=EOVERFLOW;
    return -1;
  } else if(tnode->entries[off]==NULL) {
    return DIRENT_EL;
  } else if(tnode->atime<tnode->mtime) {
    __entries_clear(tnode->entries);
    free(tnode->entries);
    tnode->entries=NULL;
    /* reread all */
    errno=tmpfs_opendir(node,NULL,NULL,0);
    return DIRENT_RR;
  }

  strcpy(dentry->d_name,tnode->entries[off]->d_name);
  dentry->d_ino=tnode->entries[off]->d_ino;
  dentry->d_off=tnode->entries[off]->d_off;
  dentry->d_namlen=tnode->entries[off]->d_namlen;
  dentry->d_rec_len=tnode->entries[off]->d_rec_len;

  return 0;
}

/* local functions */
static inline void __entries_clear(struct dirent **ent)
{
  while(*ent!=NULL)    free(*ent++);

  return;
}

static int __tmpfscmp_hash(const void *a,const void *b)
{
  return *(uint64_t*)a-*(uint64_t*)b;
}



