/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * include/ata/ata.h: ATA interface
 *  command set accords to the ATA/ATAPI-7
 */

#ifndef __ATA_H__
#define __ATA_H__

#include <stdbool.h>
#include <sys/types.h>
#include <sys/ievent.h>
#include <sys/uevent.h>
#include <datalib/list.h>
#include <block/blkdev.h>
#include <block/request.h>
#include <dma.h>
#include <timeout.h>
#include <completion.h>

#define ata_port_active(port)       ((port)->flags & ATA_PORT_ACTIVE)
#define ata_port_lock(port)         pthread_mutex_lock(&(port)->mutex)
#define ata_port_trylock(port)      pthread_mutex_trylock(&(port)->mutex)
#define ata_port_unlock(port)       pthread_mutex_unlock(&(port)->mutex)
#define ata_queued_cmd(cmd)         (cmd->tag != -1)

#define QUEUED_PROTOCOL_MASK        ((1 << ATA_PROT_DMA_QUEUED) | (1 << ATA_PROT_FPDMA_QUEUED) | \
                                     (1 << ATA_PROT_PACKET))

#define DMA_PROTOCOL_MASK           ((1 << ATA_PROT_DMA) | (1 << ATA_PROT_DMA_QUEUED) | \
                                     (1 << ATA_PROT_FPDMA_QUEUED) | (1 << ATA_PROT_STREAM_DMA))

#define PIO_PROTOCOL_MASK           ((1 << ATA_PROT_PIO) | (1 << ATA_PROT_MULTI_PIO) | \
                                     (1 << ATA_PROT_STREAM))

#define ATA_16BIT_REGS              (ATA_REG_FEATURE | ATA_REG_LBA_HIGH | \
                                     ATA_REG_LBA_MID | ATA_REG_LBA_LOW | ATA_REG_SECT_COUNT)

#define ATA_STATUS_TIMEOUT  400
#define ATA_DEFAULT_TIMEOUT 50 /* in milliseconds */

enum {
  NR_ATA_PROTOCOLS = 9,
  NR_ATA_IOREGS = 8
};

/* ATA commands */
enum {
  /* CFA feature set */
  ATA_ERASE_SECTORS = 0xC0,
  ATA_REQ_ERROR_CODE = 0x03,
  AT_TRANSLATE_SECTOR = 0x87,
  ATA_WRITE_MULTIPLE_WERASE = 0xCD,
  ATA_WRITE_SECTORS_WEARASE = 0x38,
  /* media card feature */
  ATA_CHECK_MEDIA_CARD_TYPE = 0xD1,
  /* power management feature set */
  ATA_CHECK_POWER_MODE = 0xE5,
  /* streaming feature */
  ATA_CONFIGURE_STREAM = 0x51,
  /* general feature set */
  ATA_DEVICE_CONFIGURATION = 0xB1,
  ATA_DEVICE_RESET = 0x08,
  ATA_DOWNLOAD_MICROCODE = 0x92,
  ATA_EXEC_DEV_DIAGNOSTIC = 0x90,
  ATA_FLUSH_CACHE = 0xE7,
  ATA_FLUSH_CACHE_EXT = 0xEA,
  ATA_GET_MEDIA_STATUS = 0xDA,
  ATA_IDENTIFY_DEVICE = 0xEC,
  ATA_IDENTIFY_PACKET_DEVICE = 0xA1,
  ATA_IDLE = 0xE3,
  ATA_IDLE_IMMEDIATE = 0xE1,
  ATA_INIT_DEV_PARAMS = 0x91,
  ATA_MEDIA_EJECT = 0xED,
  ATA_MEDIA_LOCK = 0xDE,
  ATA_MEDIA_UNLOCK = 0xDF,
  ATA_NOP = 0,
  ATA_PACKET = 0xA0,
  ATA_READ_BUFFER = 0xE4,
  ATA_READ_DMA = 0xC8,
  ATA_READ_DMA_EXT = 0X25,
  ATA_READ_DMA_QUEUED = 0xC7,
  ATA_READ_DMA_QUEUED_EXT = 0X26,
  ATA_READ_MULTIPLE = 0xC4,
  ATA_READ_MULTIPLE_EXT = 0x29,
  ATA_READ_NATIVE_MAX_ADDR = 0xF8,
  ATA_READ_SECTORS = 0x20,
  ATA_READ_SECTORS_EXT = 0x24,
  ATA_READ_STREAM_DMA_EXT = 0x2A,
  ATA_READ_STREAM_EXT = 0x2B,
  ATA_READ_VERIFY_SECTORS = 0x40,
  ATA_READ_VERIFY_SECTORS_EXT = 0x42,
  /* security feature set */
  ATA_DISABLE_PASSWD = 0xF6,
  ATA_ERASE_PREPARE = 0xF3,
  ATA_ERASE_UNIT = 0xF4,
  ATA_FREEZE_LOCK = 0xF5,
  ATA_SET_PASSWD = 0xF1,
  ATA_UNLOCK = 0xF2,
  /* general feature set */
  ATA_SEEK = 0x70,
  ATA_SERVICE = 0xA2,
  ATA_SET_FEATURES = 0xEF,
  ATA_SET_MAX = 0xF9,
  ATA_SET_MULTIPLE_MODE = 0xC6,
  ATA_SLEEP = 0xE6,
  ATA_SMART = 0xB0,
  ATA_STANDBY = 0xE2,
  ATA_STANDBY_IMMEDIATE = 0xE0,
  ATA_WRITE_BUFFER = 0xE8,
  ATA_WRITE_DMA = 0xCA,
  ATA_WRITE_DMA_EXT = 0x35,
  ATA_WRITE_DMA_FUA_EXT = 0x3D,
  ATA_WRITE_DMA_QUEUED = 0xCC,
  ATA_WRITE_DMA_QUEUED_EXT = 0x36,
  ATA_WRITE_DMA_QUEUED_FUA_EXT = 0x3E,
  ATA_WRITE_MULTIPLE = 0xC5,
  ATA_WRITE_MULTIPLE_EXT = 0x39,
  ATA_WRITE_MULTIPLE_FUA_EXT = 0xCE,
  ATA_WRITE_SECTORS = 0x30,
  ATA_WRITE_SECTORS_EXT = 0x34,
  ATA_WRITE_STREAM_DMA_EXT = 0x3A,
  ATA_WRITE_STREAM_EXT = 0x3B
};

/* SATA specific commands */
enum {
  SATA_READ_FPDMA_QUEUED = 0x60,
  SATA_WRITE_FPDMA_QUEUED = 0x61
};

/* enable/disable features codes */
enum {
  AF_ENABLE_8BIT_PIO = 0x01,
  AF_ENABLE_WRITE_CACHE = 0x02,
  /*
   * sector transfer mode based on sector
   * count register
   */
  AF_SECT_MULT_COUNT = 0x03,
  /* advanced power management */
  AF_ENABLE_ADVANCED_PM = 0x05,
  AF_ENABLE_STANDBY_PWRUP = 0x06,
  AF_PWRUP_SPINUP = 0x07,
  AF_ENABLE_CFA_PM1 = 0x0A,
  AF_DISABLE_MEDIA_STATUS_NOTIF = 0x31,
  AF_ENABLE_AUTOMATIC_ACOUSTIC = 0x42,
  AF_SETMAX_HOSTIF_SECT_TIMES = 0x43,
  AF_DISABLE_READ_LOOK_AHEAD = 0x55,
  AF_ENABLE_RELEASE_INTR = 0x5D,
  AF_ENABLE_SERVICE_INTR = 0x5E,
  AF_DISABLE_PWRON_REVERT = 0x66,
  AF_DISABLE_8BIT_PIO = 0x81,
  AF_DISABLE_WRITE_CACHE = 0x82,
  AF_DISABLE_ADVANCED_PM = 0x85,
  AF_DISABLE_STANDBY_PWRUP = 0x86,
  AF_DISABLE_CFA_PM1 = 0x8A,
  AF_ENABLE_MEDIA_SATUS_NOTIF = 0X95,
  AF_ENABLE_READ_LOOK_AHEAD = 0xAA,
  AF_DISABLE_AUTOMATIC_ACOUSTIC = 0xC2,
  AF_ENABLE_PWRON_REVERT = 0xCC,
  AF_DISABLE_RELEASE_INTR = 0xDD,
  AF_DISABLE_SERVICE_INTR = 0xDE
};

/* various ATA related constants */
enum {
  ATA_SIGNATURE = 0xA5,
  DEV_FPDMA_FUA = 0x80,
  DEV_LBA = 0x40,
  FEATURE_URG = 0x80,
  FEATURE_RC = 0x40,
  FEATURE_NS = 0x20,
  FEATURE_HSE = 0x10,
  DEVCTL_HOB = 0x80,
  DEVCTL_SRST = 0x04,
  DEVCTL_NIEN = 0x02
};

/* ATA configurations features */
enum {
  DEVICE_CONFIGURATION_RESTORE = 0xC0,
  DEVICE_CONFIGURATION_FREEZE_LOCK,
  DEVICE_CONFIGURATION_INDENTIFY,
  DEVICE_CONFIGURATION_SET
};

/* ATA set max features */
enum {
  MAX_SET_PASSWORD = 0x01,
  MAX_LOCK = 0x02,
  MAX_UNLOCK = 0x03,
  MAX_FREEZE_LOCK = 0x04
};

enum {
  ATA_SENSE_BUFLEN = 96,
  MAX_ATA_PORTS = 16,
  ATA_MAX_HWIF_DEVICES = 16 /* max devices per one interface */
};

/* ATA smart features */
enum {
  SMART_READ_DATA = 0xD0,
  SMART_EN_DIS_ATTR_AUTOSAVE = 0xD2,
  SMART_SAVE_ATTR_VALUES = 0xD3,
  SMART_EXEC_OFFLINE_IMMEDIATE = 0xD4,
  SMART_READ_LOG = 0xD5,
  SMART_WRITE_LOG = 0xD6,
  SMART_ENABLE_OPERATIONS = 0xD8,
  SMART_DISABLE_OPERATIONS = 0xD9,
  SMART_RETURN_STATUS = 0xDA
};

typedef enum __ata_io_protocol {
  ATA_PROT_NODATA,
  ATA_PROT_PACKET,
  ATA_PROT_PIO,
  ATA_PROT_MULTI_PIO,
  ATA_PROT_DMA,
  ATA_PROT_DMA_QUEUED,
  /* protocol supporting native command queuing */
  ATA_PROT_FPDMA_QUEUED,
  ATA_PROT_STREAM,
  ATA_PROT_STREAM_DMA
} ata_io_protocol_t;

/* ATA taskfile flags */
enum {
  TASKFILE_EXT = 0x01,
  TASKFILE_FUA = 0x02,
  TASKFILE_FEATURE = 0x04,
  TASKFILE_ERROR = 0x04,
  TASKFILE_STATUS = 0x08,
  TASKFILE_SECT_COUNT = 0x10,
  TASKFILE_LBA_LOW = 0x20,
  TASKFILE_LBA_MID = 0x40,
  TASKFILE_LBA_HIGH = 0x80,
  TASKFILE_REGS_MASK = 0xFC,
};

/*
 * ATA dma modes
 *
 * Don't support quite archaic modes such SWDMA
 */
enum {
  ATA_DMA_NOT,
  ATA_MWDMA0,
  ATA_MWDMA1,
  ATA_MWDMA2,
  ATA_MWDMA3,
  ATA_MWDMA4,
  ATA_UDMA0,
  ATA_UDMA1,
  ATA_UDMA2,
  ATA_UDMA3,         /* 33 MB/s */
  ATA_UDMA4,         /* 66 MB/s */
  ATA_UDMA5,         /* 100 MB/s */
  ATA_UDMA6,         /* 133 MB/s */
  ATA_UDMA7
};

/* ata hwif status bits */
enum {
  ATA_PORT_ACTIVE = 0x01
};

/* ATA cmd protocols */
enum {
  CMD_PROT_ATA = 0,
  CMD_PROT_ATAPI = 2
};

typedef enum __ata_pw_mode {
  ATA_PW_MODE_ACTIVE,
  ATA_PW_MODE_IDLE,
  ATA_PW_MODE_SLEEP,
  ATA_PW_MODE_STANDBY
} ata_pw_mode_t;

/* host interfaces */
typedef enum __ata_host_iface {
  ATA_IFACE_IDE,
  ATA_IFACE_AHCI
} ata_host_iface_t;

/* ata physical interfaces */
enum {
  ATA_PHYIF_IDE,
  ATA_PHYIF_SATA
};

typedef enum __ata_media_type {
  ATA_MEDIA_NOT,
  ATA_MEDIA_DISK,
  ATA_MEDIA_CD,
  ATA_MEDIA_DVD,
  ATA_MEDIA_FLOPPY
} ata_media_type_t;

enum {
  PORT_STAT_IRQ = 0x01,
  PORT_STAT_ERROR = 0x02,
  PORT_STAT_IO_DONE = 0x04
};

/* command phases */
enum {
  PHASE_INACTIVE,             /* there are not processed commands */
  PHASE_WAIT_READY,           /* wait when device will be ready for I/O */
  PHASE_INTRQ_ASSERTED,
  PHASE_TRANSFER
};

/* ATA error types */
enum {
  ATA_ERROR_NOERR,
  ATA_ERROR_DEVICE,
  ATA_ERROR_HOST,
};

typedef uint32_t port_status_t;
typedef uint32_t ata_reg_set_t;
struct __ata_host_driver;
struct __ata_driver;
struct __ata_host;
struct __ata_device;
struct __ata_cmd;

typedef struct __ata_registers {
  union {
    ulong_t command;
    ulong_t status;
  };
  ulong_t data;
  ulong_t device;
  ulong_t dev_ctl;
  union {
    ulong_t feature;
    ulong_t error;
  };
  union {
    ulong_t lba_high;
    ulong_t byte_count_high;
  };
  union {
    ulong_t lba_mid;
    ulong_t byte_count_low;
  };
  ulong_t lba_low;
  union {
    ulong_t sect_count;
    ulong_t intr_reason;
  };
} ata_registers_t;

/* ATA hardware interface */
typedef struct __ata_hwif {
  void (*outb)(ulong_t port, uint8_t val);
  uint8_t (*inb)(ulong_t port);
  void (*outw)(ulong_t port, uint16_t val);
  uint16_t (*inw)(ulong_t port);
} ata_hwif_t;

typedef struct __ata_port {
  struct __ata_host *host;          /* host servicing this interface */
  int num;
  struct __ata_device *devices[2];  /* devices connected to the interface */
  ata_registers_t regs;

  /* dma controller channel provided by the host */
  dma_channel_t *dma_chan;
  void *sg_pool;                    /* scatter/gather pool */
  ata_hwif_t *hwif;
  uint16_t irq;
  uint32_t flags;
  int state;
  /* serviced device */
  struct __ata_device *serv_dev;
  event_t *io_done;
  pthread_mutex_t mutex;
} ata_port_t;

struct ata_io_modes {
  uint8_t pio;
  uint8_t mwdma;
  uint8_t udma;
};

typedef struct __ata_port_ops {
  int (*init)(ata_port_t*);
  void (*shutdown)(ata_port_t*);
  bool (*probe)(ata_port_t*, int dev_id);
  void (*suspend)(ata_port_t*, int state);
  void (*resume)(ata_port_t*);
  void (*reset)(ata_port_t*);
  port_status_t (*read_status)(ata_port_t*);
  void (*set_pio_mode)(ata_port_t*, int dev_id, uint8_t mode);
  int (*init_dma)(ata_port_t*, uint32_t min_chunk_size,
                  uint32_t max_tsize);
  void (*set_dma_mode)(ata_port_t*, int dev_id, uint8_t mode);
  void (*enable_dma)(ata_port_t*, int dev_id, bool enable);
  int (*issue_cmd)(ata_port_t*, int tag);
  void (*set_irq)(ata_port_t*, uint16_t irq);
  void (*clear_irq)(ata_port_t*);
  void (*filter_io_modes)(ata_port_t *port, int dev_id, struct ata_io_modes *modes);
} ata_port_ops_t;

typedef struct __ata_host {
  ata_port_t *ports[MAX_ATA_PORTS];
  int nr_ports;
  ulong_t waited_ports;
  struct ata_io_modes io_modes;
  ata_host_iface_t host_iface;
  uint16_t dma_timeout;
  uint16_t flags;             /* driver specific flags */
  /* irq is shared between all ports */
  bool irq_shared;
  interrupt_event_t *ievent;
  pthread_t irq_thread;
  struct __ata_host_driver *drv;
  ata_port_ops_t *port_ops;
  void *extra_ops;           /* host interface dependend operations */
  void *private;
  void *bus_private;
} ata_host_t;

typedef struct __ata_taskfile {
  union {
    uint8_t error;
    uint16_t feature;
  };
  uint8_t nsect;
  uint8_t lbal;
  uint8_t lbam;
  uint8_t lbah;
  uint8_t hob_feature;
  uint8_t hob_nsect;
  uint8_t hob_lbal;
  uint8_t hob_lbam;
  uint8_t hob_lbah;
  uint8_t device;
  union {
    uint8_t cmd;
    uint8_t status;
  };
  uint32_t flags;
} ata_taskfile_t;

typedef struct __ata_cmd {
  ata_taskfile_t tf;
  uint8_t sense[ATA_SENSE_BUFLEN];
  uint8_t sense_len;
  uint8_t retries;
  uint8_t retry_num;
  short tag;
  int err;                    /* type of occured error */
  struct {
    unsigned io_dir : 1;      /* 0 - read, 1 - write */
    unsigned dma_prepared : 1;
    unsigned read_tf_end : 1; /* read taskfile at command completeness */
  };
  int nr_errors;
  struct __blk_request *rq;
  uint16_t timeout;
  uint64_t start_jiffies;
  struct sg_entry *sg_list;
  uint16_t sg_list_len;
  struct __ata_device *dev;   /* device the cmd run on */
  ata_io_protocol_t protocol;
  void *buf;
  /*
   * for some commands size of one transferred
   * block can differ from the native sector size
   */
  int block_shift;
  void (*done)(struct __ata_cmd *cmd);
  void *done_data;
  list_node_t node;
} ata_cmd_t;

typedef struct __ata_features {
  unsigned pio_8bit : 1;
  unsigned write_cache: 1;
  unsigned standby_pwrup : 1;
  unsigned cfa_pm1 : 1;
  unsigned media_status_notif : 1;
  unsigned automatic_acoustic : 1;
  unsigned release_intr : 1;
  unsigned service_intr : 1;
  unsigned read_look_ahead : 1;
  unsigned pwron_revert : 1;
  unsigned stream : 1;
} __attribute__ ((packed)) ata_features_t;

typedef struct __time_activities {
  uint64_t last_service_jiffies;
  uint64_t mean_serv_jiffies;
  uint64_t mean_wait_jifies;
} time_activities_t;

typedef struct __ata_media {
  ata_media_type_t type;
  unsigned present : 1;
  unsigned rdonly : 1;
  unsigned removable : 1;
  unsigned written : 1;
} ata_media_t;

typedef struct __ata_capabilities {
  union {
    struct {
      uint8_t vendor_specific;
      uint8_t dma : 1;
      uint8_t lba : 1;
      uint8_t iordy_disabling : 1;
      uint8_t iordy : 1;
      uint8_t __res4 : 1;
      uint8_t overlap_op : 1;
      uint8_t cmd_queuing : 1;
      uint8_t interl_dma : 1;
    };
    uint16_t word49;
  };
  union {
    struct {
      uint16_t standby_tmr_val_min : 1;
      uint16_t __res5 : 16;
    };
    uint16_t word50;
  };
  union {
    struct {
      /* removable media status notification support */
      uint16_t rem_media_stat_notif : 2;
      uint16_t __res17 : 13;
    };
    uint16_t word127;
  };
  union {
    struct {
      /* security support flag */
      uint16_t security : 1;
      /* variable fields */
      uint16_t __var0;
      /* enhansed security erase */
      uint16_t enh_sec_erase : 1;
      uint16_t __res21 : 2;
      uint16_t __var1 : 1;
      uint16_t __res22 : 7;
    };
    uint16_t word128;
  };
} ata_capabilities_t;

/* set of supported ata commands */
typedef struct __ata_cmd_set {
  union {
    struct {
      uint16_t smart : 1;
      uint16_t secure_mode : 1;
      uint16_t removable_media : 1;
      uint16_t power_management : 1;
      uint16_t packet : 1;
      uint16_t write_cache : 1;
      uint16_t look_ahead : 1;
      uint16_t release_intr : 1;
      uint16_t service_intr : 1;
      uint16_t device_reset : 1;
      uint16_t host_prot_area : 1;
      uint16_t __res10 : 1;
      uint16_t write_buffer : 1;
      uint16_t read_buffer : 1;
      uint16_t nop : 1;
      uint16_t __res11;
    };
    uint16_t word82;
  };
  union {
    struct {
      uint16_t download_microcode : 1;
      uint16_t rw_dma_queued : 1;
      uint16_t cfa : 1;
      uint16_t advanced_power_man : 1;
      uint16_t removable_media_status_notif : 1;
      uint16_t standby_powerup : 1;
      uint16_t setfeatures_spinup : 1;
      uint16_t __res12;
      uint16_t set_max_security : 1;
      uint16_t acoustic : 1;
      uint16_t lba48 : 1;
      uint16_t device_config_overlay : 1;
      uint16_t flush_cache : 1;
      uint16_t flush_cache_ext : 1;
      uint16_t __res13 : 2;
    };
    uint16_t word83;
  };
  union {
    struct {
      uint16_t smart_error_loging : 1;
      uint16_t smart_test_itself : 1;
      uint16_t media_serial_num : 1;
      uint16_t media_card : 1;
      uint16_t streaming : 1;
      uint16_t gen_purp_logging : 1;
      uint16_t write_fua_ext;
      uint16_t write_amd_queued_fua_ext : 1;
      uint16_t name_64bit : 1;
      uint16_t urg_bit_write_stream : 1;
      uint16_t urg_bit_read_stream : 1;
      uint16_t __res14 : 2;
      /* idle immediate with unload feature */
      uint16_t idle_imm_with_unload : 1;
      uint16_t __res15 : 2;
    };
    uint16_t word84;
  };
} ata_cmd_set_t;

/*
 * parameter information particularly
 * describing hardware features and
 * command set
 */
typedef struct __ata_param_info {
  union {
    struct {
      uint16_t __res0 : 2;
      uint16_t __resp_incompl : 1;
      uint16_t __res1 : 4;
      uint16_t removable : 1;
      uint16_t __res2 : 6;
      uint16_t cmd_prot : 2;
    };
    uint16_t word0;
  };
  uint16_t specific_conf;
  uint32_t cfa_private;
  char serial_num[20];
  char revision[8];
  char model_num[40];
  /*
   * max number of sectors that shall
   * be transfered per interrupt on R/W
   * multiple commands
   */
  uint8_t max_rw_sectors;
  ata_capabilities_t caps;
  union {
    struct {
      uint16_t __res6 : 1;
      uint16_t word76_64_valid : 1;
      uint16_t word88_valid : 1;
      uint16_t __res7 : 1;
    };
    uint16_t word53;
  };
  /* total addressable sectors */
  sector_t total_addr_sects;
  /*
   * the following word is valid only
   * for an atapi device
   */
  union {
    struct {
      uint16_t udma0 : 1;
      uint16_t udma1: 1;
      uint16_t udma2 : 1;
      uint16_t udma3 : 1;
      uint16_t udma4 : 1;
      uint16_t udma5 : 1;
      uint16_t udma6 : 1;
      /* multi dma modes */
      uint16_t pkt_mdma_mode0 : 1;
      uint16_t pkt_mdma_mode1 : 1;
      uint16_t pkt_mdma_mode2 : 1;
      uint16_t dma : 1; /* dma supporting */
      uint16_t __res8 : 4;
      uint16_t dma_dir_req : 1; /* dma dir is required for dma transfers */
    };
    uint16_t word62;
  };
  union {
    struct {
      uint16_t mdma_mode0 : 1;
      uint16_t mdma_mode1 : 1;
      uint16_t mdma_mode2 : 1;
      uint16_t __res9 : 13;
    };
    uint16_t word63;
  };
  uint8_t pio_modes;
  uint16_t min_dma_cycle; /* in nanoseconds */
  /*
   * multiword dma cycle time in nanoceconds
   * recommended by manufacturer
   */
  uint16_t recomm_dma_cycle;
  uint16_t min_pio_cycle;
  /*
   * minimum pio cycle time with
   * iordy frow control
   */
  uint16_t min_pio_cycle_iordy;
  /*
   * the following 2 fields are valid only
   * for atapi devices
   */

  /* typical time in ns from receipt packet
   * command to bus release
   */
  uint16_t pkt_bus_release_delay;
  uint16_t serv_cmd_bsy_time;
  uint8_t max_queue_depth;
  ata_cmd_set_t cmd_set;
  /* maximum supported ATA/ATAPI revision */
  short major_ver_num;
  /* security erase unit completion time */
  uint16_t erase_unit_compl_time;
  /* enhanced security erease completion time */
  uint16_t enhsec_erase_compl_time;
  /* vendor's recommended acoustic management value */
  uint16_t stream_min_req_size;
  int sect_shift;
  uint16_t atapi_byte_count;
} ata_param_info_t;

struct cmd_queue {
  list_head_t cmds;
  int depth;
};

typedef struct __ata_device {
  block_device_t *bdev;
  void *cmd_pool;
  ata_media_t media;
  int id; /* 0 - master; 1 - slave */
  int nice;
  ata_features_t set_features;
  uint8_t pio_mode;
  uint8_t dma_mode;
  bool dma_enabled;
  ata_io_protocol_t io_prot;
  uint16_t nr_failures;
  uint16_t max_failures;
  ulong_t tag_map;
  ata_pw_mode_t pw_mode;
  time_activities_t time_acts;
  /* executed queue of commands */
  struct cmd_queue cmd_queue;
  uint8_t mult_count;
  uint8_t ctl_cmd;
  struct completion cmd_compl;
  ata_cmd_t *run_cmd;
  int cmd_phase;
  int last_cmd_err;
  struct timer_desc intrq_tmr;
  struct timer_desc dma_tmr;
  ata_port_t *host_port;
  struct __ata_driver *drv;
  ata_param_info_t param_info;
  void *private;
} ata_device_t;

/* ata transport driver */
typedef struct __ata_driver {
  uint8_t media_types;  /* mask of media types supported by the driver */
  /*
   * mask of physical interfaces supported by the driver;
   * frames organization really depends on the interfaces
   */
  uint8_t phys_ifaces;
  bool (*match)(ata_host_t *host, ata_device_t *dev);
  int (*init)(ata_device_t*);
  int (*prepare_cmd)(ata_device_t *dev, ata_cmd_t *cmd);
  int (*post_issue_cmd)(ata_device_t *dev, ata_cmd_t *cmd);
  void (*rw_data)(ata_cmd_t*);
  int (*process_rq)(ata_device_t *dev, blk_request_t*);
  void (*timeout_handler)(ata_device_t*);
  void (*prep_cmd_retry)(ata_cmd_t*);
  void (*cmd_done)(ata_cmd_t*);
  void (*error_handler)(ata_device_t *dev, int err_type, uint32_t status);
  void (*irq_handler)(ata_device_t*, port_status_t);
  int (*ready_cmd_tag)(ata_device_t *dev);
  int  (*write_taskfile)(ata_device_t*, ata_taskfile_t*);
  int  (*read_taskfile)(ata_device_t*, ata_taskfile_t*);
  uint8_t (*read_status)(ata_device_t*);
} ata_driver_t;

typedef union __ata_status {
  struct {
    uint8_t err : 1;
    uint8_t reserved : 2;
    uint8_t drq : 1;
    uint8_t serv : 1;
    uint8_t df : 1;
    uint8_t drdy : 1;
    uint8_t bsy : 1;
  };
  uint8_t val;
} __attribute__ ((packed)) ata_status_t;

typedef union __ata_srv_info {
  struct {
    uint8_t cd : 1;
    uint8_t io : 1;
    uint8_t rel : 1;
    uint8_t tag : 5;
  };
  uint8_t val;
} __attribute__ ((packed)) ata_srv_info_t;

typedef union __ata_error {
  struct {
    uint8_t li : 1;
    uint8_t eom : 1;
    uint8_t abrt : 1;
    uint8_t mcr : 1;
    uint8_t idnf : 1;
    uint8_t mc : 1;
    uint8_t unc : 1;
    uint8_t icrc : 1;
  };
  uint8_t val;
} __attribute__ ((packed)) ata_error_t;

typedef struct __ata_sense_value {
  uint8_t li : 1;
  uint8_t eom : 1;
  uint8_t abrt : 1;
  uint8_t key : 5;
} __attribute__ ((packed)) ata_sense_value_t;

static inline void ata_set_port_flags(ata_port_t *port, uint32_t flags)
{
  port->flags |= flags;
}

static inline void ata_clear_port_flags(ata_port_t *port, uint32_t flags)
{
  port->flags &= ~flags;
}

static inline bool ata_queued_protocol(ata_io_protocol_t protocol)
{
  return ((1 << protocol) & QUEUED_PROTOCOL_MASK);
}

static inline bool ata_dma_protocol(ata_io_protocol_t protocol)
{
  return ((1 << protocol) & DMA_PROTOCOL_MASK);
}

static inline bool ata_pio_protocol(ata_io_protocol_t protocol)
{
  return ((1 << protocol) & PIO_PROTOCOL_MASK);
}

static inline bool ata_cmd_queue_empty(ata_device_t *dev)
{
  return list_is_empty(&dev->cmd_queue.cmds);
}

static inline bool ata_drv_support_media(ata_driver_t *drv, ata_media_type_t media)
{
  return (drv->media_types & (1 << media));
}

static inline bool ata_host_support_irq(ata_host_t *host)
{
  return (host->port_ops->clear_irq != NULL);
}

static inline bool ata_host_support_dma(ata_host_t *host)
{
  return (host->io_modes.mwdma || host->io_modes.udma);
}

#endif
