/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * core_servers/include/libata.h: ata library declarations
 *
 * ATA timings are ported from Linux
 */

#ifndef __LIBATA_H__
#define __LIBATA_H__

#include <block/request.h>
#include <errno.h>
#include <bitwise.h>
#include "ata.h"

#define ata_read_taskfile(dev, tf)  dev->drv->read_taskfile(dev, tf)
#define ata_write_taskfile(dev, tf) dev->drv->write_taskfile(dev, tf)
#define ata_read_device_status(dev) dev->drv->read_status(dev)

/* features */
enum {
  FEATURE_8BIT_PIO,
  FEATURE_WRITE_CACHE,
  FEATURE_ADVANCED_PM,
  FEATURE_STANDBY_PWRUP,
  FEATURE_CFA_PM1,
  FEATURE_MEDIA_STATUS_NOTIF,
  FEATURE_AUTOMATIC_ACOUSTIC,
  FEATURE_READ_LOOK_AHEAD,
  FEATURE_RELEASE_INTR,
  FEATURE_SERVICE_INTR,
  FEATURE_PWRON_REVERT
};

/* Timing constants */
enum {
  ATA_TIMING_SETUP  = 0x01,
  ATA_TIMING_ACT8B  = 0x02,
  ATA_TIMING_REC8B  = 0x04,
  ATA_TIMING_CYC8B  = 0x08,
  ATA_TIMING_8BIT   = ATA_TIMING_ACT8B | ATA_TIMING_REC8B |
                      ATA_TIMING_CYC8B,
  ATA_TIMING_ACTIVE = 0x10,
  ATA_TIMING_RECOVER  = 0x20,
  ATA_TIMING_CYCLE  = 0x40,
  ATA_TIMING_UDMA   = 0x80,
  ATA_TIMING_ALL    = ATA_TIMING_SETUP | ATA_TIMING_ACT8B |
                      ATA_TIMING_REC8B | ATA_TIMING_CYC8B |
                      ATA_TIMING_ACTIVE | ATA_TIMING_RECOVER |
                      ATA_TIMING_CYCLE | ATA_TIMING_UDMA
};

struct ata_timing {
  unsigned short mode;    /* ATA mode */
  unsigned short setup;   /* t1 */
  unsigned short act8b;   /* t2 for 8-bit I/O */
  unsigned short rec8b;   /* t2i for 8-bit I/O */
  unsigned short cyc8b;   /* t0 for 8-bit I/O */
  unsigned short active;    /* t2 or tD */
  unsigned short recover;   /* t2i or tK */
  unsigned short cycle;   /* t0 */
  unsigned short udma;    /* t2CYCTYP/2 */
};

int ata_probe_init_ports(ata_host_t *host, int *nr_scanned_ports);
int ata_set_device_pw_mode(ata_device_t *dev, ata_pw_mode_t mode);
int ata_enable_dma(ata_device_t *dev, bool enable);
int ata_prepare_dma(ata_device_t *dev, ata_cmd_t *cmd);
void ata_rebuild_sglist(ata_device_t *dev, uint32_t offset, uint32_t len);
int ata_load_sglist(ata_device_t *dev);
int ata_start_dma(ata_device_t *dev);
void ata_stop_dma(ata_device_t *dev);
void ata_reset_device(ata_device_t *dev);

/*
 * convert block layer request to ata I/O command
 * don't support nodata and packet protocols
 */
int ata_iocmd_from_request(blk_request_t *rq, ata_cmd_t *cmd);
int ata_service_cmd_from_request(blk_request_t *rq, ata_cmd_t *cmd);
bool ata_default_irq_handler(void *data);
/* start ata commands with enabled INTRQ assert */
int ata_start_command(ata_device_t *dev, ata_cmd_t *cmd);
int ata_get_param_info(ata_device_t *dev);
int ata_set_io_protocol(ata_device_t *dev, ata_io_protocol_t protocol);
int ata_enable_feature(ata_device_t *dev, int feature);
int ata_disable_feature(ata_device_t *dev, int feature);
short ata_alloc_tag(ata_device_t *dev);
sector_t ata_tflba_to_sector(ata_taskfile_t *tf);
void ata_sector_to_tflba(sector_t sector, ata_taskfile_t *tf);
int ata_direct_read_taskfile(ata_device_t *dev, ata_taskfile_t *tf);
int ata_direct_write_taskfile(ata_device_t *dev, ata_taskfile_t *tf);
void ata_start_transfer(ata_device_t *dev, ata_cmd_t *cmd);
uint8_t ata_direct_read_status(ata_device_t *dev);
void ata_done_cmd(ata_device_t *dev, ata_cmd_t *cmd);
bool try_prepare_cmd_restart(ata_device_t *dev, ata_cmd_t *cmd);
ata_cmd_t *find_ready_queued_cmd(ata_device_t *dev);
ata_cmd_t *ata_alloc_init_cmd(ata_device_t *dev, ata_io_protocol_t prot);
void ata_free_cmd(ata_cmd_t *cmd);
void do_blk_pio(ata_cmd_t *cmd, uint32_t nr_sects,
                void (*pio_fn)(ata_cmd_t*, void*, uint32_t));

bool wait_dev_ready(ata_device_t *dev, ata_status_t *ret_stat);
void pio_arrayb(ata_cmd_t *cmd, void *buf, uint32_t count);
void pio_arrayw(ata_cmd_t *cmd, void *buf, uint32_t count);

int ata_init_port_resources(ata_port_t *port);
void ata_free_port_resources(ata_port_t *port);

int ata_timing_compute(ata_device_t *dev, unsigned short mode, bool dma,
                       struct ata_timing *t, int T, int UT);
void ata_timing_merge(const struct ata_timing *a, const struct ata_timing *b,
          struct ata_timing *m, unsigned int what);

void ata_normalize_id_str(char *dest, char *src, int len);

static inline void ata_free_tag(ulong_t *tag_map, short tag)
{
  bit_set(tag_map, tag);
}

static inline bool ata_add_queued_cmd(ata_device_t *dev, ata_cmd_t *cmd)
{
  if (dev->cmd_queue.depth == dev->param_info.max_queue_depth)
    return false;

  list_add2tail(&dev->cmd_queue.cmds, &cmd->node);
  dev->cmd_queue.depth++;

  return true;
}

static inline void ata_del_queued_cmd(ata_device_t *dev, ata_cmd_t *cmd)
{
  list_del(&cmd->node);
  dev->cmd_queue.depth--;
}

static inline void ata_clear_irq(ata_host_t *host, ata_port_t *port)
{
  host->port_ops->clear_irq(port);
}

static inline void ata_free_cmd_tag(ata_device_t *dev, ata_cmd_t *cmd)
{
  bit_clear(&dev->tag_map, cmd->tag);
  cmd->tag = -1;
}

static inline bool ata_intrq_enabled(ata_device_t *dev)
{
  return !(dev->ctl_cmd & DEVCTL_NIEN);
}

static inline bool device_is_atapi(ata_device_t *dev)
{
  return (dev->param_info.cmd_prot == CMD_PROT_ATAPI);
}

static inline bool ata_port_busy(ata_port_t *port)
{
  return (port->serv_dev != NULL);
}

static inline void ata_set_pio_mode(ata_device_t *dev, uint8_t mode)
{
  ata_host_t *host = dev->host_port->host;

  dev->pio_mode = mode;
  host->port_ops->set_pio_mode(dev->host_port, dev->id, mode);
}

static inline void ata_set_dma_mode(ata_device_t *dev, uint8_t mode)
{
  ata_host_t *host = dev->host_port->host;

  host->port_ops->set_dma_mode(dev->host_port, dev->id, mode);
}

static inline void ata_set_port_waited(ata_port_t *port)
{
  bit_set(&port->host->waited_ports, port->num);
}

static inline void ata_clear_port_waited(ata_port_t *port)
{
  bit_clear(&port->host->waited_ports, port->num);
}

static inline bool ata_port_waited(ata_port_t *port)
{
  return bit_test(&port->host->waited_ports, port->num);
}

#endif
