/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008,2009 Dmitry Gromada <gromada@jarios.org>
 *
 * Block devices subsystem
 */

#ifndef __BLKDEV_H__
#define __BLKDEV_H__

#include <stdbool.h>
#include <sys/types.h>
#include <arch/atomic.h>
#include <datalib/list.h>
#include "defs.h"

#define BDEV_VENDOR_MAXLEN 256
#define BDEV_MODEL_MAXLEN  256
#define NR_BDEV_IFACES     4

/* status bits */
enum {
  BLKSTAT_MEDIA_ABSENT,
  BLKSTAT_MEDIA_CHANGED
};

enum {
  MEDIA_RDONLY = 0x01,
  MEDIA_WRITTEN = 0x02
};

/* block device types */
enum {
  BDEV_PSEUDO,
  BDEV_HDD,
  BDEV_CDROM,
  BDEV_DVDROM,
  BDEV_FLOPPY,
  BDEV_FLASH_STORAGE,
  BDEV_CARD_READER
};

/* block device hardware interfaces */
enum {
  BDEV_IFACE_UNKNOWN,
  BDEV_IFACE_PSEUDO,
  BDEV_IFACE_ATA,
  BDEV_IFACE_SCSI
};

struct __blk_queue;
struct __blkdev_driver;
struct __blkdev_operations;

struct bdev_id {
  char vendor[BDEV_VENDOR_MAXLEN];
  char model[BDEV_MODEL_MAXLEN];
};

/* block device interface */
struct bd_iface {
  int id; /* sequantial number among interfaces serviced by the same host driver */
  int bus;
  char *name;
  uint16_t max_devs;
  struct __blkdev_driver *driver;
  list_head_t devices;
  void *private;
};

struct bdev_media {
  int sect_shift;
  int capacity;
  uint32_t flags;
  void *private;
};

typedef struct __block_device {
  int type;                         /* type of the device */
  struct bdev_media media;          /* storage media */
  uint32_t status;                  /* status flags */
  int openers;                      /* number of clients using this block device */
  struct bd_iface *iface;           /* interface the device attached to */
  uint16_t num;                     /* device number on the interface */
  bool removable;
  struct __blkreq_queue *queue;
  struct __blkdev_operations *ops;  /* block device operations */
  void *private;                    /* device specific data */
  list_node_t node;
} block_device_t;

typedef struct __blkdev_operations {
  int (*open)(block_device_t*);
  void (*close)(block_device_t*);
  int (*flush)(block_device_t*);
  ssize_t (*ioctl)(block_device_t*, uint16_t cmd, void *buf,
                   size_t data_size, size_t buf_size);
  int (*revalidate)(block_device_t*);
  int (*get_media_status)(block_device_t*);
  int (*identify)(block_device_t*, struct bdev_id*);
} blkdev_operations_t;

static void inline set_blkdev_status(block_device_t *bdev, int stat)
{
  bdev->status |= 1 << stat;
}

static void inline clear_blkdev_status(block_device_t *bdev, int stat)
{
  bdev->status &= ~(1 << stat);
}

static bool inline test_blkdev_status(block_device_t *bdev, int stat)
{
  return (bdev->status & (1 << stat)) != 0;
}

#endif
