/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 *
 * include/block/genstor.h: generic storage device structures and
 *                          constants
 */

#ifndef __GENSTOR_H__
#define __GENSTOR_H__

#include <sys/types.h>
#include <stdbool.h>
#include <arch/atomic.h>
#include <ipc/rpc.h>
#include <queue.h>
#include <debug.h>
#include <dircons.h>
#include <lock/lock.h>
#include "blkdev.h"
#include "request.h"

#define STOR_NAME_MAXLEN 64
#define STOR_BASENAME_MAXLEN 58
#define CACHE_SIZE_UNLIMITED 0xFFFFFFFF

#define storage_cacheable(stor) ((stor)->cache_id >= 0)
#define storage_supp_cmd_abort  ((stor)->flags & STORAGE_SUPP_CMD_ABORT)
#define parts_rlock(stor)       rw_usem_try_lock(&stor->part_sem, RW_SEM_READ)
#define parts_runlock(stor)     rw_usem_unlock(&stor->part_sem, RW_SEM_READ)
#define parts_wlock(stor)       rw_usem_try_lock(&stor->part_sem, RW_SEM_WRITE)
#define parts_wunlock(stor)     rw_usem_unlock(&stor->part_sem, RW_SEM_WRITE)

/* storage flags */
enum {
  STORAGE_RDONLY = 0x01,
  /* all block devices be part of the
   * storage support command aborting
   */
  STORAGE_SUPP_CMD_ABORT = 0x02,
  STORAGE_INVALIDATED = 0x04
};

/* partition flags */
enum {
  PART_RDONLY = 0x1,
  PART_HIDDEN = 0x2
};

struct __genstor;

typedef struct __bd_part {
  uint16_t num;
  uint8_t id;
  uint16_t flags;
  sector_t start_sector;
  sector_t nr_sectors;
#ifdef CONFIG_BLOCK_STATS
  block_stats_t stats;
#endif
  void *private;              /* private partition data */
} bd_part_t;

struct __storage_operations;

typedef struct __genstor {
  char name[STOR_NAME_MAXLEN];
  sector_t capacity;
  /* device's partitions */
  bd_part_t **parts;
  uint16_t nr_parts;
  int32_t openers;            /* number of clients opened this storage */
  uint32_t part_opened;       /* times that storage's partitions was opened */
  int sect_shift;
  uint32_t flags;
  /*
   * partition semaphore, holded on read at work I/O operations with
   * partitions, and on write at partitions rescanning
   */
  rw_usem_t part_sem;
  struct __storage_operations *ops;
#ifdef CONFIG_BLOCK_STATS
  block_stats_t stats;
#endif
  uint32_t nr_active_reqs;
  int32_t cache_id;
  uint32_t curr_cache_size;
  uint32_t max_cache_size;
  void *private;              /* typically contains pointer to its request queue */
} genstor_t;

typedef struct __storage_operations {
  int (*open)(genstor_t*, uint16_t partno, uint16_t flags);
  int (*close)(genstor_t*, uint16_t partno);
  int (*scan_partitions)(genstor_t*);
  ssize_t (*read)(genstor_t*, struct __bio_rpc *bio_rpc, biorq_t *srv_biorq);
  ssize_t (*write)(genstor_t*, struct __bio_rpc *bio_rpc, biorq_t *srv_biorq);
  ssize_t (*ioctl)(genstor_t*, uint16_t partno, uint16_t cmd, void *buf,
                   size_t data_size, size_t buf_size);
  int (*sync)(genstor_t*, bd_part_t*, size_t offset, size_t size);
  int (*make_request)(genstor_t*, bd_part_t*, biorq_t*, biovec_t*, short prio);
} storage_operations_t;

int init_generic_storage(genstor_t *storage, char *name,
                         storage_operations_t *ops,
                         blkreq_queue_t *q, bool cacheable);
void release_generic_storage(genstor_t *stor);

/* default storage operations */
int generic_scan_partitions(genstor_t *storage);
int generic_open_storage(genstor_t *storage, uint16_t partno, uint16_t flags);
int generic_close_storage(genstor_t *storage, uint16_t partno);
ssize_t generic_read_storage(genstor_t *storage, struct __bio_rpc *bio_rpc,
                         biorq_t *srv_biorq);
ssize_t generic_write_storage(genstor_t *storage, struct __bio_rpc *bio_rpc,
                          biorq_t *srv_biorq);
ssize_t generic_ioctl_storage(genstor_t *storage, uint16_t partno, uint16_t cmd, void *buf,
                              size_t data_size, size_t buf_size);
int generic_sync_storage(genstor_t *storage, bd_part_t *part, size_t offset, size_t size);
int generic_make_request(genstor_t *storage, bd_part_t *part, biorq_t *bio,
                         biovec_t *bvec, short prio);

int generic_submit_blkio(blkreq_queue_t *q, blkio_t *blkio, short prio);

static inline bool check_part_exist(genstor_t *stor, uint16_t partno)
{
  return ((partno <= stor->nr_parts) &&
          (stor->parts[partno - 1] != NULL));
}

#endif
