/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 *
 * include/block/request.h: block requests; large part is adopted from
 *                          the Linux block layer
 */

/*
 * NOTE:
 *    If a request is double-end, it means a non full sector write request
 *    is processed. In this case  at first full sector is read, and then the
 *    need part is modified. The block layer must provide completion method
 *    to submit sector after modification
 */

/*
 * TO DO [dg]: replace event locked queue by prioritized queue
 */

#ifndef __REQUEST_H__
#define __REQUEST_H__

#include <sys/types.h>
#include <datalib/list.h>
#include <queue.h>
#include <unistd.h>
#include <timeout.h>
#include "defs.h"
#include "blkdev.h"
#include "config.h"

#define blk_queue_plugged(q)          ((q)->flags & (1 << BLKQUEUE_PLUGGED))
#define blk_queue_scheduled(q)        atomic_test_bit(&(q)->flags, BLKQUEUE_IOSCHED)
#define blk_queue_not_pipelined(q)    ((q)->flags & (1 << BLKQUEUE_NOT_PIPELINED))

#define set_blk_queue_plugged(q)      ((q)->flags |= (1 << BLKQUEUE_PLUGGED))
#define clear_blk_queue_plugged(q)    ((q)->flags &= ~(1 << BLKQUEUE_PLUGGED))
#define set_blk_queue_scheduled(q)    atomic_test_and_set_bit(&(q)->flags, BLKQUEUE_IOSCHED)
#define clear_blk_queue_scheduled(q)  atomic_test_and_reset_bit(&(q)->flags, BLKQUEUE_IOSCHED)

#define blk_group_wait(g)             atomic_test_bit(&(g)->flags, BLKGROUP_WAIT)

#define set_blk_group_wait(g)         atomic_test_and_set_bit(&(g)->flags, BLKGROUP_WAIT)
#define clear_blk_group_wait(g)       atomic_test_and_reset_bit(&(g)->flags, BLKGROUP_WAIT)

#define biorq_sync(biorq)           ((biorq)->flags & BIORQ_SYNC)
#define biorq_not_scheduled(biorq)  ((biorq)->flags & BIORQ_NOT_SCHEDULE)
#define biorq_double_end(biorq)     ((biorq)->flags & BIORQ_DOUBLE_END)
#define biorq_data_available(biorq) ((biorq)->flags & BIORQ_DATA_AVAILABLE)
#define biorq_buffered(biorq)       ((biorq)->flags & BIORQ_BUFFERED)
#define biorq_need_reply(biorq)     ((biorq)->flags & BIORQ_NEED_REPLY)
#define biorq_aborted(biorq)        ((biorq)->flags & BIORQ_ABORTED)
#define biorq_serv_prepared(biorq)  ((biorq)->flags & BIORQ_SERV_PREPARED)

#define biorq_type_fs(biorq)        ((biorq)->type == BIORQ_TYPE_FS)
#define biorq_type_server(biorq)    ((biorq)->type == BIORQ_TYPE_SERVER)
#define biorq_type_user(biorq)      ((biorq)->type == BIORQ_TYPE_USER)

#define blk_queued_rq(rq)           ((rq)->flags & BLKREQ_QUEUED)
#define blk_scheduled_rq(rq)        ((rq)->flags & BLKREQ_SCHEDULED)
#define blk_launched_rq(rq)         ((rq)->flags & (BLKREQ_RUN | BLKREQ_MUST_RUN))
#define blk_inactive_rq(rq)         ((rq)->flags & BLKREQ_INACTIVE)
#define blk_barrier_rq(rq)          ((rq)->flags & BLKREQ_BARRIER)
#define blk_rq_nomerge(rq)          ((rq)->flags & BLKREQ_NOMERGE)
#define blk_preflush_rq(rq)         ((rq)->flags & BLKREQ_PREFLUSH)
#define blk_postflush_rq(rq)        ((rq)->flags & BLKREQ_POSTFLUSH)
#define blk_drv_request(rq)         ((rq)->flags & BLKREQ_DRV)

#define BLKREQ_LOCKED (BITS_PER_LONG - 1)

#define blk_queue_lock(q)    pthread_mutex_lock(&(q)->mutex)
#define blk_queue_unlock(q)  pthread_mutex_unlock(&(q)->mutex)


/* Fixme [dg]: can be changed by latch? */
#define blk_group_rlock(g)    pthread_mutex_lock(&(g)->mutex)
#define blk_group_runlock(g)  pthread_mutex_unlock(&(g)->mutex)
#define blk_group_wlock(g)    pthread_mutex_lock(&(g)->mutex)
#define blk_group_wunlock(g)  pthread_mutex_unlock(&(g)->mutex)

#define blk_rq_lock(rq)       atomic_test_and_set_bit(&(rq)->flags, BLKREQ_LOCKED)
#define blk_rq_unlock(rq)     atomic_test_and_reset_bit(&(rq)->flags, BLKREQ_LOCKED)

/* block device request flags */
enum {
  BLKREQ_RW             = 0x1,    /* transfer direction: 0 - read, 1 - write */
  BLKREQ_QUEUED         = 0x2,    /* request is put in the driver I/O queue */
  BLKREQ_SCHEDULED      = 0x4,
  BLKREQ_RUN            = 0x8,    /* request is being processed by the driver */
  /*
   * the request is must be processed by the driver as soon as
   * better and must not be more scheduled
   */
  BLKREQ_MUST_RUN       = 0x10,
  BLKREQ_INACTIVE       = 0x20,
  /* request barrier, i.e any requests can't be put in the queue before it */
  BLKREQ_BARRIER        = 0x40,
  BLKREQ_NOMERGE        = 0x80,   /* request can't be merged */
  /* the hardware cache must be flushed before process this request */
  BLKREQ_PREFLUSH       = 0x100,
  /* the hardware cache must be flushed after this request completion */
  BLKREQ_POSTFLUSH      = 0x200,
  BLKREQ_DRV            = 0x400   /* this request is owned by driver */
  /* all other flags are specific for driver or block scheduler */
};

/* request types */
typedef enum __blkerq_type {
  REQ_TYPE_IO = 1,           /* client I/O request */
  REQ_TYPE_FLUSH,            /* flush request */
  /* all other types are driver specific */
} blkreq_type_t;

/* request queue flag bits */
enum {
  BLKQUEUE_PLUGGED,
  BLKQUEUE_IOSCHED,
  BLKQUEUE_NOT_PIPELINED,
  /* all other flags are driver specific */
};

/* block group flag bits */
enum {
  BLKGROUP_WAIT
};

/* file system request flags */
enum {
  BIORQ_ASYNC        = 0x10,
  BIORQ_NOT_SCHEDULE = 0x20,
  BIORQ_RT           = 0x30,
  BIORQ_ORDERED      = 0x40
};

/* private biorq flags */
enum {
  BIORQ_RW              = 0x10000,
  BIORQ_ORIG_RW         = 0x20000,
  /* means that sectors must be written back after reading and modification */
  BIORQ_DOUBLE_END      = 0x40000,
  BIORQ_DATA_AVAILABLE  = 0x80000,
  BIORQ_BUFFERED        = 0x100000,   /* data are already read from IPC object */
  BIORQ_RDONLY_BUFFER   = 0x200000,   /* the private buffer is read-only */
  BIORQ_NEED_REPLY      = 0x400000,
  BIORQ_SERV_PREPARED   = 0x800000,
  BIORQ_ABORTED         = 0x1000000
};

/* insert queue places */
enum {
  QUEUE_HEAD,
  QUEUE_TAIL
};

/* request split types */
typedef enum __blk_split_type {
  BLK_SPLIT_FRONT,
  BLK_SPLIT_TAIL
} blk_split_type_t;

typedef enum __biorq_type {
  BIORQ_TYPE_SERVER,
  BIORQ_TYPE_FS,
  BIORQ_TYPE_USER
} biorq_type_t;

struct __blkreq_queue;
struct __blkio;

/* writeback chunk of a not sector aligned request */
struct __wb_chunk {
  uint32_t off;
  uint32_t size;
};

typedef struct __biovec {
  sector_t start_sector;
  uint32_t nr_sectors;
} biovec_t;

typedef struct __biorq {
  sector_t start_sector;
  uint32_t nr_sectors;
  uint32_t flags;
  biorq_type_t type;
  union {
    uint32_t complete_sectors;
    struct __wb_chunk wb_chunk;
  };
  union {
    void *data_buf;       /* pointer to a data buffer */
    ulong_t data_off;     /* data offset in received packed */
  };
  atomic_t refcnt;
  void (*complete) (struct __blkio*);
  void *completion_data;
  union {
    void *server_private;
    pid_t owner;
  };
} biorq_t;

struct blkio_status {
  int16_t err:15,             /* error code */
      buffered:1;
} __attribute__ ((packed));

/* block I/O request */
typedef struct __blkio {
  sector_t start_sector;
  uint32_t nr_sectors;
  struct blkio_status status;
  void *buf;
  list_node_t node;
  biorq_t *biorq;
} blkio_t;

/* generic block device request */
typedef struct __blk_request {
  sector_t start_sector;
  uint32_t nr_sectors;
  uint32_t complete_sectors;
  short prio;             /* request priority */
  short type;             /* request type */
  ulong_t flags;
  atomic_t refcnt;
  struct __blkreq_queue *queue;
  short io_errors;        /* number of I/O errors occurred during processing of this request */
  union {
    void *sched_data;
    void *drv_data;
  };
  list_head_t blkios;     /* list of I/O requests associated with this request */
  blkio_t *current_blkio;
  list_node_t node;
  void (*complete) (struct __blk_request *rq);
  void *completion_data;
} blk_request_t;

struct blk_group {
  list_head_t queues;
  pthread_mutex_t mutex;
  event_t not_empty;
  ulong_t flags;
  struct __blkreq_queue *last_serv_queue;
  ulong_t filled_queues;    /* mask of not empty queues */
  ulong_t free_indexes;
  ulong_t flush_mask;       /* mask of queues need to be flushed */
  void *private;
};

typedef struct __blkreq_queue {
  prio_queue_t requests;
  struct __iosched *scheduler;
  int max_requests;
  int nr_requests;
  int nr_reqs_cong;            /* request threshold above that queue is considered as congested */
  int nr_reqs_wasted;          /* request threshold below that queue is considered as wasted */
  uint64_t done_requests;      /* total number of done requests */
  uint64_t ioerr_reqs;         /* total number of requests finished by an I/O error */
  int max_sectors;             /* max sectors per single request  */
  struct __block_device *bdev; /* block device serviced by this queue */
  blk_request_t *last_proc_rq;
  blk_request_t *last_made_rq;
  ulong_t flags;
  struct __genstor *stor;      /* storage which I/O request is dispatched to this queue */
  pthread_mutex_t mutex;
  struct blk_group *group;
  int index;                   /* index  into the block group */
  event_t done_all;            /* event to signal that all requests are processed by the driver */
  atomic_t refcnt;
  struct __blk_rq_trace *rq_trace;
  list_node_t node;            /* group node */

  /* queue specific methods */

  /* I/O request strategy routine */
  int (*io_request_fn) (struct __blkreq_queue*, blkio_t*, short prio);
  /* add request in the queue */
  int (*dispatch_request) (blk_request_t *rq, int where);
  int (*drop_request) (blk_request_t *rq);
  /* methods to merge queued requests */
  bool (*merge_requests) (blk_request_t *rq_front, blk_request_t *rq_tail);
  uint32_t (*merge_rq) (blk_request_t *rq, blkio_t *bio, int where);
} blkreq_queue_t;

int dispatch_request(blk_request_t *rq, int where);
/* dispatch request to the queue without locking event signaling */
int dispatch_request_nolock_ev(blk_request_t *rq, int where);
int drop_request(blk_request_t *rq);
int drop_request_nolock(blk_request_t *rq);
blk_request_t *get_next_blkrq(blkreq_queue_t *queue);
blk_request_t *flush_get_group_blkrq(struct blk_group *group, ulong_t qmask);
blk_request_t *get_group_blkrq_wait(struct blk_group *group, ulong_t qmask, ulong_t rec_mask);
int blk_group_wait_event(struct blk_group *group);

blk_request_t *alloc_init_blkreq(blkreq_queue_t *q, short type, sector_t start,
                                 uint32_t nr_sectors, bool rw);
biorq_t *alloc_init_biorq(sector_t start_sector, uint32_t nr_sects, biorq_type_t type, uint32_t flags);
blkio_t *alloc_init_blkio(biorq_t *biorq, sector_t sector, uint32_t nr_sects, void *buf);
int init_blk_queue(blkreq_queue_t *queue, struct __genstor *stor, block_device_t *bdev,
                   int max_requests, int max_sectors, struct blk_group *group);
int init_blk_group(struct blk_group *group);
void blk_queue_signal(blkreq_queue_t *q);
void destroy_blk_group(struct blk_group *group);
int add_queue_to_group(struct blk_group *group, blkreq_queue_t *q);
void remove_queue_from_group(blkreq_queue_t *q);
void destroy_blk_queue(blkreq_queue_t *queue);
void set_blk_queue_event(blkreq_queue_t *queue);

void put_blk_queue(blkreq_queue_t *queue);
void put_blk_request(blk_request_t *rq);
void put_biorq(biorq_t *bio);
void destroy_blkio(blkio_t *b);
bool blkio_split_ok(blkio_t *b);
blkio_t *split_blkio(blkio_t *blkio, uint32_t nr_sectors, int sect_shift,
                     blk_split_type_t where);
biorq_t *check_clusterize_biorq(biorq_t *b, size_t offset, size_t size,
                                uint32_t sect_size, blk_split_type_t where);
/*
 * combine two blkios in blkio passed through the 'b1' parameter
 * it dont check if they are mergeable
 */
void _combine_blkio(blkio_t *b1, blkio_t *b2);
void default_blk_rq_complete(blk_request_t *rq);
void default_complete_biorq(blkio_t *blkio);

static int inline blkreq_rw(blk_request_t *rq)
{
  return (rq->flags & BLKREQ_RW) != 0;
}

static int inline biorq_rw(biorq_t *biorq)
{
  return (biorq->flags & BIORQ_RW) != 0;
}

static int inline biorq_orig_rw(biorq_t *biorq)
{
  return (biorq->flags & BIORQ_ORIG_RW) != 0;
}

static inline void set_rq_flags(blk_request_t *rq, uint32_t flags)
{
  rq->flags |= flags;
}

static inline void clear_rq_flags(blk_request_t *rq, uint32_t flags)
{
  rq->flags &= ~flags;
}

static inline void set_biorq_flags(biorq_t *biorq, uint32_t flags)
{
  biorq->flags |= flags;
}

static inline void clear_biorq_flags(biorq_t *biorq, uint32_t flags)
{
  biorq->flags &= ~flags;
}

static inline void get_blk_queue(blkreq_queue_t *queue)
{
  atomic_inc(&queue->refcnt);
}


static inline void get_blk_request(blk_request_t *rq)
{
  atomic_inc(&rq->refcnt);
}

static inline void get_biorq(biorq_t *biorq)
{
  atomic_inc(&biorq->refcnt);
}

#endif
