/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org> 
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 * (c) Copyright 2008 Paul Graphov <graphov@jarios.org>
 *
 * Character device server layer 
 * 
 */

#ifndef __CHARDEV_H__
#define __CHARDEV_H__

#include <sys/types.h>
#include <datalib/usrtc.h>
#include <devio/devio.h>
#include <ipc/port.h>
#include <pthread.h>
#include <sys/uevent.h>

#define NOLOG  -1            
#define ERROR   0
#define WARNING 1
#define INFO    2

#define DIRECTIO_WRITE    0x1
#define DIRECTIO_READ     0x2
#define EVENTS_POLL       0x4
#define HW_READ           0x8
#define HW_WRITE          0x10
#define MULTIEVENTS_POLL  0x20
#define DIRECTIO_SEEK     0x40
#define HW_SEEK           0x80

#define NO_PREFETCH       0
#define QUEUE_FIFO        1
#define QUEUE_STACK       2
#define QUEUE_CUSTOM      3

#define NONBLOCKING_WRITE  0x1

#define IN  0
#define OUT 1
#define CMD 2

#define SUSPEND -1000000

#define IN_QUEUE_MASK     14
#define OUT_QUEUE_MASK    112

#define IN_QUEUE_NO_PREFETCH (NO_PREFETCH << 1)
#define IN_QUEUE_FIFO        (QUEUE_FIFO << 1)
#define IN_QUEUE_STACK       (QUEUE_STACK << 1)
#define IN_QUEUE_CUSTOM      (QUEUE_CUSTOM << 1)
#define OUT_QUEUE_NO_PREFETCH (NO_PREFETCH << 4)
#define OUT_QUEUE_FIFO        (QUEUE_FIFO << 4)
#define OUT_QUEUE_STACK       (QUEUE_STACK << 4)
#define OUT_QUEUE_CUSTOM      (QUEUE_CUSTOM << 4)

struct device_ops {
  status_t (*read)(int devno, void *buf, size_t size, size_t offset, void **interim);
  status_t (*write)(int devno, void *buf, size_t size, size_t offset, void **interim);
  status_t (*ioctl)(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize);
  void (*log)(const char *buf);
};

typedef struct __queue_ops {
  void* (*get)(void **q);
  status_t (*put)(void **q, void *ent);
  status_t (*init)(void **q);
  void (*destroy)(void **q);
} queue_ops_t;
  
struct dev_node_t {
  devid_t devno;
  const char *devpath;
};

typedef struct __device_server_t {
  pid_t pid; /* device server pid */
  const char *drvname;
  int loglevel;
  pid_t hw_pid; /* hardware server pid */
  prtid_t cmd_port;
  prtid_t io_port;
  void *cmdbuf;
  size_t cmdbuf_size;
  size_t buf_size[2];
  uint16_t cap; /* capatibility flags */
  int num_devices;
  struct device_ops *ops_tbl;
  int num_threads;
} device_server_t;

/* functions */
/*
  int init_multi_device_server(device_server_t *srv, size_t in, size_t out, size_t cmd, const char *drvname,
                             struct device_ops *ops_tbl, int max_clients, struct dev_node_t **mult_nodes,
                             int mult_nnodes, int num_threads);*/

int init_multi_device_server(device_server_t *srv, size_t in, size_t out, const char *drvname,
                             struct device_ops *ops_tbl, int num_threads);

int init_simple_device_server(size_t in, size_t out, const char *drvname,
                              const char *devname, struct device_ops *ops_tbl);

int init_simple_mt_device_server(size_t in, size_t out, const char *drvname,
                                 const char *devname, struct device_ops *ops_tbl, int num_threads);

int add_device(int devno, const char *path);

void device_server_start(void);
void set_loglevel(int loglevel);

pid_t get_sender_pid(void);
tid_t get_sender_tid(void);
uid_t get_sender_uid(void);
uid_t get_sender_gid(void);


void wake_device(int devno, int dir);

#endif /* __DEVIO_SERVER_H__ */
