/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * File system interaction with block devices
 */

#ifndef __BDIO_H__
#define __BDIO_H__

#include <stdbool.h>
#include <sys/types.h>
#include <block/defs.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <block/bdev_ioctl.h>
#include <ipc/channel.h>
#include <block/request.h>
#include <block/config.h>

#define read_blkdev_cached_sectors(bdio, mobj_id, mheads, nr_heads, prio, flags) \
        rw_blkdev_cached_sectors(0, bdio, mobj_id, mheads, nr_heads, prio, flag)
#define write_blkdev_cached_sectors(bdio, mobj_id, mheads, nr_heads, prio, flags) \
        rw_blkdev_cached_sectors(1, bdio, mobj_id, mheads, nr_heads, prio, flag)

#define MAX_RW_PRIORITY (BLKDEV_NR_PRIORITIES - 1)

enum {
  BLK_ERR = 0xFFFFFFFF
};

/* Block device I/O descriptor */
struct bdio {
  int fd;
  pid_t bd_pid;                /* block device server's pid */
  pid_t fs_pid;
  sector_t bd_capacity;
  int hw_sect_bits;
  bool rdonly;
  chid_t cmd_chan;          /* command channel of the block device server */
  prtid_t async_port;       /* asyncrounous port to notify file system about I/O completion */
  uint16_t default_flags;
  int blk_bits;             /* pow 2 of the file system I/O logical block size */
};

/* memory object head */
struct cache_head {
  uint32_t page;
  uint32_t nr_pages;
  sector_t start_sects[0];
};

/* cache I/O descriptor */
struct cio_desc {
  struct cache_head *heads;
  uint16_t blocks_per_page;
  uint16_t nr_heads;
  uint32_t heads_size;
  uint16_t *nr_proc_heads;
};

/*
 * Open block device and get most I/O data
 * On success return 0, negated error code on fail
 */
int open_blkdev(const char *path, struct bdio *bdio, bool sync);
void close_blkdev(struct bdio *bdio);

/*
 * Read/write block device sectors into/from memory object
 *
 * @dir: direction; 0 - read, 1 - write
 * @bdio: block device I/O descriptor
 * @cd: cache I/O descriptor
 * @prio: priority of the operation
 * @flags: I/O request flags
 *
 * Desciption:
 *      Initializes read block device sectors into memory object pages.
 *   The operation can be either synchrounous or asyncronous depending
 *   on flags. In case of syncrounous operation after function return data
 *   will be already in memory object pages. In the other hands if the
 *   'BIORQ_ASYNC' flag is specified, after function return data read will
 *   be still in progress. The block device will notify about completion
 *   sending message on the port which id match the 'async port' of the 'bdio'
 *   descriptor (if it was earlier set).
 *
 * Returned value:
 *    Positive value meaning number of processed heads on success
 *    Negated error code on fail:
 *      EACCES - access denied
 *      EBADF - invalid fail descriptor
 *      EINVAL - invalid parameter is passed
 *      ENOMEM - server can't allocate need amount of memory for
 *               internal structures and buffer to complete the operation
 *      EIO - Input/output error
 *      EBUSY - block device server can't process now the request
 */
int32_t rw_blkdev_cached_sectors(int dir, struct bdio *bdio, struct cio_desc *cd,
                                 short prio, uint16_t flags);

/*
 * Read file system logical blocks from block device
 * On success return number of read blocks, on error return
 * BLK_ERR, and 'errno' contains a respective error code
 */
static inline uint32_t read_blkdev_blocks(struct bdio *bdio, uint64_t block,
                                          uint32_t nr_blocks, void *buf);
/* Write file system logical blocks to block device */
static inline uint32_t write_blkdev_blocks(struct bdio *bdio, uint64_t block,
                                           uint32_t nr_blocks, void *buf);
int set_blkdev_async_port(struct bdio *bdio, prtid_t port);
int sync_blkdev(struct bdio *bdio);
/* flush hardware block device cache */
static inline int hw_flush_blkdev(struct bdio *bdio);


static inline uint32_t read_blkdev_blocks(struct bdio *bdio, uint64_t block,
                                          uint32_t nr_blocks, void *buf)
{
  ssize_t ret;

  ret = pread(bdio->fd, buf, nr_blocks << bdio->blk_bits, block << bdio->blk_bits);

  return (ret == -1) ? BLK_ERR : (ret >> bdio->blk_bits);
}

static inline uint32_t write_blkdev_blocks(struct bdio *bdio, uint64_t block,
                                           uint32_t nr_blocks, void *buf)
{
  ssize_t ret;

  ret = pwrite(bdio->fd, buf, nr_blocks << bdio->blk_bits, block << bdio->blk_bits);

  return (ret == -1) ? BLK_ERR : (ret >> bdio->blk_bits);
}

static inline struct cache_head *cache_head_iterate(struct cache_head *head,
                                                    uint16_t blocks_per_page)
{
  return (struct cache_head*)((ulong_t)head + sizeof(struct cache_head) +
                     blocks_per_page * head->nr_pages * sizeof(sector_t));
}

static inline int hw_flush_blkdev(struct bdio *bdio)
{
  return ioctl(bdio->fd, BLKIOC_FLUSH);
}

#endif /* __BDIO_H__ */
