/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org> 
 *
 * vnode structure operations 
 * 
 */

#ifndef __FSIO_VNODE_H__
#define __FSIO_VNODE_H__

#include <sys/types.h>
#include <sys/stat.h>
#include <wchar.h>
#include <datalib/usrtc.h>
#include <pthread.h>

typedef size_t uuid_t;
typedef size_t hash_t;

#define VNODE_NORM    0x1
#define VNODE_ORPHAN  0x2
#define VNODE_TMP     0x3

#define RD_DIRECTION  0
#define WR_DIRECTION  1

typedef struct __vnode_t {
  uuid_t id; /* unique id */
  hash_t hash; /* hash for fast first stage resolving */
  int ref_count; /* reference count */
  int state; /* state of vnode */
  struct vnode_ops *ops; /* operations */
  file_type_t type; /* type */
  pthread_mutex_t vn_lock;
  void *spec_obj; /* specific file system implementation structure */
  void *mem_obj; /* memory object of the file */
  usrtc_node_t *pnode; /* parent node */
} vnode_t; /* virtual node, higher fs abstraction */

struct vnode_ops {
  int (*vn_open)(vnode_t *s,int,int); /* vnode, flags, mode */
  int (*vn_close)(vnode_t *s);
  int (*vn_rdwr)(vnode_t *s,void*,size_t*,off_t,int);
  int (*vn_ioctl)(vnode_t *s,void*,size_t,size_t);
  int (*vn_select)(vnode_t *s);
  int (*vn_getattr)(vnode_t *s);
  int (*vn_setattr)(vnode_t *s);
  int (*vn_access)(vnode_t *s,gid_t gid,uid_t uid);
  int (*vn_fmap)(vnode_t *s,size_t size,off_t offset,int flags);
  /*  int (*vn_lookup)(vnode_t *s,char *name,int *pp,int amount,int flags,int mode,gid_t gid,uid_t uid); */
  int (*vn_lookup)(vnode_t *,char *,int *,int,int,int,gid_t,uid_t);
  int (*vn_create)(vnode_t*,char*,int*,int,int,gid_t,uid_t); /* last int - mode */
  int (*vn_remove)(vnode_t*);
  int (*vn_link)(vnode_t*);
  int (*vn_rename)(vnode_t*);
  int (*vn_mkdir)(vnode_t*,char*,int*,int,int,gid_t,uid_t); /* path, parsed points, amount, root value, mode */
  int (*vn_rmdir)(vnode_t*);
  int (*vn_readdir)(vnode_t*,void*,char*);
  int (*vn_symlink)(vnode_t*);
  int (*vn_readlink)(vnode_t*);
  int (*vn_fsync)(vnode_t*);
  int (*vn_unref)(vnode_t*);
  int (*vn_mount)(vnode_t*);
};

/* generic functions */


/* general functions of vnode subsystem */
/*
 * vnode_root_init() init root vnode 
 * returns 0 if all is ok, or positive 
 * value indicating to error
 */
int vnode_root_init(struct vnode_ops *ops,void *spec_obj);

/*
 * vnode_init - initialize vnode
 */
int vnode_init(vnode_t *svn,void *spec_obj);

/*
 * add reference to vnode 
 */
int vnode_addref(vnode_t *svn);

/*
 * remove reference from vnode
 */
int vnode_rmref(vnode_t *svn);

/*
 * insert vnode to cache
 */
int vnode_insert(vnode_t *svn,hash_t hash,usrtc_node_t *node);

int vnode_mount(void);

/*
 * vnode_hash
 */
hash_t vnode_hash(const char *name);

/*
 * vnode generic operations 
 */
/**
 * @fn int vnode_open(char *path,int *p,int total,int root,int flags,int mode)
 * Generic vnode open handler, it's tries to open a file in a common way, 
 * using fs-specific ops
 * 
 * @note it can preprocess some datas before calling fs-specifics to avoid 
 * time loss
 *
 * @param path - pointer to the requested path
 * @param p - pointer to the parsed pointers
 * @param total - amount of parsed pointers
 * @param root - where the root for this fs
 * @param flags - flags for open
 * @param mode - mode for a file 
 * @param uid - UID of the caller
 * @param gid - GID of the caller
 *
 * @return 0, or errno value
 */
vnode_t *vnode_open(char *path,int *p,int total,int root,
		    int flags,int mode,uid_t uid,gid_t gid);

#endif /* __FSIO_VNODE_H__ */
