/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *  
 * Fast hashing routine for ints,  longs and pointers.
 *  (C) 2002 William Lee Irwin III, IBM  
 *
 * Ported by Dmitry Gromada <gromada@jarios.org>
 *  shifts are fully replaced by multiplication in the hash_64
 *  function, because on most modern 32-bit platfoms it
 *  will not faster, and will slower on 64-bit platforms
 */

/*
 * Knuth recommends primes in approximately golden ratio to the maximum
 * integer representable by a machine word for multiplicative hashing.
 * Chuck Lever verified the effectiveness of this technique:
 * http://www.citi.umich.edu/techreports/reports/citi-tr-00-1.pdf
 *
 * These primes are chosen to be bit-sparse, that is operations on
 * them can use shifts and additions instead of multiplications for
 * machines where multiplications are slow.
 */

#ifndef __HASH_H__
#define __HASH_H__

#include <sys/types.h>

/* 2^31 + 2^29 - 2^25 + 2^22 - 2^19 - 2^16 + 1 */
#define GOLDEN_RATIO_PRIME_32 0x9e370001UL
/*  2^63 + 2^61 - 2^57 + 2^54 - 2^51 - 2^18 + 1 */
#define GOLDEN_RATIO_PRIME_64 0x9e37fffffffc0001UL

#if BITS_PER_LONG == 32
#define GOLDEN_RATIO_PRIME GOLDEN_RATIO_PRIME_32
#define hash_long(val, bits) hash_32(val, bits)
#elif BITS_PER_LONG == 64
#define hash_long(val, bits) hash_64(val, bits)
#define GOLDEN_RATIO_PRIME GOLDEN_RATIO_PRIME_64
#else
#error Wordsize not 32 or 64
#endif

static inline uint64_t hash_64(uint64_t val, unsigned int bits)
{
  uint64_t hash = val * GOLDEN_RATIO_PRIME_64;

  /* High bits are more random, so use them. */
  return hash >> (64 - bits);
}

static inline uint32_t hash_32(uint32_t val, unsigned int bits)
{
  /* On some cpus multiply is faster, on others gcc will do shifts */
  uint32_t hash = val * GOLDEN_RATIO_PRIME_32;

  /* High bits are more random, so use them. */
  return hash >> (32 - bits);
}

static inline unsigned long hash_ptr(void *ptr, unsigned int bits)
{
  return hash_long((unsigned long)ptr, bits);
}

/* `djb2' by Dan Bernstein */

static inline uint32_t djb2_hash_string(const char *str, unsigned int len)
{
  unsigned long hash = 5381;
  const char *str_ptr = str;

  while (len --)
    hash = ((hash << 5) + hash) + *str_ptr++;

  return hash;
}

/* `hashpjw' function by P.J. Weinberger */

static inline uint32_t pjw_hash_string(const char *str, unsigned int len)
{
  unsigned long int hval, g;
  const char *str_ptr = str;

  /* Compute the hash value for the given string.  */
  hval = 0;
  while (len --) {
    hval = (hval << 4) + (unsigned long) *str_ptr++;
    g = hval & ((unsigned long) 0xf << (32 - 4));
      if (g != 0) {
        hval ^= g >> (32 - 8);
        hval ^= g;
      }
  }

  return hval;
}

#endif /* __HASH_H__ */
