/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Paul Graphov <graphov@jarios.org>
 *
 * spinlock.c: spinlock implementation.
 *
 * Be careful, this mutual exclusion mechanism can cause priority
 * inversion and must NOT be used in anything real-time relate.
 * Also it is not reentrant.
 *
 */
#ifndef __SPINLOCK_H__
#define __SPINLOCK_H__

#include <stdbool.h>
#include <sys/types.h>
#include <arch/atomic.h>
#include <sched.h>


typedef volatile long spinlock_t;

/**
 * Initializer for static spinlocks
 */
#define SPINLOCK_INITIALIZER 0

/**
 * Initialize the a spinlock.
 *
 * @param[in] lock Pointer to spinlock to initialize.
 * @param[in] arg Ignored, for pthread_mutex compatibility
 */
static inline void spinlock_init(spinlock_t *lock, void *arg)
{
  
  *lock = 0;
}

/**
 * Acquire the a spinlock.
 *
 * @param[in] lock Pointer to spinlock to acquire.
 */
static inline void spinlock_lock(spinlock_t *lock)
{
  while(atomic_swap_long((ulong_t *)lock, 1) == 1) {
    sched_yield();
  }
}

/**
 * Release a spinlock.
 *
 * @param[in] lock Pointer to spinlock to release.
 */
static inline void spinlock_unlock(spinlock_t *lock)
{
  *lock = 0;
}

/**
 * Try to acquire a spinlock.
 *
 * @param[in] lock Pointer to spinlock to acquire.
 * @returns Zero on success, non-zero on failure.
 */
static inline int spinlock_trylock(spinlock_t *lock)
{
  return atomic_swap_long((ulong_t *)lock, 1);
}

/**
 * Does nothing, for pthread_mutex compatibility.
 *
 * @param[in] lock Ignored.
 */
static inline void spinlock_destroy(spinlock_t *lock)
{
  return;
}

#endif
