/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * session (resources) stuff 
 */

#ifndef __V2_SESSION_H__
#define __V2_SESSION_H__

#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <stdarg.h>
#include <dirent.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <pthread.h>
#include <dirent.h>
#include <datalib/usrtc.h>

#include <vfs/vfsdef.h>
#include <vfs/vnode.h>

#define PFS_DEFAULT_OPEN  32
#define PFS_MAX_OPEN      256

#define MAX_FSESSID  65500

#define EOESSID  65535
#define EMESSID  65534

typedef struct __pidsession_t {
  pid_t pid;
  uid_t uid;
  gid_t gid;
  uint32_t cap;
  uint16_t max_open;
  uint16_t active;
  uint16_t n_reset;
  /* bitmap for essid allocation */
  uint8_t along;
  unsigned long *bm;
  /* locking */
  pthread_mutex_t lock;
  /* tree */
  usrtc_t tree;
} pidsession_t;

typedef struct __fessid_t {
  uint16_t id;
  int mode;
  int flags;
  off_t offset;
  vnode_t *vnode;
} fessid_t;

/**
 * @struct struct fss_hooks
 * @desc used as table of hooks on determined events,
 *  h_add - called when new resource session added 
 *  h_close - called when resource session was closed 
 *  h_kill - called when all resources of task killed
 *  h_clone - called when all resource are cloned for task,
 *  to the new task ... 
 */
struct fss_hooks {
  int (*h_add)(pidsession_t *,fessid_t *);
  int (*h_close)(pidsession_t *,fessid_t *);
  int (*h_kill)(pidsession_t *);
  int (*h_clone)(pidsession_t *,pid_t);
};

/**
 * there defines are useful for locking 
 * within generic functions 
 */
#define pidsession_lock(n)  pthread_mutex_lock(&n->lock)
#define pidsession_unlock(n)  pthread_mutex_unlock(&n->lock)

/**
 * @fn int fs_session_init(void)
 * @desc This function init internal structure for 
 * filesystem that will serve all sessions
 */
int fs_session_init(void);

/**
 * @fn int fs_session_sethooks(struct fss_hooks *nhooks);
 * @desc This function sets hooks functions table for 
 * sessions layer , if nhooks NULL will return EINVAL, 
 * 0 otherwise
 */
int fs_session_sethooks(struct fss_hooks *nhooks);

/**
 * @fn pidsession_t *pidsession_create(pid_t pid,uid_t uid,gid_t gid,int max_open)
 * @desc This function create a new session for task within one fs
 * it returns a pointer to the newly allocated and inserted structure
 */
pidsession_t *pidsession_create(pid_t pid,uid_t uid,gid_t gid,int max_open);

/**
 * @fn pidsession_t *pidsession_lookup(pid_t pid)
 * @desc This function looking up for a task session on filesystem
 * it will returns NULL if no session was created yet.
 */
pidsession_t *pidsession_lookup(pid_t pid);

/**
 * @fn int pidsession_close(pid_t pid)
 * @desc This function closing a task session with file system
 * it returns 0 if all is done, or error.
 */
int pidsession_close(pid_t pid);

/**
 * @fn int pidsession_clone(pid_t pid,pid_t n_pid)
 * @desc This functions clones task session with @pid to the 
 * newly created session with @n_pid
 */
int pidsession_clone(pid_t pid,pid_t n_pid);

/**
 * @fn int pidsession_reset(pid_t pid)
 * @desc This function resets all essessions with 
 * FD_CLOEXEC flag - like POSIX says
 */
int pidsession_reset(pid_t pid);

/**
 * @fn int pidsession_change(pid_t pid,uid_t uid,gid_t gid)
 * @desc This function changes @pid session to be owned by 
 * @uid and @gid 
 */
int pidsession_change(pid_t pid,uid_t uid,gid_t gid);

/**
 * @fn int pidsession_setcap(pid_t pid,int cap)
 * @desc This function set capatibility mask for task 
 * session with @pid 
 */
int pidsession_setcap(pid_t pid,int cap);

/**
 * @fn int pidsession_setlimit(pid_t pid,int max_open)
 * @desc This function setting up limits for task session
 * with @pid 
 */
int pidsession_setlimit(pid_t pid,int max_open);

/**
 * @fn int pidsession_cloeid(pid_t pid, uint16_t id)
 * @desc This function close @id esession for @pid task session
 * definitly it's a close partially
 */
int pidsession_cloeid(pid_t pid, uint16_t id);

/**
 * @fn int pidsession_addeid(pidsession_t *sess,int mode, int flags,vnode_t *vnode)
 * @desc This function adds a new esession with given parameters 
 * vnode logic contains within this function.
 * It's a part of open logic
 */
int pidsession_addeid(pidsession_t *sess,int mode, int flags,vnode_t *vnode);

/**
 * @fn fsessid_t *pidsession_lunique(pidsession_t *psess,uint16_t unique)
 * @desc lookup for unique session for file ... returns NULL if nothing was found
 */
fessid_t *pidsession_lunique(pidsession_t *psess,uint16_t unique);

#endif /* __V2_SESSION_H__ */
