/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * general file system defines 
 */

#ifndef __V2_VFSDEF_H__
#define __V2_VFSDEF_H__

#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <stdarg.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <pthread.h>
#include <dirent.h>
#include <datalib/usrtc.h>

/* generic access permissions */
#define MAY_READ   (1 << 0)
#define MAY_WRITE  (1 << 1)
#define MAY_EXEC   (1 << 2)
#define MAY_ADD    (1 << 3)
#define MAY_DELETE (1 << 4)

/* fs independent mount flags */
#define MOUNT_RDONLY  (1 << 0)
#define MOUNT_SYNCHR  (1 << 1)
#define MOUNT_REMOUNT (1 << 2)
/* Here will be more flags */

typedef struct __vfs_info_t {
  char *fstype;
  char *author;
  char *info;
} vfs_info_t;

#include <vfs/super.h>

typedef struct __filesystem_proc_t {
  super_block_t *super; /* super block */
  prtid_t listen_port;  /* file system port */
  pid_t f_pid;          /* file system pid */
  int listeners;        /* max port listeners */
} filesystem_proc_t;

/* file system process */
int fsproc_init(filesystem_proc_t *fs_proc);
void fs_server_start(void);
/* set default mapper */
int fsproc_set_generic_mmap(filesystem_proc_t *fs_proc);

/* default mapper functions thread */
void *v2fs_generic_mapper(void *fs_proc);

/* v2 operations - generic wrappers (used internally, here just for reference) */
/**
 * lookup function, used usually by vfs and other internals 
 */
int v2_lookup(char *path,int *pp,int app,struct stat *sbuf);
/**
 * stat function, used for non-opened stats (stat)
 */
int v2_stat(super_block_t *super,char *path,int *pps,int pp,int plen,
	    int offset,struct stat *buf);
/**
 * read function - generic 
 */
void *v2_read(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,size_t *size,
              off_t offset,void *buf);
/**
 * mkdir function
 */
int v2_mkdir(super_block_t *super,char *path,int *pps,int pp,int plen,
	     int offset,uid_t uid,gid_t gid,int mode);
/**
 * unlink function 
 */
int v2_unlink(super_block_t *super,char *path,int *pps,int pp,int plen,
              int offset,uid_t uid,gid_t gid);
/**
 * unlinkdir function (rmdir)
 */
int v2_unlinkdir(super_block_t *super,char *path,int *pps,int pp,int plen,
		 int offset,uid_t uid,gid_t gid);
/**
 * open function
 */
int v2_open(super_block_t *super,char *path,int *pps,int pp,int offset,
	    int plen,pid_t pid,uid_t uid,gid_t gid,int mode,int flags,
	    struct stat *buf);
/**
 * mmap function
 */
int v2_mmap(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,size_t size,
            off_t offset,int flags,int prot,int *r_obj);
/**
 * this function used internally for checking of write access and other stub,
 * should prevent some kind of flood
 */
int v2_checkwr(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,off_t off);
/**
 * fstat function
 */
int v2_fstat(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,struct stat *buf);
/**
 * ioctl function (actually internally used, each device handling it itself)
 */
int v2_ioctl(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,
	     uni_rpc_t *rpc);
/**
 * write function
 */
int v2_write(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,
	     void *buf,size_t *size,off_t off);
/**
 * readdir function, actually reads several entries
 */
int v2_readdir(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,void *msg,
	       size_t *msg_len,size_t *size,off_t off,uint16_t *ec);
/**
 * close function
 */
int v2_close(pid_t pid,uint16_t unique);

/**
 * readlink function
 */
int v2_readlink(super_block_t *super,char *path,int *pps,int pp,int plen,
                int offset);

/**
 * ftruncate function
 */
int v2_ftruncate(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,off_t off);

/* permissions checking */
int generic_permission(vnode_t *vnode, uid_t uid, gid_t gid);
int vnode_permission(vnode_t *vnode, uid_t uid, gid_t gid);

/* functions for interactions with vfsv2 */
int vfs_register(prtid_t id,int spec_flags,vfs_info_t *fsinf);

/* file system service thread */
void *v2fs_thread(filesystem_proc_t *fs_proc);

#endif /* __V2_VFSDEF_H__ */
