/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * super block
 */

#ifndef __V2_VNODE_H__
#define __V2_VNODE_H__

#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <stdarg.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <pthread.h>
#include <dirent.h>
#include <datalib/usrtc.h>

#include <vfs/vfsdef.h>
#include <vfs/super.h>

typedef struct __fslock_t {
  int coverage;
  size_t *lsizes;
  off_t *offsets;
  int *attrs;
  pid_t *pids;
} fslock_t;

typedef struct __vnode_t {
  struct __super_block_t *super; /* super block pointer */
  struct stat vstat; /* stat of the buffer */
  usrtc_node_t *node; /* pointer to the cache structure */
  uint64_t hash; /* name hash */
  pthread_mutex_t ilock; /* locking sync */
  fslock_t *lock; /* file locked structure */
  ulong_t mem_obj; /* memory object of the file */
  int ref_count; /* reference counter */
  char *pathname; /* info on this node */
  int pp; /* number of pathname separators */
  int plen; /* length of the name */
  int *pps; /* array of a length based parsed separators */
  void *spec_data; /* specific data for filesystem implementation */
} vnode_t;

#define vnode_lock(b)  pthread_mutex_lock(&b->ilock)
#define vnode_unlock(b)  pthread_mutex_unlock(&b->ilock)

/**
 * @fn int vnode_cache_initialize(void)
 * @desc This function initialize all internals 
 * for vnode cache 
 */
int vnode_cache_initialize(void);

/**
 * @fn vnode_t *vnode_create(struct __super_block_t *super,
 *		      char *pathname,int pp,int plen,int *pps)
 * @desc create a new vnode (not in the cache) with given parameters
 */
vnode_t *vnode_create(struct __super_block_t *super,
		      char *pathname,int pp,int plen,int *pps);

/**
 * @fn vnode_t *vnode_lookup(uint64_t hash)
 * @desc looking up for an existing vnode in the cache 
 * returns NULL if no
 */
vnode_t *vnode_lookup(uint64_t hash);

/**
 * @fn int vnode_destroy(vnode_t *vnode)
 * @desc destroy vnode with all specific 
 * structures, *note* fs data specific should be 
 * released by file system close implementation
 */
int vnode_destroy(vnode_t *vnode);

/**
 * @fn int vnode_pop(vnode_t *vnode)
 * @desc removes vnode from cache
 */
int vnode_pop(vnode_t *vnode);

/**
 * @fn int vnode_push(vnode_t *vnode)
 * @desc inserts vnode to the cache
 */
int vnode_push(vnode_t *vnode);

/**
 * @fn uint64_t vnode_hash_ch(char *path)
 * @desc calculates unique 64 bit key via pathname
 */
uint64_t vnode_hash_ch(char *path);

#endif /* __V2_VNODE_H__ */
