/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * libata core implementation
 *
 * ATA timings are ported from Linux
 */

/*
 * TO DO [dg]:
 *
 * 1) Add timeout setting at PIO too.
 */

#include <ata/libata.h>
#include <pthread.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <bitwise.h>
#include <err.h>
#include <debug.h>

#define hold_port_irq(port)     pthread_mutex_lock(&(port)->mutex)

#define DECLARE_PIO_ARRAY_FUNCTION(name, elem_type, in_method, out_method)  \
  void name(ata_cmd_t *cmd, void *buf, uint32_t nbytes)                     \
  {                                                                         \
    elem_type *p = (elem_type*)buf;                                         \
    elem_type *p1 = p + nbytes / sizeof(elem_type);                         \
    ata_hwif_t *hwif = cmd->dev->host_port->hwif;                           \
    ata_registers_t regs = cmd->dev->host_port->regs;                       \
                                                                            \
    if (!blkreq_rw(cmd->rq)) {                                              \
      for (; p < p1; p++)                                                   \
        *p = hwif->in_method(regs.data);                                    \
    } else {                                                                \
      for (; p < p1; p++)                                                   \
        hwif->out_method(regs.data, *p);                                    \
    }                                                                       \
  }

#define ENOUGH(v, unit)   (((v)-1)/(unit)+1)
#define EZ(v, unit)   ((v)?ENOUGH(v, unit):0)

typedef const uint8_t (ata_iocmd_table_t)[NR_ATA_PROTOCOLS - 2][2];

struct sect_sizes {
  /* pow of logical sectors per physical one */
  uint16_t log_sects_pow : 4;
  uint16_t __res15 : 8;
  /* device logical sector longer than 256 words */
  uint16_t log_sect_longer_256 : 1;
  /* device has multiple logical sectors per track */
  uint16_t mult_log_sects : 1;
  uint16_t is_valid : 2;
};

/* enable and disable feature codes */
const uint8_t en_dis_feature_tbl[][2] = {
  {AF_DISABLE_8BIT_PIO, AF_ENABLE_8BIT_PIO},
  {AF_DISABLE_WRITE_CACHE, AF_ENABLE_WRITE_CACHE},
  {AF_DISABLE_ADVANCED_PM, AF_ENABLE_ADVANCED_PM},
  {AF_DISABLE_STANDBY_PWRUP, AF_ENABLE_STANDBY_PWRUP},
  {AF_DISABLE_CFA_PM1, AF_ENABLE_CFA_PM1},
  {AF_DISABLE_MEDIA_STATUS_NOTIF, AF_ENABLE_MEDIA_SATUS_NOTIF},
  {AF_DISABLE_AUTOMATIC_ACOUSTIC, AF_ENABLE_AUTOMATIC_ACOUSTIC},
  {AF_DISABLE_READ_LOOK_AHEAD, AF_ENABLE_READ_LOOK_AHEAD},
  {AF_DISABLE_RELEASE_INTR, AF_ENABLE_RELEASE_INTR},
  {AF_DISABLE_SERVICE_INTR, AF_ENABLE_SERVICE_INTR},
  {AF_DISABLE_PWRON_REVERT, AF_ENABLE_PWRON_REVERT}
};

/*
 * ATA command tables, command codes depends on transfer
 * protocol, direction and taskfile flags
 */

const ata_iocmd_table_t ata_iocmd_tables[] = {
  /* normal commands */
  {
    /* pio protocol */
    {ATA_READ_SECTORS, ATA_WRITE_SECTORS},
    /* pio multiple protocol */
    {ATA_READ_MULTIPLE, ATA_WRITE_MULTIPLE},
    /* dma protocol */
    {ATA_READ_DMA, ATA_WRITE_DMA},
    /* dma queued protocol */
    {ATA_READ_DMA_QUEUED, ATA_WRITE_DMA_QUEUED},
    /* fpdma queued protocol, not supported */
    {SATA_READ_FPDMA_QUEUED, SATA_WRITE_FPDMA_QUEUED},
    /* stream protocol */
    {ATA_READ_STREAM_EXT, ATA_WRITE_STREAM_EXT},
    /* stream dma protocol */
    {ATA_READ_STREAM_DMA_EXT, ATA_WRITE_STREAM_DMA_EXT}
  },
  /* extended commands */
  {
    {ATA_READ_SECTORS_EXT, ATA_WRITE_SECTORS_EXT},
    {ATA_READ_MULTIPLE_EXT, ATA_WRITE_MULTIPLE_EXT},
    {ATA_READ_DMA_EXT, ATA_WRITE_DMA_EXT},
    {ATA_READ_DMA_QUEUED_EXT, ATA_WRITE_DMA_QUEUED_EXT},
    {SATA_READ_FPDMA_QUEUED, SATA_WRITE_FPDMA_QUEUED},
    {ATA_READ_STREAM_EXT, ATA_WRITE_STREAM_EXT},
    {ATA_READ_STREAM_DMA_EXT, ATA_WRITE_STREAM_DMA_EXT}
  }
};

/*
 * PIO 0-4, MWDMA 0-2 and UDMA 0-6 timings (in nanoseconds).
 * These were taken from ATA/ATAPI-6 standard, rev 0a, except
 * for UDMA6, which is currently supported only by Maxtor drives.
 *
 * For PIO 5/6 MWDMA 3/4 see the CFA specification 3.0.
 */

static const struct ata_timing pio_timing[] = {
  { 0, 70, 290, 240, 600, 165, 150, 600, 0 },
  { 1, 50, 290,  93, 383, 125, 100, 383, 0 },
  { 2, 30, 290,  40, 330, 100,  90, 240, 0 },
  { 3, 30,  80,  70, 180,  80,  70, 180, 0 },
  { 4, 25,  70,  25, 120,  70,  25, 120, 0 },
  { 5, 15,  65,  25, 100,  65,  25, 100, 0 },
  { 6, 10,  55,  20,  80,  55,  20,  80, 0 },

  { 0xFF }
};

static const struct ata_timing dma_timing[] = {
  { ATA_MWDMA0,  60,   0,   0,   0, 215, 215, 480,   0 },
  { ATA_MWDMA1,  45,   0,   0,   0,  80,  50, 150,   0 },
  { ATA_MWDMA2,  25,   0,   0,   0,  70,  25, 120,   0 },
  { ATA_MWDMA3,  25,   0,   0,   0,  65,  25, 100,   0 },
  { ATA_MWDMA4,  25,   0,   0,   0,  55,  20,  80,   0 },

  { ATA_UDMA0,     0,   0,   0,   0,   0,   0,   0, 120 },
  { ATA_UDMA1,     0,   0,   0,   0,   0,   0,   0,  80 },
  { ATA_UDMA2,     0,   0,   0,   0,   0,   0,   0,  60 },
  { ATA_UDMA3,     0,   0,   0,   0,   0,   0,   0,  45 },
  { ATA_UDMA4,     0,   0,   0,   0,   0,   0,   0,  30 },
  { ATA_UDMA5,     0,   0,   0,   0,   0,   0,   0,  20 },
  { ATA_UDMA6,     0,   0,   0,   0,   0,   0,   0,  15 },

  { 0xFF }
};

DECLARE_PIO_ARRAY_FUNCTION(pio_arrayb, uint8_t, inb, outb)
DECLARE_PIO_ARRAY_FUNCTION(pio_arrayw, uint16_t, inw, outw)

static inline void complete_drv_cmd(ata_device_t *dev, ata_cmd_t *cmd)
{
  dev->last_cmd_err = cmd->err;
  complete(&dev->cmd_compl, cmd->err);
  put_blk_request(cmd->rq);
  ata_free_cmd(cmd);
}

static void set_tf_device(ata_device_t *dev, ata_taskfile_t *tf)
{
  tf->device = (dev->id << 4) | DEV_LBA;
}

static void set_tf_nsect(ata_taskfile_t *tf, uint32_t nsect,
                         ata_io_protocol_t prot)
{
  if (tf->flags & TASKFILE_EXT) {
    if (ata_queued_protocol(prot)) {
      tf->feature = nsect;
      tf->hob_feature = (nsect == 65536) ? 0 : nsect >> 8;
    } else {
      tf->nsect = nsect;
      tf->hob_nsect = (nsect == 65536) ? 0 : nsect >> 8;
    }
  } else {
    if (ata_queued_protocol(prot)) {
      tf->feature = (nsect == 256) ? 0 : nsect;
    } else {
      tf->nsect = (nsect == 256) ? 0 : nsect;
    }
  }
}

bool wait_dev_ready(ata_device_t *dev, ata_status_t *ret_stat)
{
  ata_status_t stat;

  while (1) {
    stat.val = ata_read_device_status(dev);
    if (stat.err || stat.df) {
      dev->run_cmd->err = ATA_ERROR_DEVICE;
      break;
    }

    if (!stat.bsy) {
      if (!stat.drq) {
        /*
         * typically in this case any device is not present
         * on that parallel port
         */
        dev->run_cmd->err = ATA_ERROR_DEVICE;
      }

      break;
    }
  }
  *ret_stat = stat;

  return (dev->run_cmd->err == 0);
}

void ata_done_cmd(ata_device_t *dev, ata_cmd_t *cmd)
{
  if (dev->cmd_phase == PHASE_TRANSFER) {
    if (ata_dma_protocol(cmd->protocol))
      reset_timeout(&dev->dma_tmr);
  } else if (ata_intrq_enabled(dev)) {
    reset_timeout(&dev->intrq_tmr);
  }
  dev->cmd_phase = PHASE_INACTIVE;
  dev->last_cmd_err = cmd->err;
  if (cmd->err) {
    cmd->nr_errors++;
  } else if (cmd->rq) {
    cmd->rq->complete_sectors = cmd->rq->nr_sectors;
  }

  if (ata_queued_cmd(cmd)) {
    ata_del_queued_cmd(dev, cmd);
  }
  dev->run_cmd = NULL;
  if (ata_cmd_queue_empty(dev)) {
    dev->host_port->serv_dev = NULL;
  }
  cmd->done(cmd);
}

void do_blk_pio(ata_cmd_t *cmd, uint32_t nr_sects,
                void (*pio_fn)(ata_cmd_t*, void*, uint32_t))
{
  blkio_t *b;
  sector_t offset;
  uint32_t n;
  void *buf;
  blk_request_t *rq = cmd->rq;

  if (rq->current_blkio) {
    b = rq->current_blkio;
  } else {
    b = container_of(list_node_first(&rq->blkios), blkio_t, node);
  }
  offset = rq->start_sector + rq->complete_sectors - b->start_sector;
  rq->complete_sectors += nr_sects;

  while (nr_sects) {
    if (offset == b->nr_sectors) {
      offset = 0;
      b = container_of(b->node.next, blkio_t, node);
    }
    rq->current_blkio = b;
    n = MIN(nr_sects, b->nr_sectors - offset);
    buf = (void*)((ulong_t)b->buf + (offset << cmd->block_shift));
    pio_fn(cmd, buf, n << cmd->block_shift);
    offset += n;
    nr_sects -= n;
  }
}

static void __do_default_io(ata_cmd_t *cmd)
{
  ata_device_t *dev = cmd->dev;
  int r = 0;
  uint32_t nsects = 0, remain_sects;
  void *buf = cmd->buf;
  void (*pio_fn)(ata_cmd_t*, void*, uint32_t);
  ata_status_t stat;

  ASSERT(!blk_drv_request(cmd->rq) || buf, "Data buffer isn't provided");

  if (ata_pio_protocol(cmd->protocol)) {
    if (ata_intrq_enabled(dev))
      remain_sects = (cmd->protocol == ATA_PROT_MULTI_PIO) ? dev->mult_count : 1;
    else
      remain_sects = (cmd->rq->nr_sectors - cmd->rq->complete_sectors);

    pio_fn = dev->set_features.pio_8bit ? pio_arrayb : pio_arrayw;

    while (remain_sects) {
      if (!wait_dev_ready(dev, &stat)) {
        /* an error is occurred */
        break;
      }
      if (cmd->protocol == ATA_PROT_MULTI_PIO) {
        nsects = MIN(dev->mult_count, remain_sects);
      } else {
        nsects = 1;
      }

      if (!blk_drv_request(cmd->rq)) {
        do_blk_pio(cmd, nsects, pio_fn);
      } else {
        pio_fn(cmd, buf, nsects << cmd->block_shift);
        cmd->rq->complete_sectors += nsects;
        buf = (void*)((ulong_t)buf + (nsects << cmd->block_shift));
      }
      remain_sects -= nsects;
    }
  } else {  // one of dma protocols
    if (cmd->retry_num) {
      ata_load_sglist(dev);
    } else {
      r = ata_prepare_dma(dev, cmd);
    }
    r = r || ata_start_dma(dev);
    if (r) {
      cmd->err = ATA_ERROR_HOST;
    }
  }
}

static void release_port_irq(ata_port_t *port)
{
  if (!ata_port_busy(port) && ata_port_waited(port)) {
    ata_clear_port_waited(port);
    event_signal(port->io_done);
  }
  ata_port_unlock(port);
}

bool try_prepare_cmd_restart(ata_device_t *dev, ata_cmd_t *cmd)
{
  bool ret = false;
  int r;
  sector_t err_sector = 0, nr_sects;
  ata_taskfile_t tf;
  blk_request_t *rq = cmd->rq;

  if (cmd->err == ATA_ERROR_DEVICE) {
    /*
     * try restart command from the sector
     * where the first error is occurred
     */
    tf.flags = cmd->tf.flags;
    r = ata_read_taskfile(dev, &tf);
    if (r)
      cmd->err = ATA_ERROR_HOST;
    else
      err_sector = ata_tflba_to_sector(&tf);
  }
  if (cmd->retry_num < cmd->retries) {
    cmd->nr_errors++;
    cmd->retry_num++;
    cmd->err = 0;
    if (err_sector && (rq != NULL) && (rq->type == REQ_TYPE_IO)) {
      nr_sects = rq->start_sector + rq->complete_sectors - err_sector;
      ata_sector_to_tflba(err_sector, &cmd->tf);
      set_tf_nsect(&cmd->tf, nr_sects, cmd->protocol);
      cmd->rq->complete_sectors = err_sector - cmd->rq->start_sector;
      ata_rebuild_sglist(dev, (err_sector - rq->start_sector) << cmd->block_shift,
                         nr_sects << cmd->block_shift);
    }

    if (dev->drv->prep_cmd_retry)
      dev->drv->prep_cmd_retry(cmd);

    ret = true;
  } else if (cmd->err == ATA_ERROR_DEVICE) {
    if (cmd->rq) {
      rq->complete_sectors = err_sector - cmd->rq->start_sector;
    }
    memcpy(&cmd->tf, &tf, sizeof(ata_taskfile_t));
  }

  return ret;
}

static bool verify_id_checksum(uint8_t id[512])
{
  uint8_t sum = 0;
  int i;

  if ((id[255] & 0xFF) == ATA_SIGNATURE) {
    for (i = 0; i < 512; i++)
      sum += id[i];
  }

  return (sum == 0);
}

static void raw_identify_to_param_info(ata_param_info_t *pi, unsigned short *buf)
{
  ulong_t s = 512;

  pi->word0 = buf[0];
  pi->specific_conf = buf[2];
  pi->cfa_private = (uint32_t)buf[8] << 16 | buf[7];
  memcpy(pi->serial_num, &buf[10], sizeof(pi->serial_num));
  memcpy(pi->revision, &buf[23], sizeof(pi->revision));
  memcpy(pi->model_num, &buf[27], sizeof(pi->model_num));
  pi->max_rw_sectors = buf[47];
  pi->caps.word49 = buf[49];
  pi->caps.word50 = buf[50];
  pi->caps.word127  = buf[127];
  pi->caps.word128 = buf[128];
  if (buf[60] == 0xFFFF && buf[61] == 0xFFF) {  /* maximum value in user addressable sectors? */
    pi->total_addr_sects = (sector_t)buf[103] << 48 |
         (sector_t)buf[102] << 32 | buf[101] << 16 | buf[100];
  } else {
    pi->total_addr_sects = (uint32_t)buf[61] << 16 | buf[60];
  }
  pi->word63 = buf[63];
  pi->pio_modes = buf[64];
  pi->min_dma_cycle = buf[65];
  pi->recomm_dma_cycle = buf[66];
  pi->min_pio_cycle = buf[67];
  pi->min_pio_cycle_iordy = buf[68];
  pi->max_queue_depth = buf[75];
  memcpy(&pi->cmd_set, &buf[82], sizeof(pi->cmd_set));
  pi->erase_unit_compl_time = buf[89];
  pi->enhsec_erase_compl_time = buf[90];
  pi->stream_min_req_size = buf[95];
  /* if device logical sector is longer than 256 words? */
  if (buf[106] & (1 << 12)) {
    /* words per logical sector */
    s = (uint32_t)buf[118] << 16 | buf[117];
    s <<= 1;
  }
  /* if device has multiple logical sector per the physical one? */
  if (buf[106] & (1 << 13)) {
    /* low for bits of the 106-th word contains binary power of logical
     * sectors per phys one
     */
    s <<= (buf[106] & 0xF);
  }
  pi->sect_shift = bit_find_msf(s);
}

static void pkt_identify_cmd_done(ata_cmd_t *cmd)
{
  ata_device_t *dev = cmd->dev;
  unsigned short *buf = (unsigned short*)cmd->buf;

  if (!cmd->err && !verify_id_checksum((uint8_t*)buf))
    cmd->err = ATA_ERROR_DEVICE;

  if (!cmd->err) {
    dev->param_info.word62 = buf[62];
    dev->param_info.pkt_bus_release_delay = buf[71];
    dev->param_info.serv_cmd_bsy_time = buf[72];
    if (!(buf[80] == 0 || buf[80] == 0xFFFF)) {
      dev->param_info.major_ver_num = bit_find_msf(buf[80]);
    } else {
      dev->param_info.major_ver_num = -1;
    }
    dev->param_info.atapi_byte_count = buf[125];
  }

  complete_drv_cmd(dev, cmd);
}

static void identify_cmd_done(ata_cmd_t *cmd)
{
  int r;
  ata_device_t *dev = cmd->dev;
  unsigned short *buf = (unsigned short*)cmd->buf;

  if (!cmd->err && !verify_id_checksum((uint8_t*)buf))
    cmd->err = ATA_ERROR_DEVICE;

  if (cmd->err) {
    complete_drv_cmd(dev, cmd);
    return;
  }
  raw_identify_to_param_info(&dev->param_info,
                     (unsigned short*)cmd->buf);

  if (buf[85] & (1 << 4)) {
    /*
     * the device support the PACKET command
     * feature set, get identify packet device data
     */
    cmd->retry_num = 0;
    cmd->tf.cmd = ATA_IDENTIFY_PACKET_DEVICE;
    cmd->done = pkt_identify_cmd_done;
    r = ata_start_command(dev, cmd);
    if (r || !ata_intrq_enabled(dev))
      ata_done_cmd(dev, cmd);
  } else {
    complete_drv_cmd(dev, cmd);
  }
}

ata_cmd_t *find_ready_queued_cmd(ata_device_t *dev)
{
  ata_cmd_t *cmd;
  int tag;

  tag = dev->drv->ready_cmd_tag(dev);
  if (tag == -1)
    return NULL;

  list_for_each_entry(&dev->cmd_queue.cmds, cmd, node) {
    if (cmd->tag == tag)
      break;
  }
  if (&cmd->node == list_head(&dev->cmd_queue.cmds)) {
    /*
     * hardware error, command with such tag wasn't issued
     * TO DO [dg]: output message to log
     */
    cmd = NULL;
  }

  return cmd;
}


/* API implementation */

/*
 * Default ata devices interrupt handler
 *
 * Check status of all devices sharing the same IRQ
 * and call specific irq handler for it
 */
bool ata_default_irq_handler(void *priv)
{
  ata_port_t *port = (ata_port_t*)priv;
  ata_device_t *dev;
  port_status_t status;
  ata_host_t *host = port->host;
  uint32_t i, n;

  if (host->irq_shared) {
    i = 0;
    n = host->nr_ports;
  } else {
    i = port->num;
    n = i + 1;
  }

  for (; i < n; i++) {
    port = host->ports[i];
    status = host->port_ops->read_status(port);
    dev = port->serv_dev;
    if ((dev == NULL) && ((status & PORT_STAT_IRQ) || !host->irq_shared))
      continue;

    hold_port_irq(port);
    ata_clear_irq(host, port);
    dev->drv->irq_handler(dev, status);
    release_port_irq(port);
  }

  return true;
}

int ata_probe_init_ports(ata_host_t *host, int *nr_scanned_ports)
{
  int i, j;
  int ret = 0;
  ata_port_t *port;
  ata_device_t *dev;

  for (i = 0; i < host->nr_ports; i++) {
    host->ports[i] = malloc(sizeof(ata_port_t));
    if (host->ports[i] == NULL) {
      ret = -ENOMEM;
    } else {
      memset(host->ports[i], 0, sizeof(ata_port_t));
      ret = ata_init_port_resources(host->ports[i]);
    }

    if (ret)
      break;
  }

  if (ret) {
    for (j = 0; j < i; j++) {
      ata_free_port_resources(host->ports[i]);
      free(host->ports[j]);
      host->ports[j] = NULL;
    }
    return ret;
  }

  for (i = 0; i < host->nr_ports; i++) {
    port = host->ports[i];
    port->host = host;
    port->num = i;
    ret = host->port_ops->init(port);
    if (ret)
      continue;

    for (j = 0; j < 2; j++) {
      if (!host->port_ops->probe(port, j))
        continue;

      if (port->devices[j] == NULL) {
        dev = malloc(sizeof(ata_device_t));
        if (dev == NULL) {
          ret = -ENOMEM;
          break;
        }
        memset(dev, 0, sizeof(ata_device_t));
        dev->id = j;
        dev->host_port = port;
        port->devices[j] = dev;
      }
    }
    if (ret)
      break;
  }

  if (nr_scanned_ports != NULL)
    *nr_scanned_ports = (i == host->nr_ports) ? host->nr_ports : (i + 1);

  return ret;
}

void ata_start_transfer(ata_device_t *dev, ata_cmd_t *cmd)
{
  ata_port_t *port = dev->host_port;

  dev->cmd_phase = PHASE_TRANSFER;
  if (dev->drv->rw_data) {
    dev->drv->rw_data(cmd);
  } else {
    __do_default_io(cmd);
  }
  if (ata_dma_protocol(cmd->protocol))
    set_timeout(&dev->dma_tmr, port->host->dma_timeout);
}

int ata_set_device_pw_mode(ata_device_t *dev, ata_pw_mode_t mode)
{
  int ret = 0;

  return ret;
}

int ata_enable_dma(ata_device_t *dev, bool enable)
{
  return 0;
}

int ata_prepare_dma(ata_device_t *dev, ata_cmd_t *cmd)
{
 return 0;
}

void ata_rebuild_sglist(ata_device_t *dev, uint32_t offset, uint32_t len)
{

}

int ata_load_sglist(ata_device_t *dev)
{
  return 0;
}

void ata_reinit_dma(ata_device_t *dev, uint32_t offset, uint32_t len)
{

}

int ata_start_dma(ata_device_t *dev)
{
  return 0;
}

void ata_stop_dma(ata_device_t *dev)
{

}

void ata_reset_device(ata_device_t *dev)
{
}

int ata_iocmd_from_request(blk_request_t *rq, ata_cmd_t *cmd)
{
  uint8_t cmd_code = 0;
  ata_taskfile_t *tf = &cmd->tf;
  int ext;
  uint8_t dev = 0;
  int io_dir;

  if (cmd->protocol == ATA_PROT_NODATA || cmd->protocol == ATA_PROT_PACKET)
    return -1;

  cmd->rq = rq;
  io_dir = blkreq_rw(rq);
  tf->flags = 0;
  if ((rq->start_sector >> 28) || (rq->nr_sectors > 256)) {
    ext = 1;
    tf->flags |= TASKFILE_EXT;
  }

  /* fill command for I/O request */
  if (tf->flags & TASKFILE_FUA) {
    if (io_dir) {
      switch (cmd->protocol) {
      case ATA_PROT_MULTI_PIO:
        cmd_code = ATA_WRITE_MULTIPLE_FUA_EXT;
        break;
      case ATA_PROT_DMA:
        cmd_code = ATA_WRITE_DMA_FUA_EXT;
        break;
      case ATA_PROT_DMA_QUEUED:
        cmd_code = ATA_WRITE_DMA_QUEUED_FUA_EXT;
        break;
      default:
        break;
      }
    }
    if (cmd->protocol == ATA_PROT_FPDMA_QUEUED)
      dev |= DEV_FPDMA_FUA;
  }

  if (!cmd_code)
    cmd_code = ata_iocmd_tables[ext][cmd->protocol - ATA_PROT_PIO][io_dir];

  /* fill taskfile */
  tf->cmd = cmd_code;
  ata_sector_to_tflba(rq->start_sector, tf);
  set_tf_nsect(tf, rq->nr_sectors, cmd->protocol);
  if (ata_queued_protocol(cmd->protocol)) {
    tf->nsect = cmd->tag << 3;
    tf->hob_nsect = 0;
  } else {
    tf->feature = tf->hob_feature = 0;
  }
  tf->flags |= (TASKFILE_SECT_COUNT | TASKFILE_LBA_LOW |
                TASKFILE_LBA_MID | TASKFILE_LBA_HIGH);
  set_tf_device(cmd->dev, tf);

  return 0;
}

sector_t ata_tflba_to_sector(ata_taskfile_t *tf)
{
  sector_t sector;

  sector = tf->lbah << 16 | tf->lbam << 8 | tf->lbal;

  if (tf->flags & TASKFILE_EXT) {
    sector |= (sector_t)(tf->hob_lbah) << 40 |
              (sector_t)(tf->hob_lbam) << 32 |
              (uint32_t)tf->hob_lbal << 24;
  } else {
    sector |= (uint32_t)(tf->device & 0xF) << 24;
  }

  return sector;
}

void ata_sector_to_tflba(sector_t sector, ata_taskfile_t *tf)
{
  tf->lbal = sector;
  tf->lbam = sector >> 8;
  tf->lbah = sector >> 16;
  if (tf->flags & TASKFILE_EXT) {
    tf->hob_lbal = sector >> 24;
    tf->hob_lbam = sector >> 32;
    tf->hob_lbah = sector >> 40;
  } else {
    /* low half of device register contains 27-24 bits of LBA */
    tf->device &= ~0xF;
    tf->device |= (sector >> 24) & 0xF;
  }
}

int ata_service_cmd_from_request(blk_request_t *rq, ata_cmd_t *cmd)
{
  switch (rq->type) {
  case REQ_TYPE_FLUSH:
    break;
  }

  return 0;
}

static void __do_nointrq_cmd(ata_cmd_t *cmd, bool *restart)
{
  struct timespec ts = {0, ATA_STATUS_TIMEOUT};
  ata_device_t *dev = cmd->dev;
  ata_status_t stat;

  nanosleep(&ts, NULL);
  if (wait_dev_ready(dev, &stat)) {
    if (ata_queued_cmd(cmd) && stat.serv) {
      cmd = find_ready_queued_cmd(dev);
      dev->run_cmd = cmd;
      if (cmd == NULL)
        return;
    }
    ata_start_transfer(dev, cmd);
    if (cmd->err)
      *restart = try_prepare_cmd_restart(dev, cmd);
  } else {
    *restart = try_prepare_cmd_restart(dev, cmd);
  }
}

int ata_start_command(ata_device_t *dev, ata_cmd_t *cmd)
{
  bool repeat;
  ata_port_t *port = dev->host_port;
  ata_cmd_t *prev_cmd;
  int prev_phase;
  ata_device_t *prev_dev;
  ata_host_t *host = port->host;
  uint32_t status;

  ASSERT(cmd->protocol == ATA_PROT_NODATA || cmd->rq,
         "Block request information is not presented for block I/O command\n");

  if (ata_queued_protocol(cmd->protocol) &&
      !ata_add_queued_cmd(dev, cmd)) {
    return -EBUSY;
  }

  prev_cmd = dev->run_cmd;
  prev_phase = dev->cmd_phase;
  prev_dev = port->serv_dev;
  dev->run_cmd = cmd;
  port->serv_dev = dev;
  dev->cmd_phase = PHASE_WAIT_READY;

  do {
    repeat = false;
    if (dev->drv->prepare_cmd) {
      if (dev->drv->prepare_cmd(dev, cmd)) {
        /* don't retry cmd which can't be even prepared */
        cmd->retry_num = cmd->retries;
        goto error;
      }
    }

    if (ata_write_taskfile(dev, &cmd->tf)) {
      cmd->err = ATA_ERROR_HOST;
      goto error;
    }

    if (host->port_ops->issue_cmd) {
      if (host->port_ops->issue_cmd(port, cmd->tag)) {
        cmd->err = ATA_ERROR_HOST;
        goto error;
      }
    }

    if (dev->drv->post_issue_cmd) {
      if (dev->drv->post_issue_cmd(dev, cmd))
        goto error;
    }

    if (ata_intrq_enabled(dev))
      set_timeout(&dev->intrq_tmr, cmd->timeout);
    else
      __do_nointrq_cmd(cmd, &repeat);

  error:
    if (cmd->err) {
      if (cmd->err == ATA_ERROR_HOST)
        status = host->port_ops->read_status(port);
      else
        status = ata_read_device_status(dev);

      dev->drv->error_handler(dev, cmd->err, status);
      if (!repeat && (cmd->retry_num < cmd->retries)) {
        repeat = true;
        cmd->err = 0;
        cmd->retry_num++;
      }
    }
  } while (repeat);

  if (cmd->err) {
    dev->run_cmd = prev_cmd;
    dev->host_port->serv_dev = prev_dev;
    dev->cmd_phase = prev_phase;
    if (ata_queued_protocol(cmd->protocol))
      ata_del_queued_cmd(dev, cmd);
  }

  return cmd->err ? -1 : 0;
}

int ata_get_param_info(ata_device_t *dev)
{
  ata_cmd_t *cmd;
  int r = -ENOMEM;
  blk_request_t *rq;

  rq = alloc_init_blkreq(NULL, REQ_TYPE_IO, 0, 1, 0);
  if (rq == NULL)
    return -ENOMEM;

  set_rq_flags(rq, BLKREQ_DRV);
  cmd = ata_alloc_init_cmd(dev, dev->io_prot);
  if (IS_ERR(cmd)) {
    r = PTR_ERR(cmd);
    goto out_rq;
  }
  cmd->rq = rq;

  cmd->buf = malloc(512);
  if (cmd->buf == NULL)
    goto out_cmd;

  cmd->block_shift = 9;
  cmd->protocol = ATA_PROT_PIO;
  cmd->retries = 2;
  cmd->tf.cmd = ATA_IDENTIFY_DEVICE;
  cmd->done = identify_cmd_done;
  set_tf_device(dev, &cmd->tf);
  r = ata_start_command(dev, cmd);
  if (r || !ata_intrq_enabled(dev))
    ata_done_cmd(dev, cmd);
  if (!ata_intrq_enabled(dev))
    r = dev->last_cmd_err ? -1 : 0;

  return r;

out_rq:
  put_blk_request(rq);
out_cmd:
  ata_free_cmd(cmd);

  return r;
}

int ata_set_io_protocol(ata_device_t *dev, ata_io_protocol_t protocol)
{
  return 0;
}

int ata_enable_feature(ata_device_t *dev, int feature)
{
  return 0;
}

int ata_disable_feature(ata_device_t *dev, int feature)
{
  return 0;
}

int ata_direct_read_taskfile(ata_device_t *dev, ata_taskfile_t *tf)
{
  ata_registers_t regs = dev->host_port->regs;
  ata_hwif_t *hwif = dev->host_port->hwif;
  uint32_t flags = tf->flags;

  hwif->outb(regs.device, tf->device);
  /* read low half of ata registers */
  hwif->outb(regs.dev_ctl, dev->ctl_cmd & ~DEVCTL_HOB);
  if (flags & TASKFILE_STATUS)
    tf->status = hwif->inb(regs.status);
  if (flags & TASKFILE_ERROR)
    tf->error = hwif->inb(regs.error);
  if (flags & TASKFILE_SECT_COUNT)
    tf->nsect = hwif->inb(regs.sect_count);
  if (flags & TASKFILE_LBA_LOW)
    tf->lbal = hwif->inb(regs.lba_low);
  if (flags & TASKFILE_LBA_MID)
    tf->lbam = hwif->inb(regs.lba_mid);
  if (flags & TASKFILE_LBA_HIGH)
    tf->lbah = hwif->inb(regs.lba_high);

  if (flags & TASKFILE_EXT) {
    /* read high half of registers */
    hwif->outb(regs.dev_ctl, dev->ctl_cmd | DEVCTL_HOB);
    if (flags & TASKFILE_SECT_COUNT)
      tf->hob_nsect = hwif->inb(regs.sect_count);
    if (flags & TASKFILE_LBA_LOW)
      tf->hob_lbal = hwif->inb(regs.lba_low);
    if (flags & TASKFILE_LBA_MID)
      tf->hob_lbam = hwif->inb(regs.lba_mid);
    if (flags & TASKFILE_LBA_HIGH)
      tf->hob_lbah = hwif->inb(regs.lba_high);
  }

  return 0;
}

int ata_direct_write_taskfile(ata_device_t *dev, ata_taskfile_t *tf)
{
  ata_registers_t regs = dev->host_port->regs;
  ata_hwif_t *hwif = dev->host_port->hwif;
  uint32_t flags = tf->flags;

  hwif->outb(regs.device, tf->device);
  hwif->outb(regs.dev_ctl, dev->ctl_cmd & ~DEVCTL_HOB);
  if (tf->flags & TASKFILE_EXT) {
    /* write high half of registers */
    if (flags & TASKFILE_FEATURE)
      hwif->outb(regs.feature, tf->hob_feature);
    if (flags & TASKFILE_SECT_COUNT)
      hwif->outb(regs.sect_count, tf->hob_nsect);
    if (flags & TASKFILE_LBA_LOW)
      hwif->outb(regs.lba_low, tf->hob_lbal);
    if (flags & TASKFILE_LBA_MID)
      hwif->outb(regs.lba_mid, tf->hob_lbam);
    if (flags & TASKFILE_LBA_HIGH)
      hwif->outb(regs.lba_high, tf->hob_lbah);
  }

  /* write low half of ata registers */
  if (flags & TASKFILE_FEATURE)
    hwif->outb(regs.feature, tf->feature);
  if (flags & TASKFILE_SECT_COUNT)
    hwif->outb(regs.sect_count, tf->nsect);
  if (flags & TASKFILE_LBA_LOW)
    hwif->outb(regs.lba_low, tf->lbal);
  if (flags & TASKFILE_LBA_MID)
    hwif->outb(regs.lba_mid, tf->lbam);
  if (flags & TASKFILE_LBA_HIGH)
    hwif->outb(regs.lba_high, tf->lbah);

  hwif->outb(regs.command, tf->cmd);

  return 0;
}

uint8_t ata_direct_read_status(ata_device_t *dev)
{
  ata_registers_t *regs = &dev->host_port->regs;
  ata_hwif_t *hwif = dev->host_port->hwif;

  return hwif->inb(regs->status);
}

ata_cmd_t *ata_alloc_init_cmd(ata_device_t *dev, ata_io_protocol_t prot)
{
  ata_cmd_t *cmd;
  short tag = -1;

  if (ata_queued_protocol(prot)) {
    tag = ata_alloc_tag(cmd->dev);
    if (tag == -1)
      return ERR_PTR(-EBUSY);
  }

  /* TO DO [dg]: replace to slab */
  cmd = malloc(sizeof(ata_cmd_t));
  if (cmd == NULL)
    return ERR_PTR(-ENOMEM);

  memset(&cmd->sense_len, 0, sizeof(ata_cmd_t) -
            offsetof(ata_cmd_t, sense_len));
  cmd->dev = dev;
  cmd->protocol = prot;
  cmd->tag = tag;
  cmd->block_shift = dev->param_info.sect_shift;

  cmd->timeout = ATA_DEFAULT_TIMEOUT;

  return cmd;
}

void ata_free_cmd(ata_cmd_t *cmd)
{
  if (cmd->buf)
    free(cmd->buf);
  if (ata_queued_cmd(cmd))
    ata_free_cmd_tag(cmd->dev, cmd);

  free(cmd);
}

short ata_alloc_tag(ata_device_t *dev)
{
  short ret = 0;

  while (ret != -1) {
    ret = bit_find_lsf(dev->tag_map);
    if (ret != -1) {
      bit_set(&dev->tag_map, ret);
      break;
    }
  }

  return ret;
}

int ata_init_port_resources(ata_port_t *port)
{
  int ret;

  ret = -pthread_mutex_init(&port->mutex, NULL);
  /* TO DO [dg]: allocate scatter/gather pool */

  return ret;
}

void ata_free_port_resources(ata_port_t *port)
{
  pthread_mutex_destroy(&port->mutex);
}

static void ata_timing_quantize(const struct ata_timing *t, struct ata_timing *q, int T, int UT)
{
  q->setup   = EZ(t->setup   * 1000,  T);
  q->act8b   = EZ(t->act8b   * 1000,  T);
  q->rec8b   = EZ(t->rec8b   * 1000,  T);
  q->cyc8b   = EZ(t->cyc8b   * 1000,  T);
  q->active  = EZ(t->active  * 1000,  T);
  q->recover = EZ(t->recover * 1000,  T);
  q->cycle   = EZ(t->cycle   * 1000,  T);
  q->udma    = EZ(t->udma    * 1000, UT);
}

void ata_timing_merge(const struct ata_timing *a, const struct ata_timing *b,
          struct ata_timing *m, unsigned int what)
{
  if (what & ATA_TIMING_SETUP  ) m->setup   = MAX(a->setup,   b->setup);
  if (what & ATA_TIMING_ACT8B  ) m->act8b   = MAX(a->act8b,   b->act8b);
  if (what & ATA_TIMING_REC8B  ) m->rec8b   = MAX(a->rec8b,   b->rec8b);
  if (what & ATA_TIMING_CYC8B  ) m->cyc8b   = MAX(a->cyc8b,   b->cyc8b);
  if (what & ATA_TIMING_ACTIVE ) m->active  = MAX(a->active,  b->active);
  if (what & ATA_TIMING_RECOVER) m->recover = MAX(a->recover, b->recover);
  if (what & ATA_TIMING_CYCLE  ) m->cycle   = MAX(a->cycle,   b->cycle);
  if (what & ATA_TIMING_UDMA   ) m->udma    = MAX(a->udma,    b->udma);
}

const struct ata_timing *ata_timing_find_mode(uint8_t mode, bool dma)
{
  const struct ata_timing *t;

  t = dma ? dma_timing : pio_timing;

  while (mode > t->mode)
    t++;

  if (mode == t->mode)
    return t;

  return NULL;
}

int ata_timing_compute(ata_device_t *dev, unsigned short mode, bool dma,
                       struct ata_timing *t, int T, int UT)
{
  const struct ata_timing *s;
  struct ata_timing p;

  /*
   * Find the mode.
   */

  if (!(s = ata_timing_find_mode(mode, dma)))
    return -EINVAL;

  memcpy(t, s, sizeof(*s));

  /*
   * If the drive is an EIDE drive, it can tell us it needs extended
   * PIO/MW_DMA cycle timing.
   */

  if (dev->param_info.word76_64_valid) {
    memset(&p, 0, sizeof(p));
    if (!dma) {
      if (mode <= 2)
        p.cycle = p.cyc8b = dev->param_info.min_pio_cycle;
      else
        p.cycle = p.cyc8b = dev->param_info.min_pio_cycle_iordy;
    } else if (mode <= ATA_MWDMA2) {
      p.cycle = dev->param_info.min_dma_cycle;
    }
    ata_timing_merge(&p, t, t, ATA_TIMING_CYCLE | ATA_TIMING_CYC8B);
  }

  /*
   * Convert the timing to bus clock counts.
   */

  ata_timing_quantize(t, t, T, UT);

  /*
   * Even in DMA/UDMA modes we still use PIO access for IDENTIFY,
   * S.M.A.R.T * and some other commands. We have to ensure that the
   * DMA cycle timing is slower/equal than the fastest PIO timing.
   */

  if (dma) {
    ata_timing_compute(dev, dev->pio_mode, false, &p, T, UT);
    ata_timing_merge(&p, t, t, ATA_TIMING_ALL);
  }

  /*
   * Lengthen active & recovery time so that cycle time is correct.
   */

  if (t->act8b + t->rec8b < t->cyc8b) {
    t->act8b += (t->cyc8b - (t->act8b + t->rec8b)) / 2;
    t->rec8b = t->cyc8b - t->act8b;
  }

  if (t->active + t->recover < t->cycle) {
    t->active += (t->cycle - (t->active + t->recover)) / 2;
    t->recover = t->cycle - t->active;
  }

  /* In a few cases quantisation may produce enough errors to
     leave t->cycle too low for the sum of active and recovery
     if so we must correct this */
  if (t->active + t->recover > t->cycle)
    t->cycle = t->active + t->recover;

  return 0;
}

void ata_normalize_id_str(char *dest, char *src, int len)
{
  int i;

  for (i = len - 1; i >= 0; i--) {
    if (src[i] && src[i] != ' ')
      break;
  }
  if (i >= 0) {
    strncpy(dest, src, i + 1);
  }
  dest[i + 1] = '\0';
}
