/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * Block layer trace implementation; trace events can be obtained
 *      by a client through rpc
 */

/*
 * Binary representation looks ugly and implementation is slightly
 * bulky, but it's required much less memory, and more events can be
 * traced
 */

#include <block/blk_trace.h>
#include <hash.h>
#include <stdlib.h>
#include <string.h>
#include <debug.h>

#define blk_trace_lock(trace) pthread_mutex_lock(&trace->mutex)
#define blk_trace_unlock(trace) pthread_mutex_unlock(&trace->mutex)

/* event control block structure */
struct __evcb {
  trace_event_type_t type;
  trace_layer_t layer;
  void *data;
  uint16_t data_size;
  void (*fill_data)(trace_entry_t*, trace_event_t*);
};

typedef struct __trace_operations {
  list_node_t *(*add_event)(trace_entry_t *te, struct __evcb *evcb);
  int (*event_size)(list_node_t *event_node);
  void (*copy_event)(list_node_t *event_node, void *buf);
  void (*free_event)(list_node_t *event_node);
} trace_operations_t;

static list_node_t *__bin_add_event(trace_entry_t *te, struct __evcb *evcb);
static int __bin_event_size(list_node_t *event_node);
static void __bin_copy_event(list_node_t *event_node, void *buf);
static void __bin_free_event(list_node_t *event_node);

trace_operations_t bin_trace_ops = {
  __bin_add_event,
  __bin_event_size,
  __bin_copy_event,
  __bin_free_event
};

static int __init_trace_sync(blk_rq_trace_t *trace)
{
  pthread_mutexattr_t attr = {PTHREAD_PROCESS_PRIVATE};

  return -pthread_mutex_init(&trace->mutex, &attr);
}

static trace_entry_t *__search_by_ptr(blk_request_t *rq)
{
  ulong_t h;
  list_head_t *head;
  blk_rq_trace_t *trace = rq->queue->rq_trace;
  trace_entry_t *te;

  h = hash_ptr(rq, TRACE_HASH_BITS);
  head = &trace->hash[0][h];
  list_for_each_entry(head, te, hash_node[0]) {
    if (te->rq == rq)
      break;
  }

  if (&te->hash_node[0] == list_head(head))
    te = NULL;

  return te;
}

static trace_entry_t *__search_by_sector(blk_rq_trace_t *trace, sector_t sector)
{
  ulong_t h;
  list_head_t *head;
  trace_entry_t *te;

  h = (ulong_t)hash_64(sector, TRACE_HASH_BITS);
  head = &trace->hash[0][h];

  if (list_is_empty(head))
    return NULL;

  list_for_each_entry(head, te, hash_node[0]) {
    if (te->data.orig_start_sect == sector)
      break;
  }

  if (&te->hash_node[0] == list_head(head))
    te = NULL;

  return te;
}

static trace_entry_t *__search_by_id(blk_rq_trace_t *trace, uint32_t id)
{
  int h;
  list_head_t *head;
  trace_entry_t *te;

  h = id & (TRACE_HASH_TABLE_WIDTH - 1);
  head = &trace->hash[1][h];

  list_for_each_entry(head, te, hash_node[1]) {
    if (te->data.id == id)
      break;
  }

  if (&te->hash_node[1] == list_head(head))
    te = NULL;

  return te;
}

static list_node_t *__bin_add_event(trace_entry_t *te, struct __evcb *evcb)
{
  trace_event_t *ev;

  ev = malloc(sizeof(trace_event_t));
  if (ev == NULL)
    return NULL;

  ev->data.type = evcb->type;
  ev->data.layer = evcb->layer;
  ev->data.flags = te->rq->flags;
  ev->data.prio = te->rq->prio;
  ev->data.ext_data_size = evcb->data_size;
  ev->data_buf = evcb->data;
  if (evcb->data && evcb->fill_data) {
    evcb->fill_data(te, ev);
  }

  return &ev->node;
}

static int __bin_event_size(list_node_t *event_node)
{
  trace_event_t *ev = container_of(event_node, trace_event_t, node);

  return (sizeof(trace_event_data_t) + ev->data.ext_data_size);
}

static void __bin_copy_event(list_node_t *event_node, void *buf)
{
  trace_event_t *ev = container_of(event_node, trace_event_t, node);
  char *p = buf;

  memcpy(p, &ev->data, sizeof(trace_event_data_t));
  if (ev->data.ext_data_size) {
    p += sizeof(trace_event_data_t);
    memcpy(p, ev->data_buf, ev->data.ext_data_size);
  }
}

static void __bin_free_event(list_node_t *event_node)
{
  trace_event_t *ev = container_of(event_node, trace_event_t, node);

  if (ev->data_buf)
    free(ev->data_buf);

  free(ev);
}


static void __del_trace_entry(blk_rq_trace_t *trace, trace_entry_t *te)
{
  list_node_t *node, *save;

  list_del(&te->hash_node[0]);
  list_del(&te->hash_node[1]);
  list_del(&te->fifo_node);
  trace->nr_traced_rqs--;
  list_for_each_safe(&te->events, node, save) {
    trace->ops->free_event(node);
  }
  trace->nr_traced_events -= te->data.nr_events;
  te->data.nr_events = 0;
  list_init_head(&te->events);
}

static void __add_trace_event(blk_request_t *rq, trace_event_type_t type, trace_layer_t layer,
                              void *data, uint16_t data_size,
                              void (*fill_data)(trace_entry_t*, trace_event_t*))
{
  blk_rq_trace_t *trace;
  trace_entry_t *te, *te1;
  list_node_t *node;
  struct __evcb evcb;

  evcb.type = type;
  evcb.layer = layer;
  evcb.data = data;
  evcb.data_size = data_size;
  evcb.fill_data = fill_data;

  trace = rq->queue->rq_trace;
  blk_trace_lock(trace);
  te = __search_by_ptr(rq);
  if (te != NULL) {
    if (trace->nr_traced_events == BLKDEV_MAX_TRACED_EVENTS) {
      /* try to delete one trace entry */
      node = list_node_first(&trace->rq_fifo);
      te1 = container_of(node, trace_entry_t, fifo_node);
      if (te == te1)
        goto out;

      __del_trace_entry(trace, te1);
      free(te1);
    }
    node = trace->ops->add_event(te, &evcb);
    if (node != NULL) {
      list_add2tail(&te->events, node);
      te->data.nr_events++;
      trace->nr_traced_events++;
      te->data.curr_start_sect = rq->start_sector;
      te->data.curr_nr_sects = rq->nr_sectors;
    }
  }
out:
  blk_trace_unlock(trace);
}

static void __fill_merge_data(trace_entry_t *te, trace_event_t *ev)
{
 merge_data_t *data = (merge_data_t*)ev->data_buf;

 data->old_start_sect = te->data.curr_start_sect;
 data->old_nr_sects = te->data.curr_nr_sects;
 data->new_start_sect = te->rq->start_sector;
 data->new_nr_sects = te->rq->nr_sectors;
}

int init_blk_rq_trace(blkreq_queue_t *q)
{
  blk_rq_trace_t *trace;
  int i;

  trace = malloc(sizeof(blk_rq_trace_t));
  if (trace == NULL)
    return -ENOMEM;

  i = __init_trace_sync(trace);
  if (i) {
    free(trace);
    return i;
  }

  for (i = 0; i < TRACE_HASH_TABLE_WIDTH; i++) {
    list_init_head(&trace->hash[0][i]);
    list_init_head(&trace->hash[1][i]);
  }
  list_init_head(&trace->rq_fifo);
  trace->nr_traced_rqs = 0;
  trace->nr_traced_events = 0;
  q->rq_trace = trace;

  trace->ops = &bin_trace_ops;

  return 0;
}

void add_trace_rq(blk_request_t *rq)
{
  blk_rq_trace_t *trace;
  trace_entry_t *te;
  ulong_t h1, h2;

  trace = rq->queue->rq_trace;
  blk_trace_lock(trace);
  if (trace->nr_traced_rqs == BLKDEV_MAX_TRACED_REQUESTS) {
    te = container_of(list_node_first(&trace->rq_fifo), trace_entry_t, fifo_node);
    __del_trace_entry(trace, te);
  } else {
    te = malloc(sizeof(trace_entry_t));
    if (te == NULL) {
      blk_trace_unlock(trace);
      return;
    }
  }
  te->data.id = trace->last_id++;
  te->rq = rq;
  te->data.orig_start_sect = te->data.curr_start_sect = rq->start_sector;
  te->data.orig_nr_sects = te->data.curr_nr_sects = rq->nr_sectors;
  te->data.nr_events = 0;
  list_init_head(&te->events);
  h1 = hash_ptr(rq, TRACE_HASH_BITS);
  h2 = te->data.id & (TRACE_HASH_TABLE_WIDTH - 1);
  /* add to the table */
  list_add2tail(&trace->hash[0][h1], &te->hash_node[0]);
  list_add2tail(&trace->hash[1][h2], &te->hash_node[1]);
  list_add2tail(&trace->rq_fifo, &te->fifo_node);
  trace->nr_traced_rqs++;
  blk_trace_unlock(trace);
}

void end_trace_rq(blk_request_t *rq)
{
  blk_rq_trace_t *trace;
  trace_entry_t *te;
  ulong_t h1, h2;

  /* rehash the request by sector */
  trace = rq->queue->rq_trace;
  blk_trace_lock(trace);
  te = __search_by_ptr(rq);
  if (te != NULL) {
    h1 = hash_ptr(rq, TRACE_HASH_BITS);
    h2 = (ulong_t)hash_64(te->data.orig_start_sect, TRACE_HASH_BITS);
    if (h1 != h2) {
      list_del(&te->hash_node[0]);
      list_add2tail(&trace->hash[0][h2], &te->hash_node[0]);
    }
    te->rq = NULL;
  }
  blk_trace_unlock(trace);
}

void trace_rq_iosched_add(blk_request_t *rq)
{
  __add_trace_event(rq, TRACE_EVENT_ADD, TRACE_LAYER_IOSCHED, NULL, 0, NULL);
}

void trace_rq_merge_rq(blk_request_t *rq_front, blk_request_t *rq_tail)
{
  merge_data_t *data;
  trace_entry_t *te;
  blk_rq_trace_t *trace = rq_front->queue->rq_trace;

  data = malloc(sizeof(merge_data_t));
  if (data == NULL)
    return;

  data->with_blkio = false;
  data->type = BLK_MERGE_TAIL;
  data->m_prio = rq_tail->prio;

  blk_trace_lock(trace);
  te = __search_by_ptr(rq_tail);
  blk_trace_unlock(trace);
  data->rq_id = te->data.id;
  __add_trace_event(rq_front, TRACE_EVENT_MERGE, TRACE_LAYER_IOSCHED,
                    data, sizeof(merge_data_t), __fill_merge_data);
}

void trace_rq_merge_blkio(blk_request_t *rq, blkio_t *blkio, short m_prio, merge_type_t where)
{
  merge_data_t *data;

  data = malloc(sizeof(merge_data_t));
  if (data == NULL)
    return;

  data->with_blkio = true;
  data->type = where;
  data->m_prio = m_prio;

  __add_trace_event(rq, TRACE_EVENT_MERGE, TRACE_LAYER_IOSCHED,
                    data, sizeof(merge_data_t), __fill_merge_data);
}

void trace_rq_dispatch(blk_request_t *rq, trace_layer_t layer)
{
  __add_trace_event(rq, TRACE_EVENT_DISPATCH, layer, NULL, 0, NULL);
}

void trace_rq_starve(blk_request_t *rq, trace_layer_t layer)
{
  __add_trace_event(rq, TRACE_EVENT_STARVE, layer, NULL, 0, NULL);
}

void trace_rq_schedule(blk_request_t *rq)
{
  __add_trace_event(rq, TRACE_EVENT_SCHEDULE, TRACE_LAYER_IOSCHED, NULL, 0, NULL);
}

void trace_rq_start_proc(blk_request_t *rq)
{
  __add_trace_event(rq, TRACE_EVENT_START_PROC, TRACE_LAYER_DRV, NULL, 0, NULL);
}

void trace_rq_end_proc(blk_request_t *rq)
{
  __add_trace_event(rq, TRACE_EVENT_END_PROC, TRACE_LAYER_DRV, NULL, 0, NULL);
}

ssize_t get_trace_events(blkreq_queue_t *q, uint64_t key, skey_type_t key_type,
                         void *buf, size_t buf_size)
{
  trace_entry_t *te;
  trace_entry_data_t *tdata;
  size_t size = buf_size, s;
  char *p = (char*)buf;
  list_node_t *node, *save;
  blk_rq_trace_t *trace = q->rq_trace;

  if (buf_size < sizeof(trace_entry_data_t))
    return -EINVAL;

  blk_trace_lock(trace);
  if (key_type == SKEY_TYPE_SECTOR) {
    te = __search_by_sector(q->rq_trace, key);
  } else {
    te = __search_by_id(q->rq_trace, key);
  }

  if (te == NULL) {
    blk_trace_unlock(trace);
    return -ESRCH;
  }

  memcpy(p, &te->data, sizeof(trace_entry_data_t));
  tdata = (trace_entry_data_t*)p;
  buf_size -= sizeof(trace_entry_data_t);
  p += sizeof(trace_entry_data_t);

  list_for_each_safe(&te->events, node, save) {
    s = trace->ops->event_size(node);
    if (s > buf_size)
      break;

    trace->ops->copy_event(node, p);
    p += s;
    buf_size -= s;
    trace->ops->free_event(node);
    te->data.nr_events--;
    trace->nr_traced_events--;
  }

  tdata->nr_events -= te->data.nr_events;
  if (!te->data.nr_events) {
    /* all memory consumed by events is already freed */
    list_init_head(&te->events);
    __del_trace_entry(trace, te);
    free(te);
  } else {
    /* link list */
    list_head(&te->events)->next = node;
    node->prev = list_head(&te->events);
  }
  blk_trace_unlock(trace);

  return (ssize_t)(size - buf_size);
}

int clear_blk_rq_trace(blkreq_queue_t *q)
{
  trace_entry_t *te;
  blk_rq_trace_t *trace = q->rq_trace;
  list_node_t *node, *save;

  blk_trace_lock(trace);
  list_for_each_safe(&trace->rq_fifo, node, save) {
    te = container_of(node, trace_entry_t, fifo_node);
    __del_trace_entry(trace, te);
    free(te);
  }
  trace->last_id = 0;
  blk_trace_unlock(trace);

  return 0;
}
