/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Copyright (c) 2002 Jens Axboe <axboe@kernel.dk>
 * Deadline I/O block scheduler
 *
 * Ported by Dmitry Gromada <gromada@jarios.org>
 */

/*
 * Main differences from original implementation:
 *
 * 1) Requests are sorted by range of sectors that make it
 *    possible to search effectively mergeable requests as
 *    well at front as at tail
 * 2) The scheduler controls request dispatching itself
 * 3) After dispatching to the driver queue requests
 *    remain in the scheduler if the driver provides
 *    merging of queued requests.
 */

#include <datalib/usrtc.h>
#include <pthread.h>
#include <sys/time.h>
#include <time.h>
#include <signal.h>
#include <debug.h>
#include <errno.h>
#include <block/iosched.h>
#include <block/blk_trace.h>
#include <sys/scheduler.h>
#include "server.h"

enum {
  READ,
  WRITE
};

#if BITS_PER_LONG == 64
#define TICKS_INFINITE (~0L)
#else
#define TICKS_INFINITE (~0LL)
#endif

static const int default_read_expire = 2;     /* max time before a read is submitted (in milliseconds) */
static const int default_write_expire = 5;    /* ditto for writes, these limits are SOFT! */
static const int default_writes_starved = 2;  /* max times reads can starve a write */
static const int default_fifo_batch = 16;     /* # of sequential requests treated as one
				                                        by the above parameters. For throughput. */

static swks_t *swks;

typedef struct __deadline_data {
	/*
	 * run time data
	 */

	/*
	 * requests are present on both sort_list and fifo
	 */
  usrtc_t rbtree[2];
  list_head_t fifo[2];

	/*
	 * next in sort order. read, write or both are NULL
	 */
	blk_request_t *next_rq[2];
	unsigned int batching;		/* number of sequential requests made */
	sector_t last_sector;		/* head position */
	unsigned int starved;		/* times reads have starved writes */

	/*
	 * settings that change how the i/o scheduler behaves
	 */
  int fifo_expire[2];
	int fifo_batch;
	int writes_starved;
  bool reinsert_on_merge;
  bool remove_on_dispatch;
} deadline_data_t;

struct __rq_data {
  uint64_t expire_ticks;
  sector_t tail_sector;
  list_node_t fifo_node;
  usrtc_node_t rbnode;
};

static int deadline_init(iosched_t *sched);
static void deadline_release(iosched_t *sched);
static int deadline_insert_request(iosched_t *sched, blk_request_t *rq);
static int deadline_remove_request(iosched_t *sched, blk_request_t *rq);
static int deadline_dispatch_rq(iosched_t *sched, blk_request_t *rq);
static int deadline_flush_queue(iosched_t *sched, bool force);
static void deadline_find_lock_mergeable(iosched_t *sched, blkio_t *blkio,
                                         blk_request_t **rq_front, blk_request_t **rq_tail);
static void deadline_merged_blkio(iosched_t *sched, blk_request_t *rq, merge_type_t where);
static void deadline_merged_requests(iosched_t *sched, blk_request_t *rq_front,
                                     blk_request_t *rq_tail);

static iosched_ops_t deadline_ops = {
  .init = deadline_init,
  .release = deadline_release,
  .insert = deadline_insert_request,
  .remove = deadline_remove_request,
  .dispatch_rq = deadline_dispatch_rq,
  .flush_queue = deadline_flush_queue,
  .find_lock_mergeable = deadline_find_lock_mergeable,
  .post_merge_blkio_fn = deadline_merged_blkio,
  .post_merge_rqs_fn = deadline_merged_requests
};

static void __remove_request(deadline_data_t *dd, blk_request_t *rq);
static void __add_to_queue(deadline_data_t *dd, blk_request_t *rq);
static blk_request_t *__latter_request(deadline_data_t *dd, blk_request_t *rq, int data_dir);

static int __compare_requests(const void *a, const void *b)
{
  blk_request_t *rq1 = (blk_request_t*)a;
  blk_request_t *rq2 = (blk_request_t*)b;

  if (rq1->start_sector + rq1->nr_sectors <= rq2->start_sector) {
    return -1;
  } else if (rq1->start_sector >= rq2->start_sector + rq2->nr_sectors) {
    return 1;
  } else {
    return 0;
  }
}

static int __compare_rq_mergeable(const void *a, const void *b)
{
  blkio_t *blkio = (blkio_t*)a;
  blk_request_t *rq = (blk_request_t*)b;

  if (blkio->start_sector + blkio->nr_sectors < rq->start_sector) {
    return -1;
  } else if (blkio->start_sector > rq->start_sector + rq->nr_sectors) {
    return 1;
  } else {
    return 0;
  }
}

static bool __deadline_is_conguested(deadline_data_t *dd, int data_dir)
{
  return (usrtc_count(&dd->rbtree[data_dir]) ==
          dd->rbtree[data_dir].usrtc_maxcount);
}

static uint64_t __next_expire_rq_ticks(deadline_data_t *dd)
{
  list_node_t *node;
  struct __rq_data *d;
  uint64_t ticks[2] = {TICKS_INFINITE, TICKS_INFINITE};
  int i;

  for (i = 0; i < 2; i++) {
    if (!list_is_empty(&dd->fifo[i])) {
      node = list_node_first(&dd->fifo[i]);
      d = container_of(node, struct __rq_data, fifo_node);
      ticks[i] = d->expire_ticks;
    }
  }

  return MIN(ticks[0], ticks[1]);
}

static inline void __replace_fifo_node(list_node_t *dst, list_node_t *src)
{
  list_del(src);
  list_add_range(src, src, dst->prev, dst->next);
  list_init_node(dst);
}

static void __add_rq_rb(deadline_data_t *dd, blk_request_t *rq, int data_dir)
{
  usrtc_node_t *node, *alias;
  struct __rq_data *rd = rq->sched_data;
  struct __rq_data *rd_alias;
  blk_request_t *rq1;

  node = &rd->rbnode;

  alias = usrtc_insert(&dd->rbtree[data_dir], node, rq);
  if (alias != node) {
    rq1 = usrtc_node_getdata(alias);

    clear_rq_flags(rq1, BLKREQ_SCHEDULED);
    if (!blk_queued_rq(rq1)) {
      __add_to_queue(dd, rq1);
    }

    /* replace rb and fifo nodes, and iosched data */
    rd_alias = rq1->sched_data;
    rq1->sched_data = NULL;
    rq->sched_data = rd_alias;

    __replace_fifo_node(&rd->fifo_node, &rd_alias->fifo_node);
    rd_alias->expire_ticks = rd->expire_ticks;
    free(rd);

    alias->usrtc_node_key = rq;
    usrtc_node_setdata(alias, rq);

    put_blk_request(rq1);
    usrtc_insert(&dd->rbtree[data_dir], node, rq);
	}
}

static void __del_rq_rb(deadline_data_t *dd, blk_request_t *rq, int data_dir)
{
  struct __rq_data *rd = rq->sched_data;

  if (dd->next_rq[data_dir] == rq)
		dd->next_rq[data_dir] = __latter_request(dd, rq, data_dir);

  usrtc_delete(&dd->rbtree[data_dir], &rd->rbnode);
}

static void __remove_request(deadline_data_t *dd, blk_request_t *rq)
{
  struct __rq_data *rd = (struct __rq_data*)rq->sched_data;
  int data_dir = blkreq_rw(rq);

  clear_rq_flags(rq, BLKREQ_SCHEDULED);
	if (!list_node_next_isbound(&rd->fifo_node)) {
    list_del(&rd->fifo_node);
  }
  __del_rq_rb(dd, rq, data_dir);
  rq->sched_data = NULL;
  free(rd);
  put_blk_request(rq);
}

static blk_request_t *__latter_request(deadline_data_t *dd, blk_request_t *rq, int data_dir)
{
  blk_request_t *latter = NULL;
  usrtc_node_t *node;
  struct __rq_data *rd = rq->sched_data;

  node = usrtc_next(&dd->rbtree[data_dir], &rd->rbnode);
  if (node != NULL)
    latter = usrtc_node_getdata(node);

  return latter;
}

static void __add_to_queue(deadline_data_t *dd, blk_request_t *rq)
{
	const int data_dir = blkreq_rw(rq);

  dd->next_rq[READ] = NULL;
	dd->next_rq[WRITE] = NULL;
  dd->next_rq[data_dir] = __latter_request(dd, rq, data_dir);

	dd->last_sector = rq->start_sector + rq->nr_sectors;

  trace_rq_dispatch(rq, TRACE_LAYER_IOSCHED);
  dispatch_request_nolock_ev(rq, QUEUE_TAIL);
}

/*
 * deadline_check_fifo returns 0 if there are no expired reads on the fifo,
 * 1 otherwise. Requires !list_is_empty(&dd->fifo[data_dir])
 */
static inline int deadline_check_fifo(deadline_data_t *dd, int ddir)
{
  list_node_t *node;
  struct __rq_data *d;
  uint64_t curr_ticks;

  node = list_node_first(&dd->fifo[ddir]);
  d = container_of(node, struct __rq_data, fifo_node);

  curr_ticks = current_sys_ticks(swks);
	/*
	 * rq is expired?
	 */
	if (curr_ticks >= d->expire_ticks)
    return 1;

	return 0;
}

static inline blk_request_t *__fifo_next_rq(deadline_data_t *dd, int data_dir)
{
  list_node_t *node;
  blk_request_t *rq;
  struct __rq_data *rd;

  node = list_node_first(&dd->fifo[data_dir]);
  rd = container_of(node, struct __rq_data, fifo_node);
  rq = usrtc_node_getdata(&rd->rbnode);

  return rq;
}


/*
 * deadline_dispatch_requests selects the best request according to
 * read/write expire, fifo_batch, etc
 */
static int __dispatch_request(deadline_data_t *dd, int force)
{
	const int reads = !list_is_empty(&dd->fifo[READ]);
	const int writes = !list_is_empty(&dd->fifo[WRITE]);
  blk_request_t *rq;
	int data_dir;
  struct __rq_data *rd;

	/*
	 * batches are currently reads XOR writes
	 */
	if (dd->next_rq[WRITE])
		rq = dd->next_rq[WRITE];
	else
		rq = dd->next_rq[READ];

	if (rq) {
		/* we have a "next request" */

		if (dd->last_sector != rq->start_sector)
			/* end the batch on a non sequential request */
			dd->batching += dd->fifo_batch;

		if (dd->batching < dd->fifo_batch)
			/* we are still entitled to batch */
      goto dispatch_request;
	}

	/*
	 * at this point we are not running a batch. select the appropriate
	 * data direction (read / write)
	 */

	if (reads) {
		if (writes && (dd->starved++ >= dd->writes_starved))
			goto dispatch_writes;

		data_dir = READ;

		goto dispatch_find_request;
	}

	/*
	 * there are either no reads or writes have been starved
	 */

	if (writes) {
dispatch_writes:
		dd->starved = 0;

		data_dir = WRITE;

		goto dispatch_find_request;
	}

	return 0;

dispatch_find_request:
	/*
	 * we are not running a batch, find best request for selected data_dir
	 */
	if (deadline_check_fifo(dd, data_dir)) {
		/*
		 * A deadline has expired or the block layer force dispatching,
     * the last request was in the other direction, or we have run
     * out of higher-sectored requests. Start again from the request
     * with the earliest expiry time.
		 */
    rq = __fifo_next_rq(dd, data_dir);
	} else if (force) {
		if (dd->next_rq[data_dir] != NULL)
      rq = dd->next_rq[data_dir];
    else
      rq = __fifo_next_rq(dd, data_dir);
	} else {
    rq = NULL;
  }

	dd->batching = 0;

  if (rq == NULL)
    return 0;

dispatch_request:
	/*
	 * rq is the selected appropriate request.
	 */
  rd = rq->sched_data;
  //ASSERT((swks->system_clock_ticks < rd->expire_ticks + 500),
  //        "rd->expire_ticks = %d, system_ticks = %d\n",
  //        rd->expire_ticks, swks->system_clock_ticks);

	dd->batching++;
  __add_to_queue(dd, rq);
  if (dd->remove_on_dispatch) {
    __remove_request(dd, rq);
  }

	return 1;
}

static bool __check_rq_mergeable_lock(blk_request_t *rq, blkio_t *blkio)
{
  bool ret = true;

  ASSERT(rq != NULL, "rq = NULL\n");

  if (blk_rq_lock(rq))
    return false;

  if (blk_rq_is_mergeable(rq)) {
    ret = (rq->start_sector + rq->nr_sectors == blkio->start_sector ||
           blkio->start_sector + blkio->nr_sectors == rq->start_sector);
  } else {
    ret = false;
  }

  if (!ret)
    blk_rq_unlock(rq);

  return ret;
}


/* interface functions implementation */

/*
 * initialize scheduler private data (deadline_data)
 */
static int deadline_init(iosched_t *sched)
{
	deadline_data_t *dd;
  blkreq_queue_t *q = sched->queue;

	dd = malloc(sizeof(deadline_data_t));
  if (dd == NULL)
    return -ENOMEM;

  sched->private = dd;
	list_init_head(&dd->fifo[READ]);
	list_init_head(&dd->fifo[WRITE]);
  usrtc_init(&dd->rbtree[READ], USRTC_REDBLACK,
             q->max_requests * 2, __compare_requests);
  usrtc_init(&dd->rbtree[WRITE], USRTC_REDBLACK,
             q->max_requests * 2, __compare_requests);
	swks = get_swks();
  dd->fifo_expire[READ] = default_read_expire;
  dd->fifo_expire[WRITE] = default_write_expire;
  dd->writes_starved = default_writes_starved;
	dd->fifo_batch = default_fifo_batch;
  dd->reinsert_on_merge = true;
  dd->remove_on_dispatch = !(sched->queue->merge_rq || sched->queue->merge_requests);

  return 0;
}

static void deadline_release(iosched_t *sched)
{
	deadline_data_t *dd = sched->private;

	ASSERT(list_is_empty(&dd->fifo[READ]), "fifo is not empty");
	ASSERT(list_is_empty(&dd->fifo[WRITE]), "fifo is not empty");

	free(dd);
}

/*
 * add rq to rbtrees and fifo, and set timer if the scheduler is empty
 */
static int deadline_insert_request(iosched_t *sched, blk_request_t *rq)
{
	deadline_data_t *dd = (deadline_data_t*)sched->private;
	const int data_dir = blkreq_rw(rq);
  struct __rq_data *rd;
  uint64_t curr_ticks;

  if (blk_scheduled_rq(rq))
    return 0;

  rd = malloc(sizeof(struct __rq_data));
  if (rd == NULL)
    return -ENOMEM;

  usrtc_node_init(&rd->rbnode, rq);
  if (__deadline_is_conguested(dd, data_dir)) {
    free(rd);
    return -EBUSY;
  }
  curr_ticks = current_sys_ticks(swks);
  rd->expire_ticks = curr_ticks + msec_to_ticks(swks, dd->fifo_expire[data_dir]);
  rq->sched_data = rd;
  get_blk_request(rq);
	list_add2tail(&dd->fifo[data_dir], &rd->fifo_node);
  __add_rq_rb(dd, rq, data_dir);
  set_rq_flags(rq, BLKREQ_SCHEDULED);
  /*
   * if the timer is already expired or expires later
   * than the request deadline, set it
   */
  if ((sched->tmr.exp_ticks <= curr_ticks) ||
      (rd->expire_ticks < sched->tmr.exp_ticks)) {

    set_timeout(&sched->tmr, dd->fifo_expire[data_dir]);
  }

  return 0;
}

static int deadline_remove_request(iosched_t *sched, blk_request_t *rq)
{
  deadline_data_t *dd = sched->private;
  int ret = 0;

  /*
   * check the req is yet in the scheduler,
   * in may be removed concurrently
   */
  if (!blk_scheduled_rq(rq))
    ret = -1;
  else
    __remove_request(dd, rq);

  return ret;
}

static int deadline_dispatch_rq(iosched_t *sched, blk_request_t *rq)
{
  deadline_data_t *dd = sched->private;

  __add_to_queue(dd, rq);
  if (dd->remove_on_dispatch) {
    __remove_request(dd, rq);
  }

  return 0;
}

static int deadline_flush_queue(iosched_t *sched, bool force)
{
  deadline_data_t *dd = sched->private;
  uint64_t ticks, curr_ticks;

  while (1) {
    while (__dispatch_request(dd, force));
    blk_queue_signal(sched->queue);
    /*
     * set timer again if there are still scheduled request,
     * at time expiration flush the queue again
     */
    ticks = __next_expire_rq_ticks(dd);
    if (ticks == TICKS_INFINITE) {
      /* there are not more scheduled requests, end flushing */
      break;
    }

    curr_ticks = current_sys_ticks(swks);
    ticks -= curr_ticks;
    if (ticks && ticks < curr_ticks) {
      set_timeout_ticks(&sched->tmr, ticks);
      break;
    }
  }

  return 0;
}

static void deadline_find_lock_mergeable(iosched_t *sched, blkio_t *blkio,
                                         blk_request_t **rq_front,
                                         blk_request_t **rq_tail)
{
  deadline_data_t *dd = sched->private;
  int data_dir = biorq_rw(blkio->biorq);
  usrtc_t *us;
  usrtc_node_t *node;
  blk_request_t *rq;
  int i;

  us = &dd->rbtree[data_dir];
  *rq_front = NULL;
  *rq_tail = NULL;

  us->usrtc_compare = __compare_rq_mergeable;
  node = usrtc_lookup(us, blkio);
  for (i = 0; (i < 2) && (node != NULL); i++) {
    rq = usrtc_node_getdata(node);
    if (__check_rq_mergeable_lock(rq, blkio)) {
      if (rq->start_sector < blkio->start_sector)
        *rq_front = rq;
      else
        *rq_tail = rq;
    }

    if (!i) {
      if (rq->start_sector <= blkio->start_sector)
        node = usrtc_next(us, node);
      else
        node = usrtc_prev(us, node);
    }
  }
  us->usrtc_compare = __compare_requests;
}

static void deadline_merged_blkio(iosched_t *sched, blk_request_t *rq,
                                  merge_type_t where)
{
	deadline_data_t *dd = (deadline_data_t*)sched->private;
  const int data_dir = blkreq_rw(rq);

  /* it's need request reposition */
  if (dd->reinsert_on_merge) {
    if (blk_scheduled_rq(rq)) {
      __del_rq_rb(dd, rq, data_dir);
      __add_rq_rb(dd, rq, data_dir);
    }
  }
}

static void deadline_merged_requests(iosched_t *sched, blk_request_t *rq,
			                               blk_request_t *next)
{
  deadline_data_t *dd = (deadline_data_t*)sched->private;
  struct __rq_data *rd = rq->sched_data;
  struct __rq_data *rd_next = next->sched_data;
  int data_dir = blkreq_rw(rq);

	/*
	 * if next expires before rq, assign its expire time to rq
	 * and move into next position (next will be deleted) in fifo
	 */
	if (blk_scheduled_rq(rq) && blk_scheduled_rq(next)) {
		if (rd_next->expire_ticks < rd->expire_ticks) {
      /* move the node at the place of next */
      __replace_fifo_node(&rd_next->fifo_node, &rd->fifo_node);
      rd->expire_ticks = rd_next->expire_ticks;
		}
    /*
	   * kill knowledge of next, this one is a goner
	   */
	  __remove_request(dd, next);
    if (dd->reinsert_on_merge) {
      __del_rq_rb(dd, rq, data_dir);
      __add_rq_rb(dd, rq, data_dir);
    }
	}
}

const char *register_deadline_iosched(iosched_ops_t **ops)
{
  *ops = &deadline_ops;

  return "deadline";
}
