/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 *
 * Generic storage implementation
 */

/*
 * NOTE: the generic block layer shouldn't handle BLKC_GET_METADATA ioctl,
 *  the server layer does that
 */

/*
 * TO DO [dg]:
 * 1) check if the block device is invalidated or removed
 * 2) release storage
 */

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <sys/blkdev.h>
#include <devio/devio.h>
#include <block/blkdev.h>
#include <block/bdev_drv.h>
#include <block/genstor.h>
#include <block/request.h>
#include <block/blk_trace.h>
#include <block/server.h>
#include <errno.h>
#include <debug.h>

static void __complete_double_end_biorq(blkio_t *blkio)
{
  biorq_t *biorq = blkio->biorq;
  void *dst;
  uint32_t len, dst_off;
  blkreq_queue_t *q = (blkreq_queue_t*)biorq->completion_data;
  int ret;
  short prio = 0;

  dst_off = biorq->wb_chunk.off;
  len = biorq->wb_chunk.size;

  dst = (void*) ((ulong_t)blkio->buf + dst_off);
  if (biorq_buffered(biorq)) {
    memcpy(dst, biorq->data_buf, len);
  } else {
    copy_biorq_data(biorq, 0, len, dst);
  }
  set_biorq_flags(biorq, BIORQ_RW | BIORQ_DATA_AVAILABLE);
  clear_biorq_flags(biorq, BIORQ_DOUBLE_END);

  if (biorq_serv_prepared(biorq)) {
    if (!biorq_aborted(biorq) && biorq_session_is_active(biorq)) {
      biorq->complete_sectors = 0;
      biorq->complete = default_complete_biorq;
    } else {
      biorq->complete = NULL;
      put_server_biorq(biorq);
    }
  } else {
    biorq->complete = NULL;
  }

  /* restart the request */
  /* Fixme [dg]: dispatch the request to the top of queue */
  set_biorq_flags(biorq, BIORQ_NOT_SCHEDULE);
  if (q->last_proc_rq) {
    prio = q->last_proc_rq->prio;
  }
  if (q->io_request_fn) {
    ret = q->io_request_fn(q, blkio, prio);
  } else {
    ret = generic_submit_blkio(q, blkio, prio);
  }

  if (ret) {
    blkio->status.err = -ret;
    if (biorq->complete)
      biorq->complete(blkio);
  }
}

int generic_open_storage(genstor_t *storage, uint16_t partno, uint16_t flags)
{
  int ret = 0;
  blkreq_queue_t *q = (blkreq_queue_t*)storage->private;
  block_device_t *bdev = q->bdev;

  if (!storage->openers && bdev->ops->open)
    ret = bdev->ops->open(bdev);

  if (!ret) {
    bdev->openers++;
    storage->openers++;
    if (partno)
      storage->part_opened++;
  }

  return ret;
}


int generic_close_storage(genstor_t *storage, uint16_t partno)
{
  blkreq_queue_t *q = (blkreq_queue_t*)storage->private;
  block_device_t *bdev = q->bdev;

  bdev->openers--;
  storage->openers--;
  if (partno)
    storage->part_opened--;

  if (!storage->openers && bdev->ops->close)
    bdev->ops->close(bdev);

  return 0;
}

ssize_t generic_read_storage(genstor_t *stor, bio_rpc_t *bio_rpc,
                             biorq_t *srv_biorq)
{
  ssize_t ret = 0;
  biorq_t *b;
  size_t size = bio_rpc->size;

  b = alloc_init_biorq(srv_biorq->start_sector, srv_biorq->nr_sectors,
                       BIORQ_TYPE_USER, srv_biorq->flags);
  if (b == NULL) {
    return -ENOMEM;
  }

  ret = server_prepare_biorqs(&b, 1, srv_biorq);
  if (ret)
    goto out;

  /* submit requests to the generic block layer */
  b->complete = default_complete_biorq;
  ret = stor->ops->make_request(stor, bio_rpc->part, b, NULL, bio_rpc->prio);

out:
  if (ret) {
    put_server_biorq(b);
  } else {
    if (!biorq_serv_prepared(b)) {
      put_biorq(b);
    }
    ret = size;
  }

  return ret;
}

ssize_t generic_write_storage(genstor_t *stor, bio_rpc_t *bio_rpc,
                              biorq_t *srv_biorq)
{
  ssize_t ret = 0;
  biorq_t *bio[3], *b, *b1;
  int i, n = 0;
  uint32_t sect_size;
  size_t size = bio_rpc->size;

  sect_size = 1 << stor->sect_shift;

  b = alloc_init_biorq(srv_biorq->start_sector, srv_biorq->nr_sectors,
                       BIORQ_TYPE_USER, srv_biorq->flags);
  if (b == NULL)
    return -ENOMEM;

  /* make requests */

  /*
   * if the request is not sector aligned, split it
   * on several parts:
   *
   * read request for non full sectors and write request for full
   * sectors, after the read requests will be done, the sectors
   * part will be copied in buffer and write request will be
   * submitted
   */
  b1 = check_clusterize_biorq(b, bio_rpc->offset, size, sect_size,
                              BLK_SPLIT_FRONT);
  if (b1 != NULL) {
    if (IS_ERR(b1)) {
      ret = PTR_ERR(b1);
      goto out;
    }
    bio[n++] = b1;
  }

  if (b1 != b) {
    bio[n++] = b;

    b1 = check_clusterize_biorq(b, bio_rpc->offset, size, sect_size,
                                BLK_SPLIT_TAIL);
    if (b1 != NULL && b1 != b) {
      if (IS_ERR(b1)) {
        ret = PTR_ERR(b1);
        goto out;
      }
      bio[n++] = b1;
    }
  }

  ret = server_prepare_biorqs(bio, n, srv_biorq);
  if (ret)
    goto out;

  /* submit requests to the generic block layer */
  for (i = 0; (i < n) && !ret; i++) {
    if (biorq_double_end(bio[i])) {
      bio[i]->completion_data = stor->private;
      bio[i]->complete = __complete_double_end_biorq;
    } else if (biorq_need_reply(bio[i])) {
      bio[i]->complete = default_complete_biorq;
    }
    ret = stor->ops->make_request(stor, bio_rpc->part, bio[i], NULL, bio_rpc->prio);
  }

out:
  if (ret) {
    for (i = 0; i < n; i++) {
      put_server_biorq(bio[i]);
    }
  } else {
    if (!biorq_serv_prepared(b)) {
      put_biorq(b);
    }
    ret = size;
  }

  return ret;
}

ssize_t generic_ioctl_storage(genstor_t *storage, uint16_t partno, uint16_t cmd, void *buf,
                              size_t data_size, size_t buf_size)
{
  ssize_t ret = 0;
  blkreq_queue_t *q = (blkreq_queue_t*)storage->private;

  if (_IOCMD_TYPE(cmd) != GENSTOR_CTLS) {
    block_device_t *bdev = q->bdev;

    if (bdev->ops->ioctl)
      ret = bdev->ops->ioctl(bdev, cmd, buf, data_size, buf_size);
    else
      ret = -ENOSYS;

    return ret;
  }

  switch (cmd) {
    case BLKCMD_RESCAN_PARTITIONS:
      if (storage->ops->scan_partitions == NULL) {
        ret = -ENOSYS;
      } else {
        if (storage->part_opened) {
          ret = -EBUSY;
        } else {
          ret = storage->ops->scan_partitions(storage);
        }
      }
      break;
    case BLKCMD_GET_TRACE_EVENTS:
    {
      struct trace_key *tkey = (struct trace_key*)buf;
      blkreq_queue_t *q = storage->private;

      ret = get_trace_events(q, tkey->value, tkey->type, buf, buf_size);
      break;
    }
    case BLKCMD_CLEAR_TRACE:
      ret = clear_blk_rq_trace((blkreq_queue_t*)storage->private);
      break;
    default:
      ret = -EINVAL;
      break;
  }

  return ret;
}

int init_generic_storage(genstor_t *storage, char *name,
                         storage_operations_t *ops,
                         blkreq_queue_t *q, bool cacheable)
{
  block_device_t *bdev = q->bdev;

  memset(storage, 0, sizeof(genstor_t));
  get_blk_queue(q);
  storage->private = q;
  storage->capacity = bdev->media.capacity;
  storage->sect_shift = bdev->media.sect_shift;
  storage->ops = ops;
  strcpy(storage->name, name);
  if (bdev->media.flags & MEDIA_RDONLY)
    storage->flags |= STORAGE_RDONLY;

  if (cacheable) {
    /* TO DO [dg]: intialize page cache */
  } else {
    storage->cache_id = -1;
  }

  return 0;
}

void release_generic_storage(genstor_t *stor)
{
}
