/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * Block I/O scheduler API implementation
 */

#include <block/genstor.h>
#include <block/iosched.h>
#include <block/blk_trace.h>
#include <errno.h>

extern const char *register_deadline_iosched(iosched_ops_t **ops);

/*
 * Array of scheduler init functions.
 * Add here your own function so that your scheduler
 * could be registered automatically at initialization
 * of block schedulers.
 *
 * Each such function must return scheduler name and
 * write pointer to operations to the 'ops' argument
 */
static const char *(*iosched_init_fn[])(iosched_ops_t **ops) = {
#ifdef CONFIG_DEADLINE_IOSCHED
  register_deadline_iosched,
#endif
  NULL
};

static list_head_t schedulers;
static const char *default_scheduler = CONFIG_DEFAULT_IOSCHED;

static short __recalc_rq_prio(iosched_t *sched, blk_request_t *rq, short m_prio)
{
  short prio;

  if (sched->ops->calc_prio)
    prio = sched->ops->calc_prio(sched, rq, m_prio);
  else
    prio = MAX(rq->prio, m_prio);

  return prio;
}

static void __reschedule_rq(iosched_t *sched, blk_request_t *rq, short prio)
{
  blkreq_queue_t *q = rq->queue;

  if (rq->prio == prio)
    return;

  if (blk_queued_rq(rq)) {
    blk_queue_lock(q);
    list_del(&rq->node);
    rq->prio = prio;
    q->nr_requests--;
    dispatch_request_nolock_ev(rq, QUEUE_TAIL);
    if (q->nr_requests == 1) {
      blk_queue_signal(q);
    }
    blk_queue_unlock(q);
  } else {
    if (sched->ops->set_prio) {
      sched->ops->set_prio(sched, rq, prio);
    } else {
      rq->prio = prio;
    }
  }
}

bool blk_rq_is_mergeable(blk_request_t *rq)
{
  blkreq_queue_t *q;
  bool ret = true;

  if (blk_launched_rq(rq) || blk_rq_nomerge(rq))
    return false;

  if (blk_queued_rq(rq)) {
    q = rq->queue;
    ret = (q->merge_rq != NULL);
  }

  return ret;
}

static uint32_t __merge_blkio(blk_request_t *rq, blkio_t *blkio, short m_prio,
                              merge_type_t where)
{
  blkio_t *b = NULL;
  int split_type;
  blkreq_queue_t *q = rq->queue;
  uint32_t nr_sects;

  /*
   * Try to merge blkio at front or tail of request.
   * If summary size of resulting request exceeds allowed
   * maximum, blkio is splitted into two blkios.
   */
  nr_sects = MIN(blkio->nr_sectors, q->max_sectors - rq->nr_sectors);
  if (nr_sects < blkio->nr_sectors) {
    if (!blkio_split_ok(blkio))
      return 0;

    split_type = (where == BLK_MERGE_FRONT) ? BLK_SPLIT_TAIL : BLK_SPLIT_FRONT;
    b = split_blkio(blkio, nr_sects, q->stor->sect_shift, split_type);
    if (b == NULL)
      return 0;
  }

  if (b == NULL)
    b = blkio;

  if (blk_queued_rq(rq)) {
    if (!(q->merge_rq && q->merge_rq(rq, blkio, where))) {
      if (b != blkio) {
        _combine_blkio(blkio, b);
        destroy_blkio(b);
      }
      nr_sects = 0;
    } else {
      trace_rq_merge_blkio(rq, b, m_prio, where);
    }
  } else {
    rq->nr_sectors += b->nr_sectors;
    if (where == BLK_MERGE_FRONT) {
      rq->start_sector = b->start_sector;
      list_add2head(&rq->blkios, &b->node);
    } else {
      list_add2tail(&rq->blkios, &b->node);
    }
    trace_rq_merge_blkio(rq, b, m_prio, where);
  }

  return nr_sects;
}

int initialize_blk_schedulers(void)
{
  int i, ret = 0;
  iosched_ops_t *ops;
  const char *name;

  list_init_head(&schedulers);
  for (i = 0; iosched_init_fn[i] != NULL; i++) {
    name = iosched_init_fn[i](&ops);
    ret = register_blk_iosched(name, ops);
    if (ret)
      break;
  }

  return ret;
}

int register_blk_iosched(const char *name, iosched_ops_t *ops)
{
  iosched_type_t *type;

  type = malloc(sizeof(iosched_type_t));
  if (type == NULL)
    return -ENOMEM;

  type->name = name;
  type->ops = ops;
  list_add2tail(&schedulers, &type->node);

  return 0;
}

int set_blk_iosched(blkreq_queue_t *q, const char *name)
{
  bool was_sched;
  iosched_type_t *type;
  iosched_t *sched = q->scheduler;
  int ret = 0;

  if (name == NULL)
    return -EINVAL;

  list_for_each_entry(&schedulers, type, node) {
    if (!strcmp(type->name, name))
      break;
  }
  if (&type->node == list_head(&schedulers))
    return -ESRCH;

  was_sched = clear_blk_queue_scheduled(q);
  get_iosched(sched);
  if (atomic_get(&sched->refcnt) > 1) {
    if (was_sched) {
      set_blk_queue_scheduled(q);
    }
    put_iosched(sched);
    return -EBUSY;
  }

  if (sched->name != name) {
    if (was_sched) {
      sched->ops->flush_queue(sched, true);
      sched->ops->release(sched);
    }

    sched->name = type->name;
    sched->ops = type->ops;
    ret = sched->ops->init(sched);
  }

  set_blk_queue_scheduled(q);
  put_iosched(sched);

  return ret;
}

int set_blk_default_iosched(blkreq_queue_t *q)
{
  if (default_scheduler == NULL)
    return -ESRCH;

  return set_blk_iosched(q, default_scheduler);
}

int disable_blk_iosched(blkreq_queue_t *q)
{
  iosched_t *sched = q->scheduler;
  bool was_sched;

  was_sched = atomic_test_and_reset_bit(&q->flags, BLKQUEUE_IOSCHED);
  get_iosched(sched);
  if (atomic_get(&sched->refcnt) > 1) {
    if (was_sched) {
      atomic_test_and_set_bit(&q->flags, BLKQUEUE_IOSCHED);
    }
    put_iosched(sched);
    return -EBUSY;
  }

  if (was_sched) {
    sched->ops->flush_queue(sched, true);
    sched->ops->release(sched);
  }
  sched->name = 0;
  sched->private = 0;

  return 0;
}

bool requests_are_mergeable(blk_request_t *rq_front, blk_request_t *rq_tail)
{
  if (rq_front == NULL || rq_tail == NULL)
    return false;

  return ((rq_tail->start_sector == rq_front->start_sector + rq_front->nr_sectors) &&
         (rq_front->nr_sectors + rq_tail->nr_sectors < rq_front->queue->max_sectors));
}

bool merge_requests(blk_request_t *rq_front, blk_request_t *rq_tail)
{
  bool ret = true;
  blkreq_queue_t *q;

  if (blk_queued_rq(rq_front) || blk_queued_rq(rq_tail)) {
    q = rq_front->queue;
    ret = q->merge_requests &&
          q->merge_requests(rq_front, rq_tail);
  } else {
    rq_front->nr_sectors += rq_tail->nr_sectors;
    list_move(list_node_last(&rq_front->blkios),
              list_head(&rq_front->blkios), &rq_tail->blkios);
  }
  if (ret) {
    trace_rq_merge_rq(rq_front, rq_tail);
  }

  return ret;
}

bool iosched_merge(iosched_t *sched, blkio_t *blkio, short m_prio)
{
  blk_request_t *rq[2];
  short prio[2];
  bool need_sched[2] = {false, false};
  bool queued = false;
  merge_type_t where[2] = {BLK_MERGE_TAIL, BLK_MERGE_FRONT};
  iosched_ops_t *ops = sched->ops;
  uint32_t n, nr_sectors = blkio->nr_sectors;
  int i;

  /*
   * [dg]: Distribute queue locking?
   */

  /*
   * try to merge the blkio at the front and tail of
   * already scheduled requests
   */
  blk_queue_lock(sched->queue);
  ops->find_lock_mergeable(sched, blkio, &rq[0], &rq[1]);
  for (i = 0; (i < 2) && nr_sectors; i++) {
    if (rq[i] == NULL)
      continue;

    queued |= blk_queued_rq(rq[i]);
    n = __merge_blkio(rq[i], blkio, m_prio, where[i]);
    if (n) {
      nr_sectors -= n;
      prio[i] = __recalc_rq_prio(sched, rq[i], m_prio);
      need_sched[i] = true;
    }
  }

  if (!nr_sectors && requests_are_mergeable(rq[0], rq[1])) {
    bool merge = true;

    if (!need_sched[1]) {
      prio[1] = rq[1]->prio;
    }
    prio[0] = __recalc_rq_prio(sched, rq[0], prio[1]);
    if (ops->pre_merge_rqs_fn)
      merge = ops->pre_merge_rqs_fn(sched, rq[0], rq[1]);

    if (merge && merge_requests(rq[0], rq[1])) {
      need_sched[1] = false;
      if (ops->post_merge_rqs_fn) {
        ops->post_merge_rqs_fn(sched, rq[0], rq[1]);
      } else {
        ops->remove(sched, rq[1]);
      }
      end_trace_rq(rq[1]);
      if (blk_queued_rq(rq[1])) {
        drop_request_nolock(rq[1]);
        if (!blk_queued_rq(rq[0])) {
          rq[0]->prio = prio[0];
          ops->dispatch_rq(sched, rq[0]);
          need_sched[0] = false;
        }
      } else {
        put_blk_request(rq[1]);
      }
    }
  }

  for (i = 0; i < 2; i++) {
    if (rq[i] == NULL)
      continue;

    if (need_sched[i]) {
      if (ops->post_merge_blkio_fn)
        ops->post_merge_blkio_fn(sched, rq[i], where[i]);

      __reschedule_rq(sched, rq[i], prio[i]);
    }
    blk_rq_unlock(rq[i]);
  }
  blk_queue_unlock(sched->queue);

  if (queued)
    blk_queue_signal(sched->queue);

  return !nr_sectors;
}
