/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 *
 * Partition scanning implementation
 */

#include <stdbool.h>
#include <sys/types.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/mman.h>
#include <block/genstor.h>
#include <block/partition.h>
#include <dircons.h>

static void __rq_complete(blk_request_t *rq)
{
  event_t *ev = (event_t*)rq->completion_data;

  event_signal(ev);
}


static inline bool is_extended_partition(bd_part_t *part)
{
  return (part->id == DOS_EXTENDED_PARTITION) || (part->id == WIN95_EXTENDED_PARTITION) ||
         (part->id == LINUX_EXTENDED_PARTITION);
}


int generic_scan_partitions(genstor_t *storage)
{
  int ret = 0;
  /* Fixme [dg]: get queue before each request */
  blkreq_queue_t *q = (blkreq_queue_t*)storage->private;
  event_t ev;
  blk_request_t *rq;
  int32_t part_num = 1, npart = 0;
  sector_t sector = 0, ebr_sect = 0, first_ebr_sect = 0;
  blkio_t *blkio, *b;
  void *buf = NULL;
  struct mbr *mbr;
  partition_t *hw_parts;
  bd_part_t *p, **parts;
  int32_t i, n = 4;
  biorq_t *biorq = NULL;

  /*
   * TO DO [dg]: check for partitions not overlaping,
   *   valid extended partition size and ids
   */

  if (!storage->capacity)
    return 0;

  ret = event_init(&ev);
  if (ret)
    return ret;

  rq = alloc_init_blkreq(q, REQ_TYPE_IO, 0, 1, false);
  if (rq == NULL) {
    ret = -ENOMEM;
    goto exit_event_destroy;
  }

  biorq = alloc_init_biorq(0, 1, BLKREQ_READ_PRIO, 0);
  if (biorq == NULL) {
    ret = -ENOMEM;
    goto exit_free_request;
  }

  buf = malloc(PAGE_SIZE);
  blkio = alloc_init_blkio(biorq, 0, true, buf);
  if (buf == NULL || blkio == NULL) {
    ret = -ENOMEM;
    goto exit_free_biorq;
  }

  parts = storage->parts;
  list_add2tail(&rq->blkios, &blkio->node);
  rq->completion_data = &ev;
  rq->complete = __rq_complete;

  while (!ret) {
    if ((n == 2) && !ebr_sect)  /* no more ebrs */
      break;

    if (ebr_sect >= storage->capacity) {
      ret = -EINVAL;
      break;
    }

    memset(buf, 0, sizeof(*mbr));
    sector = ebr_sect;
    rq->start_sector = blkio->start_sector = sector;
    rq->current_blkio = NULL;
    rq->complete_sectors = 0;
    ret = dispatch_request(rq, QUEUE_TAIL);
    if (ret)
      break;

    ebr_sect = 0;
    event_wait(&ev);
    q->ioerr_reqs += rq->io_errors;
    b = rq->current_blkio;
    if (b && b->status.err) {
      ret = -b->status.err;
      break;
    }

    mbr = (struct mbr*)buf;
    if (mbr->signature != PART_TBL_MAGIC) {
      ret = -EINVAL;
      break;
    }
    hw_parts = mbr->parts;

    for (i = 0; i < n; i++) {
      if (!hw_parts[i].fsid) {
        if (n == 2 && !i) {
          printf("Warning [part_scan]: empty extended boot record in the %s storage\n",
                  storage->name);
        }
        part_num++;
        continue;
      } else if (hw_parts[i].start_sector + hw_parts[i].nr_sectors > storage->capacity) {
        ret = -EINVAL;
        break;
      }

      if (parts == NULL) {
        parts = (bd_part_t**)malloc(sizeof(bd_part_t*) * BLKDEV_MAX_PARTS);
        if (parts == NULL) {
          ret = -ENOMEM;
          break;
        }
        memset(parts, 0, BLKDEV_MAX_PARTS * sizeof(bd_part_t*));
        storage->parts = parts;
      }

      p = parts[npart];
      if (p == NULL) {
        p = (bd_part_t*)malloc(sizeof(bd_part_t));
        if (p == NULL) {
          ret = -ENOMEM;
          break;
        }
      }

      /* fill new storage partition */
      if ((n == 4) || !i) {
        p->num = part_num;
        p->id = hw_parts[i].fsid;
        p->start_sector = hw_parts[i].start_sector + sector;
        p->nr_sectors = hw_parts[i].nr_sectors;
        p->flags = 0;
        /* TO DO [dg]: init part statistics */
        parts[npart] = p;
        if (is_extended_partition(p)) {
          if (sector || first_ebr_sect) {   /* there is EBR scanning, or EBR is found in MBR? */
            printf("Warning [part_scan]: MBR of the %s storage is invalid, it has "
                   "more than 1 extended partitions, ignore all them excepting the "
                   "first one\n", storage->name);
          } else {
            first_ebr_sect = ebr_sect = p->start_sector;
            p->flags |= PART_HIDDEN;
          }
        }

        part_num++;
        npart++;
        if (npart == BLKDEV_MAX_PARTS)
          break;
      } else if (hw_parts[i].start_sector) {
        ebr_sect = hw_parts[i].start_sector + first_ebr_sect;
      }
    }
    n = 2;
  }

  destroy_blkio(blkio);
exit_free_biorq:
  put_biorq(biorq);
exit_free_request:
  put_blk_request(rq);
exit_event_destroy:
  event_destroy(&ev);

  if (buf != NULL) {
    free(buf);
  }

  /* free unused partitions */
  if (ret) {
    if (ret == -EINVAL && (first_ebr_sect || mbr->signature == PART_TBL_MAGIC))
      printf("Warning [part_scan]: partition table of the %s storage is corrupted\n",
             storage->name);

    storage->nr_parts = MAX(storage->nr_parts, npart);
    npart = 0;
  }

  for (i = npart; i < storage->nr_parts; i++)
    free(parts[i]);

  if (npart == 0 && parts != NULL) {
    free(parts);
    storage->parts = NULL;
  }
  storage->nr_parts = npart;

  return ret;
}
