/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 *
 * Block request processing implementation
 */

/*
 * TO DO [dg]: release_queue
 */

#include <stddef.h>
#include <stdbool.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/uevent.h>
#include <sys/mman.h>
#include <block/genstor.h>
#include <block/request.h>
#include <block/iosched.h>
#include <block/blk_trace.h>
#include "server.h"

static int __make_rq_nomerge(blkreq_queue_t *q, blkio_t *blkio, short prio)
{
  int ret;
  blk_request_t *rq;
  biorq_t *biorq = blkio->biorq;
  bool write;
  iosched_t *sched = q->scheduler;

  write = biorq_rw(biorq);
  rq = alloc_init_blkreq(q, REQ_TYPE_IO, blkio->start_sector,
                         blkio->nr_sectors, write);
  if (rq == NULL)
    return -ENOMEM;

  add_trace_rq(rq);
  rq->complete = default_blk_rq_complete;
  rq->prio = prio;
  if (!write)
    rq->prio += BLKREQ_READ_PRIO;

  list_add2tail(&rq->blkios, &blkio->node);

  if (blk_queue_scheduled(q) && !biorq_not_scheduled(biorq)) {
    trace_rq_iosched_add(rq);
    blk_queue_lock(sched->queue);
    ret = sched->ops->insert(sched, rq);
    blk_queue_unlock(sched->queue);
  } else {
    trace_rq_dispatch(rq, TRACE_LAYER_GENERIC);
    ret = dispatch_request(rq, QUEUE_TAIL);
  }

  if (ret) {
    end_trace_rq(rq);
    put_blk_request(rq);
  }

  return ret;
}

/* allocate and initialize non cached blkio */
static blkio_t *__alloc_init_nc_blkio(biorq_t *biorq, sector_t sector, uint32_t nr_sects,
                                      int sect_shift)
{
  void *data;
  bool priv_buf = false;
  int ret = 0;
  blkio_t *b = NULL;

  if (!biorq_buffered(biorq) ||
      ((biorq->flags & BIORQ_RDONLY_BUFFER) && !biorq_rw(biorq))) {

    data = malloc(nr_sects << sect_shift);
    if (data == NULL) {
      ret = -ENOMEM;
    } else {
      priv_buf = true;
    }
  } else {
    data = (void*) (biorq->data_buf + ((sector - biorq->start_sector) << sect_shift));
  }

  if (!ret) {
    b = alloc_init_blkio(biorq, sector, nr_sects, data);
    if (b == NULL) {
      ret = -ENOMEM;
    }
    b->status.buffered = priv_buf;
  }

  if (ret) {
    if (b != NULL)
      destroy_blkio(b);
    if (priv_buf)
      free(data);
  }

  return b;
}

static bool __rq_match_run(list_node_t *node)
{
  blk_request_t *rq;

  rq = container_of(node, blk_request_t, node);
  if (blk_rq_lock(rq)) {
    /* don't allow to starve a request more then one time */
    set_rq_flags(rq, BLKREQ_MUST_RUN);
    return false;
  }
  return true;
}

bool blkio_split_ok(blkio_t *blkio)
{
  return !(blkio->status.buffered);
}

void default_blk_rq_complete(blk_request_t *rq)
{
  blkio_t *b;
  list_node_t *node, *save;
  biorq_t *biorq, *err_biorq = NULL;
  int err = 0;
  bool compl, err_sent = false;

  rq->queue->last_proc_rq = rq;
  rq->queue->ioerr_reqs += rq->io_errors;

  /* check if not all blkios are done */
  b = rq->current_blkio;
  if (b && b->status.err) {
    err_biorq = b->biorq;
    err = b->status.err;
  }

  trace_rq_end_proc(rq);
  end_trace_rq(rq);
  if (rq->type == REQ_TYPE_IO) {
    /*
     * NOTE: blkio can be inserted in another list in the server
     * layer, so it's need to get the next entry in advance
     */
    list_for_each_safe(&rq->blkios, node, save) {
      b = container_of(node, blkio_t, node);
      biorq = b->biorq;
      /*
       * if the I/O operations are completed with an error,
       * pass to the server layer only first encountered
       * blkio owned by the biorq in the contex of which
       * the error is occured
       */
      compl = !((biorq == err_biorq) && err_sent);
      /* TO DO [dg]: update cache state */
      if (biorq->complete && compl) {
        if (biorq == err_biorq) {
          err_sent = true;
          b->status.err = err;
        }
        biorq->complete(b);
      } else {
        destroy_blkio(b);
      }
    }
  }
  put_blk_request(rq);
}

blk_request_t *alloc_init_blkreq(blkreq_queue_t *q, short type, sector_t start,
                                 uint32_t nr_sectors, bool rw)
{
  blk_request_t *rq;

  rq = malloc(sizeof(blk_request_t));
  if (rq == NULL)
    return NULL;

  memset(rq, 0, sizeof(blk_request_t));
  rq->queue = q;
  rq->start_sector = start;
  rq->nr_sectors = nr_sectors;
  if (rw) {
    set_rq_flags(rq, BLKREQ_RW);
  }
  rq->type = type;
  list_init_head(&rq->blkios);
  atomic_set(&rq->refcnt, 1);

  return rq;
}

biorq_t *alloc_init_biorq(sector_t start_sector, uint32_t nr_sects, biorq_type_t type, uint32_t flags)
{
  biorq_t *bio;

  bio = (biorq_t*)malloc(sizeof(biorq_t));
  if (bio != NULL) {
    memset(bio, 0, sizeof(biorq_t));
    bio->start_sector = start_sector;
    bio->nr_sectors = nr_sects;
    bio->complete_sectors = 0;
    bio->flags = flags;
    bio->type = type;
    atomic_set(&bio->refcnt, 1);
  }

  return bio;
}

blkio_t *alloc_init_blkio(biorq_t *biorq, sector_t sector, uint32_t nr_sects, void *buf)
{
  blkio_t *b;

  b = malloc(sizeof(blkio_t));
  if (b == NULL)
    return NULL;

  memset(b, 0, sizeof(blkio_t));
  b->start_sector = sector;
  b->nr_sectors = nr_sects;
  b->buf = buf;

  ASSERT((biorq != NULL), "biorq is NULL\n");

  get_biorq(biorq);
  b->biorq = biorq;

  return b;
}

void destroy_blkio(blkio_t *b)
{
  if (b->status.buffered) {
    free(b->buf);
  }
  put_biorq(b->biorq);
  free(b);
}

int init_blk_queue(blkreq_queue_t *queue, genstor_t *stor,
                   block_device_t *bdev, int max_requests,
                   int max_sectors, struct blk_group *group)
{
  int ret;
  pthread_mutexattr_t attr = {PTHREAD_PROCESS_PRIVATE};

  memset(queue, 0, sizeof(blkreq_queue_t));
  queue->max_requests = max_requests;
  queue->max_sectors = max_sectors;
  queue->stor = stor;
  queue->bdev = bdev;

  ret = add_queue_to_group(group, queue);
  if (ret)
    return ret;

  ret = -ENOMEM;
  if (event_init(&queue->done_all))
    goto out_group;

  if (pthread_mutex_init(&queue->mutex, &attr))
    goto out_event;

  if (init_prio_queue(&queue->requests, BLKDEV_NR_PRIORITIES))
    goto out_mutex;

  queue->scheduler = malloc(sizeof(iosched_t));
  if (queue->scheduler == NULL)
    goto out_queue;

  memset(queue->scheduler, 0, sizeof(iosched_t));
  queue->scheduler->queue = queue;

  ret = init_blk_rq_trace(queue);
  if (!ret)
    return 0;

  if (queue->scheduler != NULL)
    free(queue->scheduler);
 out_queue:
   destroy_prio_queue(&queue->requests);
 out_mutex:
   pthread_mutex_destroy(&queue->mutex);
 out_event:
   event_destroy(&queue->done_all);
 out_group:
   remove_queue_from_group(queue);

   return ret;
}

void destroy_blk_queue(blkreq_queue_t *queue)
{
  iosched_t *sched = queue->scheduler;

  remove_queue_from_group(queue);
  if (blk_queue_scheduled(queue)) {
    sched->ops->release(sched);
  }
  free(sched);
  destroy_prio_queue(&queue->requests);
  pthread_mutex_destroy(&queue->mutex);
  event_destroy(&queue->done_all);
}

void put_blk_queue(blkreq_queue_t *queue)
{
  if (atomic_dec_and_test(&queue->refcnt)) {
    destroy_blk_queue(queue);
    free(queue);
  }
}

void put_blk_request(blk_request_t *rq)
{
  if (atomic_dec_and_test(&rq->refcnt)) {
    free(rq);
  }
}

void put_biorq(biorq_t *biorq)
{
  if (atomic_dec_and_test(&biorq->refcnt)) {
    if (biorq_buffered(biorq)) {
      free(biorq->data_buf);
    }
    free(biorq);
  }
}

void _combine_blkio(blkio_t *b1, blkio_t *b2)
{
  if (b1->start_sector < b2->start_sector) {
    b1->nr_sectors += b2->nr_sectors;
  } else {
    b1->start_sector = b2->start_sector;
    b1->buf = b2->buf;
    b1->nr_sectors += b1->nr_sectors;
  }
}

blkio_t *split_blkio(blkio_t *blkio, uint32_t nr_sectors, int sect_shift,
                     blk_split_type_t where)
{
  blkio_t *b;

  if (nr_sectors >= blkio->nr_sectors)
    return NULL;

  b = alloc_init_blkio(blkio->biorq, blkio->start_sector, nr_sectors, NULL);
  if (b != NULL) {
    if (where == BLK_SPLIT_FRONT) {
      b->buf = blkio->buf;
      blkio->start_sector += nr_sectors;
      blkio->buf = (void*) ((ulong_t)blkio->buf + (nr_sectors << sect_shift));
    } else {
      b->start_sector += blkio->nr_sectors - nr_sectors;
      b->buf = (void*) ((ulong_t)blkio->buf + (nr_sectors << sect_shift));
    }
    blkio->nr_sectors -= nr_sectors;
  }

  return b;
}

int dispatch_request_nolock_ev(blk_request_t *rq, int where)
{
  blkreq_queue_t *q = rq->queue;
  int ret = 0;
  list_node_t *node;
  list_head_t *head;
  blk_request_t *rq1;

  if (q->nr_requests == q->max_requests)
    return -EBUSY;

  if (rq->complete == NULL)
    rq->complete = default_blk_rq_complete;

  if (q->dispatch_request == NULL) {
    set_rq_flags(rq, BLKREQ_QUEUED);
    q->nr_requests++;
    if (where == QUEUE_TAIL) {
      prio_queue_add(&q->requests, &rq->node, rq->prio);
    } else {
      /*
       * add to the head of queue or after the last non
       * barrier request
       */
      head = q->requests.heads + rq->prio;
      list_for_each(head, node) {
        rq1 = container_of(rq1, blk_request_t, node);
        if (!blk_barrier_rq(rq1))
          break;
      }
      list_add(node, &rq->node);
    }
  } else {
    ret = q->dispatch_request(rq, where);
  }

  return ret;
}

int dispatch_request(blk_request_t *rq, int where)
{
  blkreq_queue_t *q = rq->queue;
  int ret = 0;
  bool signal;

  blk_queue_lock(q);
  ret = dispatch_request_nolock_ev(rq, where);
  signal = (q->nr_requests == 1);
  blk_queue_unlock(q);
  if (signal)
    blk_queue_signal(q);

  return ret;
}

int drop_request_nolock(blk_request_t *rq)
{
  blkreq_queue_t *q = rq->queue;
  int ret = -1;

  set_rq_flags(rq, BLKREQ_INACTIVE);
  if (blk_queued_rq(rq)) {
    if (q->drop_request) {
      q->drop_request(rq);
    } else {
      clear_rq_flags(rq, BLKREQ_QUEUED);
      list_del(&rq->node);
      q->nr_requests--;
      ret = 0;
    }
  }

  return ret;
}

int drop_request(blk_request_t *rq)
{
  blkreq_queue_t *q = rq->queue;
  int ret;

  blk_queue_lock(q);
  ret = drop_request_nolock(rq);
  blk_queue_unlock(q);

  return ret;
}

static blk_request_t *__get_next_blkrq(blkreq_queue_t *q)
{
  list_node_t *node;
  blk_request_t *rq = NULL;
  iosched_t *sched;

  node = prio_queue_get_next_match(&q->requests, __rq_match_run);
  if (node != NULL) {
    rq = container_of(node, blk_request_t, node);
    clear_rq_flags(rq, BLKREQ_QUEUED);
    set_rq_flags(rq, BLKREQ_RUN);
    q->nr_requests--;
    trace_rq_start_proc(rq);
    if (blk_queue_scheduled(q) && blk_scheduled_rq(rq)) {
      sched = q->scheduler;
      sched->ops->remove(sched, rq);
    }
    get_blk_request(rq);
  }

  return rq;
}

blk_request_t *get_next_blkrq(blkreq_queue_t *queue)
{
  blk_request_t *rq = NULL;
  struct blk_group *group = queue->group;
  bool wait = false;

  while (rq == NULL) {
    if (!queue->nr_requests || wait) {
      if (event_wait(&queue->group->not_empty) == -EINTR)
         break;
    }

    blk_queue_lock(queue);
    /*
     * requests can be fetched concurrently, or an event
     * can be dispatched in another queue in the group
     * check once again
     */
    if (!queue->nr_requests) {
      blk_queue_unlock(queue);
      continue;
    } else {
      rq = __get_next_blkrq(queue);
      if (!queue->nr_requests)
        atomic_test_and_reset_bit(&group->filled_queues, queue->index);

      wait = (rq == NULL);
      blk_queue_unlock(queue);
    }
  }

  return rq;
}

blk_request_t *flush_get_group_blkrq(struct blk_group *group, ulong_t qmask)
{
  blk_request_t *rq = NULL;
  blkreq_queue_t *q;
  ulong_t mask;
  bool flush, qready;
  ulong_t fmask;

  fmask = atomic_swap_long(&group->flush_mask, 0);
  blk_group_rlock(group);
  if (!(group->filled_queues || fmask)) {
    blk_group_runlock(group);
    return NULL;
  }

  list_for_each_entry(&group->queues, q, node) {
    mask = 1 << q->index;
    qready = group->filled_queues & (mask & qmask);
    flush = fmask & mask;
    if (qready || flush) {
      blk_queue_lock(q);
      if (flush && blk_queue_scheduled(q)) {
        q->scheduler->ops->flush_queue(q->scheduler, false);
        fmask &= ~mask;
        qready = group->filled_queues & (mask & qmask);
      }

      if (qready && !rq) {
        rq = __get_next_blkrq(q);
        if (!q->nr_requests)
          atomic_test_and_reset_bit(&group->filled_queues, q->index);
      }
      blk_queue_unlock(q);
      if (rq && !fmask)
        break;
    }
    mask <<= 1;
  }
  blk_group_runlock(group);
  if (rq)
    group->last_serv_queue = rq->queue;

  return rq;
}

blk_request_t *get_group_blkrq_wait(struct blk_group *group, ulong_t qmask, ulong_t rec_mask)
{
  blk_request_t *rq = NULL;
  bool wait = false;

  set_blk_group_wait(group);
  do {
    if (!(group->filled_queues || group->flush_mask) || wait) {
      if (event_wait(&group->not_empty) == -EINTR) {
        break;
      }
    }
    if (!(group->filled_queues & (qmask & rec_mask)))
      rec_mask = ~0L;

    rq = flush_get_group_blkrq(group, qmask & rec_mask);
    rec_mask  = ~0L;
    wait = (rq == NULL);
  } while (rq == NULL);
  clear_blk_group_wait(group);

  return rq;
}

int blk_group_wait_event(struct blk_group *group)
{
  set_blk_group_wait(group);
  event_wait(&group->not_empty);
  clear_blk_group_wait(group);
}

int init_blk_group(struct blk_group *group)
{
  int ret;
  pthread_mutexattr_t attr = {PTHREAD_PROCESS_PRIVATE};

  memset(group, 0, sizeof(struct blk_group));
  group->free_indexes = ~0L;
  list_init_head(&group->queues);
  ret = event_init(&group->not_empty);
  if (!ret) {
    ret = -pthread_mutex_init(&group->mutex, &attr);
    if (ret)
      event_destroy(&group->not_empty);
  }

  return ret;
}

void destroy_blk_group(struct blk_group *group)
{
  event_destroy(&group->not_empty);
  pthread_mutex_destroy(&group->mutex);
}

int add_queue_to_group(struct blk_group *group, blkreq_queue_t *q)
{
  int ret = 0;

  blk_group_wlock(group);
  q->index = bit_find_lsf(group->free_indexes);
  if (q->index != -1) {
    list_add2tail(&group->queues, &q->node);
    bit_clear(&group->free_indexes, q->index);
    q->group = group;
  }
  blk_group_wunlock(group);

  return ret;
}

void remove_queue_from_group(blkreq_queue_t *q)
{
  struct blk_group *group = q->group;

  blk_group_wlock(group);
  bit_clear(&group->filled_queues, q->index);
  bit_set(&group->free_indexes, q->index);
  list_del(&q->node);
  q->index = -1;
  q->group = NULL;
  blk_group_wunlock(group);
}

void blk_queue_signal(blkreq_queue_t *q)
{
  bit_set(&q->group->filled_queues, q->index);
  event_signal(&q->group->not_empty);
}

int generic_submit_blkio(blkreq_queue_t *q, blkio_t *blkio, short prio)
{
  bool merged;
  iosched_t *sched = q->scheduler;

  if (blk_queue_scheduled(q) && !biorq_not_scheduled(blkio->biorq)) {
    merged = iosched_merge(sched, blkio, prio);
    if (merged)
      return 0;
  } else if (q->nr_requests == q->max_requests || blk_queue_plugged(q)) {
    return -EBUSY;
  }

  return __make_rq_nomerge(q, blkio, prio);
}

int generic_make_request(genstor_t *storage, bd_part_t *part,
                         biorq_t *biorq, biovec_t *bvec, short prio)
{
  int ret = 0, n;
  blkreq_queue_t *q = (blkreq_queue_t*)storage->private;
  blkio_t *b = NULL;
  uint32_t nr_sects;
  sector_t sector;
  int shift = storage->sect_shift;

  if (part != NULL) {
    biorq->start_sector += part->start_sector;
  }
  if (bvec) {
    sector = bvec->start_sector;
    nr_sects = bvec->nr_sectors;
  } else {
    sector = biorq->start_sector;
    nr_sects = biorq->nr_sectors;
  }

  if (biorq_type_fs(biorq))
    biorq->complete = default_complete_biorq;

  get_iosched(q->scheduler);
  while (nr_sects && !ret) {
    n = MIN(q->max_sectors, nr_sects);
    b = __alloc_init_nc_blkio(biorq, sector, n, shift);
    if (b == NULL) {
      ret = -ENOMEM;
      break;
    }

    if (q->io_request_fn) {
      ret = q->io_request_fn(q, b, prio);
    } else {
      ret = generic_submit_blkio(q, b, prio);
    }

    nr_sects -= n;
    sector += n;
  }

  if (ret && b != NULL) {
    destroy_blkio(b);
  }
  put_iosched(q->scheduler);

  return ret;
}

void default_complete_biorq(blkio_t *blkio)
{
  biorq_t *biorq = blkio->biorq;
  bool reply = false;

  if (blkio->status.err) {
    set_biorq_flags(biorq, BIORQ_ABORTED);
    reply = true;
  } else {
    biorq->complete_sectors += blkio->nr_sectors;
    if (biorq->complete_sectors == biorq->nr_sectors) {
      reply = true;
    }
  }

  if (reply) {
    biorq->complete = NULL;
    reply_blkio(blkio);
  } else {
    destroy_blkio(blkio);
  }
}

biorq_t *check_clusterize_biorq(biorq_t *orig_bio, size_t offset, size_t size,
                                       uint32_t sect_size, blk_split_type_t split)
{
  biorq_t *b = NULL;
  sector_t sector;
  uint32_t sect_off = offset & (sect_size - 1);
  uint32_t wb_off;
  uint32_t wb_size;
  ulong_t data_off;
  bool use_orig = false;

  if (biorq_double_end(orig_bio))
    return NULL;

  if (split == BLK_SPLIT_FRONT) {
    if (!sect_off) {
      /* it's not need to split, the request is sector aligned */
      return NULL;
    }

    data_off = orig_bio->data_off;
    if (orig_bio->nr_sectors > 2) {
      sector = orig_bio->start_sector;
      orig_bio->start_sector++;
      wb_size = sect_size - sect_off;
      orig_bio->data_off = data_off + wb_size;
      orig_bio->nr_sectors--;
    } else {
      use_orig = true;
      wb_size = size;
    }
    wb_off = sect_off;
  } else {
    size_t s = (offset + size) & (sect_size - 1);
    size_t d;
    uint32_t c;

    if (!s) {
      /* it's not need to split, the request is sector aligned at the tail */
      return NULL;
    }

    sector = orig_bio->start_sector + orig_bio->nr_sectors - 1;
    d = (orig_bio->nr_sectors - 1) * sect_size;
    c = sect_off ? sect_size - sect_off : 0;
    wb_size = size - c - d;
    data_off = orig_bio->data_off + d;

    if (orig_bio->nr_sectors > 1) {
      orig_bio->nr_sectors--;
    } else {
      use_orig = true;
    }
    wb_off = 0;
  }

  if (use_orig) {
    b = orig_bio;
  } else {
    b = alloc_init_biorq(sector, 1, BIORQ_TYPE_USER, orig_bio->flags);
    if (b == NULL)
      return (biorq_t*)ERR_PTR(-ENOMEM);
  }

  b->wb_chunk.off = wb_off;
  b->wb_chunk.size = wb_size;
  b->data_off = data_off;
  set_biorq_flags(b, BIORQ_DOUBLE_END);
  clear_biorq_flags(b, BIORQ_RW);

  return b;
}
