/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008,2009 Dmitry Gromada <gromada@jarios.org>
 *
 * Storage server implementation
 */

/*
 * TO DO [dg]:
 *
 * 1) Replace all malloc of strutures with predefined size by
 *    allocation from slab
 *
 * 2) Access invalidation at media change
 * 3) Resource freeing at server start fail because
 *    several servers can start in the same context
 */

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <datalib/usrtc.h>
#include <datalib/list.h>
#include <ipc/rpc.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/blkdev.h>
#include <pthread.h>
#include <bitwise.h>
#include <errno.h>
#include <err.h>
#include <block/request.h>
#include <fsio/bdio.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include "server.h"

#define PAGE_MASK (PAGE_SIZE - 1)
#define BUFSIZE 4096
#define MAX_SERVCTL_DATALEN (CMDBUF_SIZE - sizeof(uni_rpc_t))
#define BLK_MAX_CLIENTS 1048576
#define BLK_MAX_FDS 1048576

typedef struct __ioctl_entry {
  uint16_t cmd;
  uint16_t devno;
  void *data;          /* buffer containing rpc followed by data */
  size_t data_size;
  size_t buf_size;
  uint16_t msgid;
  list_node_t node;
} ioctl_entry_t;

typedef struct __bio_usr_data_t {
  uint16_t msgid;
  uint32_t sect_off;
  size_t size;
  atomic_t refcnt;
  struct blk_fd *fd;
} bio_usr_data_t;

typedef struct __bio_fs_data {
  uint16_t msgid;
  int mem_obj_id;
  uint32_t obj_off;
  struct blk_fd *fd;
} bio_fs_data_t;

static const char *device_dir = "/dev/";

static void *cmd_thread_fn(void *data);
static void *io_thread_fn(void *data);
static void *reply_cmd_thread_fn(void *data);
static void *ctl_reply_thread_fn(void *data);

static int server_process_cmd(blk_thread_t *thread, ipc_port_msg_info_t *msginfo, bool *reply);
static ssize_t server_process_io(blk_thread_t *thread, ipc_port_msg_info_t *msginfo,
                                 bool *reply);
static int process_fsio(uni_rpc_t *rpc, genstor_t *stor, struct blk_fd *fd,
                        uint16_t msgid, struct cache_head *mheads,
                        uint32_t heads_size, uint16_t *nr_proc_heads);
static int check_prepare_user_ioreq(blk_thread_t *thread, biorq_t *biorq, bio_rpc_t *brpc,
                                    struct blk_fd *fd);

/* server I/O operations */

/* operations for ipc model with partially received messages */
static int receive_imsg_partial(blk_thread_t *thread);
static int get_imsg_data_partial(blk_thread_t *thread, uint32_t offset, uint32_t size, void *buf);
static int send_non_cached_data(blkdev_server_t *srv, biorq_t *biorq);
static int send_cached_data(blkdev_server_t *srv, biorq_t *biorq);

/* client related functions */
static int register_fd(blkdev_server_t *srv, uni_rpc_t *rpc);
static int unregister_fd(blkdev_server_t *srv, pid_t pid, uint16_t fid);
static struct blk_fd *lookup_get_fd(blkdev_server_t *srv, pid_t pid, uint16_t fid);
static int clone_fds(blkdev_server_t *srv, pid_t parent, pid_t clone);
static int client_killed(blkdev_server_t *srv, pid_t pid);
static void __free_client_fds(struct client *client);

static blkserv_ioops_t serv_ioops_partial = {
  NULL,
  receive_imsg_partial,
  get_imsg_data_partial,
  send_non_cached_data,
  send_cached_data
};

static bool inline __async_chan_is_valid(chid_t chan)
{
  return ((signed)chan >= 0);
}

static int __compare_pid(const void *a, const void *b)
{
  return (int) (*(pid_t*)a - *(pid_t*)b);
}

static int __compare_fid(const void *a, const void *b)
{
  return (int) (*(uint16_t*)a - *(uint16_t*)b);
}

static inline void get_client(struct client *client)
{
  atomic_inc(&client->refcnt);
}

static inline void put_client(struct client *client)
{
  if (atomic_dec_and_test(&client->refcnt)) {
    if (__async_chan_is_valid(client->async_chan)) {
      ipc_channel_close(client->async_chan);
    }
    free(client);
  }
}

static void inline get_blk_fd(struct blk_fd *fd)
{
  atomic_inc(&fd->refcnt);
}

static void inline put_blk_fd(struct blk_fd *fd)
{
  if (atomic_dec_and_test(&fd->refcnt)) {
    put_client(fd->client);
    free(fd);
  }
}

static void inline __get_bio_udata(bio_usr_data_t *data)
{
  atomic_inc(&data->refcnt);
}

static void __put_bio_udata(bio_usr_data_t *data)
{
  if (atomic_dec_and_test(&data->refcnt)) {
    put_blk_fd(data->fd);
    free(data);
  }
}

static void __list_add_raise_event(list_head_t *head, list_node_t *node, event_t *e,
                                   pthread_mutex_t *mutex)
{
  bool empty;

  pthread_mutex_lock(mutex);
  empty = list_is_empty(head);
  list_add2tail(head, node);
  pthread_mutex_unlock(mutex);
  if (empty)
   event_signal(e);

}

static void __reply_with_data(ulong_t port, void *buf, ulong_t len, int msg_id)
{
  uni_rpc_t rpc;
  iovec_t vec[2];
  int n = 1;

  rpc.rpc_call=0;
  rpc.size=0;

  vec[0].iov_base = &rpc;
  vec[0].iov_len = sizeof(uni_rpc_t);

  if (len) {
    vec[1].iov_base = buf;
    vec[1].iov_len = len;
    n++;
  }

  ipc_port_reply_nmsg_iov(port, vec, n, msg_id);
}

static void __reply_with_replcode(ulong_t port, size_t len, int msg_id)
{
  uni_rpc_t rpc;

  rpc.rpc_call = 0;
  rpc.size = len;

  ipc_port_reply_nmsg(port, sizeof(rpc), &rpc, msg_id);
}

static void __reply_with_error(ulong_t port, int err, int msg_id)
{
  uni_rpc_t rpc;

  rpc.rpc_call=err;

  ipc_port_reply_nmsg(port, sizeof(rpc), &rpc, msg_id);
}

static void inline __reply_with_rpc(ulong_t port, uni_rpc_t *rpc, int msg_id)
{
  rpc->rpc_call = 0;
  ipc_port_reply_nmsg(port, sizeof(uni_rpc_t), rpc, msg_id);
}

void reply_blkio(blkio_t *blkio)
{
  biorq_t *biorq = blkio->biorq;
  blkdev_server_t *srv;

  if (biorq_type_fs(biorq)) {
    bio_fs_data_t *bdata = biorq->server_private;
    srv = bdata->fd->client->server;

    evl_queue_add(&srv->done_srv_bios, &blkio->node);
  } else {
    bio_usr_data_t *bdata = biorq->server_private;
    srv = bdata->fd->client->server;

    /* it may be clustered request, check if it is last */
    if (atomic_get(&bdata->refcnt) == 1 || blkio->status.err) {
      __list_add_raise_event(&srv->done_user_bios, &blkio->node,
                             &srv->event, &srv->rq_mutex);
    } else {
      __put_bio_udata(bdata);
    }
  }
}

static void free_storage_parts(genstor_t *stor)
{
  unsigned int i = 0;
  bd_part_t **parts = stor->parts;

  for (i = 0; i < stor->nr_parts; i++) {
    if (parts[i] != NULL)
      free(parts[i]);
  }
  free(stor->parts);
  stor->parts = NULL;
  stor->nr_parts = 0;
}

static int unreg_parts(genstor_t *stor, unsigned int num)
{
  int ret = 0;
  unsigned int i;
  char name[STOR_NAME_MAXLEN], *p;
  bd_part_t **parts = stor->parts;

  if (stor->part_opened)
    return -EBUSY;

  strcpy(name, device_dir);
  p = name + strlen(device_dir);
  for (i = 0; i < num; i++) {
    if (parts[i] == NULL || (parts[i]->flags & PART_HIDDEN)) {
      continue;
    }

    sprintf(p, "%s%d", stor->name, parts[i]->num);
    ret = unlink(name);
    if (ret)
      break;
  }
  return ret;
}

/* register partitions */
static int reg_parts(blkdev_server_t *srv, unsigned int *nr_reg)
{
  int ret, fd = 0;
  unsigned int i;
  char name[STOR_NAME_MAXLEN];
  char *sn; /* device short name */
  genstor_t *stor = srv->stor;
  bd_part_t **parts = stor->parts;

  strcpy(name, device_dir);
  sn = name + strlen(device_dir);
  for (i = 0; i < stor->nr_parts; i++) {
    if (parts[i] == NULL || parts[i]->flags & PART_HIDDEN)
      continue;

    sprintf(sn, "%s%d", stor->name, parts[i]->num);
    fd = creat(name, (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP));
    if (fd < 0) {
      ret = fd;
      break;
    }

    ret = deviofs_assign(fd, srv->cmd_port, srv->io_port, DFS_BLKDEV, i + 1);
    /* TO DO [dg]: uncomment when it will be provided by devio fs */
#if 0
    close(fd);
#endif
    if (ret < 0)
      close(fd);

    printf("[blkserv] Block device %s is registered\n", name);
  }
  *nr_reg = i;

  return ret;
}

static void *__start_server(void *data)
{
  struct serv_reg_data *rd = (struct serv_reg_data*)data;
  long ret;

  ret = start_storage_server(rd->stor, rd->iobuf_size,
            rd->nr_cmd_threads, rd->nr_io_threads,
            &rd->compl);

  return ERR_PTR(ret);
}

static int __init_server(blkdev_server_t *srv)
{
  int ret = 0;
  pthread_mutexattr_t attr = {PTHREAD_PROCESS_PRIVATE};
  int i;

  srv->cmd_port = ipc_port_open(IPC_BLOCKED_ACCESS | IPC_PRIORITIZED_ACCESS,
                                DQUEUE_SIZE);
  srv->io_port = ipc_port_open(IPC_BLOCKED_ACCESS | IPC_PRIORITIZED_ACCESS,
                               DQUEUE_SIZE);
  if ((signed)srv->cmd_port < 0 || (signed)srv->io_port < 0)
    return -ENOMEM;

  for (i = 0; i < CLIENT_HASH_TBL_WIDTH; i++) {
    usrtc_init(&srv->client_table[i].clients, USRTC_SPLAY,
               BLK_MAX_CLIENTS, __compare_pid);
  }

  if (init_evl_queue(&srv->done_srv_bios))
    return -ENOMEM;

  if (pthread_mutex_init(&srv->rq_mutex, &attr) || event_init(&srv->event))
    return -ENOMEM;

  list_init_head(&srv->done_user_bios);
  list_init_head(&srv->ctl_reqs);
  srv->ioops = serv_ioops_partial;
  srv->mesgs_are_persistent = true;

  return ret;
}

static int start_serv_threads(blkdev_server_t *srv)
{
  int i, j, ret = 0;
  blk_thread_t *threads[2], *th;
  uint32_t size[2] = {CMDBUF_SIZE, srv->iobuf_size};
  void *(*routine[2])(void *) = {cmd_thread_fn, io_thread_fn};
  int n[2];

  threads[0] = (blk_thread_t*)malloc(srv->nr_cmd_threads * sizeof(blk_thread_t));
  if (threads[0] == NULL)
    return -ENOMEM;

  threads[1] = (blk_thread_t*)malloc(srv->nr_io_threads * sizeof(blk_thread_t));
  if (threads[1] == NULL) {
    free(threads[0]);
    return -ENOMEM;
  }

  srv->cmd_threads = threads[0];
  srv->io_threads = threads[1];
  memset(srv->cmd_threads, 0, srv->nr_cmd_threads * sizeof(blk_thread_t));
  memset(srv->io_threads, 0, srv->nr_io_threads * sizeof(blk_thread_t));

  n[0] = srv->nr_cmd_threads;
  n[1] = srv->nr_io_threads;

  for (i = 0; i < 2; i++) {
    th = threads[i];
    for (j = 0; j < n[i]; j++) {
      th[j].srv = srv;
      th[j].buf = malloc(size[i]);
      if (th[j].buf == NULL) {
        ret = -ENOMEM;
        break;
      }
      if (i || j) { /* if there is not the main thread */
        ret = -pthread_create(&th[j].id, NULL, routine[i], th + j);
        if (ret)
          break;
      }
    }
  }
  srv->cmd_threads[0].id.__tid = getpid();

  ret = -pthread_create(srv->reply_threads, NULL, &reply_cmd_thread_fn, srv);
  if (!ret)
    ret = -pthread_create(srv->reply_threads + 1, NULL, &ctl_reply_thread_fn, srv);

  return ret;
}

int start_storage_server(genstor_t *storage, uint32_t iobuf_size, int nr_cmd_threads,
                         int nr_io_threads, struct completion *start_compl)
{
  int ret = -EINVAL, fd;
  unsigned int n;
  char name[STOR_NAME_MAXLEN + 128];
  blkdev_server_t *srv;

  if (!nr_cmd_threads || (unsigned)nr_cmd_threads > MAX_BLKSERV_THREADS)
    goto end_reg;

  if ((unsigned)nr_cmd_threads > MAX_BLKSERV_THREADS)
    goto end_reg;

  srv = (blkdev_server_t*)malloc(sizeof(blkdev_server_t));
  if (srv == NULL) {
    ret = -ENOMEM;
    goto end_reg;
  }

  set_server_busy(srv);
  memset(srv, 0, sizeof(blkdev_server_t));
  srv->stor = storage;

  srv->nr_cmd_threads = nr_cmd_threads;
  srv->nr_io_threads = nr_io_threads;
  srv->iobuf_size = MAX(iobuf_size, MIN_IOBUF_SIZE);
  srv->iobuf_size = MIN(srv->iobuf_size, IPC_PORT_MSGLEN_MAX);

  ret = __init_server(srv);
  if (ret)
    goto end_reg;

  if (srv->ioops.init) {
    ret = srv->ioops.init(srv);
    if (ret)
      goto end_reg;
  }

  sprintf(name, "%s%s", device_dir, storage->name);
  if (ret)
    goto end_reg;

  fd = creat(name, (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP));
  if (fd < 0) {
    ret = -fd;
    goto end_reg;
  }

  srv->devfs_pid = getfpid(fd);
  ret = deviofs_assign(fd, srv->cmd_port, srv->io_port, DFS_BLKDEV, 0);
  /*
   * fucking deviofs doesn't support now fd closing, uncomment the line when
   * the mega feature will be available
   */
#if 0
  close(fd);
#endif
  if (ret < 0)
    goto end_reg;

  if (storage->nr_parts) {
    ret = reg_parts(srv, &n);
    if (ret) {
      printf("[blkserv] Partitions registration failed\n");
      if (!unreg_parts(storage, n)) {
        free_storage_parts(storage);
      }
      goto end_reg;
    }
  }

  ret = start_serv_threads(srv);

end_reg:
  if (start_compl != NULL) {
    complete(start_compl, -ret);
  }
  if (!ret) {
    printf("[blkserv] Block device %s is registered\n", name);
    clear_server_busy(srv);
    cmd_thread_fn(srv->cmd_threads);
  }

  return ret;
}

int start_storage_serv_thread(struct serv_reg_data *rd)
{
  int ret;
  pthread_t th;

  ret = -pthread_create(&th, NULL, __start_server, rd);
  if (!ret) {
    wait_for_completion(&rd->compl);
    ret = -rd->compl.result;
    rd->compl.result = 0;
  }

  return ret;
}

/* server I/O operations implementation */

static int receive_imsg_partial(blk_thread_t *thread)
{
  blkdev_server_t *srv = thread->srv;

  return ipc_port_rcvmsg_mi(srv->io_port, sizeof(uni_rpc_t),
                            thread->buf, &thread->msg_info);
}

static int get_imsg_data_partial(blk_thread_t *thread, uint32_t offset,
                                   uint32_t size, void *buf)
{
  return ipc_port_read_msg(thread->srv->io_port, thread->msg_info.msg_id,
                           offset + sizeof(uni_rpc_t), buf, size);
}

static int send_non_cached_data(blkdev_server_t *srv, biorq_t *biorq)
{
  bio_usr_data_t *bdata = (bio_usr_data_t*)biorq->server_private;
  uni_rpc_t rpc;
  iovec_t vecs[2];

  rpc.rpc_call = 0;
  rpc.size = bdata->size;
  vecs[0].iov_base = &rpc;
  vecs[0].iov_len = sizeof(uni_rpc_t);
  vecs[1].iov_base = biorq->data_buf + bdata->sect_off;
  vecs[1].iov_len = bdata->size;

  return ipc_port_reply_nmsg_iov(srv->io_port, vecs, 2, bdata->msgid);
}

static int send_cached_data(blkdev_server_t *srv, biorq_t *biorq)
{
  bio_usr_data_t *bdata = (bio_usr_data_t*)biorq->server_private;
  uint32_t off = bdata->sect_off;
  iovec_t vecs[2];
  uni_rpc_t rpc;
  int ret = 0;

  rpc.rpc_call = 0;
  rpc.size = 0;
  vecs[0].iov_base = &rpc;
  vecs[0].iov_len = sizeof(uni_rpc_t);

  if (biorq_data_available(biorq)) {
    vecs[1].iov_base = biorq->data_buf + off;
    vecs[1].iov_len = bdata->size;

    ret = ipc_port_reply_nmsg_iov(srv->io_port, vecs, 2, bdata->msgid);
  }

  return ret;
}

static int check_server_state(blkdev_server_t *srv)
{
  ulong_t state;
  int r = 0;

  if ((state = atomic_get(&srv->state))) {
    if (state & BLKSERV_BUSY) {
      r = -EBUSY;
    } else {
      r = -EIO;
    }
  }

  return r;
}

static inline bool check_storage_caps(genstor_t *stor, uint16_t partno,
                                      uint16_t flags)
{
  bool rdonly;

  if (partno) {
    rdonly = stor->parts[partno - 1]->flags & PART_RDONLY;
  } else {
    rdonly = stor->flags & STORAGE_RDONLY;
  }

  return !(rdonly && (flags & O_WRONLY));
}

static inline bool check_access(struct blk_fd *fd, uint16_t call)
{
  uint16_t mask;

  mask = (call == RPC_WRITE) ? O_WRONLY : O_RDONLY;

  return (fd->flags & mask);
}

/* main thread which process control requests from another services */
static void *cmd_thread_fn(void *data)
{
  int r;
  blk_thread_t *thread = (blk_thread_t*)data;
  blkdev_server_t *srv = thread->srv;
  bool reply = true;
  int msgid;

  while (true) {
    r = ipc_port_rcvmsg_mi(srv->cmd_port, CMDBUF_SIZE, thread->buf, &thread->msg_info);
    if (r < 0)
      continue;

    msgid = thread->msg_info.msg_id;
    if (thread->msg_info.msg_len < sizeof(uni_rpc_t)) {
      __reply_with_error(srv->cmd_port, EINVAL, msgid);
      continue;
    }

    r = check_server_state(srv);
    if (r) {
      __reply_with_error(srv->cmd_port, -r, msgid);
    } else {
      r = server_process_cmd(thread, &thread->msg_info, &reply);
      if (reply) {
        if (r < 0) {
          __reply_with_error(srv->cmd_port, -r, msgid);
        } else {
          __reply_with_replcode(srv->cmd_port, 0, msgid);
        }
      } else {
        reply = true;
      }
    }
  }
}

/* this thread process read/write requests from privileged used applications */
static void *io_thread_fn(void *data)
{
  ssize_t r;
  blk_thread_t *thread = (blk_thread_t*)data;
  blkdev_server_t *srv = thread->srv;
  blkserv_ioops_t *ioops = &srv->ioops;
  bool reply = true;
  int msgid;

  while (true) {
    r = ioops->receive_imsg(thread);
    if (r < 0)
      continue;

    msgid = thread->msg_info.msg_id;
    if (thread->msg_info.msg_len < sizeof(uni_rpc_t)) {
      __reply_with_error(srv->io_port, EINVAL, msgid);
      continue;
    }

    r = check_server_state(srv);
    if (r) {
      __reply_with_error(srv->io_port, r, msgid);
    } else {
      r = server_process_io(thread, &thread->msg_info, &reply);
      if (reply) {
        if (IS_ERR_VALUE(r)) {
          __reply_with_error(srv->io_port, (int)-r, msgid);
        } else {
          __reply_with_replcode(srv->io_port, r, msgid);
        }
      } else {
        reply = true;
      }
    }
  }

  return NULL;
}

/* this thread send notification to services about request completion */
static void *reply_cmd_thread_fn(void *data)
{
  blkio_t *blkio;
  blkdev_server_t *srv = (blkdev_server_t*)data;
  biorq_t *biorq;
  bio_fs_data_t *bdata;
  struct client *client;
  list_node_t *node;
  uni_rpc_t rpc;
  int r;

  while (true) {
    node = evl_queue_get_next(&srv->done_srv_bios);
    if (node == NULL)
      continue;

    blkio = container_of(node, blkio_t, node);
    biorq = blkio->biorq;
    bdata = (bio_fs_data_t*)biorq->server_private;
    client = bdata->fd->client;

    if (blkio->status.err) {
      rpc.size = 0;
      rpc.offset = biorq->start_sector;
      rpc.rpc_call = blkio->status.err;
    } else {
      rpc.size = biorq->nr_sectors;
      rpc.rpc_call = 0;
      rpc.offset = biorq->start_sector;
    }

    if (__async_chan_is_valid(client->async_chan)) {
      r = ipc_channel_msgsnd(client->async_chan, sizeof(uni_rpc_t), &rpc, 0, NULL);
      if (r == -EPIPE)
        client->async_chan = -1L;
      else {
          /* TO DO [dg]: output error message in log */
      }
    } else {
      __reply_with_rpc(srv->cmd_port, &rpc, bdata->msgid);
    }

    if (biorq->complete) {
      biorq->complete(blkio);
    }
    put_server_biorq(biorq);
    destroy_blkio(blkio);
  }

  return NULL;
}

static ssize_t __do_storage_ioctl(blkdev_server_t *srv, ioctl_entry_t *ctl)
{
  ssize_t ret = 0;
  unsigned int n;
  genstor_t *stor = srv->stor;
  uint16_t cmd = ctl->cmd;

  if (cmd == BLKCMD_RESCAN_PARTITIONS) {
    if (ctl->devno)
      return -EINVAL;

    if (!parts_wlock(stor))
      return -EBUSY;

    if (stor->parts != NULL) {
      ret = unreg_parts(stor, stor->nr_parts);
      if (ret) {
        parts_wunlock(stor);
        return ret;
      }
    }
  }

  ret = stor->ops->ioctl(stor, ctl->devno, ctl->cmd, ctl->data,
                         ctl->data_size, ctl->buf_size);

  if (cmd == BLKCMD_RESCAN_PARTITIONS) {
    if (!ret) {
      ret = reg_parts(srv, &n);
      if (ret) {
        if (!unreg_parts(stor, n))
          free_storage_parts(stor);
      }
    }
    parts_wunlock(stor);
  }

  return ret;
}

/* thread processing control request and replying on I/O requests from user applications */
static void *ctl_reply_thread_fn(void *data)
{
  blkdev_server_t *srv = (blkdev_server_t*)data;
  blkio_t *blkio;
  ioctl_entry_t *ctl;
  biorq_t *biorq;
  list_node_t *node;
  bool empty;
  genstor_t *stor = srv->stor;
  bio_usr_data_t *bdata;
  ssize_t n;

  while (true) {
    empty = true;
    pthread_mutex_lock(&srv->rq_mutex);
    if (!list_is_empty(&srv->ctl_reqs)) {
      empty = false;
      node = list_node_first(&srv->ctl_reqs);
      list_delfromhead(&srv->ctl_reqs);
      pthread_mutex_unlock(&srv->rq_mutex);

      /* process user control request */
      ctl = container_of(node, ioctl_entry_t, node);
      n = __do_storage_ioctl(srv, ctl);
      if (n < 0) {
        __reply_with_error(srv->io_port, (int)-n, ctl->msgid);
      } else if (n) {
        __reply_with_data(srv->io_port, ctl->data, n, ctl->msgid);
      } else {
        __reply_with_replcode(srv->io_port, 0, ctl->msgid);
      }

      pthread_mutex_lock(&srv->rq_mutex);
    }

    if (!list_is_empty(&srv->done_user_bios)) {
      empty = false;
      node = list_node_first(&srv->done_user_bios);
      list_delfromhead(&srv->done_user_bios);
      pthread_mutex_unlock(&srv->rq_mutex);

      /* complete user I/O request */
      blkio = container_of(node, blkio_t, node);
      biorq = blkio->biorq;
      bdata = (bio_usr_data_t*)biorq->server_private;

      /* check if there is completed without errors read request */
      if (!(biorq_rw(biorq) || blkio->status.err)) {
        if (storage_cacheable(stor))
          srv->ioops.send_cached_data(srv, biorq);
        else
          srv->ioops.send_non_cached_data(srv, biorq);
      } else if (!blkio->status.err) {
        __reply_with_replcode(srv->io_port, bdata->size, bdata->msgid);
      } else {
        __reply_with_error(srv->io_port, blkio->status.err, bdata->msgid);
      }

      if (biorq->complete) {
        biorq->complete(blkio);
      }
      put_server_biorq(biorq);
      destroy_blkio(blkio);
    } else {
      pthread_mutex_unlock(&srv->rq_mutex);
    }

    while (empty) {
      if (!event_wait(&srv->event))
        empty = false;
    }
  }

  return NULL;
}

extern int blk_requests_number;
extern int biorqs_number;
extern int blkios_number;

static int server_ioctl(struct client *client, uint16_t devno, uint16_t cmd,
                        void *buf, uint16_t data_size, uint16_t buf_size)
{
  blkdev_server_t *srv = client->server;
  struct mdata *md;
  genstor_t *stor;
  bd_part_t **parts;
  int ret = 0;

  switch (cmd) {
  case BLK_SERV_GET_MDATA:
    md = (struct mdata*)buf;
    stor = srv->stor;
    parts = stor->parts;

    if (buf_size < sizeof(struct mdata)) {
      ret = -EINVAL;
      break;
    }

    if (devno) {
      md->capacity = parts[devno - 1]->nr_sectors;
    } else {
      md->capacity = stor->capacity;
    }
    md->hw_sect_bits = stor->sect_shift;
    md->cmd_port = srv->cmd_port;
    ret = sizeof(struct mdata);
    break;
  case BLK_SERV_SET_ASYNC_PORT:
  {
     prtid_t p = *(prtid_t*)buf;

     if (__async_chan_is_valid(client->async_chan)) {
       ipc_channel_close(client->async_chan);
     }
     client->async_chan = ipc_channel_open(client->pid, p, 0);
     if ((signed)client->async_chan < 0) {
       ret = (signed)client->async_chan;
       client->async_chan = -1L;
     }
     break;
  }
  default:
    ret = -EINVAL;
    break;
  }

  return ret;
}

static inline void __init_client(struct client *client, pid_t pid,
                                 blkdev_server_t *srv)
{
  client->pid = pid;
  client->server = srv;
  client->async_chan = -1L;
  usrtc_init(&client->files, USRTC_SPLAY, BLK_MAX_FDS, __compare_fid);
}

static void __free_client_fds(struct client *client)
{
  usrtc_node_t *node;
  struct blk_fd *fd;

  for (node = usrtc_first(&client->files); node != NULL;
       node = usrtc_next(&client->files, node)) {

    fd = usrtc_node_getdata(node);
    free(node);
    put_blk_fd(fd);
  }
}

static int register_fd(blkdev_server_t *srv, uni_rpc_t *rpc)
{
  int ret = -ENOMEM;
  pid_t pid = rpc->arg0;
  struct client_list *cl_list;
  struct client *client = NULL;
  struct blk_fd *fd = NULL;
  usrtc_node_t *fd_node = NULL, *cl_node = NULL;
  genstor_t *stor = srv->stor;
  bool new_cl = false;
  uint16_t fid = rpc->arg1;
  uint16_t devno = rpc->arg2;

  if (devno) {
    if (!parts_rlock(stor))
      return -EBUSY;

    if (!check_part_exist(stor, devno)) {
      parts_runlock(stor);
      return -ENODEV;
    }
    /* arg3 contains open flags */
    if (!check_storage_caps(stor, devno, rpc->arg3)) {
      parts_runlock(stor);
      return -EPERM;
    }
  }
  cl_list = &srv->client_table[pid & CLIENT_HASH_MASK];
  clients_lock(cl_list);

  ret = -ENOMEM;
  /* lookup for the client and add it if it doesn't present */
  cl_node = usrtc_lookup(&cl_list->clients, &pid);
  if (cl_node == NULL) {
    cl_node = malloc(sizeof(usrtc_node_t));
    if (cl_node == NULL)
      goto out;

    new_cl = true;
    client = malloc(sizeof(struct client));
    if (client == NULL)
      goto out;

    atomic_set(&client->refcnt, 1);
  } else {
    client = usrtc_node_getdata(cl_node);
    get_client(client);
  }

  if (!new_cl && usrtc_lookup(&client->files, &fid)) {
    ret = -EBUSY;
    goto out;
  }

  /* add a new file descriptor */
  fd_node = malloc(sizeof(usrtc_node_t));
  if (fd_node == NULL)
    goto out;

  fd = malloc(sizeof(struct blk_fd));
  if (fd == NULL)
    goto out;

  fd->devno = rpc->arg2;
  ret = stor->ops->open(stor, fd->devno, fd->flags);
  if (ret)
    goto out;

  if (new_cl) {
    __init_client(client, pid, srv);
    get_client(client);
    usrtc_node_init(cl_node, client);
    usrtc_insert(&cl_list->clients, cl_node, &client->pid);
  }

  fd->client = client;
  fd->id = fid;
  fd->flags = rpc->arg3;
  atomic_set(&fd->refcnt, 1);
  usrtc_node_init(fd_node, fd);
  usrtc_insert(&client->files, fd_node, &fd->id);

out:
  clients_unlock(cl_list);
  if (devno) {
    parts_runlock(stor);
  }
  if (ret) {
    /* free allocated resources */
    if (fd_node != NULL) {
      free(fd_node);
      if (fd != NULL)
        free(fd);
    }
    if (client != NULL)
      put_client(client);

    if (new_cl)
      free(cl_node);
  }

  return ret;
}

static int unregister_fd(blkdev_server_t *srv, pid_t pid, uint16_t fid)
{
  int ret = -ENOENT;
  usrtc_node_t *cl_node, *fd_node = NULL;
  struct client_list *cl_list;
  struct blk_fd *fd;
  struct client *client;

  cl_list = &srv->client_table[pid & CLIENT_HASH_MASK];
  clients_lock(cl_list);

  cl_node = usrtc_lookup(&cl_list->clients, &pid);
  if (cl_node != NULL) {
    client = usrtc_node_getdata(cl_node);
    fd_node = usrtc_lookup(&client->files, &fid);
  }
  if (fd_node != NULL) {
    fd = usrtc_node_getdata(fd_node);
    usrtc_delete(&client->files, fd_node);
    free(fd_node);
    put_blk_fd(fd);
    if (!usrtc_count(&client->files)) {
      usrtc_delete(&cl_list->clients, cl_node);
      free(cl_node);
      put_client(client);
    }
    ret = 0;
  }

  clients_unlock(cl_list);

  return 0;
}

static struct blk_fd *lookup_get_fd(blkdev_server_t *srv, pid_t pid, uint16_t fid)
{
  usrtc_node_t *node;
  struct client *client;
  struct client_list *cl_list;
  struct blk_fd *fd = NULL;

  cl_list = &srv->client_table[pid & CLIENT_HASH_MASK];
  clients_lock(cl_list);
  node = usrtc_lookup(&cl_list->clients, &pid);
  if (node != NULL) {
    client = usrtc_node_getdata(node);
    node = usrtc_lookup(&client->files, &fid);
    if (node != NULL) {
      fd = usrtc_node_getdata(node);
      get_blk_fd(fd);
    }
  }
  clients_unlock(cl_list);

  return fd;
}

static int clone_fds(blkdev_server_t *srv, pid_t parent, pid_t clone)
{
  int ret = 0;
  struct client *src, *dst = NULL;
  struct blk_fd *fd1, *fd2;
  struct client_list *src_list, *dst_list = NULL;
  usrtc_node_t *node1, *node2, *dst_node = NULL;

  src_list = &srv->client_table[parent & CLIENT_HASH_MASK];
  clients_lock(src_list);
  node1 = usrtc_lookup(&src_list->clients, &parent);
  if (node1 == NULL) {
    ret = -ESRCH;
    goto out;
  }
  src = usrtc_node_getdata(node1);
  dst_list = &srv->client_table[clone & CLIENT_HASH_MASK];
  clients_lock(dst_list);

  /* add new client */
  ret = -ENOMEM;
  dst_node = malloc(sizeof(usrtc_node_t));
  if (dst_node == NULL)
    goto out;

  dst = malloc(sizeof(struct client));
  if (dst == NULL)
    goto out;

  atomic_set(&dst->refcnt, 1);
  __init_client(dst, clone, srv);

  /* copy file descriptors */
  for (node1 = usrtc_first(&src->files); node1 != NULL;
       node1 = usrtc_next(&src->files, node1)) {

    node2 = malloc(sizeof(usrtc_node_t));
    if (node2 == NULL)
      break;

    fd2 = malloc(sizeof(struct blk_fd));
    if (fd2 == NULL) {
      free(node2);
      break;
    }

    fd1 = usrtc_node_getdata(node1);
    memcpy(fd2, fd1, sizeof(blk_fd));
    atomic_set(&fd1->refcnt, 1);
    get_client(dst);
    fd2->client = dst;
    usrtc_node_init(node2, fd2);
    usrtc_insert(&dst->files, node2, &fd2->id);
  }

  if (fd1 == NULL) {
    /* file descriptors are copied successfully */
    ret = 0;
    usrtc_node_init(dst_node, dst);
    usrtc_insert(&dst_list->clients, dst_node, &dst->pid);
  }

out:
  if (dst_list) {
    clients_unlock(dst_list);
  }
  clients_unlock(src_list);
  if (ret && (dst_node != NULL)) {
    if (dst != NULL) {
      __free_client_fds(dst);
      put_client(dst);
    }
    free(dst_node);
  }

  return ret;
}

static int client_killed(blkdev_server_t *srv, pid_t pid)
{
  int ret = 0;
  struct client *client;
  usrtc_node_t *node;
  struct client_list *cl_list;

  cl_list = &srv->client_table[pid & CLIENT_HASH_MASK];
  clients_lock(cl_list);
  node = usrtc_lookup(&cl_list->clients, &pid);
  if (node == NULL) {
    ret = -ESRCH;
  } else {
    client = usrtc_node_getdata(node);
    __free_client_fds(client);
    usrtc_delete(&cl_list->clients, node);
    free(node);
    put_client(client);
  }
  clients_unlock(cl_list);

  return ret;
}

static int server_process_cmd(blk_thread_t *thread, ipc_port_msg_info_t *msginfo, bool *reply)
{
  int ret = 0;
  uni_rpc_t rpc;
  struct blk_fd *fd;
  blkdev_server_t *srv = thread->srv;
  genstor_t *stor = srv->stor;
  uint16_t msgid = msginfo->msg_id;
  pid_t spid;
  uint16_t partno, flags;
  struct cache_head *heads;
  uint16_t *proc_heads;

  memcpy(&rpc, thread->buf, sizeof(rpc));
  spid = msginfo->sender_pid;
  partno = rpc.arg2;

  if (rpc.rpc_call == RPC_ALIVE || rpc.rpc_call == RPC_STATUS)
    return 0;

  if(spid == srv->devfs_pid) {
    switch (rpc.rpc_call) {
    case RPC_REGISTER:
      ret = register_fd(srv, &rpc);
      break;
    case RPC_UNREGISTER:
      /* arg0 - fd client's pid, arg1 - file ID */
      ret = unregister_fd(srv, rpc.arg0, rpc.arg1);
      break;
    case RPC_PIDFDCLONE:
      /* arg0 - pid of the killed client */
      ret = client_killed(srv, rpc.arg0);
      break;
    case RPC_PIDFDKILL:
      /* arg0 - pid of parent, arg1 - pid of clone */
      ret = clone_fds(srv, rpc.arg0, rpc.arg1);
      break;
    default:
      ret = -EINVAL;
      break;
    }
  } else {
    /* commands from file systems */
    fd = lookup_get_fd(srv, spid, rpc.arg1);
    if (fd == NULL)
      return -EBADF;

    switch(rpc.rpc_call) {
    case RPC_READ:
    case RPC_WRITE:
      if (!check_access(fd, rpc.rpc_call)) {
        ret = -EACCES;
        break;
      }

      /* TO DO [dg]: update to the new protocol */
      heads = (struct cache_head*)((ulong_t)thread->buf + sizeof(uni_rpc_t));
      flags = rpc.arg1;
      if ((flags & BIORQ_ASYNC) && !__async_chan_is_valid(fd->client->async_chan)) {
        ret = -EINVAL;
      } else {
        ret = process_fsio(&rpc, stor, fd, msgid, heads,
            msginfo->msg_len - sizeof(uni_rpc_t), proc_heads);
      }
      break;
    case RPC_REGISTER:
    case RPC_UNREGISTER:
    case RPC_PIDFDCLONE:
    case RPC_PIDFDKILL:
      ret = -EPERM;
      break;
    default:
      ret = -EINVAL;
      break;
    }
    put_blk_fd(fd);
  }

  return ret;
}

static ssize_t server_process_io(blk_thread_t *thread, ipc_port_msg_info_t *msginfo,
                                 bool *reply)
{
  ssize_t ret = 0;
  uni_rpc_t rpc;
  int msgid = msginfo->msg_id;
  struct blk_fd *fd;
  blkdev_server_t *srv = thread->srv;
  genstor_t *stor = srv->stor;
  pid_t spid;
  void *buf = NULL;
  bool locked = false;

  memcpy(&rpc, thread->buf, sizeof(rpc));
  spid = msginfo->sender_pid;
  /* arg1 contains file id */
  fd = lookup_get_fd(srv, spid, rpc.arg1);
  if (fd == NULL)
    return -EBADF;

  if (fd->devno) {
    locked = parts_rlock(stor);
    if (!locked) {
      ret = -EBUSY;
      goto out;
    }
    if (!check_part_exist(stor, fd->devno)) {
      ret = -ENODEV;
      goto out;
    }
  }

  switch(rpc.rpc_call) {
  case RPC_READ:
  case RPC_WRITE:
  {
    bio_rpc_t bio_rpc;
    biorq_t b;
    bio_usr_data_t bdata;

    if (!check_access(fd, rpc.rpc_call)) {
      ret = -EACCES;
      break;
    }


    if (rpc.size > IPC_PORT_MSGLEN_MAX) {
      ret = -EINVAL;
      break;
    }

    bio_rpc.offset = rpc.offset;
    bio_rpc.size = rpc.size;
    bio_rpc.rw = (rpc.rpc_call == RPC_WRITE);
    bio_rpc.prio = 0;
    if (fd->devno) {
      bio_rpc.part = stor->parts[fd->devno - 1];
    } else {
      bio_rpc.part = NULL;
    }

    memset(&b, 0, sizeof(b));
    ret = check_prepare_user_ioreq(thread, &b, &bio_rpc, fd);
    if (ret)
      break;

    bdata.sect_off = bio_rpc.offset & ((1 << stor->sect_shift) - 1);
    bdata.msgid = msgid;
    bdata.size = bio_rpc.size;
    bdata.fd = fd;
    b.server_private = &bdata;

    if (rpc.rpc_call == RPC_READ) {
      ret = stor->ops->read(stor, &bio_rpc, &b);
      if (!IS_ERR_VALUE(ret)) {
        if (b.complete_sectors == b.nr_sectors) {
          if (storage_cacheable(stor)) {
            send_cached_data(srv, &b);
          } else {
            send_non_cached_data(srv, &b);
          }
        }
        *reply = false;
      }
    } else {
      ret = stor->ops->write(stor, &bio_rpc, &b);
      if (!IS_ERR_VALUE(ret) && (fd->flags & O_SYNC)) {
        *reply = false;
      }
    }
    break;
  }
  case RPC_IOCTL:
  {
    ioctl_entry_t *ctl;
    size_t in_size;
    uint16_t cmd;
    size_t size;

    cmd = _RPC_IOCMD(&rpc);
    in_size = msginfo->msg_len - sizeof(uni_rpc_t);
    /* the offset field contains output data size */
    size = MAX(in_size, rpc.offset);

    if (size > 0) {
      if (size > IPC_PORT_MSGLEN_MAX) {
        ret = -EINVAL;
        break;
      }
      buf = malloc(size);
      if (buf == NULL) {
        ret = -ENOMEM;
        break;
      }

      if (in_size) {
        ret = srv->ioops.get_imsg_data(thread, 0, in_size, buf);
        if (ret) {
          break;
        }
      }
    }

    if (_RPC_IOCTYPE(&rpc) == SERVER_CTLS) {
      ret = server_ioctl(fd->client, fd->devno, cmd, buf, in_size, size);
      if (ret > 0) {
        __reply_with_data(srv->io_port, buf, ret, msgid);
        *reply = false;
      } else if (!ret) {
        __reply_with_rpc(srv->io_port, &rpc, msgid);
        *reply = false;
      }
      break;
    }

    if (stor->ops->ioctl == NULL) {
      ret = -ENOSYS;
      break;
    }

    ret = 0;
    ctl = (ioctl_entry_t*)malloc(sizeof(ioctl_entry_t));
    if (ctl == NULL) {
      ret = -ENOMEM;
    } else {
      ctl->cmd = cmd;
      ctl->devno = fd->devno;
      ctl->data = buf;
      ctl->data_size = in_size;
      ctl->buf_size = size;
      ctl->msgid = msginfo->msg_id;
      /*
       * add to queue and then try to receive the next message,
       * control thread will reply to this message
       */

      __list_add_raise_event(&srv->ctl_reqs, &ctl->node, &srv->event, &srv->rq_mutex);
      *reply = false;
    }
    break;
  }
  case RPC_FSTAT:
  {
    struct stat stat;

    if (fd->devno)
      stat.st_size = stor->parts[fd->devno - 1]->nr_sectors << stor->sect_shift;
    else
      stat.st_size = stor->capacity << stor->sect_shift;

    stat.st_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP;
    stat.st_blksize = PAGE_SIZE;
    stat.st_uid = 0;
    stat.st_gid = 0;  /* Fixme [dg]: replace later by disk group id */

    __reply_with_data(srv->io_port, &stat, sizeof(struct stat), msginfo->msg_id);
    *reply = false;
    break;
  }
  default:
    ret = -EINVAL;
    break;
  }

out:
  if (locked) {
    parts_runlock(stor);
  }
  put_blk_fd(fd);
  if (ret && (buf != NULL))
    free(buf);

  return ret;
}

int server_prepare_biorqs(biorq_t **biorq, int nr, biorq_t *srv_biorq)
{
  int i, ret = 0;
  bio_usr_data_t *bdata;
  size_t s;
  void *buf;
  struct blk_fd *fd;
  genstor_t *stor;
  uint32_t flags;
  int write;

  flags = biorq[0]->flags;
  write = biorq_rw(biorq[0]);
  bdata = srv_biorq->server_private;
  fd = bdata->fd;
  stor = fd->client->server->stor;
  bdata = NULL;

  for (i = 0; i < nr; i++) {
    if (biorq_need_reply(biorq[i])) {
      if (bdata == NULL) {
        bdata = malloc(sizeof(bio_usr_data_t));
        if (bdata == NULL) {
          ret = -ENOMEM;
          break;
        }
        memcpy(bdata, srv_biorq->server_private,
               sizeof(bio_usr_data_t));
        get_blk_fd(fd);
        atomic_set(&bdata->refcnt, 0);
      }
      __get_bio_udata(bdata);
      biorq[i]->server_private = bdata;
      set_biorq_flags(biorq[i], BIORQ_SERV_PREPARED);
    } else {
      biorq[i]->owner = fd->client->pid;
    }

    /*
     * allocate buffer if storage is not cacheable and copy data
     * if there is a write request
     */
    if(!storage_cacheable(stor)) {
      if (biorq_double_end(biorq[i])) {
        set_biorq_flags(biorq[i], BIORQ_RDONLY_BUFFER);
        s = biorq[i]->wb_chunk.size;
      } else {
        s = biorq[i]->nr_sectors << stor->sect_shift;
      }

      buf = malloc(s);
      if (buf == NULL) {
        ret = -ENOMEM;
        break;
      }

      if (biorq_orig_rw(biorq[i])) {
        ret = copy_biorq_data(srv_biorq, biorq[i]->data_off, s, buf);
        if (ret) {
          free(buf);
          break;
        }
        set_biorq_flags(biorq[i], BIORQ_DATA_AVAILABLE);
      }
      set_biorq_flags(biorq[i], BIORQ_BUFFERED);
      biorq[i]->data_buf = buf;
    }
  }

  return ret;
}

/* This function must be rewritten due to page cache redisign */
static int process_fsio(uni_rpc_t *rpc, genstor_t *stor, struct blk_fd *fd,
                        uint16_t msgid, struct cache_head *heads,
                        uint32_t heads_size, uint16_t *nr_proc_heads)
{
#if 0
  sector_t sector;
  uint32_t sects;
  bd_part_t *part = NULL;
  sector_t capacity;
  int ret = 0;
  biorq_t *bio;
  short prio = rpc->arg3;
  uint32_t flags = rpc->arg4 | BIORQ_NEED_REPLY;
  //uint16_t blocks_per_page = rpc->arg5;
  bio_fs_data_t *bdata;
  struct cache_head *heads_end, *curr_head = heads;

  heads_end = (struct cache_head*)((ulong_t)heads + heads_size);
  while ((curr_head < heads_end) && !ret) {
    if (fd->devno) {
      part = stor->parts[fd->devno - 1];
      capacity = part->nr_sectors;
    } else {
      capacity = stor->capacity;
    }

    if (sector > capacity) {
      ret = -EINVAL;
      break;
    }

    if (rpc->rpc_call == RPC_WRITEDIO) {
      if ((stor->flags & STORAGE_RDONLY) || (part->flags & PART_RDONLY)) {
        ret = -EPERM;
        break;
      }

      flags |= (BIORQ_RW | BIORQ_ORIG_RW);
    }

    sects = MIN(sects, (capacity - sector));

    /* TO DO [dg]: map pages of this object */

    bio = alloc_init_biorq(sector, sects, BIORQ_TYPE_FS, flags);
    if (bio == NULL) {
      ret = -ENOMEM;
      break;
    }

    bdata = (bio_fs_data_t*)malloc(sizeof(bio_fs_data_t));
    if (bdata == NULL)
      ret = -ENOMEM;

    if (!ret) {
      bdata->msgid = msgid;
      bdata->mem_obj_id = 0;
      bdata->obj_off = 0;
      bio->server_private = bdata;
      bdata->fd = fd;
      get_blk_fd(fd);
      set_biorq_flags(bio, BIORQ_SERV_PREPARED);

      ret = stor->ops->make_request(stor, part, bio, NULL, prio);
    }

    if (ret)
      put_server_biorq(bio);
  }
#endif

  return 0;
}

static int check_prepare_user_ioreq(blk_thread_t *thread, biorq_t *biorq,
                                    bio_rpc_t *bio_rpc, struct blk_fd *fd)
{
  sector_t capacity;
  ipc_port_msg_info_t *msginfo = &thread->msg_info;
  genstor_t *stor  = thread->srv->stor;
  int sect_shift = stor->sect_shift;
  size_t size;
  uint32_t flags = 0;
  bd_part_t *part = bio_rpc->part;

  biorq->start_sector = bio_rpc->offset >> sect_shift;

  if (part)
    capacity = part->nr_sectors;
  else
    capacity = stor->capacity;

  if (biorq->start_sector > capacity)
    return -EINVAL;

  if (bio_rpc->rw) {
    if (fd->devno && !check_storage_caps(stor, fd->devno - 1, O_WRONLY))
      return -EPERM;

    size = msginfo->msg_len - sizeof(uni_rpc_t);
    flags |= (BIORQ_RW | BIORQ_ORIG_RW);
    if (fd->flags & O_SYNC)
      flags |= BIORQ_NEED_REPLY;
  } else {
    size = bio_rpc->size;
    flags |= BIORQ_NEED_REPLY;
  }

  if (!size) {
    bio_rpc->size = 0;
    return 0;
  }

  biorq->type = BIORQ_TYPE_SERVER;
  /* check the requested size don't exceed storage size */
  size = MIN(size, (capacity << sect_shift) - bio_rpc->offset);
  bio_rpc->size = size;
  bio_rpc->prio = 0;
  bio_rpc->part = part;
  biorq->flags = flags;

  biorq->nr_sectors = ((bio_rpc->offset + size) >> sect_shift) -
                       biorq->start_sector + 1;
  if ( !((bio_rpc->offset + size) & ((1 << sect_shift) - 1)) )
    biorq->nr_sectors--;

  return 0;
}

void put_server_biorq(biorq_t *biorq)
{
  if (biorq_serv_prepared(biorq)) {
    if (biorq_type_fs(biorq)) {
      bio_fs_data_t *bdata = (bio_fs_data_t*)biorq->server_private;

      put_blk_fd(bdata->fd);
      /* TO DO [dg]: change state of fs memory object pages and unmap pages */

      free(bdata);
    } else {
      bio_usr_data_t *bdata = (bio_usr_data_t*)biorq->server_private;

      __put_bio_udata(bdata);
    }
  }
  put_biorq(biorq);
}

bool biorq_session_is_active(biorq_t *biorq)
{
  struct blk_fd *fd;

  if (biorq_type_fs(biorq)) {
    bio_fs_data_t *bdata = biorq->server_private;

    fd = bdata->fd;
  } else {
    bio_usr_data_t *bdata = biorq->server_private;

    fd = bdata->fd;
  }

  return !(fd->flags & BLKFD_CLOSED);
}

int copy_biorq_data(biorq_t *biorq, uint32_t offset, uint32_t len, void *buf)
{
  int ret = 0;
  /* fake thread object to read data */
  blk_thread_t th;
  blkdev_server_t *srv;

  if (!(biorq_serv_prepared(biorq) || biorq_type_server(biorq)))
    return -EINVAL;

  /*
   * if there is a server biorq, it contains pointer to the
   * server thread buffer
   */
  th.buf = biorq->data_buf;
  if (biorq_type_fs(biorq)) {
    /* TO DO [dg]: implement when memory object will be available */
  } else {
    bio_usr_data_t *bdata = (bio_usr_data_t*)biorq->server_private;
    srv = bdata->fd->client->server;

    th.srv = srv;
    th.msg_info.msg_id = bdata->msgid;
    ret = srv->ioops.get_imsg_data(&th, offset, len, buf);
  }

  return ret;
}

int copy_biorq_data_sectors(biorq_t *biorq, uint32_t start_sector,
                            uint32_t nr_sectors, int sect_shift, void *buf)
{
  uint32_t offset, len;

  if (!biorq_serv_prepared(biorq))
    return -1;

  offset = (start_sector - biorq->start_sector) << sect_shift;
  len = nr_sectors << sect_shift;

  return copy_biorq_data(biorq, offset + biorq->data_off, len, buf);
}

int _copy_biorq_data_to_cache(biorq_t *biorq, int cache_id)
{
  return 0;
}
