/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org>
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 *
 * LibOS/lib/devserver.c:
 *      device server layer
 *
 */

#include <stdbool.h>
#include <sys/types.h>
#include <sys/poll.h>
#include <ipc/port.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ns/ns.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <devio/devio.h>
#include <datalib/usrtc.h>
#include <pthread.h>
#include <sys/ptd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <chardev.h>
#include <datastructs.h>
#include <spinlock.h>

#include "internal.h"

#include <stdio.h>

static deque_t works;
static event_t works_event;
static spinlock_t works_lock;

#define DEVICE_ALLOC_STEP 5
static device_t **devices = NULL;

static pthread_t *worker_threads;

static usrtc_t clients;
static spinlock_t clients_lock = SPINLOCK_INITIALIZER;

#define DEBUG

#ifdef DEBUG
static void log(int level, const char *fmt,... )
{
  static spinlock_t mutex = SPINLOCK_INITIALIZER;
  static char tmp[1024];
  static char tmp2[1024];

  va_list ap;
  if(server->loglevel < level) return;

  if(!server->ops_tbl->log) return;
  if(server->loglevel<0) return;

  // Warning! if server log method wakes some devices
  // deadlock is very possible!
  spinlock_lock(&mutex);

  memset(tmp, '\0', 256);

  va_start(ap, fmt);
  vsnprintf(tmp, 1024, fmt, ap);
  va_end(ap);

  snprintf(tmp2, 1024, "%s (tid %d): %s: %s", server->drvname, gettid(), (level == ERROR) ? "ERROR" :
           ((level == WARNING) ? "WARNING" : "INFO"), tmp);

  server->ops_tbl->log(tmp2);

  spinlock_unlock(&mutex);
  return;
}
#else
#define log(x,...)
#endif

static ulong_t get_work(void)
{
  ulong_t w;

  while(1) {
    log(INFO, "Getting work...\n");
    spinlock_lock(&works_lock);
    w = (ulong_t)fifo_pop(&works);
    spinlock_unlock(&works_lock);
    log(INFO, "work = %ld.\n", w);
    if(w) break;
    else event_wait(&works_event);
  }

  return w;
}

static int put_work(ulong_t work)
{
  int r;

  spinlock_lock(&works_lock);
  r = fifo_push(&works, (void *)work);
  spinlock_unlock(&works_lock);

  event_signal(&works_event);
  log(INFO, "Event signalled.\n");
  return r;
}

static status_t dummy_read(int devno, void *buf, size_t size, size_t offset, void **interim)
{
  return -ENOSYS;
}

static status_t dummy_write(int devno, void *buf, size_t size, size_t offset, void **interim)
{
  return -ENOSYS;
}

static status_t dummy_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize)
{
  return -ENOSYS;
}

static int pid_compare(const void *a, const void *b)
{
  return *(pid_t *)a - *(pid_t *)b;
}

static int short_compare(const void *a, const void *b)
{
  return *(uint16_t *)a - *(uint16_t *)b;
}

//TODO: Add cleanup. Add destruction function.
int init_multi_device_server(device_server_t *srv, size_t in, size_t out, const char *drvname,
                             struct device_ops *ops_tbl, int num_threads)
{
  status_t r;
  int i;

  if(!srv)
    return -5;

  server = srv;
  srv->ops_tbl=ops_tbl;
  srv->drvname = drvname;
  server->loglevel = DEFAULT_LOGLEVEL;
  server->num_devices = 0;

  if(!drvname) return -EINVAL;

  deque_init(&works);

  spinlock_init(&works_lock, NULL);
  
  if(event_init(&works_event) != 0) {
    return -EAGAIN;
  }

  worker_threads = (pthread_t*)malloc(sizeof(pthread_t) * num_threads);
  if(!worker_threads) return -ENOMEM;

  for(i = 0; i < num_threads; i++) {
    if(pthread_create(&worker_threads[i], NULL, worker, NULL) != 0) {
      return -EINVAL;
    }
  }

  devices = malloc(sizeof(device_t*) * MAX_DEVNO);

  if(!devices)
    return -ENOMEM;

  memset(devices, 0, sizeof(device_t*) * MAX_DEVNO);


  deque_init(&works);

  usrtc_init(&clients, USRTC_SPLAY, USRTC_COUNT_T_MAX, pid_compare);

  srv->num_threads = num_threads;

  /* initiate rpc buffers */

  srv->cmdbuf_size = sizeof(uni_rpc_t);
  srv->buf_size[IN] = in;
  srv->buf_size[OUT] = out;

  srv->cmdbuf=malloc(sizeof(uni_rpc_t));
  if(!srv->cmdbuf)
    return -ENOMEM;

  srv->pid = getpid();
  srv->hw_pid = -1;

  /* create ports */
  r = ipc_port_open(IPC_BLOCKED_ACCESS, SRV_QUEUE_SIZE);
  if(r < 0) {
    return -ENOMEM;
  } else {
    srv->cmd_port = r;
  }

  r = ipc_port_open(IPC_BLOCKED_ACCESS, SRV_QUEUE_SIZE);
  if(r < 0) {
    return -ENOMEM;
  } else {
    srv->io_port=r;
  }

  if(ops_tbl->read==NULL) {
    ops_tbl->read = dummy_read;
  }
  if(ops_tbl->write==NULL) {
    ops_tbl->write = dummy_write;
  }
  if(ops_tbl->ioctl==NULL) {
    ops_tbl->ioctl = dummy_ioctl;
  }

  server = srv;
  return 0;
}

int add_device(int devno, const char *path)
{
  int i;
  device_t *d;
  int ret, r, fd;

  if(devno < 0) {
    return -EINVAL;
  }

  for(i = 0; i < server->num_devices - 1; i++) {
    if(devices[i]->devno == devno) {
      return -EBUSY;
    }
  }

  d = malloc(sizeof(*d));

  if(d == NULL) {
    return -ENOMEM;
  }

  memset(d, 0, sizeof(*d));

  d->devno = devno;

  spinlock_init(&d->lock[IN], NULL);

  spinlock_init(&d->lock[OUT], NULL);

  d->pending_wake[0] = 0;
  d->pending_wake[1] = 0;

  r = event_init(&d->event);
  if(r < 0) {
    ret = -EINVAL;
    goto ad_cleanup3;
  }

  list_init_head(&d->ioqueue[IN]);
  list_init_head(&d->ioqueue[OUT]);

  fd = creat(path, (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH));
  if(fd < 0) {
    ret = fd;
    goto ad_cleanup4;
  }

  r = fstat(fd, &d->statbuf);
    if(r != 0) {
    ret = r;
    goto ad_cleanup5;
  }

  
  if(server->hw_pid == -1) {
    server->hw_pid = getfpid(fd);
  }

  r = deviofs_assign(fd, server->cmd_port, server->io_port, DFS_CHARDEV, server->num_devices);
  if(r < 0) {
    ret = r;
    goto ad_cleanup5;
  }

  devices[server->num_devices] = d;
  server->num_devices++;
  return 0;

  ad_cleanup5:
  unlink(path);

  ad_cleanup4:
  event_destroy(&d->event);

  ad_cleanup3:
  spinlock_destroy(&d->lock[OUT]);

  spinlock_destroy(&d->lock[IN]);
  free(d);

  return ret;
}

// TODO: Add cleanup
int init_simple_device_server(size_t in, size_t out, const char *drvname,
                              const char *devname, struct device_ops *ops_tbl)
{
  device_server_t *srv;
  int r;

  srv = malloc(sizeof(*srv));
  if(!srv) return -ENOMEM;

  r = init_multi_device_server(srv, in, out, drvname, ops_tbl, 1);
  if(r != 0) {
    return r;
  }

  return add_device(0, devname);
}

int init_simple_mt_device_server(size_t in, size_t out, const char *drvname,
                                 const char *devname, struct device_ops *ops_tbl, int num_threads)
{
  device_server_t *srv;
  int r;

  srv = malloc(sizeof(*srv));
  if(!srv) return -ENOMEM;

  r =  init_multi_device_server(srv, in, out, drvname, ops_tbl, num_threads);
  if(r != 0) {
    return r;
  }

  return add_device(0, devname);
}

void set_loglevel(int loglevel)
{
  server->loglevel = loglevel;
}

// Wake requests in the device waitqueue
void wake_device(int devno, int dir) {

  log(INFO, "Waking device %d, dir %d\n", devno, dir);

  put_work(MKWORK(devno, dir));
}

void do_waiting_work(int devno, int dir)
{
  msg_data_t *msg;

  spinlock_lock(&devices[devno]->lock[dir]);
  msg = devices[devno]->waiting[dir];
  if(msg) {
    devices[devno]->waiting[dir] = NULL;
  } else {
    devices[devno]->pending_wake[dir] = 1;
  }
  spinlock_unlock(&devices[devno]->lock[dir]);

  if(msg) {
    do_messages(devno, NULL, NULL, NULL, msg);
  }
}

static inline void arrange(size_t *sz, size_t szmax, const char *msg)
{
  if(*sz > szmax) {
    log(WARNING, "%s fixed from %lu to %lu\n", msg, *sz, szmax);
    *sz = szmax;
  }
}

void do_listen(void)
{
  status_t r;
  ipc_port_msg_info_t msginfo;
  uni_rpc_t rpc;
  void *data;
  msg_data_t *msg;
  size_t rsize;

  log(INFO, "Entering listening.\n");

  retry:
  data = msg = NULL;

  r = ipc_port_rcvmsg_mi(server->io_port, sizeof(uni_rpc_t), &rpc, &msginfo);

  if(r < sizeof(uni_rpc_t)) {
    log(WARNING, "Message size is less than uni_rpc_t, ignoring. pid and tid are %d and %d\n",
        msginfo.sender_pid, msginfo.sender_tid);
    goto retry;
  }

  put_work(WORK_LISTEN);

  rsize = msginfo.msg_len - sizeof(uni_rpc_t);

  // Check message size boundaries and read incoming message data if needed.
  // FIXME: remove old DIO variants when no legacy.
  switch(rpc.rpc_call) {
    case RPC_READDIO:
      rpc.rpc_call = RPC_READ;
    case RPC_READ:
      if(rpc.size < 0) {
        log(WARNING, "Invalid read size, ignoring. pid and tid are %d and %d\n",
            msginfo.sender_pid, msginfo.sender_tid);
        reply_with_error(0, EINVAL, msginfo.msg_id);
        return;
      }
      arrange(&rpc.size, server->buf_size[OUT], "read rpc.size");
      break;
    case RPC_WRITEDIO:
      rpc.rpc_call = RPC_WRITE;
    case RPC_WRITE:
      if(rpc.size < 0) {
        log(WARNING, "Invalid write size, ignoring. pid and tid are %d and %d\n",
            msginfo.sender_pid, msginfo.sender_tid);
        reply_with_error(0, EINVAL, msginfo.msg_id);
        return;
      }
      arrange(&rpc.size, rsize, "write rpc.size");
      arrange(&rpc.size, server->buf_size[IN], "write rpc.size");
      break;
    case RPC_IOCTL:
      // depending on rpc.size and rpc.offset it is in/out/inout ioctl
      if(rpc.size != rsize || rpc.size > server->buf_size[IN] || rpc.offset > server->buf_size[OUT] ||
         (_RPC_IOCDIR(&rpc) == _IOC_READ && rpc.size > 0) ||
         (_RPC_IOCDIR(&rpc) == _IOC_WRITE && rpc.offset > 0) ||
         (_RPC_IOCDIR(&rpc) == _IOC_WRFIXED && (rpc.size != rpc.offset))) {
        log(WARNING, "Invalid ioctl messave size, ignoring. pid and tid are %d and %d (%lu %lu %lu %d)\n",
            msginfo.sender_pid, msginfo.sender_tid, rpc.size, rpc.offset, rsize, _RPC_IOCDIR(&rpc));
        reply_with_error(0, EINVAL, msginfo.msg_id);
        return;
      }
      break;
    case RPC_FSTAT:
      rpc.size = 0;
      rpc.offset = 0;
      break;
    default:
      log(WARNING, "Unknown request in rpc_call field: %d, ignored.\n", rpc.rpc_call);
      reply_with_error(0, ENOSYS, msginfo.msg_id);
      return;
  }

  if(rpc.size > 0) {
    data = malloc(rpc.size);
    if(!data) {
      reply_with_error(0, ENOMEM, msginfo.msg_id);
      log(ERROR, "Not enough memory for read data.\n");
      return;
    }
    if(rpc.rpc_call == RPC_WRITE || rpc.rpc_call == RPC_IOCTL) {
      r = ipc_port_read_msg(server->io_port, msginfo.msg_id, sizeof(uni_rpc_t), data, rpc.size);
      if(r < 0) {
        reply_with_error(0, EIO, msginfo.msg_id);
        log(ERROR, "Failed to read message data.\n");
        goto cleanup;
      }
    }
  }

  process(&rpc, data, &msginfo);
  return;

  cleanup:
  if(data) free(data);
  return;
}


void *worker(void *thread_arg)
{
  ulong_t work;
  while(1) {
    work = get_work();
    log(INFO, "Got work, %ld\n", work);
    switch(work) {
      case 0:
        break;
      case WORK_LISTEN:
        do_listen();
        break;
      default:
        do_waiting_work(WDEV(work), WDIR(work));
    }
  }
}

void device_server_start()
{
  status_t p;
  ipc_port_msg_info_t msginfo;

  log(INFO, "Starting IO handler...\n");

  put_work(WORK_LISTEN);

  log(INFO, "Listening for CMD requests...\n");

  while(true) {
    p = ipc_port_rcvmsg(server->cmd_port, server->cmdbuf_size + sizeof(uni_rpc_t), server->cmdbuf);
    log(INFO, "CMD received, retval = %d\n", p);
    msginfo = ipc_port_getinfo(server->cmd_port);
    log(INFO, "msginfo: msgid = %hu; msg_len = %hu; sender_pid = %d\n", msginfo.msg_id, msginfo.msg_len, msginfo.sender_pid);
    process_cmd(msginfo);
    log(INFO, "CMD serverd\n");
  }

  /* never reached */

  return;
}


static void process(uni_rpc_t *rpc, void *data, ipc_port_msg_info_t *msginfo)
{
  int devno;
  int dir = (rpc->rpc_call == RPC_READ || rpc->rpc_call == RPC_READDIO)?OUT:IN;
  msg_data_t *msgdata;
  superio_t *superio;

  superio = get_superio(msginfo->sender_pid, rpc->arg1);

  if(!superio) {
    goto pr_cleanup;
  }

  devno = superio->devno;

  if((rpc->rpc_call == RPC_READ && ((superio->flags & O_RDONLY) == 0)) ||
     (rpc->rpc_call == RPC_WRITE && ((superio->flags & O_WRONLY) == 0))) {
    reply_with_error(IN, EBADF, msginfo->msg_id);
    goto pr_cleanup;
  }

  if(devices[devno] == NULL) {
    log(WARNING, "Inexistent device: %d\n", devno);
    reply_with_error(OUT, ENODEV, msginfo->msg_id);
    goto pr_cleanup;
  }

  // IOCTLs are not scheduled, they are served immediately.
  if(rpc->rpc_call == RPC_IOCTL) {
    void *odata = NULL;

    if(rpc->offset && (_RPC_IOCDIR(rpc) == _IOC_READ || _RPC_IOCDIR(rpc) == _IOC_WRARBIT)) {
      odata = malloc(rpc->offset);
      if(!odata) {
        log(ERROR, "Not enough memory for output ioctl buffer.\n");
        reply_with_error(OUT, ENOMEM, msginfo->msg_id);
        goto pr_cleanup;
      }
    } else if(_RPC_IOCDIR(rpc) == _IOC_WRFIXED) {
      odata = data;
    }

    process_ioctl(devno, _RPC_IOCDIR(rpc), _RPC_IOCTYPE(rpc), _RPC_IOCNR(rpc), data, rpc->size, odata, rpc->offset, msginfo);
    if(rpc->offset && (_RPC_IOCDIR(rpc) == _IOC_READ || _RPC_IOCDIR(rpc) == _IOC_WRARBIT)) {
      free(odata);
    }
    goto pr_cleanup;
  } else if(rpc->rpc_call == RPC_FSTAT) {
    reply_with_data(OUT, &devices[devno]->statbuf, sizeof(struct stat), msginfo->msg_id);
    goto pr_cleanup;
  }

  // Check if somebody is already reading | writing on the device.
  spinlock_lock(&devices[devno]->lock[dir]);
  if(!devices[devno]->num_working[dir]) {
    devices[devno]->num_working[dir] = 1;
    // Nobody. Serve request (and some queued ones maybe)
    spinlock_unlock(&devices[devno]->lock[dir]);
    do_messages(devno, rpc, data, msginfo, NULL);
  } else {
    // Somebody is already working, put message to the queue
    msgdata = make_msg(rpc, data, msginfo, NULL);
    if(!msgdata) {
      spinlock_unlock(&devices[devno]->lock[dir]);
      log(ERROR, "Not enough memory to save message.\n");
      goto pr_cleanup;
    }

    list_add2tail(&devices[devno]->ioqueue[dir], &msgdata->node);
    spinlock_unlock(&devices[devno]->lock[dir]);
  }

  return;

  pr_cleanup:
  if(rpc->size)free(data);
  return;
}

// Function for processing message current message and other enqueued ones if any.
// This function can be called from two places: from process() function when it receives a message
// and from do_waiting_work(). in latter case msg is not null otherwise in case of SUSPEND result
// this function needs to make new msg_data_t structure.
static void do_messages(int devno, uni_rpc_t *rpc, void *data, ipc_port_msg_info_t *msginfo, msg_data_t *msg)
{
  status_t ret;
  void *loc_interim;
  void *p_interim;

  // One message is passed through parameters, others are taken from the queue...
  while(1) {
    if(msg) {
      p_interim = &msg->interim;
      rpc = &msg->rpc;
      data = msg->data;
      msginfo = &msg->msginfo;
    } else {
      loc_interim = NULL;
      p_interim = &loc_interim;
    }

    devices[devno]->pending_wake[DIR(rpc)] = 0;
    ret = do_process(devno, rpc, data, msginfo, p_interim);
    spinlock_lock(&devices[devno]->lock[DIR(rpc)]);

    if(ret == SUSPEND) {
      log(INFO, "Suspended, dir: %d; devno: %d\n", DIR(rpc), devno);
      // Request is not completely served.
      if(devices[devno]->pending_wake[DIR(rpc)]) {
        // Somebody called wake_device while we were processing, we can try again.
        spinlock_unlock(&devices[devno]->lock[DIR(rpc)]);
        continue;
      } else {
        // Next time, baby...
        if(!msg) {
          msg = make_msg(rpc, data, msginfo, loc_interim);
        }
        if(!msg) {
          // No mem, drop message and try another if any.
          log(ERROR, "Not enough memory to create message.");
          free(data);
          goto try_queue;
        } else {
          devices[devno]->waiting[DIR(rpc)] = msg;
        }
        spinlock_unlock(&devices[devno]->lock[DIR(rpc)]);
        return;
      }
    } else {
      free(data);
      if(msg) free(msg);
      // Request was served, let's look at the queue if some another request came
      // while we were working and now can be served...
      try_queue:

      if(list_is_empty(&devices[devno]->ioqueue[DIR(rpc)])) {
        // Nothing in the queue, unmark busy flag and leave.
        devices[devno]->num_working[DIR(rpc)] = 0;
        spinlock_unlock(&devices[devno]->lock[DIR(rpc)]);
        return;
      } else {
        // There is some work...
        msg = list_first_entry(&devices[devno]->ioqueue[DIR(rpc)], msg_data_t, node);
        list_del(&msg->node);
        spinlock_unlock(&devices[devno]->lock[DIR(rpc)]);
        continue;
      }
    }
  }
}

static inline msg_data_t *make_msg(uni_rpc_t *rpc, void *data, ipc_port_msg_info_t *msginfo, void *interim)
{
  msg_data_t *msgdata;

  msgdata = malloc(sizeof(*msgdata));
  if(!msgdata) {
    return NULL;
  }

  memcpy(&msgdata->rpc, rpc, sizeof(*rpc));
  memcpy(&msgdata->msginfo, msginfo, sizeof(*msginfo));
  msgdata->data = data;
  msgdata->interim = interim;

  return msgdata;
}

static inline status_t do_process(int devno, uni_rpc_t *rpc, void *data, ipc_port_msg_info_t *msginfo, void **interim)
{
  if(rpc->rpc_call == RPC_WRITE) {
    return process_write(devno, rpc, data, msginfo, interim);
  } else {
    return process_read(devno, rpc, data, msginfo, interim);
  }
}

static void process_cmd(ipc_port_msg_info_t msginfo)
{
  uni_rpc_t *rpc;
  int ret;

  /* check */
  if(msginfo.msg_len < sizeof(uni_rpc_t)) {
    reply_with_error(CMD, EINVAL, msginfo.msg_id);
    return;
  }  else {
    rpc = (uni_rpc_t *)server->cmdbuf;
  }

  if(rpc->rpc_call == RPC_ALIVE || rpc->rpc_call == RPC_STATUS) {
    reply_with_error(CMD, 0, msginfo.msg_id);
    return;
  }

  if(msginfo.sender_pid != server->hw_pid) {
    reply_with_error(CMD, EPERM, msginfo.msg_id);
    return;
  }

  switch(rpc->rpc_call) {
    case RPC_REGISTER:
      ret = add_file(rpc->arg0, rpc->arg1, rpc->arg2, rpc->arg3, rpc->arg4);
      break;
    case RPC_UNREGISTER:
      ret = rm_file(rpc->arg0, rpc->arg1);
      break;
    case RPC_PIDFDKILL:
      ret = kill_client(rpc->arg0);
      break;
    case RPC_PIDFDCLONE:
      ret = clone_client(rpc->arg0, rpc->arg1);
      break;
    default:
      ret = EINVAL;
      break;
  }

  reply_with_error(CMD, ret, msginfo.msg_id);
  return;
}

static status_t process_write(int devno, uni_rpc_t *rpc, void *data, ipc_port_msg_info_t *msginfo, void **interim)
{
  status_t status = 0;
  status_t len;

  log(INFO, "Processing write...%d\n", rpc->rpc_call);

  len=server->ops_tbl->write(devices[devno]->devno, data, rpc->size, rpc->offset, interim);
  if(len == SUSPEND) {
    status = SUSPEND;
  } else if (len<0) {
    reply_with_error(IN, -len, msginfo->msg_id);
  } else {
    reply_with_replcode(IN, len, msginfo->msg_id);
  }

  return status;
}

static status_t process_read(int devno, uni_rpc_t *rpc, void *data, ipc_port_msg_info_t *msginfo, void **interim)
{
  status_t status = 0;
  status_t ret;

  log(INFO, "Processing read...\n");

  ret = server->ops_tbl->read(devices[devno]->devno, data, rpc->size, rpc->offset, interim);
  if(ret == SUSPEND) {
    status = SUSPEND;
  } else if(ret < 0) {
    reply_with_error(OUT, -ret, msginfo->msg_id);
    } else {
    reply_with_data(OUT, data, ret, msginfo->msg_id);
  }

  return status;
}

static status_t process_ioctl(int devno, int dir, int type, int nr, void *data, size_t size,
                              void *odata, size_t osize, ipc_port_msg_info_t *msginfo)
{
  status_t status = 0;
  status_t ret;

  log(INFO, "Processing ioctl...\n");

  ret = server->ops_tbl->ioctl(devices[devno]->devno, type, nr, data, size, odata, osize);
  if(ret < 0) {
    reply_with_error(OUT, -ret, msginfo->msg_id);
  } else {
    if(dir == _IOC_WRARBIT) {
      reply_with_data(OUT, odata, ret, msginfo->msg_id);
    } else if(dir == _IOC_WRITE) {
      reply_with_replcode(OUT, ret, msginfo->msg_id);
    }else {
      reply_with_data(OUT, odata, osize, msginfo->msg_id);
    }
  }

  return status;
}

pid_t get_sender_pid(void)
{
  return ((ipc_port_msg_info_t *)PTD()->custom_slots[MSGINFO_SLOT])->sender_pid;
}

tid_t get_sender_tid(void)
{
  return ((ipc_port_msg_info_t *)PTD()->custom_slots[MSGINFO_SLOT])->sender_tid;
}

uid_t get_sender_uid(void)
{
  return ((ipc_port_msg_info_t *)PTD()->custom_slots[MSGINFO_SLOT])->sender_uid;
}

uid_t get_sender_gid(void)
{
  return ((ipc_port_msg_info_t *)PTD()->custom_slots[MSGINFO_SLOT])->sender_gid;
}

/* local functions */

static void reply_with_error(int id, int err, int msg_id)
{
  int r;
  uni_rpc_t rpc;

  memset(&rpc, 0, sizeof(rpc));
  rpc.rpc_call = err;

  r=ipc_port_reply_nmsg((id == CMD)?server->cmd_port:server->io_port, sizeof(rpc), &rpc, msg_id);

  log(INFO, "Replied with error %d\n", err);

  return;
}


static void reply_with_data(int id, void *buf, ulong_t len, int msg_id)
{
  int r;
  uni_rpc_t rpc;
  iovec_t iov[2];

  memset(&rpc, 0, sizeof(rpc));

  rpc.rpc_call=0;
  rpc.size=len;

  iov[0].iov_base = &rpc;
  iov[0].iov_len = sizeof(rpc);
  iov[1].iov_base = buf;
  iov[1].iov_len = len;

  r=ipc_port_reply_nmsg_iov((id == CMD)?server->cmd_port:server->io_port, iov, buf?2:1, msg_id);

  log(INFO, "Replied with data of len %d\n", len);

  return;
}

static void reply_with_replcode(int id, ulong_t len, int msg_id)
{
  int r;
  uni_rpc_t rpc;

  memset(&rpc, 0, sizeof(rpc));

  rpc.rpc_call=0;
  rpc.size=len;
  r=ipc_port_reply_nmsg((id == CMD)?server->cmd_port:server->io_port, sizeof(rpc), &rpc, msg_id);

  log(INFO, "Replied with code %d\n", len);

  return;
}

// This function registers new opened files when client calls open()
static int add_file(pid_t pid, uint16_t unique, uint16_t devno, uint16_t flags, uint16_t mode)
{
  //TODO: remove mallocs from critical section
  int ret = 0;
  usrtc_node_t *node;
  client_t *client;
  superio_t *file;

  spinlock_lock(&clients_lock);

  node = usrtc_lookup(&clients, &pid);
  if(!node) {
    // New client, create new record
    client = malloc(sizeof(*client));
    if(!client) {
      log(ERROR, "No memory for client recor.\n");
      ret = ENOMEM;
      goto af_cleanup;
    }
    node = malloc(sizeof(*node));
    if(!node) {
      log(ERROR, "No memory for client node.\n");
      ret = ENOMEM;
      free(client);
      goto af_cleanup;
    }

    usrtc_node_init(node, client);
    client->pid = pid;
    usrtc_init(&client->files, USRTC_SPLAY, USRTC_COUNT_T_MAX, short_compare);
    usrtc_insert(&clients, node, client);
  } else {
    client = usrtc_node_getdata(node);
  }

  node = usrtc_lookup(&client->files, &unique);
  if(node) {
    ret = EBUSY;
    goto af_cleanup;
  }

  // TODO: maybe it is better to kill empty fresh created client record
  // in case when failed to create file record;
  file = malloc(sizeof(*file));
  if(!file) {
    log(ERROR, "No memory for file record.\n");
    ret = ENOMEM;
    goto af_cleanup;
  }

  node = malloc(sizeof(*node));
  if(!node) {
    log(ERROR, "No memory for file node.\n");
    ret = ENOMEM;
    free(file);
    goto af_cleanup;
  }

  usrtc_node_init(node, file);
  file->unique = unique;
  file->flags = flags;
  file->devno = devno;
  file->mode = mode;

  usrtc_insert(&client->files, node, file);

  af_cleanup:
  spinlock_unlock(&clients_lock);

  return ret;
}

// This function unregisters files (when client calls close or exec)
static int rm_file(pid_t pid, uint16_t unique)
{
  usrtc_node_t *node, *fnode;
  superio_t *superio;
  client_t *client;
  int ret;

  spinlock_lock(&clients_lock);

  node = usrtc_lookup(&clients, &pid);
  if(!node) {
    log(WARNING, "Trying to close file for inexistent client.\n");
    ret = ESRCH;
    goto rf_cleanup;
  }

  client = usrtc_node_getdata(node);

  fnode = usrtc_lookup(&client->files, &unique);
  if(!node) {
    log(WARNING, "Trying to close inexistent file.\n");
    ret = EBADF;
    goto rf_cleanup;
  }

  superio = usrtc_node_getdata(fnode);
  free(superio);

  usrtc_delete(&client->files, fnode);
  free(fnode);

  // check if it was the last file
  if(usrtc_count(&client->files) == 0) {
    usrtc_delete(&clients, node);
    free(client);
    free(node);
  }

  rf_cleanup:
  spinlock_unlock(&clients_lock);
  return ret;
}

// This function duplicates client when it calls fork()
static int clone_client(pid_t src, pid_t dest)
{
  usrtc_node_t *node, *dnode, *fnode;
  client_t *sclient, *dclient;
  superio_t *ssuperio, *dsuperio;
  int ret = 0;

  spinlock_lock(&clients_lock);

  node = usrtc_lookup(&clients, &src);
  if(!node) {
    log(WARNING, "Requested to clone inexistent client.\n");
    ret = ESRCH;
    goto cc_cleanup;
  }

  dnode = usrtc_lookup(&clients, &dest);
  if(dnode) {
    log(WARNING, "Invalid clone request, tried to overwrite existing client.\n");
    ret = EBUSY;
    goto cc_cleanup;
  }

  sclient = usrtc_node_getdata(node);
  if(usrtc_count(&sclient->files) == 0) {
    // No need to copy anything, will create record this on demand.
    goto cc_cleanup;
  }

  // All is OK, copy client;
  dnode = malloc(sizeof(*dnode));
  if(!dnode) {
    log(ERROR, "Not enough memory for client node.\n");
    ret = ENOMEM;
    goto cc_cleanup;
  }

  dclient = malloc(sizeof(*dclient));
  if(!dclient) {
    log(ERROR, "Not enough memory for client.\n");
    ret = ENOMEM;
    free(dnode);
    goto cc_cleanup;
  }

  usrtc_init(&dclient->files, USRTC_SPLAY, USRTC_COUNT_T_MAX, short_compare);

  for(fnode = usrtc_first(&sclient->files); fnode != NULL;
      fnode = usrtc_next(&sclient->files, fnode)) {

    node = malloc(sizeof(*node));
    if(!node) {
      log(ERROR, "No memory while copying fds, cannot allocate node.\n");
      ret = ENOMEM;
      goto cc_destroy;
    }

    dsuperio = malloc(sizeof(*dsuperio));
    if(!dsuperio) {
      log(ERROR, "No memory while copying fds, cannot allocate superio.\n");
      ret = ENOMEM;
      free(node);
      goto cc_destroy;
    }

    usrtc_node_init(node, dsuperio);
    ssuperio = usrtc_node_getdata(fnode);
    memcpy(dsuperio, ssuperio, sizeof(*dsuperio));

    usrtc_insert(&dclient->files, node, dsuperio);
  }

  dclient->pid = dest;
  usrtc_node_init(dnode, dclient);

  usrtc_insert(&clients, dnode, dclient);
  goto cc_cleanup;

  // Destroy half-ready record and free data if memory
  // exausted in the middle of copying process
  cc_destroy:

  while((fnode = usrtc_first(&dclient->files))) {
    dsuperio = usrtc_node_getdata(fnode);
    usrtc_delete(&dclient->files, fnode);
    free(dsuperio);
    free(fnode);
  }

  free(dclient);
  free(dnode);

  cc_cleanup:
  spinlock_unlock(&clients_lock);

  return ret;
}

// This function removes client record when it dies.
static int kill_client(pid_t pid)
{
  usrtc_node_t *node, *fnode;
  client_t *client;
  superio_t *superio;
  int ret = 0;

  spinlock_lock(&clients_lock);

  node = usrtc_lookup(&clients, &pid);
  if(!node) {
    log(WARNING, "Kill request for inexistent client.\n");
    ret = ESRCH;
    goto kc_cleanup;
  }

  client = usrtc_node_getdata(node);

  while((fnode = usrtc_first(&client->files))) {
    superio = usrtc_node_getdata(fnode);
    usrtc_delete(&client->files, fnode);
    free(superio);
    free(fnode);
  }

  usrtc_delete(&clients, node);
  free(client);
  free(node);

  kc_cleanup:
  spinlock_unlock(&clients_lock);
  return ret;
}

// This function returns access flags for file.
static superio_t *get_superio(pid_t pid, uint16_t unique)
{
  usrtc_node_t *node;
  client_t *client;
  superio_t *superio = NULL;

  spinlock_lock(&clients_lock);

  node = usrtc_lookup(&clients, &pid);
  if(!node) {
    log(WARNING, "Darklord detected, PID is %hu\n", pid);
    goto ca_cleanup;
  }

  client = usrtc_node_getdata(node);

  node = usrtc_lookup(&client->files, &unique);
  if(!node) {
    log(WARNING, "PID %hu tried to access file in Darklord way.\n", pid);
    goto ca_cleanup;
  }

  superio = usrtc_node_getdata(node);

  ca_cleanup:
  spinlock_unlock(&clients_lock);
  return superio;
}
