/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org> 
 *
 * file system resources
 * 
 */

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <wchar.h>
#include <datalib/usrtc.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ns/ns.h>
#include <pthread.h>

#include <fsio/vfs_call.h>
#include <fsio/vnode.h>
#include <fsio/fs_resources.h>

#define __lock  pthread_mutex_lock(&res_lock)
#define __unlock  pthread_mutex_unlock(&res_lock)

static usrtc_t *resources=NULL;
static pthread_mutex_t res_lock;

static int __cmp_fs_res(const void *a,const void *b);
static int __rm_file_des(file_des_t *s);

int init_fs_resources(void)
{
  resources=malloc(sizeof(*resources));
  if(!resources)
    return -ENOMEM;
  usrtc_init(resources,USRTC_SPLAY,USRTC_COUNT_T_MAX,__cmp_fs_res);

  pthread_mutex_init(&res_lock,NULL);

  return 0;
}

int add_fs_source(fs_source_t *s,pid_t pid,uid_t uid,gid_t gid)
{
  usrtc_node_t *node;

  if(!s)
    return EINVAL;

  /* check if exists */
  __lock;
  node=usrtc_lookup(resources,&pid);
  __unlock;
  if(node!=NULL)
    return EEXIST;
  /* init node structure */
  node=malloc(sizeof(*node));
  if(!node)
    return ENOMEM;
  usrtc_node_init(node,s);
  /* init resource */
  s->res=malloc(sizeof(*s->res));
  if(!s->res) {
    free(node);
    return ENOMEM;
  }
  usrtc_init(s->res,USRTC_SPLAY,USRTC_COUNT_T_MAX,__cmp_fs_res);
  s->pid=pid;
  s->uid=uid;
  s->gid=gid;
  s->max_fds=DEF_MAXFDS;
  s->n_fds=0;
  pthread_mutex_init(&s->lock,NULL);
  /* add */
  __lock;
  usrtc_insert(resources,node,&s->pid);
  __unlock;

  return 0;
}

int remove_fs_source(pid_t pid)
{
  usrtc_node_t *node;
  fs_source_t *s;
  file_des_t *des;

  __lock;
  node=usrtc_lookup(resources,&pid);

  if(!node) {
    __unlock;
    return EINVAL;
  } else 
    s=(fs_source_t *)usrtc_node_getdata(node);
  usrtc_delete(resources,node);
  __unlock;
  free(node);

  pthread_mutex_lock(&s->lock);
  while((node=usrtc_first(s->res))!=NULL) {
    des=(file_des_t *)usrtc_node_getdata(node);
    usrtc_delete(s->res,node);
    free(node);
    //    rm_file_des(des);
    free(des);
  }
  pthread_mutex_unlock(&s->lock);
  free(s->res);
  free(s);

  return 0;
}

int add_file_des(pid_t pid,uid_t uid,gid_t gid,int fd,vnode_t *vnode)
{
  usrtc_node_t *node=NULL;
  fs_source_t *r;
  file_des_t *dfd;
  int t;

  __lock;
  node=usrtc_lookup(resources,&pid);
  __unlock;
  if(!node) {
    r=malloc(sizeof(*r));
    if(!r)
      return ENOMEM;
    t=add_fs_source(r,pid,uid,gid);
    if(t!=0)
      return t;
  } else {
    r=(fs_source_t *)usrtc_node_getdata(node);
    pthread_mutex_lock(&r->lock);
    node=usrtc_lookup(r->res,&fd);
    if(node) {
      pthread_mutex_unlock(&r->lock);
      return EEXIST;
    }
    goto __addfd;
  }
  pthread_mutex_lock(&r->lock);

 __addfd:
  dfd=malloc(sizeof(*dfd));
  if(!dfd) {
    t=ENOMEM;
    goto __exit;
  }
  node=malloc(sizeof(*node));
  if(!node) {
    t=ENOMEM;
    goto __exit;
  }
  usrtc_node_init(node,dfd);
  dfd->fd=fd;
  dfd->rw_off=dfd->wr_off=0;
  dfd->owner=r;
  dfd->vnode=vnode;
  usrtc_insert(r->res,node,&dfd->fd);

 __exit:
  pthread_mutex_unlock(&r->lock);

  return t;
}

int rm_file_des(pid_t pid,int fd)
{
  file_des_t *dd;

  return __rm_file_des(dd);
}

/* local functions */

static int __rm_file_des(file_des_t *s)
{
  int r;

  r=vnode_rmref(s->vnode);
  if(r!=0)
    return r;

  return 0;
}

static int __cmp_fs_res(const void *a,const void *b)
{
  return *(int*)a - *(int*)b;
}
