/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org> 
 *
 * virtual node operations
 * 
 */

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <wchar.h>
#include <datalib/usrtc.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ns/ns.h>
#include <pthread.h>
#include <fcntl.h>

#include <fsio/vfs_call.h>
#include <fsio/vnode.h>
#include <fsio/fs_resources.h>

vnode_t *root;
usrtc_t *vnodes;
pthread_mutex_t vnodes_lock;

#define __VN_LOCK    pthread_mutex_lock(&vnodes_lock)
#define __VN_UNLOCK  pthread_mutex_unlock(&vnodes_lock)

#define vnode_lock(a)    pthread_mutex_lock(&a->vn_lock)
#define vnode_unlock(a)  pthread_mutex_unlock(&a->vn_lock)

/* local functions prototypes */
static vnode_t *__vnode_srch(char *path,int *pp,int amount);
static vnode_t *__vnode_clean_add(void);
static void __vnode_tmp_set(vnode_t *temp);
static vnode_t *__vnode_tmp_get(void);
static void __vnode_tmp_clear(void);
static int __cmp_vnodes(const void *a,const void *b);

/* vnode structure operations */
int vnode_root_init(struct vnode_ops *ops,void *spec_obj)
{
  usrtc_node_t *node;

  if(!ops)
    return EINVAL;

  root=malloc(sizeof(*root));
  if(!root)
    return ENOMEM;

  root->ops=ops;
  root->id=0;
  root->hash=0;
  root->ref_count=1; /* root node cannot be removed, take reference always more than zero*/
  root->spec_obj=spec_obj;
  root->type=FT_DIRECTORY;

  pthread_mutex_init(&root->vn_lock,NULL);

  /* init global vnodes structure */
  vnodes=malloc(sizeof(*vnodes));
  if(!vnodes) {
    pthread_mutex_destroy(&root->vn_lock);
    free(root);
    return ENOMEM;
  }

  pthread_mutex_init(&vnodes_lock,NULL);

  usrtc_init(vnodes,USRTC_SPLAY,USRTC_COUNT_T_MAX,__cmp_vnodes);

  node=malloc(sizeof(*node));
  if(!node)
    return ENOMEM;
  
  usrtc_node_init(node,root);
  root->pnode=node;
  root->hash=vnode_hash("/");
  usrtc_insert(vnodes,node,&root->hash);

  return 0;
}

int vnode_mount(void)
{
  return root->ops->vn_mount(root);
}

int vnode_init(vnode_t *svn,void *spec_obj)
{
  if(!svn)
    return EINVAL;

  svn->ops=root->ops;
  svn->id=(ulong_t)svn;
  svn->hash=0;
  svn->ref_count=0;
  if(spec_obj)
    svn->spec_obj=spec_obj;
  else
    svn->spec_obj=root->spec_obj;

  svn->pnode=malloc(sizeof(*svn->pnode));
  if(!svn->pnode)
    return ENOMEM;
  usrtc_node_init(svn->pnode,svn);

  pthread_mutex_init(&svn->vn_lock,NULL);

  return 0;
}

int vnode_addref(vnode_t *svn)
{
  if(!svn)
    return EINVAL;

  pthread_mutex_lock(&svn->vn_lock);
  svn->ref_count++;
  pthread_mutex_unlock(&svn->vn_lock);

  return 0;
}

int vnode_rmref(vnode_t *svn)
{
  if(!svn)
    return EINVAL;

  pthread_mutex_lock(&svn->vn_lock);
  if(svn->ref_count==1) { /* yep, we're should do it */
    if(svn->ops->vn_unref!=NULL)
      svn->ops->vn_unref(svn);
    __VN_LOCK;
    pthread_mutex_unlock(&svn->vn_lock);
    pthread_mutex_destroy(&svn->vn_lock);
    usrtc_delete(vnodes,svn->pnode);
    __VN_UNLOCK;
    free(svn->pnode);
    free(svn);
    return 0;
  } else 
    svn->ref_count--;
  pthread_mutex_unlock(&svn->vn_lock);

  return 0;
}

int vnode_insert(vnode_t *svn,hash_t hash,usrtc_node_t *node)
{
  if(!svn || !node)
    return EINVAL;

  usrtc_node_init(node,svn);
  svn->hash=hash;
  __VN_LOCK;
  usrtc_insert(vnodes,node,&svn->hash);
  __VN_UNLOCK;

  return 0;
}

hash_t vnode_hash(const char *name)
{
  hash_t hash=0;

  if(!name)
    return 0;

  while(*name) 
    hash=(hash << 5) + hash + *name++;
  
  return hash;
}

/* vnode generic handlers */
vnode_t *vnode_open(char *path,int *p,int total,int root,
		    int flags,int mode,uid_t uid,gid_t gid)
{
  vnode_t *temp;
  vnode_t *look;
  char *r_path;
  int *o=p;
  int r;

  /* TODO: if filesystem read-only, but flags wronly or rdwr - error! */

  temp=__vnode_tmp_get();
  if(!temp) {
    temp=__vnode_clean_add();
    if(!temp) {
      errno=ENOMEM;
      return NULL;
    }
    __vnode_tmp_set(temp);
    vnode_lock(temp);
  }

  /* now let's lookup for this */
  if(!temp->ops->vn_lookup) /* oops */ {
    vnode_unlock(temp);
    errno=ENOSYS;
    return NULL; /* FIXME: dia or error ? */
  }
  /* first look up on existing vnodes */
  r_path=path+p[root];
  o+=root;
  look=__vnode_srch(r_path,o,total-root);
  if(look!=NULL) {
    vnode_unlock(temp);
    if((flags | O_CREAT) && (flags | O_EXCL)) {
      errno=EEXIST;
      return NULL;
    }
    return look;
  }

  /*ok, now we're can look up*/
  r=temp->ops->vn_lookup(temp,r_path,o,total-root,flags,mode,gid,uid);
  if(r!=0) {
    if((flags | O_CREAT) && (flags | O_EXCL)) {
      if(flags | O_DIRECTORY) {
	r=temp->ops->vn_mkdir(temp,r_path,o,total-root,mode,uid,gid);
      } else {
	r=temp->ops->vn_create(temp,r_path,o,total-root,mode,uid,gid);
      }
      goto __ret_temp;
    }
    vnode_unlock(temp);
    errno=r;
    return NULL;
  } else {
    if(temp->ops->vn_open) {
      r=temp->ops->vn_open(temp,flags,mode);
      if(r!=0) {
	vnode_unlock(temp);
	errno=r;
	return NULL;
      }
    }
  __ret_temp:
    __vnode_tmp_clear();
    /*
     * TODO: if O_TRUNC and can be written 
     * and opened for rdwr - truncate to 0.
     */
    return temp;
  }

  errno=EINVAL;

  return NULL;
}

int vnode_close(vnode_t *vnode)
{
  return 0;
}

int vnode_read(vnode_t *vnode,void *buf,size_t *size,off_t offset)
{
  int r;

  if(!vnode->ops->vn_rdwr) { /* oops */
    return ENOSYS;
  }

  r=vnode->ops->vn_rdwr(vnode,buf,size,offset,RD_DIRECTION);
  if(r!=0)
    return r;

  return 0;
}

int vnode_write(vnode_t *vnode,void *buf,size_t *size,off_t offset)
{
  int r;

  if(!vnode->ops->vn_rdwr) { /* oops */
    return ENOSYS;
  }

  r=vnode->ops->vn_rdwr(vnode,buf,size,offset,WR_DIRECTION);
  if(r!=0)
    return r;

  return 0;
}

int vnode_mkdir(char *path,int *pp,int total,int root,int mode,uid_t uid,gid_t gid)
{
  vnode_t *temp;
  char *__path=path;
  int *_pp=pp;
  int r;

  __path+=pp[root];
  _pp+=root;
  temp=__vnode_srch(__path,_pp,total-root);
  if(temp)
    return EEXIST;

  temp=__vnode_tmp_get();
  if(!temp) {
    temp=__vnode_clean_add();
    if(!temp) 
      return ENOMEM;
    __vnode_tmp_set(temp);
    vnode_lock(temp);
  } else
    vnode_lock(temp);
  if(!temp->ops->vn_mkdir) {
    vnode_unlock(temp);
    return ENOSYS;
  } else 
    r=temp->ops->vn_mkdir(temp,__path,_pp,total-root,mode,gid,uid);
  
  vnode_unlock(temp);

  return r;
}

/* local functions */

/* vnode temproary setted up vnode */
static vnode_t *__vnode_temp=NULL;

static vnode_t *__vnode_srch(char *path,int *pp,int amount)
{
  vnode_t *child;
  usrtc_node_t *node;
  hash_t hash=vnode_hash(path);

  __VN_LOCK;
  node=usrtc_lookup(vnodes,&hash);
  __VN_UNLOCK;

  if(!node)
    return NULL;

  child=usrtc_node_getdata(node);

  return child;
}

static vnode_t *__vnode_clean_add(void)
{
  vnode_t *node;

  node=malloc(sizeof(*node));
  if(!node)
    return NULL;
  node->ops=root->ops;
  /* init mutex */
  pthread_mutex_init(&node->vn_lock,NULL);

  return node;
}

static void __vnode_tmp_set(vnode_t *temp)
{
  __vnode_temp=temp;

  return;
}

static vnode_t *__vnode_tmp_get(void)
{
  if(__vnode_temp)
    vnode_lock(__vnode_temp);

  return __vnode_temp;
}

static void __vnode_tmp_clear(void)
{
  __vnode_temp=NULL;

  return;
}

static int __cmp_vnodes(const void *a,const void *b)
{
  return (int)(*(hash_t*)a-*(hash_t*)b);
}
