/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * Implementation of timeouts for services
 */

#include <sys/swks.h>
#include <timeout.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <debug.h>

static swks_t *swks;

static int __set_timeout(struct timer_desc *tmr, uint32_t msec)
{
  struct itimerspec s;

  s.it_interval.tv_sec = 0;
  s.it_interval.tv_nsec = 0;
  s.it_value.tv_sec = msec / 1000;
  s.it_value.tv_nsec = (msec % 1000) * 1000000;

  return timer_settime(tmr->id, 0, &s, NULL);
}

int init_timeout_tmr(struct timer_desc *tmr, pthread_t thread, int signo,
                     void *sival_ptr, bool block)
{
  int r;
  struct sigevent ev;
  sigset_t sig;

  tmr->id = -1;
  swks = get_swks();
  if (block) {
    sigemptyset(&sig);
    sigaddset(&sig, signo);
    sigprocmask(SIG_BLOCK, &sig, NULL);
    tmr->blocked_signo = signo;
  }
  ev.sigev_notify = SIGEV_SIGNAL_THREAD;
  ev.sigev_signo = signo;
  ev.sigev_value.sival_ptr = sival_ptr;
  ev.tid = thread.__tid;
  tmr->exp_ticks = 0;
  r = timer_create(CLOCK_REALTIME, &ev, &tmr->id);
  if (r && block)
    sigprocmask(SIG_UNBLOCK, &sig, NULL);

  return r;
}

void destroy_timeout_tmr(struct timer_desc *tmr)
{
#if 0
  sigset_t sig;

  if (tmr->id == (unsigned)-1)
    return;

  /*
   * TO DO [dg]: uncomment when the appropriate posix
   * function will be implemented
   */
  if (tmr->blocked_signo) {
    sigemptyset(&sig);
    sigaddset(&sig, signo);
    sigprocmask(SIG_UNBLOCK, &sig, NULL);
  }
  timer_delete(tmr->id);
#endif
}

int set_timeout(struct timer_desc *tmr, uint16_t msec)
{
  ASSERT(swks != NULL, "System-wide kernel statistics is not mapped!\n");
  tmr->exp_ticks = current_sys_ticks(swks) + msec_to_ticks(swks, msec);

  return __set_timeout(tmr, msec);
}

int set_timeout_ticks(struct timer_desc *tmr, uint16_t ticks)
{
  uint32_t ms;

  ASSERT(swks != NULL, "SYstem-wide kernel statisticks is not mapped!\n");
  ms = ticks_to_msec(swks, ticks);
  tmr->exp_ticks = current_sys_ticks(swks) + ticks;

  return __set_timeout(tmr, ms);
}

int reset_timeout(struct timer_desc *tmr)
{
  struct itimerspec s;

  memset(&s, 0, sizeof(struct itimerspec));
  tmr->exp_ticks = 0;

  return timer_settime(tmr->id, 0, &s, NULL);
}

uint32_t timeout_expiration(struct timer_desc *tmr)
{
  uint32_t ms;
  uint64_t ticks;

  ASSERT(swks != NULL, "System-wide kernel statistics is not mapped!\n");
  ticks = current_sys_ticks(swks);
  if (ticks >= tmr->exp_ticks) {
    ms = MSEC_INFINITE;
  } else {
    ms = (tmr->exp_ticks - ticks) * 1000 / swks->hz;
  }

  return ms;
}

