/**
 * @file
 * Dynamic pool memory manager
 *
 * lwIP has dedicated pools for many structures (netconn, protocol control blocks,
 * packet buffers, ...). All these pools are managed here.
 */

/*
 * Copyright (c) 2001-2004 Swedish Institute of Computer Science.
 * All rights reserved. 
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 * 
 * Author: Adam Dunkels <adam@sics.se>
 *
 */

#include "lwip/opt.h"

#include "lwip/memp.h"
#include "lwip/pbuf.h"
#include "lwip/udp.h"
#include "lwip/raw.h"
#include "lwip/tcp.h"
#include "lwip/igmp.h"
#include "lwip/api.h"
#include "lwip/api_msg.h"
#include "lwip/tcpip.h"
#include "lwip/sys.h"
#include "lwip/stats.h"
#include "netif/etharp.h"
#include "lwip/ip_frag.h"
#include "arch/cc.h"

#include <string.h>
#include <stdlib.h>
#include <sys/dma.h>
#include <sys/mman.h>
#include <errno.h>

#define DMA_INIT_PAGES 128
static int __dma_left=0;
static void *__dma_phys_start=MAP_FAILED;
static void *__dma_map_base=MAP_FAILED;

const u16_t memp_sizes[MEMP_MAX] = {
#define LWIP_MEMPOOL(name,num,size,desc)  MEMP_ALIGN_SIZE(size),
#include "lwip/memp_std.h"
};

void memp_init(void)
{
  __dma_phys_start=dma_alloc_pages(DMA_INIT_PAGES);
  if( __dma_phys_start != MAP_FAILED ) {
    __dma_map_base=mmap(NULL,DMA_INIT_PAGES*PAGE_SIZE,PROT_READ | PROT_WRITE,
                        MAP_PHYS | MAP_PRIVATE, NOFD, __dma_phys_start);
    if( __dma_map_base != MAP_FAILED ) {
      __dma_left=DMA_INIT_PAGES*PAGE_SIZE;
    }
  }
}

void *
#if !MEMP_OVERFLOW_CHECK
memp_malloc(memp_t type)
#else
memp_malloc_fn(memp_t type, const char* file, const int line)
#endif
{
//  console_printf("[%d:%d] mempool: Allocating %d bytes of type '%d'\n",
//                 getpid(),gettid(),memp_sizes[type],type);
  return malloc(memp_sizes[type]);
}

void memp_free(memp_t type, void *mem)
{
  free(mem);
}

void *mem_malloc_dma(size_t size, void **phys_addr)
{
  void *_ra, *_pa;

  if( __dma_left >= size ) {
    _ra=__dma_map_base;
    _pa=__dma_phys_start;

    __dma_phys_start += size;
    __dma_map_base += size;
    __dma_left-=size;

    *phys_addr=_pa;
    return _ra;
  }
  return NULL;
}
