/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * generic file system related calls
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/iov.h>
#include <ipc/rpc.h>
#include <pthread.h>
#include <sys/v2.h>
#include <ns/ns.h>
#include <errno.h>
#include <unistd.h>
#include <datalib/usrtc.h>

#include <vfs/vfsdef.h>
#include <vfs/vnode.h>
#include <vfs/super.h>
#include <vfs/session.h>

#include "vfsint.h"

int v2_stat(super_block_t *super,char *path,int *pps,int pp,int plen,
	    int offset,struct stat *buf)
{
  int r,nc=0;
  int i;
  off_t rear_off=0;
  uint64_t shash;
  vnode_t *vnode;

  /* prepare names */
  if(offset>0) {
    rear_off=pps[offset];
    path=(char*)(path+pps[offset]);
    plen-=pps[offset];
    pps+=offset;
    pp-=offset;
    for(i=0;i<pp;i++)
      pps[i]-=rear_off;
  }
  shash=vnode_hash_ch(path);
  if((vnode=vnode_lookup(shash))==NULL) {
    vnode=vnode_create(super,path,pp,plen,pps);
    if(!vnode) {
      errno=ENOMEM;
      return -1;
    }
  } else {
    nc++;
    vnode_lock(vnode);
  }
  /* it's ok, now let's lookup */
  r=super->fs_tbl->lookup(vnode,path,pps,pp);
  if(r!=0) {
    if(!nc) vnode_destroy(vnode);
    else vnode_unlock(vnode);
    errno=r;
    return -1;
  }

  memcpy(buf,&vnode->vstat,sizeof(*buf));

  if(!nc) vnode_destroy(vnode);
  else vnode_unlock(vnode);

  return 0;
}

int v2_readlink(super_block_t *super,char *path,int *pps,int pp,int plen,
                int offset)
{
  int r,nc=0;
  int i;
  off_t rear_off=0;
  uint64_t shash;
  size_t link_size;
  vnode_t *vnode;

  /* prepare names */
  if(offset>0) {
    rear_off=pps[offset];
    path=(char*)(path+pps[offset]);
    plen-=pps[offset];
    pps+=offset;
    pp-=offset;
    for(i=0;i<pp;i++)
      pps[i]-=rear_off;
  }
  shash=vnode_hash_ch(path);
  if((vnode=vnode_lookup(shash))==NULL) {
    vnode=vnode_create(super,path,pp,plen,pps);
    if(!vnode) {
      errno=ENOMEM;
      return -1;
    }
  } else {
    nc++;
    vnode_lock(vnode);
  }
  /* it's ok, now let's lookup */
  r=super->fs_tbl->lookup(vnode,path,pps,pp);
  if(r!=0) {
    if(!nc) vnode_destroy(vnode);
    else vnode_unlock(vnode);
    errno=r;
    return -1;
  }

  if(!S_ISLNK(vnode->vstat.st_mode)) {
    errno=EINVAL;
    r=-1;
  } else {
    if(!super->fs_tbl->readlink) {
      r=-1;
      errno=ENOSYS;
    } else {
      r=super->fs_tbl->readlink(vnode,path,&link_size);
      if(!r) path[link_size]='\0'; /* some file system can avoid null char at end */
      else {
        r=-1;
        errno=r;
      }
    }
  }


  if(!nc) vnode_destroy(vnode);
  else vnode_unlock(vnode);

  return r;
}

int v2_mkdir(super_block_t *super,char *path,int *pps,int pp,int plen,
	     int offset,uid_t uid,gid_t gid,int mode)
{
  int r;
  int i;
  off_t rear_off=0;
  uint64_t shash;
  vnode_t *vnode;

  if(super->flags & V2FS_RO)  /* time to return error */
    return EROFS;

  if(!super->fs_tbl->createdir) /* filesystem doesn't support creation of directories */
    return ENOSYS;

  /* prepare names */
  if(offset>0) {
    rear_off=pps[offset];
    path=(char*)(path+pps[offset]);
    plen-=pps[offset];
    pps+=offset;
    pp-=offset;
    for(i=0;i<pp;i++)
      pps[i]-=rear_off;
  }
  shash=vnode_hash_ch(path);
  if((vnode=vnode_lookup(shash))==NULL) {
    vnode=vnode_create(super,path,pp,plen,pps);
    if(!vnode)       return ENOMEM;
  } else
    return EEXIST;

  /* it's ok, now let's create directory */
  r=super->fs_tbl->createdir(vnode,path,pps,pp,mode,uid,gid);
  /* destroy temproary allocated vnode */
  vnode_destroy(vnode);

  return r;
}

int v2_unlink(super_block_t *super,char *path,int *pps,int pp,int plen,
	     int offset,uid_t uid,gid_t gid)
{
  int r;
  int i,nc=0;
  off_t rear_off=0;
  uint64_t shash;
  vnode_t *vnode;

  if(super->flags & V2FS_RO)  /* time to return error */
    return EROFS;

  if(!super->fs_tbl->unlink) /* filesystem doesn't support creation of directories */
    return ENOSYS;

  /* prepare names */
  if(offset>0) {
    rear_off=pps[offset];
    path=(char*)(path+pps[offset]);
    plen-=pps[offset];
    pps+=offset;
    pp-=offset;
    for(i=0;i<pp;i++)
      pps[i]-=rear_off;
  }
  shash=vnode_hash_ch(path);
  if((vnode=vnode_lookup(shash))==NULL) {
    vnode=vnode_create(super,path,pp,plen,pps);
    if(!vnode) {
      errno=ENOMEM;
      return -1;
    }
  } else {
    nc++;
    vnode_lock(vnode);
  }

  /* it's ok, now let's lookup */
  if(!nc) {
    r=super->fs_tbl->lookup(vnode,path,pps,pp);
    if(r!=0) return ENOENT; /* no such file - lucky */
  }
  if(S_ISDIR(vnode->vstat.st_mode)) {
    r=EISDIR;
    goto __unlink_finish;
  }

  r=super->fs_tbl->unlink(vnode,uid,gid);

 __unlink_finish:

  if(!nc) vnode_destroy(vnode);
  else vnode_unlock(vnode);

  return r;
}

int v2_unlinkdir(super_block_t *super,char *path,int *pps,int pp,int plen,
		 int offset,uid_t uid,gid_t gid)
{
  int r;
  int i,nc=0;
  off_t rear_off=0;
  uint64_t shash;
  vnode_t *vnode;

  if(super->flags & V2FS_RO)  /* time to return error */
    return EROFS;

  if(!super->fs_tbl->unlinkdir) /* filesystem doesn't support deletion of directories */
    return ENOSYS;

  /* prepare names */
  if(offset>0) {
    rear_off=pps[offset];
    path=(char*)(path+pps[offset]);
    plen-=pps[offset];
    pps+=offset;
    pp-=offset;
    for(i=0;i<pp;i++)
      pps[i]-=rear_off;
  }
  shash=vnode_hash_ch(path);
  if((vnode=vnode_lookup(shash))==NULL) {
    vnode=vnode_create(super,path,pp,plen,pps);
    if(!vnode) {
      errno=ENOMEM;
      return -1;
    }
  } else {
    nc++;
    vnode_lock(vnode);
  }

  /* it's ok, now let's lookup */
  if(!nc) {
    r=super->fs_tbl->lookup(vnode,path,pps,pp);
    if(r!=0) return ENOENT; /* no such file - lucky */
  }
  if(!S_ISDIR(vnode->vstat.st_mode)) {
    r=ENOTDIR;
    goto __unlinkdir_finish;
  }

  r=super->fs_tbl->unlinkdir(vnode,uid,gid);

 __unlinkdir_finish:

  if(!nc) vnode_destroy(vnode);
  else vnode_unlock(vnode);

  return r;
}

int v2_open(super_block_t *super,char *path,int *pps,int pp,int offset,
	    int plen,pid_t pid,uid_t uid,gid_t gid,int mode,int flags,
	    struct stat *buf)
{
  vnode_t *vnode;
  int r,nc=0;
  uint64_t shash;
  pidsession_t *psess;
  int i;
  off_t rear_off=0;

  /* check flags */
  if((super->flags & V2FS_RO) && !(flags & O_RDONLY)) { /* time to return error */
    errno=EROFS;
    return -1;
  }

  /* prepare names */
  if(offset>0) {
    rear_off=pps[offset];
    path=(char*)(path+pps[offset]);
    plen-=pps[offset];
    pps+=offset;
    pp-=offset;
    for(i=0;i<pp;i++)
      pps[i]-=rear_off;
  }
  shash=vnode_hash_ch(path);
  if((vnode=vnode_lookup(shash))==NULL) {
    vnode=vnode_create(super,path,pp,plen,pps);
    if(!vnode) {
      errno=ENOMEM;
      return -1;
    }
  } else {
    nc++;
    vnode_lock(vnode);
  }
  /* it's ok, now let's lookup */
  r=super->fs_tbl->lookup(vnode,path,pps,pp);
  if(r!=0 && !(flags & O_CREAT)) {
    if(!nc) vnode_destroy(vnode);
    else vnode_unlock(vnode);
    errno=r;
    return -1;
  } else if(r!=0 && (flags & O_CREAT)) {
    if(!super->fs_tbl->create)      r=ENOSYS;
    else      r=super->fs_tbl->create(vnode,path,pps,pp,mode,uid,gid);
    if(r!=0) {
      if(!nc) vnode_destroy(vnode);
      else vnode_unlock(vnode);
      errno=r;
      return -1;
    } else goto __assign_essid;
  }

  if(S_ISDIR(vnode->vstat.st_mode) && !(flags & O_DIRECTORY)) {
    if(!nc) vnode_destroy(vnode);
    else vnode_unlock(vnode);
    errno=EISDIR;
    return -1;
  } else if(!S_ISDIR(vnode->vstat.st_mode) && (flags & O_DIRECTORY)) {
    if(!nc) vnode_destroy(vnode);
    else vnode_unlock(vnode);
    errno=ENOTDIR;
    return -1;
  }

  if(uid!=0) {
    r=vnode_permission(vnode,uid,gid);
    if((flags & O_RDWR) || (flags & O_TRUNC)) {
      if(!(r & MAY_WRITE)) {
	if(!nc) vnode_destroy(vnode);
	else vnode_unlock(vnode);
	errno=EACCES;
	return -1;
      }
    }
    if((flags & O_RDONLY) && !(r & MAY_READ)) {
      if(!nc) vnode_destroy(vnode);
      else vnode_unlock(vnode);
      errno=EACCES;
      return -1;
    }
    if((flags & O_DIRECTORY) && !(r & MAY_EXEC)) {
      if(!nc) vnode_destroy(vnode);
      else vnode_unlock(vnode);
      errno=EACCES;
      return -1;
    }
  }

  if(S_ISDIR(vnode->vstat.st_mode) && (flags & O_DIRECTORY)) {
    r=super->fs_tbl->opendir(vnode,path,pps,pp);
    if(r!=0) {
      if(!nc) vnode_destroy(vnode);
      else vnode_unlock(vnode);
      errno=r;
      return -1;
    }
  }

 __assign_essid:

  /* well test is made let's create a new entry */
  psess=pidsession_lookup(pid);
  if(!psess) { /* wasn't opened anything by this pid already */
    psess=pidsession_create(pid,uid,gid,256);
    if(!psess) {
      if(!nc) vnode_destroy(vnode);
      else vnode_unlock(vnode);
      errno=ENOMEM;
      return -1;
    }
  }

  /* check for O_TRUNC */
  if((!(flags & O_DIRECTORY) && (flags & O_TRUNC)) 
     && vnode->super->fs_tbl->truncate) {
    vnode->super->fs_tbl->truncate(vnode,0);
  }

  vnode_unlock(vnode);

  r=pidsession_addeid(psess,mode,flags,vnode);
  if(r<0) {
    errno=-r;
    if(!nc) vnode_destroy(vnode);
    return -1;
  }

  memcpy(buf,&vnode->vstat,sizeof(struct stat));

  if(!nc) vnode_push(vnode);

  return r;
}

int v2_mmap(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,size_t size,
            off_t offset,int flags,int prot,int *r_obj)
{
  pidsession_t *psess;
  fessid_t *fsess;
  int r=0,mmobj_id;

  psess=pidsession_lookup(pid);
  if(!psess) {
    return EBADF;
  } else
    pidsession_lock(psess);
  fsess=pidsession_lunique(psess,unique);

  if(!fsess) {
    pidsession_unlock(psess);
    return EBADF;
  }

  vnode_lock(fsess->vnode);
  if(!fsess->vnode->super->fs_tbl->mmap) { 
    if(fsess->vnode->super->is_mappable) {
      /**
       * NOTE: some notes, generic mapper is very trivial
       * and works only for read only mappings without
       * pagecache
       */
      /* check size and offset */
      if(offset+size>fsess->vnode->vstat.st_size) {
        pidsession_unlock(psess);
        vnode_unlock(fsess->vnode);
        return EOVERFLOW;
      }
      /* check proto */
      if(prot & PROT_WRITE) {
        pidsession_unlock(psess);
        vnode_unlock(fsess->vnode);
        return EINVAL;
      }
      if(!fsess->vnode->mem_obj) { /* in generic mapper it can be on first mmap */
        mmobj_id=memobj_create_wbackend(fsess->vnode->super->serve_port,
                                        fsess->vnode->vstat.st_size,
                                        MMO_LIFE_LEECH,MEMOBJ_SHARED,
                                        (long)fsess->vnode);
        /* TODO: set attributes */
        if(mmobj_id<0) {
          pidsession_unlock(psess);
          vnode_unlock(fsess->vnode);
          return errno;
        }
        fsess->vnode->mem_obj=mmobj_id;
      }
      *r_obj=fsess->vnode->mem_obj;
      vnode_unlock(fsess->vnode);
      pidsession_unlock(psess);
      return 0;
    }
    vnode_unlock(fsess->vnode);
    pidsession_unlock(psess);
    return ENOSYS;
  }
  r=fsess->vnode->super->fs_tbl->mmap(fsess->vnode,r_obj,size,offset,
                                      flags,prot);
  vnode_unlock(fsess->vnode);
  pidsession_unlock(psess);

  return r;
}

void *v2_read(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,size_t *size,
              off_t offset,void *buf)
{
  pidsession_t *psess;
  fessid_t *fsess;
  size_t nsize=*size;
  void *nbuf;

  psess=pidsession_lookup(pid);
  if(!psess) {
    errno=EBADF;
    return NULL;
  } else
    pidsession_lock(psess);
  fsess=pidsession_lunique(psess,unique);

  if(!fsess) {
    pidsession_unlock(psess);
    errno=EBADF;
    return NULL;
  }
  if((fsess->flags & O_WRONLY) && !(fsess->flags & O_RDONLY)) {
    pidsession_unlock(psess);
    errno=EBADF;
    return NULL;
  }

  /* now let's check for overflow */
  vnode_lock(fsess->vnode);
  if(offset>fsess->vnode->vstat.st_size) {
    errno=EOVERFLOW;
    vnode_unlock(fsess->vnode);
    pidsession_unlock(psess);
    return NULL;
  } 

  if(offset+nsize>fsess->vnode->vstat.st_size) 
    nsize=fsess->vnode->vstat.st_size-offset;
  if(!nsize || nsize<0) {
    errno=0;
    *size=0;
    vnode_unlock(fsess->vnode);
    pidsession_unlock(psess);
    return NULL;
  } 

  if(!fsess->vnode->super->fs_tbl->read) {
    nbuf=NULL;
    *size=0;
  } else {
    nbuf=fsess->vnode->super->fs_tbl->read(fsess->vnode,NULL,&nsize,offset);
    *size=nsize;
  }

  vnode_unlock(fsess->vnode);
  pidsession_unlock(psess);

  return nbuf;
}

int v2_ftruncate(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,off_t off)
{
  pidsession_t *psess;
  fessid_t *fsess;
  int r;

  psess=pidsession_lookup(pid);
  if(!psess) {
    return EBADF;
  } else
    pidsession_lock(psess);
  fsess=pidsession_lunique(psess,unique);

  if(!fsess) {
    pidsession_unlock(psess);
    return EBADF;
  }
  if((fsess->flags & O_RDONLY) && !(fsess->flags & O_WRONLY)) {
    pidsession_unlock(psess);
    return EBADF;
  }
  /* now let's check for O_TRUNC */
  vnode_lock(fsess->vnode);

  if(!fsess->vnode->super->fs_tbl->truncate)
    return ENOSYS;
  else
    r=fsess->vnode->super->fs_tbl->truncate(fsess->vnode,off);

  vnode_unlock(fsess->vnode);
  pidsession_unlock(psess);

  return r;
}

/* do check before copying any data from message, it will avoid some flood attacks */
int v2_checkwr(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,off_t off)
{
  pidsession_t *psess;
  fessid_t *fsess;

  psess=pidsession_lookup(pid);
  if(!psess) {
    return EBADF;
  } else
    pidsession_lock(psess);
  fsess=pidsession_lunique(psess,unique);

  if(!fsess) {
    pidsession_unlock(psess);
    return EBADF;
  }
  if((fsess->flags & O_RDONLY) && !(fsess->flags & O_WRONLY)) {
    pidsession_unlock(psess);
    return EBADF;
  }

  pidsession_unlock(psess);

  return 0;
}

int v2_fstat(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,struct stat *buf)
{
  pidsession_t *psess;
  fessid_t *fsess;

  psess=pidsession_lookup(pid);
  if(!psess) {
    return EBADF;
  } else
    pidsession_lock(psess);
  fsess=pidsession_lunique(psess,unique);

  if(!fsess) {
    pidsession_unlock(psess);
    return EBADF;
  }

  vnode_lock(fsess->vnode);

  memcpy(buf,&fsess->vnode->vstat,sizeof(struct stat));

  vnode_unlock(fsess->vnode);
  pidsession_unlock(psess);

  return 0;
}

/*
 * Note for fucking bastards - filesystem shouldn't handle all 
 * your crazy ioctl-related shit, it's a specific command interface
 * on file system layer, and you shoud care to fit all your common
 * interface to RPC header.
 * Extended iotcls goes to the device/stream service directly!
 */
int v2_ioctl(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,
	     uni_rpc_t *rpc)
{
  pidsession_t *psess;
  fessid_t *fsess;
  int ec=0;

  psess=pidsession_lookup(pid);
  if(!psess) {
    return EBADF;
  } else
    pidsession_lock(psess);
  fsess=pidsession_lunique(psess,unique);

  if(!fsess) {
    pidsession_unlock(psess);
    return EBADF;
  }

  vnode_lock(fsess->vnode);

  if(!fsess->vnode->super->fs_tbl->ioctl) 
    ec=ENODEV; /* POSIX headache require this error code (I think ENOSYS is more compliant) */
  else
    ec=fsess->vnode->super->fs_tbl->ioctl(fsess->vnode,rpc);

  vnode_unlock(fsess->vnode);
  pidsession_unlock(psess);

  return ec;
}

int v2_write(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,
	     void *buf,size_t *size,off_t off)
{
  pidsession_t *psess;
  fessid_t *fsess;
  int r;

  psess=pidsession_lookup(pid);
  if(!psess) {
    return EBADF;
  } else
    pidsession_lock(psess);
  fsess=pidsession_lunique(psess,unique);

  if(!fsess) {
    pidsession_unlock(psess);
    return EBADF;
  }
  if((fsess->flags & O_RDONLY) && !(fsess->flags & O_WRONLY)) {
    pidsession_unlock(psess);
    return EBADF;
  }

  vnode_lock(fsess->vnode);

  if(!fsess->vnode->super->fs_tbl->write) 
    r=ENOSYS;
  else {
    /* TODO: there should be a postponed logic */
    r=fsess->vnode->super->fs_tbl->write(fsess->vnode,buf,size,off);
  }

  vnode_unlock(fsess->vnode);
  pidsession_unlock(psess);

  return r;
}

int v2_readdir(pid_t pid,uid_t uid,gid_t gid,uint16_t unique,void *msg,
	       size_t *msg_len,size_t *size,off_t off,uint16_t *ec)
{
  pidsession_t *psess;
  fessid_t *fsess;
  size_t nsize=*size;
  size_t nmsg_len=*msg_len,rlen=0;
  struct dirent entr; /* reading entry */
  struct tr_dirent *tr_entr; /* trail entry */
  int i,r;

  psess=pidsession_lookup(pid);
  if(!psess) {
    return EBADF;
  } else
    pidsession_lock(psess);
  fsess=pidsession_lunique(psess,unique);

  if(!fsess) {
    pidsession_unlock(psess);
    return EBADF;
  } else {
    if(fsess->vnode->super->fs_tbl->readdir==NULL) {
      pidsession_unlock(psess);
      return ENOSYS;
    }
    vnode_lock(fsess->vnode);
  }

  if(!S_ISDIR(fsess->vnode->vstat.st_mode)) {
    vnode_unlock(fsess->vnode);
    pidsession_unlock(psess);
    return EINVAL;
  }
  /**
   * int (*readdir)(vnode_t *node,size_t *size,off_t off,struct dirent *dentry);
   */
  for(i=0;i<nsize;i++) {
    r=fsess->vnode->super->fs_tbl->readdir(fsess->vnode,i,off+i,&entr);
    if(r<0) {
      vnode_unlock(fsess->vnode);
      pidsession_unlock(psess);
      return errno;
    } else if(r==DIRENT_RR) {
      vnode_unlock(fsess->vnode);
      pidsession_unlock(psess);
      *size=0;
      *ec=r;
      return 0;
    } else if(r==DIRENT_EL) { /* reading end */
      *msg_len=rlen;
      if(i)	*size=(i);
      else	*size=0;
      *ec=r;
      vnode_unlock(fsess->vnode);
      pidsession_unlock(psess);
      return 0;
    } else {
      if(rlen+strlen((char *)entr.d_name)+sizeof(struct tr_dirent)>nmsg_len) /* not enough space */ {
	*msg_len=rlen;
	*size=i+1;
	*ec=0;
	break;
      } else {
	tr_entr=(struct tr_dirent *)((char *)msg+rlen);
	tr_entr->d_ino=entr.d_ino;
	tr_entr->d_off=entr.d_off;
	tr_entr->d_namlen=entr.d_namlen;
	tr_entr->d_rec_len=entr.d_rec_len;
	memcpy((char *)msg+rlen+sizeof(struct tr_dirent),entr.d_name,tr_entr->d_namlen);
	rlen+=sizeof(struct tr_dirent);
	rlen+=tr_entr->d_namlen;
      }
    }
  }

  *msg_len=rlen;
  *size=i;

  vnode_unlock(fsess->vnode);
  pidsession_unlock(psess);

  return 0;
}

int v2_close(pid_t pid,uint16_t unique)
{
  return pidsession_cloeid(pid,unique);
}

