/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * postponed and non-direct operations
 * queues and implementation
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <fcntl.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/iov.h>
#include <ipc/rpc.h>
#include <pthread.h>
#include <sys/v2.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>
#include <datalib/usrtc.h>

#include <vfs/vfsdef.h>
#include <vfs/vnode.h>
#include <vfs/super.h>
#include <vfs/session.h>
#include <vfs/polls.h>

#include "vfsint.h"

/* locally used structures */
struct ppq_elem {
  usrtc_node_t *node;
  queue_obj_t *qobj;
};

static queue_pp_t *pp_queue=NULL;

/* local functions */
static int __pp_qo_comp(const void *a,const void *b);
static inline int pp_queue_cut(struct ppq_elem *tc);
static inline int pp_queue_paste(struct ppq_elem *tc);
static int pp_queue_destroy(struct ppq_elem *tc,
                            void (*destroy_handler)(void *));

int pp_queue_init(void)
{
  if((pp_queue=malloc(sizeof(*pp_queue)))==NULL)
    return -1;

  /* init */
  if((pp_queue->uds=malloc(sizeof(*pp_queue->uds)))==NULL) {
    free(pp_queue);
    return -1;
  }
  usrtc_init(pp_queue->uds,USRTC_BST,USRTC_COUNT_T_MAX,__pp_qo_comp);
  pthread_mutex_init(&pp_queue->lock,NULL);

  return 0;
}

int pp_queue_insert(int msgid,uni_rpc_t *rpc)
{
  usrtc_node_t *node=NULL;
  queue_obj_t *qobj=NULL;

  if((node=malloc(sizeof(*node)))==NULL) {
  __enomem_queue:
    free(node);
    free(qobj);
    errno=ENOMEM;
    return -1;
  }
  if((qobj=malloc(sizeof(*qobj)))==NULL) {
    goto __enomem_queue;
  }
  usrtc_node_init(node,qobj);
  queue_pp_lock(pp_queue);
  qobj->op_id=usrtc_count(pp_queue->uds);
  memcpy(rpc,&qobj->rpc,sizeof(uni_rpc_t));
  qobj->msgid=msgid;
  qobj->data=NULL;
  usrtc_insert(pp_queue->uds,node,&qobj->op_id);
  queue_pp_unlock(pp_queue);

  return 0;
}

/* main thread */
void *v2_ppq_thread(filesystem_proc_t *fs)
{
  int r;
  struct ppq_elem wd;
  queue_obj_t *qobj=NULL;

  while(true) {
    while((r=pp_queue_cut(&wd))) {
      qobj=wd.qobj;
      switch(qobj->rpc.rpc_call) {
      case RPC_READ:
        break;
      case RPC_WRITE:
        break;
      default: /* EINVAL, remove */
        pp_queue_destroy(&wd,NULL);
        break;
      }
    }
    usleep(20);
  }

  return NULL;
}

/* implementation */

static int pp_queue_destroy(struct ppq_elem *tc,void (*destroy_handler)(void *))
{
  free(tc->node);
  if(tc->qobj->data && destroy_handler!=NULL)
    destroy_handler(tc->qobj->data);
  free(tc->qobj);

  return 0;
}

static inline int pp_queue_cut(struct ppq_elem *tc)
{
  queue_pp_lock(pp_queue);
  tc->node=usrtc_first(pp_queue->uds);
  if(!tc->node) {
    queue_pp_unlock(pp_queue);
    return 0;
  } else
    tc->qobj=usrtc_node_getdata(tc->node);
  usrtc_delete(pp_queue->uds,tc->node);
  queue_pp_unlock(pp_queue);

  return 1;
}

static inline int pp_queue_paste(struct ppq_elem *tc)
{
  usrtc_node_init(tc->node,tc->qobj);
  queue_pp_lock(pp_queue);
  tc->qobj->op_id=usrtc_count(pp_queue->uds);
  usrtc_insert(pp_queue->uds,tc->node,&tc->qobj->op_id);
  queue_pp_unlock(pp_queue);

  return 0;
}

static int __pp_qo_comp(const void *a,const void *b)
{
  return *(int *)b-*(int *)a;
}
