/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * sessions structures and operations 
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/iov.h>
#include <ipc/rpc.h>
#include <pthread.h>
#include <sys/v2.h>
#include <ns/ns.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <datalib/usrtc.h>

#include <vfs/vfsdef.h>
#include <vfs/vnode.h>
#include <vfs/super.h>
#include <vfs/session.h>

#include "vfsint.h"

#define BM_PREALLOC  8

static usrtc_t sessions;
static pthread_mutex_t sessions_ilock;
static int pids;
static int max_pids=65000;
static struct fss_hooks *_gl_hook=NULL;

#define sessions_lock  pthread_mutex_lock(&sessions_ilock)
#define sessions_unlock  pthread_mutex_unlock(&sessions_ilock)

static int __comppid(const void *a, const void *b);
static int __compid(const void *a, const void *b);

static unsigned long *__findset_unique(unsigned long *bm,uint8_t *ma,int *id);
static void __free_unique(unsigned long *bm,uint16_t unique);

int fs_session_init(void)
{
  usrtc_init(&sessions,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__comppid);
  pids=0;

  pthread_mutex_init(&sessions_ilock,NULL);

  return 0;
}

int fs_session_sethooks(struct fss_hooks *nhooks)
{
  if(!nhooks)
    return EINVAL;
  else
    _gl_hook=nhooks;

  return 0;
}

pidsession_t *pidsession_create(pid_t pid,uid_t uid,gid_t gid,int max_open)
{
  usrtc_node_t *node;
  pidsession_t *session;

  if(pids>=max_pids)
    return NULL;

  if((session=malloc(sizeof(*session)))==NULL) 
    return NULL;
  if((node=malloc(sizeof(*node)))==NULL) 
    return NULL;

  /* initialize bitmap */
  session->along=BM_PREALLOC;
  session->bm=malloc(sizeof(unsigned long)*BM_PREALLOC);
  if(!session->bm)    return NULL;
  memset(session->bm,0,sizeof(unsigned long)*BM_PREALLOC);
  /* init */
  session->pid=pid;
  session->uid=uid;
  session->gid=gid;
  session->max_open=max_open;
  /* reset counters */
  session->active=session->n_reset=0;
  /* initialize tree */
  usrtc_init(&session->tree,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__compid);
  /* initialize locking */
  pthread_mutex_init(&session->lock,NULL);

  sessions_lock;

  usrtc_node_init(node,session);
  usrtc_insert(&sessions,node,&session->pid);

  pids++;

  sessions_unlock;

  return session;
}

pidsession_t *pidsession_lookup(pid_t pid)
{
  usrtc_node_t *node;

  sessions_lock;

  node=usrtc_lookup(&sessions,&pid);
  sessions_unlock;

  if(!node) 
    return NULL;
  else 
    return (pidsession_t *)usrtc_node_getdata(node);  
}

int pidsession_close(pid_t pid)
{
  pidsession_t *sess;
  fessid_t *nid;
  usrtc_node_t *node;

  if((sess=pidsession_lookup(pid))==NULL)
    return 0;
  else { 
    pidsession_lock(sess);
    while((node=usrtc_first(&sess->tree))!=NULL) {
      nid=usrtc_node_getdata(node);
      /* change ref count */
      vnode_lock(nid->vnode);
      if(nid->vnode->ref_count>1) {
	nid->vnode->ref_count--;
	vnode_unlock(nid->vnode);
      } else {
	/* remove this from vnode cache */
	vnode_pop(nid->vnode);
	/* unlock */
	vnode_unlock(nid->vnode);
	if(nid->vnode->super->fs_tbl->close)
	  nid->vnode->super->fs_tbl->close(nid->vnode);
	/* destroy */
	vnode_destroy(nid->vnode);
      }
      /* remove */
      usrtc_delete(&sess->tree,node);
      free(node);
      free(nid);
    }
    /* call kill hook */
    if(_gl_hook) _gl_hook->h_kill(sess);
    pidsession_unlock(sess);
    pthread_mutex_destroy(&sess->lock);
    free(sess->bm);
    /* remove from sessions */
    sessions_lock;
    node=usrtc_lookup(&sessions,&pid);
    usrtc_delete(&sessions,node);
    sessions_unlock;
    /* release */
    free(node);
    free(sess);
  }

  return 0;
}

int pidsession_clone(pid_t pid,pid_t n_pid)
{
  pidsession_t *sess;
  pidsession_t *nsess;
  usrtc_node_t *node,*nnode,*insnode;
  fessid_t *nid,*nnid;

  if((sess=pidsession_lookup(pid))==NULL)
    return ENOENT;
  else { 
    pidsession_lock(sess);
    if((nsess=malloc(sizeof(*nsess)))==NULL) {
      pidsession_unlock(sess);
      return ENOMEM;
    }
    if((nsess->bm=malloc(sess->along*sizeof(unsigned long)))==NULL) {
    __enomem_fail:
      free(nsess);
      pidsession_unlock(sess);
      return ENOMEM;
    }
    insnode=malloc(sizeof(*insnode));
    if(!insnode) goto __enomem_fail;
    /* init tree */
    usrtc_init(&nsess->tree,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__compid);
    /* ok let's clone this all things */
    node=usrtc_first(&sess->tree);
    while(node!=NULL) {
      if((nnode=malloc(sizeof(*nnode)))==NULL || (nnid=malloc(sizeof(*nnid)))==NULL) {
	return ENOMEM; 
      }
      nid=usrtc_node_getdata(node);
      usrtc_node_init(nnode,nnid);
      nnid->id=nid->id;
      nnid->mode=nid->mode;
      nnid->flags=nid->flags;
      nnid->offset=nid->offset;
      nnid->vnode=nid->vnode;
      /* ok update ref count */
      vnode_lock(nnid->vnode);
      nnid->vnode->ref_count++;
      vnode_unlock(nnid->vnode);
      usrtc_insert(&nsess->tree,nnode,&nnid->id);
      nnode=NULL; nnid=NULL;
      /* going to the next node */
      node=usrtc_next(&sess->tree,node);
    }
    /* ok let's init nsess */
    memcpy(nsess->bm,sess->bm,sess->along*(sizeof(unsigned long)));
    nsess->pid=n_pid;
    nsess->uid=sess->uid;
    nsess->gid=sess->gid;
    pthread_mutex_init(&nsess->lock,NULL);
    usrtc_node_init(insnode,nsess);
    nsess->cap=sess->cap;
    nsess->max_open=sess->max_open;
    nsess->active=sess->active;
    nsess->n_reset=sess->n_reset;
    nsess->along=sess->along;
    /* call hook */
    if(_gl_hook) _gl_hook->h_clone(sess,n_pid);
    /* insert new resources to the structure */
    sessions_lock;
    usrtc_insert(&sessions,insnode,&nsess->pid);
    sessions_unlock;
    pidsession_unlock(sess);
  }

  return 0;
}

int pidsession_reset(pid_t pid)
{
  pidsession_t *sess;
  usrtc_node_t *node;
  fessid_t *nid;

  if((sess=pidsession_lookup(pid))==NULL)
    return 0;
  else if(sess->n_reset==0)
    return 0; 
  else { 
    pidsession_lock(sess);
    while((node=usrtc_first(&sess->tree))!=NULL) {
      nid=usrtc_node_getdata(node);
      if(nid->flags & FD_CLOEXEC) {
	/* well, shootout this */
	node=usrtc_lookup(&sess->tree,&nid->id);
	/* call close hook */
	if(_gl_hook!=NULL)
	  _gl_hook->h_close(sess,nid);
	/* change ref count */
	vnode_lock(nid->vnode);
	if(nid->vnode->ref_count>1) {
	  nid->vnode->ref_count--;
	  vnode_unlock(nid->vnode);
	} else {
	  /* remove this from vnode cache */
	  vnode_pop(nid->vnode);
	  /* unlock */
	  vnode_unlock(nid->vnode);
	  if(nid->vnode->super->fs_tbl->close)
	    nid->vnode->super->fs_tbl->close(nid->vnode);
	  /* destroy */
	  vnode_destroy(nid->vnode);
	}
	/* remove */
	usrtc_delete(&sess->tree,node);	
	__free_unique(sess->bm,nid->id);
	free(node);
	free(nid);
      } else continue;
    }
    sess->n_reset=0;
    pidsession_unlock(sess);
  }

  return 0;
}

int pidsession_change(pid_t pid,uid_t uid,gid_t gid)
{
  pidsession_t *sess;

  if((sess=pidsession_lookup(pid))==NULL)
    return ENOENT;

  pidsession_lock(sess);

  sess->uid=uid;
  sess->gid=gid;

  pidsession_unlock(sess);

  return 0;
}

int pidsession_setcap(pid_t pid,int cap)
{
  pidsession_t *sess;

  if((sess=pidsession_lookup(pid))==NULL)
    return ENOENT;

  pidsession_lock(sess);

  sess->cap=cap;

  pidsession_unlock(sess);

  return 0;
}

int pidsession_setlimit(pid_t pid,int max_open)
{
  pidsession_t *sess;

  if((sess=pidsession_lookup(pid))==NULL)
    return ENOENT;

  pidsession_lock(sess);

  sess->max_open=max_open;

  pidsession_unlock(sess);

  return 0;
}

int pidsession_cloeid(pid_t pid, uint16_t id)
{
  pidsession_t *sess;
  fessid_t *nid;
  usrtc_node_t *node;

  if((sess=pidsession_lookup(pid))==NULL) return EBADF; 
  pidsession_lock(sess);

  if((node=usrtc_lookup(&sess->tree,&id))==NULL) {
    pidsession_unlock(sess);
    return EBADF;
  }

  nid=usrtc_node_getdata(node);
  /* call close hook */
  if(_gl_hook!=NULL)
    _gl_hook->h_close(sess,nid);
  /* change ref count */
  vnode_lock(nid->vnode);
  if(nid->vnode->ref_count>1) {
    nid->vnode->ref_count--;
    vnode_unlock(nid->vnode);
  } else {
    /* remove this from vnode cache */
    vnode_pop(nid->vnode);
    /* unlock */
    vnode_unlock(nid->vnode);
    if(nid->vnode->super->fs_tbl->close)
      nid->vnode->super->fs_tbl->close(nid->vnode);
    /* destroy */
    vnode_destroy(nid->vnode);
  }
  /* check flags */
  if(nid->flags & FD_CLOEXEC)
    sess->n_reset--;
  /* remove */
  usrtc_delete(&sess->tree,node);
  free(node);
  free(nid);
  __free_unique(sess->bm,id);

  pidsession_unlock(sess);

  return 0;
}

int pidsession_addeid(pidsession_t *sess,int mode, int flags,vnode_t *vnode)
{
  usrtc_node_t *node;
  fessid_t *nid;
  int ret=0;


  if((node=malloc(sizeof(*node)))==NULL) return -ENOMEM;
  pidsession_lock(sess);

  if((nid=malloc(sizeof(*nid)))==NULL) {
    pidsession_unlock(sess);
    return -ENOMEM;
  }

  sess->bm=__findset_unique(sess->bm,&sess->along,&ret);
  if(ret<0) { /* fuck */
    free(nid);
    pidsession_unlock(sess);
    return -ENOMEM;
  }

  /* init all data */
  nid->id=ret;
  nid->offset=0;
  nid->flags=flags;
  nid->mode=mode;
  nid->vnode=vnode; /* yep */
  /* check flags */
  if(flags & FD_CLOEXEC)
    sess->n_reset++;
  /* vnode logic */
  vnode_lock(vnode);
  vnode->ref_count++;
  vnode_unlock(vnode);

  /* insert */
  usrtc_node_init(node,nid);
  usrtc_insert(&sess->tree,node,&nid->id);
  /* call add hook */
  if(_gl_hook!=NULL)
    _gl_hook->h_add(sess,nid);

  pidsession_unlock(sess);

  return ret;
}

fessid_t *pidsession_lunique(pidsession_t *psess,uint16_t unique)
{
  usrtc_node_t *node;

  node=usrtc_lookup(&psess->tree,&unique);
  if(!node) return NULL;
  else
    return (fessid_t *)usrtc_node_getdata(node);
}

/* local functions implementation */

static int __comppid(const void *a, const void *b)
{
  return *(pid_t*)a-*(pid_t*)b;
}

static int __compid(const void *a, const void *b)
{
  return *(uint16_t*)a-*(uint16_t*)b;
}

static unsigned long *__findset_unique(unsigned long *bm,uint8_t *ma,int *id)
{
  int c=-1,i,a;
  int ni=*ma;

  for(i=0;i<*ma;i++) {
    for(a=0;a<sizeof(unsigned long)*8;a++) {
      if(!(bm[i] & (1 << a))) {
	c=a;
	break;
      }
    }
    if(c>=0) break;
  }


  if(c<0) { /* reallocate */
    bm=realloc(bm,sizeof(unsigned long)*(ni+8));
    memset(bm+ni,0,sizeof(unsigned long)*ni);
    ni+=8;
    i++; bm[i]=0; c=0;
  } 

  bm[i] |= (1 << c);

  *ma=ni;

  *id=i*(sizeof(unsigned long)*8) + c;

  return bm;
}

static void __free_unique(unsigned long *bm,uint16_t unique)
{
  int i,c;

  /* find appropriate map location */
  i=unique/(sizeof(unsigned long)*8);
  c=unique%(sizeof(unsigned long)*8);
  bm[i] &= ~(1 << c);

  return;
}
