/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * vnode cache and ops within its context 
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/iov.h>
#include <ipc/rpc.h>
#include <pthread.h>
#include <sys/v2.h>
#include <ns/ns.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <datalib/usrtc.h>

#include <vfs/vfsdef.h>
#include <vfs/vnode.h>
#include <vfs/super.h>
#include <vfs/session.h>

#include "vfsint.h"

static usrtc_t vnode_hash_tree;
static usrtc_t vnode_memobj_tree;
static pthread_mutex_t vncache_lock;

#define vnode_cache_lock  pthread_mutex_lock(&vncache_lock)
#define vnode_cache_unlock  pthread_mutex_unlock(&vncache_lock)

/* local functions */
static int __comp_memobj(const void *a,const void *b);
static int __comp_hash(const void *a,const void *b);

int vnode_cache_initialize(void)
{
  /* well init all structures */
  usrtc_init(&vnode_hash_tree,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__comp_hash);
  usrtc_init(&vnode_memobj_tree,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__comp_memobj);

  pthread_mutex_init(&vncache_lock,NULL);

  return 0;
}

vnode_t *vnode_create(struct __super_block_t *super,char *pathname,int pp,int plen,int *pps)
{
  vnode_t *node;

  /* well create new node */
  if((node=malloc(sizeof(*node)))==NULL) return NULL;

  /* allocate structures */
  if((node->node=malloc(sizeof(*node->node)))==NULL) {
    free(node);
    return NULL;
  }

  if((node->lock=malloc(sizeof(*node->lock)))==NULL) {
    free(node->node);
    free(node);
    return NULL;
  }

  if((node->pps=malloc(sizeof(int)*pp))==NULL) {
    free(node->lock);
    free(node->node);
    free(node);
    return NULL;
  }

  if((node->pathname=strdup(pathname))==NULL) {
    free(node->pps);
    free(node->lock);
    free(node->node);
    free(node);
    return NULL;
  }

  /* well done */
  memcpy(node->pps,pps,sizeof(int)*pp);
  node->plen=plen;
  node->pp=pp;
  node->super=super;
  node->ref_count=0;
  node->mem_obj=0; /* TODO: add creating of the memory object */
  node->spec_data=NULL;
  node->hash=vnode_hash_ch(pathname);
  /* mutex */
  pthread_mutex_init(&node->ilock,NULL);
  /* node structure */
  usrtc_node_init(node->node,node);

  return node;
}

int vnode_destroy(vnode_t *vnode)
{
  if(!vnode->node)
    free(vnode->node);
  if(!vnode->pps)
    free(vnode->pps);
  if(!vnode->lock) {
    /* fslock destroy */
    free(vnode->lock);
  }
  pthread_mutex_destroy(&vnode->ilock);
  free(vnode);

  return 0;
}

int vnode_pop(vnode_t *vnode)
{
  vnode_cache_lock;
  usrtc_delete(&vnode_hash_tree,vnode->node);
  vnode_cache_unlock;

  return 0;
}

int vnode_push(vnode_t *vnode)
{
  vnode_cache_lock;
  usrtc_insert(&vnode_hash_tree,vnode->node,&vnode->hash);
  vnode_cache_unlock;

  return 0;
}

vnode_t *vnode_lookup(uint64_t hash)
{
  usrtc_node_t *node;

  vnode_cache_lock;
  node=usrtc_lookup(&vnode_hash_tree,&hash);
  if(!node) {
    vnode_cache_unlock;
    return NULL;
  } else {
    vnode_cache_unlock;
    return (vnode_t*)usrtc_node_getdata(node);
  }
}

uint64_t vnode_hash_ch(char *path)
{
  uint64_t hash=0;

  if(!path) return hash;

  while(*path)
    hash=(hash << 5)+hash + *path++;

  return hash;
}

/* local functions - misc */
static int __comp_memobj(const void *a,const void *b)
{
  return *(unsigned long*)a-*(unsigned long*)b;
}

static int __comp_hash(const void *a,const void *b)
{
  return *(uint64_t*)a-*(uint64_t*)b;
}
