/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * Implementation of interaction of file systems with block devices
 */

#include <fsio/bdio.h>
#include <errno.h>
#include <ipc/iov.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <fcntl.h>

int open_blkdev(const char *path, struct bdio *bdio, bool sync)
{
  int flags = 0, ret = 0;
  struct mdata md;

  if (sync)
    flags |= O_SYNC;

  errno = 0;
  bdio->fd = open(path, flags | O_RDWR);
  if (errno == EROFS) {
    bdio->fd = open(path, flags | O_RDONLY);
  }
  if (bdio->fd == -1)
    return -errno;

  bdio->bd_pid = getfpid(bdio->fd);
  ret = ioctl(bdio->fd, BLKIOC_GET_MDATA, &md);
  if (!ret) {
    bdio->fs_pid = getpid();
    bdio->bd_capacity = md.capacity;
    bdio->hw_sect_bits = md.hw_sect_bits;
    bdio->cmd_chan = ipc_channel_open(bdio->bd_pid, md.cmd_port, IPC_BLOCKED_ACCESS);
    if ((signed)bdio->cmd_chan < 0)
      ret = bdio->cmd_chan;
  } else {
    ret = -errno;
  }
  if (ret)
    close(bdio->fd);

  return ret;
}

void close_blkdev(struct bdio *bdio)
{
  ipc_channel_close(bdio->cmd_chan);
  close(bdio->fd);
  bdio->fd = -1;
}

int32_t rw_blkdev_cached_sectors(int dir, struct bdio *bdio, struct cio_desc *cd,
                                 short prio, uint16_t flags)
{
  int ret;
  iovec_t ivec, ovecs[2];
  uni_rpc_t rpc;

  rpc.rpc_call = dir ? RPC_WRITE : RPC_READ;
  rpc.arg0 = bdio->fs_pid;
  rpc.arg1 = getufsid(bdio->fd);
  rpc.arg3 = MAX((unsigned)prio, MAX_RW_PRIORITY);
  rpc.arg4 = bdio->default_flags | flags;
  rpc.arg5 = cd->blocks_per_page;

  ovecs[0].iov_base = &rpc;
  ovecs[0].iov_len = sizeof(rpc);
  ovecs[1].iov_base = cd->heads;
  ovecs[1].iov_len = cd->heads_size;
  ivec.iov_base = &rpc;
  ivec.iov_len = sizeof(rpc);

  ret = ipc_iov_iov_send(bdio->cmd_chan, ovecs, 2, &ivec, 1);
  if (ret > 0) {
    if (cd->nr_proc_heads)
      *(cd->nr_proc_heads) = rpc.size;

    if (rpc.rpc_call)
      ret = -rpc.rpc_call;
    else
      ret = rpc.size;
  }

  return ret;
}

int set_blkdev_async_port(struct bdio *bdio, prtid_t port)
{
  int ret = 0;

  ret = ioctl(bdio->fd, BLKIOC_SET_ASYNC_PORT, port);
  if (!ret)
    bdio->async_port = port;

  return ret;
}

int sync_blkdev(struct bdio *bdio)
{
  uni_rpc_t rpc;
  ssize_t ret;

  rpc.rpc_call = RPC_SYNC;
  rpc.arg0 = bdio->fs_pid;
  rpc.arg1 = getufsid(bdio->fd);
  rpc.size = 0;

  ret = ipc_channel_msgsnd(bdio->cmd_chan, sizeof(rpc), &rpc, sizeof(rpc), &rpc);
  if (ret > 0) {
    ret = -rpc.rpc_call;
  }

  return ret;
}
