/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Based on kconfig from Linux kernel:
 *   Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 *   Released under the terms of the GNU GPL v2.0.
 *
 * Copyright (C) 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 */

#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "symbol.h"
#include "confdata.h"

const char *get_config_name(void)
{
  char *name = getenv("SCONFIG_CONFIG");

  return (name != NULL ? name : ".config");
}

int read_config(const char *filename)
{
  struct symbol *sym, *choice;
  FILE *fd;
  char line[256];
  char *p, *o;
  unsigned int lineno;

  fd = fopen(filename, "r");
  if (fd == NULL)
    return 0;

  lineno = 0;
  while (fgets(line, sizeof(line), fd) != NULL) {
    lineno++;

    switch (line[0]) {
    case '#':
      if (memcmp(line, "# CONFIG_", 9) != 0)
        continue;
      p = strchr(line + 9, ' ');
      if (p == NULL)
        continue;
      *p++ = '\0';
      if (strncmp(p, "is not set", 10) != 0)
        continue;
      sym = find_symbol(get_root_symbol(), line + 9);
      if (sym == NULL) {
        fprintf(stderr, "%s:%u: warning: unknown symbol %s\n",
                filename, lineno, line + 9);
        continue;
      }
      switch (sym->type) {
      case T_BOOL:
        sym->v.yesno = false;
	break;
      case T_STRING:
        sym->v.str = NULL;
        break;
      }
      break;
    case 'C':
      if (memcmp(line, "CONFIG_", 7) != 0) {
        fprintf(stderr, "%s:%u: warning: unexpected data\n", filename, lineno);
        continue;
      }
      p = strchr(line + 7, '=');
      if (p == NULL)
        continue;
      *p++ = '\0';
      o = strchr(p, '\r');
      if (o == NULL)
        o = strchr(p, '\n');
      *o = '\0';
      sym = find_symbol(get_root_symbol(), line + 7);
      if (sym == NULL) {
        fprintf(stderr, "%s:%u: warning: unknown symbol %s\n",
                filename, lineno, line + 7);
        continue;
      }
      switch (sym->type) {
      case T_BOOL:
        if (p[0] == 'y')
          sym->v.yesno = true;
	else if (p[0] == 'n')
	  sym->v.yesno = false;
	else {
	  sym->v.yesno = false;
	  fprintf(stderr, "%s:%u: warning unknown symbol value %s\n",
	          filename, lineno, p);
	}
	break;
      case T_STRING:
        if (p[0] != '"') {
          free(sym->v.str);
          sym->v.str = strdup(p);
        }
        else {
          if (o[-1] == '"') {
            o[-1] = '\0';
            free(sym->v.str);
            sym->v.str = strdup(p + 1);
          }
        }
        break;
      case T_SWITCH:
        choice = find_symbol(sym->t.exp.list, p);
        if (choice == NULL) {
          fprintf(stderr, "%s:%u: warning: unknown symbol %s\n",
                  filename, lineno, p);
          continue;
        }
        sym->v.choice = choice;
        break;
      }
      break;
    case '\r':
    case '\n':
      break;
    default:
      fprintf(stderr, "%s:%u: warning: unexpected data\n", filename, lineno);
      continue;
    }
  }

  fclose(fd);
  return 1;
}

static void write_symbol(FILE * fd, struct symbol *sym)
{
  struct symbol *i;

  for (i = sym; i != NULL; i = i->next)
    switch (i->type) {
    case T_MENU:
      fprintf(fd, "\n#\n# %s\n#\n", i->t.menu.name);
      write_symbol(fd, i->t.menu.list);
      fprintf(fd, "\n");
      break;
    case T_BOOL:
      if (i->v.yesno)
        fprintf(fd, "CONFIG_%s=y\n", i->t.exp.id);
      else
        fprintf(fd, "# CONFIG_%s is not set\n", i->t.exp.id);
      write_symbol(fd, i->t.exp.list);
      break;
    case T_STRING:
      if ((i->v.str != NULL) && (i->v.str[0] != '\0'))
        fprintf(fd, "CONFIG_%s=\"%s\"\n", i->t.simple.id, i->v.str);
      else
        fprintf(fd, "# CONFIG_%s is not set\n", i->t.simple.id);
      break;
    case T_SWITCH:
      if (i->v.choice != NULL)
        fprintf(fd, "CONFIG_%s=%s\n", i->t.exp.id, i->v.choice->t.exp.id);
      else
	fprintf(fd, "# CONFIG_%s is not set\n", i->t.exp.id);
      break;
    }
}

int write_config(const char *filename)
{
  FILE *fd;
  time_t now;

  fd = fopen(filename, "w");
  if (fd == NULL)
    return 0;

  fprintf(fd, "#\n# Automatically generated make config: do not edit\n");
  fprintf(fd, "# ---\n");
  now = time(NULL);
  fprintf(fd, "# %s#\n", ctime(&now));
  write_symbol(fd, get_root_symbol());

  fclose(fd);
  return 1;
}

static void write_symbol_as_def(FILE * fd, struct symbol *sym)
{
  struct symbol *i;

  for (i = sym; i != NULL; i = i->next)
    switch (i->type) {
    case T_MENU:
      write_symbol_as_def(fd, i->t.menu.list);
      break;
    case T_BOOL:
      if (i->v.yesno)
        fprintf(fd, "#define CONFIG_%s 1\n", i->t.exp.id);
      write_symbol_as_def(fd, i->t.exp.list);
      break;
    case T_STRING:
      if ((i->v.str != NULL) && (i->v.str[0] != '\0'))
        fprintf(fd, "#define CONFIG_%s \"%s\"\n", i->t.simple.id, i->v.str);
      break;
    case T_SWITCH:
      if (i->v.choice != NULL)
        fprintf(fd, "#define CONFIG_%s \"%s\"\n", i->t.exp.id, i->v.choice->t.exp.id);
      break;
    }
}

int write_autoconf(const char *filename)
{
  FILE *fd;
  time_t now;

  fd = fopen(filename, "w");
  if (fd == NULL)
    return 0;

  fprintf(fd, "/*\n"
              " * Auttomatically generated C config: do not edit\n"
	      " * ---\n");
  now = time(NULL);
  fprintf(fd, " * %s */\n", ctime(&now));
  fprintf(fd, "#define AUTOCONF_INCLUDED\n");
  write_symbol_as_def(fd, get_root_symbol());

  fclose(fd);
  return 1;
}
