/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Based on kconfig from Linux kernel:
 *   Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 *   Released under the terms of the GNU GPL v2.0.
 *
 * Copyright (C) 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 */

#include <limits.h>
#include <stdio.h>

#include "lxdialog/dialog.h"
#include "symbol.h"
#include "confdata.h"

void parse_config(const char *filename);

static const char menu_instructions[] =
  "Arrow keys navigate the menu.  "
  "<Enter> selects submenus --->.  "
  "Highlighted letters are hotkeys.  "
  "Pressing <Y> includes, <N> excludes features.  "
  "Press <Esc><Esc> to exit, <?> for Help.  "
  "Legend: [*] built-in  [ ] excluded";
static const char switch_instructions[] =
  "Use the arrow keys to navigate this window or "
  "press the hotkey of the item you wish to select "
  "followed by the <SPACE BAR>.  "
  "Press <?> for additional information about this option.";
static const char inputbox_instructions[] =
  "Please enter a string value.  "
  "Use the <TAB> key to move from the input field to the buttons below it.";
static const char nohelp_text[] =
  "There is no help available for this option.\n";
static char load_config_text[] =
  "Enter the name of the configuration file you wish to load.  "
  "Accept the name shown to restore the configuration you last retrieved.  "
  "Leave blank to abort.";
static char save_config_text[] =
  "Enter a filename to which this configuration should be saved "
  "as an alternate.  Leave blank to abort.";

static void show_menu(struct symbol *first, int indent);
static void show_cases(struct symbol *first);
static void show_inputbox(struct symbol *sym);
static void show_help(struct symbol *sym);
static void show_loadcfg(void);
static void show_savecfg(void);


static char config_name[PATH_MAX + 1];

static void set_config_filename(const char *filename)
{
  unsigned int size;

  set_dialog_backtitle(filename);

  size = snprintf(config_name, sizeof(config_name), "%s", filename);
  if (size >= sizeof(config_name))
    config_name[sizeof(config_name) - 1] = '\0';
}

static const char *get_menu_prompt(const struct symbol *sym)
{
  switch (sym->type) {
  case T_MENU:
    return sym->t.menu.name;
  case T_STRING:
  case T_CASE:
    if (sym->t.simple.desc != NULL)
      return sym->t.simple.desc;
    return sym->t.simple.id;
  case T_SWITCH:
  case T_BOOL:
    if (sym->t.exp.desc != NULL)
      return sym->t.exp.desc;
    return sym->t.exp.id;
  default:
    return "?? ?? ??";
  }
}

static void make_indent(int indent)
{
  int i;

  for (i = 0; i < indent; i++)
    item_add_str("  ");
}

static void build_dlg(struct symbol *start, int indent)
{
  struct symbol *i;

  for (i = start; i != NULL; i = i->next) {
    switch (i->type) {
    case T_MENU:
      if (is_empty_symbol(i))
        continue;
      item_make("    ");
      make_indent(indent);
      item_add_str("%s  --->", get_menu_prompt(i));
      item_set_tag('m');
      break;
    case T_BOOL:
      item_make("[%c] ", (i->v.yesno ? '*' : ' '));
      make_indent(indent);
      item_add_str("%s", get_menu_prompt(i));
      item_set_tag('t');
     break;
    case T_STRING:
      if ((i->v.str == NULL) || (i->v.str[0] == '\0'))
        item_make("( ) ");
      else
        item_make("(%s) ", i->v.str);
      make_indent(indent);
      item_add_str("%s", get_menu_prompt(i));
      item_set_tag('s');
      break;
    case T_SWITCH:
      if (is_empty_symbol(i))
        continue;
      item_make("    ");
      make_indent(indent);
      item_add_str("%s (%s)  --->", get_menu_prompt(i),
                get_menu_prompt(i->v.choice));
      item_set_tag('m');
      break;
    case T_CASE:
      /* ingnore indent */
      item_make("%s", get_menu_prompt(i));
      if (i->parent->v.choice == i) {
        item_set_selected(1);
        item_set_tag('X');
      }
      break;
    }
    item_set_data(i);

    if ((i->type == T_BOOL) && i->v.yesno)
      build_dlg(i->t.exp.list, indent + 1);
  }
}

static void show_menu(struct symbol *first, int indent)
{
  const char *title;
  int res;
  int s_scroll = 0;
  struct symbol *sym = NULL;

  for (;;) {
    item_reset();
    build_dlg(first, indent);
    if (first == get_root_symbol()) {
      item_make("--- ");
      item_set_tag(':');
      item_make("    Load an Alternate Configuration File");
      item_set_tag('L');
      item_make("    Save an Laternate Configuration File");
      item_set_tag('S');

      title = "Main Menu";
    }
    else
      title = get_menu_prompt(first->parent);

    dialog_clear();
    res = dialog_menu(title, menu_instructions, sym, &s_scroll);
    if ((res == 1) || (res == KEY_ESC) || (res == -ERRDISPLAYTOOSMALL))
      break;
    if (!item_activate_selected())
      continue;
    if (!item_tag())
      continue;

    sym = item_data();
    switch (res) {
    case 0:
      switch (item_tag()) {
      case 'L':
        show_loadcfg();
        break;
      case 'S':
        show_savecfg();
        break;
      default:
        switch (sym->type) {
        case T_MENU:
          show_menu(sym->t.menu.list, 0);
          break;
        case T_SWITCH:
          show_cases(sym->t.exp.list);
          break;
        case T_STRING:
          show_inputbox(sym);
          break;
        }
      }
      break;
    case 2:
      show_help(sym);
      break;
    case 3:                    /* Y */
      if (sym == NULL)
        break;
      if (sym->type == T_BOOL)
        sym->v.yesno = true;
      break;
    case 4:                    /* N */
      if (sym == NULL)
        break;
      if (sym->type == T_BOOL)
        sym->v.yesno = false;
      break;
    case 5:                    /* M */
      break;
    case 6:                    /* space bar */
      if (sym == NULL)
        break;
      if (sym->type == T_BOOL)
        sym->v.yesno = !sym->v.yesno;
      break;
    case 7:                    /* / */
      /* TODO: search */
      break;
    }
  }
}

static void show_cases(struct symbol *first)
{
  const char *title = get_menu_prompt(first->parent);
  int res;
  struct symbol *sym;

  for (;;) {
    item_reset();
    build_dlg(first, 0);

    dialog_clear();
    res = dialog_checklist(title, switch_instructions, 15, 70, 6);

    sym = item_data();
    switch (res) {
    case 0:
      sym->parent->v.choice = sym;
      return;
    case 1:
      show_help(sym);
      break;
    case KEY_ESC:
    case -ERRDISPLAYTOOSMALL:
      return;
    }
  }
}

static void show_inputbox(struct symbol *sym)
{
  const char *title = get_menu_prompt(sym);
  int res;

  for (;;) {
    dialog_clear();
    res = dialog_inputbox(title, inputbox_instructions, 10, 75, sym->v.str);
    switch (res) {
    case 0:
      free(sym->v.str);
      sym->v.str = strdup(dialog_input_result);
      return;
    case 1:
      show_help(sym);
      break;
    case KEY_ESC:
      return;
    }
  }
}

static void help_add_symbol_name(char *str, struct symbol *sym)
{
  if (sym->type == T_MENU)
    return;

  strcat(str, "Symbol: ");
  switch (sym->type) {
  case T_STRING:
  case T_CASE:
    strcat(str, sym->t.simple.id);
    break;
  case T_SWITCH:
  case T_BOOL:
    strcat(str, sym->t.exp.id);
    break;
  }
  strcat(str, "\n");
}

static void help_add_symbol_prompt(char *str, struct symbol *sym)
{
  if (sym->type == T_MENU)
    return;

  strcat(str, "Prompt: ");
  switch (sym->type) {
  case T_STRING:
  case T_CASE:
    if (sym->t.simple.desc != NULL)
      strcat(str, sym->t.simple.desc);
    break;
  case T_SWITCH:
  case T_BOOL:
    if (sym->t.exp.desc != NULL)
      strcat(str, sym->t.exp.desc);
    break;
  }
  strcat(str, "\n");
}

static void help_add_symbol_deps(char *str, struct symbol *sym)
{
  struct symbol *i;
  bool amp;

  strcat(str, "Depends on: ");

  /* dependencies from config structure */
  if (sym->type == T_CASE)
    strcat(str, "<choice>");
  else {
    amp = 0;
    for (i = sym->parent; i != NULL; i = i->parent)
      if (i->type == T_BOOL) {
	if (amp)
	  strcat(str, " && ");
        strcat(str, i->t.exp.id);
	amp = 1;
      }
  }

  /* declared dependencies - NOT IMPLEMENTED */

  strcat(str, "\n");
}

static void show_help(struct symbol *sym)
{
  char *str;

  if (sym == NULL)
    return;

  str = malloc(85 * 25);
  str[0] = '\0';

  if (sym->type != T_MENU) {
    strcat(str, "CONFIG_");
    switch (sym->type) {
    case T_STRING:
    case T_CASE:
      strcat(str, sym->t.simple.id);
      break;
    case T_SWITCH:
    case T_BOOL:
      strcat(str, sym->t.exp.id);
      break;
    }
    strcat(str, ":\n\n");
  }

  strcat(str, nohelp_text);

  strcat(str, "\n");
  help_add_symbol_name(str, sym);
  help_add_symbol_prompt(str, sym);
  help_add_symbol_deps(str, sym);

  dialog_clear();
  dialog_textbox(get_menu_prompt(sym), str, 0, 0);
  free(str);
}

static void show_loadcfg(void)
{
  int res;

  for (;;) {
    dialog_clear();
    res = dialog_inputbox(NULL, load_config_text, 11, 55, config_name);
    switch (res) {
    case 0:
      if (dialog_input_result[0] == '\0')
        return;
      if (!read_config(dialog_input_result)) {
        dialog_clear();
        dialog_textbox(NULL, "Can't load file", 5, 38);
        break;
      }
      set_config_filename(dialog_input_result);
      return;
    case 1:
      dialog_clear();
      dialog_textbox("Load Alternate Configuration", nohelp_text, 0, 0);
      break;
    case KEY_ESC:
      return;
    }
  }
}

static void show_savecfg(void)
{
  int res;

  for (;;) {
    dialog_clear();
    res = dialog_inputbox(NULL, save_config_text, 11, 55, config_name);
    switch (res) {
    case 0:
      if (dialog_input_result[0] == '\0')
        return;
      if (!write_config(dialog_input_result)) {
        dialog_clear();
        dialog_textbox(NULL, "Can't save file", 5, 38);
        break;
      }
      set_config_filename(dialog_input_result);
      return;
    case 1:
      dialog_clear();
      dialog_textbox("Save Alternate Configuration", nohelp_text, 0, 0);
      break;
    case KEY_ESC:
      return;
    }
  }
}

int main(int argc, char *argv[])
{
  int saved_x, saved_y;
  int res;

  if (argc < 2) {
    printf("USAGE: %s <sconfig file>\n", argv[0]);
    return 1;
  }

  parse_config(argv[1]);
  read_config(get_config_name());

  getyx(stdscr, saved_y, saved_x);
  if (init_dialog(NULL)) {
    fprintf(stderr, "Your display is too small to run Menuconfig!\n");
    fprintf(stderr, "It must be at least 80x19.\n");
    return 1;
  }

  set_config_filename(get_config_name());
  do {
    show_menu(get_root_symbol(), 0);
    dialog_clear();
    //if (conf_get_changed())
    res = dialog_yesno(NULL, "Do you wish to save your "
                       "new configuration?\n" "<ESC><ESC> to continue.", 6, 60);
    //  else
    //    res = -1;
  } while (res == KEY_ESC);
  end_dialog(saved_y, saved_x);

  switch (res) {
  case 0:
    if (!write_config(get_config_name())) {
      fprintf(stderr, "Error during writing of the configuration.\n"
              "Your configuration changes were NOT saved.\n\n");
      return 1;
    }
    if (!write_autoconf("include/autoconf.h")) {
      fprintf(stderr, "Error during writing of the autoconf.h file.\n"
              "Your configuration changes were NOT FULLY saved.\n\n");
      return 1;
    }
    /* fall through */
  case -1:
    printf("\n\n" "End of configuration.\n" "Execute 'make' now.\n\n");
    break;
  default:
    printf("\n\n" "Your configuration changes were NOT saved.\n\n");
  }

  return 0;
}
