/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 * (C) Copyright 2008 Dmitry Gromada
 *
 * hwavx: (hardware and virtual devices server)
 * src/hwinodeio.c: node operations
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ipc/port.h>
#include <ipc/rpc.h>
#include <ipc/channel.h>
#include <errno.h>
#include <time.h>
#include <datalib/usrtc.h>
#include <ns/ns.h>

#include "namespace.h"

static void __free_driverio(driverio_t *driverio)
{
  free(driverio->name);
  ipc_channel_close(driverio->service);
  ipc_channel_close(driverio->input);
  if (driverio->input != driverio->output)
    ipc_channel_close(driverio->output);

  free(driverio);
}

int initial_devices_set(usrtc_t *nodes,usrtc_t *dirs)
{
  iodir_t *dir=get_root_dir(dirs);
  int r;

  /* add devices */
  if(!dir)
    return -EINVAL;
  /* console */
  r=ionode_add(dir,nodes,"/console","console");
  r+=ionode_add(dir,nodes,"/kbd","kbd");
  r+=ionode_add(dir,nodes,"/pterm0","pterm0");
  r+=ionode_add(dir,nodes,"/pterm1","pterm1");
  r+=ionode_add(dir,nodes,"/pterm2","pterm2");
  r+=ionode_add(dir,nodes,"/pterm3","pterm3");
  r+=ionode_add(dir,nodes,"/pterm4","pterm4");
  r+=ionode_add(dir,nodes,"/pterm5","pterm5");
  r+=ionode_add(dir,nodes,"/pterm6","pterm6");
  r+=ionode_add(dir,nodes,"/pterm7","pterm7");
  r+=ionode_add(dir,nodes,"/pterm8","pterm8");
  r+=ionode_add(dir,nodes,"/pterm9","pterm9");
  if(r!=0)
    return -ENOMEM;

  return 0;
}

int ionode_add(iodir_t *dir,usrtc_t *nodes,
	       char *rawname,char *symbolic_name)
{
  ulong_t hash=ionode_hash(rawname);
  usrtc_node_t *node;
  usrtc_t *ionodes;
  ionode_t *ionode;

  /* first we're need lookup nodes for existing node with this name */
  node=usrtc_lookup(nodes,&hash);
  if(node)
    return EEXIST;

  ionodes=dir->ionodes;
  if(!ionodes)
    return ENOMEM;
  /* ionode initialize */
  node=malloc(sizeof(*node));
  ionode=malloc(sizeof(*ionode));
  if(!ionode || !node)
    return ENOMEM;

  ionode->symbolic_name=strdup(symbolic_name);
  if(!ionode->symbolic_name)
    return ENOMEM;
  
  ionode->hash=hash;
  ionode->ioops=NULL;
  ionode->cap=0;
  ionode->devno=0;

  usrtc_node_init(node,ionode);
  usrtc_insert(nodes,node,&ionode->hash);

  node=malloc(sizeof(*node));
  usrtc_node_init(node,ionode);
  usrtc_insert(ionodes,node,&ionode->hash);

  return 0;
}

int ionode_rm(iodir_t *dir, usrtc_t *nodes, char *rawname)
{
  usrtc_node_t *node;
  usrtc_t *ionodes;
  ionode_t *ionode;
  ulong_t hash;

  hash=ionode_hash(rawname);
  node=usrtc_lookup(nodes,&hash);
  if(node == NULL)
    return ENODEV; 

  ionodes=dir->ionodes;
  /* free inode resources */
  ionode=usrtc_node_getdata(node);
  free(ionode->symbolic_name);
  if (ionode->ioops != NULL) {
    __free_driverio(ionode->ioops);
  }
  usrtc_delete(nodes,node);
  free(node);
  node=usrtc_lookup(ionodes,&hash);
  usrtc_delete(ionodes,node);
  free(ionode);
  free(node);

  return 0; 
}

int ionode_create(const char *msg)
{
  usrtc_t *ionodes=nshw_ionodes();
  iodir_t *find_dir;
  ionode_t *dir_node;
  char *symb_name;
  char *raw_name=(char *)msg;
  int len=strlen(msg),i;

  if(!len)
    return EINVAL;

  if((find_dir=nshw_find_rootdir(msg))==NULL) {
    return EINVAL;
  }

  dir_node=find_dir->node;
  if(dir_node->ioops!=NULL) {
    return ENOSYS;
  } else {
    for (i=0;i<len;i++) {
      if (raw_name[i] == CCNSSEP) {
        symb_name = raw_name + (i+1)*sizeof(char);
      }
    }

    return ionode_add(find_dir,ionodes,raw_name,symb_name);
  }

  return 0;
}

int ionode_delete(const char *msg)
{
  usrtc_t *ionodes=nshw_ionodes();
  iodir_t *find_dir;
  ionode_t *dir_node;
  char *raw_name=(char *)msg;
  int len=strlen(msg);

  if(!len)
    return EINVAL;

  if((find_dir=nshw_find_rootdir(msg))==NULL) {
    return EINVAL;
  }

  dir_node=find_dir->node;
  if(dir_node->ioops!=NULL) {
    return ENOSYS;
  } else {
    return ionode_rm(find_dir,ionodes,raw_name);
  }

  return 0; 
}

ionode_t *ionode_nodefind(ulong_t hash,int *err)
{
  ionode_t *ionode;
  usrtc_node_t *node;
  usrtc_t *nodes=nshw_ionodes();
  usrtc_t *dirs=nshw_iodirs();

  node=usrtc_lookup(dirs,&hash);
  if(node!=NULL) {
    *err=EISDIR;
    return NULL;
  }
  node=usrtc_lookup(nodes,&hash);
  if(node) {
    ionode=usrtc_node_getdata(node);
    *err=0;
    return ionode;
  } else {
    *err=ENOENT;
    return NULL;
  }

  return NULL;
}

int ionode_register_driverio(usrtc_t *nodes,char *rawname,char *drvname,pid_t pid,
			     ulong_t cmd_id,ulong_t input_id,ulong_t output_id,
			     int cap,devid_t devno,chid_t service,chid_t input,
			     chid_t output)
{
  driverio_t *driverio;
  ionode_t *ionode;
  usrtc_node_t *node;
  ulong_t hash=ionode_hash(rawname);

  if(!rawname || !drvname)
    return EINVAL;

  node=usrtc_lookup(nodes,&hash);
  if(!node)
    return ENODEV;
  ionode=usrtc_node_getdata(node);
  if(!ionode)
    return ENODEV;

  if(ionode->ioops!=NULL)
    return EBUSY;
  /* fillfull other capaties */
  ionode->cap=cap;
  ionode->devno=devno;
  /* init driverio */
  driverio=malloc(sizeof(*driverio));
  if(!driverio)
    return ENOMEM;

  driverio->name=strdup(drvname);
  if(!driverio->name)
    return ENOMEM;
  driverio->pid=pid;
  driverio->cmd_id=cmd_id;
  driverio->input_id=input_id;
  driverio->output_id=output_id;
  driverio->service=service;
  driverio->input=input;
  driverio->output=output;
  ionode->ioops=driverio;

  return 0;
}

int ionode_unregister_driverio(usrtc_t *nodes, char *rawname)
{
  ionode_t *ionode;
  usrtc_node_t *node;
  ulong_t hash=ionode_hash(rawname);

  if(!rawname)
    return EINVAL;

  node=usrtc_lookup(nodes,&hash);
  if(!node)
    return ENODEV;
  
  ionode=usrtc_node_getdata(node);
  if(!ionode || !ionode->ioops)
    return ENODEV;

  /* free resources */
  __free_driverio(ionode->ioops);
  ionode->ioops = NULL;
  
  return 0; 
}

ulong_t ionode_hash(const char *name)
{
  ulong_t hash=0;

  if(!name)
    return 0;

  while(*name)
    hash=(hash << 5) + hash + *name++;

  return hash;
}
