/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 *
 * hwavx: (hardware and virtual devices server)
 * src/namespace.c: high level functions for operating with name space 
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ipc/port.h>
#include <ipc/rpc.h>
#include <ipc/channel.h>
#include <errno.h>
#include <time.h>
#include <datalib/usrtc.h>
#include <ns/ns.h>
#include <devio/devio.h>

#include "namespace.h"

static usrtc_t *nodes;
static usrtc_t *dirs;
static usrtc_t *pid_sessions;

static int __hash_compare(const void *a,const void *b);
static int __pid_compare(const void *a,const void *b);
static int __int_compare(const void *a,const void *b);

int init_namespace(void)
{
  nodes=malloc(sizeof(*nodes));
  dirs=malloc(sizeof(*dirs));

  if(!nodes || !dirs) 
    return -1;

  usrtc_init(nodes,USRTC_SPLAY,USRTC_COUNT_T_MAX,__hash_compare);
  usrtc_init(dirs,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__hash_compare);

  if(dir_root_create(nodes,dirs)!=0) {
    return -1;
  }

  if(initial_devices_set(nodes,dirs)!=0) {
    return -1;
  }

  pid_sessions=malloc(sizeof(*pid_sessions));
  if(!pid_sessions) 
    return -1;

  usrtc_init(pid_sessions,USRTC_SPLAY,USRTC_COUNT_T_MAX,__pid_compare);

  return 0;
}

usrtc_t *nshw_ionodes(void)
{
  return nodes;
}

usrtc_t *nshw_iodirs(void)
{
  return dirs;
}

usrtc_t *nshw_pidsessions(void)
{
  return pid_sessions;
}

iodir_t *nshw_find_rootdir(const char *hwdir_name)
{
  char *dir=(char *)hwdir_name;
  iodir_t *iodir=NULL;
  int len,i,last_dir;

  if(!dir)
    return NULL;

  len=strlen(hwdir_name);
  last_dir=-1;

  for(i=0;i<len;i++) {
    if(dir[i]==CCNSSEP)
      last_dir=i;
  }

  if(last_dir<0)
    return NULL; /* illegal namespace name */
  else if(!last_dir) { /* root dir */
    return get_root_dir(dirs);
  } else {
    dir[last_dir]='\0';
    iodir=nshw_dirfind(dirs,dir);
    dir[last_dir]=CCNSSEP;
  }

  return iodir;
}

int nshw_open_node(pid_t pid,const char *devn,int *fd,int flags)
{
  usrtc_node_t *node;
  ionode_t *ionode;
  pid_session_t *psession;
  hwfd_t *nfd;
  ulong_t hash=ionode_hash(devn);
  usrtc_t *fds,*dirfds;

  node=usrtc_lookup(nodes,&hash);
  if(!node)
    return ENOENT;

  ionode=usrtc_node_getdata(node);
  if(!ionode->ioops) 
    return ENODEV;

  node=usrtc_lookup(pid_sessions,&pid);
  if(!node) { /* this guy firstly do this - insert psession */
    node=malloc(sizeof(*node));
    psession=malloc(sizeof(*psession));
    if(!node || !psession)      return ENOMEM;
    /* init psession */
    psession->pid=pid;
    psession->lfd=psession->ldfd=0;
    fds=malloc(sizeof(*fds));
    dirfds=malloc(sizeof(*dirfds));
    if(!fds || !dirfds)      return ENOMEM;
    usrtc_init(fds,USRTC_SPLAY,USRTC_COUNT_T_MAX,__int_compare);
    usrtc_init(dirfds,USRTC_SPLAY,USRTC_COUNT_T_MAX,__int_compare);
    psession->hw_fds=fds;
    psession->hw_dirfds=dirfds;
    usrtc_node_init(node,psession); /* add session */
    usrtc_insert(pid_sessions,node,&psession->pid);
  } else {
    psession=usrtc_node_getdata(node);
  }

  /* create new fd entry */
  node=NULL;
  node=malloc(sizeof(*node));
  nfd=malloc(sizeof(*nfd));
  if(!node || !nfd)
    return ENOMEM;
  /* initiate new fd entry */
  nfd->fd=psession->lfd; /* TODO: operations on fd numbering */
  psession->lfd++;
  nfd->node=ionode;
  nfd->rd_set=nfd->wr_set=0;
  nfd->flags=flags;
  /* simply insert this */
  usrtc_node_init(node,nfd);
  usrtc_insert(psession->hw_fds,node,&nfd->fd);

  *fd=nfd->fd;

  return 0;
}

int nshw_close_node(pid_t pid,int fd)
{
  usrtc_node_t *node;
  pid_session_t *psession;
  hwfd_t *nfd;

  node=usrtc_lookup(pid_sessions,&pid);
  if(!node)
    return EINVAL;
  else
    psession=usrtc_node_getdata(node);
  node=usrtc_lookup(psession->hw_fds,&fd);
  if(!node)
    return EINVAL;
  else
    nfd=usrtc_node_getdata(node);

  usrtc_delete(psession->hw_fds,node);
  free(node);
  free(nfd);

  return 0;
}

int nshw_open_dir(pid_t pid,const char *dirn,int *fd)
{
  usrtc_node_t *node,*nnode;
  iodir_t *iodir;
  pid_session_t *psession;
  hwdirfd_t *ndirfd;
  ulong_t hash=ionode_hash(dirn);
  usrtc_t *fds,*dirfds;

  nnode=usrtc_lookup(nodes,&hash);

  node=usrtc_lookup(dirs,&hash);
  if(!nnode && !node)
    return ENOENT;
  if(nnode && !node)
    return EISDIR;

  iodir=usrtc_node_getdata(node);

  node=usrtc_lookup(pid_sessions,&pid);
  if(!node) { /* this guy firstly do this - insert psession */
    node=malloc(sizeof(*node));
    psession=malloc(sizeof(*psession));
    if(!node || !psession)      return ENOMEM;
    /* init psession */
    psession->pid=pid;
    psession->lfd=psession->ldfd=0;
    fds=malloc(sizeof(*fds));
    dirfds=malloc(sizeof(*dirfds));
    if(!fds || !dirfds)      return ENOMEM;
    usrtc_init(fds,USRTC_SPLAY,USRTC_COUNT_T_MAX,__int_compare);
    usrtc_init(dirfds,USRTC_SPLAY,USRTC_COUNT_T_MAX,__int_compare);
    psession->hw_fds=fds;
    psession->hw_dirfds=dirfds;
    usrtc_node_init(node,psession); /* add session */
    usrtc_insert(pid_sessions,node,&psession->pid);
  } else {
    psession=usrtc_node_getdata(node);
  }

  /* create new fd entry */
  node=NULL;
  node=malloc(sizeof(*node));
  ndirfd=malloc(sizeof(*ndirfd));
  if(!node || !ndirfd)
    return ENOMEM;
  ndirfd->dentry=malloc(sizeof(*ndirfd)+256);
  if(!ndirfd->dentry)
    return ENOMEM;
  /* initiate new dir fd entry */
  ndirfd->fd=psession->ldfd; /* TODO: operations on fd numbering */
  psession->ldfd++;
  ndirfd->iodir=iodir;
  ndirfd->current_point=usrtc_first(iodir->ionodes);
  memset(ndirfd->dentry->ename,'\0',256);
  ndirfd->dentry->name_len=0;
  ndirfd->dentry->ntype=0;
  /* simply insert this */
  usrtc_node_init(node,ndirfd);
  usrtc_insert(psession->hw_dirfds,node,&ndirfd->fd);

  *fd=ndirfd->fd;

  return 0;
}

int nshw_close_dir(pid_t pid,int fd)
{
  usrtc_node_t *node;
  pid_session_t *psession;
  hwdirfd_t *ndirfd;

  node=usrtc_lookup(pid_sessions,&pid);
  if(!node)
    return EINVAL;
  else
    psession=usrtc_node_getdata(node);
  node=usrtc_lookup(psession->hw_dirfds,&fd);
  if(!node)
    return EINVAL;
  else
    ndirfd=usrtc_node_getdata(node);

  usrtc_delete(psession->hw_dirfds,node);
  free(node);
  free(ndirfd->dentry);
  free(ndirfd);

  return 0;
}

size_t nshw_read(pid_t pid,int fd,size_t size) /* TODO: implement this */
{
  return ENOSYS;
}

size_t nshw_write(pid_t pid,int fd,size_t size) /* TODO: implement this */
{
  return ENOSYS;
}

int nshw_seek(pid_t pid,int fd,size_t size) /* TODO: implement this */
{
  return ENOSYS;
}

hw_dirent_t *nshw_readdir(pid_t pid,int fd,int *res)
{
  usrtc_node_t *node,*snode;
  pid_session_t *psession;
  ionode_t *ionode;
  hwdirfd_t *ndirfd;
  hw_dirent_t *dentry;
  int len=0;

  node=usrtc_lookup(pid_sessions,&pid);
  if(!node) {
    *res=EINVAL;
    return NULL;
  }
  else
    psession=usrtc_node_getdata(node);
  node=usrtc_lookup(psession->hw_dirfds,&fd);
  if(!node) {
    *res=EINVAL;
    return NULL;
  }
  else
    ndirfd=usrtc_node_getdata(node);

  if(usrtc_count(ndirfd->iodir->ionodes)==0) { /* there are no entries */
    dentry=ndirfd->dentry;
    dentry->ntype|=HWDIR;
    strcpy(dentry->ename,".");
    *res=HWEOD;
    return dentry;
  }

  /* okay let's work with it */
  ionode=usrtc_node_getdata(ndirfd->current_point);
  len=strlen(ionode->symbolic_name);
  dentry=ndirfd->dentry;
  dentry->ntype=0;
  strcpy(dentry->ename,ionode->symbolic_name);
  dentry->name_len=len;
  /* determine type */
  snode=usrtc_lookup(dirs,&ionode->hash);
  if(!snode)
    dentry->ntype|=HWNODE;
  else
    dentry->ntype|=HWDIR;
  if(ionode->ioops)
    dentry->ntype|=HWREGISTERED;

  if(ndirfd->current_point==usrtc_last(ndirfd->iodir->ionodes)) {
    *res=HWEOD;
  } else {
    *res=0;
    ndirfd->current_point=usrtc_next(ndirfd->iodir->ionodes,ndirfd->current_point);
  }

  return dentry;
}

static int __hash_compare(const void *a,const void *b)
{
  ulong_t *h1=(ulong_t*)a;
  ulong_t *h2=(ulong_t*)b;
  ulong_t hh1=*h1,hh2=*h2;

  return hh1-hh2;
}

static int __pid_compare(const void *a,const void *b)
{
  pid_t *h1=(pid_t*)a;
  pid_t *h2=(pid_t*)b;
  pid_t hh1=*h1,hh2=*h2;

  return hh1-hh2;
}

static int __int_compare(const void *a,const void *b)
{
  int *h1=(int*)a;
  int *h2=(int*)b;
  int hh1=*h1,hh2=*h2;

  return hh1-hh2;
}
