/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Nameserver V2
 *
 */

#include <sys/types.h>
#include <limits.h>
#include <uts.h>
#include <sys/mman.h>
#include <sys/proc.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <datalib/usrtc.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <pthread.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ns/ns.h>

#include "nsx.h"

#ifdef NSDEBUG
#include <dircons.h> 
#endif

#define PROT_THREADS  4

#define NSBUF  8192

const char *glob_name="nsv2.x";
static int glob_type=0;
static int glob_srv_type=0;
const char *glob_environ[4] = {
  "OS","Jari OS",
  "KERNEL","muistring",
};

/* special trees */
usrtc_t *t_list;    /* general task list */
usrtc_t *sys_list;  /* subsystems lists */

#ifdef CONFIG_NSENVFS
time_t glc_time; /* changed time of root directory */
envfs_node_t *fsroot; /* <-- points to directory */
#endif

/* IPC */
prtid_t prtsrv;
prtid_t prtevn;
static pthread_mutex_t tlist_lock;
static pthread_mutex_t syslist_lock;
static pthread_t pth_atom[PROT_THREADS];
/* environs */
static task_event_ctl_arg_t t_event;

/* some macros */
#define __global_pid_lock  pthread_mutex_lock(&tlist_lock)
#define __global_pid_unlock  pthread_mutex_unlock(&tlist_lock)
#define __global_type_lock  pthread_mutex_lock(&syslist_lock)
#define __global_type_unlock  pthread_mutex_unlock(&syslist_lock)

/* local functions proto */
static int __reg_task(pid_t pid,uid_t uid,gid_t gid,int type,uint8_t subsystem,
		      char *path,char **environ,int env_amount);
static int __clone_task(pid_t pid,pid_t npid);
static int __kill_task(pid_t pid);
static void __free_environ(v2task_t *task);
static pid_t __getsubsys(int type);
static int __regport(pid_t pid,char *name,int id);
static int __rmport(pid_t pid,char *name);
static int __chgpath_task(pid_t pid,char *path);
static int __srchport(pid_t pid,char *name);
static int __chngenv_task(pid_t pid,int num,char *nenv);
static int __setenv_task(pid_t pid,int num,char *nenv);
static int __task_setattrs(pid_t pid,int type,int subsys);
static int __setenvironment_task(pid_t pid,char **environ,int env_amount);
static int __getenvironment_task(pid_t pid,char *ebuf,int *rsz,uni_rpc_t *rpc);
static int __pidcmp(const void *a,const void *b);
static int __typecmp(const void *a,const void *b);
static int __prtcmp(const void *a,const void *b);
static void *ns_thread(void *a);
static char* __uname(uni_rpc_t *rpc);

int main(int argc,char **argv)
{
  int i=0,r;
  ipc_port_msg_info_t msgi;
  task_event_descr_t evn_desc;

#ifdef NSDEBUG
  open_console();
#endif

  /* allocate resources */
  if((t_list=malloc(sizeof(*t_list)))==NULL)     abort();
  if((sys_list=malloc(sizeof(*sys_list)))==NULL)     abort();
#ifdef CONFIG_NSENVFS
  if((fsroot=malloc(sizeof(*fsroot)))==NULL) abort();
#endif
  /* initialize data */
  glob_type=(V2T_TRUSTED | V2T_MKSERVICE);
  glob_srv_type=SYS_NS;
  
  /* open port */
  prtsrv=ipc_port_open(IPC_BLOCKED_ACCESS,512);
  if(prtsrv<0)     abort();

  /* well init event */
  prtevn=ipc_port_open(0,512);
  if(prtevn<0)     abort();
  t_event.ev_mask=0;
  t_event.ev_mask|=TASK_EVENT_TERMINATION;
  t_event.port=prtevn;

  /* initialize */
  usrtc_init(t_list,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__pidcmp);
  usrtc_init(sys_list,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__typecmp);
  /* locks */
  pthread_mutex_init(&tlist_lock,NULL);
  pthread_mutex_init(&syslist_lock,NULL);
  
  i=__reg_task(1,0,0,glob_type,glob_srv_type,(char *)glob_name,NULL,0);
  if(i!=0) abort();
  i=__setenvironment_task(1,(char **)glob_environ,4);
  if(i!=0) abort();
#ifdef NSDEBUG
  console_printf("[nsv2] Added self. All initial datas allocated.\n");
#endif

  for(i=0;i<PROT_THREADS;i++) {
    r=pthread_create(&pth_atom[i],NULL,ns_thread,NULL);
    if(r!=0)      abort();
  }

  while(1) {
    r=ipc_port_rcvmsg_mi_cs(prtevn,sizeof(task_event_descr_t),&evn_desc,&msgi,IPC_BLOCKED_ACCESS);
    if(r<0) continue;
    if(evn_desc.ev_mask & TASK_EVENT_TERMINATION) {
      __kill_task(evn_desc.pid);
    }
  }

  /* never reached */

  return 0;
}

/* main thread */
static void *ns_thread(void *a)
{
  int r,i,col=0;
  ipc_port_msg_info_t msgi;
  uni_rpc_t rpc;
  int msgid;
  pid_t c_pid,a_pid;
  uid_t c_uid;
  gid_t c_gid;
  uni_rpc_t *erpc;
  char *msg,*ee,*ty;
  char **env;
  iovec_t send[2];

  if((msg=malloc(NSBUF))==NULL)    return NULL;
  else memset(msg,0,NSBUF);
  if((ee=malloc(ARG_MAX*sizeof(uintptr_t)))==NULL)    return NULL;
  else memset(ee,0,ARG_MAX*sizeof(uintptr_t));

  while(1) {
    r=ipc_port_rcvmsg_mi(prtsrv,0,NULL,&msgi);
    if(r<0) continue;
    msgid=msgi.msg_id;
    c_pid=msgi.sender_pid;
    c_uid=msgi.sender_uid;
    c_gid=msgi.sender_gid;
    if(r>0 && r>=sizeof(uni_rpc_t)) {
      ipc_port_read_msg(prtsrv,msgid,0,&rpc,sizeof(rpc)); /* reads header */
      /* ok let's go deeper */
      switch(rpc.rpc_call) {
      case RPC_REGISTER:
	if(c_uid!=0) 
	  rpc.rpc_call=EPERM;
	else {
	  if(r+sizeof(uni_rpc_t)<rpc.arg5+sizeof(uni_rpc_t))
	    rpc.rpc_call=EINVAL;
	  else {
	    ipc_port_read_msg(prtsrv,msgid,sizeof(uni_rpc_t),msg,rpc.arg5); /* read name */
	    msg[rpc.arg5]='\0';
	    if(rpc.arg6>0) {
	      env=(char **)ee; col=0;
	      ty=msg+rpc.arg5+sizeof(char);
	      ipc_port_read_msg(prtsrv,msgid,sizeof(uni_rpc_t)+rpc.arg5+sizeof(char),ty,r-sizeof(uni_rpc_t)+rpc.arg5+sizeof(char));
	      if(rpc.arg6>ARG_MAX) rpc.arg6=ARG_MAX;
	      for(i=0;i<rpc.arg6;i++) {
		env[i]=ty+col;
		col+=sizeof(char)+strlen(env[i]);
		if(col>(r-sizeof(uni_rpc_t)+rpc.arg5+sizeof(char))) {
		  rpc.rpc_call=EINVAL;
		  goto __repl;
		}
	      }
	    } else env=NULL;
	    rpc.rpc_call=__reg_task(rpc.arg0/*pid*/,rpc.arg1/*uid*/,rpc.arg2/*gid*/,
				    rpc.arg3/*type*/,rpc.arg4/*subsystem*/,msg,env,
				    rpc.arg6/*env amount*/);
	  }
	}
      __repl:
	r=ipc_port_reply_nmsg(prtsrv,sizeof(rpc),&rpc,msgid);
	break;
      case RPC_SETENVIRON:
	if(c_uid!=0)  /* ignore arg0 */
	  a_pid=c_uid;
	else  /* privelenged call */
	  a_pid=rpc.arg0;
	if(rpc.arg6>0) {
	  env=(char **)ee; col=0;
	  ty=msg;
	  ipc_port_read_msg(prtsrv,msgid,sizeof(uni_rpc_t),ty,r-sizeof(uni_rpc_t));
	  if(rpc.arg6>ARG_MAX) rpc.arg6=ARG_MAX;
	  for(i=0;i<rpc.arg6;i++) {
	    env[i]=ty+col;
	    col+=sizeof(char)+strlen(env[i]);
	    if(col>(r-sizeof(uni_rpc_t)+rpc.arg5+sizeof(char))) {
	      rpc.rpc_call=EINVAL;
	      goto __repl;
	    }
	  }
	} else env=NULL;
	rpc.rpc_call=__setenvironment_task(a_pid,env,rpc.arg6);
	goto __repl;
	break;
      case RPC_GETENVIRON:
	col=0;
	erpc=(uni_rpc_t*)msg;
	erpc->rpc_call=__getenvironment_task(c_pid,msg+sizeof(uni_rpc_t),&col,erpc);
	r=ipc_port_reply_nmsg(prtsrv,col,msg,msgid);
	break;
      case RPC_UPDENV:
	if(rpc.arg1==0 || rpc.arg0>ARG_MAX) {
	  rpc.rpc_call=EINVAL;
	  goto __repl;
	}
	ipc_port_read_msg(prtsrv,msgid,sizeof(uni_rpc_t),msg,rpc.arg1);
	msg[rpc.arg1]='\0';
	rpc.rpc_call=__chngenv_task(c_pid,rpc.arg0,msg);
	goto __repl;
	break;
      case RPC_SETENV:
	if(rpc.arg1==0) {
	  rpc.rpc_call=EINVAL;
	  goto __repl;
	}
	ipc_port_read_msg(prtsrv,msgid,sizeof(uni_rpc_t),msg,rpc.arg1);
	msg[rpc.arg1]='\0';
	rpc.rpc_call=__setenv_task(c_pid,rpc.arg0,msg);
	goto __repl;
	break;
      case RPC_INFO:
	col=rpc.arg0;
	r=__getsubsys(col);
	if(!r) {
	  rpc.rpc_call=errno;
	  rpc.arg0=0;
	} else {
	  rpc.rpc_call=0;
	  rpc.arg0=r;
	}
	r=ipc_port_reply_nmsg(prtsrv,sizeof(rpc),&rpc,msgid);
	break;
      case RPC_IOCTL:
	switch(rpc.arg0) {
	case NX_REGPORT:
	  if(r<sizeof(uni_rpc_t)+rpc.arg1)
	    rpc.rpc_call=EINVAL;
	  else {
	    ipc_port_read_msg(prtsrv,msgid,sizeof(rpc),msg,rpc.arg1); /* reads msg */
	    msg[rpc.arg1]='\0';
	    rpc.rpc_call=__regport(c_pid,msg,rpc.arg2/*id*/);
	  }
	  break;
	case NX_RMPORT:
	  if(r<sizeof(uni_rpc_t)+rpc.arg1)
	    rpc.rpc_call=EINVAL;
	  else {
	    ipc_port_read_msg(prtsrv,msgid,sizeof(rpc),msg,rpc.arg1); /* reads msg */
	    msg[rpc.arg1]='\0';
	    rpc.rpc_call=__rmport(c_pid,msg);
	  }
	  break;
	case NX_SETATTR:
	  if(c_uid!=0)
	    rpc.rpc_call=EPERM;
	  else {
	    rpc.rpc_call=__task_setattrs(c_pid,rpc.arg1,rpc.arg2);
	  }
	  break;
	case NX_SRCHPORT:
	  if(r<sizeof(uni_rpc_t)+rpc.arg1)
	    rpc.rpc_call=EINVAL;
	  else {
	    ipc_port_read_msg(prtsrv,msgid,sizeof(rpc),msg,rpc.arg1); /* reads msg */
	    msg[rpc.arg1]='\0';
	    r=__srchport(rpc.arg2/*pid*/,msg);
	    if(r<0) rpc.rpc_call=EIO;
	    else {
	      rpc.rpc_call=0;
	      rpc.arg0=r;
	    }
	  }
	  break;
	case NX_CHGPATH: /* this call comes when we're executing something */
	  if(c_pid!=0)
	    rpc.rpc_call=EPERM;
	  else {
	    ipc_port_read_msg(prtsrv,msgid,sizeof(rpc),msg,rpc.arg1); /* reads msg */
	    msg[rpc.arg1]='\0';
	    r=__chgpath_task(rpc.arg2/*pid*/,msg);
	  }
	  break;
	case NX_CLONE:
	  rpc.rpc_call=__clone_task(rpc.arg1,rpc.arg2);
	  break;
	case NX_ENVFSUP:
#ifdef CONFIG_NSENVFS
	  rpc.rpc_call=0; /* TODO: done this */
#else
	  rpc.rpc_call=ENOSYS;
#endif
	  break;
	default:
	  rpc.rpc_call=EINVAL;
	  break;
	}
	r=ipc_port_reply_nmsg(prtsrv,sizeof(rpc),&rpc,msgid);
	break;
      case RPC_UNAME:
	ee = __uname(&rpc);
	if (ee == NULL)
	  r = ipc_port_reply_nmsg(prtsrv, sizeof(uni_rpc_t), &rpc, msgid);
	else {
          send[0].iov_base = &rpc;
	  send[0].iov_len = sizeof(uni_rpc_t);
	  send[1].iov_base = ee;
	  send[1].iov_len = rpc.size;
	  r = ipc_port_reply_nmsg_iov(prtsrv, send, 2, msgid);
	}
	break;
      default:
	rpc.rpc_call=EINVAL;
	r=ipc_port_reply_nmsg(prtsrv,sizeof(rpc),&rpc,msgid);
	break;
      }
    } else {
      rpc.rpc_call=EINVAL;
      r=ipc_port_reply_nmsg(prtsrv,sizeof(rpc),&rpc,msgid);
    }
  }

  return NULL;
}

/* local functions */
static int __regport(pid_t pid,char *name,int id)
{
  usrtc_node_t *node;
  v2task_t *task;
  v2regport_t *nport;
  int err=0;

  __global_pid_lock;
  if((node=usrtc_lookup(t_list,&pid))==NULL) {
    __global_pid_unlock;
    return ENOENT;
  } else {
    task=usrtc_node_getdata(node);
    v2task_lock(task);
  }
  __global_pid_unlock;

  if(task->ports!=NULL) {
    node=usrtc_lookup(task->ports,name);
    if(node) { 
      err=EEXIST;
      goto __exit;
    }
  } else {
    /* allocate and initialize */
    if((task->ports=malloc(sizeof(*task->ports)))==NULL) {
      err=ENOMEM;
      goto __exit;
    } else usrtc_init(task->ports,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__prtcmp);
  }
  /* try to alloc new port structure */
  if((nport=malloc(sizeof(*nport)))==NULL) {
    err=ENOMEM;
    goto __exit;
  }
  if((nport->node=malloc(sizeof(*nport->node)))==NULL) {
    free(nport);
    err=ENOMEM;
    goto __exit;
  }
  if((nport->name=strdup(name))==NULL) {
    free(nport->node);
    free(nport);
    err=ENOMEM;
    goto __exit;
  }
  /* initialize structure */
  nport->id=id;
  nport->parent=task;
  usrtc_node_init(nport->node,nport);
  /* insert */
  usrtc_insert(task->ports,nport->node,nport->name);
  err=0;

 __exit:
  v2task_unlock(task);

  return err;
}

static int __rmport(pid_t pid,char *name)
{
  usrtc_node_t *node;
  v2task_t *task;
  v2regport_t *nport;
  int err=0;

  __global_pid_lock;
  if((node=usrtc_lookup(t_list,&pid))==NULL) {
    __global_pid_unlock;
    return ENOENT;
  } else {
    task=usrtc_node_getdata(node);
    v2task_lock(task);
  }
  __global_pid_unlock;

  if(task->ports!=NULL) {
    node=usrtc_lookup(task->ports,name);
    if(!node) { 
      err=ENOENT;
      goto __exit;
    }
    nport=usrtc_node_getdata(node);
    usrtc_delete(task->ports,node);
    free(nport->name);
    free(nport->node);
    free(nport);
    err=0;
  } else err=ENOENT;
  
 __exit:
  v2task_unlock(task);

  return err;
}

static int __chgpath_task(pid_t pid,char *path)
{
  usrtc_node_t *node;
  v2task_t *task;

  __global_pid_lock;
  if((node=usrtc_lookup(t_list,&pid))==NULL) {
    __global_pid_unlock;
    return ENOENT;
  } else {
    task=usrtc_node_getdata(node);
    v2task_lock(task);
  }
  __global_pid_unlock;

  task->type=0; /* we're reset this on exec */

  if(task->path) free(task->path);
  if((task->path=strdup(path))==NULL) {
    v2task_unlock(task);
    return ENOMEM;
  }

  v2task_unlock(task);

  return 0;
}

static int __task_setattrs(pid_t pid,int type,int subsys)
{
  usrtc_node_t *node;
  v2task_t *task;

  __global_pid_lock;
  if((node=usrtc_lookup(t_list,&pid))==NULL) {
    __global_pid_unlock;
    return ENOENT;
  } else {
    task=usrtc_node_getdata(node);
    v2task_lock(task);
  }
  __global_pid_unlock;

  task->type=0; /* we're reset this on exec */
  task->subsystem=0;

  task->type=type; /* we're reset this on exec */
  task->subsystem=subsys;
  if(task->subsystem!=SYS_NONE) {
    /* let's add this */
    node=NULL;
    __global_type_lock;
    node=usrtc_lookup(sys_list,&task->subsystem);
    if(node) {
      __global_type_unlock;
      v2task_unlock(task);
      return EEXIST;
    }
    if((node=malloc(sizeof(*node)))==NULL) {
      __global_type_unlock;
      v2task_unlock(task);
      return ENOMEM;
    }
    /* make a deal */
    usrtc_node_init(node,task);
    usrtc_insert(sys_list,node,&task->subsystem);
    __global_type_unlock;
  }

  v2task_unlock(task);

  return 0;
}

static int __srchport(pid_t pid,char *name)
{
  usrtc_node_t *node;
  v2task_t *task;
  v2regport_t *nport;
  int err=0;

  __global_pid_lock;
  if((node=usrtc_lookup(t_list,&pid))==NULL) {
    __global_pid_unlock;
    return -1;
  } else {
    task=usrtc_node_getdata(node);
    v2task_lock(task);
  }
  __global_pid_unlock;

  if(task->ports!=NULL) {
    node=usrtc_lookup(task->ports,name);
    if(!node) { 
      err=-1;
      goto __exit;
    }
    nport=usrtc_node_getdata(node);
    err=nport->id;
  } else {
    err=-1;
  }
  
 __exit:
  v2task_unlock(task);

  return err;
}

static pid_t __getsubsys(int type)
{
  usrtc_node_t *node;
  v2task_t *task;

  __global_type_lock;
  if((node=usrtc_lookup(sys_list,&type))==NULL) {
    __global_type_unlock;
    errno=ENOENT;
    return 0;
  } else {
    task=usrtc_node_getdata(node);
  }
  __global_type_unlock;

  return task->pid;
}

static int __setenv_task(pid_t pid,int num,char *nenv)
{
  int i,err=0;
  usrtc_node_t *node;
  v2task_t *task;

  __global_pid_lock;
  if((node=usrtc_lookup(t_list,&pid))==NULL) {
    __global_pid_unlock;
    return ENOENT;
  } else {
    task=usrtc_node_getdata(node);
    v2task_lock(task);
  }
  __global_pid_unlock;

  i=num*2;
  if(task->environ[i]) {
    v2task_unlock(task);
    return EEXIST;
  }

  task->environ[i]=strdup(nenv);
  if(!task->environ[i]) {
    err=ENOMEM;
    goto __endenvset;
  }
  nenv+=(strlen(task->environ[i])+sizeof(char));
  task->environ[i+1]=strdup(nenv);
  if(!task->environ[i+1]) {
    err=ENOMEM;
    goto __endenvset;
  }
  task->env_amount+=2;

 __endenvset:
  v2task_unlock(task);

  return err;
}

static int __chngenv_task(pid_t pid,int num,char *nenv)
{
  int i;
  usrtc_node_t *node;
  v2task_t *task;

  __global_pid_lock;
  if((node=usrtc_lookup(t_list,&pid))==NULL) {
    __global_pid_unlock;
    return ENOENT;
  } else {
    task=usrtc_node_getdata(node);
    v2task_lock(task);
  }
  __global_pid_unlock;

  if(!task->environ) {
    v2task_unlock(task);
    return EINVAL;
  }
  i=num*2;
  if(!task->environ[i]) {
    v2task_unlock(task);
    return ENOENT;
  }
  if(task->environ[i+1]) free(task->environ[i+1]);
  task->environ[i+1]=strdup(nenv);

  v2task_unlock(task);

  return 0;
}

static int __getenvironment_task(pid_t pid,char *ebuf,int *rsz,uni_rpc_t *rpc)
{
  int i,off=0,rs=0;
  usrtc_node_t *node;
  v2task_t *task;

  __global_pid_lock;
  if((node=usrtc_lookup(t_list,&pid))==NULL) {
    __global_pid_unlock;
    *rsz=sizeof(uni_rpc_t);
    rpc->size=0;
    return ENOENT;
  } else {
    task=usrtc_node_getdata(node);
    v2task_lock(task);
  }
  __global_pid_unlock;

  if(!task->env_amount) {
    *rsz=sizeof(uni_rpc_t);
    rpc->size=0;
  } else {
    rpc->size=task->env_amount;
    for(i=0;i<task->env_amount;i++) {
      if(task->environ[i]==NULL) continue;
      strcpy(ebuf+off,task->environ[i]);
      ebuf[off+strlen(task->environ[i])]='\0';
      off+=(strlen(task->environ[i])+sizeof(char));
    }
    rs=off+sizeof(uni_rpc_t);
    *rsz=rs;
  }

  v2task_unlock(task);

  return 0;
}

static int __setenvironment_task(pid_t pid,char **environ,int env_amount)
{
  int i;
  usrtc_node_t *node;
  v2task_t *task;

  __global_pid_lock;
  if((node=usrtc_lookup(t_list,&pid))==NULL) {
    __global_pid_unlock;
    return ENOENT;
  } else {
    task=usrtc_node_getdata(node);
    v2task_lock(task);
  }
  __global_pid_unlock;

  __free_environ(task);
  if(!env_amount && !environ) { /* it's mean that we will reset all */
    task->env_amount=0;
    if((task->environ=malloc(ARG_MAX*sizeof(uintptr_t)))==NULL) {
    __enomem:
      v2task_unlock(task);
      return ENOMEM;
    } else
      memset(task->environ,0,sizeof(uintptr_t)*ARG_MAX);
#ifdef CONFIG_NSENVFS
    /* TODO: Add/update envfs entry */
#endif /* CONFIG_NSENVFS */
  } else {
    task->env_amount=env_amount;
    for(i=0;i<task->env_amount;i++) 
      if((task->environ[i]=strdup(environ[i]))==NULL) 	goto __enomem;
#ifdef CONFIG_NSENVFS
    /* there should be some changes in envfs <- TODO: */
#endif /* CONFIG_NSENVFS */
  }

  v2task_unlock(task);

  return 0;
}

static void __free_environ(v2task_t *task)
{
  int i;

  if(!task->environ)    return;

  for(i=0;i<task->env_amount;i++)     if(task->environ[i]) free(task->environ[i]);
  free(task->environ);

  return;
}

static int __kill_task(pid_t pid)
{
  int i,p;
  v2task_t *parent;
  v2regport_t *port;
  usrtc_node_t *node;

  __global_pid_lock;
  node=usrtc_lookup(t_list,&pid);
  if(!node) {
    __global_pid_unlock;
    return ENOENT;
  } else {
    parent=usrtc_node_getdata(node);
    v2task_lock(parent);
    usrtc_delete(t_list,node);
    __global_pid_unlock;
  }

  /* free environ */
  __free_environ(parent);
  /* free node */
  free(parent->node);
  if(parent->path) free(parent->path);
  /* free ports */
  if(parent->ports) {
    if((p=usrtc_count(parent->ports))>0) {
      for(i=0;i<p;i++) {
	node=usrtc_first(parent->ports);
	port=usrtc_node_getdata(node);
	usrtc_delete(parent->ports,node);
	/* free port */
	free(port->node);
	free(port->name);
	free(port);
      }
    }
    free(parent->ports);
  }
  v2task_unlock(parent);
  pthread_mutex_destroy(&parent->lock);
  free(parent);

  return 0;
}

static int __clone_task(pid_t pid,pid_t npid)
{
  int i;
  v2task_t *parent,*child;
  usrtc_node_t *node;

  __global_pid_lock;
  node=usrtc_lookup(t_list,&pid);
  __global_pid_unlock;
  if(!node) return ENOENT;
  else parent=usrtc_node_getdata(node);

  v2task_lock(parent);

  /* try to allocate */
  if((child=malloc(sizeof(*child)))==NULL) {
    v2task_unlock(parent);
    return ENOMEM;
  }
  if((child->path=strdup(parent->path))==NULL) {
    v2task_unlock(parent);
    free(child);
    return ENOMEM;
  }
  if((child->environ=malloc(ARG_MAX*sizeof(uintptr_t)))==NULL) {
    free(child->path);
    free(child);
    v2task_unlock(parent);
    return ENOMEM;
  } else
    memset(child->environ,0,ARG_MAX*sizeof(uintptr_t));
  if((child->node=malloc(sizeof(*child->node)))==NULL) {
  __enomem:
    v2task_unlock(parent);
    if(child->environ) free(child->environ);
    free(child->path);
    free(child);
    return ENOMEM;
  }

#ifdef CONFIG_NSENVFS
  if((child->fsnod=malloc(sizeof(*child->fsnod)))==NULL) {
    free(child->node);
    goto __enomem;
  }
  child->c_time=time(NULL);
  if(child->c_time==(time_t)-1)    task->c_time=0;
#endif
  /* okay, copying environ */
  child->env_amount=parent->env_amount;
  if(child->env_amount>0) {
    for(i=0;i<child->env_amount;i++) {
      child->environ[i]=strdup(parent->environ[i]);
      //      if(!parent->environ[i]) abort();
      if(child->environ[i]==NULL) {
	goto __enomem;
      }
    }
#ifdef CONFIG_NSENVFS
    /* there should be some changes in envfs <- TODO: */
#endif /* CONFIG_NSENVFS */
  }

  /* initialize other data */
  child->ports=NULL;
  child->pid=npid;
  child->uid=parent->uid;
  child->gid=parent->gid;
  child->type=child->type;
  child->subsystem=SYS_NONE; /* always */
  pthread_mutex_init(&child->lock,NULL);
#ifdef CONFIG_NSENVFS
  /* there should be some changes in envfs <- TODO: */
#endif /* CONFIG_NSENVFS */
  /* node initialization */
  usrtc_node_init(child->node,child);
  __global_pid_lock;
  usrtc_insert(t_list,child->node,&child->pid);
  task_ctl(child->pid,0,SYS_PR_CTL_ADD_EVENT_LISTENER,(uint64_t)&t_event);
  __global_pid_unlock;

  v2task_unlock(parent);

  return 0;
}

/* NOTE: this functions called from RPC handle, i.e. it should copying all data */
static int __reg_task(pid_t pid,uid_t uid,gid_t gid,int type,uint8_t subsystem,
		      char *path,char **environ,int env_amount)
{
  int i;
  v2task_t *task;
  usrtc_node_t *node;

  __global_pid_lock;
  node=usrtc_lookup(t_list,&pid);
  __global_pid_unlock;
  if(node) {
    return EEXIST;
  }

  if(subsystem) {
    __global_type_lock;
    node=usrtc_lookup(sys_list,&subsystem);
    __global_type_unlock;
    if(node) {
      return EEXIST;
    }
    if((node=malloc(sizeof(*node)))==NULL) return ENOMEM;
  }

  if((task=malloc(sizeof(*task)))==NULL) return ENOMEM;
  if((task->path=strdup(path))==NULL) {
    free(task);
    return ENOMEM;
  }
  if((task->environ=malloc(ARG_MAX*sizeof(uintptr_t)))==NULL) {
    free(task->path);
    free(task);
    return ENOMEM;
  } else
    memset(task->environ,0,sizeof(uintptr_t)*ARG_MAX);
  if(env_amount>0) {
    task->env_amount=env_amount;
  } else {
    task->env_amount=0;
  }
  if((task->node=malloc(sizeof(*task->node)))==NULL) {
  __enomem:
    if(task->environ) free(task->environ);
    free(task->path);
    free(task);
    return ENOMEM;
  }

#ifdef CONFIG_NSENVFS
  if((task->fsnod=malloc(sizeof(*task->fsnod)))==NULL) {
    free(task->node);
    goto __enomem;
  }
  task->c_time=time(NULL);
  if(task->c_time==(time_t)-1)    task->c_time=0;
#endif

  /* okay, copying environ */
  if(task->env_amount>0) {
    for(i=0;i<task->env_amount;i++) 
      if((task->environ[i]=strdup(environ[i]))==NULL) 	goto __enomem;
#ifdef CONFIG_NSENVFS
    /* there should be some changes in envfs <- TODO: */
#endif /* CONFIG_NSENVFS */
  }

  /* initialize other data */
  task->ports=NULL;
  task->pid=pid;
  task->uid=uid;
  task->gid=gid;
  task->type=type;
  task->subsystem=subsystem;
  pthread_mutex_init(&task->lock,NULL);
#ifdef CONFIG_NSENVFS
  /* there should be some changes in envfs <- TODO: */
#endif /* CONFIG_NSENVFS */
  /* node initialization */
  usrtc_node_init(task->node,task);
  if(subsystem) {
    usrtc_node_init(node,task);
    __global_type_lock;
    usrtc_insert(sys_list,node,&task->subsystem);
    __global_type_unlock;
  }
  __global_pid_lock;
  usrtc_insert(t_list,task->node,&task->pid);
  task_ctl(task->pid,0,SYS_PR_CTL_ADD_EVENT_LISTENER,(uint64_t)&t_event);
  __global_pid_unlock;

  return 0;
}

/* trees comparation */
static int __pidcmp(const void *a,const void *b)
{
  return *(pid_t*)a-*(pid_t*)b;
}

static int __typecmp(const void *a,const void *b)
{
  return *(uint8_t*)a-*(uint8_t*)b;
}

static int __prtcmp(const void *a,const void *b)
{
  return strcmp((char*)a,(char*)b);
}

static char* __uname(uni_rpc_t *rpc)
{
  static char buf[NX_BUFSIZE];
  size_t len;
  char *p;

  len  = strlen(UTS_SYSNAME) + 1;
  len += strlen(UTS_NODENAME) + 1;
  len += strlen(UTS_RELEASE) + 1;
  len += strlen(UTS_VERSION) + 1;
  len += strlen(UTS_MACHINE) + 1;

  if (rpc->size < len) {
    rpc->size = len;
    rpc->rpc_call = EOVERFLOW;
    return NULL;
  }

  memset(buf, 0, NX_BUFSIZE);
  p = buf;

  strcpy(p, UTS_SYSNAME);
  p += strlen(UTS_SYSNAME) + 1;
  strcpy(p, UTS_NODENAME);
  p += strlen(UTS_NODENAME) + 1;
  strcpy(p, UTS_RELEASE);
  p += strlen(UTS_RELEASE) + 1;
  strcpy(p, UTS_VERSION);
  p += strlen(UTS_VERSION) + 1;
  strcpy(p, UTS_MACHINE);
  p += strlen(UTS_MACHINE) + 1;

  rpc->size = len;
  rpc->rpc_call = 0;
  return buf;
}
