/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 *
 * termio.x:
 * keyb.c: physical input ops
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <wchar.h>
#include <ns/ns.h>
#include <sys/scheduler.h>
#include <sys/keyboard.h>
#include <chardev.h>
#include <datastructs.h>
#include <datalib/list.h>
#include <sys/kcontrol.h>

#include "keyb.h"
#include "debug.h"
#include "ptm.h"
#include "termio.h"
#include "en_layout.h"
#include "ru_layout.h"

#define KBDBUF_SIZE 128

int kbdfd;
char *kbdbuf;
int kbuf_pos=0;
int comb_flags, state_flags;
int ptmbuf_pos;

spinlock_t freekbdlock;
LIST_DEFINE(freekbdbuf);

wchar_t *ptm_buf;

layout_fn_t layout_fn[2] = {en_scan_to_ascii, ru_scan_to_ascii};
int layout_nr = 2;

extern ptm_t **ptms;
extern int current_ptm;
static void prealloc_keyseqs(int num);

void init_keyb(void)
{
  int r;
  char tmp[256];
  uint8_t rate;

  ptm_print("[termio] allocating keyboard buffer\n");

  kbdbuf = malloc(KBDBUF_SIZE);
  if(!kbdbuf) {
    ptm_print("[termio] buffer allocating error.\n");
    abort();
  }

  ptm_buf = malloc(128 * sizeof(wchar_t));
  if(!ptm_buf) {
    ptm_print("[termio] failed to allocate ptm stdin buffer.\n");
    abort();
  } else {
    memset(ptm_buf, 0, 128 * sizeof(wchar_t));
  }

  kbdfd = open("/dev/kbd", O_RDWR);
  if(kbdfd < 0) {
    sprintf(tmp, "[termio] opening (/dev/)kbd failed, errno = %d.\n", errno);
    ptm_print(tmp);
    abort();
  }

  prealloc_keyseqs(100);
  
  rate = TM_RATE(DEF_TM_RATE) | TM_DELAY(DEF_TM_DELAY);

  r = ioctl(kbdfd, KIOC_SETLEDS, &rate);
  
  if(r != 0) {
    sprintf(tmp, "[termio] failed to set typematic options (%d)\n", r);
    ptm_print(tmp);
  }

  kbuf_pos = 0;
  comb_flags = 0;
  state_flags = 0;
  ptmbuf_pos = 0;

  return;
}

keyseq_t *get_keyseq(void)
{
  keyseq_t *res;

  spinlock_lock(&freekbdlock);

  if(list_is_empty(&freekbdbuf)) {
    res = malloc(sizeof(keyseq_t));
  } else {
    res = list_first_entry(&freekbdbuf, keyseq_t, list);
    list_delfromhead(&freekbdbuf);
  }

  spinlock_unlock(&freekbdlock);
  return res;
}

void put_keyseq(keyseq_t *seq)
{
  spinlock_lock(&freekbdlock);
  list_add2tail(&freekbdbuf, &seq->list);
  spinlock_unlock(&freekbdlock);
}

static void prealloc_keyseqs(int num)
{
  keyseq_t *seq;
  while(num-- > 0) {
    seq = malloc(sizeof(*seq));
    if(!seq) {
      break;
    }
    put_keyseq(seq);
  }
}

keyseq_t *first_keyseq(ptm_t *ptm)
{
  if(list_is_empty(&ptm->kbdbuf)) {
    return NULL;
  } else {
    return list_first_entry(&ptm->kbdbuf, keyseq_t, list);
  }
}

keyseq_t *last_keyseq(ptm_t *ptm)
{
  if(list_is_empty(&ptm->kbdbuf)) {
    return NULL;
  } else {
    return list_last_entry(&ptm->kbdbuf, keyseq_t, list);
  }
}

void pop_keyseq(keyseq_t *seq)
{
  list_del(&seq->list);
  put_keyseq(seq);
}

void push_keyseq(ptm_t *ptm, keyseq_t *seq)
{
  list_add2tail(&ptm->kbdbuf, &seq->list);
}

void flush_kbdbuf(ptm_t *ptm)
{
  keyseq_t *seq;
  while((seq = first_keyseq(ptm)) != NULL) {
    pop_keyseq(seq);
  }
}

void setleds(void)
{
  uint8_t leds = state_flags & (CAPS_MASK | SCROLL_MASK | NUM_MASK);
  int r;

  r = ioctl(kbdfd, KIOC_SETLEDS, &leds);
}

const char bsbuf[] = "\b\b\b\b\b\b\b\b\b\b\b\b";

// check that only flags from 'y' present in 'x'
// (at least one)
#define EXCL(x,y) (((x) & (y)) && (!((x) & ~(y))))

void *keyb_thread(void *data)
{
  keyseq_t *seq;
  int r,i,a,acp;
  uint8_t *rbuf;
  int s, n;

  task_scheduler_control(0,0,SYS_SCHED_CTL_SET_PRIORITY,28);
  while(1) {
    s = 0; n = 0;
    r = read(kbdfd, kbdbuf, 128);
    kbuf_pos = r;
    rbuf = (unsigned char *)kbdbuf;
    a=0;
    for(i=0;i<kbuf_pos;i++) {
      switch(rbuf[i]) {
        // Shifts
        case KC_LSHIFT:
          comb_flags |= LSHIFT_MASK;
          if(comb_flags & LCTRL_MASK) {
            ADDMOD(ptms[current_ptm]->kbd_layout, 1, layout_nr);
          }
          break;
        case KC_RSHIFT:
          comb_flags |= RSHIFT_MASK;
          break;
        case (KC_LSHIFT | RELEASE_MASK):
          comb_flags &= ~LSHIFT_MASK;
          break;          
        case (KC_RSHIFT | RELEASE_MASK):
          comb_flags &= ~RSHIFT_MASK;
          break;
          
        // Alts
        case KC_LALT:
          comb_flags |= LALT_MASK;
          break;
        case (KC_LALT | RELEASE_MASK):
          comb_flags &= ~LALT_MASK;
          break;
        case KC_RALT:
          comb_flags |= RALT_MASK;
          break;
        case (KC_RALT | RELEASE_MASK):
          comb_flags &= ~RALT_MASK;
          break;

        // Ctrls
        case KC_LCTRL:
          comb_flags |= LCTRL_MASK;
          if(comb_flags & LSHIFT_MASK) {
            ADDMOD(ptms[current_ptm]->kbd_layout, 1, layout_nr);
          }
          break;
        case (KC_LCTRL | RELEASE_MASK):
          comb_flags &= ~LCTRL_MASK;
          break;
        case KC_RCTRL:
          comb_flags |= RCTRL_MASK;
          break;
        case (KC_RCTRL | RELEASE_MASK):
          comb_flags &= ~RCTRL_MASK;
          break;


          // Locks  
        case KC_NUMLOCK:
          state_flags ^= NUM_MASK;
          setleds();
          break;
        case KC_CAPSLOCK:
          state_flags ^= CAPS_MASK;
          setleds();
          break;
        case KC_SCROLLLOCK:
          state_flags ^= SCROLL_MASK;
          setleds();
          break;          

        default:
          if(EXCL(comb_flags, SHIFT_MASK)) {
            // Shift-PgUp & Shift-PgDn scrolling. 
            if(rbuf[i] == KC_PGUP) {
              ptm_scroll(SCROLL_REL, 12);
            } else if(rbuf[i] == KC_PGDN){
              ptm_scroll(SCROLL_REL, -12);
            }
            break;
          }
          if(comb_flags == 0 && state_flags & SCROLL_MASK) {
            // SCROLL LOCK, comfort scrolling:
            if(rbuf[i] == KC_UP) {
              ptm_scroll(SCROLL_REL, 1); //Up - up one line
            } else if(rbuf[i] == KC_DOWN){
              ptm_scroll(SCROLL_REL, -1); //Down - down one line
            } else if(rbuf[i] == KC_PGUP){
              ptm_scroll(SCROLL_REL | SCROLL_PAGES, 1); //PgUp - up one page
            } else if(rbuf[i] == KC_PGDN){
              ptm_scroll(SCROLL_REL | SCROLL_PAGES, -1); //PgDown - down one page
            } else if(rbuf[i] == KC_HOME){
              ptm_scroll(SCROLL_TOP, 0);  //Home - up to top
            } else if(rbuf[i] == KC_END){
              ptm_scroll(SCROLL_BTM, 0);  //End - down to writing position
            }
            break;
          }
          if(EXCL(comb_flags, ALT_MASK)) {
            // Codes for ALT-[F1..F10], switch active ptm
            // This hack will not work with F11 and F12 as their codes aren't next to F10
            if(rbuf[i] >= KC_F1 && rbuf[i] <= KC_F10) {
              acp=rbuf[i] - KC_F1;
              ptm_switch_active(acp);
              break;
            }
          }
      }

      // Serve character. Maybe translate to ascii, maybe echo to the terminal.
      // TODO: this code is terrible and error-prone, change of keyboard buffer
      // data structure is needed.
      seq = get_keyseq();
      if(!seq) {
        klog("FATAL: no memory for keyboard buffer");
      }
      ptm_kbdbuf_lock(ptms[current_ptm]);
      if(ptms[current_ptm]->kbd_mode & TC_XLATE) {
        ptm_t *ptm = ptms[current_ptm];
        if(!(rbuf[i] & RELEASE_MASK)) {
          int n = 0;
          n = layout_fn[ptm->kbd_layout](seq->data, rbuf[i], comb_flags, state_flags);
          if(!n) {
            n = fb_scan_to_ascii(seq->data, rbuf[i], comb_flags, state_flags);
          }
          if(n) {
            // check for bacspace
            if(n == 1 && *seq->data == '\b') {
              // get last output from keyboard buffer and
              // remove corresponding symbol from ptm.
              
              // put seq back
              put_keyseq(seq);
              seq = last_keyseq(ptm);
              if(!seq || seq->data[0] == '\n') {
                goto out;
              }
              if((ptm->termios.c_lflag & ECHO) && seq->scr_len) {
                ptm_puts(ptm, (char *)bsbuf, seq->scr_len);
              }
              pop_keyseq(seq);
            } else {
              seq->len = n;
              if(n == 1 && seq->data[0] == '\n') {
                ptm->canon_enter++;
              }
              if(ptm->termios.c_lflag & ECHO) {
                seq->scr_len = ptm_puts(ptm, seq->data, n);
              } else {
                seq->scr_len = 0;
              }
              push_keyseq(ptm, seq);
              wake_device(current_ptm, OUT);
            }
          }
        } else {
          put_keyseq(seq);
        }
      } else {
        seq->scr_len = 0;
        seq->len = 1;
        seq->data[0] = rbuf[i];
        push_keyseq(ptms[current_ptm], seq);
        wake_device(current_ptm, OUT);
      }
      out:
      ptm_kbdbuf_unlock(ptms[current_ptm]);        
    }
  }
  
  /* never reached */
  return NULL;
}
