/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 * (c) Copyright 2008,2009 Paul Graphov <graphov@jarios.org>
 *
 * termio.x:
 * ptm.c: physical terminals ops
 *
 */

#include <sys/types.h>
#include <stdbool.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <wchar.h>
#include <datalib/usrtc.h>
#include <ns/ns.h>
#include <sys/scheduler.h>
#include <ipc/rpc.h>
#include <devio/devio.h>
#include <termios.h>
#include <pthread.h>
#include <sys/kcontrol.h>

#include <chardev.h>
#include <spinlock.h>

#include <dircons.h>
#include <sys/ioctl.h>
#include <sys/console.h>
#include <datalib/list.h>

#include "debug.h"
#include "ptm.h"
#include "termio.h"
#include "en_layout.h"

#define VGACONS_MAXLEN (1024 / sizeof(cschar_t))
#define CPTM (ptms[current_ptm])

#define DEFAULT_COLOR 0x07

#define PTMS     10

#define TECHO_OFF  0x1

#define PTM_BUF_HEIGHT 16

spinlock_t conbuf_mutex;

typedef struct __inqueue_t {
  ulong_t msgid;
  size_t size;
} inqueue_t;

ptm_t **ptms;

int current_ptm=0;

ptm_t deb_ptm;
int deb_lineno;

static status_t ptm_write(int devno, void *buf, size_t size, size_t offset, void **interim);
static status_t ptm_read(int devno, void *buf, size_t size, size_t offset, void **interim);
static status_t ptm_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize);

static void do_scroll(int how, int lineno);
static void do_repaint(void);
static void do_scrollup(int lineno);
static void do_scrolldown(int lineno);
  
void put_string(const char *s);

void console_clear_initial(int width, int height);

struct device_ops ptm_ops = {
  .read = ptm_read,
  .write = ptm_write,
  .ioctl = ptm_ioctl,
  .log = put_string,
};

void ascii_puts(const wchar_t *msg);
wchar_t wbuf[1024];
void put_string(const char *s)
{
  int k;

  k = mbstowcs(wbuf, s, 1024);
  ascii_puts(wbuf);
}

device_server_t *srv;
static cschar_t *console_buf=NULL;
static int console_buf_size;
static uni_rpc_t cons_rpc;
int buf_pos=0;

cschar_t CSPACE = {L' ', DEFAULT_COLOR, 0};

int ptm_init(int width,int height)
{
  int i, j, r;
  int stdoutlen = sizeof(cschar_t) * width * height * PTM_BUF_HEIGHT;
  struct dev_node_t **ptm_nodes;
  char buff[12];

  ptms = malloc(sizeof(uintptr_t) * PTMS);
  if(!ptms) {
    return -1;
  }

  console_clear_initial(width, height);

  spinlock_init(&conbuf_mutex, NULL);

  for(i = 0; i < PTMS; i++) {

    ptms[i] = malloc(sizeof(ptm_t));
    if(!ptms[i]) {
      return -1;
    }
    
    ptms[i]->stdout = malloc(stdoutlen);
    if(!ptms[i]->stdout) {
      return -1;
    }

    for(j = 0; j < height * width; j++) {
      memcpy(&ptms[i]->stdout[j], &CSPACE, sizeof(CSPACE));
    }
    
    list_init_head(&ptms[i]->kbdbuf);

    memset(&ptms[i]->termios, 0, sizeof(ptms[i]->termios));
    // All other termios fields doesn't matter now...
    ptms[i]->termios.c_lflag |= ECHO | ICANON;

    ptms[i]->kbd_mode = TC_XLATE;

    
    ptms[i]->kbd_layout = 0;

    ptms[i]->wy = 0;
    ptms[i]->upscroll = 0;
    ptms[i]->fmark = 1;
    ptms[i]->uwpos = 0;
    ptms[i]->flags = 0;
    ptms[i]->wx = 0;
    ptms[i]->cattr = DEFAULT_COLOR;
    ptms[i]->xattr = 0;
    ptms[i]->width = width;
    ptms[i]->height = height;
    ptms[i]->canon_enter = false;

    ptms[i]->cur_visible = 1;

    spinlock_init(&ptms[i]->kbdlock, NULL);
    spinlock_init(&ptms[i]->buflock, NULL);
  }

  ptm_nodes = malloc(sizeof(uintptr_t)*PTMS);
  srv = malloc(sizeof(*srv));
  if(!srv || !ptm_nodes) return -1;

  r = init_multi_device_server(srv, 65536, 65535, SERVICE_NAME, &ptm_ops, 3);
  if(r != 0) {
    ascii_printf(L"Failed to init device server.\n");
    abort();
  }
  
  for(i = 0; i < PTMS; i++) {
    snprintf(buff, 12, "/dev/ptm%d", i);
    r = add_device(i, buff);
    if(r < 0) {
      ascii_printf(L"Failed to add device \"%s\".\n", buff);
      abort();
    }
  }

  set_loglevel(1);

  if(r != 0) {
    return r;
  }

  console_buf_size = width * height * 2;
  console_buf = malloc(sizeof(*console_buf) * console_buf_size);
  if(!console_buf) {
    return -1;
  }

  memset(&cons_rpc, 0, sizeof(uni_rpc_t));

  return 0;
}

void console_clear_initial(int width, int height)
{
  clear_desc_t clr;
  int r;

  clr.rect.x1 = 0;
  clr.rect.y1 = 0;
  clr.rect.x2 = width - 1;
  clr.rect.y2 = height - 1;
  clr.color = 0x70;
  
  r = ioctl(consolefd, CIOC_CLEAR, &clr);
  
  return;
}

static void console_clear_rect(rect_t *rect, uint8_t color)
{
  clear_desc_t clr;
  int r;

  memcpy(&clr.rect, rect, sizeof(*rect));
  clr.color = color;
  
  r = ioctl(consolefd, CIOC_CLEAR, &clr);

  return;  
}

void console_clear(int x1, int y1, int x2, int y2)
{
  rect_t rects[3];
  int rectn;
  int i;

  rects[0].x1 = x1;
  rects[0].y1 = y1;
  
  if(y1 == y2) {
    rectn = 1;
    rects[0].x2 = x2;
    rects[0].y2 = y2;
  } else {
    rects[0].x2 = CPTM->width-1;
    rects[0].y2 = y1;
    rects[1].x1 = 0;
    rects[1].y1 = y2;
    rects[1].x2 = x2;
    rects[1].y2 = y2;
    if(y2 == y1 + 1) {
      rectn = 2;
    } else {
      rectn = 3;
      rects[2].x1 = 0;
      rects[2].y1 = y1+1;
      rects[2].x2 = CPTM->width-1;
      rects[2].y2 = y2-1;
    }
  }

  for(i = 0; i < rectn; i++) {
    console_clear_rect(&rects[i], 0x70);
  }
  
  return;
}

void console_scrolldisable(void)
{
  ioctl(consolefd, CIOC_DISABLEAUTOSCROLL);

  return;
}

void console_scrollenable(void)
{
  ioctl(consolefd, CIOC_ENABLEAUTOSCROLL);

  return;
}

void console_flush(void)
{
  if(buf_pos) write(consolefd, console_buf, buf_pos * sizeof(*console_buf));
  buf_pos = 0;
  return;
}

void console_scroll(int startline, int stopline, int lineno)
{
  scroll_desc_t sd;
  
  sd.startline = startline;
  sd.stopline = stopline;
  sd.lines = lineno;

  ioctl(consolefd, CIOC_SCROLL, &sd);
}

void ptm_add_symb(cschar_t sym)
{
  if(buf_pos >= console_buf_size) console_flush();
  
  console_buf[buf_pos] = sym;
  buf_pos++;

  return;
}

void console_set_cursor(int x, int y)
{
  coord_t coord;

  coord.x = x;
  coord.y = y;

  ioctl(consolefd, CIOC_SETCURPOSITION, &coord);

  return;  
}

void console_set_cursor_visibility(int val)
{
  ioctl(consolefd, val ? CIOC_SHOWCURSOR : CIOC_HIDECURSOR);

  return;  
}

#define ENROUND(x, y) (x) = ((x) + (y)) % (y);

void fill_stdout(int rline, int maxline)
{
  cschar_t x;
  int ll = 0, i;

  for(i = 0; i < maxline;) {
    ENROUND(rline, CPTM->height*PTM_BUF_HEIGHT);
    if(ll == CPTM->wx && rline == (CPTM->wy + CPTM->uwpos) % (CPTM->height*PTM_BUF_HEIGHT) && i == maxline-1) break;
    x = CPTM->stdout[(rline * CPTM->width) + ll];
    ll++;
    if(x.ch > 31) ptm_add_symb(x);
    else {
      x.ch = L' ';
      ptm_add_symb(x);
    }
    if(ll==CPTM->width) {
      if(i != maxline - 1) {
        x.ch = L'\n';
        ptm_add_symb(x);
      }
      i++; rline++; ll=0;
    }
  }
}

#define abs(x) (((x) < 0) ? (-(x)) : (x))

void ptm_move_cursor(ptm_t *ptm, int how, int a, int b)
{
  if(ptm == ptms[current_ptm]) console_flush();

  switch(how) {
    case CUR_UP:
      ptm->wy -= a;
      break;
    case CUR_DOWN:
      ptm->wy += a;
      break;
    case CUR_LEFT:
      ptm->wx -= a;
      break;
    case CUR_RIGHT:
      ptm->wx += a;
      break;
    case CUR_UP_FST_COL:
      ptm->wy -= a;
      ptm->wx = 1;
      break;
    case CUR_DOWN_FST_COL:
      ptm->wy += a;
      ptm->wx = 1;
      break;
    case CUR_SET_COL:
      ptm->wx = a-1;
      break;
    case CUR_SET_ROW:
      ptm->wy = a-1;
      break;
    case CUR_SET_ROW_COL:
      ptm->wy = a-1;
      ptm->wx = b-1;
  }
  if(ptm->wx < 0) ptm->wx = 0;
  else if(ptm->wx > ptm->width) ptm->wx = ptm->width - 1;
  if(ptm->wy < 0) ptm->wy = 0;
  else if(ptm->wy > ptm->height) ptm->wy = ptm->height - 1;
  if(ptm->fmark <= ptm->wy) ptm->fmark = ptm->wy + 1;

  if(ptm == ptms[current_ptm]) console_set_cursor(ptm->wx, ptm->wy);
}

void ptm_scroll(int how, int lineno) {
  spinlock_lock(&conbuf_mutex);
  do_scroll(how, lineno);
  spinlock_unlock(&conbuf_mutex);
}

static void do_scroll(int how, int lineno)
{
  if(CPTM->fmark<CPTM->height) return;

  if(how & SCROLL_PAGES) {
    how &= ~SCROLL_PAGES;
    lineno *= CPTM->height;
  }
  
  if(how == SCROLL_ABS) {
    lineno = lineno - CPTM->upscroll;
  } else if(how == SCROLL_TOP) {
    lineno = CPTM->fmark - CPTM->height - CPTM->upscroll;
  } else if(how == SCROLL_BTM) {
    lineno = -CPTM->upscroll;
  }

  if(lineno > 0) {
    if(CPTM->upscroll + lineno > CPTM->fmark - CPTM->height)
      lineno = CPTM->fmark - CPTM->height - CPTM->upscroll;
  } else {
    if(CPTM->upscroll < -lineno) lineno = -CPTM->upscroll;
  }

  if(lineno == 0) return;

  if(abs(lineno) < CPTM->height) {
    if(lineno > 0) do_scrollup(lineno);
    else do_scrolldown(-lineno);
  } else {
    CPTM->upscroll += lineno;
    do_repaint();
  }
}

static void do_repaint(void)
{
  int rline, maxline;
  
  rline=CPTM->uwpos - CPTM->upscroll;
  if(CPTM->fmark < CPTM->height) {
    maxline = CPTM->fmark;
  } else {
    maxline = CPTM->height;
  }
  
  fill_stdout(rline, maxline);
  
  console_clear(0, 0, CPTM->width-1, CPTM->height-1);
  console_set_cursor(0, 0);
  console_scrolldisable();
  console_flush();
  console_scrollenable();
}

static void do_scrollup(int lineno)
{
  uint32_t rline;

  if(buf_pos != 0) {
    console_flush();
  }

  CPTM->upscroll += lineno;

  rline = CPTM->uwpos - CPTM->upscroll;

  fill_stdout(rline, lineno);

  /* let's flush buffer */
  console_scrolldisable();
  console_scroll(0, ptms[current_ptm]->height - 1, -lineno);
  console_set_cursor(0, 0);
  console_flush();
  console_scrollenable();
  return;
}

static void do_scrolldown(int lineno)
{
  uint32_t rline;

  if(buf_pos != 0) {
    console_flush();
  }

  rline = CPTM->uwpos - CPTM->upscroll + CPTM->height;

  CPTM->upscroll -= lineno;

  fill_stdout(rline, lineno);
  
  /* let's flush buffer */
  console_scrolldisable();
  console_scroll(0, ptms[current_ptm]->height - 1, lineno);
  console_set_cursor(0, ptms[current_ptm]->height - lineno);
  console_flush();
  console_scrollenable();
  return;
}

void ptm_switch_active(int ptmno)
{
  int old_ptm;
  spinlock_lock(&ptms[ptmno]->buflock);
  spinlock_lock(&conbuf_mutex);

  if(ptmno > 10) /* given incorrect ptm number*/
    goto out;
  if(ptmno == current_ptm) /* nothing to do */
    goto out;
  
  spinlock_lock(&ptms[current_ptm]->buflock);
  
  old_ptm = current_ptm;
  current_ptm = ptmno;
  CPTM->upscroll = 0;

  if(buf_pos != 0) /* be sure that */
    console_flush();
  

  do_repaint();

  if(ptms[old_ptm]->cur_visible != ptms[current_ptm]->cur_visible) {
    console_set_cursor_visibility(ptms[current_ptm]->cur_visible);
  }

  spinlock_unlock(&ptms[old_ptm]->buflock);

  out:
  spinlock_unlock(&conbuf_mutex);
  spinlock_unlock(&ptms[ptmno]->buflock);
}

static inline int ptm_put(ptm_t *ptm, wchar_t c)
{
  cschar_t *out = ptm->stdout;
  uint32_t rlen = ptm->height * PTM_BUF_HEIGHT;
  uint32_t loc = ((ptm->wy + ptm->uwpos) % rlen) * ptm->width+ptm->wx;
  int move_up = 0;
  int misalign;
  int pnum = 0;
  int i;

  out[loc].ch = c;
  out[loc].cattr = ptm->cattr;
  out[loc].xattr = ptm->xattr;

  switch(c) {
    case (L'\n'):
      if(ptm == ptms[current_ptm]) {
        ptm_add_symb(out[loc]);
      }
      ptm->wx = 0;
      ptm->wy++;
      if(ptm->wy >= ptm->height) {
        move_up = 1;
        ptm->wy = ptm->height - 1;
      }
      if(ptm->fmark < rlen) ptm->fmark++;
      break;
    case (L'\b'):
      if(ptm->wx > ptm->verbx || ptm->wy > ptm->verby) {
        ptm_add_symb(out[loc]);
        ptm->wx--;
      }
      break;
    case (L'\t'):
      misalign = ((ptm->wx + 8) & (~7)) - ptm->wx;
      if(ptm == ptms[current_ptm]) {
        for(i = 0; i < misalign; i++) {
          ptm_add_symb(CSPACE);
        }
      }
      ptm->wx += misalign;
      pnum = misalign;
      break;
    default:
      if(ptm == ptms[current_ptm]) ptm_add_symb(out[loc]);
      ptm->wx++;
      pnum = 1;
      break;
  }

  if(ptm->wx >= ptm->width) {
    if(ptm->fmark < rlen) {
      ptm->fmark++;
    }
    ptm->wx=0;
    ptm->wy++;
    if(ptm->wy >= ptm->height) {
      move_up = 1;
      ptm->wy = ptm->height - 1;
    }
  }
  if(move_up) {
    int j;
    ADDMOD(ptm->uwpos, 1, rlen);
    for(j = 0; j < ptm->width; j++)
      memcpy(&ptm->stdout[j + ptm->width*((ptm->wy+ptm->uwpos) % rlen)], &CSPACE, sizeof(CSPACE));
  }

  ptm->upscroll = 0;

  return pnum;
}

void ptm_print(char *msg)
{
  size_t size = strlen(msg);

  ptm_puts(ptms[current_ptm], msg, size);

  return;
}

void ptm_setmode(ptm_t *ptm, int val)
{
  int i;

  for(i = 0; i < ptm->nparams; i++) {
    switch(ptm->params[i]) {
      case 25:
        if(ptm->cur_visible != val) {
          if(ptm == ptms[current_ptm]) console_set_cursor_visibility(val);
          ptm->cur_visible = val;
        }
        break;
    }
  }
}

// convert linear offset inside writing window to buffer offset
#define PTWINL(l)                               \
  ((ptm->uwpos * ptm->width + (l)) % llen)

// convert coordinates to linear window offset
#define XYTOL(x,y)\
  (ptm->width*(y) + (x))

void ptm_clear(ptm_t *ptm, int how)
{
  uint32_t llen = ptm->height * ptm->width * PTM_BUF_HEIGHT;
  uint32_t startx, starty, stopx, stopy, i, j;

  /*
    how:
    0 = clear from cursor to the end of the screen
    1 = clear from the beginning of the screen to the cursor
    2 = clear entire screen
  */
  
  if(how == 0) {
    startx = ptm->wx;
    starty = ptm->wy;
  } else {
    startx = 0;
    starty = 0;
  }

  if(how == 1) {
    stopx = ptm->wx + 1;
    stopy = ptm->wy;
  } else {
    stopx = ptm->width;
    stopy = ptm->height - 1;
  }
  
  for(i = XYTOL(startx, starty); i < XYTOL(stopx, stopy); i++) {
    j = PTWINL(i);
    ptm->stdout[j].ch = L' ';
    ptm->stdout[j].cattr = ptm->cattr;
    ptm->stdout[j].xattr = ptm->xattr;
  }
  if(ptm == ptms[current_ptm]) {
    console_clear(startx, starty, stopx, stopy);
    console_set_cursor(0, 0);
  }
}

#define SETFG(x,c)                              \
  {                                             \
    (x) &= 0xF0;                                \
    (x) |= ((c) & 0x0F);                        \
  }

#define SETBG(x,c)                              \
  {                                             \
    (x) &= 0x0F;                                \
    (x) |= (((c) & 0x0F) << 4);                 \
  }


int ptm_puts(ptm_t *ptm,char *buf,size_t size)
{
  int i;
  wchar_t wc;
  int pnum = 0;

  spinlock_lock(&ptm->buflock);
  if(ptm == ptms[current_ptm]) {
    spinlock_lock(&conbuf_mutex);
    console_scrollenable();
    // scroll to bottom if needed.
    if(CPTM->upscroll > 0) do_scroll(SCROLL_BTM, 0);
  }    
  
  for(i=0;i<size;) {
    int k;
    k = mbrtowc(&wc, &buf[i], size - i, NULL);
    if(k <= 0) {
      break;
    }
    // Simple state machine, scans for some of vt-102 sequences.
    switch(ptm->esc_state) {
      case ST_NORM:
        if(wc == L'\e') {
          ptm->esc_state = ST_ESC;
        } else {
          pnum += ptm_put(ptm, wc);
        }
        break;
      case ST_ESC:
        if(wc == L'[') {
          ptm->esc_state = ST_BRACE;
          memset(ptm->params, 0, sizeof(ptm->params));
          ptm->nparams = 0;
          ptm->partial = -1;
          ptm->quest = 0;
        } else if(wc == L'\e') {
        } else if(wc < 32) {
          ptm_put(ptm, wc);
        } else {
          ptm->esc_state = ST_NORM;
        }
        break;
      case ST_BRACE:
        if(wc == L'?') {
          ptm->quest = 1;
          ptm->esc_state = ST_PARAMS;
          break;
        }
      case ST_PARAMS:
        if(wc == L'-') {
          ptm->neg_partial = 1;
          ptm->esc_state = ST_NUM;
          break;
        }
      case ST_NUM:
        if(wc >= L'0' && wc <= '9') {
          if(ptm->partial == -1) ptm->partial = 0;
          ptm->partial *= 10;
          ptm->partial += wc - L'0';
          if(ptm->partial > 100) ptm->partial = 100;
          break;
        } else if(wc == L'\e') {
          ptm->esc_state = ST_ESC;
        } else if(wc < 32) {
          ptm_put(ptm, wc);
        } else {
          if(ptm->partial == -1) {
            if(!ptm->neg_partial) {
              ptm->partial = 0;
            } else {
              ptm->esc_state = ST_NORM;
              break;
            }
          }
          ptm->params[ptm->nparams++] = ptm->partial * (ptm->neg_partial?(-1):1);
          if(ptm->nparams >= NPAR && wc == L';') {
            ptm->esc_state = ST_NORM;
            break;
          }

          if(wc == L';') {
             ptm->partial = -1;
             ptm->neg_partial = 0;
          } else {
            int j;
            ptm->esc_state = ST_NORM;
            switch(wc) {
              case L'm':
                for(j = 0; j < ptm->nparams; j++) {
                  uint16_t p = ptm->params[j];
                  if(p >= 30 && p <= 37) {
                    SETFG(ptm->cattr, p - 30);
                  } else if(p >= 40 && p <= 47){
                    SETBG(ptm->cattr, p - 40);                    
                  } else if(p == 0) {
                    ptm->cattr = DEFAULT_COLOR;
                  }
                }
                break;
              case L'J':
                ptm_clear(ptm, ptm->params[0]);
                break;
              case L'A':
                ptm_move_cursor(ptm, CUR_UP, ptm->params[0], 0);
                break;
              case L'B':
              case L'e':
                ptm_move_cursor(ptm, CUR_DOWN, ptm->params[0], 0);
                break;
              case L'C':
              case L'a':  
                ptm_move_cursor(ptm, CUR_RIGHT, ptm->params[0], 0);
                break;
              case L'D':
                ptm_move_cursor(ptm, CUR_LEFT, ptm->params[0], 0);
                break;
              case L'E':
                ptm_move_cursor(ptm, CUR_DOWN_FST_COL, ptm->params[0], 0);
                break;
              case L'F':
                ptm_move_cursor(ptm, CUR_UP_FST_COL, ptm->params[0], 0);
                break;
              case L'G':
              case L'`':
                ptm_move_cursor(ptm, CUR_SET_COL, ptm->params[0], 0);
                break;
              case L'd':
                ptm_move_cursor(ptm, CUR_SET_ROW, ptm->params[0], 0);
                break;
              case L'H':
              case L'f':
                ptm_move_cursor(ptm, CUR_SET_ROW_COL, ptm->params[0], ptm->params[1]);
                break;
              case L'h':
                ptm_setmode(ptm, 1);
                break;
              case L'l':
                ptm_setmode(ptm, 0);
                break;
            }
          }
        }
        break;
    }
    i += k;
  }

  if(ptm == ptms[current_ptm]) {
    console_flush();
    spinlock_unlock(&conbuf_mutex);
  }
  
  spinlock_unlock(&ptm->buflock);

  return pnum;
}

void ptm_verbatim(ptm_t *ptm)
{
  ptm->verbx = ptm->wx;
  ptm->verby = ptm->wy;
}

static status_t ptm_write(int devno, void *buf,size_t size,size_t offset, void **interim)
{
  if(devno < PTMS) {
    ptm_puts(ptms[devno], (char *)buf,size);
    ptm_verbatim(ptms[devno]);
  } else
    return -EINVAL;

  return size;
}

#include <sys/kcontrol.h>

static status_t ptm_ioctl(int devno, int type, int nr, void *buf, size_t size, void *obuf, size_t osize)
{
  status_t ret = 0;
  
  if(type != TERM_IOCTL) return -123;

  if(devno >= PTMS) return -456;

  ptm_kbdbuf_lock(ptms[devno]);

  switch(nr) {
    case TERM_GETSIZE:
      if(osize < sizeof(coord_t)) {
        ret = -EINVAL;
        break;
      }
      ((coord_t *)obuf)->x = ptms[devno]->width;
      ((coord_t *)obuf)->y = ptms[devno]->height;
      ret = 0;
      break;
    case TERM_GETATTR:
      if(osize < sizeof(struct termios)) {
        ret = -EINVAL;
        break;
      }
      memcpy(obuf, &ptms[devno]->termios, sizeof(struct termios));
      ret = 0;
      break;
    case TERM_SETATTR:
      if(size < sizeof(struct termios)) {
        ret = -EINVAL;
        break;
      }
      memcpy(&ptms[devno]->termios, buf, sizeof(struct termios));
      ret = 0;
      break;
    case TERM_GETKBMODE:
      if(osize < 1) return -EINVAL;
      *(uint8_t*)obuf = ptms[devno]->kbd_mode;
      break;
    case TERM_SETKBMODE:
      if(size < 1) return -EINVAL;
      if(*(uint8_t*)buf != ptms[devno]->kbd_mode) {
        flush_kbdbuf(ptms[devno]);
      }
      ptms[devno]->kbd_mode = *(uint8_t*)buf;
      break;
    default:
      ret = -ENOSYS;
  }

  ptm_kbdbuf_unlock(ptms[devno]);
    
  return ret;
}

status_t read_noncanon(ptm_t *ptm, uint8_t *buf, int size)
{
  int pos;
  keyseq_t *seq;
  
  //got = ptm_kbdbuf_read_l(ptm, buf, size);
  ptm_kbdbuf_lock(ptm);
  
  while(size && ((seq = first_keyseq(ptm)) != NULL)) {
    if(size >= seq->len) {
      MEMCPY_SMALL(&buf[pos], seq->data, seq->len);
      pos += seq->len;
      size -= seq->len;
      pop_keyseq(seq);
    } else {
      MEMCPY_SMALL(&buf[pos], seq->data, size);
      MEMCPY_SMALL(seq->data, &seq->data[size], seq->len - size);
      seq->len -= size;
      pos += size;
      size = 0;
    }
  }

  ptm_kbdbuf_unlock(ptm);
  
  if(!pos) {
    return SUSPEND;
  } else {
    return pos;
  }
}

status_t read_canon(ptm_t *ptm, uint8_t *buf, int size, int *ready)
{
  int left;
  status_t ret;
  keyseq_t *seq;

  left = size - *ready;
  
  if(!ptm->canon_enter) {
    return SUSPEND;
  }
  
  ptm_kbdbuf_lock(ptm);

  while(left && ((seq = first_keyseq(ptm)) != NULL)) {
    if(left >= seq->len) {
      MEMCPY_SMALL(&buf[*ready], seq->data, seq->len);
      *ready += seq->len;
      left -= seq->len;
      if(seq->data[0] == '\n') {
        ptm->canon_enter--;
        left = 0;
      }
      pop_keyseq(seq);
    } else {
      MEMCPY_SMALL(&buf[*ready], seq->data, size);
      MEMCPY_SMALL(seq->data, &seq->data[size], seq->len - size);
      seq->len -= size;
      *ready += size;
      left = 0;
    }
  }

  if(left > 0) {
    ret = SUSPEND;
  } else {
    ret = *ready;
  }
  
  /*
  while(*ready < size) {
    if(!ptm->read_middle) {
      got = ptm_kbdbuf_getc(ptm, &tmp);
      if(got) {
        ptm->read_middle = true;
        got = ptm_kbdbuf_getc(ptm, &tmp);
      }
      if(!got) {
        ret = SUSPEND;
        break;
      }
    }
    got = ptm_kbdbuf_getc(ptm, &tmp);
    if(!got) {
      ptm->canon_enter = 0;
      ret = SUSPEND;
      break;
    }
    if(tmp == '\n') {
      buf[*ready] = '\n';
      ret = *ready + 1;
      ptm->canon_enter = 0;
      break;
    } else if(tmp == '\0') {
      ptm->read_middle = false;
      continue;
    }
    buf[*ready] = tmp;
    (*ready)++;
  }
  */
  
  ptm_kbdbuf_unlock(ptm);
  return ret;
}

static status_t ptm_read(int devno, void *buf, size_t size, size_t offset, void **interim)
{
  status_t ret;

  if(devno >= PTMS) return -EINVAL;

  if(ptms[devno]->termios.c_lflag & ICANON) {
    ret = read_canon(ptms[devno], buf, size, (int *)interim);
  } else {
    ret = read_noncanon(ptms[devno], buf, size);
  }

  return ret;
}

