/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 * (c) Copyright 2008,2009 Paul Graphov <graphov@jarios.org>
 *
 * termio.x:
 * ptm.h: physical terminals ops
 *
 */

#ifndef __PTM_H__
#define __PTM_H__

#include <sys/types.h>
#include <datastructs.h>
#include <datalib/circular.h>
#include <sys/console.h>
#include <termios.h>
#include <spinlock.h>
#include <sys/kcontrol.h>
#include <datalib/list.h>

#include "keyb.h"

#define PTM_KBDBUF_SIZE 1023
#define MBBUF_SIZE 16

#define SCROLL_ABS 0
#define SCROLL_REL 1
#define SCROLL_TOP 2
#define SCROLL_BTM 3

#define SCROLL_PAGES 0x80

#define NPAR 15

// maximum length of key sequence
#define MAX_KEY_LEN 12

typedef struct __keyseq {
  list_node_t list;
  unsigned char len;
  unsigned char scr_len;
  char data[MAX_KEY_LEN];
} keyseq_t;

typedef enum {
  CUR_UP = 0,
  CUR_DOWN,
  CUR_LEFT,
  CUR_RIGHT,
  CUR_UP_FST_COL,
  CUR_DOWN_FST_COL,
  CUR_SET_COL,
  CUR_SET_ROW,
  CUR_SET_ROW_COL,
} cur_dir_t;

// Esc sequences are processed via state machine
typedef enum {
  ST_NORM = 0,
  ST_ESC,
  ST_BRACE,
  ST_PARAMS,
  ST_NUM,
} escm_state_t;

typedef struct __ptm_t {
  /* output data */
  cschar_t *stdout; /* real terminal buffer */
  spinlock_t buflock;
  int16_t wx;    /* writing position line */
  int16_t wy;
  int16_t verbx; /* verbatim text limit */
  int16_t verby; /* no backspace behind */
  int16_t upscroll;
  int16_t uwpos;
  uint16_t fmark;   /* cycle buffer fill marker */
  uint8_t cur_visible; /* is cursor visible */

  /* terminal state */
  uint16_t flags;   /* state of the terminal */
  uint16_t width,height;
  uint8_t cattr,xattr;

  /*read lock, write lock is not needed as only one writer */
  spinlock_t kbdlock;
  list_head_t kbdbuf;
  uint8_t kbd_layout;
  // TODO: next two must be eliminated by better data structure
  // for keyboard buffer.
  bool canon_enter; // if ENTER was pressed and canonical read
                    // can be done

  //Esc sequence related:
  escm_state_t esc_state;
  int16_t partial;
  uint8_t neg_partial;
  uint8_t nparams;
  uint8_t quest;
  int16_t params[NPAR];

  // terminal state
  struct termios termios;
  uint8_t kbd_mode;
} ptm_t;

/*
size_t ptm_kbdbuf_write(ptm_t *ptm, uint8_t *msg, int size);
size_t ptm_kbdbuf_read_l(ptm_t *ptm, uint8_t *msg, int size);
size_t ptm_kbdbuf_unputc(ptm_t *ptm, uint8_t *c);
size_t ptm_kbdbuf_putc(ptm_t *ptm, uint8_t c);
size_t ptm_kbdbuf_getc(ptm_t *ptm, uint8_t *msg);
void ptm_kbdbuf_zero(ptm_t *ptm);
*/

keyseq_t *get_keyseq(void);
keyseq_t *first_keyseq(ptm_t *ptm);
keyseq_t *last_keyseq(ptm_t *ptm);
void pop_keyseq(keyseq_t *seq);
void push_keyseq(ptm_t *ptm, keyseq_t *seq);
void flush_kbdbuf(ptm_t *ptm);

#define MEMCPY_SMALL(dest, src, n)              \
  {                                             \
    int i;                                      \
    for(i = 0; i < (n); i++) {                  \
      ((char *)(dest))[i] = ((char *)(src))[i]; \
    }                                           \
  }

static inline void ptm_kbdbuf_lock(ptm_t *ptm)
{
  spinlock_lock(&ptm->kbdlock);
}

static inline void ptm_kbdbuf_unlock(ptm_t *ptm)
{
  spinlock_unlock(&ptm->kbdlock);
}

int ptm_init(int width, int height);
void ptm_clear(ptm_t *ptm, int how);
void ptm_verbatim(ptm_t *ptm);

int ptm_puts(ptm_t *ptm, char *buf, size_t size);

void ptm_process_keyboard(ptm_t *ptm, uint8_t msg);

void *ptm_thread(void *data);

void ptm_print(char *msg);

void ptm_scroll(int how,int lineno);

void ptm_switch_active(int ptmno);

void console_set_cursor(int x, int y);

#endif /* __PTM_H__ */
