/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system,
 * mount table operations
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <wchar.h>
#include <datalib/usrtc.h>
#include <errno.h>
#include <unistd.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <ns/ns.h>
#include <pthread.h>
#include <termio/term.h>

/* services headers */
#include <fsio/vfs_call.h>

/* local headers */
#include "vfs_glob.h"
#include "vfs_rpc.h"

usrtc_t *mount_table;
usrtc_t *ft_cache;

static int __cmp_mountfs(const void *a,const void *b);

int init_mounts(void)
{
  mount_table=malloc(sizeof(*mount_table));
  ft_cache=malloc(sizeof(*ft_cache));
  if(!mount_table || !ft_cache)
    return -1;

  usrtc_init(mount_table,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__cmp_mountfs);
  usrtc_init(ft_cache,USRTC_LIST,USRTC_COUNT_T_MAX,__cmp_mountfs);

  return 0;
}

int add_mount_task(void *msg,size_t len,int kmsgid)
{
  serv_task_t *job;
  vfs_mount_request_t *mount_req;
  uni_rpc_t *rpc=(uni_rpc_t *)msg;
  int dlen,mlen,flen;
  int o;

  /* check size */
  if(len<sizeof(*rpc)+((rpc->arg1+rpc->arg2)*sizeof(wchar_t))+rpc->arg3) 
    return EINVAL;
  if(!rpc->arg1 || !rpc->arg2)
    return EINVAL;
  if(!rpc->arg3)
    return EINVAL;
  /* TODO: check rights */
  /*
   *if(access_check(RPC_MOUNT,rpc->arg5,rpc->arg6))
   *  return EPERM;
   */
  mount_req=malloc(sizeof(*mount_req));
  if(!mount_req)    return ENOMEM;
  mount_req->flags=rpc->arg0;
  dlen=(rpc->arg1+1)*sizeof(wchar_t);
  mlen=(rpc->arg2+1)*sizeof(wchar_t);
  flen=rpc->arg3+1;
  mount_req->devno=malloc(dlen);
  mount_req->point=malloc(mlen);
  if(!mount_req->devno || !mount_req->point) {
    free(mount_req);
    return ENOMEM;
  } else {
    memset(mount_req->devno,0,dlen);
    memset(mount_req->point,0,mlen);
  }
  mount_req->fstype=malloc(flen);
  if(!mount_req->fstype)
    return ENOMEM;
  else
    memset(mount_req->fstype,0,flen);
  dlen--;mlen--;flen--;
  memcpy(mount_req->devno,(void *)((char *)msg+sizeof(*rpc)),dlen);
  memcpy(mount_req->point,(void *)((char *)msg+sizeof(*rpc)+dlen),mlen);
  memcpy(mount_req->fstype,(void *)((char *)msg+sizeof(*rpc)+dlen+mlen),flen);

  job=malloc(sizeof(*job));
  if(!job)
    return ENOMEM;
  job->vfs_cmd=RPC_MOUNT;
  job->kmsgid=kmsgid;
  job->data=mount_req;
  o=add_service_job(job);
  if(o!=0) {
    free(job);
    free(mount_req->devno);
    free(mount_req->point);
    free(mount_req->fstype);
    free(mount_req);
    return o;
  }

  return 0;
}

int vfs_mount(vfs_mount_request_t *mount_req)
{
  usrtc_node_t *node;
  vfs_mount_point_t *mount;
  uni_rpc_t *rpc;
  int rsize;
  char *msg;
  int r;

  node=usrtc_lookup(ft_cache,mount_req->fstype);
  if(!node)
    return ENOSYS; /* TODO: loads file system */
  else
    mount=usrtc_node_getdata(node);

  /* now let's send a request to file system to make mount */
  rsize=sizeof(uni_rpc_t);
  if(mount_req->devno) 
    rsize+=strlen(mount_req->devno)*sizeof(char);
  /* construct */
  msg=malloc(rsize);
  if(!msg)
    return ENOMEM;
  rpc=(uni_rpc_t*)msg;
  rpc->rpc_call=RPC_MOUNT;
  rpc->arg0=mount_req->flags;
  rpc->arg1=rsize-sizeof(uni_rpc_t);

  if(mount_req->devno!=NULL)
    memcpy((void*)(msg+sizeof(uni_rpc_t)),mount_req->devno,rsize-sizeof(uni_rpc_t));

  r=ipc_channel_msgsnd(mount->s_channel,rsize,msg,sizeof(uni_rpc_t),rpc);
  if(rpc->rpc_call!=0) {
    /* TODO: delete file mounts table cache entry */
    free(msg);
    return rpc->rpc_call;
  }

  free(msg);

  /* all is ok, remove from file cache */
  mount=(vfs_mount_point_t*)node;
  usrtc_delete(ft_cache,node);

  usrtc_node_init(node,mount);
  mount->hash=hash_s(mount_req->point);
  mount->use_count=0;
  mount->mount_path=strdup(mount_req->point);
  mount->over_off=NULL;
  if(mount_req->devno!=NULL) {
    /* TODO: implement logic with devices mounts */
  } else
    mount->sys_arrangel=NULL;

  /* ok insert this */
  usrtc_node_init(node,mount);
  usrtc_insert(mount_table,node,mount->mount_path);

  return 0;
}

int register_fs(void *msg,size_t len,int kmsgid)
{
  usrtc_node_t *node;
  vfs_mount_point_t *vfs_point;
  vfs_reg_head_t reg_info;
  vfs_fs_info_t *fsinfo;
  serv_task_t *job;
  uni_rpc_t *rpc;
  char *p;
  size_t len_head;

  /* check rpc header */
  rpc=(uni_rpc_t *)msg;
  if(rpc->arg0<sizeof(reg_info))
    return EINVAL;
  else
    len_head=rpc->arg0;
  p=msg;p+=sizeof(*rpc);
  /*TODO: check rights */
  /*if(access_check(rpc->arg5,rpc->arg6))
   *  return EPERM;
   */
  memcpy(&reg_info,p,sizeof(reg_info));
  len_head-=sizeof(reg_info);
  reg_info.fs_name=malloc(len_head+sizeof(wchar_t));
  if(!reg_info.fs_name)
    return ENOMEM;
  else
    memset(reg_info.fs_name,'\0',len_head+sizeof(wchar_t));
  p+=sizeof(reg_info);
  memcpy(reg_info.fs_name,p,len_head);
  /* ok let's add */
  node=malloc(sizeof(*node));
  vfs_point=malloc(sizeof(*vfs_point));
  if(!node || !vfs_point) {
    free(reg_info.fs_name);
    return ENOMEM;
  }
  /* fill out */
  vfs_point->hash=0;
  job=malloc(sizeof(*job));
  fsinfo=malloc(sizeof(*fsinfo));
  if(!job || !fsinfo) {
    free(reg_info.fs_name);
    free(node);
    free(vfs_point);
    return ENOMEM;
  }
  vfs_point->info=fsinfo;
  fsinfo->fs_name=reg_info.fs_name;
  vfs_point->fspid=reg_info.pid;
  vfs_point->w_port=reg_info.w_port;
  vfs_point->s_port=reg_info.s_port;
  /* create channels */
  vfs_point->w_channel=ipc_channel_open(vfs_point->fspid,
					vfs_point->w_port,IPC_BLOCKED_ACCESS);
  vfs_point->s_channel=ipc_channel_open(vfs_point->fspid,
					vfs_point->s_port,IPC_BLOCKED_ACCESS);

  /* add to file system table cache */
  usrtc_node_init(node,vfs_point);
  usrtc_insert(ft_cache,node,fsinfo->fs_name);

  job->vfs_cmd=RPC_REGISTER;
  job->data=vfs_point;
  job->kmsgid=kmsgid;

  return add_service_job(job);
}

vfs_mount_point_t *vfs_lookup_end_point(char *path,int *mt_off,
					int amount,int *markers)
{
  int i,len;
  usrtc_node_t *node,*fnode;
  vfs_mount_point_t *pt=NULL;
  char *tbuf,*pbuf;

  if(usrtc_count(mount_table)<1)
    return NULL;

  pbuf=path;
  len=strlen(path);
  tbuf=malloc(len+1);
  memset(tbuf,'\0',len+1);

  /* simple case */
  if(amount<2) {
    node=usrtc_lookup(mount_table,L"/");
    pt=usrtc_node_getdata(node);
    *mt_off=0;
    return pt;
  }

  fnode=usrtc_lookup(mount_table,"/");

  for(i=0;i<amount-1;i++) {
    tbuf=strncat(tbuf,pbuf,markers[i+1]-markers[i]);
    node=usrtc_lookup(mount_table,tbuf);
    if(node==NULL) {
      pt=usrtc_node_getdata(fnode);
      goto __lok_ex;
    } else {
      fnode=node;
      pbuf+=markers[i+1];
    }
  }

 __lok_ex:
  *mt_off=i-1;
  pt=usrtc_node_getdata(fnode);

  free(tbuf);

  return pt;
}

static int __cmp_mountfs(const void *a,const void *b)
{
  char *aa=(char *)a;
  char *bb=(char *)b;

  return strcmp(aa,bb);
}
