/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system,
 * queues
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/uevent.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <datalib/usrtc.h>
#include <errno.h>
#include <unistd.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <ns/ns.h>
#include <pthread.h>
#include <termio/term.h>

/* local headers */
#include "vfs_glob.h"
#include "vfs_rpc.h"

usrtc_t *job_queue;
ulong_t glob_job_queue_lock;
/* syncronizations primitives */
pthread_mutex_t tqueue_lock;
event_t equeue_atom;

static int __cmp_kmsgid(const void *a,const void *b);

#define QUEUE_ATOMS  10

#define __lock_queue    pthread_mutex_lock(&tqueue_lock)
#define __unlock_queue  pthread_mutex_unlock(&tqueue_lock);

#define __atom_wait     event_wait(&equeue_atom)

static void *queue_worker(void *args);

int init_queue(void)
{
  int i,r;
  pthread_t qap[QUEUE_ATOMS];

  job_queue=malloc(sizeof(*job_queue));
  if(!job_queue)
    return -1;
  usrtc_init(job_queue,USRTC_LIST,USRTC_COUNT_T_MAX,__cmp_kmsgid);

  pthread_mutex_init(&tqueue_lock,NULL);
  event_init(&equeue_atom);

  for(i=0;i<QUEUE_ATOMS;i++) {
    r=pthread_create(&qap[i],NULL,queue_worker,(void *)job_queue);
    if(r!=0) 
      return -1;
  }

  return 0;
}

int queue_atom_run(void) 
{
  return event_signal(&equeue_atom);
}

int add_task(int vfs_cmd,vfs_path_loc_t *loc,
	     uid_t uid,gid_t gid,pid_t pid,
	     int kmsgid,void *data,size_t data_len)
{
  usrtc_node_t *node;
  vfs_task_t *task;

  node=malloc(sizeof(*node));
  task=malloc(sizeof(*task));
  if(!node || !task)
    return ENOMEM;

  task->vfs_cmd=vfs_cmd;
  task->loc=loc;
  task->uid=uid;
  task->gid=gid;
  task->pid=pid;
  task->kmsgid=kmsgid;
  task->data=data;
  task->data_len=data_len;

  usrtc_node_init(node,task);

  __lock_queue;
  usrtc_insert(job_queue,node,task);
  __unlock_queue;

  return 0;
}

vfs_task_t *take_task(void)
{
  usrtc_node_t *node;
  vfs_task_t *task;

  node=usrtc_first(job_queue);
  if(node!=NULL) {
    task=usrtc_node_getdata(node);

    __lock_queue;
    usrtc_delete(job_queue,node);
    __unlock_queue;

    free(node);
    return task;
  } else 
    return NULL;
}

static void *queue_worker(void *args)
{
  int r,i;
  vfs_task_t *myjob;
  uni_rpc_t rpc;
  task_res_t *resource;
  char *tbuf;
  int *proots;
  uni_rpc_t *trpc;
  int fs_root;
  vfs_mount_point_t *dest_fs;
  vfs_fd_t *tvfd;
  int len;
  iovec_t r_vals[2];

  /*
   * TODO: add permission and limits check
   */

  while(1) {
    r=__atom_wait;
    if(r==0) {
      myjob=take_task();
      if(!myjob)
	continue;
      switch(myjob->vfs_cmd) {
      case RPC_OPEN:
	/* there can be case while we're opening something from kernel */
	/*
	 * common situation, there are nothing :
	 *  - resource is absent 
	 */
	resource=lookup_task_resource(myjob->pid);
	if(!resource) {
	  resource=new_task_resource(NULL,myjob->pid,myjob->uid,myjob->gid);
	} 
	trpc=(uni_rpc_t *)myjob->data;
	proots=(int *)((char *)myjob->data+sizeof(*trpc));
	tbuf=(char *)((char *)myjob->data+sizeof(*trpc)+(trpc->arg5*sizeof(uint16_t)));
	dest_fs=vfs_lookup_end_point(tbuf,&fs_root,trpc->arg5,proots);
	if(!dest_fs) { /* not mounted anything, or incorrect */
	  rpc.rpc_call=ENOENT;
	  i=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,myjob->kmsgid);
	  if(i<0) 
	    if(i==-EPIPE) {
	      gc_cleanup_pid(myjob->pid);
	    }
	  free(myjob->data);
	  free(myjob);
	  continue;
	} else { /* ok, all is done */
	  trpc->offset=fs_root;
	  /* now let's make a request to destination filesystem */
	  i=ipc_channel_msgsnd(dest_fs->s_channel,myjob->data_len,
			       myjob->data,sizeof(rpc),&rpc); 
	  if(rpc.rpc_call!=0) { /* error */
	    i=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,myjob->kmsgid);
	    if(i<0) 
	      if(i==-EPIPE) {
		gc_cleanup_pid(myjob->pid);
	      }
	    free(myjob->data);
	    free(myjob);
	    continue;
	  } else { /* yummie, let's add this to resources */
	    lock_task_resource(resource);
	    i=add_descriptor_task_resource(resource,trpc->size,dest_fs,trpc->arg0);
	    rpc.rpc_call=i;
	    rpc.arg2=dest_fs->fspid;
	    rpc.arg3=dest_fs->w_port;
	    i=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,myjob->kmsgid);
	    unlock_task_resource(resource);
	    free(myjob->data);
	    free(myjob);
	    continue;
	  }
	}
	break;
      case RPC_CLOSE:
	resource=lookup_task_resource(myjob->pid);
	if(!resource) {
	__close_notexist:
	  rpc.rpc_call=EINVAL;
	__close_ok:
	  i=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,myjob->kmsgid);
	  if(i<0) 
	    if(i==-EPIPE) {
	      gc_cleanup_pid(myjob->pid);
	    }
	  free(myjob->data);
	  free(myjob);
	  continue;
	} 
	trpc=(uni_rpc_t *)myjob->data;
	lock_task_resource(resource);
	tvfd=task_resource_lookup(resource,trpc->size);
	if(!tvfd) {
	  unlock_task_resource(resource);
	  goto __close_notexist;
	}
	dest_fs=tvfd->owner;
	trpc->arg2=resource->pid;
	trpc->arg3=resource->uid;
	trpc->arg4=resource->gid;
	i=ipc_channel_msgsnd(dest_fs->s_channel,myjob->data_len,
			     myjob->data,sizeof(rpc),&rpc);
	if(!rpc.rpc_call)
	  rpc.rpc_call=task_resource_delete(resource,tvfd);
	unlock_task_resource(resource);
	goto __close_ok;
	break;
      case RPC_GETCWD:
	resource=lookup_task_resource(myjob->pid);
	if(!resource) {
	  resource=new_task_resource(NULL,myjob->pid,myjob->uid,myjob->gid);
	} 
	trpc=(uni_rpc_t *)myjob->data;
	if(trpc->size<=0) { /* tra-lala EINVAL */
	  rpc.rpc_call=EINVAL;
	__err_getcwd:
	  i=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,myjob->kmsgid);
	  if(i<0) 
	    if(i==-EPIPE) {
	      gc_cleanup_pid(myjob->pid);
	    }
	  free(myjob->data);
	  free(myjob);
	  continue;
	} else {
	  len=strlen(resource->cwd);
	  lock_task_resource(resource);
	  if(len+1>trpc->size) { /* POSIX says that we should return an error in this case ERANGE */
	    unlock_task_resource(resource);
	    rpc.rpc_call=ERANGE;
	    goto __err_getcwd;
	  } else { /* now it's possible to make a reply */
	    trpc->size=len;
	    r_vals[0].iov_base=trpc;
	    r_vals[0].iov_len=sizeof(*trpc);
	    r_vals[1].iov_base=resource->cwd;
	    r_vals[1].iov_len=len;
	    i=ipc_port_reply_nmsg_iov(serv_port,r_vals,2,myjob->kmsgid);
	    if(i<0 && i==-EPIPE) {
	      gc_cleanup_pid(myjob->pid);
	    }
	  }
	  unlock_task_resource(resource);
	  free(myjob->data);
	  free(myjob);
	}
	break;
      case RPC_OPENDIR:
      case RPC_CLOSEDIR:
      case RPC_STAT:
      case RPC_DUP:
      case RPC_DUP2:
      case RPC_SYNC:
	rpc.rpc_call=ENOSYS;
	i=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,myjob->kmsgid);
	if(i<0) 
	  if(i==-EPIPE) {
	    gc_cleanup_pid(myjob->pid);
	  }
	free(myjob->data);
	free(myjob);
	break;
      default:
	rpc.rpc_call=EINVAL;
	i=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,myjob->kmsgid);
	if(i<0) 
	  if(i==-EPIPE) {
	    gc_cleanup_pid(myjob->pid);
	  }
	free(myjob->data);
	free(myjob);
	break;
      }
    } else
      continue;
  }

  return NULL;
}

static int __cmp_kmsgid(const void *a,const void *b)
{
  vfs_task_t *aa=(vfs_task_t *)a;
  vfs_task_t *bb=(vfs_task_t *)b;

  return aa->kmsgid - bb->kmsgid;
}
