/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system,
 * resource storage
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <datalib/usrtc.h>
#include <errno.h>
#include <unistd.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <ns/ns.h>
#include <pthread.h>
#include <termio/term.h>

/* local headers */
#include "vfs_glob.h"
#include "vfs_rpc.h"

usrtc_t *res_list;
pthread_mutex_t res_lock;

/* misc local functions prototypes */
static int __cmp_res_func(const void *a,const void *b);
static int __cmp_fds_func(const void *a,const void *b);

int init_resources(void)
{
  res_list=malloc(sizeof(*res_list));
  if(!res_list) {
    return -1;
  }

  usrtc_init(res_list,USRTC_SPLAY,USRTC_COUNT_T_MAX,__cmp_res_func);

  pthread_mutex_init(&res_lock,NULL);

  return 0;
}

/*
 * lookup_task_resource: try to find existing pid resource 
 * It returns NULL, or pointer to pid resource, 
 * take care *it increments reference counter!*
 */
task_res_t *lookup_task_resource(pid_t pid)
{
  task_res_t *resource=NULL;
  task_res_t fr;
  usrtc_node_t *node;

  fr.pid=pid;
  pthread_mutex_lock(&res_lock);
  node=usrtc_lookup(res_list,&fr);
  pthread_mutex_unlock(&res_lock);
  if(node==NULL) {
    pthread_mutex_unlock(&res_lock);
    return NULL;
  } else {
    resource=(task_res_t *)usrtc_node_getdata(node);
    resource->ref_count++;
    pthread_mutex_unlock(&res_lock);
  }

  return resource;
}

/*
 * new_task_resource: add a new resource entry for pid
 * make sure that this pid doesn't exist, there are no 
 * checking.
 * It returns pointer to new resource and increments reference count.
 * Take care! if CWD is NULL *ROOT* directory will be set as CWD, also
 * this function *freeing cwd* on any error.
 */
task_res_t *new_task_resource(char *cwd,pid_t pid,uid_t uid,gid_t gid)
{
  task_res_t *nres=NULL;
  usrtc_node_t *node;

  pthread_mutex_lock(&res_lock);

  node=malloc(sizeof(*node));
  nres=malloc(sizeof(*nres));
  if(!node || !nres) {
    free(cwd);
    pthread_mutex_unlock(&res_lock);
    return NULL;
  }
  usrtc_node_init(node,nres); /* init node */

  /* make a deal with CWD */
  if(!cwd) {
    nres->cwd=strdup("/");
    if(!nres->cwd)
      goto __noroom;
  } else 
    nres->cwd=cwd;
  
  /* initialize new resource */
  nres->fds=malloc(sizeof(*nres->fds));
  if(!nres->fds) {
  __noroom:
    free(nres);
    free(node);
    free(cwd);
    pthread_mutex_unlock(&res_lock);
    return NULL;
  }
  usrtc_init(nres->fds,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__cmp_fds_func);
  pthread_mutex_init(&nres->lock,NULL);
  nres->pid=pid;
  nres->uid=uid;
  nres->gid=gid;
  nres->ref_count=1;
  nres->state=0;

  /* now insert this to global list */
  usrtc_insert(res_list,node,nres);
  pthread_mutex_unlock(&res_lock);

  return nres;
}

/* NOTE: resource should be locked ! */
int add_descriptor_task_resource(task_res_t *res,int fd,
				 vfs_mount_point_t *dest_fs,
				 int flags)
{
  usrtc_node_t *node;
  vfs_fd_t *vfd;

  node=malloc(sizeof(*node));
  vfd=malloc(sizeof(*vfd));
  if(!node || !fd) 
    return ENOMEM;
  
  usrtc_node_init(node,vfd);
  vfd->fd=fd;
  vfd->owner=dest_fs;
  vfd->flags=flags;
  usrtc_insert(res->fds,node,&vfd->fd);

  return 0;
}

/* lookup */
vfs_fd_t *task_resource_lookup(task_res_t *res,int fd)
{
  usrtc_node_t *node;

  node=usrtc_lookup(res->fds,&fd);
  if(!node)
    return NULL;
  else
    return (vfs_fd_t*)usrtc_node_getdata(node);
}

/* delete, TODO: optimize this one  */
int task_resource_delete(task_res_t *res,vfs_fd_t *tvfd)
{
  usrtc_node_t *node;

  node=usrtc_lookup(res->fds,&tvfd->fd);
  usrtc_delete(res->fds,node);
  free(tvfd);
  free(node);

  return 0;
}

void lock_task_resource(task_res_t *res)
{
  if(!res)
    return;
  pthread_mutex_lock(&res->lock);

  return;
}

void unlock_task_resource(task_res_t *res)
{
  if(!res)
    return;
  pthread_mutex_unlock(&res->lock);

  return;
}


/* local functions - misc (usrtc used) */
static int __cmp_fds_func(const void *a,const void *b)
{
  int *aa=(int *)a;
  int *bb=(int *)b;

  return *aa-*bb;
}

static int __cmp_res_func(const void *a,const void *b)
{
  task_res_t *aa=(task_res_t*)a;
  task_res_t *bb=(task_res_t*)b;

  return aa->pid - bb->pid;
}
