/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system,
 * main entry 
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <datalib/usrtc.h>
#include <errno.h>
#include <unistd.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <ns/ns.h>
#include <pthread.h>
#include <termio/term.h>

/* local headers */
#include "vfs_glob.h"
#include "vfs_rpc.h"

/* defines */
#define KQUEUE_SIZE  512
#define KSERV_FLAGS  (IPC_BLOCKED_ACCESS)
#define KPORT_MSGSZ  4096

/* globals */
prtid_t serv_port;

static void __serv_repl_err(uni_rpc_t *rpc,int err,pid_t pid);

int main(int argc,char **argv)
{
  int r;
  char *serv_buf;
  uni_rpc_t *irpc;
  term_loc_t wterm;
  pid_t cpid;
  vfs_path_loc_t *loc;
  ipc_port_msg_info_t minfo;
  void *data=NULL;

  serv_port=ipc_port_open(KSERV_FLAGS,KQUEUE_SIZE);
  if(serv_port<0) {
    abort();
  }

  serv_buf=malloc(KPORT_MSGSZ);
  loc=malloc(sizeof(*loc));
  if(!serv_buf) {
    ipc_port_close(serv_port);
    abort();
  } else
    memset(serv_buf,0,KPORT_MSGSZ);

  /* init all */
  r=init_queue();
  if(r!=0) {
  __main_failed:
    free(serv_buf);
    ipc_port_close(serv_port);
    abort();
  }
  r=init_resources();
  if(r!=0) {
    goto __main_failed;
  }
  r=init_mounts();
  if(r!=0) {
    goto __main_failed;
  }
  r=init_service();
  if(r!=0) {
    goto __main_failed;
  }

  /* make registration on ns */
  wterm.type=PHY_TERM;
  wterm.flags=0;
  wterm.id=0;
  ns_register("vfs.x",0,0,wterm,PCR_SERVER,VFS_SERVER); /* register on ns */

  /* start to listen */
  while(1) {
    r=ipc_port_rcvmsg(serv_port,KPORT_MSGSZ,serv_buf);
    if(r<0) {
      /* fucking deal */
    } else {
      data=NULL;
      minfo=ipc_port_getinfo(serv_port);
      cpid=minfo.sender_pid;
      irpc=(uni_rpc_t *)serv_buf;
      if(minfo.msg_len<sizeof(*irpc)) {
	__serv_repl_err(irpc,EINVAL,cpid);
	continue;
      }
      switch(irpc->rpc_call) {
      case RPC_ALIVE:
	__serv_repl_err(irpc,RPC_ALIVE,cpid);
	break;
      case RPC_REGISTER:
	r=register_fs(serv_buf,minfo.msg_len,minfo.msg_id);
	if(r!=0) {
	  __serv_repl_err(irpc,r,cpid);
	} else
	  service_atom_run();
	break;
      case RPC_OPEN:
      case RPC_CLOSE:
      case RPC_OPENDIR:
      case RPC_CLOSEDIR:
      case RPC_STAT:
      case RPC_GETCWD:
	/* copy message data */
	data=malloc(r);
	if(!data) {
	  __serv_repl_err(irpc,ENOMEM,cpid);
	  continue;
	}
	memcpy(data,serv_buf,r);
	/* add task to the worker queue */
	r=add_task(irpc->rpc_call,loc,0,0,minfo.sender_pid,
		   minfo.msg_id,data,r);
	if(r!=0)
	  __serv_repl_err(irpc,r,cpid);
	else
	  queue_atom_run();
	break;
      case RPC_DUP:
      case RPC_DUP2:
      case RPC_SYNC:
	__serv_repl_err(irpc,ENOSYS,cpid);
	break;
      case RPC_MOUNT:
	r=add_mount_task(serv_buf,minfo.msg_len,minfo.msg_id);
	if(r!=0) {
	  __serv_repl_err(irpc,r,cpid);
	} else
	  service_atom_run();
	break;
      case RPC_UMOUNT:
      case RPC_PIDFDGET:
      case RPC_PIDFDCLONE:
	__serv_repl_err(irpc,ENOSYS,cpid);
	break;
      default:
	__serv_repl_err(irpc,EINVAL,cpid);
	break;
      }
      /* uffgh */
    }
  }

  /* shutdown */

  return 0;
}

static void __serv_repl_err(uni_rpc_t *rpc,int err,pid_t pid)
{
  int r;

  rpc->rpc_call=err;
  r=ipc_port_reply(serv_port,sizeof(*rpc),(void *)rpc);
  if(r<0 && r==-EPIPE)
    gc_cleanup_pid(pid);

  return;
}
