/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system,
 * queues
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/uevent.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <datalib/usrtc.h>
#include <errno.h>
#include <unistd.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <ns/ns.h>
#include <pthread.h>
#include <termio/term.h>

/* local headers */
#include "vfs_glob.h"
#include "vfs_rpc.h"

#define SERV_ATOMS  4

usrtc_t *serv_queue;
pthread_mutex_t squeue_lock;
event_t sequeue_atom;

/* local functions prototypes */
static int __srv_queue_cmp(const void *a,const void *b);
static void *__service_worker(void *args);
static serv_task_t *take_job(void);

#define __atom_wait  event_wait(&sequeue_atom)

int init_service(void)
{
  int i,r;
  pthread_t qap[SERV_ATOMS];

  serv_queue=malloc(sizeof(*serv_queue));
  if(!serv_queue)
    return -1;

  usrtc_init(serv_queue,USRTC_LIST,USRTC_COUNT_T_MAX,__srv_queue_cmp);

  pthread_mutex_init(&squeue_lock,NULL);
  event_init(&sequeue_atom);

  for(i=0;i<SERV_ATOMS;i++) {
    r=pthread_create(&qap[i],NULL,__service_worker,(void *)serv_queue);
    if(r!=0)
      return -1;
  }

  return 0;
}

int add_service_job(serv_task_t *job)
{
  usrtc_node_t *node;

  node=malloc(sizeof(*node));
  if(!node)
    return ENOMEM;
  else
    usrtc_node_init(node,job);

  usrtc_insert(serv_queue,node,job);

  return 0;
}

static serv_task_t *take_job(void)
{
  usrtc_node_t *node;
  serv_task_t *task;

  node=usrtc_first(serv_queue);
  if(node!=NULL) {
    task=usrtc_node_getdata(node);

    service_queue_lock();
    usrtc_delete(serv_queue,node);
    service_queue_unlock();

    free(node);
    return task;
  } else 
    return NULL;
}

void service_queue_lock(void)
{
  pthread_mutex_lock(&squeue_lock);

  return;
}

void service_queue_unlock(void)
{
  pthread_mutex_unlock(&squeue_lock);

  return;
}

void service_atom_run(void)
{
  event_signal(&sequeue_atom);
  return;
}

static void *__service_worker(void *args)
{
  int r;
  serv_task_t *myjob;
  uni_rpc_t rpc;
  vfs_mount_request_t request;
  vfs_mount_point_t *pp;

  while(1) {
    r=__atom_wait;
    if(r!=0)      continue;
    myjob=take_job();
    if(!myjob)       continue;    
    /* do some job */
    switch(myjob->vfs_cmd) {
    case RPC_REGISTER:
      rpc.rpc_call=0;
      ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,myjob->kmsgid);
      pp=(vfs_mount_point_t*)(myjob->data);
      if(!wcscmp(pp->info->fs_name,L"initfs")) {
	request.flags=0;
	request.devno=NULL;
	request.point=strdup("/");
	request.fstype=(char *)wcsdup(L"initfs");
	r=vfs_mount(&request);
	free(request.point);
	free(request.fstype);
      }
      break;
    case RPC_MOUNT:
      rpc.rpc_call=vfs_mount((vfs_mount_request_t *)myjob->data);
      ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,myjob->kmsgid);
      break;
    default:
      rpc.rpc_call=ENOSYS;
      ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,myjob->kmsgid);
      break;
    }
  }

  return NULL;
}

static int __srv_queue_cmp(const void *a,const void *b)
{
  serv_task_t *aa=(serv_task_t *)a;
  serv_task_t *bb=(serv_task_t *)b;

  return aa->kmsgid - bb->kmsgid;
}
