/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * filesystem operations for iterate
 * with file system, includes proto and
 * internal point lookup
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/iov.h>
#include <ipc/rpc.h>
#include <pthread.h>
#include <datalib/usrtc.h>
#include <limits.h>

#include "vfsconfig.h"
#include "resource.h"
#include "fsops.h"

#ifdef CONFIG_VFSV2_DEBUG
#define CONFIG_V2_DEBUG  1
#endif

#ifdef CONFIG_V2_DEBUG
/* debug */
#include "fslog.h"
#endif

/* local functions */
static int __vfs_resolve_symlink(vfs_path_t *path,struct stat *buf);
static ulong_t __vfs_symhash(char *path);

/* operations without fds */
int vfs_lookup(vfs_path_t *path,struct stat *buf)
{
  link_point_t *link;
  uni_rpc_t rpc;
  iovec_t iov[3];
  iovec_t r_iov[2];
  int edeep=0;
  int eloop=0,r;

  link=vfs_point_lookup(path,&eloop,&edeep);
  if(!link)
    return ENOENT;

  if(eloop) {
    memcpy(buf,&link->pp_stat,sizeof(struct stat));
    rpc.rpc_call=0;
  } else {
    /* prepare RPC header */
    rpc.rpc_call=RPC_STAT;
    rpc.arg5=path->pp; /* number of points */
    rpc.arg6=path->plen; /* path length */
    rpc.offset=edeep; /* offset of moved root */
    iov[0].iov_base=&rpc;
    iov[0].iov_len=sizeof(rpc);
    /* parsed points */
    iov[1].iov_base=path->pps;
    iov[1].iov_len=path->pp*sizeof(int);
    /* real path */
    iov[2].iov_base=path->path;
    iov[2].iov_len=path->plen;
    /* prepare receive iovec */
    r_iov[0].iov_base=&rpc; /* rpc header */
    r_iov[0].iov_len=sizeof(rpc);
    r_iov[1].iov_base=buf;
    r_iov[1].iov_len=sizeof(struct stat);
    r=ipc_iov_iov_send(link->end_point->channel,iov,3,r_iov,2);
    if(r<0)       return -r;
  }

  return rpc.rpc_call;
}

int vfs_lookup_resolve(vfs_path_t *path,struct stat *buf)
{
  link_point_t *link;
  uni_rpc_t rpc;
  iovec_t iov[3];
  iovec_t r_iov[2];
  int edeep=0;
  int eloop=0,r;
  int err;

  link=vfs_point_lookup(path,&eloop,&edeep);
  if(!link)
    return ENOENT;

  if(eloop) {
    memcpy(buf,&link->pp_stat,sizeof(struct stat));
    rpc.rpc_call=0;
  } else {
    /* prepare RPC header */
    rpc.rpc_call=RPC_STAT;
    rpc.arg5=path->pp; /* number of points */
    rpc.arg6=path->plen; /* path length */
    rpc.offset=edeep; /* offset of moved root */
    iov[0].iov_base=&rpc;
    iov[0].iov_len=sizeof(rpc);
    /* parsed points */
    iov[1].iov_base=path->pps;
    iov[1].iov_len=path->pp*sizeof(int);
    /* real path */
    iov[2].iov_base=path->path;
    iov[2].iov_len=path->plen;
    /* prepare receive iovec */
    r_iov[0].iov_base=&rpc; /* rpc header */
    r_iov[0].iov_len=sizeof(rpc);
    r_iov[1].iov_base=buf;
    r_iov[1].iov_len=sizeof(struct stat);
    r=ipc_iov_iov_send(link->end_point->channel,iov,3,r_iov,2);
    if(r<0)       return -r;
    if(rpc.rpc_call==0 && S_ISLNK(buf->st_mode)) {
      /* oops, it's a link - should resolve */
      r=__vfs_resolve_symlink(path,buf);
      err=r;
    } else {
      err=rpc.rpc_call;
      goto __resolve_end;
    }
  }

 __resolve_end:
  return err;
}

/* This call modify vfs_path_t fully with normalization */
int vfs_readlink_solve(vfs_path_t *path)
{
  char *rpath=NULL;
  char *spath=NULL,*gp;
  char *src,*dst,s;
  char *end;
  int err=0,len,j,i;
  int ppln=0;

  if((spath=strdup(path->path))==NULL) { /* copy old location */
    err=ENOMEM;
    goto __out_solve;
  }
  err=vfs_readlink(path); /* read this link */
  if(err!=0) goto __out_solve;

  /* first we will normalize path according to original path */
  /* this code repeat code from ulibc(jarifs part), but it's internally used */
  /* and has other behaivor - this is a code style FIXME */
  if((rpath=malloc(PATH_MAX))==NULL) {
    err=ENOMEM;
    goto __out_solve;
  } else memset(rpath,0,PATH_MAX);

  /* setup pointers */
  src=path->path;
  dst=rpath;

  /* take a look for path start */
  if(*src=='/') {
    src++;
    *dst++='/';
  } else {
    gp=spath+(strlen(spath)-1);
    while(*gp!='/') {
      ppln++;
      gp-=sizeof(char);
    }
    strncpy(dst,spath,strlen(spath)-ppln);
    dst=rpath+strlen(rpath);
    if(*(dst-1)!='/') *dst++='/';
  }
  /* well, going to change it */
  while (*src != '\0') {
    end = strchr(src, '/');
    if (end == NULL)
      end = src + strlen(src);
    s = *end;
    *end = '\0';

    if ((strlen(src) == 0) || (strcmp(src, ".") == 0))
      goto next_part;
    if (strcmp(src, "..") == 0) {
      dst -= 2;
      while ((dst >= rpath) && (*dst != '/'))        dst--;
      if (dst < rpath) {
        /* Can't step back so far */
        *end = s;
        err = ENOENT;
        goto __out_solve;
      }
    } else {
      if (dst + strlen(src) >= rpath + PATH_MAX) {
        /* Path is too long */
        *end = s;
        err = ENAMETOOLONG;
        goto __out_solve;
      }
      strcat(dst, src);
      dst += strlen(src);
    }

    *dst++ = '/';
    *dst = '\0';

  next_part:
    *end = s;
    src = end;
    if (*src == '/')
      src++;
  }

  /* Cut off trailing slash */
  if (strcmp(rpath, "/") != 0)
    dst--;
  *dst = '\0';

  /* NOTE: *for reviewers* we always sure that path->pps
   * and path->path is a really one pointer i.e.
   * <pps><path> and it's has a size to overflow in any case
   * it's a bit sophiscated, but can avoid many memory ops
   */
  len=strlen(rpath);
  j=0;

  path->pp=0;
  for(i=0;i<len;i++)
    if(rpath[i]=='/') {
      path->pps[j]=i; j++;
    }
  path->pp=j;

  /* ok, now let's shift all this stuff */
  path->path=(char *)path->pps+(sizeof(int)*path->pp);
  strcpy(path->path,rpath);
  path->plen=len;
  /* new path is ready to lookup */
  //  klog(path->path);

 __out_solve:
  free(rpath);
  free(spath);

  return err;
}

/* This call modify vfs_path_t -> path field with symlink content, but it doesn't change other things */
int vfs_readlink(vfs_path_t *path)
{
  link_point_t *link;
  uni_rpc_t rpc;
  iovec_t iov[3];
  iovec_t r_iov[2];
  int edeep=0;
  int eloop=0,r;

  link=vfs_point_lookup(path,&eloop,&edeep);
  if(!link)
    return ENOENT;

  if(eloop) {
    return EINVAL;
  } else {
    /* prepare RPC header */
    rpc.rpc_call=RPC_READLINK;
    rpc.arg5=path->pp; /* number of points */
    rpc.arg6=path->plen; /* path length */
    rpc.offset=edeep; /* offset of moved root */
    iov[0].iov_base=&rpc;
    iov[0].iov_len=sizeof(rpc);
    /* parsed points */
    iov[1].iov_base=path->pps;
    iov[1].iov_len=path->pp*sizeof(int);
    /* real path */
    iov[2].iov_base=path->path;
    iov[2].iov_len=path->plen;
    /* prepare receive iovec */
    r_iov[0].iov_base=&rpc; /* rpc header */
    r_iov[0].iov_len=sizeof(rpc);
    r_iov[1].iov_base=path->path;
    r_iov[1].iov_len=PATH_MAX;
    r=ipc_iov_iov_send(link->end_point->channel,iov,3,r_iov,2);
    if(r<0)       return -r;
  }

  return rpc.rpc_call;
}

/*
 * NOTE: this functions tries to take info on each ongoing link,
 * it has a two limits - max deep symlink level and infinite loop
 */
static int __vfs_resolve_symlink(vfs_path_t *path,struct stat *buf)
{
  ulong_t *path_h=NULL,chash;
  int i,r,j=0,k;
  int err=0;

  if((path_h=malloc(sizeof(ulong_t)*SYMLEVEL))==NULL) {
    err=ENOMEM;
    goto __out_ressym;
  }

  /* ok, we will encount our limit */
  for(i=0;i<SYMLEVEL;i++) {
    path_h[j++]=__vfs_symhash(path->path); /* save hash */
    /* modify vfs_path_t for new data */
    if((r=vfs_readlink_solve(path))!=0) {
      err=r;
      goto __out_ressym;
    }
    chash=__vfs_symhash(path->path);
    for(k=0;k<j;k++) {
      if(chash==path_h[k])
        err=ELOOP; /* oops it was already */
        goto __out_ressym;
      }
    if((r=vfs_lookup(path,buf))!=0) {
      err=r;
      goto __out_ressym;
    }
    if(!S_ISLNK(buf->st_mode)) { /* resolved */
      err=0;
      goto __out_ressym;
    }
  }

  err=ELOOP;

 __out_ressym:
  free(path_h);

  return err;
}

int vfs_mkdir(vfs_path_t *path,int mode,uid_t uid,gid_t gid)
{
  link_point_t *link;
  uni_rpc_t rpc;
  iovec_t iov[3];
  int edeep=0;
  int eloop=0,r;

  link=vfs_point_lookup(path,&eloop,&edeep);
  if(!link) 
    return ENOENT;
  if(eloop)
    return EEXIST;

  /* prepare RPC header */
  rpc.rpc_call=RPC_CREATEDIR;
  rpc.arg0=mode; /* mode for directory */
  rpc.arg3=uid; /* UID of creator */
  rpc.arg4=gid; /* GID of creator */
  rpc.arg5=path->pp; /* number of points */
  rpc.arg6=path->plen; /* path length */
  rpc.offset=edeep; /* offset of moved root */
  iov[0].iov_base=&rpc;
  iov[0].iov_len=sizeof(rpc);
  /* parsed points */
  iov[1].iov_base=path->pps;
  iov[1].iov_len=path->pp*sizeof(int);
  /* real path */
  iov[2].iov_base=path->path;
  iov[2].iov_len=path->plen;
  /* send it */
  r=ipc_iov_send(link->end_point->channel,iov,3,&rpc,sizeof(rpc));
  if(r<0)     return -r;

  return rpc.rpc_call;
}

int vfs_unlink(vfs_path_t *path,uid_t uid,gid_t gid)
{
  link_point_t *link;
  uni_rpc_t rpc;
  iovec_t iov[3];
  int edeep=0;
  int eloop=0,r;

  link=vfs_point_lookup(path,&eloop,&edeep);
  if(!link) 
    return ENOENT;
  if(eloop)
    return EISDIR;

  /* prepare RPC header */
  rpc.rpc_call=RPC_UNLINK;
  rpc.arg3=uid; /* UID of caller */
  rpc.arg4=gid; /* GID of caller */
  rpc.arg5=path->pp; /* number of points */
  rpc.arg6=path->plen; /* path length */
  rpc.offset=edeep; /* offset of moved root */
  iov[0].iov_base=&rpc;
  iov[0].iov_len=sizeof(rpc);
  /* parsed points */
  iov[1].iov_base=path->pps;
  iov[1].iov_len=path->pp*sizeof(int);
  /* real path */
  iov[2].iov_base=path->path;
  iov[2].iov_len=path->plen;
  /* send it */
  r=ipc_iov_send(link->end_point->channel,iov,3,&rpc,sizeof(rpc));
  if(r<0)     return -r;

  return rpc.rpc_call;
}

int vfs_unlinkdir(vfs_path_t *path,uid_t uid,gid_t gid)
{
  link_point_t *link;
  uni_rpc_t rpc;
  iovec_t iov[3];
  int edeep=0;
  int eloop=0,r;

  link=vfs_point_lookup(path,&eloop,&edeep);
  if(!link) 
    return ENOENT;
  if(eloop)
    return EBUSY;

#ifdef CONFIG_V2_DEBUG
  console_printf("[v2] vfs_unlinkdir-> `%s' \n",path->path);
#endif
  /* prepare RPC header */
  rpc.rpc_call=RPC_UNLINKDIR;
  rpc.arg3=uid; /* UID of caller */
  rpc.arg4=gid; /* GID of caller */
  rpc.arg5=path->pp; /* number of points */
  rpc.arg6=path->plen; /* path length */
  rpc.offset=edeep; /* offset of moved root */
  iov[0].iov_base=&rpc;
  iov[0].iov_len=sizeof(rpc);
  /* parsed points */
  iov[1].iov_base=path->pps;
  iov[1].iov_len=path->pp*sizeof(int);
  /* real path */
  iov[2].iov_base=path->path;
  iov[2].iov_len=path->plen;
  /* send it */
  r=ipc_iov_send(link->end_point->channel,iov,3,&rpc,sizeof(rpc));
  if(r<0)     return -r;
#ifdef CONFIG_V2_DEBUG
  console_printf("[v2] vfs_unlinkdir->%d\n",rpc.rpc_call);
#endif

  return rpc.rpc_call;
}

/* operations fd related */
/* aff_am - offset within vfs_path */
int vfs_open(user_link_t *ulink,link_point_t *fslink,vfs_path_t *mould,
	     int aff_am,uid_t uid,gid_t gid,pid_t pid)
{
  uni_rpc_t rpc;
  iovec_t iov[3];
  int r,a;

  /* prepare RPC call header */
  rpc.rpc_call=RPC_OPEN;
  rpc.arg0=ulink->flags;
  rpc.arg1=ulink->mode;
  rpc.arg2=pid;
  rpc.arg3=uid;
  rpc.arg4=gid;
  rpc.arg5=mould->pp;
  rpc.arg6=mould->plen;
  rpc.size=ulink->u_fd;
  rpc.offset=aff_am;
  iov[0].iov_base=&rpc;
  iov[0].iov_len=sizeof(rpc);
  iov[1].iov_base=mould->pps;
  iov[1].iov_len=mould->pp*sizeof(int);
  iov[2].iov_base=mould->path;
  iov[2].iov_len=mould->plen;

  r=ipc_iov_send(fslink->end_point->channel,iov,3,&rpc,sizeof(rpc));
  if(r<0)     a=-r;
  else    a=rpc.rpc_call;
  /* TODO: add links redirection */
  
  ulink->ufsid=rpc.arg4;
  ulink->pid=rpc.arg2; /*pid*/
  ulink->port_id=rpc.arg3; /*portid*/

  return a;
}

int vfs_close(user_gate_t *sgate,user_link_t *slink)
{
  uni_rpc_t rpc;
  int r,a;

  /* complete rpc */
  rpc.rpc_call=RPC_CLOSE;
  rpc.arg0=slink->u_fd;
  rpc.arg1=slink->ufsid;
  rpc.arg2=sgate->ug_pid;
  rpc.arg3=sgate->ug_uid;
  rpc.arg4=sgate->ug_gid;

  r=ipc_channel_msgsnd(slink->point->end_point->channel,sizeof(rpc),&rpc,sizeof(rpc),&rpc);
  if(r<0)     a=-r;
  else    a=rpc.rpc_call;

  return a;
}

/* system operations */
int vfs_mount(link_point_t *ulink,vfs_path_t *d_loc,int spec_flags)
{
  iovec_t iov[3];
  uni_rpc_t rpc;
  int r,a;

  rpc.rpc_call=RPC_MOUNT;
  rpc.arg0=spec_flags;

  if(!d_loc) { /* seems to be a system fs */
    rpc.arg1=0;
    r=ipc_channel_msgsnd(ulink->end_point->channel,sizeof(rpc),&rpc,sizeof(rpc),&rpc);
    if(r<0)     a=-r;
    else    a=rpc.rpc_call;
  } else { /* regular fs */
    /* first prepare RPC header */
    rpc.arg1=0;
    rpc.arg2=d_loc->pp;
    rpc.arg3=d_loc->plen;
    iov[0].iov_base=&rpc;
    iov[0].iov_len=sizeof(rpc);
    /* points */
    iov[1].iov_base=d_loc->pps;
    iov[1].iov_len=d_loc->pp*sizeof(int);
    /* path */
    iov[2].iov_base=d_loc->path;
    iov[2].iov_len=d_loc->plen;
    r=ipc_iov_send(ulink->end_point->channel,iov,3,&rpc,sizeof(rpc));
    if(r<0)     a=-r;
    else    a=rpc.rpc_call;
  }

  return a;
}

/* sync all io */
void vfs_sync(link_point_t *ulink)
{
  uni_rpc_t rpc;
  int r;

  if(!ulink)    return;

  rpc.rpc_call=RPC_SYNC;
  r=ipc_channel_msgsnd(ulink->end_point->channel,sizeof(rpc),&rpc,sizeof(rpc),&rpc);
  /* TODO: add cleaner logic if channel was broken */

  return;
}

/* clone resources */
void vfs_pidfdclone(link_point_t *ulink,pid_t ppid,pid_t npid)
{
  uni_rpc_t rpc;
  int r;

  if(!ulink)    return;

  rpc.rpc_call=RPC_PIDFDCLONE;
  rpc.arg0=ppid;
  rpc.arg1=npid;
  r=ipc_channel_msgsnd(ulink->end_point->channel,sizeof(rpc),&rpc,sizeof(rpc),&rpc);
  /* TODO: add cleaner logic if channel was broken */

  return;
}

/* reset resources */
void vfs_pidfdreset(link_point_t *ulink,pid_t ppid)
{
  uni_rpc_t rpc;
  int r;

  if(!ulink)    return;

  rpc.rpc_call=RPC_PIDFDRESET;
  rpc.arg0=ppid;
  r=ipc_channel_msgsnd(ulink->end_point->channel,sizeof(rpc),&rpc,sizeof(rpc),&rpc);
  /* TODO: add cleaner logic if channel was broken */

  return;
}

/* kill resources */
void vfs_pidfdkill(link_point_t *ulink,pid_t ppid)
{
  uni_rpc_t rpc;
  int r;

  if(!ulink)    return;

  rpc.rpc_call=RPC_PIDFDKILL;
  rpc.arg0=ppid;
  r=ipc_channel_msgsnd(ulink->end_point->channel,sizeof(rpc),&rpc,sizeof(rpc),&rpc);
  /* TODO: add cleaner logic if channel was broken */

  return;
}

/* local functions implementation */
static ulong_t __vfs_symhash(char *path)
{
  ulong_t hash=0;

  if(!path) return hash;
  while(*path)
    hash=(hash << 5)+hash + *path++;

  return hash;
}
