/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * vfs points manipulation
 *
 */

#include <sys/types.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ipc/channel.h>
#include <ipc/port.h>
#include <pthread.h>
#include <unistd.h>
#include <datalib/usrtc.h>

#include <fsio/fsspec.h>

#include "resource.h"
#include "fsops.h"

#ifdef CONFIG_VFSV2_DEBUG
#define CONFIG_V2_DEBUG  1
#endif

#ifdef CONFIG_V2_DEBUG
/* debug */
#include <dircons.h>
#endif

typedef struct __vfs_pnode_t {
  int level;
  int nnodes;
  int cnodes;
  pthread_mutex_t lock;
  link_point_t *point;
  vfs_path_t *path;
  usrtc_t *srch;
} vfs_pnode_t;

#define pnode_lock(p)    pthread_mutex_lock(&p->lock)
#define pnode_unlock(p)  pthread_mutex_unlock(&p->lock)

static link_point_t *raw_list[V2_MAX_LINKS];
static pthread_mutex_t list_lock;
static int n_links=0;
static vfs_pnode_t *vfs_root=NULL;

#define raw_list_lock    pthread_mutex_lock(&list_lock)
#define raw_list_unlock  pthread_mutex_unlock(&list_lock)

/* local ops */
static int __comp_pnode_str(const void *a,const void *b);
static vfs_pnode_t *__vfs_pnode_lookup(vfs_path_t *vfs_path,int *ts,int *ed);
static link_point_t *__vfs_find_via_typenm(char *fstype);

/* initialization function */
int vfs_point_init(void)
{
  int i=0;

  pthread_mutex_init(&list_lock,NULL);

  for(i=0;i<V2_MAX_LINKS;i++) 
    raw_list[i]=NULL;
  
  n_links=0;

  return 0;
}

/* register new point (actually doesn't mounts it) */
int vfs_point_register(pid_t fs_pid,prtid_t fs_port,char *fs_type,
		       char *fs_author,char *fs_info,int spec_flag)
{
  int i=0;
  link_point_t *n_link=NULL;
  vfs_info_t *n_info=NULL;
  vfs_point_t *n_point=NULL;

  if(!fs_type)
    return EINVAL;

  raw_list_lock;

  for(i=0;i<V2_MAX_LINKS;i++) {
    if(raw_list[i]==NULL) {
      n_link=malloc(sizeof(*n_link));
      raw_list[i]=n_link;
      break;
    }
  }
  if(i==V2_MAX_LINKS) {
    raw_list_unlock;
    return ENOSLOT;
  }
  if(!n_link) {
    raw_list_unlock;
    return ENOMEM;
  }
  /* NOTE: fs-> author and info may be nil */
  /* take an info */
  n_info=malloc(sizeof(*n_info));
  if(!n_info) {
  __enomem:
    free(n_link);
    raw_list[i]=NULL;
    raw_list_unlock;
    return ENOMEM;
  }
  if((n_info->fstype=strdup(fs_type))==NULL) 
    goto __enomem;
  if(!fs_author)
    n_info->author=NULL;
  else if((n_info->author=strdup(fs_author))==NULL) {
    free(n_info->fstype);
    goto __enomem;
  }
  if(!fs_info)
    n_info->info=NULL;
  else if((n_info->info=strdup(fs_info))==NULL) {
    free(n_info->fstype);
    if(n_info->author)      free(n_info->author);
    goto __enomem;
  }
  if((n_point=malloc(sizeof(*n_point)))==NULL) {
    free(n_info->fstype);
    if(n_info->author)      free(n_info->author);
    if(n_info->info)      free(n_info->info);
    free(n_info);
    goto __enomem;
  }
  /* pid, port, channel */
  n_point->pid=fs_pid;
  n_point->channel=ipc_channel_open(fs_pid,fs_port,IPC_BLOCKED_ACCESS);
  if(n_point->channel<0) {
    /* fucking kernel ? libc ? */
    free(n_point);
    free(n_info->fstype);
    if(n_info->author)      free(n_info->author);
    if(n_info->info)      free(n_info->info);
    free(n_info);
    free(n_link);
    raw_list[i]=NULL;
    raw_list_unlock;
    return EIO; /* it seems to be input/output error */
  }
  n_point->port=fs_port;
  /* init link */
  n_link->ref_count=0;
  n_link->vfs_flags=0;
  n_link->end_point=n_point;
  n_link->fs_point_info=n_info;
  n_link->owner=NULL;
  n_link->vfs_flags|=spec_flag;
  n_link->state=0; /* not mounted yet */

  raw_list_unlock;

  return 0;
}

/* unregister fs */
int vfs_point_unregister(link_point_t *fs_linkpoint)
{
  /* end_point */
  /* for those who has bugmania - in this case, within this contex *I DONT CARE ABOUT ERROR ON PORT CLOSE* */
  ipc_channel_close(fs_linkpoint->end_point->channel);
  free(fs_linkpoint->end_point);
  /* owner */
  if(fs_linkpoint->owner) /* FIXME: after adding specific MAC rule - aa! WARN! */
    free(fs_linkpoint->owner);
  /* fsinfo */
  free(fs_linkpoint->fs_point_info->fstype);
  if(fs_linkpoint->fs_point_info->info)
    free(fs_linkpoint->fs_point_info->info);
  if(fs_linkpoint->fs_point_info->author)
    free(fs_linkpoint->fs_point_info->author);

  return 0;
}

/* unregister fs by pid */
int vfs_point_unregister_p(pid_t caller)
{
  int i;
  link_point_t *pm=NULL;

  raw_list_lock;
  /* look up for registered file system */
  for(i=0;i<V2_MAX_LINKS;i++) {
    if(raw_list[i]==NULL)
      break;
    else {
      if(raw_list[i]->end_point->pid==caller && !raw_list[i]->state) {
	pm=raw_list[i];
	raw_list[i]=NULL;
	break;
      } else if(raw_list[i]->end_point->pid==caller && raw_list[i]->state) {
	raw_list_unlock;
	return EBUSY; /* you cannot be unregistered while file system is mounted */
      } else
	continue;
    }
  }
  raw_list_unlock;
  if(pm==NULL)
    return ENOENT; /* freaky, fucking deal */

  return vfs_point_unregister(pm);
}

int vfs_point_sync(void)
{
  int i;

  /* NOTE: to those who likes shit flame - going via raw list really faster then via pnodes */
  /* pnodes is a tree - not binary , it's a hiearchical structure */

  raw_list_lock;
  for(i=0;i<V2_MAX_LINKS;i++) {
    if(raw_list[i]==NULL)
      break;
    else {
      if(raw_list[i]->state > 0) {
	vfs_sync(raw_list[i]);
      } else
	continue;
    }
  }
  raw_list_unlock;

  return 0;
}

int vfs_point_mount_l(link_point_t *pm,vfs_path_t *m_loc,int flags)
{
  struct stat sbuf;
  vfs_pnode_t *root;
  vfs_pnode_t *new;
  vfs_path_t *npath;
  usrtc_t *srch;
  usrtc_node_t *anode;
  int sr;
  int tbit=0,dloop=0;
  struct stat vstat;

  raw_list_lock;

  /* lookup for root point */
  if(!vfs_root) { /* no root file system */
    if(m_loc->pp>1) {
      raw_list_unlock;
      return ENOENT;
    }
    if(strcmp(m_loc->path,(const char *)V2_SEP)) {
      raw_list_unlock;
      return ENOENT; /* fucking deal, you are trying to mount something without root */
    }
    if((srch=malloc(sizeof(*srch)))==NULL) {
      raw_list_unlock;
      return ENOMEM;
    } else {
#ifdef CONFIG_VFSV2_REALTIME
      usrtc_init(srch,USRTC_AVL,USRTC_COUNT_T_MAX,__comp_pnode_str);
#else
      usrtc_init(srch,USRTC_SPLAY,USRTC_COUNT_T_MAX,__comp_pnode_str);
#endif
    }
    /* it's a simple case to make mount */
    if((vfs_root=malloc(sizeof(*vfs_root)))==NULL) {
      raw_list_unlock;
      free(srch);
      return ENOMEM;
    }
    if((vfs_root->path=malloc(sizeof(*vfs_root->path)))==NULL) {
      free(srch);
      free(vfs_root);
      vfs_root=NULL;
      return ENOMEM;
    }
    if((vfs_root->path->path=strdup((const char*)"/"))==NULL) {
      free(vfs_root->path);
      free(srch);
      free(vfs_root);
      vfs_root=NULL;
      return ENOMEM;
    }
    pthread_mutex_init(&vfs_root->lock,NULL);
    pnode_lock(vfs_root);
    vfs_root->level=1;
    vfs_root->nnodes=0;
    vfs_root->cnodes=0;
    vfs_root->srch=srch;
    vfs_root->point=pm;
    pm->state=1; /* mounted */
    pm->vfs_flags|=flags;
    /* make a correct stat buffer */
    vstat.st_uid=0;
    vstat.st_gid=0;
    vstat.st_mode=0;
    vstat.st_mode|=(00000755);
    vstat.st_mode|=S_IFDIR;
    memcpy(&pm->pp_stat,&vstat,sizeof(struct stat));
    pnode_unlock(vfs_root);
  } else {
#ifdef CONFIG_V2_DEBUG
    console_printf("[v2] {point_mount_l} try to lookup ... \n");
#endif
    root=__vfs_pnode_lookup(m_loc,&tbit,&dloop);
#ifdef CONFIG_V2_DEBUG
    console_printf("[v2] {point_mount_l} root=%p ... \n",root);
#endif
    if(!root) {
      raw_list_unlock;
      return ENOENT;
    }
    if(tbit==1) {
      raw_list_unlock;
      return EEXIST;
    }
    sr=vfs_lookup(m_loc,&sbuf);
    if(sr!=0) {
      raw_list_unlock;
      return sr;
    } 
    if(!S_ISDIR(sbuf.st_mode)) {
      raw_list_unlock;
      return ENOTDIR;
    }
    if((anode=malloc(sizeof(*anode)))==NULL) {
      raw_list_unlock;
      return ENOMEM;
    }
    if((srch=malloc(sizeof(*srch)))==NULL) {
      free(anode);
      raw_list_unlock;
      return ENOMEM;
    } else {
#ifdef CONFIG_VFSV2_REALTIME
      usrtc_init(srch,USRTC_AVL,USRTC_COUNT_T_MAX,__comp_pnode_str);
#else
      usrtc_init(srch,USRTC_SPLAY,USRTC_COUNT_T_MAX,__comp_pnode_str);
#endif
    }
    /* clone vfs_path */
    if((npath=malloc(sizeof(*npath)))==NULL) {
    __enomem_t:
      free(anode);
      free(srch);
      raw_list_unlock;
      return ENOMEM;
    } else {
      npath->pp=m_loc->pp;
      npath->plen=m_loc->plen;
      npath->pps=NULL;
      npath->path=NULL;
    }
    if((npath->pps=malloc(sizeof(int)*npath->pp))==NULL) {
      free(npath);
      goto __enomem_t;
    } else
      memcpy(npath->pps,m_loc->pps,sizeof(int)*npath->pp);
    if((npath->path=malloc(sizeof(char)*npath->plen))==NULL) {
      free(npath->pps);
      free(npath);
      goto __enomem_t;
    } else
      strncpy(npath->path,m_loc->path,npath->plen);
    /* create pnode */
    if((new=malloc(sizeof(*new)))==NULL) {
      free(npath->path);
      free(npath->pps);
      free(npath);
      goto __enomem_t;
    }
    pnode_lock(root);
    root->nnodes++; /* increment parent counter */
    new->level=dloop; 
    new->nnodes=0;
    new->cnodes=0;
    new->srch=srch; /* init internal childs tree */
    new->point=pm; /* assign link_point_t from raw list */
    new->path=npath; /* */
    pm->state=1;
    pm->vfs_flags|=flags;
    memcpy(&pm->pp_stat,&sbuf,sizeof(struct stat));
    pthread_mutex_init(&new->lock,NULL);
    /* now let's add this one */
    usrtc_node_init(anode,new);
    usrtc_insert(root->srch,anode,npath->path);
    pnode_unlock(root);
  }

  raw_list_unlock;

  return 0;
}

/* mount fs point */
int vfs_point_mount(char *fstype,vfs_path_t *m_loc,int flags)
{
  int i;
  link_point_t *pm=NULL;
  //  link_point_t *parent=NULL;

  raw_list_lock;

  /* look up for registered file system */
  for(i=0;i<V2_MAX_LINKS;i++) {
    if(raw_list[i]==NULL)
      break;
    else {
      if(!strcmp(raw_list[i]->fs_point_info->fstype,fstype)) {
	pm=raw_list[i];
	break;
      } else
	continue;
    }
  }

  if(!pm) {
    raw_list_unlock;
    return ENOENT; /* no such file system */
  }

  raw_list_unlock;

  return vfs_point_mount_l(pm,m_loc,flags);
}

/* umount point */
int vfs_point_umount(link_point_t *fs_linkpoint)
{
  return 0;
}

/* lookup a link point,if NULL then there are no fses  */
/* TODO: optimize this bottle neck */
link_point_t *vfs_point_lookup(vfs_path_t *vfs_path,int *ts,int *ed)
{
  vfs_pnode_t *rp=NULL;
  link_point_t *lpoint=NULL;
  int tts=0,eed=0; 

  *ts=0;

  rp=__vfs_pnode_lookup(vfs_path,&tts,&eed);
  if(rp!=NULL) {
    *ts=tts; *ed=eed;
    lpoint=rp->point;
  }

  return lpoint;
}

link_point_t *vfs_lookup_nmtype(char *fstype)
{
  return __vfs_find_via_typenm(fstype);
}

/* some local functions */
static int __comp_pnode_str(const void *a,const void *b)
{
  return strcmp((const char *)a,(const char *)b);
}

static link_point_t *__vfs_find_via_typenm(char *fstype)
{
  int i;
  link_point_t *lnk=NULL;

  raw_list_lock;

  for(i=0;i<V2_MAX_LINKS;i++) {
    lnk=raw_list[i];
    if(!lnk)      break;
    else if(lnk->state==0 && !strcmp(fstype,lnk->fs_point_info->fstype)) {
      break;
    }
  }

  raw_list_unlock;

  return lnk;
}

/* TODO: optimize this bottle neck */
static vfs_pnode_t *__vfs_pnode_lookup(vfs_path_t *vfs_path,int *ts,int *ed)
{
  vfs_pnode_t *point=NULL;
  //  vfs_path_t mould;
  vfs_pnode_t *root;
  usrtc_node_t *node;
  char *tn;
  int i=0;

  if(!vfs_root)
    return NULL;
  else
    point=vfs_root;

  if(!vfs_root->nnodes) { /* one thing */
#ifdef CONFIG_V2_DEBUG
    console_printf("[v2] {pp_lookup with vfs_root->nnodes==nil} \n");
    console_printf("[v2] {pp_lookup with vfs_root=%p} \n",vfs_root);
#endif
    point=vfs_root;
    *ts=0;
    if(!strcmp(vfs_path->path,V2_SEP))
      *ts=1;
    *ed=0;
  } else {
    tn=malloc(strlen(vfs_path->path)+sizeof(char));
    if(!tn)       return point;
    else      memset(tn,'\0',strlen(vfs_path->path)+sizeof(char));
    root=vfs_root;
#ifdef CONFIG_V2_DEBUG
    console_printf("[v2] {pp_lookup} root-> (%p)\n",root->point/*->fs_point_info->fstype*/);
#endif
    /* the one case - pp == 1 */
    if(vfs_path->pp<2) {
      strcpy(tn,vfs_path->path);
      node=usrtc_lookup(root->srch,tn);
      if(!node) { /* nothing was found */
	point=root;
	*ts=0; *ed=0;
      } else {
	*ts=1; *ed=0;
	/* modify vfs_path ROOT */
	vfs_path->path[sizeof(char)*2]='\0';
	vfs_path->plen=1;
	vfs_path->pp=1;
	point=usrtc_node_getdata(node);
      }
    } else {
#ifdef CONFIG_V2_DEBUG
      console_printf("[v2] vfs_path->pp=%d [",vfs_path->pp);
      for(i=0;i<vfs_path->pp;i++) console_printf("%d;",vfs_path->pps[i]);
      console_printf("]\n");
#endif
      for(i=0;i<vfs_path->pp;i++) {
	/* root was taken - going on */
	if((i+1)<vfs_path->pp)
	  strncpy(tn,vfs_path->path,vfs_path->pps[i+1]);
	else
	  strcpy(tn,vfs_path->path);
#ifdef CONFIG_V2_DEBUG
	console_printf("[v2] {deep_lookup} tn = %s (i=%d)\n",tn,i);
#endif
	node=usrtc_lookup(root->srch,tn);
	if(!node) { /* nothing was found */
	  point=root;
	  *ts=0; *ed=i;
	  break;
	} else {
	  *ts=1; *ed=i;
	  if((i+1)<vfs_path->pp) {
	    root=usrtc_node_getdata(node);
	  } else {
	    /* modify vfs_path go ROOT  */
	    *ed=0;
	    vfs_path->path[sizeof(char)*2]='\0';
	    vfs_path->plen=1;
	    vfs_path->pp=1;
	    point=usrtc_node_getdata(node);
	    break;
	  }
	}

      }
    }
    
    free(tn); /* ed -> offset to lookup, ts==0 if this is not mount point, ts==1 otherwise */
  }

#ifdef CONFIG_V2_DEBUG
  console_printf("[v2] {pp_lookup} point->point-> (%p)<<END\n",point->point/*->fs_point_info->fstype*/);
#endif


  return point;
}
