/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * resources structures
 *
 */

#ifndef __V2_RESOURCE_H__
#define __V2_RESOURCE_H__

#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <stdarg.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <pthread.h>
#include <datalib/usrtc.h>

#include <vfs/vfsdef.h>

/* global defines */
#define V2_SEP  "/"

/* some defines */
#define V2_MAX_LINK_DEEP  16
#define V2_MAX_OPEN       512
#define V2_MAX_LINKS      32

#define V2_MAX_CLIENTS  256

/* mount resources */

typedef struct __vfs_point_t {
  pid_t pid;      /* pid of a filesystem */
  prtid_t port;   /* port id of a file system */
  chid_t channel; /* opened channel to a file system */
} vfs_point_t;

/*
 * path name structure description:
 * |<---------plen--------->|
 * /dir0000000/dir111111/file
 * ^          ^         ^
 * pps[0]     pps[1]    pps[2]
 * where:
 * pps[max] - end, max = pp
 */
typedef struct __vfs_path_t {
  int pp;        /* amount of points */
  int plen;      /* path length in bytes */
  ulong_t hash;  /* hash of the path */
  int *pps;      /* parsed points */
  char *path;    /* path name*/
} vfs_path_t;

typedef struct __vfs_point_owner_t {
  uid_t o_uid;
  gid_t o_gid;
  void *mac_rule;
} vfs_point_owner_t;

typedef struct __link_point_t {
  int ref_count;
  int vfs_flags;
  int state;
  struct stat pp_stat;
  vfs_point_t *end_point;
  vfs_point_owner_t *owner; /* if null,then supervisor */
  vfs_info_t *fs_point_info;
} link_point_t;

/* initialization function */
int vfs_point_init(void);
/* register new point (actually doesn't mounts it) */
int vfs_point_register(pid_t fs_pid,prtid_t fs_port,char *fs_type,
		       char *fs_author,char *fs_info,int spec_flag);
/* unregister fs */
int vfs_point_unregister(link_point_t *fs_linkpoint);
/* unregister fs by pid */
int vfs_point_unregister_p(pid_t caller);
/* mount fs point */
int vfs_point_mount(char *fstype,vfs_path_t *m_loc,int flags);
/* mount fs point via it's link pointer */
int vfs_point_mount_l(link_point_t *pm,vfs_path_t *m_loc,int flags);
/* umount point */
int vfs_point_umount(link_point_t *fs_linkpoint);
/* lookup a link point,if NULL then there are no fses  */
/* ts wil changed to 1 if this is a mount point - > directory */
/* ed == edeep -> how many offset to the root of destination fs */
link_point_t *vfs_point_lookup(vfs_path_t *vfs_path,int *ts,int *ed);
/* lookup for not mounted filesystem with specified type */
link_point_t *vfs_lookup_nmtype(char *fstype);
/* sync all io on fses */
int vfs_point_sync(void);

#define ULNK_CLONE  0xa
#define ULNK_SETMD  0xb
#define ULNK_GETMD  0xc

/* user task resources */
typedef struct __user_link_t {
  int u_fd;             /* file descriptor id */
  int flags;            /* flags of this file descriptor */
  int mode;             /* mode of this file descriptor */
  int ufsid;            /* id of resource session on fs/node */
  int port_id;          /* port id */
  pid_t pid;            /* pid of node, int case of specific fses */
  link_point_t *point;  /* link point */
} user_link_t;

typedef struct __user_gate_t {
  pid_t ug_pid;
  uid_t ug_uid;
  gid_t ug_gid;
  /* lock */
  pthread_mutex_t ulock;
  /* data structure */
  usrtc_node_t *unode;
  /* limits and permissions */
  int max_open;
  /* user can open, unlink, opendir, mmap, mkdir, rmdir */
  int allowed_actors;
  /* cwd */
  int cwd_len;
  char *cwd;
  /* resources */
  int mex;
  int rcon;
  user_link_t **u_links;
} user_gate_t;

/* user gate interface */
/* allowed actors */
#define UGATE_AAOPEN     (1 << 0)
#define UGATE_AAUNLINK   (1 << 1)
#define UGATE_AAOPENDIR  (1 << 2)
#define UGATE_AAMMAP     (1 << 3)
#define UGATE_AAMKDIR    (1 << 4)
#define UGATE_AARMDIR    (1 << 5)
#define UGATE_AADIRECT   (1 << 6)

/* commands to influence */
#define UGATE_CLONE  0x0 /* clone usergate */
#define UGATE_RESET  0x1 /* reset usergate - close all links with FD_NOEXEC */
#define UGATE_KILL   0x2 /* kill and clear usergate */

/* commands to setting specific attributes */
#define UGATE_SETMAXOPEN  0x0
#define UGATE_GETMAXOPEN  0x1
#define UGATE_SETACT      0x2
#define UGATE_GETACT      0x3
#define UGATE_GETCWD      0x4
#define UGATE_SETCWD      0x5

/* command for operations with user links */
#define ULINK_FDUP    0x0
#define ULINK_FGETFL  0x1
#define ULINK_FSETFL  0x2
#define ULINK_FGETFD  0x4
#define ULINK_FSETFD  0x5

#define MAGIC_ULI  128

/* useful macros */
#define ugate_lock(u)    pthread_mutex_lock(&u->ulock)
#define ugate_unlock(u)  pthread_mutex_unlock(&u->ulock)

/* initialize gate interface */
int user_gate_init(void);

/**
 * @fn int create_user_gate(user_gate_t *ugate,pid_t pid,uid_t uid,gid_t gid);
 * @brief Create user gate if it doesn't created. Standart case for those
 * who ran from grub
 * @param ugate - a pointer to the allocated structure
 */
int user_gate_create(user_gate_t *ugate,pid_t pid,uid_t uid,gid_t gid);
/* set attribute to the user gate */
int user_gate_setattr(user_gate_t *ugate,int cmd,...);
/* get attribute of the user gate */
int user_gate_getattr(user_gate_t *ugate,int cmd,...);
/* do something with user gate */
int user_gate_influence(user_gate_t *ugate,int cmd,pid_t pid,...);
/* find user gate */
user_gate_t *user_gate_take(pid_t pid);

/* operations with user links */
#define user_link_init(p,fd,upoint,umode,uflags)    \
  p->u_fd=fd; p->point=upoint; p->mode=umode;       \
  p->flags=uflags;

#define user_link_lookup(gate,fd)  gate->u_links[fd]
#define user_link_max(gate)  gate->mex

/* store user link */
int user_link_store(user_gate_t *ugate,user_link_t *ulink);
/* remove user link */
int user_link_pop(user_gate_t *ugate,user_link_t *ulink);
/* int user_link shake */
int user_link_shake(user_gate_t *ugate,user_link_t *ulink,int cmd,...);

#endif /* __V2_RESOURCE_H__ */
