/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * user gate interface (it doesn't implement any protocol, 
 * intercommunication features)
 *
 */

#include <sys/types.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <pthread.h>
#include <unistd.h>
#include <errno.h>
#include <datalib/usrtc.h>

#include "resource.h"

/* internal functions and vars prototypes  */
static usrtc_t *gates=NULL;
static int cur_gates=0;
static pthread_mutex_t gate_lock;

static int __gate_compare(const void *a,const void *b);

#define __glock  pthread_mutex_lock(&gate_lock)
#define __gunlock  pthread_mutex_unlock(&gate_lock)

int user_gate_init(void)
{
  gates=malloc(sizeof(*gates));
  if(!gates)
    return ENOMEM;
  else
    cur_gates=0;

  usrtc_init(gates,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__gate_compare);

  /* init lock */
  pthread_mutex_init(&gate_lock,NULL);

  return 0;
}

/**
 * @fn int create_user_gate(user_gate_t *ugate,pid_t pid,uid_t uid,gid_t gid);
 * @brief Create user gate if it doesn't created. Standart case for those
 * who ran from grub
 * @param ugate - a pointer to the allocated structure
 */
int user_gate_create(user_gate_t *ugate,pid_t pid,uid_t uid,gid_t gid)
{
  int i;

  if(!ugate)
    return EINVAL;

  /* check for limits */
  if(cur_gates>=V2_MAX_CLIENTS)
    return EBUSY;

  /* now let's fill ops */
  ugate->ug_pid=pid;
  ugate->ug_uid=uid;
  ugate->ug_gid=gid;
  /* initialize thread lock */
  pthread_mutex_init(&ugate->ulock,NULL);
  /* set defaults limits */
  ugate->max_open=V2_MAX_OPEN;
  /* if it's supervisor allow all */
  ugate->allowed_actors=0;
  if(uid==0 && gid==0) 
    ugate->allowed_actors=(UGATE_AAOPEN | UGATE_AAUNLINK | UGATE_AAOPENDIR |
			   UGATE_AAMMAP | UGATE_AAMKDIR | UGATE_AARMDIR | UGATE_AADIRECT);
  else /* we don't know - allow open | opendir only */
    ugate->allowed_actors=UGATE_AAOPEN|UGATE_AAOPENDIR;
  /* now let's a deal with cwd (current working directory) */
  /* this function usually called via grub started friends, so - they're root is / */
  ugate->cwd=strdup(V2_SEP);
  if(!ugate->cwd)
    return ENOMEM;
  else
    ugate->cwd_len=strlen(ugate->cwd);
  /* setup data structure in location */
  ugate->unode=malloc(sizeof(*ugate->unode));
  if(!ugate->unode)
    return ENOMEM;

  /* let's a deal with resources */
  ugate->u_links=malloc(sizeof(uintptr_t)*MAGIC_ULI);
  if(!ugate->u_links)
    return ENOMEM;
  ugate->mex=MAGIC_ULI;
  ugate->rcon=0;

  for(i=0;i<MAGIC_ULI;i++) 
    ugate->u_links[i]=NULL;

  /* now let's add this */
  usrtc_node_init(ugate->unode,ugate);
  __glock;
  usrtc_insert(gates,ugate->unode,&ugate->ug_pid);
  cur_gates++;
  __gunlock;
  
  return 0;
}

/* set attribute to the user gate */
int user_gate_setattr(user_gate_t *ugate,int cmd,...)
{
  va_list vp;
  char *cwd;

  if(!ugate)
    return EINVAL;

  switch(cmd) {
  case UGATE_SETMAXOPEN:
    pthread_mutex_lock(&ugate->ulock);
    va_start(vp,cmd);
    ugate->max_open=va_arg(vp,int);
    va_end(vp);
    pthread_mutex_unlock(&ugate->ulock);
    break;
  case UGATE_SETACT:
    pthread_mutex_lock(&ugate->ulock);
    va_start(vp,cmd);
    ugate->allowed_actors=va_arg(vp,int);
    va_end(vp);
    pthread_mutex_unlock(&ugate->ulock);
    break;
  case UGATE_SETCWD:
    pthread_mutex_lock(&ugate->ulock);
    va_start(vp,cmd);
    cwd=va_arg(vp,char *);
    va_end(vp);
    if(cwd==NULL) {
      pthread_mutex_unlock(&ugate->ulock);
      return EINVAL;
    }
    free(ugate->cwd);
    ugate->cwd=strdup(cwd);
    if(!ugate->cwd) {
      pthread_mutex_unlock(&ugate->ulock);
      return ENOMEM;
    }
    ugate->cwd_len=strlen(ugate->cwd);
    pthread_mutex_unlock(&ugate->ulock);
    break;
  default:
    return EINVAL;
    break;
  }

  return 0;
}

/* get attribute of the user gate */
int user_gate_getattr(user_gate_t *ugate,int cmd,...)
{
  va_list vp;
  char *cwd;
  int *mop;

  if(!ugate)
    return EINVAL;

  switch(cmd) {
  case UGATE_GETMAXOPEN:
    pthread_mutex_lock(&ugate->ulock);
    va_start(vp,cmd);
    mop=va_arg(vp,int *);
    va_end(vp);
    if(!mop)
      goto __getat_einval;
    else
      *mop=ugate->max_open;
    pthread_mutex_unlock(&ugate->ulock);
    break;
  case UGATE_GETACT:
    pthread_mutex_lock(&ugate->ulock);
    va_start(vp,cmd);
    mop=va_arg(vp,int *);
    va_end(vp);
    if(!mop)
      goto __getat_einval;
    else
      *mop=ugate->allowed_actors;
    pthread_mutex_unlock(&ugate->ulock);
    break;
  case UGATE_GETCWD:
    pthread_mutex_lock(&ugate->ulock);
    va_start(vp,cmd);
    cwd=va_arg(vp,char *);
    va_end(vp);
    if(cwd==NULL) {
    __getat_einval:
      pthread_mutex_unlock(&ugate->ulock);
      return EINVAL;
    }
    strcpy(cwd,ugate->cwd);
    pthread_mutex_unlock(&ugate->ulock);
    break;
  default:
    return EINVAL;
    break;
  }

  return 0;
}

/* do something with user gate */
int user_gate_influence(user_gate_t *ugate,int cmd,pid_t pid,...)
{
  pid_t n_pid;
  user_gate_t *n_ugate;
  va_list vp;
  int ec=0,i,ym;
  int nlist=0,j;
  int *aam;
  int __aff_amount=0;
  uintptr_t *list=NULL;

  switch(cmd) {
  case UGATE_CLONE: /* fork use it */
    if(ugate==NULL) {
      ugate=user_gate_take(pid);
    }
    if(!ugate)
      return EINVAL;
    __glock;
    pthread_mutex_lock(&ugate->ulock);
    va_start(vp,pid);
    n_pid=va_arg(vp,int);
    nlist=va_arg(vp,int);
    list=va_arg(vp,uintptr_t *);
    aam=va_arg(vp,int *);
    va_end(vp);
    /* ok, let's create new clone */
    if((n_ugate=malloc(sizeof(*n_ugate)))==NULL) { /* new structure */
    __clone_enomem:
      ec=ENOMEM;
      goto __clone_failed;
    }
    if((n_ugate->unode=malloc(sizeof(*n_ugate->unode)))==NULL) { /* data structure */
      free(n_ugate);
      goto __clone_enomem;
    }
    if((n_ugate->cwd=strdup(ugate->cwd))==NULL) { /* cwd clone */
      free(n_ugate->unode);
      free(n_ugate);
      goto __clone_enomem;
    }
    if((n_ugate->u_links=malloc(sizeof(uintptr_t)*ugate->mex))==NULL) { /* links clone */
      free(n_ugate->cwd);
      free(n_ugate->unode);
      free(n_ugate);
      goto __clone_enomem;
    }
    /* ok, clone all filedescriptors */
    for(i=0;i<ugate->mex;i++) {
      if(ugate->u_links[i]!=NULL) {
	if((n_ugate->u_links[i]=malloc(sizeof(user_link_t)))==NULL) /* fuck */
	  continue;
	else {
	  memcpy(n_ugate->u_links[i],ugate->u_links[i],sizeof(user_link_t)); /* copying */
	  n_ugate->u_links[i]->point->ref_count++;
	  ym=0;
	  for(j=0;list[j]!=0;j++) {
	    if(list[j]==(uintptr_t)n_ugate->u_links[i]->point) {
 	      ym++;
	      break;
	    }
	  }
	  if(!ym) {
	    list[j]=(uintptr_t)n_ugate->u_links[i]->point;
	    __aff_amount++;
	  }
	}
	
      } else
	n_ugate->u_links[i]=NULL;
    }
    *aam=__aff_amount;
    /* init mutex */
    pthread_mutex_init(&n_ugate->ulock,NULL);
    /* now we're should copy other stuff */
    n_ugate->ug_pid=n_pid;
    n_ugate->ug_uid=ugate->ug_uid;
    n_ugate->ug_gid=ugate->ug_gid;
    n_ugate->max_open=ugate->max_open;
    n_ugate->allowed_actors=ugate->allowed_actors;
    n_ugate->cwd_len=ugate->cwd_len;
    n_ugate->mex=ugate->mex;
    n_ugate->rcon=ugate->rcon;
    /* ok let's adds this */
    usrtc_node_init(n_ugate->unode,n_ugate);
    usrtc_insert(gates,n_ugate->unode,&n_ugate->ug_pid);
    ec=0;
    
  __clone_failed:
    pthread_mutex_unlock(&ugate->ulock);
    __gunlock;
    break;
  case UGATE_RESET: /* exec use it, it's mean that we're remove all FD_CLOEXEC fds */
    if(ugate==NULL) {
      ugate=user_gate_take(pid);
    }
    if(!ugate)
      return EINVAL;
    pthread_mutex_lock(&ugate->ulock);
    /* take arguments */
    va_start(vp,pid);
    nlist=va_arg(vp,int);
    list=va_arg(vp,uintptr_t *);
    aam=va_arg(vp,int *);
    va_end(vp);
    /* ok, let's go */
    for(i=0;i<ugate->mex;i++) {
      if(ugate->u_links[i]==NULL) 
	continue;
      else {
	if(ugate->u_links[i]->flags & FD_CLOEXEC) {
	  ym=0;
	  for(j=0;list[j]!=0;j++) {
	    if(list[j]==(uintptr_t)ugate->u_links[i]->point) {
 	      ym++;
	      break;
	    }
	  }
	  if(!ym) {
	    list[j]=(uintptr_t)ugate->u_links[i]->point;
	    ugate->u_links[i]->point->ref_count--;
	    __aff_amount++;
	  }
	  free(ugate->u_links[i]);
	  ugate->u_links[i]=NULL;
	  ugate->rcon--;
	}
      }
    }
    *aam=__aff_amount;
    pthread_mutex_unlock(&ugate->ulock);
    ec=0;
    break;
  case UGATE_KILL: /* when process is died or exiting kill it resources */
    if(ugate==NULL) {
      ugate=user_gate_take(pid);
    }
    if(!ugate)
      return EINVAL;
    pthread_mutex_lock(&ugate->ulock);
    va_start(vp,pid);
    nlist=va_arg(vp,int);
    list=va_arg(vp,uintptr_t *);
    aam=va_arg(vp,int *);
    va_end(vp);
    __glock;
    if(nlist>0) { /* task exit was abnormally */
      for(i=0;i<ugate->mex;i++) {
	if(ugate->u_links[i]!=NULL) {
	  ym=0;
	  for(j=0;list[j]!=0;j++) {
	    if(list[j]==(uintptr_t)ugate->u_links[i]->point) {
 	      ym++;
	      break;
	    }
	  }
	  if(!ym) {
	    list[j]=(uintptr_t)ugate->u_links[i]->point;
	    ugate->u_links[i]->point->ref_count--;
	    __aff_amount++;
	  }
	  free(ugate->u_links[i]);
	} else
	  continue;
      }
    } 
    usrtc_delete(gates,ugate->unode); /* first remove it */
    __gunlock;
    /* now free resources */
    free(ugate->cwd);
    free(ugate->u_links);
    pthread_mutex_unlock(&ugate->ulock);
    pthread_mutex_destroy(&ugate->ulock);
    free(ugate);
    *aam=__aff_amount;
    ec=0; /* ok */
    break;
  default:
    return EINVAL;
    break;
  }

  return ec;
}

/* find user gate */
user_gate_t *user_gate_take(pid_t pid)
{
  usrtc_node_t *node;

  __glock;
  node=usrtc_lookup(gates,&pid);
  __gunlock;
  if(node!=NULL) 
    return usrtc_node_getdata(node);
  else
    return NULL;
}

/* local functions implementation */
static int __gate_compare(const void *a,const void *b)
{
  return *(pid_t*)a-*(pid_t*)b;
}
