/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS virtual file system v2,
 * v2 VFS main entry and proto low level operations,
 * it can use general resources
 */

#include <sys/types.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <pthread.h>
#include <unistd.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <sys/v2.h>
#include <sys/proc.h>
#include <ns/ns.h>
#include <errno.h>
#include <datalib/usrtc.h>

#include <fsio/fsspec.h>

#ifdef CONFIG_VFSV2_DEBUG
#define CONFIG_V2_DEBUG  1
#endif

#ifdef CONFIG_V2_DEBUG
/* debug */
#include "fslog.h"
#endif

#include "resource.h"
#include "fsops.h"

/* defines */
#define KQUEUE_SIZE  512
#define KSERV_FLAGS  (IPC_BLOCKED_ACCESS)
#define KPORT_MSGSZ  4096

#ifndef CONFIG_VFS_THREADS
#define CONFIG_VFS_THREADS  6
#endif

/* 64k for IPC buffer */
#define MSG_SIZE  65536
#define IMS_SIZE  4096  /* internal file layer max message size */

/* local sets */
static prtid_t serv_port,prtevn;
static user_gate_t *pt_gate[CONFIG_VFS_THREADS];
static user_link_t *pt_link[CONFIG_VFS_THREADS];
static char *pt_data[CONFIG_VFS_THREADS];
static int tpoint[CONFIG_VFS_THREADS];
static pthread_t pth_atom[CONFIG_VFS_THREADS];
/* local system events sets */
static task_event_ctl_arg_t t_event;

/* local functions */
static void *vfs_thread(void *arg);
static void vfs_killresources(pid_t pid);
static inline int __vfs_read_mould(prtid_t port,int msgid,
                                   uni_rpc_t *rpc,vfs_path_t *mould,
                                   char *msg);

int main(int argc,char **argv)
{
  int r,i;
  ipc_port_msg_info_t msgi;
  task_event_descr_t evn_desc;

#ifdef CONFIG_V2_DEBUG
  log_init();
#endif

  serv_port=ipc_port_open(KSERV_FLAGS,KQUEUE_SIZE);
  if(serv_port<0) {
    abort();
  }

  /* well init event */
  prtevn=ipc_port_open(0,512);
  if(prtevn<0)     abort();
  t_event.ev_mask=0;
  t_event.ev_mask|=TASK_EVENT_TERMINATION;
  t_event.port=prtevn;

  /* initialize user gates */
  r=user_gate_init();
  if(r!=0)
    abort();

  /* initialize vfs points */
  r=vfs_point_init();
  if(r!=0)
    abort();

  /* make registration on ns */
  ns_register(getpid(),0,0,(V2T_TRUSTED|V2T_MKSERVICE),SYS_VFS,"vfsv2.x",NULL,0); /* register on ns */

#ifdef CONFIG_V2_DEBUG
  log("[VFSv2] Jari OS VFS registered.\n");
#endif

  /* initialize per thread data */
  for(i=0;i<CONFIG_VFS_THREADS;i++) {
    pt_gate[i]=malloc(sizeof(user_gate_t));
    pt_link[i]=malloc(sizeof(user_link_t));
    if(!pt_gate[i] || !pt_link[i]) { /* not all data can be allocated */
      abort();
    }
    pt_data[i]=malloc(MSG_SIZE);
    if(!pt_data[i]) {
      abort();
    }
  }

  /* initialize all threads */
  for(i=0;i<CONFIG_VFS_THREADS;i++) {
    tpoint[i]=i;
    r=pthread_create(&pth_atom[i],NULL,vfs_thread,(void *)&tpoint[i]);
    if(r!=0)      abort();
  }

#ifdef CONFIG_V2_DEBUG
  log("[VFSv2] Service threads up and running, going to listen system events.\n");
#endif

  /* start to listen system queries */
  while(1) {
    r=ipc_port_rcvmsg_mi_cs(prtevn,sizeof(task_event_descr_t),&evn_desc,&msgi,IPC_BLOCKED_ACCESS);
    if(r<0) continue;
    if(evn_desc.ev_mask & TASK_EVENT_TERMINATION) {
      vfs_killresources(evn_desc.pid);
    }
  }

  return 0;
}

static void vfs_killresources(pid_t pid)
{
  user_gate_t *sgate;
  uintptr_t *aff_list;
  char msg[IMS_SIZE];
  int a,aff_am,i;

  sgate=user_gate_take(pid);
  if(!sgate) {
    goto remove_listener;
  }
  memset(msg,0,IMS_SIZE);
  aff_list=(uintptr_t*)msg;
  aff_am=0;
  a=user_gate_influence(sgate,UGATE_KILL,pid,256,aff_list,&aff_am);
  if(a!=0) {
    goto remove_listener;
  }
  for(i=0;i<aff_am;i++)
    vfs_pidfdkill((link_point_t *)aff_list[i],pid);

remove_listener:
  task_ctl(0,0,SYS_PR_CTL_DEL_EVENT_LISTENER,pid);
  return;
}

static void vfs_fcntl(pid_t pid,uid_t uid,gid_t gid,uni_rpc_t *rpc)
{
  user_gate_t *sgate;
  user_link_t *alink;
  int md;
  int r;

  sgate=user_gate_take(pid);
  if(!sgate) {
    rpc->rpc_call=ESRCH;
    return;
  }
  ugate_lock(sgate);

  switch(rpc->arg0) {
  case F_DUPFD: /* going with fd to clone (arg1) and fd(arg2) to affect */
    alink=user_link_lookup(sgate,rpc->arg1);
    if(alink==NULL) {
      rpc->rpc_call=EINVAL;
      goto __end_fcntl;
    }
    r=user_link_shake(sgate,alink,ULNK_CLONE,(int)rpc->arg2);
    rpc->rpc_call=r;
    break;
  case F_GETFD:
    alink=user_link_lookup(sgate,rpc->arg1);
    if(alink==NULL) {
      rpc->rpc_call=EINVAL;
      goto __end_fcntl;
    }
    r=user_link_shake(sgate,alink,ULNK_GETMD,&md);
    rpc->rpc_call=r;
    rpc->arg0=md;
    break;
  case F_SETFD:
  case F_GETFL:
  case F_SETFL:
  case F_GETLK:
  case F_SETLK:
  case F_SETLKW:
  case F_GETOWN:
  case F_SETOWN:
    rpc->rpc_call=ENOSYS;
  default:
    rpc->rpc_call=EINVAL;
    break;
  }

 __end_fcntl:
  ugate_unlock(sgate);

  return;
}

static void *vfs_thread(void *arg)
{
  int a,aa,r,msgid;
  int holf,ii,i;
  ipc_port_msg_info_t msgi;
  uni_rpc_t rpc;
  user_gate_t *wgate,*sgate;
  user_link_t *wlink,*uslink;
  link_point_t *sslink;
  vfs_path_t mould;
  l_v2IO_t *slink;
  iovec_t r_iov[2];
  struct stat stat_buf;
  char *msg,*fs_type,*fs_author,*fs_info;
  char *temp_cwd;
  uintptr_t *aff_list;
  int aff_am;
  pid_t c_pid;
  uid_t c_uid;
  gid_t c_gid;

  aa=*(int*)arg;

  /* take pre-allocated resources */
  wgate=pt_gate[aa];
  wlink=pt_link[aa];
  msg=pt_data[aa];

  while(1) {
    r=ipc_port_rcvmsg_mi(serv_port,0,NULL,&msgi);
    msgid=msgi.msg_id;
    c_pid=msgi.sender_pid;
    c_uid=msgi.sender_uid;
    c_gid=msgi.sender_gid;
    if(r>0 && r>=sizeof(uni_rpc_t)) {
      a=ipc_port_read_msg(serv_port,msgid,0,&rpc,sizeof(rpc)); /* reads header */
      if(a!=0) continue;
      switch(rpc.rpc_call) {
	/* system jobs */
      case RPC_REGISTER: /* register a file system */
	if(c_uid!=0 || c_gid!=0) {
	  rpc.rpc_call=EPERM;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	} else {
	  if(sizeof(rpc)+rpc.arg2+rpc.arg3+rpc.arg4 > r) {
	    rpc.rpc_call=EINVAL;
	    r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	    break;
	  }
	  /* there are need to read fs_type and optional info */
	  fs_type=msg;
	  a=ipc_port_read_msg(serv_port,msgid,sizeof(rpc),fs_type,rpc.arg2);
	  if(a!=0) continue;
	  fs_type[rpc.arg2]='\0'; /* fill with end symbol */
	  if(rpc.arg3!=0) {
	    fs_author=fs_type+rpc.arg2+sizeof(char);
	    a=ipc_port_read_msg(serv_port,msgid,sizeof(rpc)+rpc.arg2,fs_author,rpc.arg3);
	    if(a!=0) continue;
	    fs_author[rpc.arg3]='\0';
	  } else
	    fs_author=NULL;
	  if(rpc.arg4!=0) {
	    if(!fs_author)
	      fs_info=fs_type+rpc.arg2+sizeof(char);
	    else
	      fs_info=fs_author+rpc.arg3+sizeof(char);
	    /* read this from a message */
	    a=ipc_port_read_msg(serv_port,msgid,sizeof(rpc)+rpc.arg2+rpc.arg3,fs_info,rpc.arg4);
	    if(a!=0) continue;
	    fs_author[rpc.arg4]='\0';
	  } else
	    fs_info=NULL;
	  r=vfs_point_register(c_pid,rpc.arg0,fs_type,fs_author,fs_info,rpc.arg1);
	  rpc.rpc_call=r;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	  /* check for initial core fses */
	  if(!strcmp(fs_type,"initfs")) {
	    sslink=vfs_lookup_nmtype(fs_type);
	    r=vfs_mount(sslink,NULL,0);
	    if(!r) {
	      mould.pp=1; /* complete mould for root */
	      mould.pps=(int *)msg;
	      mould.path=msg+sizeof(int);
	      mould.plen=1;
	      mould.pps[0]=1;
	      mould.path[0]='/';
	      mould.path[1]='\0';
	      r=vfs_point_mount_l(sslink,&mould,rpc.arg0);
	    }
	  } else if(!strcmp(fs_type,"tmpfs")) {
	    sslink=vfs_lookup_nmtype(fs_type);
	    r=vfs_mount(sslink,NULL,0);
	    if(!r) {
	      mould.pp=1; /* complete mould for root */
	      mould.pps=(int *)msg;
	      mould.path=msg+sizeof(int);
	      mould.plen=strlen("/tmp");
	      mould.pps[0]=1;
	      strcpy(mould.path,"/tmp");
	      mould.path[mould.plen]='\0';
	      r=vfs_point_mount_l(sslink,&mould,rpc.arg0);
	    }
	  } else if(!strcmp(fs_type,"deviofs")) {
	    sslink=vfs_lookup_nmtype(fs_type);
	    r=vfs_mount(sslink,NULL,0);
	    if(!r) {
	      mould.pp=1; /* complete mould for root */
	      mould.pps=(int *)msg;
	      mould.path=msg+sizeof(int);
	      mould.plen=strlen("/dev");
	      mould.pps[0]=1;
	      strcpy(mould.path,"/dev");
	      mould.path[mould.plen]='\0';
	      r=vfs_point_mount_l(sslink,&mould,rpc.arg0);
	    }
	  }
	}
	break;
      case RPC_UNREGISTER: /* remove file system */
	/*
	 * NOTE: only file system can initiate this call
	 * file system ran with supervisor or root user - chiki
	 */
	if(c_uid!=0 && c_uid!=1)
	  rpc.rpc_call=EPERM;
	else
	  rpc.rpc_call=vfs_point_unregister_p(c_pid);
	r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	break;
      case RPC_MOUNT: /* mount file system */
	if(c_uid!=0 && c_uid!=1) { /* what the fuck are you doing ? */
	  rpc.rpc_call=EPERM;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	} else {
	  if(r<sizeof(rpc)+rpc.arg1+((rpc.arg2*sizeof(int))+rpc.arg3) +
	     ((rpc.arg4*sizeof(int))+rpc.arg5)) { /* somebody wants to make an error */
	    rpc.rpc_call=EINVAL;
	    r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	  } else if(!(rpc.arg0 & V2FS_SYS) && rpc.arg4==0) { /* you cannot mount regular fs without mount subject */
	    rpc.rpc_call=EINVAL; /* FIXME: change error code */
	    r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	  } else {
	    /*
	     * NOTE: well we will read message with msg buffer in the following maner:
	     * <fstype><points><path><points><path>
	     * ^-msg
	     */
	    /* reads fs_type */
	    fs_type=msg;
	    a=ipc_port_read_msg(serv_port,msgid,sizeof(rpc),fs_type,rpc.arg1);
	    fs_type[rpc.arg1+1]='\0';
	    /* lookup for registered fses */
	    if((sslink=vfs_lookup_nmtype(fs_type))==NULL) { /* not found */
	      rpc.rpc_call=ENOENT;
	      r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	    } else { /* did it ?! */
	      if(rpc.arg0 & V2FS_SYS) { /* all the friends will be happy */
		a=vfs_mount(sslink,NULL,rpc.arg0);
		if(a!=0)
		  rpc.rpc_call=a;
		else {
		  /* complete a path mould */
		  mould.pp=rpc.arg2; /* take a points amount */
		  mould.pps=(int *)(msg+rpc.arg1+sizeof(char));
		  a=ipc_port_read_msg(serv_port,msgid,sizeof(rpc)+rpc.arg1,mould.pps,mould.pp*sizeof(int));
		  /* take a path */
		  mould.plen=rpc.arg3;
		  mould.path=msg+rpc.arg1+sizeof(char)+mould.pp*sizeof(int);
		  a=ipc_port_read_msg(serv_port,msgid,sizeof(rpc)+rpc.arg1+mould.pp*sizeof(int),
				      mould.path,mould.plen);
		  mould.path[mould.plen+1]='\0';
		  a=vfs_point_mount_l(sslink,&mould,rpc.arg0); /* now VFS will attach filesystem */
		  rpc.rpc_call=a;
		}
		r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	      } else { /* we should read source of filesystem */
		mould.pp=rpc.arg4; /* take a points amount */
		mould.pps=(int *)(msg+rpc.arg1+sizeof(char));
		a=ipc_port_read_msg(serv_port,msgid,sizeof(rpc)+rpc.arg1+rpc.arg3+rpc.arg2*sizeof(int),
				    mould.pps,mould.pp*sizeof(int));
		/* take a path */
		mould.plen=rpc.arg5;
		mould.path=msg+rpc.arg1+sizeof(char)+mould.pp*sizeof(int);
		a=ipc_port_read_msg(serv_port,msgid,
				    sizeof(rpc)+rpc.arg1+rpc.arg3+rpc.arg2*sizeof(int)+mould.pp*sizeof(int),
				    mould.path,mould.plen);
		mould.path[mould.plen+1]='\0';
		a=vfs_mount(sslink,&mould,rpc.arg0);
		if(a!=0)
		  rpc.rpc_call=a;
		else {
		  mould.pp=rpc.arg2; /* take a points amount */
		  mould.pps=(int *)(msg+rpc.arg1+sizeof(char));
		  a=ipc_port_read_msg(serv_port,msgid,sizeof(rpc)+rpc.arg1,mould.pps,mould.pp*sizeof(int));
		  /* take a path */
		  mould.plen=rpc.arg3;
		  mould.path=msg+rpc.arg1+sizeof(char)+mould.pp*sizeof(int);
		  a=ipc_port_read_msg(serv_port,msgid,sizeof(rpc)+rpc.arg1+mould.pp*sizeof(int),
				      mould.path,mould.plen);
		  mould.path[mould.plen+1]='\0';
		  a=vfs_point_mount_l(sslink,&mould,rpc.arg0); /* now VFS will attach filesystem */
		  rpc.rpc_call=a;
		}
		r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	      }
	    }
	  }
	}
	break;
      case RPC_UMOUNT: /* umount file system */
	rpc.rpc_call=ENOSYS; /* TODO: vfs_umount */
        /* NOTE: umount file system should be
         * initiated from file system linker and fslinker
         * should care about rights and other stuff
         * VFS just takes end point, check ref count and
         * remove it from lookup table, after this done,
         * VFS will send RPC call to this file system service
         * that shall make umount, if this file system is mounted
         * in other lookup path, VFS will not send RPC call to this
         * file system service.
         */
	r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	break;
      case RPC_SYNC: /* global sync call */
	rpc.rpc_call=vfs_point_sync();
	r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	break;
	/* user gates (fd resources) user ops */
      case RPC_PIDFDGET: /* get file descriptors */
	/*
	 * NOTE: this call can be quered with nil user gates
	 * on this case we're should provide gate
	 */
	if(rpc.size==0) {
	  rpc.rpc_call=EINVAL;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	  break;
	} else {
	  /*
	   * This call works only for caller
	   * Also caller should set size of the his range of fds
	   * ulibc has a default size, and in this case all will works
	   * are ok , on other case ERANGE will cause
	   */
	  if((sgate=user_gate_take(c_pid))==NULL) {
	    holf=user_gate_create(wgate,c_pid,c_uid,c_gid);
	    if(holf!=0) {
	      rpc.rpc_call=holf;
	      r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	    } else {
	      user_gate_setattr(wgate,UGATE_SETCWD,"/");
	      /* now we will reply */
	      rpc.rpc_call=0;
	      rpc.size=0;
	      /* attach event for this pid */
	      task_ctl(c_pid,0,SYS_PR_CTL_ADD_EVENT_LISTENER,(uint64_t)&t_event);
	      r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
              wgate=NULL;
	      wgate=malloc(sizeof(*wgate)); /* take a new resource */
	    }
	  } else { /* real works on it */
	    if(sgate->rcon==0) { /* there are nothing opened */
	      rpc.rpc_call=0;
	      rpc.size=0;
	      r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	    } else {
	      if(rpc.size<sgate->rcon) {
		rpc.rpc_call=ERANGE;
		r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	      } else {
		slink=(l_v2IO_t*)msg; a=0;
		for(ii=0;ii<user_link_max(sgate);ii++) { /* TODO: optimize this one */
		  if(sgate->u_links[ii]!=NULL) {
		    slink->fd=sgate->u_links[ii]->u_fd;
		    slink->lid=sgate->u_links[ii]->ufsid;
		    slink->pid=sgate->u_links[ii]->pid;
		    slink->prtid=sgate->u_links[ii]->port_id;
		    slink++;
		    a++;
		    if(a>sgate->rcon)
		      break;
		  } else
		    continue;
		}
		/* now we're ready to send */
		rpc.rpc_call=0;
		rpc.size=sgate->rcon;
		r_iov[0].iov_base=&rpc;
		r_iov[0].iov_len=sizeof(rpc);
		r_iov[1].iov_base=msg;
		r_iov[1].iov_len=sizeof(l_v2IO_t)*sgate->rcon;
		r=ipc_port_reply_nmsg_iov(serv_port,r_iov,2,msgid);
	      }
	    }
	  }
	}
	break;
      case RPC_PIDFDCLONE: /* clone file descriptors */
	if(c_uid!=0 || c_gid!=0) {
	  rpc.rpc_call=EPERM;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	} else {
	  /**
	     #if 0
	     do something with user gate
	     int user_gate_influence(user_gate_t *ugate,int cmd,pid_t pid,...);
	     #endif
	  */
	  sgate=user_gate_take(rpc.arg0);
	  if(!sgate) {
	    rpc.rpc_call=ESRCH;
	    goto __clone_end;
	  }
	  aff_list=(uintptr_t*)msg;
	  aff_am=0;
	  memset(aff_list,0,256*sizeof(uintptr_t));
	  a=user_gate_influence(sgate,UGATE_CLONE,rpc.arg0,rpc.arg1,256,aff_list,&aff_am);
	  if(a!=0) {
	    rpc.rpc_call=a;
	    goto __clone_end;
	  }
	  for(i=0;i<aff_am;i++)
	    vfs_pidfdclone((link_point_t *)aff_list[i],rpc.arg0,rpc.arg1);
	  rpc.rpc_call=0;
	  /* add system event */
	  task_ctl(rpc.arg1,0,SYS_PR_CTL_ADD_EVENT_LISTENER,(uint64_t)&t_event);

	__clone_end:
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	}
	break;
      case RPC_PIDFDRESET: /* reset file descriptors */
	if(c_uid!=0 || c_gid!=0) {
	  rpc.rpc_call=EPERM;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	} else {
	  sgate=user_gate_take(rpc.arg0);
	  if(!sgate) {
	    rpc.rpc_call=ESRCH;
	    goto __reset_end;
	  }
	  aff_list=(uintptr_t*)msg;
	  aff_am=0;
	  a=user_gate_influence(sgate,UGATE_RESET,rpc.arg0,256,aff_list,&aff_am);
	  if(a!=0) {
	    rpc.rpc_call=a;
	    goto __reset_end;
	  }
	  for(i=0;i<aff_am;i++)
	    vfs_pidfdreset((link_point_t *)aff_list[i],rpc.arg0);
	  rpc.rpc_call=0;

	__reset_end:
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	}
	break;
      case RPC_PIDFDKILL:
	if(rpc.arg0!=c_pid && c_uid!=0) {
	  rpc.rpc_call=EPERM;
	  goto __kill_end;
	} else if(rpc.arg0!=c_pid && c_uid==0)
	  c_pid=rpc.arg0; /* exchange this */
	sgate=user_gate_take(rpc.arg0);
	if(!sgate) {
	  rpc.rpc_call=ESRCH;
	  goto __kill_end;
	}
	aff_list=(uintptr_t*)msg;
	aff_am=0;
	a=user_gate_influence(sgate,UGATE_KILL,c_pid,256,aff_list,&aff_am);
	if(a!=0) {
	  rpc.rpc_call=a;
	  goto __kill_end;
	}
	for(i=0;i<aff_am;i++)
	  vfs_pidfdkill((link_point_t *)aff_list[i],c_pid);
	rpc.rpc_call=0;

      __kill_end:
	r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	break;
	/* user regular operations */
      case RPC_GETCWD:
	/*
	 * NOTE: this call can be quered with nil user gates
	 * on this case we're should provide gate
	 */
	if(rpc.size==0) {
	  rpc.rpc_call=EINVAL;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	  break;
	}
	if((sgate=user_gate_take(c_pid))==NULL) {
	  holf=user_gate_create(wgate,c_pid,c_uid,c_gid);
	  if(holf!=0) {
	    rpc.rpc_call=holf;
	    r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	  } else {
	    user_gate_setattr(wgate,UGATE_SETCWD,"/");
	    /* now we will reply */
	    rpc.rpc_call=0;
	    r_iov[0].iov_base=&rpc;
	    r_iov[0].iov_len=sizeof(rpc);
	    r_iov[1].iov_base=wgate->cwd;
	    r_iov[1].iov_len=wgate->cwd_len;
	    r=ipc_port_reply_nmsg_iov(serv_port,r_iov,2,msgid);
            wgate=NULL;
	    wgate=malloc(sizeof(*wgate)); /* take a new resource */
	  }
	} else {
	  rpc.rpc_call=0;
	  if(sgate->cwd_len>rpc.size) { /* POSIX say that this should be like that */
	    rpc.rpc_call=ERANGE;
	    r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	  } else {
	    rpc.rpc_call=0;
	    r_iov[0].iov_base=&rpc;
	    r_iov[0].iov_len=sizeof(rpc);
	    r_iov[1].iov_base=sgate->cwd;
	    r_iov[1].iov_len=sgate->cwd_len+1;
	    r=ipc_port_reply_nmsg_iov(serv_port,r_iov,2,msgid);
	  }
	}
	break;
      case RPC_FCNTL:
        vfs_fcntl(c_pid,c_uid,c_gid,&rpc);
        r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
        break;
      case RPC_SETPWD:
	if((sgate=user_gate_take(c_pid))==NULL) {
	  rpc.rpc_call=EINVAL;
	  goto __setcwd_end;
	} else {
          if(__vfs_read_mould(serv_port,msgid,&rpc,&mould,msg)!=0) {
            rpc.rpc_call=EINVAL;
            goto __setcwd_end;
          }
          temp_cwd=strdup(mould.path);
	  r=vfs_lookup(&mould,&stat_buf);
	  if(r!=0) {
	    rpc.rpc_call=r;
	    goto __setcwd_end;
	  } else {
	    if(!S_ISDIR(stat_buf.st_mode)) {
	      rpc.rpc_call=ENOTDIR;
	      goto __setcwd_end;
	    }
	    if(c_uid!=0) { /* check it */
	      /* dummy */
              /* TODO: add permission checking */
	    }
	    user_gate_setattr(sgate,UGATE_SETCWD,temp_cwd);
	    rpc.rpc_call=0;
            free(temp_cwd);
	  }
	}
      __setcwd_end:
	r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	break; /*user_gate_setattr(wgate,UGATE_SETCWD,"/");*/
      case RPC_OPEN:
      __rpc_open:
	if(r<sizeof(rpc)+rpc.arg6+rpc.arg5*sizeof(int)) 
	  rpc.rpc_call=EINVAL;
	else {
	  /* take ugate */
	  if((sgate=user_gate_take(c_pid))==NULL) {
	    rpc.rpc_call=EINVAL;
	  } else {
	    /* check flags and actors */
	    if(c_uid!=0 || c_gid!=0) { /* we will check */
	      if((rpc.arg0 & O_DIRECTORY) && !(sgate->allowed_actors & UGATE_AAOPENDIR)) {
		rpc.rpc_call=EPERM;
		goto __open_end;
	      }
	      if(!(rpc.arg0 & O_DIRECTORY) && !(sgate->allowed_actors & UGATE_AAOPEN)) {
		rpc.rpc_call=EPERM;
		goto __open_end;
	      }
	      if((rpc.arg0 & O_DIRECTIO) && !(sgate->allowed_actors & UGATE_AADIRECT)) {
		rpc.rpc_call=EPERM;
		goto __open_end;
	      }
	      if(sgate->max_open==sgate->rcon) { /* check for limits */
		rpc.rpc_call=EMFILE;
		goto __open_end;
	      }
	    }
	    /* take a mould of request */
            if(__vfs_read_mould(serv_port,msgid,&rpc,&mould,msg)!=0) {
              rpc.rpc_call=EIO;
              goto __open_end;
            }
	    /* now we're should determine request destination */
            /* it could be symlink */
	    aff_am=ii=0;
	    sslink=vfs_point_lookup(&mould,&ii,&aff_am);
	    if(!sslink) {
	      rpc.rpc_call=ENOENT;
	      goto __open_end;
	    }
	    if(ii && !(rpc.arg0 & O_DIRECTORY)) { /* you trying to open file, but it's a directory */
	      rpc.rpc_call=EISDIR;
	      goto __open_end;
	    } else if(ii && c_uid!=0) { /* gone checking */
	      rpc.rpc_call=EPERM; /* TODO: add DAC checking */
	      goto __open_end;
	    }
            r=vfs_lookup(&mould,&stat_buf);
            if(!r && S_ISLNK(stat_buf.st_mode)) {
              if((r=vfs_lookup_resolve(&mould,&stat_buf))!=0) {
                rpc.rpc_call=r;
                goto __open_end;
              }
              /* ok, copypaste - it's ugly */
              aff_am=ii=0;
              sslink=vfs_point_lookup(&mould,&ii,&aff_am);
              if(!sslink) {
                rpc.rpc_call=ENOENT;
                goto __open_end;
              }
              if(ii && !(rpc.arg0 & O_DIRECTORY)) { /* you trying to open file, but it's a directory */
                rpc.rpc_call=EISDIR;
                goto __open_end;
              } else if(ii && c_uid!=0) { /* gone checking */
                rpc.rpc_call=EPERM; /* TODO: add DAC checking */
                goto __open_end;
              }
            }
	    /* prepare user link - not injected to usergate */
	    wlink->point=sslink;
	    wlink->u_fd=rpc.size;
	    wlink->flags=rpc.arg0;
	    wlink->mode=rpc.arg1;
	    r=vfs_open(wlink,sslink,&mould,aff_am,c_uid,c_gid,c_pid);
	    if(r!=0) {
	      rpc.rpc_call=r;
	      goto __open_end;
	    }
	    /* inject to user gate */
	    user_link_store(sgate,wlink);
	    rpc.rpc_call=0;
	    /* ok now edit rpc header to reply */
	    rpc.arg2=wlink->pid;
	    rpc.arg3=wlink->port_id;
	    rpc.arg4=wlink->ufsid;
	    wlink=malloc(sizeof(*wlink)); /* allocate new link */
	  }
	}
      __open_end:
	r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	break;
      case RPC_CLOSEDIR:
      case RPC_CLOSE:
	if((sgate=user_gate_take(c_pid))==NULL) {
	  rpc.rpc_call=EINVAL;
	} else {
	  if((uslink=user_link_lookup(sgate,rpc.arg0))==NULL) {
	    rpc.rpc_call=EBADF;
	    goto __close_end;
	  }
	  /* remove this one */
	  user_link_pop(sgate,uslink);
	  vfs_close(sgate,uslink);
	  free(uslink); /* free this resource */
	}
      __close_end:
	r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	break;
      case RPC_UNLINK: /* unlink goes w/o fd */
	if(sizeof(rpc)+rpc.arg6+rpc.arg5*sizeof(int)>r) {
        __unlink_fail:
	  rpc.rpc_call=EINVAL;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
          break;
	} else {
	  /* we're working without fd in this case */
          if(__vfs_read_mould(serv_port,msgid,&rpc,&mould,msg)!=0)
            goto __unlink_fail;
	  /* make a call */
	  r=vfs_unlink(&mould,c_uid,c_gid);
	  rpc.rpc_call=r;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	}
	break;
      case RPC_CREATEDIR: /* w/o fd */
	if(sizeof(rpc)+rpc.arg6+rpc.arg5*sizeof(int)>r) {
        __createdir_fail:
	  rpc.rpc_call=EINVAL;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
          break;
	} else {
	  /* we're working without fd in this case */
          if(__vfs_read_mould(serv_port,msgid,&rpc,&mould,msg)!=0)
            goto __createdir_fail;
	  /* make a call */
	  r=vfs_mkdir(&mould,rpc.arg0,c_uid,c_gid);
	  rpc.rpc_call=r;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	}
	break;
      case RPC_UNLINKDIR: /* w/o fd */
	if(sizeof(rpc)+rpc.arg6+rpc.arg5*sizeof(int)>r) {
        __unlinkdir_fail:
	  rpc.rpc_call=EINVAL;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
          break;
	} else {
	  /* we're working without fd in this case */
          if(__vfs_read_mould(serv_port,msgid,&rpc,&mould,msg)!=0)
            goto __unlinkdir_fail;
	  /* make a call */
	  r=vfs_unlinkdir(&mould,c_uid,c_gid);
	  rpc.rpc_call=r;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	}
	break;
      case RPC_OPENDIR:
	rpc.arg0|=O_DIRECTORY;
	goto __rpc_open;
	break;
      case RPC_STAT:  /* there are simililar tasks, but lstat is a `raw' stat */
      case RPC_LSTAT: /* instead of stat, stat shall resolve symlink */
	if(sizeof(rpc)+rpc.arg6+rpc.arg5*sizeof(int)>r) {
        __stat_fail:
	  rpc.rpc_call=EINVAL;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
          break;
	} else {
	  /* we're working without fd in this case */
          if(__vfs_read_mould(serv_port,msgid,&rpc,&mould,msg)!=0)
            goto __stat_fail;
	  /* make a lookup */
          if(rpc.rpc_call==RPC_LSTAT)
            r=vfs_lookup(&mould,&stat_buf);
          else
            r=vfs_lookup_resolve(&mould,&stat_buf);
	  if(!r) { /* it's all is ok */
	    rpc.rpc_call=0;
	    r_iov[0].iov_base=&rpc;
	    r_iov[0].iov_len=sizeof(rpc);
	    r_iov[1].iov_base=&stat_buf;
	    r_iov[1].iov_len=sizeof(struct stat);
	    r=ipc_port_reply_nmsg_iov(serv_port,r_iov,2,msgid);
	  } else { /* error has occured */
	    rpc.rpc_call=r;
	    r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	  }
	}
	break;
      case RPC_READLINK:
        if(sizeof(rpc)+rpc.arg6+rpc.arg5*sizeof(int)>r) {
        __readlink_fail:
	  rpc.rpc_call=EINVAL;
	  r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
          break;
        } else {
	  /* we're working without fd in this case */
          if(__vfs_read_mould(serv_port,msgid,&rpc,&mould,msg)!=0)
            goto __readlink_fail;
          r=vfs_readlink(&mould); /* try to read link */
          if(!r) {
	    rpc.rpc_call=0;
	    r_iov[0].iov_base=&rpc;
	    r_iov[0].iov_len=sizeof(rpc);
	    r_iov[1].iov_base=mould.path;
	    r_iov[1].iov_len=strlen(mould.path);
	    r=ipc_port_reply_nmsg_iov(serv_port,r_iov,2,msgid);
          } else {
	    rpc.rpc_call=r;
	    r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
          }
        }
        break;
      default:
	rpc.rpc_call=EINVAL;
	r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
	break;
      }
    } else if(r>0 && r<sizeof(rpc)) { /* invalid data  */
      rpc.rpc_call=EINVAL;
      r=ipc_port_reply_nmsg(serv_port,sizeof(rpc),&rpc,msgid);
    } else /* error occured */ /* TODO: add IPC error handling */
      continue;
  }

  return NULL;
}

/* local functions */

static inline int __vfs_read_mould(prtid_t port,int msgid,
                                   uni_rpc_t *rpc,vfs_path_t *mould,
                                   char *msg)
{
  int a;

  mould->pp=rpc->arg5;
  mould->plen=rpc->arg6;
  mould->pps=(int *)(msg);

  /* read parsed points */
  if((a=ipc_port_read_msg(port,msgid,sizeof(*rpc),
                          mould->pps,mould->pp*sizeof(int)))!=0)
    return -1;

  /* read path */
  mould->path=(char *)(msg+(sizeof(int)*mould->pp));
  if((a=ipc_port_read_msg(port,msgid,sizeof(*rpc)+mould->pp*sizeof(int),
                          mould->path,mould->plen))!=0)
    return -1;
  mould->path[mould->plen]='\0';

  return 0;
}
