/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * Block devices server
 */

/*
 * TO DO [dg]:
 * 1) release storage
 * 2) interface registration
 */

#include <sys/types.h>
#include <fcntl.h>
#include <sys/scheduler.h>
#include <block/genstor.h>
#include <block/blkdev.h>
#include <block/bdev_drv.h>
#include <block/server.h>
#include <block/request.h>
#include <block/iosched.h>
#include <unistd.h>
#include <sys/mman.h>
#include <dircons.h>
#include <errno.h>
#include <err.h>
#include <completion.h>
#include "server.h"

enum {
  DEFAULT_MAX_REQUESTS = 65536,
  DEFAULT_MAX_SECTORS = 256
};

struct bdev_serv_stat *serv_stat;

static storage_operations_t genstor_ops = {
  generic_open_storage,
  generic_close_storage,
  generic_scan_partitions,
  generic_read_storage,
  generic_write_storage,
  generic_ioctl_storage,
  generic_sync_storage,
  generic_make_request
};

static const char *bd_iface_names[] = {
  "unknown",
  "pseudo",
  "ata",
  "scsi"
};

static const char *bdev_names[] = {
  "unk_disk",
  "pd",
  "ad",
  "sd"
};

static void load_drivers(void);
static int register_device_servers(blkdev_driver_t *drv, void *host_lib_handle,
                                   const char *arg_line);

#ifdef CONFIG_BDEV_STATLINK

extern blkdev_driver_t *init_ata_driver(void);

init_bdev_driver_t init_driver_routines[] = {
#ifdef CONFIG_ATA_STATLINK
  init_ata_driver,
#endif
  NULL
};

/* load all drivers linked statically with the block subsystem */
static void load_drivers(void)
{
  blkdev_driver_t *drv = NULL;
  int i;

  for (i = 0; init_driver_routines[i] != NULL; i++) {
    drv = init_driver_routines[i]();
    if (drv == NULL)
      break;

    if (register_device_servers(drv, NULL, NULL)) {
      /* TO DO [dg]: output warning in log */
      break;
    }
  }
}

#else

static void load_drivers(void)
{
  /* stub */
}

#endif

static int register_dev_server(blkdev_driver_t *drv, struct bd_iface *iface,
                               block_device_t *bdev, char *name,
                               struct blk_group *group, struct serv_reg_data *rd)
{
  genstor_t *stor = NULL;
  blkreq_queue_t *q;
  bool cache;
  int ret = -ENOMEM;
  bool err_printed = false;

  /* Fixme [dg]: assign depending on the interface bus */
  cache = false;
  stor = malloc(sizeof(genstor_t));
  if (stor == NULL)
    goto out;

  q = malloc(sizeof(blkreq_queue_t));
  if (q != NULL)
    ret = init_blk_queue(q, stor, bdev, DEFAULT_MAX_REQUESTS, DEFAULT_MAX_SECTORS, group);
  if (ret)
    goto out;

  ret = init_generic_storage(stor, name, &genstor_ops, q, cache);
  if (!ret)
    ret = drv->prepare_rq_proc(q);

  if (ret)
    goto out;

  ret = generic_scan_partitions(stor);
  if (!ret || ret == -EINVAL) {
    /* register server in new thread */
    rd->stor = stor;
    ret = start_storage_serv_thread(rd);
    if (ret) {
      printf("[bdev.x]: storage server %s registration failed\n",
             stor->name);
      err_printed = true;
    }
  } else {
    printf("[bdev.x]: Error at partition scanning for device %d on "
           "%s interface\n", bdev->num, iface->name);
    err_printed = 1;
  }

out:
  if (ret) {
    if (!err_printed) {
      printf("[bdev.x]: Cannot allocate resources for storage, "
             "interface %s, device %d\n", iface->name, bdev->num);
    }
    if (stor != NULL) {
      release_generic_storage(stor);
      free(stor);
    }
    if (q != NULL)
      put_blk_queue(q);
  }

  return ret;
}

static int register_device_servers(blkdev_driver_t *drv, void *host_lib_handle,
                                   const char *arg_line)
{
  struct bd_iface *iface;
  int n, nr_ifaces;
  int ret = 0;
  pid_t pid = -1;
  struct serv_reg_data rd;
  char name[STOR_NAME_MAXLEN];
  struct blk_group *group;
  bool err_printed = false;
  block_device_t *bdev;
  struct completion compl;
  int index, i;
  int *ndevs = NULL;
  char s[16];
  bool init;

  rd.iobuf_size = 0;
  rd.nr_cmd_threads = 1;
  rd.nr_io_threads = 1;

  printf("[bdev.x]: scanning ata interfaces\n");
  nr_ifaces = drv->scan_ifaces(host_lib_handle);
  if (nr_ifaces < 0) {
    printf("[bdev.x]: the physical interface is not supported "
           "by the block device driver, or an error occurred at "
           "interface initialization, error code %d\n", -nr_ifaces);
    return nr_ifaces;
  } else if (nr_ifaces) {
    printf("OK\n");
  }

#if 0 /* anonymous shared mapping are not supported now */
  ndevs = mmap(NULL, PAGE_SIZE, PROT_READ | PROT_WRITE,
               MAP_ANONYMOUS | MAP_SHARED, 0, 0);
#endif
  if (/*ndevs == MAP_FAILED ||*/ init_completion(&compl))
    goto out;

  get_iface_ndevs(ndevs);
  for (n = 0; n < nr_ifaces; n++) {
    err_printed = false;
    pid = fork();
    if (pid == -1) {
      /* TO DO [dg]: print the interface name in the message */
      printf("[bdev.x]: Can't create a process for the block device interface\n");
      err_printed = true;
      ret = -errno;
      break;
    }

    if (pid) {
      /*
       * wait while the child reserves indexes for found devices
       * so that to eliminate naming stochastic on different
       * startups
       */
      wait_for_completion(&compl);
    } else {
      index = -1;
      ret = init_completion(&rd.compl);
      if (ret)
        break;

      ret = -ENOMEM;
      iface = malloc(sizeof(struct bd_iface));
      if (iface == NULL)
        break;

      memset(iface, 0, sizeof(struct bd_iface));
      list_init_head(&iface->devices);
      iface->id = n;
      group = malloc(sizeof(struct blk_group));
      if (group != NULL)
        ret = init_blk_group(group);

      if (ret)
        break;

      index = find_reserve_iface_index(drv->iface_type);
      /* naming the interface */
      sprintf(s, "%d", index);
      i = strlen(bd_iface_names[drv->iface_type]) + strlen(s) + 1;
      iface->name = malloc(i);
      if (iface->name == NULL) {
        ret = -ENOMEM;
        break;
      }
      sprintf(iface->name, "%s%s", bd_iface_names[drv->iface_type], s);

      /*
       * init and register the found interface
       * and all devices located on it
       */
      ret = drv->init_iface(host_lib_handle, iface, arg_line);
      if (ret)
        break;

      init = true;
      i = 0; // ndevs[drv->iface_type];
      /* reserve indexes for the devices */
#if 0
      list_for_each_entry(&iface->devices, bdev, node)
        ndevs[drv->iface_type]++;
#endif
      complete(&compl, 0);
      list_for_each_entry(&iface->devices, bdev, node) {
        if (ret) {
          /*
           * There are not enough resources to register
           * all found devices.
           * Suspend all not registered devices after the first
           * occured error so that to register them later
           */
          drv->suspend_device(bdev);
          continue;
        }

        if (bdev->status && BLKSTAT_MEDIA_ABSENT)
          continue;

        /* name for the new storage */
        if (i < 28) {
          sprintf(name, "%s%c", bdev_names[drv->iface_type], 'a' + i);
        } else {
          sprintf(name, "%s%d", drv->name, bdev->num);
        }
        i++;

        ret = register_dev_server(drv, iface, bdev, name, group, &rd);
        if (ret) {
          printf("[bdev.x]: Warning, registration of device %d of "
                 "the %s interface failed\n", i, iface->name);
          err_printed = true;
          drv->suspend_device(bdev);
        }
      }

      ret = start_interface_server(iface, index);
      if (ret) {
        printf("[bdev.x]: Cannot start %s block device "
               "interface server, %d error\n", iface->name, -ret);
        err_printed = true;
      }
      break;
    }
  }

out:
  if (pid) {
    destroy_completion(&compl);
    if (ndevs != NULL)
       munmap(ndevs, PAGE_SIZE);
  } else {
    if (!init)
      complete(&compl, ret);

    free_iface_index(drv->iface_type, index);
  }

  if (ret && !pid) {
    if (!err_printed) {
      printf("[bdev.x]: Unable to allocate memory for internal structures, or "
              "synchronization objects to register the block device insterface\n");
    }
  }

  return ret;
}

int main(int argc, char *arg[], char *env[])
{
  start_service();
  int r;

  /* open terminal */
  r = (open("/dev/ptm0", O_RDONLY) == -1);
  r += (open("/dev/ptm0", O_WRONLY) == -1);
  r += (open("/dev/ptm0", O_WRONLY) == -1);
  if (r || init_subsys_reginfo())
    abort();

  if (initialize_blk_schedulers())
    abort();

  load_drivers();
  if (start_subsystem_server())
    abort();

  return 0;
}
