/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 * (c) Copyright 2009 Stanislav M. Ivankin <grepz@jarios.org>
 *
 * bootstrap service
 *
 */
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/kcontrol.h>
#include <sys/scheduler.h>
#include <fcntl.h>
#include <limits.h>
#include <uts.h>
#include <sys/mman.h>
#include <sys/proc.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <datalib/usrtc.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <pthread.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ns/ns.h>

#include <sexpr/sexp.h>
#include <sexpr/sexp_ops.h>

#include "log.h"
#include "btsd.h"
#include "syntax.h"

static char used_conf[TOKEN_LENGTH];
static usrtc_t *modules_buf;
static usrtc_t *runlevels_buf;

static int __cmp_module(const void *a, const void *b);
static void __generic_delete(void *data, usrtc_node_t *node);

static btsd_runlevel_conf_t *__get_runlevel(usrtc_t *rlist, const char *used);
static void __run_levels(btsd_runlevel_conf_t *runlevel_conf);

int process_expr(sexp_t *expr)
{
  int ret;
  usrtc_node_t *node;
  btsd_runlevel_conf_t *rconf;
  char *token_name;

  if (expr->ty != SEXP_LIST || expr->list->ty != SEXP_VALUE)
    return BTSD_SYNTAX_SEXP_TYPE_ERR;

  token_name = expr->list->val;

  ret = BTSD_SYNTAX_OK;
  if (!strcmp(token_name, BTSD_MODULES)) {
    ret = eval_modules(expr, modules_buf);
  } else if (!strcmp(token_name, BTSD_RUNLEVEL)) {
    ret = eval_runlevel_conf(expr, &rconf);
    if (ret || !(node = malloc(sizeof(node)))) {
      /* TODO: Free properly */
    }
    usrtc_node_init(node, rconf);
    usrtc_insert(runlevels_buf, node, rconf);
  } else if (!strcmp(token_name, BTSD_LOAD_RUNLEVEL)) {
    if (*used_conf == '\0') {
      token_get_value(expr, SEXP_DQUOTE, BTSD_LOAD_RUNLEVEL,
		      used_conf, TOKEN_LENGTH);
    }
  } else
    ret = BTSD_SYNTAX_UNKNOWN_SEXP_ERR;

  return ret;
}

int main(int arch, char *argv[])
{
  char buf[4096];
  int ret, fd, len;
  pcont_t *cont;
  sexp_t *sx;
  btsd_runlevel_conf_t *used_runlevel;

  ns_register(getpid(), 0, 0, (V2T_TRUSTED|V2T_SYSDAEMON),
	      0, "bootsd.x", NULL, 0);

  fd = open(CONFFILE, O_RDONLY);
  if (fd < 0)
    return EXIT_FAILURE;

  memset(used_conf, '\0', TOKEN_LENGTH);
  sx   = NULL;
  cont = NULL;
  len = -1;

  modules_buf   = init_module_list();
  runlevels_buf = init_module_list();
  if (!modules_buf || !runlevels_buf)
    return EXIT_FAILURE;

  while (len != 0) {
    len = read(fd, buf, 16);
    buf[len] = '\0';
    if (cont == NULL)
      cont = init_continuation(buf);
    if (len > 0)
      sx = (sexp_t *)iparse_sexp(buf, len, cont);

    if (sx != NULL) {
      ret = process_expr(sx);
      if (ret)
	return EXIT_FAILURE;
      destroy_sexp(sx);
      sx = (sexp_t *)iparse_sexp(buf, len, cont);
    }
  }

  destroy_continuation(cont);
  close(fd);

  if (*used_conf == '\0')
    return EXIT_FAILURE;
  used_runlevel = __get_runlevel(runlevels_buf, used_conf);
  if (!used_runlevel)
    return EXIT_FAILURE;

  __run_levels(used_runlevel);

  return EXIT_SUCCESS;
}

static btsd_runlevel_conf_t *__get_runlevel(usrtc_t *rlist, const char *used)
{
  usrtc_node_t *node;
  node = usrtc_lookup(rlist, used);
  return node ? usrtc_node_getdata(node) : NULL;
}

static void __run_levels(btsd_runlevel_conf_t *runlevel_conf)
{
  usrtc_node_t *rnode, *mod_node;
  btsd_runlevel_t *rl;
  btsd_module_t *module;
  unsigned int i;
  pid_t pid;
  int status;

  rnode = usrtc_first(runlevel_conf->runlevels);
  while (rnode) {
    rl = usrtc_node_getdata(rnode);

    btsd_debug("[BTSD] Running level (%d).\n", rl->runlevel);

    for (i = 0; i < rl->modules_num; i++) {
      mod_node = usrtc_lookup(modules_buf, rl->modules[i].name);
      if (!mod_node) {
	btsd_debug("[BTSD] Can't find module definition '%s'.\n",
                   rl->modules[i].name);
	abort();
      }

      module = usrtc_node_getdata(mod_node);
      btsd_debug("[BTSD] Running '%s:%s' with %d params.\n",
	   module->name, module->binary, rl->modules[i].param_num);
      if (!module) abort();
      pid = fork();
      switch (pid) {
      case -1:
	abort();
	/* TODO: */
	break;
      case 0:
        //        task_scheduler_control(0, 0, SYS_SCHED_CTL_SET_PRIORITY,64);
        btsd_debug("[BTSD] Executing module '%s'.\n", module->binary);
	if (execve(module->binary,rl->modules[i].params, NULL) < 0) {
          btsd_debug("[BTSD] Failed executing module '%s'.\n", module->binary);
	  abort();
	  /* TODO: */
	}
	break;
      default:
        btsd_debug("[BTSD] Waiting for '%s:%d'.\n", module->binary, pid);
	pid = waitpid(pid, &status, 0);
        btsd_debug("[BTSD] waitpid ended '%s'.\n", module->binary);
        usleep(2000);
      }
    }

    rnode = usrtc_next(runlevel_conf->runlevels, rnode);
    btsd_debug("[BTSD] Executing next level.\n");
  }

  btsd_debug("[BTSD] Running process ended.\n");
  for (;;)
    sleep(10);
}

void delete_list(usrtc_t *list)
{
  usrtc_node_t *node;

  for (node = usrtc_first(list); node != NULL;
       node = usrtc_next(list, node))
    usrtc_delete_free(list, node);

  free(list);
}

usrtc_t *init_list(int (*list_cmp)(const void *a, const void *b),
		   void (*list_node_delete)(void *module, usrtc_node_t *node))
{
  usrtc_t *list;
  list = malloc(sizeof(*list));
  if (!list)
    return NULL;
  usrtc_init(list, USRTC_LIST, USRTC_COUNT_T_MAX, list_cmp);
  usrtc_set_allocator(list, NULL, list_node_delete, NULL);
  return list;
}

usrtc_t *init_module_list(void)
{
  return init_list(__cmp_module, __generic_delete);
}

static int __cmp_module(const void *a, const void *b)
{
  char *aa = ((btsd_module_t *)a)->name;
  char *bb = ((btsd_module_t *)b)->name;
  return strcmp(aa, bb);
}

static void __generic_delete(void *data, usrtc_node_t *node)
{
  free(data);
  free(node);
}
