/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Stanislav M. Ivankin <grepz@jarios.org>
 *
 * bootstrap service
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>

#include "btsd.h"
#include "syntax.h"

btsd_syntax_err_t parse_mod_params(sexp_t *expr, btsd_runlevel_module_t *lvl_mod);

static btsd_syntax_err_t eval_module(sexp_t *expr, btsd_module_t **module);
static btsd_syntax_err_t eval_runlevel(sexp_t *expr, btsd_runlevel_t **runlevel);

btsd_syntax_err_t eval_modules(sexp_t *expr, usrtc_t *mlist)
{
  int i, ret;
  usrtc_node_t *node;
  sexp_t *modsx;
  btsd_module_t *module;

  modsx = expr->list->next;
  for (i = 0; i < sexp_list_length(expr) - 1; i++) {
    ret = eval_module(modsx, &module);
    if (ret)
      return ret;

    node = malloc(sizeof(*node));
    usrtc_node_init(node, module);
    usrtc_insert(mlist, node, module);

    modsx = modsx->next;
  }

  return BTSD_SYNTAX_OK;
}

btsd_syntax_err_t eval_runlevel_conf(sexp_t *expr, btsd_runlevel_conf_t **rconf)
{
  usrtc_node_t *node;
  btsd_runlevel_conf_t *rconf_ptr;
  btsd_runlevel_t *runlevel;
  char name[TOKEN_LENGTH];
  int i, ret;
  sexp_t *runsx;

  ret = token_get_value(expr, SEXP_DQUOTE, BTSD_RUNLEVEL,
			name, TOKEN_LENGTH);
  if (ret) return ret;

  rconf_ptr = malloc(sizeof(*rconf_ptr));
  if (!rconf_ptr)
    return BTSD_SYNTAX_MEM_ERR;
  rconf_ptr->runlevels = init_module_list();
  if (!rconf_ptr->runlevels)
    goto __runlevel_conf_fail;

  strcpy(rconf_ptr->name, name);

  runsx = expr->list->next->next;
  for (i = 0; i < sexp_list_length(expr) - 2; i++) {
    ret = eval_runlevel(runsx, &runlevel);
    if (ret)
      goto __runlevel_conf_fail;

    node = malloc(sizeof(*node));
    if (!node) {
      /* TODO: Make function to delete runlevel, also set
       * deallocator to level usrtc list
       */
      free(runlevel->modules);
      free(runlevel);
      goto __runlevel_conf_fail;
    }

    usrtc_node_init(node, runlevel);
    usrtc_insert(rconf_ptr->runlevels, node, runlevel);

    runsx = runsx->next;
  }

  *rconf = rconf_ptr;

  return BTSD_SYNTAX_OK;

 __runlevel_conf_fail:
  free(rconf_ptr);
  return ret;
}

/* Local evals */

static btsd_syntax_err_t eval_module(sexp_t *expr, btsd_module_t **module)
{
  /* This way to minimize goto's */
  char token_val[TOKEN_LENGTH];
  btsd_module_t *mod;
  btsd_syntax_err_t ret;
  sexp_t *sx;

  *module = NULL;

  ret = list_get_token(expr, &sx, 0);
  if (ret || (ret = is_valid_token(sx)))
    return ret;

  ret = token_get_value(expr, SEXP_DQUOTE, BTSD_MODULE_BINARY,
			token_val, TOKEN_LENGTH);
  if (ret)
    return ret;

  mod = malloc(sizeof(*mod));
  if (!mod)
    return BTSD_SYNTAX_MEM_ERR;
  strcpy(mod->name, sx->val);
  strcpy(mod->binary, token_val);

  *module = mod;

  return BTSD_SYNTAX_OK;
}

static btsd_syntax_err_t eval_runlevel(sexp_t *expr, btsd_runlevel_t **runlevel)
{
  int ret, i, mods_num;
  sexp_t *sx;
  int level;
  btsd_runlevel_module_t *lvl_mod;
  btsd_runlevel_t *runlvl;

  *runlevel = NULL;

  ret = list_get_token(expr, &sx, 0);
  if (ret || (ret = is_valid_token(sx)))
    return ret;

  errno = 0;
  level = atoi(sx->val);
  if (level <= 0 || level >= 100)
    return BTSD_SYNTAX_VAL_ERR;

  mods_num = sexp_list_length(expr->list->next);
  lvl_mod = malloc(sizeof(*lvl_mod)*mods_num);
  if (!lvl_mod)
    return BTSD_SYNTAX_MEM_ERR;
  memset(lvl_mod, 0, sizeof(*lvl_mod));

  sx = expr->list->next->list;
  for (i = 0; i < sexp_list_length(expr->list->next); i++) {
    if (sx->ty == SEXP_VALUE) {
      if (strlen(sx->val) > TOKEN_LENGTH) {
	ret = BTSD_SYNTAX_SEXP_LENGTH_ERR;
	goto __eval_runlevel_fail;
      }
      strcpy(lvl_mod[i].name, sx->val);
    } else if (sx->ty == SEXP_LIST) {
      /* TODO: Convert to upper level handling and check types */
      if (strlen(sx->list->val) > TOKEN_LENGTH) {
	ret = BTSD_SYNTAX_SEXP_LENGTH_ERR;
	goto __eval_runlevel_fail;
      }
      strcpy(lvl_mod[i].name, sx->list->val);

      if (sx->list->next) {
	ret = parse_mod_params(sx->list->next, &lvl_mod[i]);
	if (ret)
	  goto __eval_runlevel_fail;
      }
    } else
      return BTSD_SYNTAX_SEXP_TYPE_ERR;

    sx = sx->next;
  }

  runlvl = malloc(sizeof(*runlvl));
  if (!runlvl) {
    free(lvl_mod->params);
    goto __eval_runlevel_fail;
  }

  runlvl->runlevel    = level;
  runlvl->modules_num = mods_num;
  runlvl->modules     = lvl_mod;

  *runlevel = runlvl;

  return BTSD_SYNTAX_OK;

 __eval_runlevel_fail:
  free(lvl_mod);

  return ret;
}

btsd_syntax_err_t parse_mod_params(sexp_t *expr, btsd_runlevel_module_t *lvl_mod)
{
  int i, num;
  sexp_t *sx;

  num = sexp_list_length(expr);
  if (!num)
    return BTSD_SYNTAX_OK;

  lvl_mod->params = malloc(sizeof(uintptr_t)*(num + 1));
  for (i = 0; i <= num; i++)
    lvl_mod->params[i] = malloc(sizeof(char) * DQUOTE_LENGTH);

  lvl_mod->param_num = num;

  sx = expr->list;
  for (i = 0; i < num; i++) {
    if (strlen(sx->val) > DQUOTE_LENGTH)
      goto __parse_params_fail;
    strcpy(lvl_mod->params[i], sx->val);
    sx = sx->next;
  }

  lvl_mod->params[num] = NULL;

  return BTSD_SYNTAX_OK;

 __parse_params_fail:
  /* FIXME: */
  free(lvl_mod->params);
  lvl_mod->params = NULL;

  return BTSD_SYNTAX_SEXP_LENGTH_ERR;
}

/* Generic */

btsd_syntax_err_t list_get_token(sexp_t *expr, sexp_t **sx, int n)
{
  int length, i;
  sexp_t *ptr;

  if (expr->ty != SEXP_LIST || expr->list->ty != SEXP_VALUE)
    return BTSD_SYNTAX_SEXP_TYPE_ERR;

  length = sexp_list_length(expr->next);
  if (n > length)
    return BTSD_SYNTAX_VAL_ERR;

  ptr = expr->list;
  for (i = 0; i < n; i++)
    ptr = ptr->next;

  *sx = ptr;

  return BTSD_SYNTAX_OK;
}

btsd_syntax_err_t token_get_value(sexp_t *expr, atom_t type,
				  const char *token_name, char *data,
				  unsigned int maxlen)
{
  sexp_t *token;

  token = find_sexp(token_name, expr);
  if (!token)
    return BTSD_SYNTAX_SEXP_NOTFOUND_ERR;

  if (token->ty != SEXP_VALUE && token->aty != type)
    return BTSD_SYNTAX_SEXP_TYPE_ERR;

  if (maxlen && strlen(token->next->val) > maxlen)
    return BTSD_SYNTAX_SEXP_LENGTH_ERR;

  strcpy(data, token->next->val);

  return BTSD_SYNTAX_OK;
}
