/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 MString Core Team <http://mstring.jarios.org>
 * (c) Copyright 2009 Michael Tsymbalyuk <mtzaurus@gmail.com>
 *
 * services/system/procmgr/procmgr.c: General logic of the Process Manager.
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <datalib/usrtc.h>
#include <errno.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <ipc/rpc.h>
#include <ns/ns.h>
#include <errno.h>
#include <devio/devio.h>
#include <sys/poll.h>
#include <sys/thread.h>
#include <debug.h>
#include <ns/ns.h>
#include <sys/procmgr.h>
#include <ipc/rpc.h>
#include <sys/proc.h>
#include <sys/syscall.h>
#include <sys/scheduler.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/uevent.h>
#include <lwip/tcp.h>
#include <pthread.h>
#include <netif/etherif.h>

#include "nm.h"
#include "lwip/netif.h"
#include "lwip/init.h"
#include "lwip/ip_addr.h"
#include "lwip/sockets.h"
#include "ipv4/lwip/ip.h"

#define NUM_INPUT_THREADS  1

typedef int spinlock_t;

static int __num_input_threads=NUM_INPUT_THREADS;
static event_t __input_events[NUM_INPUT_THREADS];
static pthread_t __input_threads[NUM_INPUT_THREADS];
static spinlock_t __ithread_locks[NUM_INPUT_THREADS];
static list_head_t __ithread_lists[NUM_INPUT_THREADS];
static pthread_t __tcp_tmr_thread;

#define spinlock_lock_irq(s)
#define spinlock_unlock_irq(s)

#define spinlock_lock(s)
#define spinlock_unlock(s)

#define lock_ithreads_array()
#define unlock_ithreads_array()

/**********************************************************************
*            Logic related to Ethernet-related drivers (inner)
***********************************************************************/
static network_host_ethops_t ethops_inner = {
  .etherif_init_device = etherif_init_device,
};

/**********************************************************************
*                Logic related to Network host (inner)
***********************************************************************/
err_t ip_input(struct pbuf *p, struct netif *inp)
{
  int id=0;

  p->stack_priv=inp;

  spinlock_lock_irq(&__ithread_locks[id]);
  list_add2tail(&__ithread_lists[id],&p->node);
  spinlock_unlock_irq(&__ithread_locks[id]);

  event_signal(&__input_events[id]);
  return 0;
}

static int __nhi_register_interface(struct netif *nif,netif_creds_t *creds)
{
  int r;
  struct ip_addr mask={INADDR_NONE};

  r=(netif_add(nif,IP_ADDR_ANY,&mask,IP_ADDR_ANY,NULL,creds->init,
               ip_input) == nif ? 0 : -1);
  if( !r ) {
    /* Continue interface initialization. */
    nif->pbuf_alloc=pbuf_alloc;
    nif->pbuf_free=pbuf_free;
    nif->pbuf_header=pbuf_header;
  }
  return r;
}

static int __nhi_unregister_interface(struct netif *nif)
{
  return 0;
}

static struct __network_host nethost_inner = {
  .register_interface=__nhi_register_interface,
  .unregister_interface=__nhi_unregister_interface,
  .eth_ops=&ethops_inner,
};

/**********************************************************************
*                Logic related to input threads (inner)
***********************************************************************/
static void *__nhi_input_thread(void *d)
{
  int id=(long)d;

  while( 1 ) {
    list_node_t *ln;
    struct pbuf *pb;
    struct netif *nif;

    event_wait(&__input_events[id]);

    spinlock_lock_irq(&__ithread_locks[id]);
    ln=list_node_first(&__ithread_lists[id]);
    list_del(ln);
    spinlock_unlock_irq(&__ithread_locks[id]);
    pb=(struct pbuf *)ln;
    nif=(struct netif *)pb->stack_priv;

    lwip_ip_input(pb,nif);
  }
}

static void *__tcp_timer_thread(void *d)
{
  struct timespec tv={0,TCP_FAST_INTERVAL*1000000};

  while(1) {
    nanosleep(&tv,NULL);
    tcp_tmr();
  }
  return NULL;
}
/**************************************************************************************/

static int __initialize_netmgr(void)
{
  int i,r;

  lwip_init();

  /* Start TCP timer threads. */
  if( pthread_create(&__tcp_tmr_thread,NULL,__tcp_timer_thread,NULL) ) {
    console_printf("[NetMgr] Can't start TCP timer thread !\n");
    return -1;
  }

  /* Initialize input threads. */
  for(i=0;i<NUM_INPUT_THREADS;i++) {
    list_init_head(&__ithread_lists[i]);

    r=event_init(&__input_events[i]);
    if( r ) {
      console_printf("[NetMgr] Can't create event for input thread !\n");
      return r;
    }

    if( pthread_create(&__input_threads[i],NULL,__nhi_input_thread,
                       (void *)(long)i) ) {
      console_printf("[NetMgr] Can't create input thread !\n");
      return -1;
    }
  }
  return 0;
}

static int __netmgr_main_loop(void)
{
  struct netif *nif;
  char *ifname="eth0";
  struct ip_addr ipaddr,netmask;

  console_printf("[NetMgr] entering main loop.\n");
  nif=netif_find(ifname);
  if( !nif ) {
    console_printf("[NetMgr] Can't find interface '%s' !\n",ifname);
    return -1;
  }

  console_printf("[NetMgr] successfully located network interface '%s'.\n",
                 ifname);

  ipaddr.addr=htonl(0xC0A80001);   /* 192.168.0.1   */
  netmask.addr=htonl(0xffffff00);  /* 255.255.255.0 */

  netif_set_ipaddr(nif,&ipaddr);
  netif_set_gw(nif,&ipaddr);
  netif_set_netmask(nif,&netmask);
  netif_set_default(nif);
  netif_set_up(nif);
  console_printf("[NetMgr] Interface '%s' is up and running.\n",ifname);

  sleep(1000);
  return 0;
}

const network_host_t *get_network_host(void)
{
  return &nethost_inner;
}

int main(void)
{
  start_service();

  open_console();
  console_printf("[NetMgr] Starting with PID=%d\n",getpid());

  if( __initialize_netmgr() ) {
    return -1;
  }
  netmgr_load_drivers();

  __netmgr_main_loop();

  console_printf("NETWORK MANAGER RETURNED FROM MAIN LOOP !!!!!!\n");
  for(;;) {
    sleep(100);
  }
}
