/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 MString Core Team <http://mstring.jarios.org>
 * (c) Copyright 2009 Michael Tsymbalyuk <mtzaurus@gmail.com>
 *
 * services/system/procmgr/elf.c: General logic of ELF loader.
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/thread.h>
#include <arch/system.h>  /* Not valid on host system. */
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>

#include "pm.h"
#include "elf.h"
#include "elf_e.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <autoconf.h>

/* Callbacks for performing some ELF-specific actions. */
typedef struct __elf_map_callback {
  void *priv_ctx;
  int (*sections_map_end)(elf64_sh_t *sections,int first_section,
                          int last_section,void *map_addr,
                          void *grant_addr,
                          size_t size,void *ctx);
} elf_map_callback_t;

typedef struct __elf_map_ctx_parm {
  elf64_sh_t *sections;
  int num_sections;
  elf_map_callback_t *cb;
  char *map_base,*server_base;
} elf_map_ctx_parm_t;

typedef struct __elf_map_ctx {
  char *first_addr,*end_addr;
  long flags;
  pid_t target;
  int fd;
  io_ctx_t *ioctx;
  elf64_sh_t *sections;
  int num_sections;
  elf_map_callback_t *cb;
  char *map_base,*server_base;
} elf_map_ctx_t;

static int __setup_dynamic_stats(elf64_sh_t *sections,int num_sections,
                                 dynamic_elf_stat_t *dstats);

unsigned long elf64_hash(char *n)
{
  unsigned char *name = (unsigned char *)n;
  unsigned long h = 0, g;
  while (*name) {
    h = (h << 4) + *name++;
    if ((g = h & 0xf0000000))
      h ^= g >> 24;
    h &= 0x0fffffff;
  }
  return h;
}

static shl_descr_data_t *__get_loaded_library_data(int cache_id,
                                                   pm_elf_ctx_t *elfctx)
{
  int i;

  /* Ugly, but quick to implement. */
  for( i=0; i<elfctx->num_libs; i++ ) {
    if( elfctx->lib_cache_ids[i] == cache_id ) {
      return elfctx->descr_data[i];
    }
  }

  return NULL;
}

void __set_loaded_library_data(pm_elf_ctx_t *elfctx,shared_lib_descr_t *cached_lib,
                               shl_descr_data_t *data)
{
  int i;

  /* Ugly, but quick to implement. */
  for( i=0; i<elfctx->num_libs; i++ ) {
    if( elfctx->lib_cache_ids[i] == cached_lib->cache_id ) {
      elfctx->descr_data[i] = data;
    }
  }
}

static int __validate_elf_header(elf_header_t *elfhdr)
{
  return *(int32_t*)elfhdr->e_ident == ELF_MAGIC &&
         elfhdr->e_ident[EI_CLASS] == ELFCLASS64;
}

static int __elf_load_program_body(pid_t target,int fd,io_ctx_t *ioctx,
                                   elf_map_ctx_parm_t *parm);

static int __elf_lookup(int fd,io_ctx_t *ioctx, void **ctx,pm_exec_type_t type)
{
  elf_header_t *elfhdr;
  pm_elf_ctx_t *elf_ctx;
  int t,r=1;

  if( ioctx->stats.st_size >= ELF_MIN_SIZE ) {
    elfhdr=(elf_header_t *)ioctx->iobuf.buf;
    if( *(int32_t*)elfhdr->e_ident == ELF_MAGIC ) {
      if( elfhdr->e_shnum > ELF_MAX_SECTIONS ) {
        return -EINVAL;
      }
      if( elfhdr->e_ident[EI_CLASS] == ELFCLASS64 ) {
        elf_ctx=malloc(sizeof(*elf_ctx));
        if( elf_ctx ) {
          /* Read all sections in advance. */
          elf_ctx->sections = malloc(elfhdr->e_shentsize*elfhdr->e_shnum);          
          if( !elf_ctx->sections ) {
            free(elf_ctx);
            return -ENOMEM;
          }
          t=pread(fd,elf_ctx->sections,elfhdr->e_shentsize*elfhdr->e_shnum,
                  elfhdr->e_shoff);
          if( t != elfhdr->e_shentsize*elfhdr->e_shnum ) {
            free(elf_ctx->sections);
            free(elf_ctx);
            return -ENOEXEC;
          }

          elf_ctx->num_sections=elfhdr->e_shnum;
          elf_ctx->elf_header=*elfhdr;
          elf_ctx->type=type;
          elf_ctx->weak_symbols = NULL;
          elf_ctx->mapped_vmas = NULL;
          *ctx=elf_ctx;
          elf_ctx->rt_launch=1;
          r=0;
        } else {
          r=-ENOMEM;
        }
      }
    }
  }

  return r;
}

static int __elf_check_shared_libraries(pid_t target,int fd,io_ctx_t *ioctx,
                                        pm_elf_ctx_t *elfctx)
{
  int i,r,l,num_libraries;
  dynamic_elf_stat_t *dstats=&elfctx->dyn_stats;
  elf64_dyn_t *dyns;

  __log("\n[ELF] Checking for dynamic libraries.\n");

  num_libraries = elfctx->sections[dstats->dynamic_id].sh_size / sizeof(elf64_dyn_t);
  elfctx->dynstr = malloc(elfctx->sections[dstats->dynstr_id].sh_size);
  if( !elfctx->dynstr ) {
    return -ENOMEM;
  }

  /* Populate string section where libriry names are placed. */
  r=pread(fd,elfctx->dynstr,(size_t)elfctx->sections[dstats->dynstr_id].sh_size,
          (off_t)elfctx->sections[dstats->dynstr_id].sh_offset);
  if( r != elfctx->sections[dstats->dynstr_id].sh_size ) {
    return -EIO;
  }

  /* Read all descriptors for dynamic symbols. */
  r=ioctx_read_offt(fd,ioctx,elfctx->sections[dstats->dynamic_id].sh_size,
                    elfctx->sections[dstats->dynamic_id].sh_offset);
  if( r != elfctx->sections[dstats->dynamic_id].sh_size ) {
    return -EIO;
  }

  /* Lookup all required libraries */
  dyns = (elf64_dyn_t*)ioctx->iobuf.buf;
  for(i=0,l=0; i<num_libraries; i++) {
    switch(dyns[i].d_tag) {
      case DT_NEEDED:
        if( l >= MAX_LIBS_PER_EXECUTABLE ) {
          __log_err("[ELF] Limit for shared libraries exceeded !\n");
          return -ENOEXEC;
        }

        r=check_library(&elfctx->dynstr[dyns[i].d_un.d_val]);
        if( r < 0 ) {
          __log_err("[ELF] Failed to validate library '%s' !\n",
                    &elfctx->dynstr[dyns[i].d_un.d_val]);
          return -ENOEXEC;
        }

        elfctx->lib_cache_ids[l] = r;
        elfctx->descr_data[l] = NULL;
        l++;
        break;
    }
  }

  elfctx->num_libs = l;
  return 0;
}

static int __setup_dynamic_stats(elf64_sh_t *sections,int num_sections,
                                 dynamic_elf_stat_t *dstats)
{
  int i,dynamic;

  memset(dstats,0,sizeof(*dstats));
  dynamic=0;

  for( i=1; i<num_sections; i++ ) {
    if( sections[i].sh_type == SHT_DYNAMIC ) {
      dstats->dynamic_id=i;
      dynamic |= 0x1;
    } else if( sections[i].sh_type == SHT_STRTAB && (sections[i].sh_flags & ESH_ALLOC) ) {
      dstats->dynstr_id=i;
      dynamic |= 0x2;
    } else if( sections[i].sh_type == SHT_DYNSYM ) {
      dstats->dynsym_id=i;
      dynamic |= 0x4;
    } else if( sections[i].sh_type == SHT_HASH ) {
      dstats->hash_id=i;
      dynamic |= 0x8;
    }
  }

  return dynamic;
}

static int __elf_pre_load(pid_t target,int fd,io_ctx_t *ioctx,void *ctx)
{
  int r,dynamic;
  pm_elf_ctx_t *elfctx=(pm_elf_ctx_t *)ctx;
  elf_header_t *elfheader=&elfctx->elf_header;
  elf64_sh_t *sections=elfctx->sections;
  dynamic_elf_stat_t *dstats=&elfctx->dyn_stats;
  const int OBLIG_SECTIONS = 4;

  /* First,determine if this executable is a dynamically-linked ELF. */
  dynamic=__setup_dynamic_stats(sections,elfheader->e_shnum,dstats);
  elfctx->dynamic = dynamic != 0;

  if( elfctx->dynamic ) {
    /* Make sure executable provided all obligatory sections. */
    if( dynamic != ((1<<OBLIG_SECTIONS)-1) ) {
      __log_err("[ELF] Target executable is missing some obligatory sections: %p instead of %p!\n",
                dynamic,(1<<OBLIG_SECTIONS)-1);
      return -ENOEXEC;
    }

    /* Make sure we can load all shared libraries. */
    if( (r=__elf_check_shared_libraries(target,fd,ioctx,elfctx)) ) {
      __log_err("[ELF] Shared libraries check failed with code %d.\n",r);
      return -ENOEXEC;
    }
  }
  return 0;
}

static int __can_merge_sections(elf64_sh_t *section1,elf64_sh_t *section2)
{
  #define _TF (ESH_WRITE | ESH_ALLOC)
  int can_merge = ((section1->sh_flags & _TF) ^ (section2->sh_flags & _TF))==0;

  if( can_merge ) {
    /* Only bounding VMAs can be merged. */
    if( section1->sh_addr < section2->sh_addr ) {
      can_merge = (PAGE_ALIGN(section1->sh_addr + section1->sh_size) >= (section2->sh_addr & ~PAGE_MASK));
    } else {
      can_merge = (PAGE_ALIGN(section2->sh_addr + section2->sh_size) >= (section1->sh_addr & ~PAGE_MASK));
    }
  }
  return can_merge;
}

static int __map_one_vma(int first,int last,elf_map_ctx_t *mapctx)
{
  char *our_vma,*target_vma;
  int size=(mapctx->end_addr-mapctx->first_addr);
  int prot=PROT_READ;
  int r,res;
  elf64_sh_t *sections=mapctx->sections;
  char *target_map_addr;

  if( mapctx->flags & ESH_WRITE ) {
    prot |= PROT_WRITE;
  }
  if( mapctx->flags & ESH_EXEC ) {
    prot |= PROT_EXEC;
  }

  target_map_addr = (char *)((long)mapctx->first_addr + (long)mapctx->map_base);
  target_vma=mmap_pid(mapctx->target,target_map_addr,size,prot,
                      MAP_PRIVATE | MAP_ANON | MAP_FIXED | MAP_CANRECPAGES,0,0);

  if( (long)target_vma & PAGE_MASK ) {
    __log_err("[ELF] __map_one_vma(): Failed to map [%p-%p] for target %d, prot=%p ! ERRNO=%d\n",
              target_map_addr,target_map_addr+size,
              mapctx->target,prot,errno);
    return -ENOMEM;
  }

  r=MAP_ANON | MAP_PRIVATE | MAP_POPULATE;
  if( mapctx->map_base || mapctx->server_base ) {
    /* In case of non-null 'map_base' we just use 'target_map_addr' calculated
     * earlier. Otherwise, we use 'server_base'.
     */
    if( mapctx->server_base ) {
      target_map_addr = (void *)((long)target_map_addr + (long)mapctx->server_base);
    }
    r |= MAP_FIXED;
  } else {
    target_map_addr = NULL;
  }

  our_vma=mmap(target_map_addr,size,prot | PROT_WRITE,r,0,0);
  if( (long)our_vma & PAGE_MASK ) {
    __log_err("[ELF] __map_one_vma(): Failed to allocate %d bytes at %p!\n",size,target_map_addr);
    return -ENOMEM;
  }

  if( mapctx->cb ) {
    r=mapctx->cb->sections_map_end(sections,first,last,our_vma,
                                   target_vma,
                                   size,mapctx->cb->priv_ctx);
    if( r ) {
      return r;
    }
  }

  /* Read all sections into our VMA. */
  res=0;
  do {
    elf64_sh_t *section=&sections[first];
    long vma_offset=(long)section->sh_addr-(long)mapctx->first_addr;

    /* No need to read anything from the file. */
    if( section->sh_type == SHT_NOBITS ) {
      memset(our_vma+vma_offset,0,section->sh_size);
      __log_err("[ELF] .bss mapped and zeroized at %p (%d bytes).\n",
                our_vma+vma_offset,section->sh_size);
    } else {
      r=pread(mapctx->fd,our_vma+vma_offset,section->sh_size,section->sh_offset);
      if( r != section->sh_size ) {
        __log_err("[ELF] Failed to read a section %d. r=%d, errno=%d\n",
                  first,r,errno);
        res=-EIO;
        break;
      }
    }
    first++;
  } while(first<=last);

  /* Delay granting since we will have to perform some manipulations
   * with our VMA.
   */
  if( !res && !mapctx->cb ) {
    res=grant_pages(our_vma,size,mapctx->target,(uintptr_t)target_vma);
    if( res ) {
      __log_err("[ELF] Failed to grant pages %p ==> %p, errno=%d\n",
                our_vma,target_vma,errno);
    }
  }

  /* No special handling for our VMA, so we can unmap it. */
  if( !mapctx->cb ) {
    munmap(our_vma,size);
  }

  return res;
}

static int __map_sections(elf_map_ctx_parm_t *parm,int first,int last,
                          pid_t target,int fd,io_ctx_t *ioctx)
{
  int r,i,first_idx;
  long last_addr,first_addr;
  elf64_sh_t *section=&parm->sections[first];
  long flags;
  elf_map_ctx_t mapctx;

  first_addr=section->sh_addr & ~(PAGE_SIZE-1);
  last_addr=first_addr+section->sh_size;
  first_idx=first;
  flags=section->sh_flags;

  mapctx.target=target;
  mapctx.fd=fd;
  mapctx.ioctx=ioctx;
  mapctx.sections=parm->sections;
  mapctx.cb=parm->cb;
  mapctx.map_base = parm->map_base;
  mapctx.server_base = parm->server_base;

  for(i=first+1;i<=last;i++) {
    section=&parm->sections[i];

    if( (section->sh_addr & ~PAGE_MASK) > PAGE_ALIGN(last_addr) ) {
      mapctx.first_addr=(char *)first_addr;
      mapctx.end_addr=(char *)(PAGE_ALIGN(last_addr));
      mapctx.flags=flags;
      r=__map_one_vma(first_idx,i-1,&mapctx);
      if( !r ) {
        return r;
      }
      first_idx=i;
      first_addr=section->sh_addr;
      last_addr=first_addr+section->sh_size;
      flags=section->sh_flags;
    } else {
      flags |= section->sh_flags;
      last_addr=section->sh_addr+section->sh_size;
    }
  }

  mapctx.first_addr=(void *)first_addr;
  mapctx.end_addr=(void *)PAGE_ALIGN(last_addr);
  mapctx.flags=flags;

  return __map_one_vma(first_idx,last,&mapctx);
}

/* This terrible hack is used for telling the ProcMgr that we must use
 * address of target symbol in _target_ space, but write it to relocation
 * address mapped in _our_ address space.
 * It will go away after moving lots of ProcMgr's logic to executer's
 * context.
 * Use it only for relocating symbols in ELF executable, not libraries !
 */
static bool __use_ugly_hack_address=false;
static long __ugly_hack_address;

static int __elf_relocate_symbol(elf64_rela_t *rel,elf64_sym_t *rsym,
                                 int *lib_cache_ids,int num_libs,
                                 long obj_base,char *dynstr,
                                 pm_elf_ctx_t *elfctx,bool resolve_weaks,
                                 int our_id)
{
  long load_addr;
  uint64_t *reladdr = (uint64_t *)(obj_base + (long)rel->r_offset);
  int h,idx,lib_idx;
  uint32_t *hash_syms,*hash_buckets;
  shl_descr_data_t *ldata;
  elf64_sym_t *s,*target_sym;

  if( __use_ugly_hack_address ) {
    obj_base = __ugly_hack_address;
  }

  switch( ELF64_R_TYPE(rel->r_info) ) {
    default:
      __log_err("[ELF] Unknown relocation type: %d\n",ELF64_R_TYPE(rel->r_info));
      return -EINVAL;
    case R_X86_64_RELATIVE:
      *reladdr = (obj_base + rel->r_addend);
      break;
    case R_X86_64_COPY:
    case R_X86_64_GLOB_DAT:
    case R_X86_64_JUMP_SLOT:
    case R_X86_64_64:
    case R_X86_64_PC32:
      if( rsym->st_value && !resolve_weaks &&
          ELF64_R_TYPE(rel->r_info) != R_X86_64_COPY) { /* Our own symbol. */
        target_sym = rsym;
        load_addr = obj_base;
      } else { /* External symbol. */
        h = elf64_hash(&dynstr[rsym->st_name]);
        target_sym=NULL;

        for( lib_idx=0; !target_sym && lib_idx < num_libs; lib_idx++ ) {
          ldata = __get_loaded_library_data(lib_cache_ids[lib_idx],elfctx);

          if( !ldata ) {
            __log_err("[ELF] Failed to lookup already cached library (cache id=%d) !\n",
                      lib_cache_ids[lib_idx]);
            return -EINVAL;
          }

          load_addr = (long)ldata->load_addr;
          hash_buckets = &ldata->hash[2];
          hash_syms = hash_buckets + ldata->hash[0];

          idx=hash_buckets[h % ldata->hash[0]];
          for(s=&ldata->dynsyms[idx];idx != SHN_UNDEF;s=&ldata->dynsyms[idx]) {
            if( !strcmp(&ldata->symnames[s->st_name],&dynstr[rsym->st_name]) &&
                s->st_value ) {
              /* We accept only already defined symbols. */
              if( resolve_weaks ) {
                /* Outer symbols are more preferable than ours. */
                if( ELF64_SYM_BINDATTR(rsym->st_info) == STB_WEAK &&
                    lib_cache_ids[lib_idx] != our_id ) {
                  __log("[ELF] Better relocation found for weak symbol '%s' : new cache_id = %d (our is %d)\n",
                        &dynstr[rsym->st_name],lib_cache_ids[lib_idx],our_id);
                  target_sym=s;
                  break;
                }
              } else {
                target_sym=s;
                break;
              }
            }
            idx=hash_syms[idx];
          }
        }
      }

      if( !target_sym ) {
        __log_err("[ELF] Can't resolve dynamic symbol: '%s'\n",&dynstr[rsym->st_name]);
        return -EINVAL;
      } else {
        /* Perform actual relocation. */
        uint64_t symaddr = target_sym->st_value + load_addr;
        uint32_t naddr;

        switch( ELF64_R_TYPE(rel->r_info) ) {
          case R_X86_64_GLOB_DAT:
          case R_X86_64_JUMP_SLOT:
            *reladdr = (uint64_t)symaddr;
            break;
          case R_X86_64_PC32:
            naddr = (uint32_t)(symaddr + rel->r_addend - rel->r_offset);
            __log_err("[ELF] Patching R_X86_64_PC32 relocation at %p (symaddr = %p) to %p\n",
                      reladdr,symaddr,naddr);
            __log_err("      rel->r_addend = %p, rel->r_offset = %p\n",
                      rel->r_addend,rel->r_offset);
            *((uint32_t *)reladdr) = naddr;
            break;
          case R_X86_64_64:
            *reladdr = ((uint64_t)symaddr + rel->r_addend);
            break;
          case R_X86_64_COPY:
            memcpy(reladdr,(void *)symaddr,target_sym->st_size);
            __log_err("==> Copied symbol '%s' to %p from %p (%d bytes)\n",
                      &dynstr[rsym->st_name],reladdr,symaddr,target_sym->st_size);
            break;
        }

        __log("[ELF] Relocated symbol '%s' (R-location = %p) to %p\n",
              &dynstr[rsym->st_name],reladdr,*reladdr);

        /* Weak symbols need to be linked in the list for further processing. */
        if( ELF64_SYM_BINDATTR(rsym->st_info) == STB_WEAK ) {
          weaksym_descr_t *wd,*wdp;

          if( !resolve_weaks ) {
            __log("[ELF] Weak symbol '%s' found in cache id = %d\n",
                  &dynstr[rsym->st_name],our_id);

            wd = (weaksym_descr_t *)malloc(sizeof(*wd));
            if( !wd ) {
              __log_err("[ELF] Can't allocate memory for a weak symbol descriptor !\n");
              return -ENOMEM;
            }

            wd->name = &dynstr[rsym->st_name];
            wd->sym = *rsym;
            wd->rsym = *rel;
            wd->cache_id = our_id;
            wd->obj_base = obj_base;
            wd->dynstr = dynstr;
            wdp=elfctx->weak_symbols;
            elfctx->weak_symbols=wd;
            wd->next=wdp;
          }
        }
      }
      break;
  }
  return 0;
}

/* obj_base: reflecfs load address of a module in which relocations
 *           take place.
 */
static int __elf_relocate_symbols(elf64_sh_t *sections,int num_sections,
                                  int *lib_cache_ids,int num_libs,long obj_base,
                                  int fd,io_ctx_t *ioctx,elf64_sym_t *dsyms,
                                  char *dynstr,pm_elf_ctx_t *elfctx,
                                  int our_id)
{
  int i,j,r=0;

  for( j=0; j < num_sections; j++ ) {
    if( sections[j].sh_type == SHT_RELA ) {
      elf64_rela_t *relas;

      /* Read relocations. */
      r=ioctx_read_offt(fd,ioctx,sections[j].sh_size,sections[j].sh_offset);
      if( r != sections[j].sh_size ) {
        __log_err("[ELF] Can't read library's .rela section !\n");
        r=-EIO;
        goto out;
      }

      relas=(elf64_rela_t *)ioctx->iobuf.buf;
      for( r=0,i=0; !r && i<sections[j].sh_size/sizeof(elf64_rela_t); i++ ) {
        elf64_sym_t *rsym=&dsyms[ELF64_R_SYM(relas[i].r_info)];

        r=__elf_relocate_symbol(&relas[i],rsym,lib_cache_ids,num_libs,
                                obj_base,dynstr,elfctx,false,our_id);
      }
    }
  }
out:
  return r;
}

static int __elf_resolve_externals(pid_t target,int fd,io_ctx_t *ioctx,
                                   pm_elf_ctx_t *elfctx,void *load_addr)
{
  weaksym_descr_t *wd,*wd_prev;
  elf64_sym_t *dsyms;
  int r;

  /* Perform relocations for all symbols in this ELF. */
  dsyms = (elf64_sym_t*)malloc(elfctx->sections[elfctx->dyn_stats.dynsym_id].sh_size);
  if( !dsyms ) {
    __log_err("[ELF] Can't allocate %d bytes for executable's .dynsym section !\n",
              elfctx->sections[elfctx->dyn_stats.dynsym_id].sh_size);
    return -ENOMEM;
  }

  r = pread(fd,dsyms,(size_t)elfctx->sections[elfctx->dyn_stats.dynsym_id].sh_size,
            elfctx->sections[elfctx->dyn_stats.dynsym_id].sh_offset);
  if( r != elfctx->sections[elfctx->dyn_stats.dynsym_id].sh_size ) {
    __log_err("[ELF] Can't read executable's .dynsym section !\n");
    goto out;
  }

  /* UGLY HACK !!!! [mt] */
  __use_ugly_hack_address = true;
  __ugly_hack_address = 0;

  __log("[ELF] Performing ELF dynamic relocation (num_libs: %d, num_sections: %d, load addr: %p)\n",
            elfctx->num_libs,elfctx->num_sections,load_addr);
  r=__elf_relocate_symbols(elfctx->sections,elfctx->num_sections,
                           elfctx->lib_cache_ids,elfctx->num_libs,
                           (long)load_addr,fd,ioctx,
                           dsyms,elfctx->dynstr,elfctx,
                           -1); /* Use fake 'library ID' */
  __log("[ELF] ELF Relocation finished with code %d\n",r);

  /* UGLY HACK !!!! [mt] */
  __use_ugly_hack_address = false;

  if( !r && elfctx->weak_symbols ) {
    __log("[ELF] Processing weak symbols.\n");
    for(wd_prev = NULL,wd=elfctx->weak_symbols; !r && wd; wd=wd->next) {
      r=__elf_relocate_symbol(&wd->rsym,&wd->sym,
                              elfctx->lib_cache_ids,elfctx->num_libs,
                              wd->obj_base,wd->dynstr,elfctx,true,
                              wd->cache_id);
      __log("[ELF] Weak symbol '%s' handled with code %d\n",wd->name,r);
      if( wd_prev ) {
        free(wd_prev);
      }
      wd_prev = wd;
    }
    free(wd_prev);
    __log("[ELF] Done processing weak symbols.\n");
  }  
out:
  free(dsyms);
  return r;
}

typedef struct __elf_shlib_cb_ctx {
  shared_lib_descr_t *cached_lib;
  pm_elf_ctx_t *elfctx;
  int rounds;
  void *load_addr,*last_addr;
} elf_shlib_cb_ctx_t;

static int __elf_shlib_sections_map_end(elf64_sh_t *sections,int first_section,
                                        int last_section,void *map_addr,
                                        void *grant_addr,
                                        size_t size,void *ctx)
{
  elf_shlib_cb_ctx_t *lctx = (elf_shlib_cb_ctx_t*)ctx;
  struct __vma_range *new_vma,*prev_vma;

  if( !lctx->rounds ) {
    lctx->load_addr = map_addr;
  }

  lctx->rounds++;
  lctx->last_addr = (void *)((char *)map_addr + size);

  __log("__elf_shlib_sections_map_end(): MAP [%p:%p)\n",map_addr,map_addr+size);

  new_vma = (struct __vma_range *)malloc(sizeof(*new_vma));
  if( !new_vma ) {
    __log_err("[ELF] Failed to allocate %d bytes for VMA range !\n",
              sizeof(*new_vma));
    return -ENOMEM;
  }  

  new_vma->start = map_addr;
  new_vma->size = size;
  new_vma->grant_addr = grant_addr;

  prev_vma = lctx->elfctx->mapped_vmas;
  lctx->elfctx->mapped_vmas = new_vma;
  new_vma->next = prev_vma;

  return 0;
}

static int __elf_load_shared_library(pid_t target,io_ctx_t *ioctx,
                                     shared_lib_descr_t *cached_lib,
                                     int depth,pm_elf_ctx_t *elfctx)
{
  int i,r,needed_libs;
  elf_header_t lib_ehdr;
  dynamic_elf_stat_t dstats;
  elf64_sh_t *sections;
  char *dynstr;
  elf64_dyn_t *dyns;
  elf_map_ctx_parm_t p;
  elf_map_callback_t cb;
  elf_shlib_cb_ctx_t privctx;
  elf64_sym_t *dsyms;
  shl_descr_data_t *lib_descr_data;
  int deps_ids[MAX_LIBS_PER_EXECUTABLE];

  if( __get_loaded_library_data(cached_lib->cache_id,elfctx) ) {
    return 0;
  }

  if( !--depth ) {
    __log_err("[ELF] Max recursion depth reached during libraries resolve.\n");
    return -EBUSY;
  }

  r=pread(cached_lib->fd,&lib_ehdr,sizeof(lib_ehdr),0);
  if( r != sizeof(lib_ehdr) ) {
    __log_err("[ELF] Can't read ELF header for library !\n");
    return -EIO;
  }

  if( !__validate_elf_header(&lib_ehdr) ) {
    __log_err("[ELF] Invalid ELF header in library !\n");
    return -EIO;
  }

  __log("[ELF] Starting to load library '%s' (sections: %d, depth=%d)\n",
        cached_lib->filename,lib_ehdr.e_shnum,depth);
  sections = malloc(lib_ehdr.e_shentsize*lib_ehdr.e_shnum);
  if( !sections ) {
    __log_err("[ELF] Can't allocate memory for library's sections !\n");
    return -ENOMEM;
  }

  r=pread(cached_lib->fd,sections,lib_ehdr.e_shentsize*lib_ehdr.e_shnum,
          lib_ehdr.e_shoff);
  if( r != lib_ehdr.e_shentsize*lib_ehdr.e_shnum ) {
    __log_err("[ELF] Can't read ELF sections for library !\n");
    return -EIO;
  }

  /* TODO: [mt] check return value properly ! */
  __setup_dynamic_stats(sections,lib_ehdr.e_shnum,&dstats);

  dynstr=malloc(sections[dstats.dynstr_id].sh_size);
  if( !dynstr ) {
    __log_err("[ELF] Can't allocate memory for library's  string section !\n");
    r=-ENOMEM;
    goto out_free;
  }

  /* Populate string section where libriry names are placed. */
  r=pread(cached_lib->fd,dynstr,(size_t)sections[dstats.dynstr_id].sh_size,
          (off_t)sections[dstats.dynstr_id].sh_offset);
  if( r != sections[dstats.dynstr_id].sh_size ) {
    __log_err("[ELF] Can't read library's string section !\n");
    r=-EIO;
    goto out_free;
  }

  dsyms=(elf64_sym_t*)malloc(sections[dstats.dynsym_id].sh_size);
  if( !dsyms ) {
    __log_err("[ELF] Can't allocate buffer for library's .dynsym section !\n");
    r=-ENOMEM;
    goto out_free;
  }

  /* Read all descriptors for dynamic symbols. */
  r=pread(cached_lib->fd,dsyms,(size_t)sections[dstats.dynsym_id].sh_size,
          sections[dstats.dynsym_id].sh_offset);
  if( r !=  sections[dstats.dynsym_id].sh_size ) {
    __log_err("[ELF] Can't read library's .dynsym section !\n");
    r=-EIO;
    goto out_free;
  }

  /* Load library body in advance to simplify resolve logic. */
  privctx.cached_lib = cached_lib;
  privctx.elfctx=elfctx;
  privctx.rounds=0;

  cb.sections_map_end = __elf_shlib_sections_map_end;
  cb.priv_ctx = &privctx;

  p.sections = sections;
  p.num_sections = lib_ehdr.e_shnum;
  p.cb = &cb;
  p.map_base = elfctx->libs_map_base;
  p.server_base = NULL;

  r = __elf_load_program_body(target,cached_lib->fd,ioctx,&p);
  if( r ) {
    __log_err("[ELF] Can't read librarie's body ! r=%d\n",r);
    goto out_free;
  }

  /* Adjust library base address for next library. */
  elfctx->libs_map_base = privctx.last_addr;

  dyns=(elf64_dyn_t *)malloc(sections[dstats.dynamic_id].sh_size);
  if( !dyns ) {
    __log_err("[ELF] Can't allocate buffer for storing '.dynamic' symbols !\n");
    goto out_free;
  }

  /* Read descriptors for all symbols in this library. */
  r=pread(cached_lib->fd,dyns,(size_t)sections[dstats.dynamic_id].sh_size,
          sections[dstats.dynamic_id].sh_offset);
  if( r != sections[dstats.dynamic_id].sh_size ) {
    __log_err("[ELF] Can't read library's .dynamic section !\n");
    r=-EIO;
    goto out_free;
  }

  lib_descr_data = (shl_descr_data_t *)malloc(sizeof(shl_descr_data_t));
  if( !lib_descr_data ) {
    __log_err("[ELF] Can't allocate descriptor for shared library !\n");
    r=-ENOMEM;
    goto out_free;
  }

  /* Read hash table for this library. */
  lib_descr_data->hash = malloc(sections[dstats.hash_id].sh_size);
  if( !lib_descr_data->hash ) {
    __log_err("[ELF] Can't allocate hash buffer of %d bytes for shared library !\n",
              sections[dstats.hash_id].sh_size);
    r=-ENOMEM;
    goto out_free;
  }
  lib_descr_data->hash_size=sections[dstats.hash_id].sh_size;

  r=pread(cached_lib->fd,lib_descr_data->hash,
          (size_t)sections[dstats.hash_id].sh_size,
          sections[dstats.hash_id].sh_offset);
  if( r != sections[dstats.hash_id].sh_size ) {
    __log_err("[ELF] Can't read library's .hash section !\n");
    r=-EIO;
    goto out_free;
  }

  /* Mark library as already loaded for target process. */
  lib_descr_data->num_sections = lib_ehdr.e_shnum;
  lib_descr_data->sections = sections;
  lib_descr_data->load_addr = privctx.load_addr;
  lib_descr_data->dynsyms = dsyms;
  lib_descr_data->num_dynsyms = sections[dstats.dynsym_id].sh_size/sizeof(elf64_sym_t);
  lib_descr_data->symnames = dynstr;

  __set_loaded_library_data(elfctx,cached_lib,lib_descr_data);

  /* Now we should load all shared library this library depends on.
   */
  deps_ids[0] = cached_lib->cache_id; /* We always depend on ourself */
  needed_libs = 1;

  for( i=0, r=0; !r && i<sections[dstats.dynamic_id].sh_size/sizeof(elf64_dyn_t); i++ ) {
    switch(dyns[i].d_tag) {
      /* Another shared library we depend on.
       */
      case DT_NEEDED:
        r=check_library(&dynstr[dyns[i].d_un.d_val]);
        __log("[ELF] External library found: '%s', cache id=%d\n",
              &dynstr[dyns[i].d_un.d_val],r);
        if( r >= 0 ) {
          shared_lib_descr_t *dep_lib=get_cached_library(r);
          if( dep_lib ) {
            /* Try to load new library in case it wasn't loaded earlier.
             */
            if( !__get_loaded_library_data(dep_lib->cache_id,elfctx) ) {
              r=__elf_load_shared_library(target,ioctx,dep_lib,depth,elfctx);
            } else {
              r=0;
            }
            deps_ids[needed_libs] = dep_lib->cache_id;
            needed_libs++;
          } else {
            __log_err("[ELF] Can't lookup cached library for target library '%s'!\n",
                      &dynstr[dyns[i].d_un.d_val]);
            r=-EINVAL;
          }
        }
        break;
    }
  }

  /* All external dependencies are seem to be met, so we can perform
   * relocations for those symbols that require it.
   */
  __log("[ELF] Relocating symbols in '%s'\n",cached_lib->filename);
  r=__elf_relocate_symbols(sections,lib_ehdr.e_shnum,deps_ids,needed_libs,
                           (long)lib_descr_data->load_addr,
                           cached_lib->fd,ioctx,dsyms,dynstr,elfctx,
                           cached_lib->cache_id);
  __log("[ELF] Relocation done: %d\n",r);
  if( !r ) {
    __log( "  ** Library '%s' loaded with cache id = %d.\n",
           cached_lib->filename,cached_lib->cache_id);
    free(dyns);
    return 0;
  }
out_free:
  free(sections);
  return r;
}

static int __elf_load_shared_libraries(pid_t target,int fd,io_ctx_t *ioctx,
                                       pm_elf_ctx_t *elfctx)
{
  int i,r;
  shared_lib_descr_t *cached_lib;

  elfctx->libs_map_base = (char *)SHLIB_RANGE_START;
  for( i=0,r=0; !r && i<elfctx->num_libs; i++ ) {
    cached_lib = get_cached_library(elfctx->lib_cache_ids[i]);

    if( !cached_lib ) {
      __log_err("[ELF] Can't lookup cached library %d !\n",elfctx->lib_cache_ids[i]);
      return -EINVAL;
    }
    __log("[ELF ] Loading shared library '%s'\n",cached_lib->filename);
    r=__elf_load_shared_library(target,ioctx,cached_lib,MAX_LIBS_RESOLVE_DEPTH,elfctx);
  }
  return r;
}

static int __grant_areas(pm_elf_ctx_t *elfctx,pid_t target)
{
  int r=0;
  struct __vma_range *vma;

  if( elfctx->mapped_vmas ) {
    for(vma=elfctx->mapped_vmas; !r && vma; vma=vma->next) {
      r=grant_pages(vma->start,vma->size,target,(uintptr_t)vma->grant_addr);
      __log("[ELF] Granting %p to %p (%d bytes), r=%d\n",vma->start,vma->grant_addr,vma->size,r);
      if( r ) {
        __log_err("[ELF] Failed to grant %d VMA pages %p ==> %p, errno=%d\n",
                  vma->size,vma->start,vma->grant_addr,errno);
      }
    }
  }
  return r;
}

static int __elf_load_dynamic_image(pid_t target,int fd,io_ctx_t *ioctx,
                                    pm_elf_ctx_t *elfctx)
{
  int r;
  elf_map_ctx_parm_t p;
  elf_shlib_cb_ctx_t privctx;
  elf_map_callback_t cb;

  __log("Loading dynamic ELF (%d libraries to attach) ...\n",
        elfctx->num_libs);

  /* Step 1: load program's body. */
  privctx.cached_lib = NULL;
  privctx.elfctx=elfctx;
  privctx.rounds=0;

  cb.sections_map_end = __elf_shlib_sections_map_end;
  cb.priv_ctx = &privctx;

  p.sections = elfctx->sections;
  p.num_sections = elfctx->num_sections;
  p.cb = &cb;
  p.map_base = NULL;
  p.server_base = (void *)DYNIMAGE_RANGE_START;

  r=__elf_load_program_body(target,fd,ioctx,&p);
  if( r ) {
    __log_err("[ELF] Failed to load ELF program body ! r=%d\n",r);
    return r;
  }

  /* Step 2: load shared libraries. */
  r=__elf_load_shared_libraries(target,fd,ioctx,elfctx);
  if( r ) {
    __log_err("[ELF] Failed to load shared libraries for ELF program ! r=%d\n",r);
    return r;
  }

  __log("[ELF] Resolving externals for ELF executable itself (load address: %p)\n",
            privctx.load_addr);

  /* Step 3: resolve program's external dependencies. */
  r=__elf_resolve_externals(target,fd,ioctx,elfctx,(void *)DYNIMAGE_RANGE_START);
  if( !r ) {
    __log("[ELF] Granting pages to target process ...\n");
    r = __grant_areas(elfctx,target);
    __log("[ELF] Done granting pages: %d\n",r);
  }

  return r;
}

static int __elf_load_program_body(pid_t target,int fd,io_ctx_t *ioctx,
                                   elf_map_ctx_parm_t *parm)
{
  int i,r,__first_id;

  __first_id=0;
  if( !parm->sections[__first_id].sh_addr ) {
    __first_id++;
  }

  for(r=0,i=__first_id+1;!r && i<parm->num_sections;i++) {
    if( !parm->sections[i].sh_addr ) {
      r=__map_sections(parm,__first_id,i-1,target,fd,ioctx);
      break;
    }
    if( !__can_merge_sections(&parm->sections[i-1],&parm->sections[i]) ) {
      r=__map_sections(parm,__first_id,i-1,target,fd,ioctx);
      __first_id=i;

      if( i != parm->num_sections-1 ) {
        continue;
      } else {
        r=__map_sections(parm,i,i,target,fd,ioctx);
        break;
      }
    }
  }

  if( r ) {
    return r;
  }

  return 0;
}

static int __elf_load_image(pid_t target,int fd,io_ctx_t *ioctx,void *ctx,
                            exec_fmt_task_ctx_t *task_attrs)
{
  pm_elf_ctx_t *elfctx=(pm_elf_ctx_t *)ctx;
  int r;
  elf_map_ctx_parm_t p;

  if( elfctx->dynamic ) {
    r = __elf_load_dynamic_image(target,fd,ioctx,elfctx);
  } else {
    p.sections = elfctx->sections;
    p.num_sections = elfctx->num_sections;
    p.cb = NULL;
    p.map_base = p.server_base = NULL;

    r = __elf_load_program_body(target,fd,ioctx,&p);
  }

  if( !r ) {
    task_attrs->entrypoint=elfctx->elf_header.e_entry;
  }
  return r;
}

static void __elf_cleanup(void *ctx)
{
  pm_elf_ctx_t *elfctx=(pm_elf_ctx_t *)ctx;
  struct __vma_range *vma,*vma_prev;

  /* Unmap mappings involved into ELF processing. */
  if( elfctx->mapped_vmas ) {
    for(vma=elfctx->mapped_vmas,vma_prev=NULL;vma;vma=vma->next) {
      if( vma_prev ) {
        free(vma_prev);
      }
      munmap(vma->start,vma->size);
      vma_prev = vma;
    }
    free(vma_prev);
  }

  free(elfctx->sections);
  free(elfctx);
}

executable_format_t elf_exec_fmt = {
  .id="ELF",
  .lookup=__elf_lookup,
  .pre_load=__elf_pre_load,
  .load_image=__elf_load_image,
  .cleanup=__elf_cleanup,
};
