/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 MString Core Team <http://mstring.jarios.org>
 * (c) Copyright 2005,2008 Tirra <madtirra@jarios.org>
 *
 * include/kernel/elf.h: elf defines
 *
 */


#ifndef __KERNEL_ELF_H__
#define __KERNEL_ELF_H__

//#include <sys/types.h>

#define EI_NIDENT  16

/* sections flags */
#define ESH_WRITE  0x1 /* writable */
#define ESH_ALLOC  0x2 /* allocatable */
#define ESH_EXEC   0x4 /* executable */
/* section types */
#define SHT_NULL           0 /* Section header table entry unused */
#define SHT_PROGBITS       1 /* Program specific (private) data */
#define SHT_SYMTAB         2 /* Link editing symbol table */
#define SHT_STRTAB         3 /* A string table */
#define SHT_RELA           4 /* Relocation entries with addends */
#define SHT_HASH           5 /* A symbol hash table */
#define SHT_DYNAMIC        6 /* Information for dynamic linking */
#define SHT_NOTE           7 /* Information that marks file */
#define SHT_NOBITS         8 /* Section occupies no space in file */
#define SHT_REL            9 /* Relocation entries, no addends */
#define SHT_SHLIB          10 /* Reserved, unspecified semantics */
#define SHT_DYNSYM         11 /* Dynamic linking symbol table */
#define SHT_INIT_ARRAY     14 /* Array of ptrs to init functions */
#define SHT_FINI_ARRAY     15 /* Array of ptrs to finish functions */
#define SHT_PREINIT_ARRAY  16 /* Array of ptrs to pre-init funcs */
#define SHT_GROUP          17 /* Section contains a section group */
#define SHT_SYMTAB_SHNDX   18 /* Indicies for SHN_XINDEX entries */

#define ELF_MAGIC  0x464C457F

#define EI_CLASS    4  /* File class */
#define EI_DATA     5  /* Data encoding */
#define EI_VERSION  6  /* File version */

#define ELFCLASS32  1  /* 32-bit ELF executable */
#define ELFCLASS64  2  /* 64-bit ELF executable */

/* elf image header */
typedef struct __elf_header_t 
{
  unsigned char e_ident[EI_NIDENT]; /* ELF64 magic number */
  uint16_t e_type; /* elf type */
  uint16_t e_machine; /* elf required architecture */
  uint32_t e_version; /* elf object file version */
  uintptr_t e_entry; /* entry point virtual address */
  uintptr_t e_phoff; /* program header file offset */
  uintptr_t e_shoff; /* section header file offset */
  uint32_t e_flags; /* processor specific object tags */
  uint16_t e_ehsize; /* elf header size in bytes */
  uint16_t e_phentsize; /* program header table entry size */
  uint16_t e_phnum; /* program header count */
  uint16_t e_shentsize; /* section header table entry size */
  uint16_t e_shnum; /* section header count */
  uint16_t e_shstrndx; /* section header string table index */
} elf_header_t;

/* program header */
typedef struct {
  uint32_t p_type; /* program segment type */
  uint32_t p_flags; /* segment flags */
  uintptr_t p_offset; /* segment image offset */
  uintptr_t p_vaddr; /* segment virtual address */
  uintptr_t p_paddr; /* segment physical address */
  uint64_t p_filesz; /* segment size on image */
  uint64_t p_memsz; /* segment size on memory */
  uint64_t p_align; /* segment alignment on image and memory */
} elf64_pr_t;

/* section header */
typedef struct {
  uint32_t sh_name; /* section name, actually is a index in string table */
  uint32_t sh_type; /* type of a section */
  uint64_t sh_flags; /* section flags */
  uintptr_t sh_addr; /* section virtual address at execution */
  uintptr_t sh_offset; /* section image offset */
  ulong_t sh_size; /* size of section in bytes */
  uint32_t sh_link; /* index of another section */
  uint32_t sh_info; /* additional section information */
  ulong_t sh_addralign; /* section alignment */
  size_t sh_entsize; /* entry size if section holds table*/
} elf64_sh_t;

typedef struct elf64_sym {
  uint32_t      st_name;
  unsigned char st_info;
  unsigned char st_other;
  uint16_t    st_shndx;
  uint64_t    st_value;
  uint64_t    st_size;
} elf64_sym_t;

typedef struct elf64_dyn {
  uint64_t d_tag;
  union {
    uint64_t d_val;
    uint64_t d_ptr;
  } d_un;
} elf64_dyn_t;

typedef struct __elf64_rela {
  uint64_t r_offset,r_info,r_addend;
} elf64_rela_t;

enum elf64_dyn_tags {
  DT_NEEDED = 1,
  DT_PLTRELSZ = 2, /* Total size, in bytes, of the relocation entries
                      associated with the procedure linkage table. */
  DT_PLTGOT = 3,   /* Contains an address associated with the linkage table. The
                      specific meaning of this field is processor-dependent. */
  DT_HASH = 4,     /* Address of the symbol hash table. */
  DT_STRTAB = 5,   /* Address of the dynamic string table. */
  DT_SYMTAB = 6,   /* Address of the dynamic symbol table. */
  DT_RELA = 7,     /* Address of a relocation table with Elf64_Rela entries. */
  DT_RELASZ = 8,   /* Total size, in bytes, of the DT_RELA relocation table. */
  DT_RELAENT = 9,  /* Size, in bytes, of each DT_RELA relocation entry. */
  DT_STRSZ = 10,   /* Total size, in bytes, of the string table. */
  DT_SYMENT = 11,  /* Size, in bytes, of each symbol table entry. */
  DT_INIT = 12,    /* Address of the initialization function. */
  DT_FINI = 13,    /* Address of the termination function. */
  DT_SONAME = 14,  /* The string table offset of the name of this shared object. */
  DT_RPATH = 15,   /* The string table offset of a shared library search path string. */
  DT_SYMBOLIC = 16,/* The presence of this dynamic table entry modifies the
                      ignored symbol resolution algorithm for references within the
                      library. Symbols defined within the library are used to
                      resolve references before the dynamic linker searches the
                      usual search path. */
  DT_REL = 17,     /* Address of a relocation table with Elf64_Rel entries. */
  DT_RELSZ = 18,   /* Total size, in bytes, of the DT_REL relocation table. */
  DT_RELENT = 19,  /* Size, in bytes, of each DT_REL relocation entry. */
  DT_PLTREL = 20,  /* Type of relocation entry used for the procedure linkage
                      table. The d_val member contains either DT_REL or DT_RELA. */
  DT_DEBUG = 21,   /* Reserved for debugger use. */
  DT_TEXTREL = 22, /* The presence of this dynamic table entry signals that
                      the ignored relocation table contains relocations for
                      a non-writable segment. */
  DT_JMPREL = 23,  /* Address of the relocations associated with the procedure
                      linkage table. */
  DT_BIND_NOW = 24,/* The presence of this dynamic table entry signals that the
                      ignored dynamic loader should process all relocations for
                      this object before transferring control to the program. */
  DT_INIT_ARRAY = 25, /* Pointer to an array of pointers to initialization functions. */
  DT_FINI_ARRAY = 26, /* Pointer to an array of pointers to termination functions. */
  DT_INIT_ARRAYSZ = 27, /* Size, in bytes, of the array of initialization functions. */
  DT_FINI_ARRAYSZ = 28, /* Size, in bytes, of the array of termination functions. */
  DT_LOOS = 0x60000000, /* Defines a range of dynamic table tags that are reserved for
                           environment-specific use. */
  DT_HIOS = 0x6FFFFFFF,
  DT_LOPROC = 0x70000000, /* Defines a range of dynamic table tags that are reserved for
                             processor-specific use. */
  DT_HIPROC = 0x7FFFFFFF,
};

#define SHN_UNDEF 0 /* Used to mark an undefined or
                       meaningless section reference */

#define ELF64_SYM_TYPE(x)  ((x) & 0xf) /* Symbol type occupes least 4 bits. */
#define ELF64_SYM_BINDATTR(x) (((x) >> 4) & 0xf)

/* ELF symbol bindings. */
#define STB_LOCAL  0 /* Not visible outside the object file */
#define STB_GLOBAL 1 /* Global symbol, visible to all object files */
#define STB_WEAK   2 /* Global scope, but with lower precedence than global symbols */

#define ELF64_R_SYM(i)((long long)(i) >> 32)
#define ELF64_R_TYPE(i)((i) & 0xffffffffLL)
#define ELF64_R_INFO(s, t)(((s) << 32) + ((t) & 0xffffffffLL))

/* Symbol types */
#define STT_NOTYPE 0 /* No type specified (e.g., an absolute symbol) */
#define STT_OBJECT 1 /* Data object */
#define STT_FUNC   2 /* Function entry point */

/* Types of relocations. */
#define R_X86_64_NONE      0       /* No relocation. */
#define R_X86_64_64        1       /* Add 64 bit symbol value. */
#define R_X86_64_PC32      2       /* PC-relative 32 bit signed sym value. */
#define R_X86_64_GOT32     3       /* PC-relative 32 bit GOT offset. */
#define R_X86_64_PLT32     4       /* PC-relative 32 bit PLT offset. */
#define R_X86_64_COPY      5       /* Copy data from shared object. */
#define R_X86_64_GLOB_DAT  6       /* Set GOT entry to data address. */
#define R_X86_64_JUMP_SLOT 7       /* Set GOT entry to code address. */
#define R_X86_64_RELATIVE  8       /* Add load address of shared object. */
#define R_X86_64_GOTPCREL  9       /* Add 32 bit signed pcrel offset to GOT. */
#define R_X86_64_32        10      /* Add 32 bit zero extended symbol value */
#define R_X86_64_32S       11      /* Add 32 bit sign extended symbol value */
#define R_X86_64_16        12      /* Add 16 bit zero extended symbol value */
#define R_X86_64_PC16      13      /* Add 16 bit signed extended pc relative symbol value */
#define R_X86_64_8         14      /* Add 8 bit zero extended symbol value */
#define R_X86_64_PC8       15      /* Add 8 bit signed extended pc relative symbol value */
#define R_X86_64_DTPMOD64  16      /* ID of module containing symbol */
#define R_X86_64_DTPOFF64  17      /* Offset in TLS block */
#define R_X86_64_TPOFF64   18      /* Offset in static TLS block */
#define R_X86_64_TLSGD     19      /* PC relative offset to GD GOT entry */
#define R_X86_64_TLSLD     20      /* PC relative offset to LD GOT entry */
#define R_X86_64_DTPOFF32  21      /* Offset in TLS block */
#define R_X86_64_GOTTPOFF  22      /* PC relative offset to IE GOT entry */
#define R_X86_64_TPOFF32   23      /* Offset in static TLS block */

#endif /* __KERNEL_ELF_H__*/
