/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 */

#include <sys/types.h>
#include <fcntl.h>
#include <getopt.h>
#include <framework.h>
#include <fsio/bdio.h>

char *bdev_name = NULL;
const char *test_usage="Test usage: t_blkbe -n [bdev name]";

static int test_init(void)
{
  int opt;
  int ret = 0;

  while (!ret) {
    opt = getopt(cmdline_args_num, cmdline_args, "n:");
    if (opt == -1)
      break;

    switch (opt) {
    case 'n':
      bdev_name = malloc(strlen(optarg) + 1);
      if (bdev_name == NULL)
        ret = -1;
      else
        strcpy(bdev_name, optarg);
      break;
    default:
      printf("Wrong arguments. %s\n", test_usage);
      ret = -1;
      break;
    }
  }

  if (ret)
    return ret;

  if (!bdev_name) {
    printf("Device name or capacity is not determined. %s\n",
           test_usage);
    return -1;
  }

  return 0;
}

bool __user_read_write(int dir, char *buf, size_t size, size_t off)
{
  int fd;
  int r;
  const char *str[2] = {"read", "write"};
  char msg[64];

  sprintf(msg, "%s error", str[dir]);
  fd = open(bdev_name, O_RDWR, 0);
  if (!check_test_condition_ext(fd, 0, COMP_GREQ, 0, "failed to open block device", false))
    return false;

  lseek(fd, off, SEEK_SET);
  if (dir) {
    for (r = 0; r < size; r++)
      buf[r] = r;

    r = write(fd, buf, size);
  } else {
    r = read(fd, buf, size);
  }
  r = check_test_condition_ext(r, size, COMP_EQUAL, 0, msg, false);
  close(fd);

  return r;
}

bool __fs_read_write(int dir, char *buf, uint32_t block,
                     uint32_t nr_blocks, uint16_t blk_bits)
{
  struct bdio bdio;
  int r;

  r = open_blkdev(bdev_name, &bdio, false);
  if (!check_test_condition_ext(r, 0, COMP_EQUAL, 0,
                                "file system like device opening fail", false)) {
    return false;
  }

  bdio.blk_bits = blk_bits;     /* 1024 bytes per block */
  r = read_blkdev_blocks(&bdio, block, nr_blocks, buf);
  r  = check_test_condition_ext(r, nr_blocks, COMP_EQUAL, 0,
                                "file system like read error", false);
  close_blkdev(&bdio);

  return r;
}

void TC_read_stream(long *args)
{
  char buf[4096], buf1[4096];

  if (!__user_read_write(1, buf, 4096, 8192))
    return;

  if (__fs_read_write(0, buf1, 8, 4, 10)) {
    if (!memcmp(buf1, buf, 4096))
      test_pass();
    else
      test_fail("Data don't tally\n");
  }
}

void TC_write_stream(long *args)
{
  char buf[4096], buf1[4096];

  if (!__fs_read_write(1, buf1, 16, 4, 10))
    return;

  if (__user_read_write(0, buf, 4096, 16384)) {
    if (!memcmp(buf1, buf, 4096))
      test_pass();
    else
      test_fail("Data don't tally\n");
  }
}

test_struct_t test_struct = {
  .init = test_init,
  .pause_seconds = 1,
  .cases = {
    {
      .name = "read_stream",
      .function = TC_read_stream,
      .description = "Test file system like stream read\n"
    },
    {
      .name = "write_stream",
      .function = TC_write_stream,
      .description = "Test file system like write\n"
    },

    CASE_LIST_END
  }
};
