#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <pthread.h>
#include <sys/scheduler.h>
#include <sys/proc.h>
#include <sys/thread.h>
#include <sys/ptd.h>
#include <sys/keyboard.h>
#include <datalib/circular.h>
#include <termios.h>

#define TBUFLEN 32768

#define TERM_SLOT 6
#define BUFF_SLOT 5

#define STARTULEN 7
#define MAXULEN 200
#define WIDTH 80
#define HEIGHT 24

#define START_DELAY 600000;

struct termios old_termios;
uint8_t old_kbdmode;

int delay;
int ulen;
bool paused;

circular_t keypresses;

int xs[MAXULEN];
int ys[MAXULEN];

int field[WIDTH][HEIGHT];

pthread_mutex_t umutex;

int vx, vy, lvx, lvy;

static void printxy(int x, int y, char c)
{
  printf("\e[%d;%dH%c", y+1, x+1, c);
  fflush(stdout);
}

void clrscr(void)
{
  printf("\e[2J");
  fflush(stdout);
}

int set_raw(void)
{
  int r;
  struct termios t;
  uint8_t kbmode = 0;

  r = tcgetattr(STDOUT_FILENO, &t);

  if(r < 0) {
    fprintf(stderr, "cannot get term attrs %d %d\n", r, errno);
    return -1;
  }

  memcpy(&old_termios, &t, sizeof(t));

  t.c_lflag &= ~(ICANON | ECHO);

  r = tcsetattr(STDOUT_FILENO, TCSANOW, &t);

  if(r < 0) {
    fprintf(stderr, "cannot set term attrs %d %d\n", r, errno);
    return -1;
  }

  r = ioctl(STDIN_FILENO, TIOC_GETKBMODE, &kbmode);

  if(r < 0) {
    fprintf(stderr, "cannot set kbmode %d %d\n", r, errno);
    return -1;
  }

  old_kbdmode = kbmode;

  kbmode = 0;
  
  r = ioctl(STDIN_FILENO, TIOC_SETKBMODE, &kbmode);

  if(r < 0) {
    fprintf(stderr, "cannot set kbmode %d %d\n", r, errno);
    return -1;
  }

  return 0;
}

char get_char(void)
{
  char c;
  
  read(STDIN_FILENO, &c, 1);
  
  return c;
}

int is_killed(int nx, int ny)
{
  int i;
  if(nx >= WIDTH-1) return 1;
  if(nx < 1) return 1;
  if(ny >= HEIGHT-1) return 1;
  if(ny < 1) return 1;
  for(i = 0; i < ulen; i++) {
    if(xs[i] == nx && ys[i] == ny) return 1;
  }
  return 0;
}

uint16_t random(uint16_t y)
{
  static ulong_t a = 3, b = 11, c = 17;
  uint16_t x = (1 + a + 23 * b + 17 * c);
  
  a = b;
  b = c;
  c = x;
  
  return x % y;
}

void put_food(void)
{
  int i;
  int flag;
  uint16_t fx, fy;
  while(1) {
    fx = 1 + random(WIDTH-2);
    fy = 1 + random(HEIGHT-2);
    flag = 0;
    for(i = 0; i < ulen; i++) {
      if(fx == xs[i] || fy == ys[i] || field[fx][fy] != 0) flag = 1;
    }
    if(!flag) break;
  }
  printxy(fx, fy, 'r');
  field[fx][fy] = 1;
}
                                                
void paint_border(void)
{
  char tmp[WIDTH+1];
  int i;

  memset(tmp, '-', sizeof(tmp));
  tmp[0] = '+';
  tmp[WIDTH-1] = '+';
  tmp[WIDTH] = 0;
    
  printf("\e[%d;1H%s\e[1;1H%s", HEIGHT, tmp, tmp);
  for(i = 2; i < HEIGHT; i++) {
    printf("\e[%d;1H|\e[%d;%dH|", i, i, WIDTH);
  }
  fflush(stdout);

}

void print_status(void)
{
  char str[WIDTH+1];

  //move cursor
  printf("\e[%d;1H", HEIGHT + 1);

  memset(str, ' ', WIDTH);

  if(paused) {
    snprintf(str, WIDTH, "Game is paused, press \"SPACE\" to resume or \"x\" to quit");
  } else {
    snprintf(str, WIDTH, "       Length: %d", HEIGHT + 1, ulen);
  }

  str[strlen(str)] = ' ';
  str[WIDTH-1] = 0;
  
  printf(str);
  
  fflush(stdout);
}

void udav_init(void)
{
  int i;

  clrscr();
  paint_border();

  delay = START_DELAY;

  circular_flush(&keypresses);
  
  vx = 1;
  vy = 0;

  paused = true;

  ulen = STARTULEN;

  memset(field, 0, sizeof(field));
  print_status();

  for(i = 0; i < STARTULEN; i++) {
    xs[i] = STARTULEN-i+4;
    ys[i] = HEIGHT/2;
    printxy(xs[i], ys[i], 'o');
  }

  put_food();
  put_food();
  put_food();
  put_food();
  put_food();
  put_food();
  put_food();
}

static void udav_exit(void)
{
  tcsetattr(STDOUT_FILENO, TCSANOW, &old_termios);
  ioctl(STDIN_FILENO, TIOC_SETKBMODE, &old_kbdmode);
  
  clrscr();
  printf("\nThanks for playing udav.\n\n");
  exit(0);
}


void udav_move(void)
{
  int nx, ny, i;
  int enl = 0;
  char code;
  int got = 0;

  print_status();
  
  while(got == 0 && circular_getc(&keypresses, &code)){
    if(paused) {
      switch(code) {
        case KC_X:
          udav_exit();
          break;
        case KC_SPACE:
          paused = 0;
          break;
        default:
          //ignore
          break;
      }
    } else {
      switch(code) {
        case KC_LEFT:
        if(vx != 1) {
          vx = -1;
          vy = 0;
          got = 1;
        }
        break;
        case KC_DOWN:
          if(vy != -1) {
            vy = 1;
            vx = 0;
            got = 1;
          }
          break;
        case KC_RIGHT:
          if(vx != -1) {
            vx = 1;
            vy = 0;
            got = 1;
          }
          break;
        case KC_UP:
          if(vy != 1) {
            vy = -1;
            vx = 0;
            got = 1;
          }
          break;
        case KC_SPACE:
        case KC_ESC:
          paused = 1;
          print_status();
          got = 1;
          break;
      }
    }
  }

  if(paused) return;
  
  nx = xs[0] + vx;
  ny = ys[0] + vy;

  if(is_killed(nx, ny)){
    udav_init();
    return;
  }

  pthread_mutex_lock(&umutex);

  lvx = vx;
  lvy = vy;
  
  if(field[nx][ny]) {
    field[nx][ny] = 0;
    put_food();
    xs[ulen] = xs[ulen-1];
    ys[ulen] = ys[ulen-1];
    enl = 1;
    delay -= delay/50;
  }

  printxy(nx, ny, 'o');
  if(!enl)printxy(xs[ulen-1], ys[ulen-1], ' ');
  
  for(i = ulen-1; i > 0; i--) {
    xs[i] = xs[i-1];
    ys[i] = ys[i-1];
  }

  if(enl) ulen++;

  xs[0] = nx;
  ys[0] = ny;
  pthread_mutex_unlock(&umutex);
}

void *control(void *arg)
{
  char c;

  while(1) {
    c = get_char();
    circular_putc(&keypresses, c);
  }
}

int main(void)
{
  pthread_t ct;

  if(!isatty(STDIN_FILENO) || !isatty(STDOUT_FILENO)) {
    fprintf(stderr, "Udav only works on terminals, exiting.\n");
    exit(1);
  }
  
  pthread_mutex_init(&umutex, NULL);

  if(circular_init(&keypresses, 64) != 0) {
    fprintf(stderr, "Failed to allocate keyboard buffer.\n");
    exit(1);
  };

  printf("\e[?25l");
  fflush(stdout);

  set_raw();
  
  udav_init();

  if(pthread_create(&ct, NULL, control, NULL) != 0) {
    fprintf(stderr, "Failed to create thread.\n");
    exit(1);
  }
  
  while(1) {
    usleep(delay);
    udav_move();
  }

  return 0;
}
