/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Copyright (C) 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 */

%{
#include <stdbool.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#define __INSIDE_PARSER 1
#include "symbol.h"

#define YYERROR_VERBOSE 1

int yylex(void);
void yyerror(const char *msg);
void yyset_in(FILE *fd);

static const char *file;
void error(int status, const char *fmt, ...);
void warning(const char *fmt, ...);

static bool inside_switch = false;
%}

%union {
  char *string;                 /* string buffer */
  int   cmd;                    /* command id */
}

%token <string>T_QSTRING
%token <string>T_ID

%token <cmd>T_MENU
%token <cmd>T_BOOL
%token <cmd>T_STRING
%token <cmd>T_DESC
%token <cmd>T_SWITCH
%token <cmd>T_CASE

%type <cmd>menu
%type <cmd>simple
%type <cmd>exp_switch
%type <cmd>exp_boolean
%type <string>id
%type <string>qstring

%%
input:          stmt_list
     ;

stmt_list:      /* empty */
         |      stmt_list menu
         |      stmt_list simple
         |      stmt_list exp_switch
	 |      stmt_list exp_boolean
         ;

menu:           menu_type qstring '{' stmt_list '}' {
                  set_menu_name($2);
                }
    ;

menu_type:      T_MENU {
                  add_symbol(T_MENU);
                }
         ;

simple:         var_type id '{' var_body '}' {
                  set_simple_id($2);
                }
      ;

var_type:       T_STRING {
                  if (inside_switch)
                    yyerror("syntax error");
                  add_symbol(T_STRING);
                }
        |       T_CASE {
                  if (!inside_switch)
                    yyerror("syntax error");
                  add_symbol(T_CASE);
                }
        ;

var_body:       /* empty */
        |       var_body var_desc
        ;

var_desc:       T_DESC qstring {
                  set_simple_desc($2);
                }
        ;

exp_switch:     switch_type id '{' switch_body '}' {
                  set_exp_id($2);
                  inside_switch = false;
                }
          ;

switch_type:    T_SWITCH {
                  add_symbol(T_SWITCH);
                  inside_switch = true;
                }
           ;

switch_body:       /* empty */
	   |       switch_body exp_desc
           |       switch_body simple
           ;

exp_desc:       T_DESC qstring {
                  set_exp_desc($2);
                }
        ;

exp_boolean:    bool_type id '{' bool_body '}' {
                  set_exp_id($2);
		}
           ;

bool_type:      T_BOOL {
	          add_symbol(T_BOOL);
		}
	;

bool_body:      /* empty */
	 |      bool_body exp_desc
         |      bool_body menu
         |      bool_body simple
         |      bool_body exp_switch
	 |      bool_body exp_boolean
	 ;

id:             T_ID { $$ = $1; }
  |             T_QSTRING {
                  warning("string literal inappropriate");
                  $$ = $1;      /* but use it anyway */
                }
  ;

qstring:        T_QSTRING { $$ = $1; }
       |        T_ID {
                  warning("non-string literal inappropriate");
                  $$ = $1;      /* but use it anyway */
                }
       ;

%%

int lineno = 1;

void parse_config(const char *filename)
{
  FILE *fd;

  fd = fopen(filename, "r");
  if (fd == NULL) {
    perror("Can't open config file");
    exit(EXIT_FAILURE);
  }

  yyset_in(fd);
  file = filename;

  yyparse();
  if (yynerrs != 0)
    exit(EXIT_FAILURE);
}

void yyerror(const char *msg)
{
  error(EXIT_FAILURE, msg);
}

void error(int status, const char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "%s:%d: error: ", file, lineno);
  fprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);

  exit(status);
}

void warning(const char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "%s:%d: warning: ", file, lineno);
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
}
