/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Based on kconfig from Linux kernel:
 *   Copyright (C) 2002 Roman Zippel <zippel@linux-m68k.org>
 *   Released under the terms of the GNU GPL v2.0.
 *
 * Copyright (C) 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 */

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "symbol.h"

static struct symbol *root = NULL;

static struct symbol *cur = NULL;
static enum
{
  SIBLING,
  CHILD
} next_is = SIBLING;

struct symbol *get_root_symbol(void)
{
  return root;
}

void add_symbol(int type)
{
  struct symbol *e;

  e = calloc(1, sizeof(struct symbol));
  e->type = type;
  e->prev = e->next = NULL;

  switch (type) {
  case T_MENU:
    e->t.menu.name = NULL;
    e->t.menu.list = NULL;
    break;
  case T_STRING:
  case T_CASE:
    e->t.simple.id = NULL;
    e->t.simple.desc = NULL;
    break;
  case T_SWITCH:
  case T_BOOL:
    e->t.exp.id = NULL;
    e->t.exp.desc = NULL;
    e->t.exp.list = NULL;
    break;
  }

  if (root == NULL) {
    e->parent = NULL;
    root = e;
  }
  else {
    if (next_is == SIBLING) {
      e->parent = cur->parent;
      cur->next = e;
      e->prev = cur;
    }
    else {
      e->parent = cur;
      switch (cur->type) {
      case T_MENU:
        cur->t.menu.list = e;
	break;
      case T_SWITCH:
      case T_BOOL:
        cur->t.exp.list = e;
	break;
      }
    }
  }

  cur = e;
  switch (type) {
  case T_MENU:
  case T_SWITCH:
  case T_BOOL:
    next_is = CHILD;
    break;
  default:
    next_is = SIBLING;
  }
}

bool set_menu_name(char *name)
{
  if (next_is == SIBLING)
    cur = cur->parent;

  cur->t.menu.name = name;
  next_is = SIBLING;
  return true;
}

bool set_simple_id(char *id)
{
  cur->t.simple.id = id;
  switch (cur->type) {
  case T_STRING:
    cur->v.str = NULL;
    break;
  }
  next_is = SIBLING;
  return true;
}

bool set_simple_desc(char *desc)
{
  if (cur->t.simple.desc != NULL)
    free(cur->t.simple.desc);
  cur->t.simple.desc = desc;
  return true;
}

bool set_exp_id(char *id)
{
  if (next_is == SIBLING)
    cur = cur->parent;

  cur->t.exp.id = id;
  if (cur->type == T_SWITCH)
      cur->v.choice = cur->t.exp.list;
  next_is = SIBLING;
  return true;
}

bool set_exp_desc(char *desc)
{
  struct symbol *exp;

  exp = cur;
  if (next_is == SIBLING)
    exp = cur->parent;

  if (exp->t.exp.desc != NULL)
    free(exp->t.exp.desc);
  exp->t.exp.desc = desc;
  return true;
}

bool free_symbol(struct symbol * sym)
{
  if (sym == NULL)
    return true;

  switch (sym->type) {
  case T_MENU:
    free(sym->t.menu.name);
    break;
  case T_STRING:
    free(sym->t.simple.id);
    free(sym->t.simple.desc);
    break;
  case T_SWITCH:
  case T_BOOL:
    free(sym->t.exp.id);
    free(sym->t.exp.desc);
  default:
    return false;
  }

  return true;
}

bool is_empty_symbol(const struct symbol * sym)
{
  bool rc;
  const struct symbol *i;

  if (sym == NULL)
    return true;

  switch (sym->type) {
  case T_STRING:
  case T_CASE:
  case T_BOOL:
    rc = false;
    break;
  case T_MENU:
    rc = true;
    for (i = sym->t.menu.list; i != NULL; i = i->next)
      rc = rc && is_empty_symbol(i);
    break;
  case T_SWITCH:
    rc = (sym->t.exp.list == NULL);
    break;
  }

  return rc;
}

struct symbol *find_symbol(struct symbol *first, const char *id)
{
  struct symbol *rc;
  struct symbol *i;

  rc = NULL;
  for (i = first; i != NULL; i = i->next) {
    switch (i->type) {
    case T_MENU:
      rc = find_symbol(i->t.menu.list, id);
      break;
    case T_STRING:
    case T_CASE:
      if (strcmp(i->t.simple.id, id) == 0)
        rc = i;
      break;
    case T_SWITCH:
    case T_BOOL:
      if (strcmp(i->t.exp.id, id) == 0)
        rc = i;
      else
        rc = find_symbol(i->t.exp.list, id);
      break;
    }
    if (rc != NULL)
      break;
  }

  return rc;
}
