/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 *
 * Jari OS default shell.
 *
 */

#define _POSIX_C_SOURCE 200112L

#include <limits.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

struct builtin_cmd {
  char *cmd;
  char *help;
  int (*handler)(int argc, char *argv[]);
};

static void set_environment(const char *filename);
static const char* get_prompt(void);
static size_t split_line(const char *line, char ***args);
static void free_args(char *args[]);

static int exit_issued = 0;

static int __cd(int argc, char *argv[]);
static int __exit(int argc, char *argv[]);
static int __help(int argc, char *argv[]);
static int __pwd(int argc, char *argv[]);
static int __set(int argc, char *argv[]);
static int __unset(int argc, char *argv[]);

/* Please, add new commands in alphabetic order */
static struct builtin_cmd builtins[] = {
  { "cd",       "Change the shell working directory",           __cd },
  { "exit",     "Exit the shell",                               __exit },
  { "help",     "Display information about builtin commands",   __help },
  { "logout",   "Exit a login shell",                           __exit },
  { "pwd",      "Print the name of the current working directory", __pwd },
  { "set",      "Set values of environment variables",          __set },
  { "unset",    "Unset environment variables",                  __unset },

  { NULL,       NULL,                                           NULL }
};

int main(int argc, char *argv[])
{
  char line[LINE_MAX];
  size_t nr_args, i;
  char **args, **argsv;
  pid_t pid;
  int status, rc;

  (void)argc;
  (void)argv;

  setenv("SHELL", argv[0], 1);
  set_environment("/etc/josh/env");

  for (;;) {
    printf("%s", get_prompt());
    fflush(stdout);

    memset(line, 0, sizeof(line));
    args = NULL;

    if (fgets(line, sizeof(line), stdin) == NULL) {
      if (ferror(stdin))
        fprintf(stderr, "josh: fgets() failed: %d\n", errno);
      continue;
    }
    if (line[strlen(line) - 1] == '\n')
      line[strlen(line) - 1] = '\0';

    nr_args = split_line(line, &args);
    if (nr_args == 0)
      goto end_loop;

    /* Is it built-in command ? */
    for (i = 0; builtins[i].cmd != NULL; i++)
      if (strcmp(builtins[i].cmd, args[0]) == 0)
        break;
    if (builtins[i].handler != NULL) {
      errno = 0;
      rc = builtins[i].handler(nr_args, args);
      if (rc == EXIT_FAILURE)
        fprintf(stderr, "josh: %s failed: %d\n", args[0], errno);
      goto end_loop;
    }

    errno = 0;
    pid = fork();
    if (pid == -1) {
      fprintf(stderr, "josh: fork() failed: %d\n", errno);
      goto end_loop;
    }
    if (pid == 0) {
      argsv=args; argsv++; /* TODO: fix it in procmgr */
      execvp(args[0], argsv);
      fprintf(stderr, "josh: execvp() failed: %d\n", errno);
      exit(EXIT_FAILURE);
    }
    if (waitpid(pid, &status, 0) == (pid_t)-1)
      fprintf(stderr, "josh: waitpid() failed: %d\n", errno);
    //rc = WEXITSTATUS(status);
    rc = status;

end_loop:
    free_args(args);

    if (exit_issued)
      break;
  }

  return rc;
}

static void set_environment(const char *filename)
{
  FILE *fd;
  char line[LINE_MAX];

  fd = fopen(filename, "r");
  if (fd == NULL)
    return;

  memset(line, 0, sizeof(line));
  while (fgets(line, sizeof(line), fd) != NULL) {
    if (line[strlen(line) - 1] == '\n')
      line[strlen(line) - 1] = '\0';
    putenv(strdup(line));
    memset(line, 0, sizeof(line));
  }

  fclose(fd);
}

static const char* get_prompt(void)
{
  static char prompt[PATH_MAX];

  memset(prompt, 0, PATH_MAX);

  if (getenv("USER") != NULL) {
    strcpy(prompt, "\e[34m");
    strcpy(prompt, getenv("USER"));
    strcat(prompt, "@");
  }

  strcat(prompt, "\e[31m");
  if (getcwd(prompt + strlen(prompt), PATH_MAX - strlen(prompt) - 1) != NULL)
    strcat(prompt, ":");

  strcat(prompt, ">\e[0m ");
  return prompt;
}

static size_t split_line(const char *line, char ***args)
{
  size_t argc;
  size_t i, j;
  char **p;
  char *str;

  if ((line == NULL) || (args == NULL))
    return 0;

  argc = 0;
  *args = malloc(sizeof(char**) * (argc + 1));
  if (*args == NULL)
    return 0;
  (*args)[argc] = NULL;

  for (i = 0; i < strlen(line); i++) {
    if (isspace(line[i]))
      continue;

    for (j = i; j < strlen(line); j++) {
      if (isspace(line[j]))
        break;
    }

    str = malloc(j - i + 1);
    if (str == NULL)
      break;
    memcpy(str, line + i, j - i);
    str[j - i] = '\0';

    p = realloc(*args, sizeof(char**) * (argc + 2));
    if (p == NULL)
      break;
    *args = p;

    (*args)[argc] = str;
    argc++;
    (*args)[argc] = NULL;

    i = j;
  }

  return argc;
}

static void free_args(char *args[])
{
  size_t i;

  for (i = 0; args[i] != NULL; i++)
    free(args[i]);
  free(args);
}

static int __cd(int argc, char *argv[])
{
  int rc;

  if (argc > 1)
    rc = chdir(argv[1]);
  else
    rc = chdir(getenv("HOME"));

  return (rc == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

static int __exit(int argc, char *argv[])
{
  exit_issued = 1;

  if (argc == 1)
    return EXIT_SUCCESS;
  return atoi(argv[1]);
}

static int __help(int argc, char *argv[])
{
  size_t i;

  for (i = 0; builtins[i].cmd != NULL; i++) {
    if (argc == 1)
      printf("%s\n    %s\n\n", builtins[i].cmd, builtins[i].help);
    else {
      if (strcmp(builtins[i].cmd, argv[1]) == 0) {
         printf("%s\n    %s\n\n", builtins[i].cmd, builtins[i].help);
	 return EXIT_SUCCESS;
      }
    }
  }

  if (argc != 1)
    fprintf(stderr, "josh: help: no help topics match '%s'\n", argv[2]);

  return EXIT_SUCCESS;
}

static int __pwd(int argc, char *argv[])
{
  char *p;

  (void)argc;
  (void)argv;

  p = malloc(PATH_MAX);
  if (p == NULL)
    return EXIT_FAILURE;
  if (getcwd(p, PATH_MAX) == NULL) {
    free(p);
    return EXIT_FAILURE;
  }

  printf("%s\n", p);
  free(p);
  return EXIT_SUCCESS;
}

static int __set(int argc, char *argv[])
{
  size_t i;

  if (argc != 1) {
    if (setenv(argv[1], argv[2], 1) == -1)
      return EXIT_FAILURE;
    return EXIT_SUCCESS;
  }

  for (i = 0; environ[i] != NULL; i++)
    printf("%s\n", environ[i]);
  return EXIT_SUCCESS;
}

static int __unset(int argc, char *argv[])
{
  int i;

  for (i = 1; i < argc; i++)
    unsetenv(argv[i]);

  return EXIT_SUCCESS;
}
