/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dan Kruchinin <dk@jarios.org>
 *
 * tests/mm/grant_pages_test.c - General grant_pages test.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <test/framework.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <ipc/channel.h>

#define CHECK_MMAP_ADDR(addr, fn)                               \
  do {                                                          \
    (addr) = (fn);                                              \
    if ((addr) == MAP_FAILED) {                                 \
      printf("Failed to map (%s): [%d]\n", #fn, errno); \
      abort();                                                  \
    }                                                           \
  } while (0)

#define CHECK_ERR(fn, errok)                    \
  do {                                          \
    int __ret;                                  \
    bool __cond;                                \
                                                \
    __ret = (fn);                                                       \
    __cond = __ret < 0;                                                 \
    printf("[%s] returned %d\n", #fn, __ret);                   \
    if ((__cond && !(errok)) || (!__cond && (errok))) {                 \
      abort();                                                          \
    }                                                                   \
  } while (0)

#define PATTERN_ONE   0xABCDE
#define PATTERN_TWO   0xEDCBA
#define PATTERN_THREE 0xFF0FF
#define PATTERN_FOUR  0xDEAD

struct repl {
  uintptr_t start_addr;
  size_t size;
  int pattern;
  int cmd;
};

static inline void __fill_with_pattern(void *base, int pages, long pattern)
{
  int i;
  
  for (i = 0; i < pages; i++) {
    *(long *)((char *)base + i * PAGE_SIZE) = pattern;
  }
}

static inline int __check_pattern(void *base, int pages, long pattern)
{
  int i;

  for (i = 0; i < pages; i++) {
    if (*(long *)((char *)base + i * PAGE_SIZE) != pattern)
      return -1;
  }

  return 0;
}

void TC_main_test(long *unused)
{
  int chpid;
  void *srcaddr, *dstaddr;
  prtid_t port;

  port = ipc_port_open(IPC_BLOCKED_ACCESS, 1);
  if (port < 0) {
    printf("Can not open port!\n");
    abort();
  }

  printf("chpid = %d\n", chpid);
  chpid = fork();
  if (chpid < 0) {
    printf("Fork failed! [%d]\n", chpid);
    abort();
  }
  else if (chpid > 0) {
    int i, r;
    ipc_port_msg_info_t msg_info;
    struct repl repl;
    int cmd = 0;
    void *srcaddr2;
    
    printf("Child pid: %d\n", chpid);
    CHECK_MMAP_ADDR(srcaddr, mmap(NULL, PAGE_SIZE * 4, PROT_READ | PROT_WRITE,
                                  MAP_PRIVATE | MAP_ANON | MAP_POPULATE, 0, 0));
    for (i = 0; i < 5; i++) {
      r = __syscall5(SC_PORT_RECEIVE, port, IPC_BLOCKED_ACCESS, (uintptr_t)&cmd,
                     sizeof(int), (uintptr_t)&msg_info);
      if (r < 0) {
        printf("[%d] Failed to receive: %d\n", i, errno);
        abort();
      }
      if (cmd == 1) {        
        CHECK_MMAP_ADDR(dstaddr, mmap_pid(chpid, NULL, PAGE_SIZE * 4, PROT_READ | PROT_WRITE,
                                          MAP_PRIVATE | MAP_ANON | MAP_CANRECPAGES, 0, 0));
        
        __fill_with_pattern(srcaddr, 4, PATTERN_ONE);
        printf("Grant 4 pages(va_from = %p) to %d\n", srcaddr, chpid);
        CHECK_ERR(grant_pages(srcaddr, PAGE_SIZE * 4, chpid, (uintptr_t)dstaddr), false);
        CHECK_ERR(grant_pages(srcaddr, PAGE_SIZE * 4, chpid, (uintptr_t)dstaddr), true);
        *(int *)srcaddr = 1;
        CHECK_ERR(grant_pages(srcaddr, PAGE_SIZE * 4, chpid, (uintptr_t)dstaddr), true);
        
        repl.start_addr = (uintptr_t)dstaddr;
        repl.size = PAGE_SIZE * 4;
        repl.pattern = PATTERN_ONE;
        repl.cmd = 1;
      }
      else if (cmd == 2) {
        CHECK_MMAP_ADDR(srcaddr2, mmap((void *)0x60000000UL, PAGE_SIZE * 8, PROT_READ | PROT_WRITE,
                                       MAP_PRIVATE | MAP_ANON, 0, 0));
        __fill_with_pattern(srcaddr, 4, PATTERN_TWO);
        __fill_with_pattern(srcaddr2, 8, PATTERN_TWO);

        CHECK_MMAP_ADDR(dstaddr, mmap_pid(chpid, NULL, PAGE_SIZE * 12, PROT_READ | PROT_WRITE,
                                          MAP_PRIVATE | MAP_ANON | MAP_CANRECPAGES, 0, 0));
        CHECK_ERR(grant_pages(srcaddr, PAGE_SIZE * 4, chpid, (uintptr_t)dstaddr), false);
        CHECK_ERR(grant_pages(srcaddr2, PAGE_SIZE * 8, chpid, (uintptr_t)dstaddr + PAGE_SIZE * 4), false);
        *(int *)srcaddr = 1;
        *(int *)srcaddr2 = 1;
        CHECK_ERR(grant_pages(srcaddr, PAGE_SIZE * 4, chpid, (uintptr_t)dstaddr), true);
        CHECK_ERR(grant_pages(srcaddr2, PAGE_SIZE * 8, chpid, (uintptr_t)dstaddr), true);

        repl.start_addr = (uintptr_t)dstaddr;
        repl.size = PAGE_SIZE * 12;
        repl.pattern = PATTERN_TWO;
        repl.cmd = 2;
      }
      else if (cmd == 3) {
        __fill_with_pattern(srcaddr, 4, PATTERN_THREE);
        __fill_with_pattern(srcaddr2, 6, PATTERN_THREE);
        CHECK_MMAP_ADDR(dstaddr, mmap_pid(chpid, NULL, PAGE_SIZE * 12, PROT_READ,
                                          MAP_PRIVATE | MAP_ANON | MAP_CANRECPAGES, 0, 0));
        CHECK_ERR(grant_pages(srcaddr, PAGE_SIZE * 4, chpid, (uintptr_t)dstaddr), false);
        CHECK_ERR(grant_pages(srcaddr2, PAGE_SIZE * 6, chpid, (uintptr_t)dstaddr + PAGE_SIZE * 4), false);

        repl.start_addr = (uintptr_t)dstaddr;
        repl.size = PAGE_SIZE * 10;
        repl.pattern = PATTERN_THREE;
        repl.cmd = 3;
      }
      else if (cmd == 4) {
        __fill_with_pattern(srcaddr, 2, PATTERN_FOUR);
        CHECK_ERR(grant_pages(srcaddr, PAGE_SIZE * 2, chpid, (uintptr_t)dstaddr + PAGE_SIZE * 10), false);
        *(int *)srcaddr = 1;
        CHECK_ERR(grant_pages(srcaddr, PAGE_SIZE * 2, chpid, (uintptr_t)dstaddr + PAGE_SIZE * 10), true);

        repl.start_addr = (uintptr_t)dstaddr + PAGE_SIZE * 10;
        repl.size = PAGE_SIZE * 2;
        repl.pattern = PATTERN_FOUR;
        repl.cmd = 4;
      }
      else if (cmd == 5) {
        repl.cmd = 0;
      }

      r = __syscall4(SC_PORT_REPLY, port, msg_info.msg_id, (uintptr_t)&repl, sizeof(struct repl));
      if (r < 0) {
        printf("Can not reply to %d! [ret = %d]\n", chpid, r);
        abort();
      }
    }

    __syscall1(SC_EXIT, 0);
  }
  else {
    pid_t parent = getppid();
    chid_t channel;
    int ret, snd;
    struct repl rpl;
    
    sleep(1);
    channel = __syscall3(SC_OPEN_CHANNEL, parent, port, IPC_BLOCKED_ACCESS);
    if (channel < 0) {
      printf("Can not open channel!");
      abort();
    }

    snd = 1;
    rpl.cmd = -1;
    while (rpl.cmd) {
      ret = __syscall5(SC_PORT_SEND, channel, (uintptr_t)&snd, sizeof(int),
                       (uintptr_t)&rpl, sizeof(struct repl));
      if (ret < 0) {
        printf("Failed to send a message!\n");
        abort();
      }
      if (rpl.cmd > 0) {
        if (__check_pattern((void *)rpl.start_addr,
                            PAGE_ALIGN(rpl.size) >> PAGE_WIDTH, rpl.pattern) < 0) {
          abort();
        }

        CHECK_ERR(munmap((void *)rpl.start_addr, rpl.size), false);
        snd++;
      }
    }

    __syscall1(SC_EXIT, 0);
  }
}

void TC_check_errors(long *args)
{
  void *addr;
  int chpid;
  prtid_t port;
  
  printf("1) Give kernel an address I haven't ever mapped\n");
  CHECK_MMAP_ADDR(addr, mmap((void *)0x101f000, PAGE_SIZE, PROT_READ, MAP_ANON | MAP_PRIVATE | MAP_POPULATE, 0, 0));
  CHECK_ERR(grant_pages((void *)0x60000000UL, PAGE_SIZE, 7, 0x55555), true);

  printf("2) Give to kernel invalid pid\n");
  CHECK_ERR(grant_pages((void *)PAGE_ALIGN((uintptr_t)&addr), PAGE_SIZE, 123, 0x55555), true);

  printf("3) Give to kernel an address I don't have permissions write to\n");
  CHECK_ERR(grant_pages((void *)0x666, PAGE_SIZE, 7, 0x100000), true);

  printf("4) Give to kernel valid address of another process and valid mapping I already have\n");
  CHECK_MMAP_ADDR(addr, mmap(NULL, PAGE_SIZE * 4, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANON, 0, 0));
  CHECK_ERR(grant_pages(addr, PAGE_SIZE, 7, (uintptr_t)&addr), true);
  
  printf("4) Let's create new process and play with it...\n");
  port = ipc_port_open(IPC_BLOCKED_ACCESS, 1);
  if (port < 0) {
    printf("Can not create port!\n");
    abort();
  }
  
  chpid = fork();
  if (chpid < 0) {
    printf("Can not create new process!\n");
    abort();
  }
  else if (chpid > 0) {
    int cmd = 0, r;
    void *remaddr;
    ipc_port_msg_info_t msg_info;
    
    printf("... Child pid: .. %d\n", chpid);    
    r = __syscall5(SC_PORT_RECEIVE, port, IPC_BLOCKED_ACCESS, (uintptr_t)&cmd,
                   sizeof(int), (uintptr_t)&msg_info);
    if (r < 0) {
      printf("Failed to receive: %d\n", errno);
      abort();
    }
    if (cmd != 1) {
      printf("Invalid cmd: %d\n", cmd);
      abort();
    }

    printf("Map to child %d VM range without MAP_GENERIC_FLAG\n");
    CHECK_MMAP_ADDR(remaddr, mmap_pid(chpid, NULL, PAGE_SIZE * 2, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE, 0, 0));
    CHECK_ERR(grant_pages(addr, PAGE_SIZE * 2, chpid, (uintptr_t)remaddr), true);
    CHECK_MMAP_ADDR(remaddr, mmap_pid(chpid, NULL, PAGE_SIZE * 2, PROT_READ, MAP_ANON | MAP_PRIVATE | MAP_CANRECPAGES, 0, 0));
    CHECK_ERR(grant_pages(addr, PAGE_SIZE * 2 + 1, chpid, (uintptr_t)remaddr), true);
    CHECK_ERR(grant_pages((void *)0x66666666UL, PAGE_SIZE * 2, chpid, (uintptr_t)remaddr), true);
    CHECK_ERR(grant_pages(addr, PAGE_SIZE * 2, chpid, (uintptr_t)remaddr), true);
    cmd = 0;
    r = __syscall4(SC_PORT_REPLY, port, msg_info.msg_id, (uintptr_t)&cmd, sizeof(int));
    if (r < 0) {
      printf("Can not reply to %d! [ret = %d]\n", chpid, r);
      abort();
    }

    printf("Munmapping...\n");
    CHECK_ERR(munmap(addr, PAGE_SIZE * 4), false);
    printf("dooooone\n");
  }
  else {
    pid_t parent_pid = getppid();
    chid_t channel;
    int ret, repl = -1, snd;

    sleep(1);
    channel = __syscall3(SC_OPEN_CHANNEL, parent_pid, port, IPC_BLOCKED_ACCESS);
    if (channel < 0) {
      printf("Can not open channel!");
      abort();
    }

    snd = 1;
    while (repl) {
      ret = __syscall5(SC_PORT_SEND, channel, (uintptr_t)&snd, sizeof(int),
                       (uintptr_t)&repl, sizeof(int));
      if (ret < 0) {
        printf("Failed to send a message!\n");
        abort();
      }
    }

    __syscall1(SC_EXIT, 0);
  }

  printf("Ok, child exited\n");
}

test_struct_t test_struct = {
  .pause_seconds = 1,
  .cases = {
    {
      .name = "grant_pages_retcodes",
      .function = TC_check_errors,
      .description = "Let's try to break grant_pages function.",
      .nolaunch_on_fails = false,
      .abort_on_fail = true,
      .arg_list = ARG_LIST { "no args", ARG_LIST_END },
      .args = ARGS_CHAIN {
        ARGS { 0 },
        ARGS_END,
      },
    },
    {
      "grant_pages_maintest", TC_main_test, "Main grant_pages test",
      false, true, ARG_LIST { "no args", ARG_LIST_END },
      ARGS_CHAIN { ARGS { 0 }, ARGS_END },
    },

    CASE_LIST_END,
  },
};
