/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Dan Kruchinin <dk@jarios.org>
 *
 * tests/mm/mmap_test.c - General mmap/munmap test.
 *
 */


#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <test/framework.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/mman.h>

#define DEFAULT_NUM_REGIONS 8

static jmp_buf env;

static void sigsegvhnd(int unused)
{
  longjmp(env, 1);
}

static void TC_munmap_chunks(long *unused)
{
  void *addr, *first_addr = NULL;
  int i, tmp;
  char buf[128];

  for (i = 0; i < DEFAULT_NUM_REGIONS; i++) {
    addr = mk_mmap(NULL, PAGE_SIZE * 2, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE, 0, 0);
    if (!first_addr)
      first_addr = addr;

    sprintf(buf, "failed to mk_mmap a region #%d of %d pages", i, PAGE_SIZE * (i + 1)); 
    if (!check_test_condition_ext((long)addr, (long)MAP_FAILED, COMP_NOT_EQU, 0, buf, true))
      return;

    memset(addr, 0, PAGE_SIZE * 2);
  }

  i = munmap((char *)first_addr + PAGE_SIZE * 2, PAGE_SIZE * 3);
  if (!check_test_condition_ext(i, 0, COMP_EQUAL, 0, NULL, true))
    return;
    
  signal(SIGSEGV, sigsegvhnd);
  if (!(i = setjmp(env))) {
    tmp = *(int *)((char *)first_addr + PAGE_SIZE * 2) = 0xDEAD;    
  } else {
    sprintf(buf, "I expected a fault on unmapped address %p, but it wasn't!\n", 
            ((char *)first_addr + PAGE_SIZE * 2)); 
    if (!check_test_condition_ext(i, 0, COMP_GREQU, 0, buf, true))
      return;
  }
  
  i = munmap((char *)first_addr, PAGE_SIZE * 2 * DEFAULT_NUM_REGIONS);
  if (!check_test_condition_ext(i, 0, COMP_EQUAL, 0, NULL, true))
    return;

  if (!(i = setjmp(env))) {
    tmp = *(int *)((char *)first_addr) = 0xDEAD;
  } else {
    sprintf(buf, "I expected a fault on unmapped address %p, but it wasn't!\n", first_addr);
    check_test_condition_ext(i, 0, COMP_GREQU, 0, buf, true);
  }
}

static void TC_munmap_milti(long *args)
{
  void *addr, *first_addr = NULL;
  int i, tmp;
  size_t total_sz = 0;
  int prot[2] = { PROT_READ, PROT_READ | PROT_WRITE };
  char buf[128];

  for (i = 0; i < DEFAULT_NUM_REGIONS; i++) {
    addr = mk_mmap(NULL, PAGE_SIZE * (i + 1), prot[i % 2], MAP_ANON | MAP_PRIVATE, 0, 0);
    if (!first_addr)
      first_addr = addr;

    total_sz += PAGE_SIZE * (i + 1);
    sprintf(buf, "Failed to mk_mmap a region #%d of %d pages\n", i, PAGE_SIZE * (i + 1));
    if (!check_test_condition_ext((long)addr, (long)MAP_FAILED, COMP_NOT_EQU, 0, buf, true))
      return;

    if (!(i % 2)) {
      tmp = *(int *)addr;
    } else {
      memset(addr, 1, PAGE_SIZE * (i + 1));
    }
  }
  if (!args[0]) {
    i = munmap(first_addr, total_sz);
    if (!check_test_condition_ext(i, 0, COMP_EQUAL, 0, NULL, true))
      return;
    
    signal(SIGSEGV, sigsegvhnd);
    if (!(i = setjmp(env))) {
      tmp = *(int *)((char *)first_addr + PAGE_SIZE * 2) = 0xDEAD;    
    } else {
      sprintf(buf, "I expected a fault on unmapped address %p, but it wasn't!\n",
              ((char *)first_addr + PAGE_SIZE * 2));
      if (!check_test_condition_ext(i, 0, COMP_GREQU, 0, buf, true))
        return;

      signal(SIGSEGV, SIG_DFL);
    }
  } else {
    i = munmap(first_addr, total_sz - PAGE_SIZE * 2);
    if (!check_test_condition_ext(i, 0, COMP_EQUAL, 0, NULL, true))
      return;
    
    signal(SIGSEGV, sigsegvhnd);
    if (!(i = setjmp(env))) {
      tmp = *(int *)((char *)first_addr + (total_sz - PAGE_SIZE * 2));
    } else {
      sprintf(buf, "Unexpected page-fault on address %p\n",
              (char *)first_addr + (total_sz - PAGE_SIZE * 2));
      if (!check_test_condition_ext(i, 0, COMP_LEQU, 0, buf, true))
        return;
    }
    i = munmap((char *)first_addr + (total_sz - PAGE_SIZE * 2), PAGE_SIZE * 2);
    if (!check_test_condition_ext(i, 0, COMP_EQUAL, 0, NULL, true))
      return;

    if (!(i = setjmp(env))) {
      tmp = *(int *)((char *)first_addr + (total_sz - PAGE_SIZE * 2));
    } else {
      sprintf(buf, "I expected a fault on unmapped address %p, but it wasn't!\n",
              ((char *)(total_sz - PAGE_SIZE * 2)));
      if (!check_test_condition_ext(i, 0, COMP_GREQU, 0, buf, true))
        return; 

      signal(SIGSEGV, SIG_DFL);
    }
  }
}

static void TC_map_munmap(long *args)
{
  void *addr;
  int i;
  char buf[128];

  addr = mk_mmap((void *)args[0], args[1], args[2], args[3], 0, 0);
  if (!check_test_condition_ext((long)addr, (long)MAP_FAILED, COMP_NOT_EQU, 0, NULL, true))
    return;

  memset(addr, 1, args[1]);
  for (i = 0; i < args[1]; i += PAGE_SIZE) {
    if (*((char *)addr + i) != 1) {
      test_fail("Mismatch: %p!\n", (char *)addr + i);
      return;
    }    
  }

  signal(SIGSEGV, sigsegvhnd);
  if (!(i = setjmp(env))) {
    *((char *)addr + args[1]) = 0;
  } else {
    sprintf(buf, "I expected a fault on %p address, but it wasn't :(\n",
            (char *)addr + i);
    if (!check_test_condition_ext(i, 0, COMP_GREQU, 0, buf, true))
        return;

    signal(SIGSEGV, SIG_DFL);
  }

  i = munmap(addr, args[1]);
  if (!check_test_condition_ext(i, 0, COMP_EQUAL, 0, NULL, true))
    return;

  signal(SIGSEGV, sigsegvhnd);
  if (!(i = setjmp(env))) {
    *(int *)addr = 0xDEAD;    
  } else {
    sprintf(buf, "I expected a failt on unmapped address %p, but it wasn't!\n", addr);
    if (!check_test_condition_ext(i, 0, COMP_GREQU, 0, buf, true))
      return; 

    signal(SIGSEGV, SIG_DFL);
  }
}

static void TC_map_retcodes(long *args)
{
  void *ret;
  int err;

  ret = mk_mmap((void *)args[0], args[1], (int)args[2], (int)args[3], (int)args[4], (off_t)args[5]);
  err = errno;
  printf("ret = %d, err = %d\n", ret, err);
  check_test_condition_ext((long)ret, (long)MAP_FAILED, COMP_EQUAL, args[6], NULL, true);
}

test_struct_t test_struct = {
  .pause_seconds = 1,
  .cases = {
    {
      .name = "mmap_retcodes",
      .function = TC_map_retcodes,
      .description = "Testing a set of errors returned by mmap function.",
      .nolaunch_on_fails = false,
      .abort_on_fail = true,
      .arg_list = ARG_LIST { "address", "length", "protocol", "flags", "fd", "offset", "expected error code", ARG_LIST_END },
      .args = ARGS_CHAIN {
        ARGS { 0, 0, PROT_WRITE, MAP_PRIVATE | MAP_ANON, 0, 0, EINVAL },
        ARGS { PAGE_SIZE - 1, 1, PROT_WRITE, MAP_FIXED | MAP_PRIVATE | MAP_ANON, 0, 0, EINVAL },
        ARGS { 0, PAGE_SIZE, PROT_WRITE, MAP_PRIVATE | MAP_ANON, 0, PAGE_SIZE - 1, EINVAL },
        ARGS { (16UL << 40UL) + PAGE_SIZE, PAGE_SIZE, PROT_READ, MAP_FIXED | MAP_PRIVATE | MAP_ANON, 0, 0, ENOMEM },
        ARGS { 0, PAGE_SIZE, PROT_READ, MAP_PRIVATE, 666, 0, ENOENT },
        ARGS { 0, (2UL << 40UL), PROT_READ, MAP_PHYS | MAP_ANON | MAP_PRIVATE, 0, (16UL << 40UL), EOVERFLOW },
        ARGS { 0, (2UL << 40UL), PROT_READ, MAP_ANON | MAP_PRIVATE | MAP_SHARED, 0, 0, EINVAL },
        ARGS { 0, (2UL << 40UL), PROT_READ, MAP_ANON | MAP_SHARED, 0, 0, ENOTSUP },
        ARGS_END,
      },
    },
    {
      "mmap_munmap", TC_map_munmap,
      "Testing general functionality of mmap and munmap function", false, true,
      ARG_LIST { "address", "length", "protocol", "flags", ARG_LIST_END },
      ARGS_CHAIN
      {
        ARGS { 0, PAGE_SIZE, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE  },
        ARGS { 0, PAGE_SIZE * 2, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE  },
        ARGS { 0, 16UL << 20, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE  },
        ARGS { 0x60000000, 1UL << 20, PROT_READ | PROT_WRITE, MAP_FIXED | MAP_ANON | MAP_PRIVATE  },
        ARGS_END,
      },
    },
    {
      "munmap_multi", TC_munmap_milti,
      "Testing of umapping several VM ranges", false, true,
      ARG_LIST { "fake", ARG_LIST_END },
      ARGS_CHAIN
      {
        ARGS { 0 },
        ARGS { 1 },
        ARGS_END,
      },
    },    
    {
      "maunmap_chunks", TC_munmap_chunks,
      "Testing of unmapping one large VM range by small chunks", false, true,
      ARG_LIST { "fake", ARG_LIST_END },
      ARGS_CHAIN
      {
        ARGS { 0 },
        ARGS_END,
      }
    },

    CASE_LIST_END,
  },  
};
