/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 MString Core Team <http://mstring.berlios.de>
 * (c) Copyright 2008,2009 Dmitry Gromada <gromada@jarios.org>
 *
 * Block device library test
 */

/*
 * NOTES:
 *    1) TC_iosched_merge_exceed test can fail in case of large requests and
 *       slow target platform. For example, it fails under the following conditions:
 *       a) software emulator
 *       b) kvm as platform, 256 sectors per request, 5 ms schedule timeout for write requests.
 *    2) Argument 'offset' is need for tests for tracer and I/O scheduler if the block
 *       device is a partition, but not an entire device.
 */

#include <test/framework.h>
#include <stdlib.h>
#include <sys/blkdev.h>
#include <sys/mman.h>
#include <sys/swks.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/scheduler.h>
#include <getopt.h>
#include <limits.h>
#include <ipc/rpc.h>

#define close(fd) ({0;})

#define NO_CHECKED_TRACE_ENTRY_ID 0xFFFFFFFF
#define MIN_SECT_SHIFT 6
#define MAX_SECT_SHIFT 16
#define MIN_BDEV_CAPACITY 2048

#define BLKIOC_GET_TRACE_EVENTS _IOWRARB(GENSTOR_CTLS, CTL_GET_TRACE_EVENTS, \
                                         struct key_desc)
#define PART_TBL_MAGIC 0xAA55

#define DOS_EXTENDED_PARTITION    0x5
#define WIN95_EXTENDED_PARTITION  0xF
#define LINUX_EXTENDED_PARTITION  0x85

/* ioctls, copypasted from fsio/bdio.h */

#define SERVER_CTLS 0

/* ioctls */
#define BLKIOC_GET_MDATA        _IOR(SERVER_CTLS, CTL_GET_MDATA, struct mdata)
#define BLKIOC_SET_ASYNC_PORT   _IOW(SERVER_CTLS, CTL_SET_ASYNC_PORT, ulong_t)

/* commands */
#define BLK_SERV_GET_MDATA        _IOCMD(SERVER_CTLS, CTL_GET_MDATA)
#define BLK_SERV_SET_ASYNC_PORT   _IOCMD(SERVER_CTLS, CTL_SET_ASYNC_PORT)

#define BDEV_CTLS 2

/* ioctls */
#define BLKIOC_FLUSH            _IO(BDEV_CTLS, CTL_FLUSH)

/* commands */
#define BLKCMD_FLUSH            _IOCMD(BDEV_CTLS, CTL_FLUSH)

/* ioctl numbers */
enum {
  CTL_GET_MDATA,
  CTL_SET_ASYNC_PORT,
};

enum {
  CTL_FLUSH
};

/* partition table entry types */
enum {
  PART_TBL_ENTRY_PRIMARY,
  PART_TBL_ENTRY_LOGICAL,
  PART_TBL_ENTRY_EBR
};

typedef uint64_t sector_t;

/* metadata structure, copypasted from fsio/bdio.h */

struct mdata {
  sector_t capacity;
  int hw_sect_bits;
  prtid_t cmd_port;
};

struct key_desc {
  uint64_t key;
  skey_type_t key_type;
};

/* partition related structures, copypasted from block/partiton.h */

typedef struct __part_cs_sector {
  uint32_t cyl : 6,
    sect : 10;
} __attribute__ ((packed)) part_cs_sector_t;

typedef struct __partition {
  uint8_t boot_ind;        /* boot identificator: 0x80 active */
  uint8_t start_head;
  part_cs_sector_t start_cs_sect;
  uint8_t fsid;            /* file system identifier */
  uint8_t end_head;
  part_cs_sector_t end_sect;
  uint32_t start_sector;  /* logical number of the first partition sector */
  uint32_t nr_sectors;    /* partition size in sectors */
} __attribute__ ((packed)) partition_t;

/* master boot record */
struct mbr {
  uint8_t __res[0x01BE];  /* boot code and free area */
  partition_t parts[4];
  uint16_t signature;
};

#if BITS_PER_LONG==64
sector_t str_to_capacity(char *s)
{
  return strtoul(s, NULL, 0);
}
#else
sector_t str_to_capacity(char *s)
{
  return strtoull(s, NULL, 0);
}
#endif /* BITS_PER_LONG */

static bool bdev_zeroed = false;
static char *bdev_name;
static sector_t bdev_capacity;
static uint32_t max_rq_sectors;
static int bdev_sect_shift = 9;
static sector_t first_dev_sector;
static bool rescan_failed;
static bool blk_trace_failed;
static const char *help_message="Test usage: t_libblkdev [options]:\n\n"
    "-n, --name <device name> (o)                    name of device to test\n"
    "-c, --capacity <capacity>                       device capacity in sectors\n"
    "-s, --sector-shift <bits>                       bits per sector size\n"
    "-m, --max-sectors <sectors number> (o)          max number of sectors per on request\n"
    "-o, --offset                                    sector the device starts on the entire one\n"
    "-z, --zeroed                                    the device is entirely filled by zeros\n"
    "--no-rw-stream                                  don't launch read/write stream tests\n"
    "--no-opclose-inval                              don't launch test for open/close invalid "
                                                     "functionality\n"
    "--no-iosched                                    don't launch tests for block I/O scheduler\n"
    "--no-blktrace                                   don't launch tests for tracer\n"
    "--no-part-scan                                  don't launch tests for partitions scanning\n"
    "\nOptions marked by (o) are obligatory to specify\n";

/* executed test flags */
static bool rw_stream_tests = true;
static bool part_scan_tests = true;
static bool iosched_tests = true;
static bool blk_trace_tests = true;
static bool open_close_inval_test = true;

typedef struct __rw_desc {
  uint16_t oflags;
  long woff;
  long wsize;
  char *wbuf;
  long exp_wret;
  long roff;
  long rsize;
  char *rbuf;
  long exp_rret;
  int rw_pause;
} rw_desc_t;

typedef struct __sect_range {
  sector_t start;
  uint32_t len;
} sect_range_t;

typedef struct __merge_test_opts {
  sect_range_t *front_range;
  sect_range_t *orig_range;
  sect_range_t *tail_range;
  bool rw;
  bool orig_is_first;
} merge_test_opts_t;

typedef struct __rw_thread_desc {
  int fd;
  sect_range_t *range;
  void *buf;
  int rw;
} rw_thread_desc_t;

typedef bool (*check_mergings_fn_t)(int fd, sect_range_t *front_range,
                                    sect_range_t *orig_range, sect_range_t *tail_range);

static bool measure_read_time;

static int test_init(void)
{
  int opt;
  int ret = 0;
  int index;
  int fd;

  static struct option long_options[] = {
    {"name", 1, 0, 'n'},
    {"capacity", 1, 0, 'c'},
    {"sector-shift", 1, 0, 's'},
    {"max-sectors", 1, 0, 'm'},
    {"offset", 1, 0, 'c'},
    {"zeroed", 0, 0, 'z'},
    {"no-rw-stream", 0, 0, 0},
    {"no-opclose-inval", 0, 0, 0},
    {"no-iosched", 0, 0, 0},
    {"no-blktrace", 0, 0, 0},
    {"no-part-scan", 0, 0, 0},
    {"help", 0, 0, 0}
  };

  while (!ret) {
    opt = getopt_long_only(cmdline_args_num, cmdline_args, "n:c:s:m:o:z", long_options, &index);
    if (opt == -1)
      break;

    switch (opt) {
    case 'n':
      bdev_name = malloc(strlen(optarg) + 1);
      if (bdev_name == NULL)
        ret = -1;
      else
        strcpy(bdev_name, optarg);
      break;
    case 'c':
      bdev_capacity = str_to_capacity(optarg);
      break;
    case 's':
      bdev_sect_shift = atoi(optarg);
      break;
    case 'm':
      max_rq_sectors = atoi(optarg);
      break;
    case 'o':
      first_dev_sector = strtol(optarg, NULL, 0);
      if (first_dev_sector == LONG_MIN || first_dev_sector == LONG_MAX) {
        printf("Wrong first sector value!\n");
        ret = -1;
      }
      break;
    case 'z':
      bdev_zeroed = false;
      break;
    case 0:
      switch (index) {
      case 6:
        rw_stream_tests = false;
        break;
      case 7:
        open_close_inval_test = false;
        break;
      case 8:
        iosched_tests = false;
        break;
      case 9:
        blk_trace_tests = false;
        break;
      case 10:
        part_scan_tests = false;
        break;
      case 11:
        printf(help_message);
        exit(0);
      }
      break;
    default:
      printf(help_message);
      exit(0);
    }
  }

  if (ret)
    return ret;

  if ((bdev_sect_shift < MIN_SECT_SHIFT) || (bdev_sect_shift > MAX_SECT_SHIFT)) {
    printf("Sector shift is wrong, it must be from %d upto %d\n", MIN_SECT_SHIFT, MAX_SECT_SHIFT);
    return -1;
  }
  if (!(bdev_name && max_rq_sectors)) {
    printf("Device name or maximum setors per request is not determined!\n%s",
           help_message);
    exit(0);
  }

  fd = open(bdev_name, O_RDWR);
  ioctl(fd, BLKIOC_CLEAR_TRACE);
  close(fd);

  return 0;
}


static void __prepare_rw_test(rw_desc_t *desc)
{
  int i;

  desc->wbuf = malloc(desc->wsize);
  desc->rbuf = malloc(desc->rsize);

  memset(desc->rbuf, 0, desc->rsize);
  for (i = 0; i < desc->wsize; i++)
  {
    desc->wbuf[i] = i & 0xFF;
  }
}

static void __free_rw_test_resources(rw_desc_t *desc)
{
  free(desc->wbuf);
  free(desc->rbuf);
}

static void fill_part(partition_t *part, sector_t start, sector_t size,
                      sector_t offset, int entry_type, uint8_t id)
{
  part->fsid = id;
  switch (entry_type) {
  case PART_TBL_ENTRY_PRIMARY:
    part->start_sector = start;
    part->nr_sectors = size;
    break;
  case PART_TBL_ENTRY_EBR:
    part->start_sector = start - offset;
    part->nr_sectors = size;
    break;
  case PART_TBL_ENTRY_LOGICAL:
    part->start_sector = offset;
    part->nr_sectors = size - offset;
    break;
  }
}

static bool __generic_rw_stream_test(rw_desc_t *desc, int fd)
{
  size_t r = 0;
  bool ret = true;
  int fd1;
  swks_t *swks;
  uint64_t ticks1, ticks2;
  uint32_t ms;

  if (fd == -1) {
    fd1 = open(bdev_name, O_RDWR | desc->oflags);
    if (fd1 == -1) {
      test_fail("Device opening failed, error = %s\n", error_strings[-r]);
      return false;
    }
  } else {
    fd1 = fd;
  }

  lseek(fd1, desc->woff, SEEK_SET);
  r = write(fd1, desc->wbuf, desc->wsize);
  if (!check_test_condition_ext(r, desc->exp_wret, COMP_EQUAL, 0,
                                "write error", false))
    ret = false;

  if (ret && desc->rsize) {
    if (desc->rw_pause)
      sleep(desc->rw_pause);

    lseek(fd1, desc->roff, SEEK_SET);
    if (measure_read_time) {
      swks = get_swks();
      /* set maximum real-time priority so that the process couldn't be forced out */
      task_scheduler_control(0, 0, SYS_SCHED_CTL_SET_PRIORITY, 1);
      ticks1 = current_sys_ticks(swks);
    }
    r = read(fd1, desc->rbuf, desc->rsize);
    if (measure_read_time) {
      ticks2 = current_sys_ticks(swks);
      /* restore normal priority */
      task_scheduler_control(0, 0, SYS_SCHED_CTL_SET_PRIORITY, 96);
      ms = ticks_to_msec(swks, ticks2 - ticks1);
      printf("ticks1 = %ld, ticks2 = %ld, swks->hz = %d, read time = %d milliseconds\n",
             ticks1, ticks2, swks->hz, ms);
    }
    ret = check_test_condition_ext(r, desc->exp_rret, COMP_EQUAL, 0,
                                   "read error", false);
  }

  if (fd == -1)
    close(fd);

  return ret;
}

static bool __open_close(const char *name, bool exist, bool quiet)
{
  int r;
  int fd;

  if (!quiet) {
    printf("Testing for openning\n");
  }
  fd = open(name, O_RDWR, 0);
  if (!exist)
    return ((fd == -1) && (errno = ENODEV));

  if (fd == -1) {
    if (!quiet)
      test_fail("error is %s\n", error_strings[errno]);

    return false;
  }

  if (!quiet) {
    printf("Testing for close\n");
  }
  r = close(r);
  if (!quiet) {
    check_test_condition_ext(r, 0, COMP_EQUAL, 0, NULL, true);
  }
  return !r;
}

static inline trace_event_data_t *__trace_event_first(trace_entry_data_t *td)
{
  return (trace_event_data_t*)((ulong_t)td + sizeof(trace_entry_data_t));
}

static inline trace_event_data_t *__trace_event_next(trace_event_data_t *ev)
{
  return (trace_event_data_t*)((ulong_t)ev + sizeof(trace_event_data_t) +
                               ev->ext_data_size);
}

static trace_event_data_t *__search_event(void *buf, void *min_ev_pos,
                                          trace_event_type_t type)
{
  trace_entry_data_t *te;
  trace_event_data_t *ev, *ret_ev = NULL;
  int i;

  te = (trace_entry_data_t*)buf;
  ev = __trace_event_first(te);

  for (i = 0; i < te->nr_events; i++) {
    if ((ev->type == type) && (ev > (trace_event_data_t*)min_ev_pos)) {
      ret_ev = ev;
      break;
    }
    ev = __trace_event_next(ev);
  }

  return ret_ev;
}

/* Auxiliary functions to test request merging */

void *__rw_thread_fn(void *data)
{
  rw_thread_desc_t *desc = (rw_thread_desc_t*)data;

  lseek(desc->fd, (desc->range->start - first_dev_sector) << bdev_sect_shift, SEEK_SET);
  if (desc->rw)
    write(desc->fd, desc->buf, desc->range->len << bdev_sect_shift);
  else
    read(desc->fd, desc->buf, desc->range->len << bdev_sect_shift);

  return NULL;
}

static inline void __fill_merge_test_opts(merge_test_opts_t *opts, sect_range_t *front_range,
                                          sect_range_t *orig_range, sect_range_t *tail_range,
                                          bool rw, bool orig_is_first)
{
  opts->front_range = front_range;
  opts->orig_range = orig_range;
  opts->tail_range = tail_range;
  opts->rw = rw;
  opts->orig_is_first = orig_is_first;
}

static bool __check_merge_range(merge_data_t *md, sect_range_t *prev,
                                sect_range_t *curr)
{
  return ((md->old_start_sect == prev->start) && (md->old_nr_sects == prev->len) &&
          (md->new_start_sect == curr->start) && (md->new_nr_sects == curr->len));
}

static bool __check_trace_entry_ranges(trace_entry_data_t *edata, sect_range_t *orig_range,
                                       sect_range_t *curr_range)
{
  bool ret = true;

  if (!((edata->orig_start_sect == orig_range->start) &&
       (edata->orig_nr_sects == orig_range->len) &&
       (edata->curr_start_sect == curr_range->start) &&
       (edata->curr_nr_sects == curr_range->len))) {

    printf("Sector ranges of merged request are invalid: "
            "original start sector = %ld, original number of sectors = %d, "
            "final start sector = %ld, final number of sectos = %d; "
            "but must be: %ld, %d, %ld, %d\n",
            edata->orig_start_sect, edata->orig_nr_sects,
            edata->curr_start_sect, edata->curr_nr_sects,
            orig_range->start, orig_range->len,
            curr_range->start, curr_range->len);
    ret = false;
  }

  return ret;
}

static bool __check_merging_single(int fd, sect_range_t *front_range,
                                   sect_range_t *orig_range, sect_range_t *tail_range)
{
  int i;
  bool ret = true;
  sect_range_t range = *orig_range, prev_range;
  trace_entry_data_t *edata;
  trace_event_data_t *ev = NULL;
  merge_data_t *md;
  char trace_buf[1024];
  merge_type_t where[2] = {BLK_MERGE_TAIL, BLK_MERGE_FRONT};

  i = get_blkdev_trace_events(fd, orig_range->start, SKEY_TYPE_SECTOR,
                              trace_buf, sizeof(trace_buf));
  if (i < 0) {
    printf("Unable to get trace events, error = %s\n", error_strings[-i]);
    return false;
  }

  /* calculate final range */
  if (front_range) {
    range.start = front_range->start;
    range.len += orig_range->start - front_range->start;
  }
  if (tail_range) {
    range.len += (tail_range->start + tail_range->len) -
                 (orig_range->start + orig_range->len);
  }
  edata = (trace_entry_data_t*)trace_buf;
  ret = __check_trace_entry_ranges(edata, orig_range, &range);

  /* check merging events */
  range = prev_range = *orig_range;
  ev = (trace_event_data_t*)edata;
  for (i = 0; (i < 2) && ret; i++) {
    if (!i && front_range) {
      range.start = front_range->start;
      range.len += orig_range->start - front_range->start;
    } else if (i && tail_range) {
      range.len += (tail_range->start + tail_range->len) -
                   (orig_range->start + orig_range->len);
    } else {
      continue;
    }
    ev = __search_event(edata, ev, TRACE_EVENT_MERGE);
    if (ev == NULL) {
      ret = false;
      printf("Event for merging is absent\n");
    } else {
      md = (merge_data_t*)ev->ext_data;
      if (!(md->with_blkio && (md->type = where[i]) &&
            __check_merge_range(md, &prev_range, &range))) {

        if (!md->with_blkio) {
          printf("Request is not merged with block I/O structure\n");
        } else if (md->type != where[i]) {
          printf("Merge location is invalid\n");
        }
        ret = false;
      }
    }
    prev_range = range;
  }

  return ret;
}

static bool __check_merging_clustered(int fd, sect_range_t *front_range,
                                      sect_range_t *orig_range, sect_range_t *tail_range)
{
  int i;
  bool ret = true;
  sect_range_t range = *orig_range, prev_range;
  trace_entry_data_t *edata;
  trace_event_data_t *ev = NULL;
  merge_data_t *md;
  char trace_buf[1024];

  i = get_blkdev_trace_events(fd, front_range->start, SKEY_TYPE_SECTOR,
                                    trace_buf, sizeof(trace_buf));
  if (i == -1) {
    printf("Unable to get trace events, error = %s\n", error_strings[-errno]);
    return false;
  }

  /* final range */
  range.start = front_range->start;
  range.len = (tail_range->start + tail_range->len) - range.start;

  edata = (trace_entry_data_t*)trace_buf;
  ret = __check_trace_entry_ranges(edata, front_range, &range);

  /* check merging events */
  range = prev_range = *front_range;
  ev = (trace_event_data_t*)edata;
  for (i = 0; (i < 2) && ret; i++) {
    if (!i) {
      range.len += orig_range->len;
    } else {
      range.len += tail_range->len;
    }
    ev = __search_event(edata, ev, TRACE_EVENT_MERGE);
    if (ev == NULL) {
      ret = false;
      printf("Event for merging is absent\n");
    } else {
      md = (merge_data_t*)ev->ext_data;
      if ((!i && !md->with_blkio) || (i && md->with_blkio)) {
        printf("Type of merged object is invalid\n");
        ret = false;
      }
      if (ret && (md->type != BLK_MERGE_TAIL)) {
        printf("Request is not merged with block I/O structure\n");
        ret = false;
      }
      if (ret && i && (md->rq_id - edata->id != 1)) {
        printf("Id of merged request is invalid\n");
        ret = false;
      }
      ret = ret && __check_merge_range(md, &prev_range, &range);
    }
    prev_range = range;
  }

  return ret;
}

static bool __check_merging_exceed(int fd, sect_range_t *front_range,
                                   sect_range_t *orig_range, sect_range_t *tail_range)
{
  int i, r;
  bool ret = true;
  trace_entry_data_t *edata;
  char trace_buf[1024];
  sect_range_t new_orig_range;
  sect_range_t *begin_ranges[3] = {front_range, &new_orig_range, tail_range};
  sect_range_t final_ranges[3];
  sector_t total_sectors;

  total_sectors = tail_range->start + tail_range->len - front_range->start;
  final_ranges[0].start = front_range->start;
  final_ranges[2].start = tail_range->start - (max_rq_sectors - tail_range->len);
  final_ranges[0].len = final_ranges[2].len = max_rq_sectors;
  new_orig_range.start = front_range->start + max_rq_sectors;
  new_orig_range.len = final_ranges[2].start - new_orig_range.start;
  final_ranges[1] = new_orig_range;

  /*
   * it must be 3 processed requests, no check more merging events,
   * they was be tested earlier
   */

  for (i = 0; (i < 3) && ret; i++) {
    if (!final_ranges[i].len)
      continue;

    r = get_blkdev_trace_events(fd, begin_ranges[i]->start, SKEY_TYPE_SECTOR,
                                trace_buf, sizeof(trace_buf));
    if (i == 1 && total_sectors <= max_rq_sectors * 2) {
      if (!r) {
        printf("New request was made instead of merging\n");
        ret = false;
      }
      continue;
    } else {
      if (r < 0) {
        printf("Unable to get trace events, error = %s\n", error_strings[-r]);
        return false;
      }
    }

    edata = (trace_entry_data_t*)trace_buf;
    ret = __check_trace_entry_ranges(edata, begin_ranges[i], &final_ranges[i]);
  }

  return ret;
}

/* perform mergeable operations summary not exceeding maximum request size */
static bool __test_mergeable_ioops(int fd, merge_test_opts_t *test_opts,
                                   char *buffer, bool pass_on_success,
                                   check_mergings_fn_t check_fn)
{
  bool ret = true;
  rw_thread_desc_t desc[3];
  sect_range_t *ranges[3];
  int i, n = 0;
  sector_t min_sect;
  char *buf1;
  pthread_t th[3];

  if (test_opts->orig_is_first) {
    ranges[n++] = test_opts->orig_range;
  }
  if (test_opts->front_range) {
    ranges[n++] = test_opts->front_range;
    min_sect = test_opts->front_range->start;
  } else {
    min_sect = test_opts->orig_range->start;
  }
  if (test_opts->tail_range) {
    ranges[n++] = test_opts->tail_range;
  }
  if (!test_opts->orig_is_first) {
    ranges[n++] = test_opts->orig_range;
  }

  for (i = 0; i < n; i++) {
    buf1 = buffer + ((ranges[i]->start - min_sect) << bdev_sect_shift);
    desc[i].fd = fd;
    desc[i].range = ranges[i];
    desc[i].buf = buf1;
    desc[i].rw = test_opts->rw;
    pthread_create(&th[i], NULL, __rw_thread_fn, &desc[i]);
  }

  for (i = 0; i < n; i++)
    pthread_join(th[i], NULL);

  ret = check_fn(fd, test_opts->front_range, test_opts->orig_range,
                 test_opts->tail_range);

  if (!ret) {
    test_fail(NULL);
  } else if (pass_on_success) {
    test_pass();
  }

  return ret;
}

static void TC_open_close(long *args)
{
  __open_close(bdev_name, true, false);
}

static void TC_fstat(long *args)
{
  int fd, r;
  struct stat st;

  fd = open(bdev_name, O_RDONLY);
  if (!check_test_condition_ext(fd, 0, COMP_GREQU, 0, "open error", false))
    return;

  r = fstat(fd, &st);
  if (check_test_condition_ext(r, 0, COMP_EQUAL, 0, NULL, false)) {
    if (!bdev_capacity)
      bdev_capacity = st.st_size >> bdev_sect_shift;

    if ((st.st_size >> bdev_sect_shift) == bdev_capacity &&
        st.st_blksize == PAGE_SIZE) {
      test_pass();
    } else {
      test_fail("File statistics are wrong\n");
    }
  }
  close(fd);
}

static void rw_stream_test(long *args, int rw_delay)
{
  long i;
  bool pass = true;
  rw_desc_t desc = { 0, args[0], args[1], NULL, args[1], args[2],
                     args[3], NULL, args[3], rw_delay };

  __prepare_rw_test(&desc);

  if (__generic_rw_stream_test(&desc, -1)) {
    /* check if the regions overlapp and compare data */
    if (desc.roff >= desc.woff) {
      if (desc.roff - desc.woff < desc.wsize) {
        desc.woff = desc.roff - desc.woff;
        desc.roff = 0;
        desc.rsize = MIN(desc.rsize, desc.wsize - desc.woff);
      } else {
        desc.rsize = 0;
      }
    } else if (desc.woff - desc.roff < desc.rsize) {
      desc.roff = desc.woff - desc.roff;
      desc.woff = 0;
      desc.rsize = MIN(desc.wsize, desc.rsize - desc.roff);
    } else {
      desc.rsize = 0;
    }

    if (desc.rsize) {
      pass = pass && !memcmp(desc.wbuf + desc.woff, desc.rbuf + desc.roff, desc.rsize);
      if (!pass) {
        for (i = 0; i < desc.rsize; i++) {
          if (desc.wbuf[desc.woff + i] != desc.rbuf[desc.roff + i]) {
            printf("First error is occurred at %d index, wbuf[%d] = %d, rbuf[%d] = %d\n",
                   desc.roff + i, desc.woff + i, desc.wbuf[desc.woff + i], desc.roff + i,
                   desc.rbuf[desc.roff + i]);
              break;
          }
        }
        test_fail("Data don't tally\n");
      }

      if (pass && bdev_zeroed) {
        /* check that data was writte/readen with valid offset */
        for (i = 0; (i < desc.roff) && (desc.rbuf[i] == 0); i++);
        pass = pass && (i == desc.roff);

        for (i = desc.roff + desc.rsize; (i < args[3]) && (desc.rbuf[i] == 0); i++);
        pass = pass && (i == args[3]);

        if (!pass)
          printf("Data was read from a wrong offset\n");
      }
    }

    if (pass)
      test_pass();
  }

  __free_rw_test_resources(&desc);
}

static void TC_read_write_stream(long *args)
{
  rw_stream_test(args, 1);
}

void TC_rw_stream_multirequest(long *args)
{
  uint32_t size;
  long a[4];

  size = (max_rq_sectors * 6 << bdev_sect_shift) + 100;
  a[0] = a[2] = 8100;
  a[1] = a[3] = size;

  measure_read_time = true;
  rw_stream_test(a, 3);
  measure_read_time = false;
}

static void TC_read_write_stream_exceed(long *args)
{
  long i;
  rw_desc_t desc = {
    0,
    (bdev_capacity - 1) << bdev_sect_shift,
    2 << bdev_sect_shift,
    NULL,
    1 << bdev_sect_shift,
    (bdev_capacity - 1) << bdev_sect_shift,
    2 << bdev_sect_shift,
    NULL,
    1 << bdev_sect_shift,
    1
  };

  __prepare_rw_test(&desc);

  if (__generic_rw_stream_test(&desc, -1)) {
    if (!memcmp(desc.wbuf, desc.rbuf, desc.exp_rret)) {
      for (i = desc.exp_rret; (i < desc.rsize) && !desc.rbuf[i]; i++);
      if (i == desc.rsize) {
        test_pass();
      } else {
        test_fail("Size of read data is greater than the returned result!\n");
      }
    } else {
      test_fail("Written and read data don't compare!\n");
    }
  }

  __free_rw_test_resources(&desc);
}

static bool check_mdata(char *name, sector_t capacity, bool quiet)
{
  int r;
  int fd;
  struct mdata md;
  bool pass;

  fd = open(name, O_RDONLY);
  if (fd == -1) {
    if (!quiet)
      test_fail("open fail error");

    return false;
  }

  r = ioctl(fd, BLKIOC_GET_MDATA, &md);
  if (r == -1) {
    if (!quiet)
      test_fail("get server metadata error: %s\n", error_strings[errno]);
  } else {
    pass = (md.capacity == capacity) && (md.hw_sect_bits == bdev_sect_shift);
    if (!quiet) {
      if (pass) {
        test_pass();
      } else {
        test_fail("Metadata are invalid: capacity = %ld, sector shift = %d, "
                  "instead of %ld and %d respectively\n", md.capacity, md.hw_sect_bits,
                   capacity, bdev_sect_shift);
      }
    }
  }
  close(fd);

  return pass;
}

static void TC_check_get_mdata(long *args)
{
  check_mdata(bdev_name, bdev_capacity, false);
}

static void TC_rescan_partitions(long *args)
{
  struct mbr mbr;
  int i, j, r;
  int fd;
  sector_t start = 0x10, nr_sects = 1;
  rw_desc_t desc;
  partition_t *parts;
  char buf[32];
  bool pass = true;
  sector_t ebr_offs[] = {5, 16, 4};     /* offsets of first partition from ebr */
  sector_t ebr_starts[] = {96, 200, 700, 0};
  sector_t ebr_sizes[] = {50, 140, 300, 0};
  sector_t ebr_ids[] = {LINUX_EXTENDED_PARTITION, WIN95_EXTENDED_PARTITION,
                        DOS_EXTENDED_PARTITION, 0};
  sector_t prim_sizes[] = {8, 17, 31};
  uint8_t id;

  /* write boot records to the disk */
  printf("Writing mbr containing the linux extended partition\n");

  rescan_failed = true;
  fd = open(bdev_name, O_RDWR, 0);
  if (!check_test_condition_ext(fd, 0, COMP_GREQU, 0, "open error", false))
    return;

  memset(&mbr, 0, sizeof(mbr));
  mbr.signature = PART_TBL_MAGIC;
  parts = mbr.parts;
  j = 0;
  for (i = 0; i < 4; i++) {
    if (i == 2) {
      fill_part(&parts[i], ebr_starts[0], (ebr_sizes[0] + ebr_sizes[1] + ebr_sizes[2]),
                0, PART_TBL_ENTRY_EBR, LINUX_EXTENDED_PARTITION);
    } else {
      fill_part(&parts[i], start, prim_sizes[j], 0, PART_TBL_ENTRY_PRIMARY, 1);
      start += prim_sizes[j];
      j++;
    }
  }

  memset(&desc, 0, sizeof(desc));
  desc.woff = 0;
  desc.wsize = sizeof(mbr);
  desc.wbuf = (char*)&mbr;
  desc.exp_wret = sizeof(mbr);

  if (!__generic_rw_stream_test(&desc, fd))
    return;

  printf("Write 3 ebrs with different accessible ids, "
         "the second of the ebrs has no active partition\n");
  r = 0;
  memset(&mbr, 0, sizeof(mbr));
  mbr.signature = PART_TBL_MAGIC;
  for (j = 0; (j < 3) && !r; j++) {
    desc.woff = ebr_starts[j] << bdev_sect_shift;
    id = (j == 1) ? 0 : 1;
    fill_part(parts, ebr_starts[j], ebr_sizes[j],
              ebr_offs[j], PART_TBL_ENTRY_LOGICAL, id);
    fill_part(&parts[1], ebr_starts[j+1], ebr_sizes[j+1], ebr_starts[0],
              PART_TBL_ENTRY_EBR, ebr_ids[j+1]);
    if (!__generic_rw_stream_test(&desc, fd))
      r = 1;
  }
  if (r)
    return;

  sleep(1);
  printf("Try to rescan\n");
  r = ioctl(fd, BLKIOC_RESCAN_PARTITIONS);
  if (check_test_condition_ext(r, 0, COMP_EQUAL, 0, NULL, false)) {
    bool b;
    nr_sects = 1;

    /* verify that paritions are registered correctly */
    for (i = 1; i <= 7; i++) {
      sprintf(buf, "%s%d", bdev_name, i);
      b = (i == 3 || i == 6);
      if (!__open_close(buf, !b, true)) {
        pass = false;
        if (b) {
          printf("Device %s is registered but must not\n", buf);
        } else {
          printf("Device %s is not registered\n", buf);
        }
      } else if (!b) {
        if (i <= 4)
          nr_sects = (i < 3) ? prim_sizes[i - 1] : prim_sizes[i - 2];
        else
          nr_sects = ebr_sizes[i - 5] - ebr_offs[i - 5];

        if (!check_mdata(buf, nr_sects, true)) {
          pass = false;
          printf("%s storage metadata are wrong\n", buf);
        }
      }
    }
    if (pass) {
      rescan_failed = false;
      test_pass();
    } else {
      test_fail(NULL);
    }
  }
  close(fd);
}

static void TC_unreg_rescan_parts(long *args)
{
  struct mbr mbr;
  int i, r;
  int fd;
  rw_desc_t desc;
  partition_t *parts;
  char buf[32];
  bool pass = true;

  if (rescan_failed) {
    printf("It is not matter because of rescan partitions test fail\n");
    return;
  }

  fd = open(bdev_name, O_RDWR, 0);
  if (!check_test_condition_ext(fd, 0, COMP_GREQU, 0, "open error", false))
    return;

  memset(&mbr, 0, sizeof(mbr));
  mbr.signature = PART_TBL_MAGIC;
  parts = mbr.parts;
  parts[3].fsid = 1;
  parts[3].start_sector = 0x10;
  parts[3].nr_sectors = 0x20;

  memset(&desc, 0, sizeof(desc));
  desc.woff = 0;
  desc.wsize = sizeof(mbr);
  desc.wbuf = (char*)&mbr;
  desc.exp_wret = sizeof(mbr);

  if (!__generic_rw_stream_test(&desc, fd))
    return;

  sleep(1);
  printf("Try to rescan\n");
  r = ioctl(fd, BLKIOC_RESCAN_PARTITIONS);
  if (check_test_condition_ext(r, 0, COMP_EQUAL, 0, NULL, false)) {
    /* verify that paritions are registered correctly */
    for (i = 1; i <= 8; i++) {
      sprintf(buf, "%s%d", bdev_name, i);
      if (!__open_close(buf, i == 4, true)) {
        pass = false;
        if (i != 4) {
          printf("Device %s is registered but must not\n", buf);
        } else {
          sprintf("Device %s is not registered\n", buf);
        }
      } else if (i == 4) {
        if (!check_mdata(buf, 0x20, false)) {
          pass = false;
          printf("%s storage metadata are wrong\n", buf);
        }
      }
    }
    if (pass) {
      test_pass();
    } else {
      test_fail(NULL);
    }
  }

  close(fd);
}

void TC_read_write_partition(long *args)
{
  long i, r;
  int fd1, fd2;
  rw_desc_t desc;
  char name[32];
  uint32_t sect_size = 1 << bdev_sect_shift;
  struct mdata md1, md2;

  if (rescan_failed) {
    printf("It is not matter because of rescan partitions test fail\n");
    return;
  }

  memset(&desc, 0, sizeof(desc));

  desc.wsize = 32 << bdev_sect_shift;
  desc.exp_wret = sect_size,
  desc.rsize = sect_size,
  desc.exp_rret = sect_size,
  desc.rw_pause = 1;

  __prepare_rw_test(&desc);

  /* clear 4-th and 6-th partitions */
  memset(desc.wbuf, 0, desc.wsize);
  sprintf(name, "%s%d", bdev_name, 1);
  fd1 = open(name, O_RDWR, 0);
  if (fd1 == -1) {
    test_fail("Cannot open partition 1\n");
    goto out_free_resources;
  }

  r = ioctl(fd1, BLKIOC_GET_MDATA, &md1);
  if (!check_test_condition_ext(r, 0, COMP_EQUAL, 0, "Getting server metadata failed", false))
    goto out_close_bdev1;

  sprintf(name, "%s%d", bdev_name, 2);
  fd2 = open(name, O_RDWR, 0);
  if (fd2 == -1) {
    test_fail("Cannot open partition 2\n");
    goto out_close_bdev1;
  }

  r = ioctl(fd2, BLKIOC_GET_MDATA, &md2);
  if (!check_test_condition_ext(r, 0, COMP_EQUAL, 0, "Getting server metadata failed", false))
    goto out;

  write(fd1, desc.wbuf, md1.capacity << bdev_sect_shift);
  write(fd2, desc.wbuf, md2.capacity << bdev_sect_shift);

  desc.woff = (md1.capacity - 1) << bdev_sect_shift;
  desc.wsize = md1.capacity << bdev_sect_shift;
  for (i = 0; i < desc.wsize; i++) {
    desc.wbuf[i] = i & 0xFF;
  }

  if (__generic_rw_stream_test(&desc, fd1)) {
    /* check that data on the adjacent partition are not affected */
    lseek(fd2, 0, SEEK_SET);
    r = read(fd2, desc.rbuf, sect_size);
    if (r < 0) {
      test_fail("Fail at reading from the adjacent partition, errno = %s\n", error_strings[-r]);
    } else {
      /* check that all bytes are 0 */
      for (i = 0; (i < sect_size) && !desc.rbuf[i]; i++);
      if (i == sect_size) {
        test_pass();
      } else {
        test_fail("Data on the adjacent partition are affected, rbuf[%d] = %d\n",
                  i, desc.rbuf[i]);
      }
    }
  }

out:
  close(fd2);
out_close_bdev1:
  close(fd1);
out_free_resources:
  __free_rw_test_resources(&desc);
}

static void TC_write_sync(long *args)
{
  rw_desc_t desc = {
    O_SYNC,
    300,
    1536,
    NULL,
    1536,
    300,
    1536,
    NULL,
    1536,
    0
  };

  printf("Testing synchrounous writing. Read just after writing without any delay and "
                 "check data were tally\n");
  __prepare_rw_test(&desc);
  memset(desc.wbuf, 0xFF, desc.wsize);

  if (__generic_rw_stream_test(&desc, -1)) {
    if (!memcmp(desc.wbuf, desc.rbuf, desc.exp_rret)) {
      test_pass();
    } else {
      test_fail(NULL);
    }
  }

  __free_rw_test_resources(&desc);
}

static void TC_open_close_invalid(long *args)
{
  int r, fd;
  uni_rpc_t rpc;
  chid_t chan;
  pid_t pid;
  bool pass = false;

  fd = open(bdev_name, O_RDWR, 0);
  if (fd == -1) {
    test_fail("failed to open device, errno = %s\n", error_strings[-r]);
    return;
  }

  printf("Testing for direct closing by-passing the file system\n");
  memset(&rpc, 0, sizeof(rpc));
  rpc.rpc_call = RPC_UNREGISTER;
  rpc.arg0 = getpid();
  rpc.arg1 = getufsid(fd);
  pid = getfpid(fd);
  /* opening channel to write to the server's command port directly */
  chan = ipc_channel_open(pid, 0, IPC_BLOCKED_ACCESS);
  if ((signed)chan < 0) {
    test_fail("Cannot open channel, %s error\n", error_strings[-chan]);
  } else {
    ipc_channel_msgsnd(chan, sizeof(rpc), &rpc, sizeof(rpc), &rpc);
    ipc_channel_close(chan);
    pass = check_test_condition(rpc.rpc_call, EPERM, NULL, true);
  }

  if (pass)
    close(fd);
}

static void TC_read_write_stream_invalid(long *args)
{
  char buf[512];
  int ret, fd;

  printf("Trying to read with offset exceeding the block device capacity\n");
  fd = open(bdev_name, O_RDWR, 0);
  if (fd == -1) {
    printf("Device opening failed, %s error\n", error_strings[errno]);
    return;
  }

  lseek(fd, bdev_capacity * 2 << bdev_sect_shift, SEEK_SET);
  ret = read(fd, buf, 512);
  check_test_condition_ext(ret, -1, COMP_EQUAL, EINVAL, NULL, true);

  close(fd);
}

void TC_ioctl_invalid(long *args)
{
  int fd1;
  //ulong_t p;

  printf("Testing error behavior of block server ioctl\n");
  fd1 = open(bdev_name, O_RDWR, 0);
  if (!check_test_condition_ext(fd1, 0, COMP_GREQU, 0, "device opening failed", false))
    return;

#if 0   /* uncomment when security subsystem will be implemented */
  p = ipc_port_open(0, DQUEUE_SIZE);
  printf("Testing for set asynchrounous port without having privileges\n");
  ret = ioctl(fd1, BLKIOC_SET_ASYNC_PORT, &p);
  check_test_condition_ext(ret, -1, COMP_EQUAL, EPERM, NULL, true);
#endif

#ifdef PART_SCAN_TESTS
  if (!rescan_failed) {
    int fd2;
    int ret;
    char name[32];

    printf("Trying to rescan partitions when some partitions are yet opened\n");
    sprintf(name, "%s%d", bdev_name, 4);
    fd2 = open(name, O_RDWR, 0);
    if (!check_test_condition_ext(fd2, 0, COMP_GREQU, 0, "partition opening failed", false))
      goto out;

    ret = ioctl(fd1, BLKIOC_RESCAN_PARTITIONS);
    check_test_condition_ext(ret, -1, COMP_EQUAL, EBUSY, NULL, true);
    close(fd2);
  }

out:
#endif

  close(fd1);
}

static bool __check_nomerged_trace_entry(int fd, uint32_t id, sector_t start_sector,
                                         skey_type_t key_type, uint32_t nr_sectors, void *err_buf)
{
  bool ret = true;
  trace_entry_data_t *edata;
  trace_event_data_t *ev;
  uint32_t nr_events;
  int i;
  int nr_start_proc = 0, nr_end_proc = 0;
  int nr_add_iosched = 0, nr_merge = 0;
  char trace_buf[1024];
  uint64_t key;

  edata = (trace_entry_data_t*)trace_buf;
  key = (key_type == SKEY_TYPE_ID) ? id : start_sector;
  i = get_blkdev_trace_events(fd, key, key_type, trace_buf, sizeof(trace_buf));
  if (i < 0) {
    sprintf(err_buf, "Trace entry get error: %s\n", error_strings[-i]);
    return false;
  }

  nr_events = edata->nr_events;
  if (nr_events < 1) {
    sprintf(err_buf, "Invalid number of events: %d, must be not less than 2\n",
            nr_events);
    return false;
  }

  if (key_type == SKEY_TYPE_ID || id != NO_CHECKED_TRACE_ENTRY_ID) {
    if (edata->id != id) {
      sprintf(err_buf, "Trace entry id is invalid: %d instead of %d\n", edata->id, id);
      return false;
    }
  }

  if (!((edata->orig_start_sect == start_sector) &&
        (edata->curr_start_sect == start_sector) &&
        (edata->orig_nr_sects == nr_sectors) &&
        (edata->curr_nr_sects == nr_sectors))) {

    ret = false;
    sprintf(err_buf, "Trace entry data are invalid: "
            "original start sector = %ld, original number of sectors = %d, "
            "final start sector = %ld, final number of sectos = %d; "
            "but must be: %ld, %d, %ld, %d\n",
            edata->orig_start_sect, edata->orig_nr_sects,
            edata->curr_start_sect, edata->curr_nr_sects,
            start_sector, nr_sectors, start_sector, nr_sectors);
  } else {
    ev = __trace_event_first(edata);

    /* check thare start processing and end processing event */
    for (i = 0; i < nr_events; i++) {
      if (ev->type > TRACE_EVENT_SCHEDULE) {
        sprintf(err_buf, "Invalid event type: %d\n", ev->type);
        ret = false;
        break;
      }

      switch (ev->type) {
      case TRACE_EVENT_START_PROC:  nr_start_proc++;  break;
      case TRACE_EVENT_END_PROC:    nr_end_proc++;    break;
      case TRACE_EVENT_ADD:         nr_add_iosched++; break;
      case TRACE_EVENT_MERGE:       nr_merge++;       break;
      default:                                        break;
      }
      ev = __trace_event_next(ev);
    }
    if (ret) {
      if (nr_start_proc != 1) {
        sprintf(err_buf, "Number of start processing events is %d, but must be 1\n",
                nr_start_proc);
        err_buf += strlen(err_buf);
        ret = false;
      }
      if (nr_end_proc != 1) {
        sprintf(err_buf, "Number of start processing events is %d, but must be 1\n",
                nr_end_proc);
        err_buf += strlen(err_buf);
        ret = false;
      }
      if (nr_add_iosched != 1) {
        sprintf(err_buf, "Number of add to scheduler events is %d, but must be 1\n",
                nr_add_iosched);
        err_buf += strlen(err_buf);
        ret = false;
      }
      if (nr_merge) {
        sprintf(err_buf, "Number of merge events is %d, but must be 0\n",
                nr_merge);
        err_buf += strlen(err_buf);
        ret = false;
      }
    }
  }

  return ret;
}

void TC_blk_trace(long *args)
{
  int fd;
  int err, i, r;
  char buf[1024];
  char err_buf[1024];
  int size;
  trace_entry_data_t *edata;
  int fails = 0;

  printf("Trying to get block layer trace events\n");

  fd = open(bdev_name, O_RDWR, 0);
  if (fd == -1) {
    printf("Device opening failed, %s error\n", error_strings[errno]);
    return;
  }

  printf("Get by id number\n");
  if (__check_nomerged_trace_entry(fd, 0, first_dev_sector, SKEY_TYPE_ID, 1, err_buf)) {
    test_pass();
  } else {
    fails++;
    test_fail(err_buf);
  }

  printf("Get by sector\n");
  if (__check_nomerged_trace_entry(fd, 4, 2 + first_dev_sector, SKEY_TYPE_SECTOR, 1, err_buf)) {
    test_pass();
  } else {
    fails++;
    test_fail(err_buf);
  }

  printf("Read events in 2 attempts, check residuary events "
                 "remain available\n");
  /* it must be totally 4 events */
  err = 0;
  size = sizeof(trace_entry_data_t) + 2 * sizeof(trace_event_data_t);
  for (i = 0; (i < 2) && !err; i++) {
    r = get_blkdev_trace_events(fd, 1 + first_dev_sector, SKEY_TYPE_SECTOR, buf, size);
    if (r < 0) {
      fails++;
      test_fail("%s error at events reading\n", error_strings[-r]);
      err = 1;
    } else {
      edata = (trace_entry_data_t*)buf;
      if (edata->nr_events != 2) {
        test_fail("Invalid number of events at %d attemt: %d instead of %d\n",
                  i, edata->nr_events, 2);
        err = 1;
      }
    }
  }
  if (!err)
    test_pass();

  blk_trace_failed = (fails != 0);

  close(fd);
}

void TC_blk_trace_invalid(long *args)
{
  int fd;
  int ret;
  char buf[1024];
  struct key_desc kd;

  fd = open(bdev_name, O_RDWR, 0);
  if (fd == -1) {
    printf("Device opening failed, %s error\n", error_strings[errno]);
    return;
  }

  printf("Try to get already read events, check that they don't remain\n");
  ret = get_blkdev_trace_events(fd, 0, SKEY_TYPE_ID, buf, sizeof(buf));
  check_test_condition(ret, -ESRCH, NULL, true);

  printf("Test trace clearing and try to get events of not traced request\n");
  errno = 0;
  ret = ioctl(fd, BLKIOC_CLEAR_TRACE);
  if (check_test_condition_ext(ret, 0, COMP_EQUAL, 0, NULL, false)) {
    ret = get_blkdev_trace_events(fd, 10, SKEY_TYPE_ID, buf, sizeof(buf));
    check_test_condition(ret, -ESRCH, NULL, true);
  }

  printf("Try to read events in too small buffer\n");
  kd.key = 0;
  kd.key_type = SKEY_TYPE_ID;
  ret = ioctl(fd, BLKIOC_GET_TRACE_EVENTS, 10, &kd, buf);
  check_test_condition_ext(ret, -1, COMP_EQUAL, EINVAL, NULL, true);

  close(fd);
}

static void __test_multiple_mergings(int fd, merge_test_opts_t *test_opts,
                                     check_mergings_fn_t check_fn)
{
  uint32_t buf_size;
  char *buf;
  bool ret;
  int i;

  buf_size = (test_opts->tail_range->start + test_opts->tail_range->len -
              test_opts->front_range->start) << bdev_sect_shift;

  buf = malloc(buf_size);

  for (i = 0; i < buf_size; i++) {
    buf[i] = i & 255;
  }
  ret = __test_mergeable_ioops(fd, test_opts, buf, false, check_fn);
  if (ret) {
    memset(buf, 0, buf_size);
    test_opts->rw = false;
    ret = __test_mergeable_ioops(fd, test_opts, buf, false, check_fn);
    if (ret) {
      /* compare data */
      for (i = 0; i < buf_size; i++) {
        if (buf[i] != (char)(i & 255))
          break;
      }
      if (i == buf_size) {
        test_pass();
      } else {
        test_fail("Data don't tally, i = %d, buf[%d] = %d\n", i, i, buf[i]);
      }
    }
  }

  free(buf);
}

static void TC_iosched_merge_single(long *args)
{
  int fd;
  uint32_t max_size = max_rq_sectors << bdev_sect_shift;
  char *buf;
  sect_range_t front_range, orig_range, tail_range;
  merge_test_opts_t test_opts;

  fd = open(bdev_name, O_RDWR | O_SYNC);
  if (fd == -1) {
    printf("Device opening failed, %s error\n", error_strings[errno]);
    return;
  }

  ioctl(fd, BLKIOC_CLEAR_TRACE);

  buf = malloc(max_size);

  printf("Check for merging at tail of 2 requests to adjacent sectors\n");
  orig_range.start = first_dev_sector;
  orig_range.len = tail_range.len = max_rq_sectors / 2;
  tail_range.start = max_rq_sectors / 2 + first_dev_sector;
  __fill_merge_test_opts(&test_opts, NULL, &orig_range, &tail_range, true, true);
  __test_mergeable_ioops(fd, &test_opts, buf, true, __check_merging_single);

  printf("Check for merging at front of 2 requests to adjacent sectors\n");
  front_range = orig_range;
  orig_range = tail_range;
  __fill_merge_test_opts(&test_opts, &front_range, &orig_range, NULL, true, true);
  __test_mergeable_ioops(fd, &test_opts, buf, true, __check_merging_single);

  printf("Check for merging request twice, at front and at tail in both "
                 "directions and check that data are correct\n");

  front_range.start = first_dev_sector;
  orig_range.start = max_rq_sectors / 4 + first_dev_sector;
  tail_range.start = max_rq_sectors / 2 + first_dev_sector;
  front_range.len = orig_range.len = max_rq_sectors / 4;
  tail_range.len = max_rq_sectors / 4;
  __fill_merge_test_opts(&test_opts, &front_range, &orig_range, &tail_range, true, true);
  __test_multiple_mergings(fd, &test_opts, __check_merging_single);

  free(buf);

  close(fd);
}

static void TC_iosched_merge_clustered(long *args)
{
  int fd;
  sect_range_t front_range, orig_range, tail_range;
  merge_test_opts_t test_opts;
  sector_t base = max_rq_sectors + first_dev_sector;

  printf("Check for merging with clustered request, check that all "
                 "requests will merged in single, and check that data are correct\n");

  fd = open(bdev_name, O_RDWR | O_SYNC);
  if (fd == -1) {
    printf("Device opening failed, %s error\n", error_strings[errno]);
    return;
  }

  ioctl(fd, BLKIOC_CLEAR_TRACE);

  front_range.start = base;
  orig_range.start = base + max_rq_sectors / 4;
  tail_range.start = base + max_rq_sectors / 2;
  front_range.len = orig_range.len = max_rq_sectors / 4;
  tail_range.len = max_rq_sectors / 4;
  __fill_merge_test_opts(&test_opts, &front_range, &orig_range, &tail_range, true, false);
  __test_multiple_mergings(fd, &test_opts, __check_merging_clustered);

  close(fd);
}

static void TC_iosched_merge_exceed(long *args)
{
  int fd;
  sect_range_t front_range, orig_range, tail_range;
  merge_test_opts_t test_opts;
  sector_t base = 4 * max_rq_sectors + first_dev_sector;

  printf("Testing for merging request when product of merging exceeds "
         "maximum request size, and some sectors remain after all merging. "
         "Check it will be merged with 2 requests, and an additional request "
         "will be done; check data were written correctly\n");

  fd = open(bdev_name, O_RDWR | O_SYNC);
  if (fd == -1) {
    printf("Device opening failed, %s error\n", error_strings[errno]);
    return;
  }

  ioctl(fd, BLKIOC_CLEAR_TRACE);

  front_range.start = base;
  orig_range.start = base + 3 * max_rq_sectors / 4;
  tail_range.start = base + 3 * max_rq_sectors / 2;
  front_range.len = tail_range.len = 3 * max_rq_sectors / 4;
  orig_range.len = 3 * max_rq_sectors / 4;
  __fill_merge_test_opts(&test_opts, &front_range, &orig_range, &tail_range, true, false);
  __test_multiple_mergings(fd, &test_opts, __check_merging_exceed);

  printf("Testing for merging request when product of merging exceeds "
         "maximum request size, check it will be totally merged with 2 requests; "
         "check data were written correctly\n");

  base = 8 * max_rq_sectors + first_dev_sector;

  ioctl(fd, BLKIOC_CLEAR_TRACE);

  front_range.start = base;
  orig_range.start = base + 3 * max_rq_sectors / 4;
  tail_range.start = base + 5 * max_rq_sectors / 4;
  front_range.len = tail_range.len = 3 * max_rq_sectors / 4;
  orig_range.len = max_rq_sectors / 2;
  __fill_merge_test_opts(&test_opts, &front_range, &orig_range, &tail_range, true, false);
  __test_multiple_mergings(fd, &test_opts, __check_merging_exceed);

  close(fd);
}

static void TC_iosched_nomerge(long *args)
{
  int fd;
  char *buf = malloc(65536);
  bool r;
  uint32_t nr_sects;

  printf("Do overlapped requests and check they were not be merged\n");

  fd = open(bdev_name, O_RDWR);
  if (fd == -1) {
    printf("Device opening failed, %s error\n", error_strings[errno]);
    return;
  }

  ioctl(fd, BLKIOC_CLEAR_TRACE);
  lseek(fd, 0, SEEK_SET);
  nr_sects = max_rq_sectors / 2;
  write(fd, buf, nr_sects << bdev_sect_shift);
  lseek(fd, nr_sects / 2 << bdev_sect_shift, SEEK_SET);
  write(fd, buf, nr_sects << bdev_sect_shift);

  sleep(1);
  r = __check_nomerged_trace_entry(fd, NO_CHECKED_TRACE_ENTRY_ID, first_dev_sector,
                                   SKEY_TYPE_SECTOR, nr_sects, buf);
  r = r && __check_nomerged_trace_entry(fd, NO_CHECKED_TRACE_ENTRY_ID,
                                        nr_sects / 2 + first_dev_sector,
                                        SKEY_TYPE_SECTOR, nr_sects, buf);

  if (r) {
    test_pass();
  } else {
    test_fail(buf);
  }

  printf("Do requests in different directions and check it will not be "
         "merged\n");
  ioctl(fd, BLKIOC_CLEAR_TRACE);
  lseek(fd, 0, SEEK_SET);
  nr_sects /= 2;
  write(fd, buf, nr_sects << bdev_sect_shift);
  read(fd, buf, nr_sects << bdev_sect_shift);

  sleep(1);
  r = __check_nomerged_trace_entry(fd, NO_CHECKED_TRACE_ENTRY_ID, first_dev_sector,
                                   SKEY_TYPE_SECTOR, nr_sects, buf);
  r = r && __check_nomerged_trace_entry(fd, NO_CHECKED_TRACE_ENTRY_ID,
                                        nr_sects + first_dev_sector,
                                        SKEY_TYPE_SECTOR, nr_sects, buf);
  if (r) {
    test_pass();
  } else {
    test_fail(buf);
  }

  close(fd);
}

static bool tc_must_be_launched(test_function_t fn)
{
  bool ret = true;

  if (fn == TC_read_write_stream || fn == TC_rw_stream_multirequest ||
      fn == TC_read_write_stream_exceed || fn == TC_write_sync) {
    ret = rw_stream_tests;
  } else if (fn == TC_rescan_partitions || fn == TC_unreg_rescan_parts) {
    ret = part_scan_tests && !rescan_failed;
  } else if (fn == TC_read_write_partition) {
    ret = rw_stream_tests && part_scan_tests && rescan_failed;
  } else if (fn == TC_open_close_invalid) {
    ret = open_close_inval_test;
  } else if (fn == TC_blk_trace || fn == TC_blk_trace_invalid) {
    ret = blk_trace_tests && !blk_trace_failed;
  } else if (fn == TC_iosched_merge_single || fn == TC_iosched_merge_clustered ||
             fn == TC_iosched_merge_exceed || fn == TC_iosched_nomerge) {
    ret = iosched_tests && !blk_trace_failed;
  }

  return ret;
}

test_struct_t test_struct = {
  .init = test_init,
  .pause_seconds = 1,
  .must_launched = tc_must_be_launched,
  .cases = {
#if 0
    {
      .name = "open_close",
      .function = TC_open_close,
      .abort_on_fail = true
    },
    {
      .name = "fstat",
      .function = TC_fstat,
      .description = "Test getting block device statistics\n",
      .abort_on_fail = true
    },
    /* stream read/write tests */
    { "read_write_stream", TC_read_write_stream,
      "Testing for stream read/write with different offset and size\n", false, false,
      ARG_LIST { "write offset", "write size", "read offset", "read size", ARG_LIST_END },
      ARGS_CHAIN {
        ARGS { 0, 0, 0, 0 },
        ARGS { 0, 512, 0, 512 },
        ARGS { 512, 512, 512, 250 },
        /* testsing to perform read not sector-aligned request, not full sector */
        ARGS { 1024, 512, 1200, 250 },
        /* testing that the valid amount of data is written */
        ARGS { 2560, 512, 2048, 1536 },
        /* testing to perform write multisector (> 2) not sector-aligned request */
        ARGS { 2500, 800, 2048, 1536 },
        /* testsing to preform read not sector-aligned request, not full sector */
        ARGS { 4096, 256, 4096, 512 },
        /* testing to perform read and write 2-sector requests */
        ARGS { 6144, 600, 6144, 800 },
        ARGS_END
      }
    },
#endif
    {
      "rw_stream_multirequest", TC_rw_stream_multirequest,
      "Testing to read and write when user request size exceeds the "
      "maximum size of a block request\n", false, false
    },
#if 0
    { "read_write_stream_exceed", TC_read_write_stream_exceed,
      "Testing write and read when desired area exceed the disk size\n" },
    { "write_sync", TC_write_sync },
    { "test getting metadata", TC_check_get_mdata, "Test getting block device metadata\n" },
    /* test cases for partition functionality */
    { "rescan_partitions", TC_rescan_partitions,
      "Testing for partitions rescanning, create primary and logical partitions\n", true, false },
    { "read_write_partition", TC_read_write_partition,
      "Testing write and read to a partition, write to end of a partition and chech that "
      "data of the ajacent partition will not affected\n", true, false },
#endif
#if 0       /* uncomment when close will be accessible */
    { "rescan_unreg_parts", TC_unreg_rescan_parts,
      "Testing for partition unregistration and rescanning\n", true, false },
#endif
#if 0
    { "open_close_invalid", TC_open_close_invalid },
    { "ioctl_invalid", TC_ioctl_invalid, false, false },
    { "read_write_stream_invalid", TC_read_write_stream_invalid },
    /* blk trace group of cases */
    { "blk_trace", TC_blk_trace },
    { "blk_trace_invalid", TC_blk_trace_invalid },
    /* io scheduler tests */
    { "iosched_merge_single", TC_iosched_merge_single },
    { "iosched_merge_clustered", TC_iosched_merge_clustered },
    { "iosched_merge_exceed (can fail depending on request size "
      ", scheduling timeout and target platform)", TC_iosched_merge_exceed },
    {"iosched_nomerge", TC_iosched_nomerge},
#endif
    CASE_LIST_END
  }
};
