/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 MString Core Team <http://mstring.berlios.de>
 * (c) Copyright 2009 Michael Tsymbalyuk <mtzaurus@gmail.com>
 *
 * Signal subsystem test.
 *
 */

#include <test/framework.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <ipc/port.h>
#include <ipc/channel.h>
#include <sys/syscall.h>
#include <errno.h>

enum sleeper_action {
  SA_RECV=0,
  SA_SEND=1,
  SA_NANOSLEEP=2,
  SA_SPIN=3,
};

#define NUM_SLEEPERS  4

static struct timespec __sleep_interval = {0,300000000};

#define __sleep() nanosleep(&__sleep_interval,NULL)

static void *__signal_sleeper(void *arg)
{
  int i,r;
  long chid;
  struct timespec ts={100,0};

  switch( (long)arg ) {
    case SA_NANOSLEEP:
      printf("[%d] NANOsleeping forever ...\n",gettid());
      nanosleep(&ts,NULL);
      printf("[%d] Insufficient return from nanosleep !\n");
      for(;;);
    case SA_SEND:
      chid=__syscall3(SC_OPEN_CHANNEL,getpid(),0,IPC_BLOCKED_ACCESS);
      if( chid < 0 ) {
        test_fail("Can't connect to port N 0 ! errno=%d\n",errno);
      } else {
        printf("[%d] Sleeping on SEND (channel=%d)...\n",gettid(),
                       chid);
        r=__syscall5(SC_PORT_SEND,chid,(syscall_arg_t)&i,sizeof(i),(syscall_arg_t)&r,
                     sizeof(r));
        printf("SEND(): insufficient return occured. r=%d\n",
                       r);
      }
      for(;;);
    case SA_RECV:
      printf("[%d] Sleeping on RECV on port 1 ...\n",gettid());
      r=ipc_port_rcvmsg(1,sizeof(i),&i);
      printf("RECV(): insufficient return occured. r=%d\n",r);
      for(;;);
    case SA_SPIN:
      printf("[%d] Spinning ...\n",gettid());
      for(;;);
  }

  return NULL;
}

static void __start_sleepers(void)
{
  long i;
  pthread_t t;

  /* Create 2 IPC ports. */
  for(i=0;i<2;i++) {
    if( ipc_port_open(IPC_BLOCKED_ACCESS,32) != i ) {
      printf("Can't create IPC port N %d: (errno=%d)\n",
                     i,errno);
      test_fail(NULL);
      return;
    }
  }

  for(i=0;i<NUM_SLEEPERS;i++) {
    if( pthread_create(&t,NULL,__signal_sleeper,(void *)i) ) {
      printf("Failed to launch a sleeper N %d !\n",i);
      test_fail(NULL);
      return;
    }
  }
}

static void TC_sigsegv_master(long *args)
{
  int pid,status,chid,i;

  printf("Testing process termination via 'SIGSEGV'  sending to the main thread.\n");

  pid=fork();
  switch( pid ) {
    case 0:
      printf("[%d] Starting sleepers.\n",getpid());
      __start_sleepers();
      __sleep();
      __sleep();
      printf("[%d] Generating page fault.\n",getpid());
      *(char *)0x500=10;
      for(;;);
    case -1:
      test_fail("Can't launch tester !\n");
      break;
    default:
      __sleep();
      printf("Now establish a connection to a tester's port.\n");
      chid=__syscall3(SC_OPEN_CHANNEL,pid,0,IPC_BLOCKED_ACCESS);
      if( chid < 0 ) {
        test_fail("Can't connect to port N 0 ! errno=%d\n",errno);
      } else {
        printf("[%d] Now send a message (waiting for -EPIPE) (channel=%d)...\n",
                       getpid(),chid);
        printf("     Kernel should wake us up as a part of the IPC port shutdown sequence.\n");
        status=__syscall5(SC_PORT_SEND,chid,(syscall_arg_t)&i,sizeof(i),(syscall_arg_t)&i,
                          sizeof(i));
        printf("[%d] Got woken up ! r=%d\n",getpid(),status);
        if( status == -EPIPE ) {
          test_pass();
        } else {
          test_fail("Expected -EPIPE, but another error was returned !\n");
        }
        printf("[%d] Waiting for tester to exit ...\n",getpid());
        waitpid(pid,&status,0);
        printf("[%d] Tester exited successfully !\n",getpid());
        test_pass();
      }
      break;
  }
}

static void *__faulty_sleeper(void *p)
{
  printf("TID [%d] Sleeping for a while to wake up later and generate a fault.\n",
                 gettid());
  __sleep();
  printf("TID [%d] Generating page fault.\n",gettid());
  *(char *)300=10;
  return NULL;
}

static void TC_sigsegv_thread(long *args)
{
  int r,pid,status;
  pthread_t t;
  void *ptr=(void *)0x100000;

  printf("Testing process survival when 'SIGSEGV' is sent to a thread other than the main thread.\n");

  pid=fork();
  switch( pid ) {
    case 0:
      printf("[%d] Starting 1 sleepers\n",getpid());
      if( pthread_create(&t,NULL,__faulty_sleeper,NULL) ) {
        printf("[%d] Can't create a sleeper\n",getpid());
        test_fail(NULL);
      } else {
        printf("Waiting for the faulter thread to exit ...\n");
        r=pthread_join(t,&ptr);
        printf("Thread exited ! exit ptr=%p (0x100000 means test faulure)\n");
        if( (ptr) == (void *)0x100000 ) {
          test_fail("Insufficient thread exit value !");
        } else {
          test_pass();
        }
      }
      __syscall1(SC_EXIT,0);
    case -1:
      test_fail("Can't launch tester !\n");
      break;
    default:
      printf("Waiting for tester to exit ...\n");
      waitpid(pid,&status,0);
      printf("Tester exited successfully !\n");
      test_pass();
      for(;;);
      break;
  }
}

test_struct_t test_struct = {
  .pause_seconds = 1,
  .cases = {
    {
      .name = "sigsegv_root_thread",
      .function = TC_sigsegv_master,
      .abort_on_fail = true,
      .arg_list = NULL,
    },
    {
      .name = "sigsegv_thread",
      .function = TC_sigsegv_thread,
      .abort_on_fail = true,
      .arg_list = NULL,
    },
    CASE_LIST_END
  }
};
