/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008,2009 MadTirra <madtirra@jarios.org>
 *
 * Jari OS specific defines and functions for GNU coreutils
 *
 */

#ifndef __SYSTEM_H__
#define __SYSTEM_H__

#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "intprops.h"
#include "lib/version-etc.h"

#define SAFE_READ_ERROR  (ssize_t)-1

#define TYPE_SIGNED(t) (! ((t) 0 < (t) -1))

#define O_BINARY  0

#define OFF_T_MAX TYPE_MAXIMUM (off_t)

/* Return zero if T can be determined to be an unsigned type.
   Otherwise, return 1.
   When compiling with GCC, INT_STRLEN_BOUND uses this macro to obtain a
   tighter bound.  Otherwise, it overestimates the true bound by one byte
   when applied to unsigned types of size 2, 4, 16, ... bytes.
   The symbol signed_type_or_expr__ is private to this header file.  */
#if __GNUC__ >= 2
# define signed_type_or_expr__(t) TYPE_SIGNED (__typeof__ (t))
#else
# define signed_type_or_expr__(t) 1
#endif

/* Bound on length of the string representing an integer type or expression T.
   Subtract 1 for the sign bit if T is signed; log10 (2.0) < 146/485;
   add 1 for integer division truncation; add 1 more for a minus sign
   if needed.  */
#define INT_STRLEN_BOUND(t) \
  ((sizeof (t) * CHAR_BIT - signed_type_or_expr__ (t)) * 146 / 485 \
   + signed_type_or_expr__ (t) + 1)

#define INT_BUFSIZE_BOUND(t) (INT_STRLEN_BOUND (t) + 1)

/* some getopt defines */
#define GETOPT_HELP_OPTION_DECL  "help", no_argument, NULL, 'h'
#define GETOPT_VERSION_OPTION_DECL  "version", no_argument, NULL, 'V'

#define case_GETOPT_HELP_CHAR case 'h': \
  usage(EXIT_SUCCESS); \
  break

#define case_GETOPT_VERSION_CHAR(pname,authors) case 'V':		\
  version_etc(stdout,pname,"Jari OS GNU coreutils",Version,authors, NULL); \
  exit(EXIT_SUCCESS);							\
  break

#define no_argument        0
#define required_argument  1

#define IF_LINT(Code) Code

#define ptr_align(s,a)    ((((uintptr_t)s)+((a)-1)) & ~((a)-1))

/* file io stuff */
ssize_t safe_read(int fd,void *buf,ssize_t buf_size);
ssize_t full_write(int fd,void *buf,ssize_t n);
size_t io_blksize(struct stat fstat);

/* memory related functions */
size_t getpagesize(void);
void *xmalloc(size_t size);

/* reporting errors and misc stuff */
void error(int status,int errnum,const char *message,...);
const char *proper_name(const char *nm);

#include "lib/dirname.h"
#include "lib/openat.h"

static inline bool
dot_or_dotdot (char const *file_name)
{
  if (file_name[0] == '.')
    {
      char sep = file_name[(file_name[1] == '.') + 1];
      return (! sep || ISSLASH (sep));
    }
  else
    return false;
}

/* A wrapper for readdir so that callers don't see entries for `.' or `..'.  */
static inline struct dirent const *
readdir_ignoring_dot_and_dotdot (DIR *dirp)
{
  while (1)
    {
      struct dirent const *dp = readdir (dirp);
      if (dp == NULL || ! dot_or_dotdot (dp->d_name))
	return dp;
    }
}

/* Return true if DIR is determined to be an empty directory.  */
static inline bool
is_empty_dir (int fd_cwd, char const *dir)
{
  DIR *dirp;
  struct dirent const *dp;
  int saved_errno;
  int fd = openat (fd_cwd, dir,
		   (O_RDONLY | O_DIRECTORY
		    | O_NOCTTY | O_NONBLOCK));

  if (fd < 0)
    return false;

  dirp = fdopendir (fd);
  if (dirp == NULL)
    {
      close (fd);
      return false;
    }

  errno = 0;
  dp = readdir_ignoring_dot_and_dotdot (dirp);
  saved_errno = errno;
  closedir (dirp);
  if (dp != NULL)
    return false;
  return saved_errno == 0 ? true : false;
}

static inline unsigned char to_uchar (char ch) { return ch; }

#endif
