#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <getopt.h>
#include <errno.h>
#include <unistd.h>

#include <sys/fslinker.h>

#include <sexpr/sexp.h>
#include <sexpr/sexp_ops.h>
#include <josconf/josconf.h>

#include "fslink.h"

static const char *fslink_opts = "o:t:c:";
static char *progname;

static inline void fslink_usage(void)
{
  fprintf(stderr, "Usage: %s [-o options] [-c cmd] <link subject> <link point>\n"
	  "\t-t <type>        - FS type.\n"
	  "\t-o <options>     - "
	  "Comma separated string of options for fslink target.\n"
	  "\t-c <link|unlink> - link/unlink command.\n", progname);
  exit(EXIT_FAILURE);
}

static inline int __alloc_config_elem(sexp_t *expr, int elem_n,
				      char **data, atom_t type);
static fs_link_entry_t *__process_fsl_entry(sexp_t *sx,
					    const char *fstype,
					    const char *path1,
					    const char *path2,
					    const char *options);
static int process_config_file(fs_link_entry_t **entry,
			       const char *fstype, const char *path1,
			       const char *path2, const char *options);

int main(int argc, char *argv[])
{
  int opt, i, ret;
  char *options, *pathp[2], *fstype;
  fs_link_req_t request;
  fs_link_entry_t *entry;

  options  = NULL;
  fstype   = NULL;
  pathp[0] = NULL;
  pathp[1] = NULL;
  progname = argv[0];
  request  = FSLINK_REQ_MOUNT;

  while ((opt = getopt(argc, argv, fslink_opts)) != -1) {
    switch (opt) {
    case 'o':
      options = optarg;
      break;
    case 'c':
      /* TODO: */
      break;
    case 't':
      fstype = optarg;
      break;
    case ':':
      return EXIT_FAILURE;
    default:
      fprintf(stderr, "%s: unknown option -%c\n", progname, opt);
    }
  }

  switch (argc - optind) {
  case 1:
  case 2:
    break;
  default:
    fprintf(stderr, "Set atleast one link point.\n");
    fslink_usage();
    return EXIT_FAILURE;
  }

  for (i = 0 ; optind < argc; optind++, i++) {
    pathp[i] = argv[optind];
    if (access(argv[optind], F_OK))
      fprintf(stderr, "Can't access to '%s'.\n", argv[optind]);
  }

  /* Get necessary information from config file */
  if (!fstype || !pathp[1]) {
    if (!process_config_file(&entry, fstype, pathp[0], pathp[1], options)) {
      fprintf(stderr, "Can't get configuration information. "
	      "Please be more specific.\n");
      return EXIT_FAILURE;
    }
  } else {
    entry = fslink_fill_struct(fstype, pathp[0], pathp[1], options);
    if (!entry) {
      fprintf(stderr,
	      "Can't assemble request information. Error = (%d).\n", errno);
      return EXIT_FAILURE;
    }
  }

  fprintf(stdout, "Structure assembled successfully.\n"
	  "FS type = '%s'\n", entry->fstype);

  if (entry->options)
    fprintf(stdout, "Options = '%s'\n", entry->options);
  else
    fprintf(stdout, "Options = 'NONE'\n");

  fprintf(stdout, "PP1: %d; %d; %s\n",
	  entry->mounts.plen, entry->mounts.p_num, entry->mounts.path);
  fprintf(stdout, "PP2: %d; %d; %s\n",
	  entry->mountp.plen, entry->mountp.p_num, entry->mountp.path);

  ret = fslink_send_request(entry, request);
  if (ret) {
    fprintf(stderr, "Error accured while sending request = %d.\n", ret);
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}

static inline int __alloc_config_elem(sexp_t *expr, int elem_n,
				      char **data, atom_t type)
{
  sexp_t *sx;

  if (sexp_list_get_element(expr, &sx, elem_n) ||
      sexp_get_value_alloc(sx, type, data))
    return 1;

  return 0;
}

static fs_link_entry_t *__process_fsl_entry(sexp_t *sx,
					    const char *fstype,
					    const char *path1,
					    const char *path2,
					    const char *options)
{
  /* TODO: Options */
  fs_link_entry_t *entry;
  char *c_fstype, *c_mounts, *c_mountp;

  c_fstype = c_mounts = c_mountp = NULL;
  entry = NULL;

  /*
    char pstr[1024];
    print_sexp(pstr, 1024, sx);
    fprintf(stdout, "line of config: %s\n", pstr);
  */

  if (__alloc_config_elem(sx, 1, &c_mounts, SEXP_DQUOTE) ||
      __alloc_config_elem(sx, 2, &c_mountp, SEXP_DQUOTE))
    goto __process_entry_fail1;

  if (path1 && path2) {
    if (strcmp(c_mounts, path1) || strcmp(c_mountp, path2))
      goto __process_entry_fail1;
  } else
    if (strcmp(c_mounts, path1) && strcmp(c_mountp, path1))
      goto __process_entry_fail1;

  if (__alloc_config_elem(sx, 0, &c_fstype, SEXP_BASIC))
    goto __process_entry_fail1;

  if (fstype && strcmp(c_fstype, fstype))
    goto __process_entry_fail1;

  entry = malloc(sizeof(*entry));
  if (!entry)
    goto __process_entry_fail1;

  memset(entry, 0, sizeof(*entry));

  /* FS type information */
  entry->fstype     = c_fstype;
  entry->fstype_len = strlen(c_fstype);

  /* Link path structures allocation */
  if (fslink_pp_fill(c_mounts, &entry->mounts) ||
      fslink_pp_fill(c_mountp, &entry->mountp))
    goto __process_entry_fail2;

  /* TODO: Options */
  if (!options) {
  }

  return entry;

 __process_entry_fail2:

  fslink_free_pp(&entry->mounts);
  fslink_free_pp(&entry->mountp);
  free(entry);

 __process_entry_fail1:

  free(c_mounts);
  free(c_mountp);
  free(c_fstype);

  return NULL;
}

static int process_config_file(fs_link_entry_t **entry,
			       const char *fstype, const char *path1,
			       const char *path2, const char *options)
{
  FILE *fh;
  sexp_t *sx;
  fs_link_entry_t *bufentry;
  char conf_str[BUFSIZ];

  fh = fopen(CONFFILE, "r");
  if (!fh) return -1;

  bufentry = NULL;

  while (1) {
    fgets(conf_str, BUFSIZ, fh);
    if (feof(fh) != 0) break;

    sx = parse_sexp(conf_str, BUFSIZ);
    if (!sx || sexp_list_length(sx) != FSL_CONF_ELEMS)
      continue;

    bufentry = __process_fsl_entry(sx, fstype, path1, path2, options);
    if (bufentry) {
      destroy_sexp(sx);
      *entry = bufentry;
      return 1;
    }

    destroy_sexp(sx);
  }

  fclose(fh);

  return 0;
}
