#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <getopt.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>

#include <pci/pci.h>
#include <pci/names.h>

#define PRINT_RESOURCES 0x01
#define PRINT_IRQLINES  0x02
#define PRINT_CAPS      0x04
#define PRINT_NAMES     0x08
#define PRINT_CLASS     0x10

#define PRINT_ALL       0xFF

static const char *valid_opts = "nsarchiv:p:";

static pci_ident_t ident;
static int options = 0;

static void print_usage(void)
{
  printf("PCI bud device lister.\n"
         "  Usage: lspci [options] [-v vendor] [-p product]\n"
         "  vendor and product must be four digit hexadecimals\n"
         "  -r    show resources\n"
         "  -i    show interrupt line\n"
         "  -c    show capabilities\n"
         "  -n    show vendor and product names\n"
         "  -s    show device class\n"
         "  -a    show all above\n"
         "  example: lspci -a -v 8086 -p 1237\n" );
}

#define GB 1024L * 1024L * 1024L
#define MB 1024L * 1024L
#define KB 1024L
static void human_print(ulong_t val)
{
  char *suffix;
  if(val >= 3 * GB) {
    val /= GB;
    suffix = " Gbyte";
  } else if(val >= 3 * MB) {
    val /= MB;
    suffix = " Mbyte";
  } else if(val >= 3 * KB) {
    val /= 1024;
    suffix = " Kbyte";
  } else {
    suffix = " byte";
  }
  printf("%lu%s", val, suffix);
  if(val != 1) {
    printf("s");
  }
}

static void print_names(const pci_dev_t *dev)
{
  vendor_desc_t *vendor;
  product_desc_t *product;
  subprod_desc_t *subprod;

  vendor = lookup_vendor(dev->ident.vendor_id);
  if(!vendor) {
    printf("\tUnknown vendor/product: 0x%x:0x%x\n",
           dev->ident.vendor_id, dev->ident.device_id);
    goto out;
  };
  printf("\tVendor: %s\n", vendor->name);

  product = lookup_product(vendor, dev->ident.device_id);
  if(!product) {
    printf("\tUnknown product: 0x%x\n", dev->ident.device_id);
    goto out;
  }
  printf("\tProduct: %s\n", product->name);
 
  subprod = lookup_subprod(product, dev->ident.subsys_id, dev->ident.subsys_vendor_id);
  if(!subprod) {
    goto out;
  }
  printf("\tSubproduct: %s\n", subprod->name);
  
  out:
  return;
}

static void print_class(const pci_dev_t *dev)
{
  class_desc_t *class;
  subclass_desc_t *subclass;
  prog_if_desc_t *prog_if;
  
  class = lookup_class(dev->ident.class_id);
  if(!class) {
    printf("\tUnknown device class: 0x%x\n", dev->ident.class_id);
    goto out;
  }
  printf("\tClass: %s\n", class->name);
  
  subclass = lookup_subclass(class, dev->ident.subclass_id);
  if(!subclass) {
    printf("\tUnknown device subclass: 0x%x\n", dev->ident.subclass_id);
    goto out;
  }
  printf("\tSubclass: %s\n", subclass->name);
  
  prog_if = lookup_prog_if(subclass, dev->ident.prog_if_id);
  if(!prog_if) {
    printf("\tUnknown programming interface: 0x%x\n", dev->ident.prog_if_id);
    goto out;
  }
  printf("\tProgramming interface: %s\n", prog_if->name);

  out:
  return;
}

static void print_resources(const pci_dev_t *dev)
{
  int i;
  if(dev->num_resources <= 0) {
    printf("\tNo resources");
  } else {
    for(i = 0; i < dev->num_resources; i++) {
      ulong_t len;
      //some paranoia
      if(i > 5) {
        continue;
      }
      len = dev->resources[i].end - dev->resources[i].start + 1;
      if(dev->resources[i].flags & PCI_RESOURCE_IO) {
        printf("\tResource %d (I/O ports):\n", i);
        printf("\t\t  start: %x\n", dev->resources[i].start);
        printf("\t\t    end: %x\n", dev->resources[i].end);
        printf("\t\t length: %lu ports\n", len);
      } else if(dev->resources[i].flags & PCI_RESOURCE_MEM64) {
        printf("\tResource %d (memory, 64 bit address):\n", i);
        printf("\t\t  start: %lx\n", dev->resources[i].start);
        printf("\t\t    end: %lx\n", dev->resources[i].end);
        printf("\t\t length: ");
        human_print(len);
        printf("\n");
      } else if(dev->resources[i].flags & PCI_RESOURCE_MEM) {
        printf("\tResource %d (memory, 32 bit address):\n", i);
        printf("\t\t  start: %x\n", dev->resources[i].start);
        printf("\t\t    end: %x\n", dev->resources[i].end);
        printf("\t\t length: ");
        human_print(len);
        printf("\n");
      }
    }
  }  
}

static void print_capabilities(const pci_dev_t *dev)
{
  int i;
  
  if(dev->num_caps <= 0) {
    printf("\tNo capabilities.\n");
  } else {
    printf("\tDevice capabilities:\n");
    for(i = 0; i < dev->num_caps; i++) {
      switch(dev->capabilities[i].cap_id) {
        case PCI_CAP_ID_PM:
          printf("\t\tPower Management\n");
          break;
        case PCI_CAP_ID_AGP:
          printf("\t\tAccelerated Graphics Port\n");
          break;
        case PCI_CAP_ID_VPD:
          printf("\t\tVital Product Data\n");
          break;
        case PCI_CAP_ID_SLOTID:
          printf("\t\tSlot Identification\n");
          break;
        case PCI_CAP_ID_MSI:
          printf("\t\tMessage Signalled Interrupts\n");
          break;
        case PCI_CAP_ID_CHSWP:
          printf("\t\tCompactPCI HotSwap\n");
          break;
        case PCI_CAP_ID_PCIX:
          printf("\t\tPCI-X\n");
          break;
        case PCI_CAP_ID_HT:
          printf("\t\tHyperTransport\n");
          break;
        case PCI_CAP_ID_VNDR:
          printf("\t\tVendor Specific\n");
          break;
        case PCI_CAP_ID_DBG:
          printf("\t\tDebug Port\n");
          break;
        case PCI_CAP_ID_CCRC:
          printf("\t\tCompactPCI Central Resource Control\n");
          break;
        case PCI_CAP_ID_SHPC:
          printf("\t\tPCI Standard Hot-Plug Controller\n");
          break;
        case PCI_CAP_ID_SSVID:
          printf("\t\tBridge subsystem vendor/device ID\n");
          break;
        case PCI_CAP_ID_AGP3:
          printf("\t\tAGP Target PCI-PCI bridge\n");
          break;
        case PCI_CAP_ID_EXP:
          printf("\t\tPCI-Express\n");
          break;
        case PCI_CAP_ID_MSIX:
          printf("\t\tMSI-X\n");
          break;
        default:
          printf("\t\tUnknown capability, ID: 0x%x [o.k. if you are in emulator]\n", dev->capabilities[i].cap_id);
          break;
      }
    }
  }
}

static void print_device(const pci_dev_t *dev)
{
  printf("Bus %hhu, device %hhu, function %hhu;  ", dev->addr.bus, dev->addr.device, dev->addr.function);
  printf("Vendor ID: 0x%4x; Device ID: 0x%4x\n", dev->ident.vendor_id, dev->ident.device_id);

  if(options & PRINT_NAMES) {
    print_names(dev);
  }

  if(options & PRINT_CLASS) {
    print_class(dev);
  }
  
  if(options & PRINT_RESOURCES) {
    print_resources(dev);
  }
  
  if(options & PRINT_IRQLINES) {
    printf("\tInterrupt line: %d\n", dev->irq_line);
  }

  if(options & PRINT_CAPS) {
    print_capabilities(dev);
  }
}

static void do_lookup(void)
{
  int nd;
  pci_addr_t *devs;
  pci_dev_t pdev;
  int ret, k;

  nd = pci_lookup(&ident, NULL, 0);
  if(nd < 0) {
    fprintf(stderr, "PCI access failed, errno = %d.\n", errno);
    exit(1);
  }

  if(nd == 0) {
    printf("No devices.\n");
    return;
  }

  devs = malloc(nd * sizeof(*devs));
  if(!devs) {
    fprintf(stderr, "Not enough memory.\n");
    exit(1);
  }

  ret = pci_lookup(&ident, devs, nd);

  if(ret != nd) {
    goto cleanup;
  }

  for(k = 0; k < nd; k++) {
    
    ret = pci_get_dev(&devs[k], &pdev);
    if(ret != 0) {
      printf("Failed to get device, errno = %d\n", errno);
      goto cleanup;
    }
    
    print_device(&pdev);
  }

  cleanup:
  free(devs);
  exit(1);
}
  
  
int main(int argc, char **argv)
{
  int opt;
  int val;
  char *endptr;
  
  memset(&ident, 0, sizeof(ident));

  while((opt = getopt(argc, argv, valid_opts)) != -1) {
    switch(opt) {
      case 'v':
      case 'p':
        val = strtoul(optarg, &endptr, 16);
        if(endptr != optarg + 4) {
          fprintf(stderr, "Failed to parse number, \"%x\" use -h for help.\n", val);
          return -1;
        }
        if(opt == 'p') {
          ident.device_id = val;
          ident.lookup_opt |= PCI_LOOKUP_P;
        } else {
          ident.vendor_id  = val;
          ident.lookup_opt |= PCI_LOOKUP_V;
        }
        break;
      case 'r':
        options |= PRINT_RESOURCES;
        break;
      case 'i':
        options |= PRINT_IRQLINES;
        break;
      case 'c':
        options |= PRINT_CAPS;
        break;
      case 'n':
        options |= PRINT_NAMES;
        break;
      case 's':
        options |= PRINT_CLASS;
        break;
      case 'h':
        print_usage();
        return 0;
      case 'a':
        options |= PRINT_ALL;
        break;
      default:
        fprintf(stderr, " use -h for help.\n");
        return -1;
    }
  }

  printf("\n");
  
  do_lookup();

  return 0;
}
