#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <stdarg.h>
#include <sys/mman.h>
#include <sys/syscall.h>
#include <wchar.h>

#include <sys/ioport.h>
#include <vm86/x86emu.h>
#include <int10/int10.h>

#include "internal.h"

char msg[1024];
wchar_t wbuf[1024];

//extern void (*printk)(const char*, ...);
#define printk(...)

static uint64_t portmask[1024];

#define PORT_BIT(n)(portmask[(n)/64] & ((1ULL<<63)>>((n)%64)))

//For next three functions n is first port and m is ports number
static inline void set_port_bits(uint16_t n, uint16_t m)
{
  int i;               
  for(i = 0; i < m; i++) {
    //printk("(%d)", n+i);
    portmask[(n+i)/64] |= (1ULL << 63) >> ((n+i)%64);
  }
}

static int alloc_ports(uint16_t n, uint16_t m) {
  uint32_t i, j;
  int r;
  
  for(i = 0; i < m; i++ ) {
    if(!PORT_BIT(n+i)) {
      for(j = i+1; j < m; j++) 
        if(PORT_BIT(j)) break;
      r = alloc_ioport_range(n+i, j-i);
      if(r) return r;
      set_port_bits(n+i, j-i);
      if(r != 0) return r;
      i = j;
    }
  }
  return 0;
}

static void free_ports(void)
{
  uint32_t i, j;
  int r;

  for(i = 0; i < 65536; i++) {
    if(PORT_BIT(i)) {
      for(j = i+1; j < 65536; j++) {
        if(!PORT_BIT(j)) break;
      }
      r = free_ioport_range(i, j-i);
      i = j; // We need not j-1 as jth port is surely free.
    }
  }
}

uint8_t *pages[EMU_PAGES];

static uint8_t X86API my_inb(uint16_t addr);
static uint16_t X86API my_inw(uint16_t addr);
static uint32_t X86API my_inl(uint16_t addr);
static void X86API my_outb(uint16_t addr, uint8_t val);
static void X86API my_outw(uint16_t addr, uint16_t val);
static void X86API my_outl(uint16_t addr, uint32_t val);
static uint8_t X86API my_rdb(uint32_t addr);
static uint16_t X86API my_rdw(uint32_t addr);
static uint32_t X86API my_rdl(uint32_t addr);
static void X86API my_wrb(uint32_t addr, uint8_t val);
static void X86API my_wrw(uint32_t addr, uint16_t val);
static void X86API my_wrl(uint32_t addr, uint32_t val);

X86EMU_pioFuncs my_pioFuncs =
{
  my_inb,
  my_inw,
  my_inl,
  my_outb,
  my_outw,
  my_outl
};

X86EMU_memFuncs my_memFuncs =
{
  my_rdb,
  my_rdw,
  my_rdl,
  my_wrb,
  my_wrw,
  my_wrl
};

static uint8_t X86API my_inb(uint16_t addr)
{
  if(alloc_ports(addr, 1) != 0) {
    printk("port not allocated: %hu\n", addr);
    abort();
  }

  return inb(addr);
}
static uint16_t X86API my_inw(uint16_t addr)
{
  int ret;
  if((ret = alloc_ports(addr, 2)) != 0) {
    printk("port not allocated, port: %hu, ret = %d\n", addr, ret);
    abort();
  }

  return inw(addr);
}

static uint32_t X86API my_inl(uint16_t addr)
{
  if(alloc_ports(addr, 4) != 0) {
    printk("port not allocated\n");
    abort();
  }

  return inl(addr);
}

static void X86API my_outb(uint16_t addr, uint8_t val)
{
  if(alloc_ports(addr, 1) != 0) {
    printk("port not allocated\n");
    abort();
  }

  outb(addr, val);
}
static void X86API my_outw(uint16_t addr, uint16_t val)
{
  if(alloc_ports(addr, 2) != 0) {
    printk("port not allocated\n");
    abort();
  }

  outw(addr, val);
}

static void X86API my_outl(uint16_t addr, uint32_t val)
{
  if(alloc_ports(addr, 4) != 0) {
    printk("port not allocated\n");
    abort();
  }

  outl(addr, val);
}

static uint8_t X86API my_rdb(uint32_t addr)
{
  return *(uint8_t*)(pages[addr/PAGE_SIZE] + addr%PAGE_SIZE);
}

static uint16_t X86API my_rdw(uint32_t addr)
{
  return *(uint16_t*)(pages[addr/PAGE_SIZE] + addr%PAGE_SIZE);
}

static uint32_t X86API my_rdl(uint32_t addr)
{
  return *(uint32_t*)(pages[addr/PAGE_SIZE] + addr%PAGE_SIZE);
}

static void X86API my_wrb(uint32_t addr, uint8_t val)
{
  *(uint8_t*)(pages[addr/PAGE_SIZE] + addr%PAGE_SIZE) = val;
}

static void X86API my_wrw(uint32_t addr, uint16_t val)
{
  *(uint16_t*)(pages[addr/PAGE_SIZE] + addr%PAGE_SIZE) = val;
}

static void X86API my_wrl(uint32_t addr, uint32_t val)
{
  *(uint32_t*)(pages[addr/PAGE_SIZE] + addr%PAGE_SIZE) = val;
}

int map_phys_pages(void)
{
  int i;

  //Map first page, int vectors
  pages[0] = mmap(NULL, 4096, PROT_READ|PROT_WRITE, MAP_ANON|MAP_PHYS|MAP_PRIVATE, NOFD, 0);
  if(pages[0] == MAP_FAILED) {
    printk("Failed to map first page\n");
    return -1;
  }

  //Map VGA bios
  pages[VGA_BIOS_PAGE] = mmap(NULL, VGA_BIOS_LEN, PROT_READ|PROT_WRITE, MAP_ANON|MAP_PHYS|MAP_PRIVATE, NOFD, VGA_BIOS_START);
  if(pages[VGA_BIOS_PAGE] == MAP_FAILED) {
    printk("Failed to map VGA BIOS pages, errno = %d\n", errno);
    return -1;
  }


  for(i = 1; i < VGA_BIOS_PAGES; i++) {
    pages[VGA_BIOS_PAGE + i] = pages[VGA_BIOS_PAGE] + PAGE_SIZE*i;
  }

  //Map system bios
  pages[SYS_BIOS_PAGE] = mmap(NULL, SYS_BIOS_LEN, PROT_READ|PROT_WRITE, MAP_ANON|MAP_PHYS|MAP_PRIVATE, NOFD, SYS_BIOS_START);
  if(pages[SYS_BIOS_PAGE] == MAP_FAILED) {
    printk("Failed to map system BIOS pages");
    return -1;
  }
  for(i = 1; i < SYS_BIOS_PAGES; i++) {
    pages[SYS_BIOS_PAGE + i] = pages[SYS_BIOS_PAGE] + PAGE_SIZE*i;
  }

  //Map one page for "int 10" code

  pages[1] = mmap(NULL, 4096, PROT_READ|PROT_WRITE, MAP_ANON|MAP_PRIVATE, NOFD, 0);
  if(pages[1] == MAP_FAILED) {
    printk("Failed to map one page for code" );
    return -1;
  }

  return 0;
}

void unmap_phys_pages(void) {
  int i, j;

  for(i = 0; i < EMU_PAGES; i++) {
    if(pages[i]) {
      for(j = i+1; j < EMU_PAGES; j++) {
        if(!pages[j]) break;
      }
      munmap(pages[i], PAGE_SIZE * (j - i));
      i = j;
    }
  }
}

int int10_init(void)
{
  if(map_phys_pages() != 0) {
    printk("int10:Failed to initialize physical memory\n");
    unmap_phys_pages();
    return -1;
  } else {
    X86EMU_setupPioFuncs(&my_pioFuncs);
    X86EMU_setupMemFuncs(&my_memFuncs);
  }

  *pages[1]     = 0xCD; // int
  *(pages[1]+1) = 0x10; // 10
  *(pages[1]+2) = 0xF4; // hlt

  memset(portmask, 0, sizeof(portmask));
  
  return 0;
}

int int10_done(void)
{
  printk("Initing\n");
  free_ports();
  unmap_phys_pages();
  return 0;
}

int int10_set_vbe_mode(int mode, vbe_ctrl_t **ctrl_info, vbe_mode_t **mode_info)
{
  uint32_t phys;
  char *oem_string;
  /* Get VESA BIOS controller information. */
  M.x86.R_EAX  = 0x4F00;        /* int10 function number */
  M.x86.R_EDI  = 0x100;         /* ES:DI pointer to at least 512 bytes */
  M.x86.R_IP   = 0;             /* address of "int10; hlt" */
  M.x86.R_SP   = PAGE_SIZE;     /* SS:SP pointer to stack */
  M.x86.R_CS   = PAGE_SIZE >> 4;
  M.x86.R_DS   = PAGE_SIZE >> 4;
  M.x86.R_ES   = PAGE_SIZE >> 4;
  M.x86.R_SS   = PAGE_SIZE >> 4;
  X86EMU_exec();
  
  *ctrl_info = (vbe_ctrl_t*)(pages[1] + 0x100);
  if(M.x86.R_AX != 0x4F) {
    printk("No vesa bios found");
    return -ENODEV;
  } 

  phys = ((*ctrl_info)->oem_string & 0x0FFFF)  + (((*ctrl_info)->oem_string >> 12) & 0xFFFF0);
  oem_string = (char*)(pages[phys/PAGE_SIZE] + (phys % PAGE_SIZE));
  printk("Found VESA BIOS version %d.%d\nOEM %p\n", (int)((*ctrl_info)->version >> 8),
               (int)((*ctrl_info)->version &  0xFF), (*ctrl_info)->oem_string);
  if ((*ctrl_info)->version < 0x0200)
    {
      printk("VESA BIOS 2.0 or later required.\n");
      return -EINVAL;
    }

  /* Get mode information. */
  M.x86.R_EAX  = 0x4F01;        /* int10 function number */
  M.x86.R_ECX  = mode;          /* VESA mode */
  M.x86.R_EDI  = 0x800;         /* ES:DI pointer to at least 256 bytes */
  M.x86.R_IP   = 0;             /* address of "int10; hlt" */
  M.x86.R_SP   = PAGE_SIZE;   /* SS:SP pointer to stack */
  M.x86.R_CS   =
  M.x86.R_DS   =
  M.x86.R_ES   =
  M.x86.R_SS   = PAGE_SIZE >> 4;
  X86EMU_exec();

  *mode_info = (vbe_mode_t*)(pages[1] + 0x800);
  if (M.x86.R_AX != 0x004F ||
      ((*mode_info)->mode_attributes & 0x0091) != 0x0091)
    {
      printk("Mode %x not supported\n", mode);
      return -ERANGE;
    }

  /* Switch mode. */
  M.x86.R_EAX  = 0x4F02;        /* int10 function number */
  M.x86.R_EBX  = mode & 0xf7ff; /* VESA mode; use current refresh rate */
  M.x86.R_EBX |= (1<<14);       /* use flat buffer model */
  M.x86.R_IP   = 0;             /* address of "int10; hlt" */
  M.x86.R_SP   = PAGE_SIZE;   /* SS:SP pointer to stack */
  M.x86.R_CS   = PAGE_SIZE >> 4;
  M.x86.R_DS   = PAGE_SIZE >> 4;
  M.x86.R_ES   = PAGE_SIZE >> 4;
  M.x86.R_SS   = PAGE_SIZE >> 4;
  X86EMU_exec();
  
  printk("VBE mode 0x%x successfully set.\n", mode);
  return 0;
}

int int10_pan(unsigned *x, unsigned *y)
{
  printk("Trying execution of ``set display start'' using x86emu\n");

  /* Get VESA BIOS controller information. */
  M.x86.R_EAX  = 0x4F07;	/* int10 function number */
  M.x86.R_BL   = 0x00;          /* set display start */
  M.x86.R_CX   = *x;            /* first displayed pixel in scanline */
  M.x86.R_DX   = *y;            /* first displayed scanline */
  M.x86.R_IP   = 0;		/* address of "int10; hlt" */
  M.x86.R_SP   = PAGE_SIZE;	/* SS:SP pointer to stack */
  M.x86.R_CS   =
  M.x86.R_DS   =
  M.x86.R_ES   =
  M.x86.R_SS   = PAGE_SIZE >> 4;
  X86EMU_exec();

  if (M.x86.R_AX != 0x004F)
    {
      printk("Panning to %d,%d not supported\n", *x, *y);
      return -EINVAL;
    }

  printk("Successful.\n");
  return 0;
}
