/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Graphov Paul <graphov@jarios.org>
 * (c) Copyright 2009 Dmitry Gromada <gromada@jarios.org>
 *
 * libpci.c: basic pci device functions.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <pci.h>

#include "internal.h"

static int pcifd = -1;

#define PCI_PATH "/dev/pci"

int pci_init(void)
{
  if(pcifd >= 0) {
    return 0;
  }
  pcifd = open(PCI_PATH, O_RDONLY);
  if(pcifd < 0) {
    return -1;
  } else {
    return 0;
  }
}

void pci_close(void)
{
  close(pcifd);
  pcifd = -1;
}

MK_READ_FUNC(uint8_t, 1, byte);
MK_READ_FUNC(uint16_t, 2, word);
MK_READ_FUNC(uint32_t, 4, dword);

MK_WRITE_FUNC(uint8_t, 1, byte);
MK_WRITE_FUNC(uint16_t, 2, word);
MK_WRITE_FUNC(uint32_t, 4, dword);

int pci_lookup(pci_ident_t *ident, pci_addr_t *buf, int size)
{
  pci_addr_t *addrbuf = (pci_addr_t *)buf;
  int ret;
  pci_ident_t empty;
  
  pci_init();

  if(ident == NULL) {
    ident = &empty;
    memset(ident, 0, sizeof(*ident));
  }
  
  if(pcifd < 0) {
    errno = EBADF;
    return -1;
  }
  
  if(buf) {
    if(size < 1) {
      errno = EINVAL;
      return -1;
    }
    ret = ioctl(pcifd, PIOC_LOOKUP, size * (sizeof(*addrbuf)), ident, addrbuf);
    if(ret < 0) {
      return ret;
    } else {
      return ret / sizeof(*addrbuf);
    }
  } else {
    return ioctl(pcifd, PIOC_LOOKUP_CNT, ident);
  }
}

int pci_lookup_prod_vend(uint16_t vendor, int product, pci_addr_t *buf, int size)
{
  pci_ident_t ident;

  ident.vendor_id = vendor;
  ident.device_id = product;
  ident.lookup_opt = PCI_LOOKUP_VP;
  
  return pci_lookup(&ident, buf, size);
}

#include <sys/kcontrol.h>

int pci_get_dev(pci_addr_t *addr, pci_dev_t *dev)
{
  int ret;

  ret = ioctl(pcifd, PIOC_GET_DEV, sizeof(*dev), addr, dev);

  if(ret == sizeof(*dev)) {
    ret = 0;
  }
  
  return ret;
}

int pci_search_devices(pci_dev_t **devs, pci_devid_t *ids,
                       uint16_t nids)
{
  long i;
  int r = 0, ret = 0;
  pci_dev_t *d = NULL;
  pci_addr_t addrs[MAX_SCANNED_PCI_DEVICES];
  /* number of free elements in the 'addrs' buffer */
  uint32_t n = MAX_SCANNED_PCI_DEVICES;

  for (i = 0; (i < nids) && n; i++) {
    r = pci_lookup_prod_vend(ids[i].vendor, ids[i].product, addrs + ret,
                             n * sizeof(pci_addr_t));
    if (r == -1) {
      ret = -1;
      break;
    }

    ret += r;
    n -= r;
  }

  if (ret > 0) {
    d = malloc(ret * sizeof(pci_dev_t));
    if (d == NULL) {
      ret = -1;
      errno = ENOMEM;
    } else {
      for (i = 0; i < ret; i++) {
        r = pci_get_dev(addrs + i, d + i);
        if (r == -1) {
          ret = r;
          break;
        }
      }
    }
  }

  if ((ret == -1) && (d != NULL)) {
    free(d);
  } else {
    *devs = d;
  }

  return ret;
}
