/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008 MString Core Team <http://mstring.berlios.de>
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 *
 *  Test framework implementation
 */

#include <sys/types.h>
#include <stddef.h>
#include <stdlib.h>
#include <ipc/rpc.h>

#include "framework.h"

typedef bool (*comp_func_t)(long, long);

static int current_case;
static int nr_failed_cases;
static int total_fails;
static int failed_cases[MAX_TESTS];
static bool current_passed;

int cmdline_args_num;
char **cmdline_args;
char **env_vars;

static bool lless(long a, long b)
{
  return (a < b);
}

static bool lless_equ(long a, long b)
{
  return (a <= b);
}

static bool lequ(long a, long b)
{
  return (a == b);
}

static bool lnot_equ(long a, long b)
{
  return (a != b);
}

bool lgreater(long a, long b)
{
  return (a > b);
}

bool lgreq(long a, long b)
{
  return (a >= b);
}

static comp_func_t comp_funcs[] = {
  lless,
  lless_equ,
  lequ,
  lnot_equ,
  lgreater,
  lgreq
};

int main(int argc, char *argv[], char *envp[])
{
  int i, j, r = 0;
  char buf[1024], *p;
  test_case_t *tc;
  long *v;
  bool abrt = false;

  cmdline_args_num = argc;
  cmdline_args = argv;
  env_vars = envp;

  if (test_struct.init)
    r = test_struct.init();

  if (r) {
    printf("Test initialization failed\n");
    return 1;
  }

  for (current_case = 0; (current_case < MAX_TESTS) && !abrt; current_case++) {
    if (test_struct.cases[current_case].function == NULL)
      break;

    tc = test_struct.cases + current_case;
    if (total_fails && tc->nolaunch_on_fails)
      continue;

    if (test_struct.must_launched && !test_struct.must_launched(tc->function))
      continue;

    printf("Launch '%s' test\n", tc->name);
    if (tc->description) {
      printf(tc->description);
    }
    current_passed = true;
    if (tc->arg_list == NULL) {
      tc->function(NULL);
      if (!current_passed && tc->abort_on_fail) {
        abrt = true;
        break;
      }
    } else {
      /* call the case whith each argument chain */
      for (i = 0; (v = tc->args[i]); i++) {
        buf[0] = 0;
        p = buf;
        for (j = 0; tc->arg_list[j] != NULL; j++) {
          sprintf(p, "%s = %d, ", tc->arg_list[j], v[j]);
          p += strlen(p);
        }
        *(p - 2) = '\n';   /* reduce last ',' and ' ' */
        *(p - 1) = '\0';
        printf(buf);
        errno = 0;
        tc->function(v);
        if (!current_passed && tc->abort_on_fail) {
          abrt = true;
          break;
        }
        current_passed = true;
      }
    }

    if (test_struct.pause_seconds)
      sleep(test_struct.pause_seconds);
  }

  if (test_struct.complete)
    test_struct.complete();

  if (total_fails) {
    printf("\nTotally %d tests are failed, names of failed cases:\n", total_fails);
    for (i = 0; i < nr_failed_cases; i++) {
      j = failed_cases[i];
      printf("%s\n", test_struct.cases[j].name);
    }
  } else {
    printf("\nAll tests passed\n");
  }

  return 0;
}

void test_fail(char *msg,...)
{
  char tmp[MAX_MSG_LENGTH], *p;
  va_list ap;
  int n, len;

  if (msg == NULL) {
    printf("FAIL\n\n");
  } else {
    sprintf(tmp, "FAIL: ");
    n = strlen(tmp);
    p = tmp + n;

    va_start(ap, msg);
    len = vsnprintf(p, MAX_MSG_LENGTH - n - 1, msg, ap);
    va_end(ap);

    p[len++] = '\n';
    if (len + n < MAX_MSG_LENGTH)
      p[len] = '\0';

    printf(tmp);
  }

  total_fails++;
  if (!nr_failed_cases || failed_cases[nr_failed_cases - 1] != current_case)
    failed_cases[nr_failed_cases++] = current_case;

  current_passed = false;
}

bool check_test_condition(long res, long exp_res, const char *err_msg,
                          bool pass_on_true)
{
  char buf[MAX_MSG_LENGTH], *p = buf;
  bool ret = true;

  buf[0] = '\0';
  if (res == exp_res) {
    if (pass_on_true)
      test_pass();
  } else {
    ret = false;
    if (err_msg != NULL) {
      sprintf(buf, "%s ", err_msg);
      p += strlen(buf);
    }
    sprintf(p, "expected result is %ld, but %ld is returned\n",
                exp_res, res);
    test_fail(buf);
  }

  return ret;
}

bool check_test_condition_ext(long res, long exp_res, comp_type_t ctype,
                              int exp_errno, const char *err_msg, bool pass_on_true)
{
  char buf[MAX_MSG_LENGTH], *p = buf;
  bool ret = true;
  const char *s[6] = {"less than ", "not greater than", "", "not equal to", 
                      "greater than ", "not less than "};
  const char *comma = ", ", *empty = "";
  const char *prefix = empty;

  buf[0] = '\0';
  if (comp_funcs[ctype](res, exp_res) && (exp_errno == errno)) {
    if (pass_on_true)
      test_pass();
  } else {
    ret = false;
    if (err_msg != NULL) {
      sprintf(buf, "%s", err_msg);
      prefix = comma;
    }
    if (!comp_funcs[ctype](res, exp_res)) {
      p += strlen(p);
      sprintf(p, "%sexpected result is %s%ld, but %ld is returned",
              prefix, s[ctype], exp_res, res);
      prefix = comma;
    }
    if (exp_errno != errno) {
      p += strlen(p);
      if (exp_errno) {
        sprintf(p, "%sexpected error is %s, but %s is returned",
                prefix, error_strings[exp_errno], error_strings[errno]);
      } else {
        sprintf(p, "%s%s error is returned", prefix, error_strings[errno]);
      }
    }
    strcat(buf, "\n");
    test_fail(buf);
  }

  return ret;
}
