/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Based on Linux byteorder/swab.h
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Graphov Paul <graphov@jarios.org>
 *
 * byteorder.h: operations for working with binary data in machine
 * byte-order independent way
 */

#ifndef __BYTEORDER_H__
#define __BYTEORDER_H__

#define BIG_ENDIAN 4321
#define LITTLE_ENDIAN 1234

#include <sys/types.h>
#include <arch/byteorder.h>

#ifndef HAVE_ARCH_SWAB16
static inline uint16_t __swab16(uint16_t x)
{
	return x << 8 | x >> 8;
}
#else
#define __swab16(x) __arch_swab16(x)
#endif

#ifndef HAVE_ARCH_SWAB32
static inline uint32_t __swab32(uint32_t x)
{
	return                       x << 24 |
		(x & (uint32_t)0x0000ff00UL) << 8  |
		(x & (uint32_t)0x00ff0000UL) >> 8  |
                               x >> 24;
}
#else
#define __swab32(x) __arch_swab32(x)
#endif

#ifndef HAVE_ARCH_SWAB64
static inline uint64_t __swab64(uint64_t x)
{
	return
                                        x << 56 |
		(x & (uint64_t)0x000000000000ff00ULL) << 40 |
		(x & (uint64_t)0x0000000000ff0000ULL) << 24 |
		(x & (uint64_t)0x00000000ff000000ULL) << 8  |
    (x & (uint64_t)0x000000ff00000000ULL) >> 8  |
		(x & (uint64_t)0x0000ff0000000000ULL) >> 24 |
		(x & (uint64_t)0x00ff000000000000ULL) >> 40 |
                                        x >> 56;
}
#else
#define __swab64(x) __arch_swab64(x)
#endif

#define ___constant_swab16(x) \
	((uint16_t)( \
		(((uint16_t)(x) & (uint16_t)0x00ffU) << 8) | \
		(((uint16_t)(x) & (uint16_t)0xff00U) >> 8) ))

#define ___constant_swab32(x) \
	((uint32_t)( \
		(((uint32_t)(x) & (uint32_t)0x000000ffUL) << 24) | \
		(((uint32_t)(x) & (uint32_t)0x0000ff00UL) <<  8) | \
		(((uint32_t)(x) & (uint32_t)0x00ff0000UL) >>  8) | \
		(((uint32_t)(x) & (uint32_t)0xff000000UL) >> 24) ))

#define ___constant_swab64(x) \
	((uint64_t)( \
		(uint64_t)(((uint64_t)(x) & (uint64_t)0x00000000000000ffULL) << 56) | \
		(uint64_t)(((uint64_t)(x) & (uint64_t)0x000000000000ff00ULL) << 40) | \
		(uint64_t)(((uint64_t)(x) & (uint64_t)0x0000000000ff0000ULL) << 24) | \
		(uint64_t)(((uint64_t)(x) & (uint64_t)0x00000000ff000000ULL) <<  8) | \
    (uint64_t)(((uint64_t)(x) & (uint64_t)0x000000ff00000000ULL) >>  8) | \
		(uint64_t)(((uint64_t)(x) & (uint64_t)0x0000ff0000000000ULL) >> 24) | \
		(uint64_t)(((uint64_t)(x) & (uint64_t)0x00ff000000000000ULL) >> 40) | \
		(uint64_t)(((uint64_t)(x) & (uint64_t)0xff00000000000000ULL) >> 56) ))

#if BYTE_ORDER == BIG_ENDIAN

#define cpu_to_le16 __swab16
#define cpu_to_le32 __swab32
#define cpu_to_le64 __swab64
#define const_cpu_to_le16 __constant_swab16
#define const_cpu_to_le32 __constant_swab32
#define const_cpu_to_le64 __constant_swab64
#define le_to_cpu16 __swab16
#define le_to_cpu32 __swab32
#define le_to_cpu64 __swab64
#define const_le_to_cpu16 __constant_swab16 
#define const_le_to_cpu32 __constant_swab32
#define const_le_to_cpu64 __constant_swab64
#define cpu_to_be16(x) x
#define cpu_to_be32(x) x
#define cpu_to_be64(x) x
#define const_cpu_to_be16(x) x
#define const_cpu_to_be32(x) x
#define const_cpu_to_be64(x) x
#define be_to_cpu16(x) x
#define be_to_cpu32(x) x
#define be_to_cpu64(x) x
#define const_be_to_cpu16(x) x
#define const_be_to_cpu32(x) x
#define const_be_to_cpu64(x) x

#elif BYTE_ORDER ==  LITTLE_ENDIAN

#define cpu_to_be16 __swab16
#define cpu_to_be32 __swab32
#define cpu_to_be64 __swab64
#define const_cpu_to_be16 __constant_swab16
#define const_cpu_to_be32 __constant_swab32
#define const_cpu_to_be64 __constant_swab64
#define be_to_cpu16 __swab16
#define be_to_cpu32 __swab32
#define be_to_cpu64 __swab64
#define const_be_to_cpu16 __constant_swab16 
#define const_be_to_cpu32 __constant_swab32
#define const_be_to_cpu64 __constant_swab64
#define cpu_to_le16(x) x
#define cpu_to_le32(x) x
#define cpu_to_le64(x) x
#define const_cpu_to_le16(x) x
#define const_cpu_to_le32(x) x
#define const_cpu_to_le64(x) x
#define le_to_cpu16(x) x
#define le_to_cpu32(x) x
#define le_to_cpu64(x) x
#define const_le_to_cpu16(x) x
#define const_le_to_cpu32(x) x
#define const_le_to_cpu64(x)
#else
#error "BYTE_ORDER must be set to BIG_ENDIAN or LITTLE_ENDIAN by arch/byteorder.h"
#endif

#endif
