/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Paul Graphov <graphov@jarios.org>
 *
 * circular.h: circular buffer, fixed length buffer with independent
 * read and write pointers that support interleaving read and write
 * operation without any synchronisation, also flush may overlap with
 * getc.
 *
 */

typedef struct __circular_t {
  size_t size;
  int write_pos;
  int read_pos;
  char *data;
} circular_t;

status_t circular_init(circular_t *circ, size_t size);
void circular_destroy(circular_t *circ);

size_t circular_read(circular_t *circ, char *buf, size_t size);
size_t circular_write(circular_t *circ, char *buf, size_t size);

//next two are faster if only one char is to be moved.
static inline size_t circular_getc(circular_t *circ, char *c)
{
  if(circ->write_pos != circ->read_pos) {
    *c = circ->data[circ->read_pos];
    circ->read_pos++;
    if(circ->read_pos == circ->size) {
      circ->read_pos = 0;
    }
    return 1;
  }
  return 0;
}

static inline size_t circular_putc(circular_t *circ, char c)
{
  size_t new_write_pos = (circ->write_pos + 1) % circ->size;
  if(new_write_pos != circ->read_pos) {
    circ->data[circ->write_pos] = c;
    circ->write_pos = new_write_pos;
    return 1;
  }
  return 0;
}

// get one character back from writing end;

static inline size_t circular_unputc(circular_t *circ, char *c)
{
  if(circ->write_pos != circ->read_pos) {
    circ->write_pos--;
    if(circ->write_pos < 0) {
      circ->write_pos = circ->size - 1;
    }
    *c = circ->data[circ->write_pos];
    return 1;
  }
  return 0;
}

void circular_flush(circular_t *circ);
