/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org> 
 *
 * name server related structures
 * 
 */

#ifndef __DEVIO_DEVIO_H__
#define __DEVIO_DEVIO_H__

#include <sys/types.h>
#include <ipc/rpc.h>
#include <ipc/channel.h>

#define HWEOF  0x1   /* hardware EOF */
#define HWEOD  0xff  /* hardware EOD (end of directory) */

#define HWDIR         0x1
#define HWNODE        0x2
#define HWREGISTERED  0x4

/* flags to hw_open() */
#define HWRD_ONLY  0x1
#define HWWR_ONLY  0x2
#define HWRDWR     0x4

/* capability flags */
#define DIRECTIO_WRITE    0x1
#define DIRECTIO_READ     0x2
#define EVENTS_POLL       0x4
#define HW_READ           0x8
#define HW_WRITE          0x10
#define MULTIEVENTS_POLL  0x20
#define DIRECTIO_SEEK     0x40
#define HW_SEEK           0x80

/* hardware ioctl buffer directions */
enum {
  IOCTL_INBUF,
  IOCTL_OUTBUF
};

enum {
  DFS_BLKDEV = 0 ,
  DFS_CHARDEV = 1,
  DFS_LINK = 2,
  DFS_SYMLINK = 3,
  DFS_DIR = 4,
};

#define DEV_ASSIGN  0x1f

typedef uint16_t devid_t;

/* io memory area */
typedef struct __iomem_area_t {
  ulong_t base;
  size_t size;
} iomem_area_t;

/* types specified for hw_* ops */
typedef struct __hw_dirent {
  uint16_t name_len;
  uint16_t ntype;
  char ename[256];
} hw_dirent_t;

typedef struct __hw_direct_io_t {
  pid_t pid;
  ulong_t cmd;
  ulong_t in;
  ulong_t out;
  chid_t chc;
  chid_t chi;
  chid_t cho;
  devid_t devno;
  size_t buf_size;
  size_t offset;
  void *iobuf;
  char *devn;
} hw_direct_io_t;

/* functions prototypes */
/* driver related functions */
/* 
 * register driver device node:
 * devn: full path from root to register 
 * pid: your pid
 * cmd_id,input_id,output_id:  your ports
 * cap: capatibilities flag
 * driver_name: your driver name 
 * id: device number (unique to your driver only)
 *
 */
int hw_register_driver(const char *devn,pid_t pid,ulong_t cmd_id,
           ulong_t input_id,ulong_t output_id,
           int cap,char *driver_name,devid_t id);

/* 
 * remove device node
 * devn: full path from root to unregister
 */
int hw_unregister_device(const char *devn);

/* direct IO operations */
/*
 * open direct IO to the driver's device node:
 * devn: full path from root to device 
 * flags: opening flags
 * superio: pointer to direct IO structure descriptor
 */
int hw_dio_open(const char *devn,int flags,hw_direct_io_t *superio);

/*
 * Direct IO read:
 * superio: pointer to direct io descriptor structure
 * buf: pointer to buf
 * size: size of the reading.
 */
size_t hw_dio_read(hw_direct_io_t *superio,uni_rpc_t *rpc,void *buf,size_t size);

/*
 * Direct IO write:
 * superio: pointer to direct io descriptor structure 
 * buf: pointer to buf
 * size: size amount to write
 */
size_t hw_dio_write(hw_direct_io_t *superio,uni_rpc_t *rpc,const void *buf,size_t size);

/* 
 * Direct IO IOCTL:
 *
 */
int hw_dio_ioctl(hw_direct_io_t *superio,uni_rpc_t *rpc,void *buffer,uint32_t buflen,int dir);

/*
 * Direct IO close
 */
int hw_dio_close(hw_direct_io_t *superio);

/* operations with devices events */
int hw_register_event(const char *devn,ulong_t listen_port);
int hw_unregister_event(const char *devn);

/* devices operations */
/* 
 * open device: 
 * devn: full path to device
 * Return: 
 * fd or errno negative value
 */
int hw_open(const char *devn);

/*
 * create device node
 * devn - symbolic path with name to create device
 * Return: 0 if ok , errno if something wrong
 */
int hw_create(const char *devn);

/*
 * unlink device node
 * devn - symbolic path with name to unlink device
 * Return: 0 if ok, errno if something wrong
 */
int hw_unlink(const char *devn);

/*
 * create device directory
 * devn - symbolic path with directory name to create device
 * Return: 0 if ok , errno if something wrong
 */
int hw_mkdir(const char *devn);

/*
 * open hardware directory 
 * devn - symbolic name 
 * Return: fd of opened directory, or errno if something wrong
 */
int hw_opendir(const char *devn);

/*
 * read directory entry
 * fd - fd if opened directory
 * dentry - entry
 * Return - errno, 0 or EOD
 */
int hw_readdir(int fd, hw_dirent_t *dentry);

/*
 * close hardware directory
 * fd - fd if opened directory
 * Return: 0, or errno if something wrong*
 */
int hw_closedir(int fd);

int deviofs_assign(int fildes, uint16_t cmd_port, uint16_t io_port, uint16_t dev_type, uint16_t dev_id);


#endif /* __DEVIO_DEVIO_H__ */

