/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Dmitry Gromada <gromada82@gmail.com>
 *
 * LibOS:
 * include/sys/irq.h: API prototypes for irq control
 */

/* TO DO: implement get/set irq balancer */

#ifndef __IRQ_H__
#define __IRQ_H__

#include <sys/types.h>
#include <arch/irq.h>

#ifndef NR_IRQ_VECTORS
#define NR_IRQ_VECTORS 16
#endif

/* irq trigger modes */
#define IRQ_TRIG_LEVEL 1
#define IRQ_TRIG_EDGE  0

/* irq polarities */
#define IRQ_POLAR_HIGH 0
#define IRQ_POLAR_LOW  1

/* irq controller ioctls */
enum {
  IRQCON_ROUTE,
  IRQCON_MASK,
  IRQCON_SET_PRIO,
  IRQCON_BALANCE_CONTROL,
  IRQCON_ADD_LISTENER,
  IRQCON_RM_LISTENER,
  /* 
   * set irq mode, hardware specific ioctl, mustn't be called from
   * arch independent code
   */
  IRQCON_SET_MODE
};

/* irq balancer control commands */
enum __irq_balance_cmd {
  BALCTL_SET_IRQ_FLAGS,
  BALCTL_GET_IRQ_FLAGS,
  BALCTL_GET_BALANCERS,
  BALCTL_SET_BALANCER,
  BALCTL_SET_CPU_LIMIT,   /* set irq the irq limit for a cpu */
  BALCTL_GET_CPU_LIMIT    /* get irq limit for a cpu */
} irq_balance_cmd_t;

/* irq flags */
enum {
   IRQ_NOT_MIGRATE = 0x1  /* not migrate this irq on another CPU */
};

typedef struct __irq_mode {
  int trig;        /* trigger mode */
  int deliv;      /* delivery mode */
  int polarity;
} irq_mode_t;

/* per cpu irq handling limit */
typedef struct __cpu_irq_limit {
   uint16_t cpu;
   uint16_t nirqs;
} cpu_irq_limit_t;


/*
 * hw_route_irq - route irq to the specified CPU
 *
 * @irq: irq to route
 * @cpu: CPU logical ID to which route this irq, CPUs are enumerated starting from 0
 *
 *  Return value:
 *    0 on success
 *    -EINVAL if cpu or irq is invalid
 *    -ENOSYS if this functionality is not supprorted
 */
int hw_route_irq(uint16_t irq, uint16_t cpu);


/*
 * hw_set_irq_prio - set irq priority
 *
 * @irq: irq to set priority to
 * @prio: static interrupt priority, higher value of this argument corresponds
 *        to higher hardware priority
 *
 * Return value:
 *    0 on success
 *    negated value of error otherwise:
 *      -EINVAL if priorities or irq id are invalid
 *      -ENOSYS if this functionality is not supported 
 */
int hw_set_irq_prio(uint16_t irq, uint16_t prio);

/*
 * Control irq balancer
 *
 * @cmd: control command
 * @buf: location containing control data
 *    or where data will be stored, must point to
 *    the following structure depending on 'cmd':
 */
int hw_irq_balance_ctl(uint16_t cmd, void *buf);

/*
 * mask or unmask interrupt depending on the mask argument:
 *    0 - unmask
 *    not 0 - mask
 *
 *  Return value:
 *    0 on success
 *    negated value of error otherwise:
 *      -EINVAL if an invalid irq number is passed
 *      -ENOSYS if this functionality is not supported
 */
int hw_mask_irq(uint16_t irq, int mask);

/*
 * Add the calling processes to listeners of this irq
 *
 * @irq: irq that the process will handle
 *
 * Return value:
 *  0 on success
 *  negated value of error otherwise
 *    -EINVAL if an invalid irq number  is passed
 *    -ENOSYS if this functionality is not supported
 */
int hw_add_irq_listener(uint16_t irq);

/*
 * Remove the calling process from listeners of this irq
 *
 * @irq: irq that the process handles
 *
 * Return value:
 *  0 on success
 *  negated value of error otherwise
 *    -EINVAL if an invalid irq number  is passed
 *    -ENOSYS if this functionality is not supported
 */
int hw_rm_irq_listener(uint16_t irq);

/* if some field of mode is set to -1 it is ignored */
int hw_set_irq_mode(uint16_t irq, irq_mode_t *mode);

#endif /* __IRQ_H__ */
