/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2008 Jari Operating System core team <http://jarios.org>
 */

/**
 * @file include/eza/arch/atomic.h
 * A set of AMD64-dependent dependent atomic operations.
 *
 * @author Dan Kruchinin
 */

#ifndef __AMD64_ATOMIC_H__
#define __AMD64_ATOMIC_H__

#include <sys/types.h>
#include <stdbool.h>

#define __LOCK_PREFIX

#ifndef always_inline
#define always_inline inline __attribute__((always_inline))
#endif

typedef volatile long atomic_t;

/**
 * @def atomic_set(a, val)
 * Set a value of an atomic variable @a a
 *
 * @param a   - A pointer to atomic variable
 * @param val - The value to set.
 */
#define atomic_set(a, val) (*(a) = (val))

/**
 * @def atomic_get(a)
 * Get a value of an atomic variable @a a
 *
 * @param a - A pointer to atomic variable
 * @return The value of atomic variable @a a
 */
#define atomic_get(a) (*(a))

/**
 * @fn static always_inline void atomic_add(atomic_t *a, long add)
 * Atomically add signed number @a add to an atomic variable @a a
 *
 * @param[out] a   - A pointer to atomic variable @a add will be added to
 * @param[in]  add - A signed number to add.
 */
static always_inline void atomic_add(atomic_t *a, long add)
{
  __asm__ volatile ( __LOCK_PREFIX "addq %1, %0\n\t"
                    : "+m" (*a)
                    : "ir" (add));
}

/**
 * @fn static always_inline void atomic_inc(atomic_t *a)
 * Atomically increment the value of atomic variable @a a
 *
 * @param[out] a - A pointer to atomic variable.
 */
static always_inline void atomic_inc(atomic_t *a)
{
  __asm__ volatile (__LOCK_PREFIX "incq %0\n\t"
                    : "+m" (*a));
}

/**
 * @fn static always_inline void atomic_sub(atomic_t *a, long sub)
 * Atomically substract signed number @a sub from an atomic variable @a a
 *
 * @param[out] a   - A pointer to atomic vatiable @a sub will be substracted from.
 * @param[in]  sub - A signed number to substract.
 */
static always_inline void atomic_sub(atomic_t *a, long sub)
{
  __asm__ volatile (__LOCK_PREFIX "subq %1, %0\n\t"
                    : "+m" (*a)
                    : "ir" (sub));
}

/**
 * @fn static always_inline void atomic_dec(atomic_t *a)
 * Atomically decrement the value of atomic variable @a a
 *
 * @param[out] a - A pointer to atomic variable.
 */
static always_inline void atomic_dec(atomic_t *a)
{
  __asm__ volatile (__LOCK_PREFIX "decq %0\n\t"
                    : "+m" (*a));
}

/**
 * @fn static always_inline bool atomic_dec_and_test(atomic_t *a)
 * Atomically decrements target atomic variable and tests if its
 * value is zero.
 * @return true - if atomic variable is zero, false - if not.
 */
static always_inline bool atomic_dec_and_test(atomic_t *a)
{
  uint8_t ret;
  __asm__ volatile (__LOCK_PREFIX "decq %0\n\t"
                    "sete %1"
                    : "+m" (*a), "=m" (ret));

  return (ret != 0);
}

/**
 * @fn static always_inline bool atomic_inc_and_test(atomic_t *a,long sub)
 * Atomically increments target atomic variable by a given amount of bytes
 * and tests if its
 * value is zero.
 * @return true - if atomic variable is zero, false - if not.
 */
static always_inline bool atomic_inc_and_test(atomic_t *a)
{
  uint8_t ret;
  __asm__ volatile (__LOCK_PREFIX "incq %0; sete %1"
                    : "+m" (*a), "=m" (ret));

  return (ret != 0);
}

/**
 * @fn static always_inline bool atomic_sub_and_test(atomic_t *a,long sub)
 * Atomically decrements target atomic variable by a given amount of bytes
 * and tests if its
 * value is zero.
 * @return true - if atomic variable is zero, false - if not.
 */
static always_inline bool atomic_sub_and_test(atomic_t *a, long sub)
{
  uint8_t ret;
  __asm__ volatile (__LOCK_PREFIX "subq %2, %0\n\t"
                    "sete %1"
                    : "+m" (*a), "=m" (ret)
                    : "ir" (sub));

  return ret;
}

static always_inline bool atomic_test_bit(unsigned long *v, unsigned long bit) {
  bool res;

  __asm__ __volatile__( __LOCK_PREFIX "bt %2,%1\n"
                        "adc $0,%0\n"
                        :"=r"(res), "=m"(*v)
                        :"r"(bit), "r"(0))  ;
  return res ? 1 : 0;
}

static always_inline bool atomic_test_and_set_bit(unsigned long *v, unsigned long bit) {
  bool res;

  __asm__ __volatile__( __LOCK_PREFIX "bts %2,%1\n"
                        "adc $0,%0\n"
                        :"=r"(res), "=m"(*v)
                        :"r"(bit), "r"(0))  ;
  return res ? 1 : 0;
}

static always_inline bool atomic_test_and_reset_bit(unsigned long *v,unsigned long bit) {
  bool res;

  __asm__ __volatile__( __LOCK_PREFIX "btr %2,%1\n"
                        "adc $0,%0\n"
                        :"=r"(res), "=m"(*v)
                        :"r"(bit), "r"(0))  ;
  return res ? 1 : 0;
}

static inline unsigned long atomic_swap_long(unsigned long *v,unsigned long new)
{
  unsigned long c;

  __asm__ __volatile__(
    "lock; xchg %1,%2\n"
    :"=r"(c)
    : "m"(*v),"r"(new) );
  return c;
}

#endif
