/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Michael Tsymbalyuk <mtzaurus@gmail.com>
 *
 * pthread.h: Data types and function prototypes related to POSIX threads.
 *
 */

#ifndef __PTHREAD_H__
#define __PTHREAD_H__

#include <sys/types.h>
#include <unistd.h>
#include <signal.h>

#define __SO_UNINITIALIZED_ID  0

/* Supported synchronization primitives. */
enum {
  __SO_MUTEX=0,  /**< Userspace mutex **/
  __SO_SEMAPHORE=1, /**< Userspace semaphore **/
  __SO_CONDVAR=2 /**< Userspace conditional variable **/
};

/* Control commands related to sync primitives. */
enum {
  /* Mutex-related commands. */
  __SYNC_CMD_MUTEX_LOCK=0x100,    /**< Lock target mutex. **/
  __SYNC_CMD_MUTEX_UNLOCK=0x101,  /**< Unlock target mutex. **/
  __SYNC_CMD_MUTEX_TRYLOCK=0x102, /**< Try to lock target mutex. **/
};

enum {
  PTHREAD_PROCESS_PRIVATE=0,  /**< Private primitive. **/
  PTHREAD_PROCESS_SHARED=1    /**< Shared primitive. **/
};

#define SYNC_CTRL(id,cmd,arg)  __syscall3(SC_SYNC_CONTROL,(id),(cmd),(arg))
#define MUTEX_CTRL(m,cmd,arg)  SYNC_CTRL((m)->__id,(cmd),(arg))
#define SYNC_DESTROY(s)  __syscall1(SC_SYNC_DESTROY,(s)->__id)

int pthread_mutex_destroy(pthread_mutex_t *mutex);
int pthread_mutex_init(pthread_mutex_t *mutex,const pthread_mutexattr_t *attr);
int pthread_mutexattr_setpshared(pthread_mutexattr_t *attr,int pshared);
int pthread_mutex_lock(pthread_mutex_t *mutex);
int pthread_mutex_unlock(pthread_mutex_t *mutex);
int pthread_mutex_trylock(pthread_mutex_t *mutex);

#define PTHREAD_MUTEX_INITIALIZER  {__SO_UNINITIALIZED_ID,    \
    {(PTHREAD_PROCESS_PRIVATE)} }

#define POSIX_ERR(e) ((e) >=0 ? (e) : (-e))

#define PTHREAD_CANCEL_ENABLE   1
#define PTHREAD_CANCEL_DISABLE  0

#define PTHREAD_CANCEL_DEFERRED      1
#define PTHREAD_CANCEL_ASYNCHRONOUS  0

#define PTHREAD_CANCELLED  ~(unsigned long)0

static inline tid_t pthread_self(void)
{
  return gettid();
}

int pthread_create(pthread_t *thread,const pthread_attr_t *attr,
                   void *(*start_routine)(void *),void *arg);
int pthread_join(pthread_t thread,void **value_ptr);
void pthread_exit(void *value_ptr);
int pthread_setcancelstate(int state,int *oldstate);
int pthread_setcanceltype(int type,int *oldtype);
int pthread_cancel(pthread_t thread);

static int inline pthread_kill(pthread_t thread, int sig)
{
  int r;

  thread_kill(0, thread.__tid, sig);

  return POSIX_ERR(r);
}

#endif
