/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Michael Tsymbalyuk <mtzaurus@gmail.com>
 *
 * LibC:
 * include/signal.h: Generic data types and prototypes for dealing with
 *                   POSIX signals.
 */

#ifndef __SIGNAL_H__
#define __SIGNAL_H__

#include <arch/signal.h>

/* Signals */
#define	SIGHUP		1	  /* Hangup (POSIX) */
#define	SIGINT		2	  /* Interrupt (ANSI) */
#define	SIGQUIT		3	  /* Quit (POSIX) */
#define	SIGILL		4	  /* Illegal instruction (ANSI) */
#define	SIGTRAP		5	  /* Trace trap (POSIX) */
#define	SIGABRT		6	  /* Abort (ANSI) */
#define	SIGIOT		6	  /* IOT trap (4.2 BSD) */
#define	SIGBUS		7	  /* BUS error (4.2 BSD) */
#define	SIGFPE		8	  /* Floating-point exception (ANSI) */
#define	SIGKILL		9	  /* Kill, unblockable (POSIX) */
#define	SIGUSR1		10	  /* User-defined signal 1 (POSIX) */
#define	SIGSEGV		11	  /* Segmentation violation (ANSI) */
#define	SIGUSR2		12	  /* User-defined signal 2 (POSIX) */
#define	SIGPIPE		13	  /* Broken pipe (POSIX) */
#define	SIGALRM		14	  /* Alarm clock (POSIX) */
#define	SIGTERM		15	  /* Termination (ANSI) */
#define	SIGSTKFLT	16	  /* Stack fault */
#define	SIGCHLD		17	  /* Child status has changed (POSIX) */
#define	SIGCLD		SIGCHLD	  /* Same as SIGCHLD (System V) */
#define	SIGCONT		18	  /* Continue (POSIX) */
#define	SIGSTOP		19	  /* Stop, unblockable (POSIX) */
#define	SIGTSTP		20	  /* Keyboard stop (POSIX) */
#define	SIGTTIN		21	  /* Background read from tty (POSIX) */
#define	SIGTTOU		22	  /* Background write to tty (POSIX) */
#define	SIGURG		23	  /* Urgent condition on socket (4.2 BSD) */
#define	SIGXCPU		24	  /* CPU limit exceeded (4.2 BSD) */
#define	SIGXFSZ		25	  /* File size limit exceeded (4.2 BSD) */
#define	SIGVTALRM	26	  /* Virtual alarm clock (4.2 BSD) */
#define	SIGPROF         27	  /* Profiling alarm clock (4.2 BSD) */
#define	SIGWINCH        28	  /* Window size change (4.3 BSD, Sun) */
#define	SIGIO           29	  /* I/O now possible (4.2 BSD) */
#define	SIGPOLL         SIGIO	  /* Pollable event occurred (System V) */
#define	SIGPWR          30	  /* Power failure restart (System V) */
#define SIGSYS          31	  /* Bad system call */

#define _NSIG           65        /* Biggest signal number + 1
                                   * (including real-time signals).
				   */

#define SIGRTMIN        32
#define SIGRTMAX        (_NSIG - 1)


/* Fake signal function */
#define SIG_ERR (sighandler_t)-1  /* Error occured */
#define SIG_IGN (sighandler_t)0   /* Ignore signal */
#define SIG_DFL (sighandler_t)1   /* Default action */


#ifndef __PID_T_DEFINED
typedef int             pid_t;
#define __PID_T_DEFINED 1
#endif

/* Type of a signal handler */
typedef void (*sighandler_t)(int);

typedef union sigval {
  int       sival_int;
  void     *sival_ptr;
} sigval_t;

typedef struct __siginfo {
  int       si_signo;    /* Signal number */
  int       si_errno;    /* An errno value */
  int       si_code;     /* Signal code */
  pid_t     si_pid;      /* Sending process ID */
  uid_t     si_uid;      /* Real user ID of sending process */
  int       si_status;   /* Exit value or signal */
  clock_t   si_utime;    /* User time consumed */
  clock_t   si_stime;    /* System time consumed */
  sigval_t  si_value;    /* Signal value */
  int       si_int;      /* POSIX.1b signal */
  void     *si_ptr;      /* POSIX.1b signal */
  void     *si_addr;     /* Memory location which caused fault */
  int       si_band;     /* Band event */
  int       si_fd;       /* File descriptor */
} siginfo_t;

/* Available flags for the 'siginfo' structure. */
#define SA_NOCLDSTOP  0x1
#define SA_NOCLDWAIT  0x2
#define SA_RESETHAND  0x4
#define SA_ONSTACK    0x8
#define SA_RESTART    0x10
#define SA_NODEFER    0x20
#define SA_SIGINFO    0x40

typedef struct sigaction {
  /* Signal handlers. */
  void (*sa_handler)(int);
  void (*sa_sigaction)(int, siginfo_t*, void*);

  /* Additional set of signals to be blocked */
  sigset_t sa_mask;

  /* Special flags */
  int sa_flags;

  /* Restore handler */
  void (*sa_restorer)(void);
} sigaction_t;

struct sigevent {
  int sigev_notify;
  int sigev_signo;
  union sigval sigev_value;
  tid_t tid;
};

/* Values for the 'how' argument to sigprocmask() */
#define SIG_BLOCK       0         /* Block signals */
#define SIG_UNBLOCK     1         /* Unblock signals */
#define SIG_SETMASK     2         /* Set a new mask of blocking signals */

#define SIGEV_SIGNAL         0 /* Send a signal to a process. */
#define SIGEV_SIGNAL_THREAD  1 /* Send a signal to the specified thread. */

#define valid_signal(s)  ((s) >= 0 && (s) <= SIGRTMAX)

/* Functions for dealing with POSIX signal sets. */
int sigemptyset(sigset_t *set);
int sigfillset(sigset_t *set);
int sigaddset(sigset_t *set, int signum);
int sigdelset(sigset_t *set, int signum);

struct timespec;

/* Prototypes for POSIX signal-related functions. */
sighandler_t signal(int signum, sighandler_t handler);
int kill (pid_t pid, int sig);
int sigprocmask(int how, const sigset_t *set, sigset_t *oldset);
int sigwait(const sigset_t *set,int *sig);
int sigwaitinfo(const sigset_t *set,siginfo_t *info);
int sigaction(int signum,const struct sigaction *act,struct sigaction *oldact);
int sigtimedwait(const sigset_t *set,siginfo_t *info,const struct timespec *timeout);

int raise(int sig);

/* JariOS-specific signal-related functions. */
status_t kill_siginfo(pid_t pid, int sig, siginfo_t *siginfo);
status_t thread_kill(pid_t pid, tid_t tid, int sig);

#endif  /* __SIGNAL_H__ */
