/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Michael Tsymbalyuk <mtzaurus@gmail.com>
 * (c) Copyright 2008 Dan Kruchinin <dan,kruchinin@gmail.com>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 * (c) Copyright 2009 Stanislav M. Ivankin <grepz@jarios.org>
 *
 * LibC:
 * include/string.h: contains main libc types and prototypes for dealing
 *                   with strings.
 */

#ifndef __STRING_H__
#define __STRING_H__

#include <stddef.h>
#include <arch/string.h>

#ifndef ARCH_MEMSET
static inline void *memset(void *s, int c, size_t n)
{
    if ((uintptr_t)s & 0x1) {
        register char *to = (char *)s;

        while (n-- > 0)
            *to++ = c;
    }
    else { /* address is aligned by 2 */
        register uint16_t *to = (uint16_t *)s;
        register uint16_t tmp = ((c & 0xff) | ((c & 0xff) << 8));
        size_t len = n >> 1;

        while (len-- > 0)
            *to++ = tmp;
        if (n & 0x1)
            *(char *)to = c;
    }

    return s;
}
#else
extern void *memset(void *s,int c,size_t n);
#endif /* ARCH_MEMSET */


#ifndef ARCH_MEMCPY
static inline void *memcpy(void *dst, const void *src, size_t n)
{
    register char *to = (char *)dst;
    register const char *from = src;

    while(n-- > 0)
        *to++ = *from++;

    return dst;
}
#else
extern void *memcpy(void *dst,const void *src,size_t n);
#endif /* ARCH_MEMCPY */

/* TODO: Arch optimized vairant */
static inline void *memccpy(void *dst, const void *src, int c, size_t n)
{
  register char *to = (char *)dst;
  register const char *from = src;

  while(n-- > 0)
    if ((*to++ = *from++) == c)
      return to;

  return NULL;
}

#ifndef ARCH_MEMMOVE
static inline void* memmove(void *dest, const void *src, size_t n)
{
  register const char *from = src;
  register char *to = dest;
  register size_t i;

  if ((from == NULL) || (to == NULL))
    return NULL;
  if (from == to)
    return dest;

  if ((from > to) || (from + n < to))
    memcpy(dest, src, n);
  else {
    for (i = n; i > 0; i--)
      to[i - 1] = from[i - 1];
  }

  return dest;
}
#else
extern void *memmove(void *dst, const void *src, size_t n);
#endif

#ifndef ARCH_MEMCMP
static inline int memcmp(const void *s1, const void *s2, size_t n)
{
    register const char *p1, *p2;

    p1 = (const char *)s1;
    p2 = (const char *)s2;
    for (; n > 0; p1++, p2++, n--) {
        if (*p1 != *p2)
            return (*p1 - *p2);
    }

    return 0;
}
#else
extern int memcmp(const void *s1,const void *s2,size_t n);
#endif /* ARCH_MEMCMP */

#ifndef ARCH_STRLEN
static inline size_t strlen(const char *s)
{
    size_t n = 0;
    register const char *p = s;

    for (; *p != '\0'; p++, n++);
    return n;
}
#else
extern size_t strlen(const char *s);
#endif /* ARCH_STRLEN */


#ifndef ARCH_STRNCPY
static inline char *strncpy(char *dst, const char *src, size_t n)
{
    size_t i;

    for (i = 0; (i < n) && (src[i] != '\0'); i++)
        dst[i] = src[i];
    for ( ; i < n; i++)
        dst[i] = '\0';

    return dst;
}
#else
extern char *strncpy(char *dst,const char *src,size_t n);
#endif /* ARCH_STRNCPY */

#ifndef ARCH_STRCPY
static inline char *strcpy(char *dst, const char *src)
{
    register char *to = dst;
    register const char *from = src;

    while (*from != '\0')
        *to++ = *from++;

    *to = '\0';
    return dst;
}
#else
extern char *strcpy(char *dst,const char *src);
#endif /* ARCH_STRCPY */


#ifndef ARCH_STRNCAT
static inline char *strncat(char *dst, const char *src, size_t n)
{
    register char *to = dst;
    register const char *from = src;

    for (; *to != '\0'; to++);
    while (n-- > 0)
        *to++ = *from++;

    *to = '\0';
    return dst;
}
#else
extern char *strncat(char *dst,const char *src,size_t n)
#endif /* ARCH_STRNCAT */


#ifndef ARCH_STRCAT
static inline char *strcat(char *dst, const char *src)
{
    register char *to = dst;
    register const char *from = src;

    for (; *to != '\0'; to++);
    while (*from != '\0')
        *to++ = *from++;

    *to = '\0';
    return dst;
}
#else
extern char *strcat(char *dst,const char *src);
#endif /* ARCH_STRCAT */


#ifndef ARCH_STRNCMP
static inline int strncmp(const char *s1, const char *s2, size_t n)
{
    register const char *p1 = s1;
    register const char *p2 = s2;

    for (; n > 0; n--, p1++, p2++) {
        if (*p1 != *p2)
            return (*p1 - *p2);
    }

    return 0;
}
#else
#define strncmp(s1, s2, n) arch_strncmp(s1, s2, n)
#endif /* ARCH_STRNCMP */


#ifndef ARCH_STRCMP
static inline int strcmp(const char *s1, const char *s2)
{
  register const char *p1 = s1;
  register const char *p2 = s2;

  for (; (*p1 != '\0') && (*p2 != '\0'); p1++, p2++) {
    if (*p1 != *p2)
      return (*p1 - *p2);
  }

  return ((int)(*p1) - (int)(*p2));
}
#else
#define strcmp(s1, s2) arch_strcmp(s1, s2)
#endif /* ARCH_STRCMP */


#ifndef ARCH_STRCHR
static inline char *strchr(const char *s, int c)
{
  register const char *p = s;

  for (; (*p != '\0') && (*p != c); p++);
  return (*p == '\0') ? NULL : (char *)(s + (p - s));
}
#else
#define strchr(s, c) arch_strchr(s, c)
#endif /* ARCH_STRCHR */


#ifndef ARCH_STRRCHR
static inline char *strrchr(const char *s, int c)
{
  register const char *p = s;

  for (; (*p != '\0') && (*p != c); p++);
  if (*p == '\0')
    return NULL;
  while (*p++ == c);

  return (char *)(s + (p - s));
}
#else
#define strrchr(s, c) arch_strrchr(s, c)
#endif /* ARCH_STRRCHR */

/* TODO: Make arch dependent */
static inline char *strpbrk (const char *s, const char *accept)
{
  while (*s != '\0') {
    const char *a = accept;
    while (*a != '\0')
      if (*a++ == *s)
	return (char *) s;
    ++s;
  }

  return NULL;
}

#ifndef ARCH_STRSTR
static inline char *strstr(const char *haystack, const char *needle)
{
  register const char *s1 = haystack;
  register const char *s2 = needle;

  for (; (*s1 != '\0') && (*s1 != *s2); s1++);
  while ((*s1 != '\0') && (*s1 != '\0') && (*s1++ == *s2++));
  if ((*s1 == '\0') || (*s2 == '\0'))
    return NULL;

  return (char *)(haystack + (s1 - haystack));
}
#else
#define strstr(haystack, needle) arch_strstr(haystack, needle)
#endif /* ARCH_STRSTR */

char *strdup(const char *s);
/* This is GNU extension */
char* strndup(const char *s, size_t n);

void* memmove(void *dest, const void *src, size_t n);
int strcasecmp(const char *s1, const char *s2);
int strncasecmp(const char *s1, const char *s2, size_t n);
/* memchr and GNU extension */
void *memchr(const void *s,int c,size_t n);
void *memrchr(const void *s,int c,size_t n);

/* these are GNU extensions */
char *stpcpy(char *dest,const char *src);
void* mempcpy(void *dst, const void *src, size_t n);

/* BSD extension */
char *strsep (char **stringp, const char *delim);

char *strerror(int errnum);
int strerror_r(int errnum, char *buf, size_t buflen);

size_t strspn(const char *s1, const char *s2);
size_t strcspn(const char *s1, const char *s2);
int strcoll(const char *s1, const char *s2);

#endif  /*__STRING_H__*/
