/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 * (c) Copyright 2008 Michael Tsymbalyuk <mtzaurus@gmail.com>
 *           - initial version of thread stack allocator;
 * LibAlloc:
 * lib/imcache.c: Simple implementation of virtual address space manager.
 *                It was used in ilix and works well on many cases, but it's 
 *                just one of the memory managers/allocators implemented on 
 *                uspace, and it's possible to choose one what you need.
 *
 */

#include <stdbool.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <alloc/imcache.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <lock/lock.h>
#include <datalib/usrtc.h>
#include <arch/system.h>
#include <sys/thread.h>
#include <stdlib.h>

#define FIRST_TASK_MAX_STACK 1048576

/* global allocator pointers */
/* last used pointer to mmap */
uintptr_t last_used_ptr=0;

mcache_t *mhead;

/* Stack allocatin stuff. */
stack_cache_t stack_cache;
uintptr_t stack_free_address;
bool stack_inited=false;

static ulong_t glob_mcache_lock=0;
static ulong_t glob_vpage_lock=0;
static ulong_t stack_lock=0;

/* local functions prototypes */
static mcache_t *add_alloc_mcache(mcache_t *head);
static int __memcache_compare(const void *a,const void *b);
static int __memcache_compare_ptr(const void *a,const void *b);

/* allocate virtual pages, not actually mmap, it's just used by mmap functions */
void *alloc_virt_pages(size_t pages)
{
  mmap_area_t *node;

  usrtc_node_t *unode;
  mcache_t *scache;
  int i=0;
  int key_page=0;

  sem_lock(&glob_vpage_lock);

  if(!mhead) {
    mhead=add_alloc_mcache(mhead);
    if(!mhead) {
      errno=-ENOMEM;

      sem_unlock(&glob_vpage_lock);

      return NULL;
    }

    /* initialize mcache */

    mhead->head=mhead;
    mhead->next=NULL;
    mhead->prev=NULL;
    /* initialize data structures */
    usrtc_init(&(mhead->free),USRTC_REDBLACK,(1 << 16),__memcache_compare);
    usrtc_init(&(mhead->alloc),USRTC_REDBLACK,(1 << 16),__memcache_compare_ptr);
    usrtc_init(&(mhead->raw),USRTC_LIST,(1 << 16),__memcache_compare);
    /* now we're need to add nodes */
    while((uint8_t*)node<(uint8_t*)mhead+PAGE_SIZE-sizeof(*node)-sizeof(*mhead)) {
      node=(mmap_area_t *)((char *)mhead + sizeof(*mhead) + i*(sizeof(*node)));
      /* initialize node */
      node->addr=NULL;
      node->pages=0;
      usrtc_node_init(&node->node,node);
      usrtc_insert(&(mhead->raw),&node->node,&node->pages);
      i++;
    }
    /* okay we're need to mark that one page area was used */
    unode=usrtc_lookup(&mhead->raw,&key_page);
    usrtc_delete(&mhead->raw,unode); /* delete from raw space */
    node=usrtc_node_getdata(unode);
    node->pages=1;
    node->addr=(void *)mhead;
    usrtc_insert(&mhead->alloc,unode,node->addr); /* insert to allocated space */
    scache=mhead; 
    goto _retry_alloc;
  }

  scache=mhead; /* we will look out for memory area */

  while(scache->next!=NULL) {
  _retry_alloc:
    if(usrtc_count(&(scache->raw)) > 0) { /* surprise */
      unode=usrtc_first(&(scache->raw));
      usrtc_delete(&scache->raw,unode); /* delete from raw space */
      node=usrtc_node_getdata(unode);
      node->pages=pages;
      node->addr=(void *)last_used_ptr; 
      usrtc_insert(&scache->alloc,unode,node->addr); /* insert to allocated space */
      last_used_ptr+=PAGE_SIZE*pages;

      sem_unlock(&glob_vpage_lock);

      return node->addr;
    } else if(usrtc_count(&scache->free) > 0 && 
	      (unode=usrtc_lookup(&scache->free,&pages)) != NULL) {
      node=usrtc_node_getdata(unode);
      /* delete from freed data struct */
      usrtc_delete(&scache->free,unode);
      /* insert to used space */
      usrtc_insert(&scache->alloc,unode,node->addr);

      sem_unlock(&glob_vpage_lock);

      return node->addr;
    }
    if(scache->next!=NULL)
      scache=scache->next; 
    else
      break;
  }

  /* oops, not accessible anything */
  scache=add_alloc_mcache(mhead);
  if(!scache) {
    sem_unlock(&glob_vpage_lock);

    return NULL;
  }

  /* initialize */
  /* initialize data structures */
  usrtc_init(&(scache->free),USRTC_REDBLACK,(1 << 16),__memcache_compare);
  usrtc_init(&(scache->alloc),USRTC_REDBLACK,(1 << 16),__memcache_compare_ptr);
  usrtc_init(&(scache->raw),USRTC_LIST,(1 << 16),__memcache_compare);
  i=0; node=NULL;
  /* now we're need to add nodes */
  while((uint8_t *)node<(uint8_t *)scache+PAGE_SIZE-sizeof(*node)-sizeof(*scache)) {
    node=(mmap_area_t *)((char *)scache + sizeof(*scache) + i*(sizeof(*node)));
    /* initialize node */
    node->addr=NULL;
    node->pages=0;
    usrtc_node_init(&node->node,node);
    usrtc_insert(&(scache->raw),&node->node,&node->pages);
    i++;
  }
  /* okay we're need to mark that one page area was used */
  unode=usrtc_first(&scache->raw);
  usrtc_delete(&scache->raw,unode); /* delete from raw space */
  node=usrtc_node_getdata(unode);
  node->pages=1;
  node->addr=(void *)scache;
  usrtc_node_init(unode,node);
  usrtc_insert(&scache->alloc,unode,node->addr); /* insert to allocated space */
  
  goto _retry_alloc;

  sem_unlock(&glob_vpage_lock);

  return NULL;
}

status_t free_virt_pages(void *addr)
{
  mcache_t *scache = mhead;
  usrtc_node_t *node;
  mmap_area_t *marea;

  sem_lock(&glob_vpage_lock);

  if(!mhead) {
    sem_unlock(&glob_vpage_lock);
    return -EINVAL;
  }

  while(scache->next!=NULL) {
    node=usrtc_lookup(&scache->alloc,addr);
    if(node!=NULL)
      break;

    scache=scache->next;
  }

  if(!node) {
    sem_unlock(&glob_vpage_lock);
    return -EINVAL;
  }

  if(scache != scache->head) {
    usrtc_delete(&scache->alloc,node);
    marea=usrtc_node_getdata(node);
    usrtc_insert(&scache->free,node,&marea->pages);

    sem_unlock(&glob_vpage_lock);

    return 0;
  }

  sem_unlock(&glob_vpage_lock);

  return -EINVAL;
}

status_t validate_virt_addr(void *addr) /* TODO: align addr */
{
  mcache_t *scache = mhead;
  usrtc_node_t *node;

  sem_lock(&glob_vpage_lock);

  if(!mhead) {
    sem_unlock(&glob_vpage_lock);
    return -EINVAL;
  }

  while(scache->next!=NULL) {
    node=usrtc_lookup(&scache->alloc,addr);
    if(node!=NULL)
      break;

    scache=scache->next;
  }

  if(!node && addr > (void *)(__heap)) { 
    sem_unlock(&glob_vpage_lock);

    return 0;
  } else {
    sem_unlock(&glob_vpage_lock);

    return -EINVAL;
  }
}

/* implementation */

static mcache_t *add_alloc_mcache(mcache_t *head)
{
  mcache_t *ret;
  uintptr_t lup;

  sem_lock(&glob_mcache_lock);

  if(!mhead) {
    lup=(uintptr_t)&__heap;
    mhead=(void *)mmap(NULL,PAGE_SIZE,PROT_READ|PROT_WRITE,MAP_ANON|MAP_PRIVATE,NOFD,0);
    if(!mhead) 
      {
	for(;;);
      }
    if(!mhead) {
      sem_unlock(&glob_mcache_lock);

      return NULL;
    }

    last_used_ptr+=PAGE_SIZE;

    sem_unlock(&glob_mcache_lock);

    return mhead;
  } else {
    ret=mhead;
    while(ret->next!=NULL)
      ret=ret->next;
    ret->next=mmap(NULL,PAGE_SIZE,PROT_READ|PROT_WRITE,MAP_ANON|MAP_PRIVATE,NOFD,0);

    if(ret->next==NULL) {
      sem_unlock(&glob_mcache_lock);

      return NULL;
    }

    last_used_ptr+=PAGE_SIZE;
    ret->head=mhead;
    ret->next->prev=ret;
    ret->next->next=NULL;
    ret=ret->next;
  }

  sem_unlock(&glob_mcache_lock);

  return ret;
}

static int __memcache_compare(const void *a,const void *b)
{
  int *aa=(int *)a;
  int *bb=(int *)b;

  return *aa-*bb;
}

static int __memcache_compare_ptr(const void *a, const void *b)
{
  return a-b;
}

static int __stack_node_compare(const void *a,const void *b)
{
  const stack_node_t *node1=(const stack_node_t*)a;
  const stack_node_t *node2=(const stack_node_t*)b;

  if( node1->end_addr < node2->start_addr ) {
    return -1;
  } else if(node1->start_addr > node2->end_addr) {
    return 1;
  }

  return 0;
}

static stack_node_t *__alloc_stack_node(void)
{
  stack_node_t *node=(stack_node_t *)malloc(sizeof(stack_node_t));

  if( node ) {
    usrtc_node_init(&node->node,node);
    node->start_addr=0;
    node->end_addr=0;
  }

  return node;
}

static void __free_stack_node(stack_node_t *node)
{
  free(node);
}

void *alloc_stack_pages(size_t pages)
{
  void *addr=NULL;
  stack_node_t *node;

  if( !pages ) {
    return NULL;
  }

  sem_lock(&stack_lock);

  if( !stack_inited ) {
    stack_free_address=USER_MAX_VIRT - FIRST_TASK_MAX_STACK;
    usrtc_init(&stack_cache.free,USRTC_REDBLACK,USRTC_COUNT_T_MAX,
               __stack_node_compare);    
    usrtc_init(&stack_cache.active,USRTC_REDBLACK,USRTC_COUNT_T_MAX,
               __stack_node_compare);
    stack_inited=true;
  }

  if( usrtc_count( &stack_cache.free ) > 0 ) {
    goto insert_stack;
  } else {
    ulong_t size=pages<<PAGE_WIDTH;
    node=__alloc_stack_node();

    if( !node ) {
      goto out;
    }

    node->start_addr=stack_free_address-size;
    node->end_addr=stack_free_address;

    addr=mmap((void*)NULL,size,PROT_READ|PROT_WRITE,MAP_STACK|MAP_ANON|MAP_PRIVATE,NOFD,0);
    if( !addr ) {
      __free_stack_node(node);
      goto out;
    }
    //addr=(void*)stack_free_address; /* Stack grows down. */
    stack_free_address=node->start_addr-PAGE_SIZE; /* one guard page */
  }

insert_stack:
  usrtc_insert(&stack_cache.active,&node->node,(void *)node->start_addr);
out:
  sem_unlock(&stack_lock);
  return addr;
}
