/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 *
 * LibAlloc:
 * lib/smalloc.c: Simple implementation of malloc(), free() and related functions
 *                it's relatively simple, seems to be changed to more 
 *                efficient allocator
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <alloc/smalloc.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <lock.h>
#include <datalib/usrtc.h>

static ulong_t smalloc_lock;

sm_cache_t *scache[CACHES_AMOUNT];

static int __cache_list_comp(const void *a,const void *b);

static sm_cache_t *__init_scache(sm_cache_t *cache,int idx)
{
  int i;
  usrtc_t *frees,*allocs;
  sm_chunk_t *chunk;
  usrtc_node_t *node;
  uint8_t *swin,*uwin;

  if(cache)     return NULL; /* be safe */
  if(idx>(CACHES_AMOUNT-1)) return NULL; /* be safe too */

  cache=(sm_cache_t*)mmap(NULL,4096,MMAP_RW,NOFD,NULL);
  if(!cache) 
    return NULL;

  /* init values */
  i=idx+4;
  cache->size=(1 << i);
  cache->next=NULL;
  /* init structures */
  frees=&cache->free;
  allocs=&cache->alloc;
  usrtc_init(frees,USRTC_LIST,USRTC_COUNT_T_MAX,__cache_list_comp);
  usrtc_init(allocs,USRTC_REDBLACK,USRTC_COUNT_T_MAX,__cache_list_comp);
  /* now we're need to init nodes */
  swin=(uint8_t *)cache;
  uwin=swin;
  swin+=4096-(sizeof(sm_chunk_t)+(1 << i));
  for(uwin+=sizeof(*cache);uwin<=swin;uwin+=(sizeof(sm_chunk_t)+(1 << i))) {
    chunk=(sm_chunk_t*)uwin;
    node=&chunk->node;
    usrtc_node_setdata(node,chunk);
    usrtc_insert(frees,node,chunk);
  }

  return cache;
}

static int __calc_cache_size(size_t size)
{
  int i;

  if(size>MAX_CHUNK_SIZE)    return -EINVAL;
  if(size<=16) return 0;

  for(i=0;i<CACHES_AMOUNT;i++) {
    if(size>(1 << (i+4)) && size<=(1 << (i+5)))
      return i+1;
  }

  return -EINVAL;
}

void *smalloc(size_t size)
{
  int asize=__calc_cache_size(size);
  sm_cache_t *cache;
  usrtc_t *frees,*allocs;
  usrtc_node_t *node;
  uint8_t *p;

  sem_lock(&smalloc_lock);

  if(asize<0) {
    errno=EINVAL;
    return NULL;
  }

  if(!scache[asize]) {
    scache[asize]=__init_scache(scache[asize],asize);
    if(!scache[asize]) {
      sem_unlock(&smalloc_lock);
      return NULL;
    }
  } 

  cache=scache[asize];

  frees=&cache->free;
  allocs=&cache->alloc;
  if(usrtc_count(frees) > 0) {
    node=usrtc_first(frees);
    usrtc_delete(frees,node);
    usrtc_insert(allocs,node,usrtc_node_getdata(node));
    p=(uint8_t *)usrtc_node_getdata(node);
    sem_unlock(&smalloc_lock);
    return p+sizeof(sm_chunk_t);
  } else {
    while(cache) {
      if(cache->next==NULL) {
	cache->next=__init_scache(cache->next,asize);
	if(cache->next==NULL) {
	  sem_unlock(&smalloc_lock);
	  return NULL; /* oops, failed */
	}
	else cache=cache->next; /* yep, got it */
      } else 
	cache=cache->next;
      
      /* try to allocate */
      frees=&cache->free;
      allocs=&cache->alloc;
      if(usrtc_count(frees) > 0) {
	node=usrtc_first(frees);
	usrtc_delete(frees,node);
	usrtc_insert(allocs,node,usrtc_node_getdata(node));
	p=(uint8_t *)usrtc_node_getdata(node);
	sem_unlock(&smalloc_lock);
	return p+sizeof(sm_chunk_t);
      }

    }
  }

  sem_unlock(&smalloc_lock);

  return NULL;
}

void sfree(void *p)
{
  sm_cache_t *cache;
  sm_chunk_t *chunk;
  usrtc_node_t *node;
  usrtc_t *frees,*allocs;
  uint8_t *pp=p;

  if(!p) return; /* be safe */

  sem_lock(&smalloc_lock);

  pp-=sizeof(sm_chunk_t);
  chunk=(sm_chunk_t*)pp;

  /* safe checks */
  cache=chunk->cache;
  allocs=&cache->alloc;
  frees=&cache->free;
  node=usrtc_lookup(allocs,(const void*)pp);
  if(!node) {
    sem_unlock(&smalloc_lock);
    return;
  }
  usrtc_delete(allocs,node);
  usrtc_insert(frees,node,(const void*)pp);

  sem_unlock(&smalloc_lock);

  return;
}

void *scalloc(size_t n,size_t size)
{
  void *p;

  p=smalloc(n*size);
  if(!p)    return NULL;
  memset(p,'\0',n*size);

  return p;
}

static int __cache_list_comp(const void *a,const void *b)
{
  return a-b;
}
