/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 *
 * LibC:
 * ippq.c: ipc port queue
 *
 */

#include <sys/types.h>
#include <sys/syscall.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include <ipc/port.h>
#include <datalib/usrtc.h>

#include "__lowlevel.h"

#define DEFCNT  16

static ippq_t **__toippqs=NULL;
static int __ntoippqs=0;
/* local functions prototypes */
static int __init_global_toippqs(void);
static int __incr_global_toippqs(void);
static ippqid_t __lookup_global_toippqs(void);
/* default implementations */
static int __qfifo_cmp(const void *a,const void *b);
static int __qstack_cmp(const void *a,const void *b);
static int __qfifo_put(ippq_t *qu,void *data,size_t size,
		       ipc_port_msg_info_t inf);
static ippq_ent_t* __qfifo_get(ippq_t *,void *);
static ippq_ent_t* __qstack_get(ippq_t *,void *);
#define ___qfifo_get(a,b)   (ippq_ent_t *)usrtc_node_getdata(usrtc_first(a->queue));
#define ___qstack_get(a,b)   (ippq_ent_t *)usrtc_node_getdata(usrtc_last(a->queue));


#define __qstack_put  __qfifo_put

ippqid_t ippq_create(ulong_t flags,size_t kqueue_size,
		     ippq_type_t type,int qflags)
{
  ippqid_t id;
  prtid_t port_id;

  if(!__ntoippqs)
    if(__init_global_toippqs()!=0)
      return -ENOMEM;

  port_id=ipc_port_open(flags,kqueue_size);
  if(port_id<0)
    return port_id;
  else {
    id=__lookup_global_toippqs();
    if(id<0) {
    __close_goq:
      ipc_port_close(port_id);
      return -ENOMEM;
    }
  }

  __toippqs[id]->queue=malloc(sizeof(usrtc_t));
  if(!__toippqs[id]->queue) 
    goto __close_goq;
  __toippqs[id]->id=id;
  __toippqs[id]->port_id=port_id;
  __toippqs[id]->flags=qflags;
  __toippqs[id]->qtype=type;
  __toippqs[id]->data=NULL;
  __toippqs[id]->queue_put=NULL;
  __toippqs[id]->queue_get=NULL;
  __toippqs[id]->rep_hook=NULL;
  __toippqs[id]->del_hook=NULL;

  switch(__toippqs[id]->qtype) {
  case FIFO:
    usrtc_init(__toippqs[id]->queue,USRTC_LIST,USRTC_COUNT_T_MAX,__qfifo_cmp);
    __toippqs[id]->queue_put=__qfifo_put;
    __toippqs[id]->queue_get=__qfifo_get;
    break;
  case STACK:
    usrtc_init(__toippqs[id]->queue,USRTC_LIST,USRTC_COUNT_T_MAX,__qstack_cmp);
    __toippqs[id]->queue_put=__qstack_put;
    __toippqs[id]->queue_get=__qstack_get;
    break;
  case CUSTOM:
  default:
    /* dummy, user should to do it */
    break;
  }

  return id;
}

int ippq_destroy(ippqid_t qid)
{
  ippq_ent_t *ent;
  usrtc_node_t *node;

  if(__toippqs[qid]==NULL)
    return -1;

  node=usrtc_first(__toippqs[qid]->queue);
  while((node=usrtc_next(__toippqs[qid]->queue,node))!=NULL) {
    ent=(ippq_ent_t *)usrtc_node_getdata(node);
    if(!(__toippqs[qid]->flags & IPPQ_IMMEDIATE) && __toippqs[qid]->del_hook!=NULL)
      __toippqs[qid]->del_hook(__toippqs[qid],ent);
    usrtc_delete(__toippqs[qid]->queue,node);
    free(ent->data);
    free(ent);
    free(node);
  }

  ipc_port_close(__toippqs[qid]->port_id);

  free(__toippqs[qid]->queue);
  if(__toippqs[qid]->data)
    free(__toippqs[qid]->data);
  free(__toippqs[qid]);
  __toippqs[qid]=NULL;

  return 0;
}

int ippq_rcvmsg(ippqid_t qid,size_t size,void *buf)
{
  int r;

  if(__toippqs[qid]==NULL)
    return -1;

  r=ipc_port_rcvmsg(__toippqs[qid]->port_id,size,buf);
  if(r<0)
    return -1;

  /* we should reply now */
  if(__toippqs[qid]->flags & IPPQ_IMMEDIATE) 
    __toippqs[qid]->rep_hook(__toippqs[qid],buf,size,ipc_port_getinfo(__toippqs[qid]->port_id));
  
  __toippqs[qid]->queue_put(__toippqs[qid],(void *)buf,size,ipc_port_getinfo(__toippqs[qid]->port_id));

  return 0;
}

ippq_ent_t *ippq_getmsg(ippqid_t qid)
{
  ippq_ent_t *ent;
  usrtc_node_t *node;

  if(__toippqs[qid]==NULL)
    return NULL;

  ent=__toippqs[qid]->queue_get(__toippqs[qid],__toippqs[qid]->data);
  if(__toippqs[qid]->qtype==FIFO)
    node=usrtc_first(__toippqs[qid]->queue);
  else if(__toippqs[qid]->qtype==STACK)
    node=usrtc_last(__toippqs[qid]->queue);
  usrtc_delete(__toippqs[qid]->queue,node);
  free(node);

  return ent;
}

int ippq_rplmsg(ippqid_t qid,size_t size,void *buf,ippq_ent_t *ent)
{
  int r;

  if(__toippqs[qid]==NULL)
    return -1;

  r=__reply_ipc_port(__toippqs[qid]->port_id,ent->msginfo.msg_id,
		     buf,size);

  if(r<0)
    return r;

  return 0;
}

/* hooks load */
int ippq_set_rephook(ippqid_t qid,int (*rep_hook)(ippq_t *,void *,
						size_t,ipc_port_msg_info_t))
{
  if(__toippqs[qid]==NULL)
    return -1;

  __toippqs[qid]->rep_hook=rep_hook;

  return 0;
}

int ippq_set_delhook(ippqid_t qid,int (*del_hook)(ippq_t *,ippq_ent_t *))
{
  if(__toippqs[qid]==NULL)
    return -1;

  __toippqs[qid]->del_hook=del_hook;

  return 0;
}

int ippq_set_puthook(ippqid_t qid,int (*queue_put)(ippq_t *,void *,size_t,
						 ipc_port_msg_info_t))
{
  if(__toippqs[qid]==NULL)
    return -1;

  __toippqs[qid]->queue_put=queue_put;

  return 0;
}

int ippq_set_gethook(ippqid_t qid,ippq_ent_t* (*queue_get)(ippq_t *,void *))
{
  if(__toippqs[qid]==NULL)
    return -1;

  __toippqs[qid]->queue_get=queue_get;

  return 0;
}

void *ippq_ent_get_data(size_t *size,ippq_ent_t *ent)
{
  if(!ent) {
    *size=0;
    return NULL;
  }
  *size=ent->dsize;

  return ent->data;
}

void ippq_ent_destroy(ippq_ent_t *ent)
{
  if(!ent)
    return;

  free(ent->data);
  free(ent);

  return;
}

/* local functions */
static int __init_global_toippqs(void)
{
  __toippqs=malloc(sizeof(unative_t)*DEFCNT);
  if(!__toippqs)
    return -1;
  else
    memset(__toippqs,0,sizeof(unative_t)*DEFCNT);

  __ntoippqs=DEFCNT;

  return 0;
}

static int __incr_global_toippqs(void)
{
  __toippqs=realloc(__toippqs,sizeof(unative_t)*(__ntoippqs+DEFCNT));
  __ntoippqs+=DEFCNT;

  return 0;
}

static ippqid_t __lookup_global_toippqs(void)
{
  int r;
  ippqid_t qid=-1;

  for(r=0;r<__ntoippqs;r++) {
    if(__toippqs[r]==NULL) {
      __toippqs[r]=malloc(sizeof(ippq_t));
      if(!__toippqs[r]) 
	return -1;
      qid=r;
      break;
    }
  }

  if(qid<0) {
    if(!__incr_global_toippqs()) {
      qid=r+1;
      for(r+=1;r<__ntoippqs;r++) 
	__toippqs[r]=NULL;
      __toippqs[qid]=malloc(sizeof(ippq_t));
      if(!__toippqs[qid])
	return -1;
    }
  }

  return qid;
}

static int __qfifo_put(ippq_t *qu,void *data,size_t size,
		       ipc_port_msg_info_t inf)
{
  void *ndata;
  usrtc_node_t *node;
  ippq_ent_t *ent;

  if(!qu)
    return -1;

  if(!data || !size)
    return -1;

  ndata=malloc(size);
  node=malloc(sizeof(*node));
  if(!ndata || !node)
    return -1;

  ent=malloc(sizeof(*ent));
  if(!ent) {
    free(ndata);
    free(node);
    return -1;
  }

  memcpy(ndata,data,size);
  ent->data=ndata;
  ent->dsize=size;
  ent->msginfo=inf;

  usrtc_node_init(node,ent);
  usrtc_insert(qu->queue,node,&ent->msginfo.msg_id);

  return 0;
}

static ippq_ent_t* __qfifo_get(ippq_t *q,void *d)
{
  return ___qfifo_get(q,d);
}

static ippq_ent_t* __qstack_get(ippq_t *q,void *data)
{
  return ___qstack_get(q,d);
}

static int __qfifo_cmp(const void *a,const void *b)
{
  uint16_t *aa=(uint16_t *)a;
  uint16_t *bb=(uint16_t *)b;

  return *aa-*bb;
}

static int __qstack_cmp(const void *a,const void *b)
{
  uint16_t *aa=(uint16_t *)a;
  uint16_t *bb=(uint16_t *)b;

  return *bb-*aa;
}

