/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 *
 */

#include <limits.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <ipc/rpc.h>
#include <ipc/iov.h>

#include "__iovpath.h"

static int __normalize_path(char *path, size_t size)
{
  char dir[PATH_MAX];
  char *src, *dst, *end, s;

  if ((path == NULL) || (size == 0)) {
    errno = ENOENT;
    return -1;
  }
  memset(dir, 0, PATH_MAX);

  src = path;
  dst = dir;

  if (*src == '/') {
    src++;
    *dst++ = '/';
  } else {
    getcwd(dir, PATH_MAX);
    dst = dir + strlen(dir);
    if (*(dst - 1) != '/')
      *dst++ = '/';
  }

  while (*src != '\0') {
    end = strchr(src, '/');
    if (end == NULL)
      end = src + strlen(src);
    s = *end;
    *end = '\0';

    if ((strlen(src) == 0) || (strcmp(src, ".") == 0))
      goto next_part;
    if (strcmp(src, "..") == 0) {
      dst -= 2;
      while ((dst >= dir) && (*dst != '/'))
        dst--;
      if (dst < dir) {
	/* Can't step back so far */
        *end = s;
	errno = ENOENT;
	return -1;
      }
    } else {
      if (dst + strlen(src) >= dir + PATH_MAX) {
        /* Path is too long */
	*end = s;
	errno = ENAMETOOLONG;
	return -1;
      }
      strcat(dst, src);
      dst += strlen(src);
    }

    *dst++ = '/';
    *dst = '\0';

next_part:
    *end = s;
    src = end;
    if (*src == '/')
      src++;
  }

  /* Cut off trailing slash */
  if (strcmp(dir, "/") != 0)
    dst--;
  *dst = '\0';

  if (strlen(dir) >= size) {
    errno = ENOMEM;
    return -1;
  }

  strcpy(path, dir);
  return 0;
}

static int *__parse_path(int *amount, const char *path)
{
  int len, pp;
  int i, j;
  int *r;

  len = strlen(path);

  pp = 0;
  for (i = 0; i < len; i++)
    if (path[i] == '/')
      pp++;

  r = malloc(pp * sizeof(int));
  if (r == NULL)
    return NULL;

  j = 0;
  for (i = 0; i < len; i++)
    if (path[i] == '/') {
      r[j] = i;
      j++;
    }

  *amount = pp;
  return r;
}

int make_iovpath(const char *path, uni_rpc_t *rpc, iovec_t *ppoints, iovec_t *ppath)
{
  char *str;
  int err, plen, *pps, pp;

  if ((path == NULL) || (*path == '\0')) {
    errno = EINVAL;
    return -1;
  }
  if ((rpc == NULL) || (ppoints == NULL) || (ppath == NULL)) {
    errno = EINVAL;
    return -1;
  }
  ppoints->iov_base = ppath->iov_base = NULL;

  str = calloc(1, PATH_MAX);
  if (str == NULL)
    return -1;
  strcpy(str, path);

  if (__normalize_path(str, PATH_MAX) == -1) {
    err = errno;
    free(str);
    errno = err;
    return -1;
  }
  plen = strlen(str);

  pps = __parse_path(&pp, str);
  if (pps == NULL) {
    err = errno;
    free(str);
    errno = err;
    return -1;
  }

  rpc->arg5 = pp;
  rpc->arg6 = plen;
  ppoints->iov_base = pps;
  ppoints->iov_len = pp * sizeof(int);
  ppath->iov_base = str;
  ppath->iov_len = plen;

  return 0;
}

void free_iovpath(iovec_t *ppoints, iovec_t *ppath)
{
  if ((ppoints == NULL) || (ppath == NULL)) {
    errno = EINVAL;
    return;
  }

  free(ppoints->iov_base);
  free(ppath->iov_base);
}
