/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Paul Graphov <graphov@jarios.org>
 *
 */

#include <limits.h>
#include <sys/types.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ioctl.h>

#include <ipc/rpc.h>
#include <ipc/iov.h>

#include "__fildesc.h"

/*
  First note:
  If size field (_IOC_SIZE) of request is all ones in binary
  it means that size is passed to ioctl as a parameter.

  Second note:
  There are four ioctl directions now in Jari. (2 bits)
  first two are trivial - _IOC_READ and _IOC_WRITE are used when
  one way transfer is performed. In case of _IOC_READ buffer is
  not transferred to the device driver, it is only used to receive
  its answer. _IOC_WRITE is the opposite - buffer is only used to
  send data.
  Next two directions are used for two way transfers.
  _IOC_WRFIXED is used when the same buffer is used to send and
  receive data. For example it can be come structure which is
  partially filled in by the client and the driver fills remaining
  parts.
  _IOC_WRARBIT is used when two separate buffers are used. In this
  case size of the receiving buffer is always passed as a
  parameter. If sending buffer is also of arbitrary size (all ones)
  then it goes first.

  Third note:
  If size of some buffer is zero by definition of ioctl constant
  corresponding buffer argument should be omitted. If it is defined
  to be passed as parameter and corresponding parameter is zero it
  SHOULD NOT be omitted.

  So function is called this way:
  ioctl( fd, request, [out buf size], [in buf size], [out buf], [in buf])

*/

int ioctl(int fildes, int request, ...)
{
  int r=0, n1=1, n2=1;
  iovec_t ivec[2], ovec[2];
  va_list arglist;
  uni_rpc_t rpc;
  void *buf, *ibuf;
  size_t size;
  int need_buf = 1;

  if(!valid_fd(fildes)) {
    errno = EBADF;
    return -1;
  }

  va_start(arglist, request);

  rpc.rpc_call = RPC_IOCTL;
  rpc.arg1 = tfds[fildes].ufsid;

  _RPC_IOCTYPE(&rpc) = _IOC_TYPE(request);
  _RPC_IOCNR(&rpc) = _IOC_NR(request);
  _RPC_IOCDIR(&rpc) = _IOC_DIR(request);

  size = _IOC_SIZE(request);
  if(size == _IOC_SIZEMAX) {
    size = va_arg(arglist, size_t);
  } else if(size == 0) {
    need_buf = 0;
  }

  switch(_IOC_DIR(request)) {
    case _IOC_WRITE:
      rpc.size = size;
      rpc.offset = 0;
      break;
    case _IOC_READ:
      rpc.offset = size;
      rpc.size = 0;
      break;
    case _IOC_WRFIXED:
      rpc.size = size;
      rpc.offset = size;
      break;
    case _IOC_WRARBIT:
      rpc.size = size;
      rpc.offset = va_arg(arglist, size_t);
      break;
  }

  if(rpc.size > 0 || ((_IOC_DIR(request) == _IOC_WRARBIT) && need_buf)) {
    buf = va_arg(arglist, void*);
  }

  if(rpc.offset > 0 && _IOC_DIR(request) != _IOC_WRFIXED) {
    ibuf = va_arg(arglist, void*);
  }

  va_end(arglist);

  if(rpc.size == 0 && rpc.offset == 0) {
    r=ipc_channel_msgsnd(tfds[fildes].chid,sizeof(rpc),(void *)&rpc,
                         sizeof(rpc),(void *)&rpc);
  } else {
    ivec[0].iov_base = ovec[0].iov_base = &rpc;
    ivec[0].iov_len = ovec[0].iov_len = sizeof(rpc);

    switch(_IOC_DIR(request)) {
      case _IOC_WRITE:
        if(rpc.size > 0) {
          ovec[1].iov_base = buf;
          ovec[1].iov_len = rpc.size;
          n1++;
        }
        break;
      case _IOC_READ:
        if(rpc.offset > 0) {
          ivec[1].iov_base = ibuf;
          ivec[1].iov_len = rpc.offset;
          n2++;
        }
        break;
      case _IOC_WRFIXED:
        if(rpc.size > 0) {
          ivec[1].iov_base = buf;
          ivec[1].iov_len = rpc.size;
          ovec[1].iov_base = buf;
          ovec[1].iov_len = rpc.size;
          n2++;
          n1++;
        }
        break;
      case _IOC_WRARBIT:
        if(rpc.size > 0) {
          ovec[1].iov_base = buf;
          ovec[1].iov_len = rpc.size;
          n1++;
        }
        if(rpc.offset > 0) {
          ivec[1].iov_base = ibuf;
          ivec[1].iov_len = rpc.offset;
          n2++;
        }
        break;
    }
    r=ipc_iov_iov_send(tfds[fildes].chid, ovec, n1, ivec, n2);
  }

  if(r<0) {
    errno = -r;
    return -1;
  } else if(rpc.rpc_call != 0) {
    errno = rpc.rpc_call;
    return -1;
  } else {
    return rpc.size;
  }
}
