/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Dmitry Balakshin <balakshin@jarios.org>
 *
 */

#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <ipc/rpc.h>
#include <ipc/iov.h>

#include "__iovpath.h"
#include "__init.h"
#include "__fildesc.h"

static int __stat(unsigned rpc_call, const char *path, struct stat *buf);

int stat(const char *path, struct stat *buf)
{
  return __stat(RPC_STAT, path, buf);
}

int lstat(const char *path, struct stat *buf)
{
  return __stat(RPC_LSTAT, path, buf);
}

static int __stat(unsigned rpc_call, const char *path, struct stat *buf)
{
  int err;
  uni_rpc_t rpc;
  iovec_t send[3], recv[2];

  if ((rpc_call != RPC_STAT) && (rpc_call != RPC_LSTAT)) {
    errno = EINVAL;
    return -1;
  }

  if ((path == NULL) || (*path == '\0')) {
    errno = ENOENT;
    return -1;
  }
  if (buf == NULL) {
    errno = EINVAL;
    return -1;
  }

  rpc.rpc_call = rpc_call;
  send[0].iov_base = &rpc;
  send[0].iov_len = sizeof(rpc);

  if (make_iovpath(path, &rpc, send + 1, send + 2) == -1) {
    err = errno;
    goto done;
  }

  recv[0].iov_base = &rpc;
  recv[0].iov_len  = sizeof(uni_rpc_t);
  recv[1].iov_base = buf;
  recv[1].iov_len  = sizeof(struct stat);

  err = ipc_iov_iov_send(vfs_channel, send, 3, recv, 2);
  if (err < 0) {
    err = -err;
    goto done;
  }
  if (rpc.rpc_call != 0) {
    err = rpc.rpc_call;
    goto done;
  }

  err = 0;

done:
  free_iovpath(send + 1, send + 2);
  errno = err;
  return (err == 0) ? 0 : -1;
}

int fstat(int fildes, struct stat *buf)
{
  int err;
  uni_rpc_t rpc;
  iovec_t send, recv[2];

  if (!valid_fd(fildes)) {
    errno = EBADF;
    return -1;
  }
  if (buf == NULL) {
    errno = EINVAL;
    return -1;
  }

  rpc.rpc_call = RPC_FSTAT;
  rpc.arg1 = tfds[fildes].ufsid;

  send.iov_base = &rpc;
  send.iov_len = sizeof(uni_rpc_t);

  recv[0].iov_base = &rpc;
  recv[0].iov_len = sizeof(uni_rpc_t);
  recv[1].iov_base = buf;
  recv[1].iov_len = sizeof(struct stat);

  err = ipc_iov_iov_send(tfds[fildes].chid, &send, 1, recv, 2);
  if (err < 0) {
    errno = -err;
    return -1;
  }
  if (rpc.rpc_call != 0) {
    errno = rpc.rpc_call;
    return -1;
  }

  return 0;
}
