/* This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Michael Tsymbalyuk <mtzaurus@gmail.com>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 *
 * libc:
 *   general/libc/proc/exec.c
 *
 * execXX() family functions
 */

#include <limits.h>
#include <sys/types.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>
#include <alloca.h>
#include <stdlib.h>
#include <unistd.h>

#include <ns/ns.h>
#include <ipc/channel.h>
#include <ipc/iov.h>
#include <sys/syscall.h>
#include <sys/procmgr.h>

#include "__procmgr.h"

static int __setup_args(const char *filename, char *const argv[],
                               pm_exec_request_t *rq);



int execv(const char *path, char *const argv[])
{
  return execve(path, argv, environ);
}

int execve(const char *path, char *const argv[], char *const envp[])
{
  int rc;
  pm_exec_request_t rq;
  pm_cmd_header_t rqh;
  iovec_t iov[2];
  pid_t pid;

  if ((path == NULL) || (*path == '\0')) {
    errno = ENOENT;
    return -1;
  }
  if (strlen(path) > PATH_MAX) {
    errno = ENAMETOOLONG;
    return -1;
  }
  if (argv != NULL) {
    for (rc = 0; argv[rc] != NULL; rc ++)
            ;
    if (rc > ARG_MAX) {
      errno = E2BIG;
      return -1;
    }
  }
  if (envp != NULL) {
    for (rc = 0; envp[rc] != NULL; rc++)
            ;
    if (rc > ARG_MAX) {
      errno = E2BIG;
      return -1;
    }
  }

  if (__proc_mgr_chid == (chid_t)-1) {
    open_pm_connection();
    if (__proc_mgr_chid == (chid_t)-1) {
      errno = EPIPE;
      return -1;
    }
  }

  rc = __setup_args(path, argv, &rq);
  if (rc < 0) {
    errno = -rc;
    return -1;
  }
  rq.argsize = rc;

  rqh.cmd = __PROCMGR_CMD_EXEC;

  iov[0].iov_base = &rqh;
  iov[0].iov_len = sizeof(rqh);
  iov[1].iov_base = &rq;
  iov[1].iov_len = sizeof(rq);

  rc = ipc_iov_send(__proc_mgr_chid, iov, 2, &pid, sizeof(pid));

  /* We return here only in case of errors. */
  if (rc < 0)
    errno = -rc;
  if (pid < 0)
    errno = -pid;

  return -1;
}

int execl(const char *path, const char *arg0, ... /*, (char *)0 */)
{
  va_list ap;
  const char *argv[ARG_MAX + 1];
  size_t i = 0;

  va_start(ap, arg0);
  while ((arg0 != NULL) && (i < ARG_MAX)) {
    argv[i++] = arg0;
    arg0 = va_arg(ap, const char *);
  }
  argv[i] = NULL;
  va_end(ap);

  return execv(path, argv);
}

int execle(const char *path, const char *arg0, ... /*, (char *)0, char *const envp[]*/)
{
  va_list ap;
  const char *argv[ARG_MAX + 1];
  const char *const *envp;
  size_t i;

  va_start(ap, arg0);

  i = 0;
  while ((arg0 != NULL) && (i < ARG_MAX)) {
    argv[i++] = arg0;
    arg0 = va_arg(ap, const char *);
  }
  argv[i] = NULL;

  /* skip to (char *)0 */
  while (arg0 != NULL)
    arg0 = va_arg(ap, const char *);
  envp = va_arg(ap, const char * const *);

  va_end(ap);

  return execve(path, argv, envp);

}

int execlp(const char *file, const char *arg0, ... /*, (char *)0 */)
{
  va_list ap;
  const char *argv[ARG_MAX + 1];
  size_t i = 0;

  va_start(ap, arg0);
  while ((arg0 != NULL) && (i < ARG_MAX)) {
    argv[i++] = arg0;
    arg0 = va_arg(ap, const char *);
  }
  argv[i] = NULL;
  va_end(ap);

  return execvp(file, argv);
}

int execvp(const char *file, char *const argv[])
{
  char *path = getenv("PATH");
  char *buf, *p;

  if (file == NULL) {
    errno = ENOENT;
    return -1;
  }

  if (path == NULL)
    return execv(file, argv);

  /* 'file' must be just a filename, without any path */
  if (strchr(file, '/') != NULL)
    return execv(file, argv);

  buf = alloca(PATH_MAX);
  if (buf == NULL) {
    errno = ENOMEM;
    return -1;
  }

  while (*path != '\0') {
    memset(buf, 0, PATH_MAX);

    p = strchr(path, ':');

    if (p == path) {
    /* The "::" means current directory */
      if (getcwd(buf, PATH_MAX) == NULL)
        return -1;
      path++;
    } else if (p == NULL) {
      strcpy(buf, path);
      path += strlen(path);
    } else {
      memcpy(buf, path, p - path);
      path += p - path + 1;
    }
    if (buf[strlen(buf) - 1] != '/')
	strcat(buf, "/");
    strcat(buf, file);

    if (access(buf, X_OK) == 0)
      return execv(buf, argv);
  }

  errno = ENOENT;
  return -1;
}


static int __setup_args(const char *filename, char *const argv[],
                               pm_exec_request_t *rq)
{
  int space_left=PATH_MAX+PM_ARGV_SIZE;
  int arg=0,offset=0;
  int len,i=0;

  offset=strlen(filename)+1;
  if( offset > space_left ) {
    return -ENOMEM;
  }

  strcpy(rq->args,filename);
  space_left-=offset;
  rq->arg_offsets[0]=0;
  arg++;
  if(argv==NULL) goto __endparse;

  while(argv[i]!=NULL) {
    len=strlen(argv[i])+1;

    if( len > space_left ) {
      return -ENOMEM;
    }

    strcpy(rq->args+offset,argv[i]);
    space_left-=len;
    rq->arg_offsets[i+1]=offset;
    i++;arg++;
    offset+=len;
  }

  if( arg >= ARG_MAX ) {
    return -E2BIG;
  }

 __endparse:

  rq->numargs=arg;
  return offset;
}
