/* This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 *
 * LibC:
 * general/libc/stdio/vswprintf.c
 */

/*
 * Not implemented:
 *	Flags: '
 *	Specifiers: f, F, e, E, g, G, a, A, n
 *	Usage of wide characters with 'c' and 's' specifiers.
 *	A field width/precision with an asterisk ('*')
 */

#include <stdint.h>
#include <sys/types.h>
#include <ctype.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <stdio.h>
#include <wchar.h>
#include <wctype.h>

enum flag {
	FLAG_1000		= (1 << 0),		/* ' */
	FLAG_JUSTIFY		= (1 << 1),		/* - */
	FLAG_SIGN		= (1 << 2),		/* + */
	FLAG_SPACE_PREFIX	= (1 << 3),		/* <space> */
	FLAG_ALTERNATIVE	= (1 << 4),		/* # */
	FLAG_ZERO		= (1 << 5)		/* 0 */
};

enum len_mod {
	MOD_NONE,
	MOD_INT8,		/* hh */
	MOD_INT16,		/* h */
	MOD_L,			/* l */
	MOD_LL,			/* ll */
	MOD_INTMAX,		/* j */
	MOD_SIZE,		/* z */
	MOD_PTRDIFF		/* t */
};

enum spec {
	SPEC_INT,		/* d, i */
	SPEC_OCT,		/* o */
	SPEC_UINT,		/* u */
	SPEC_HEX,		/* x */
	SPEC_HEX_UPPERCASE,	/* X */
	SPEC_CHAR,		/* c */
	SPEC_STRING,		/* s */
	SPEC_PTR		/* p */
};

struct arg_ref{
	int valid;		/* 0 - unused, 1 - can be used */
	unsigned value;
	int use_arg;		/* 1 - value is the argument number
				 * 0 - value is the actual number to use
				 */
};

struct unit {
	unsigned int	flags;
	struct arg_ref	width;
	struct arg_ref	precision;
	enum len_mod	mod;
	enum spec	spec;
};

static unsigned str2uint(const wchar_t *str);
static void int2str(const struct unit *unit, uintmax_t value, wchar_t *buf, size_t size);
static int parse_unit(const wchar_t *format, struct unit *unit);
static wchar_t* as_int(const struct unit *unit, const void *val);
static wchar_t* as_string(const struct unit *unit, const char *val);
static wchar_t* as_wstring(const struct unit *unit, const wchar_t *val);
static wchar_t* as_char(const struct unit *unit, char val);
static wchar_t* as_wchar(const struct unit *unit, wchar_t val);
static wchar_t* as_ptr(const struct unit *unit, const void *val);


/* Do not use strtol()/atoi() because they alter errno.
 * We expect only positive values, so -1 as result means error.
 */
static unsigned str2uint(const wchar_t *str)
{
	int rc = 0, power = 1;
	size_t len, i;

	len = wcslen(str);

	for (i = len; i != 0; i--) {
		if ((str[i - 1] < L'0') || (str[i - 1] > L'9'))
			return (unsigned)-1;
		if ((str[i - 1] - L'0') * power < 0)
			return (unsigned)-1;			/* overflow */
		rc += (str[i - 1] - L'0') * power;
		power *= 10;
	}

	return rc;
}

/* Buffer is expected to be large enough */
static void int2str(const struct unit *unit, uintmax_t value, wchar_t *buf, size_t size)
{
	const wchar_t DIGITS[16] = L"0123456789abcdef";
	unsigned base;
	size_t i, power;
	wchar_t c;

	switch (unit->spec) {
	case SPEC_INT:
	case SPEC_UINT:
		base = 10;
		break;
	case SPEC_OCT:
		base = 8;
		break;
	case SPEC_HEX:
	case SPEC_HEX_UPPERCASE:
		base = 16;
		break;
	default:
		/* should not be reached */
		return;
	}

	i = 0;
	power = 0;
        do {
		c = DIGITS[value % base];
		if (unit->spec == SPEC_HEX_UPPERCASE)
			c = towupper(c);
		buf[size - i - 1] = c;

		if (((unit->spec == SPEC_INT) || (unit->spec == SPEC_UINT)) &&
				(unit->flags & FLAG_1000)) {
			if ((power != 0) && (power % 3 == 0)) {
				i++;
				buf[size - i - 1] = L'\'';
			}
		}

		value /= base;
		power++;
		i++;

	} while (value != 0);

	memmove(buf, buf + (size - i), i);
	buf[i] = L'\0';
}

static int parse_unit(const wchar_t *format, struct unit *unit)
{
	const wchar_t *p = format;
	wchar_t buf[16];
	size_t i;
	int ignore_precision;

	if (*p != L'%')
		return 0;
	p++;

	/* Zero or more flags */
	unit->flags = 0;
	for (;; p++) {
		if (*p == L'\0')
			return -1;

		switch (*p) {
		case L'\'':
			unit->flags |= FLAG_1000;
			break;
		case L'-':
			unit->flags |= FLAG_JUSTIFY;
			break;
		case L'+':
			unit->flags |= FLAG_SIGN;
			break;
		case L' ':
			unit->flags |= FLAG_SPACE_PREFIX;
			break;
		case L'#':
			unit->flags |= FLAG_ALTERNATIVE;
			break;
		case L'0':
			unit->flags |= FLAG_ZERO;
			break;
		default:
			goto done_flags;
		}
	}
done_flags:
	/* Some flags discard other flags */
	if ((unit->flags & FLAG_SIGN) && (unit->flags & FLAG_SPACE_PREFIX))
		unit->flags &= ~FLAG_SPACE_PREFIX;
	if ((unit->flags & FLAG_JUSTIFY) && (unit->flags & FLAG_ZERO))
		unit->flags &= ~FLAG_ZERO;

	/* Optional minimum field width */
	unit->width.valid = 0;

	if (*p == L'.')
		goto done_width;
	unit->width.use_arg = 0;
#if 0
	if (*p == L'*') {
		unit->width.use_arg = 1;
		p++;
		if (*p == L'\0')
			return -1;
	}
#endif
	/* get the length of the number */
	i = 0;
	while ((p[i] >= L'0') && (p[i] <= L'9'))
		i++;
	if (i == 0) {
		if (unit->width.use_arg)
			return -1;
		goto done_width;	/* no field width */
	}
	if (p[i] == L'\0')
		return -1;
	if (i < 16) {
		memset(buf, 0, sizeof(buf));
		wmemcpy(buf, p, i);
		unit->width.value = str2uint(buf);
		if (unit->width.value != (unsigned)-1)
			unit->width.valid = 1;
	}
	p += i;

done_width:
	/* Optional precision */
	unit->precision.valid = 0;

	if (*p != L'.')
		goto done_precision;
	p++;
	if (*p == L'\0')
		return -1;

	unit->precision.use_arg = 0;
#if 0
	if (*p == L'*') {
		unit->precision.use_arg = 1;
		p++;
		if (*p == L'\0')
			return -1;
	}
#endif
	ignore_precision = 0;
	if (*p == L'-') {
		ignore_precision = 1;
		p++;
		if (*p == L'\0')
			return -1;
	}

	/* get the length of the number */
	i = 0;
	while ((p[i] >= L'0') && (p[i] <= L'9'))
		i++;
	if (i == 0) {
		/* special case - null digit string */
		if (unit->precision.use_arg)
			return -1;
		unit->precision.value = 0;
		if (!ignore_precision)
			unit->precision.valid = 1;
		goto done_precision;
	}
	if (p[i] == L'\0')
		return -1;
	if (i < 16) {
		memset(buf, 0, sizeof(buf));
		wmemcpy(buf, p, i);
		unit->precision.value = str2uint(buf);
		if (unit->precision.value != (unsigned)-1) {
			if (!ignore_precision)
				unit->precision.valid = 1;
		}
	}
	p += i;

done_precision:
	/* Optional length modifier */
	unit->mod = MOD_NONE;
	switch (*p) {
	case L'h':
		unit->mod = MOD_INT16;
		if (*(p + 1) == L'h') {
			unit->mod = MOD_INT8;
			p++;
		}
		break;
	case L'l':
		unit->mod = MOD_L;
		if (*(p + 1) == L'l') {
			unit->mod = MOD_LL;
			p++;
		}
		break;
	case L'j':
		unit->mod = MOD_INTMAX;
		break;
	case L'z':
		unit->mod = MOD_SIZE;
		break;
	case L't':
		unit->mod = MOD_PTRDIFF;
		break;
	}

	if (unit->mod != MOD_NONE) {
		p++;
		if (*p == L'\0')
			return -1;
	}

	/* Conversion specifier character */
	switch (*p) {
	case L'd':
	case L'i':
		unit->spec = SPEC_INT;
		if (unit->precision.valid)
			unit->flags &= ~FLAG_ZERO;
		break;
	case L'o':
		unit->spec = SPEC_OCT;
		if (unit->precision.valid)
			unit->flags &= ~FLAG_ZERO;
		break;
	case L'u':
		unit->spec = SPEC_UINT;
		if (unit->precision.valid)
			unit->flags &= ~FLAG_ZERO;
		break;
	case L'x':
		unit->spec = SPEC_HEX;
		if (unit->precision.valid)
			unit->flags &= ~FLAG_ZERO;
		break;
	case L'X':
		unit->spec = SPEC_HEX_UPPERCASE;
		if (unit->precision.valid)
			unit->flags &= ~FLAG_ZERO;
		break;
	case L'c':
		unit->spec = SPEC_CHAR;
		break;
	case L's':
		unit->spec = SPEC_STRING;
		break;
	case L'p':
		unit->spec = SPEC_PTR;
		break;
	case L'C':
		unit->spec = SPEC_CHAR;
		unit->mod = MOD_L;
		break;
	case L'S':
		unit->spec = SPEC_STRING;
		unit->mod = MOD_L;
		break;
	default:
		return -1;
	}

	return (p - format) + 1;
}

static wchar_t* as_int(const struct unit *unit, const void *val)
{
	unsigned char v8	= *((const unsigned char*)val);
	unsigned short v16	= *((const unsigned short*)val);
	unsigned v		= *((const unsigned*)val);
	unsigned long vl	= *((const unsigned long*)val);
	unsigned long long vll	= *((const unsigned long long*)val);
	uintmax_t vintmax	= *((const uintmax_t*)val);
	size_t vsize		= *((const size_t*)val);
	ptrdiff_t vptrdiff	= *((const ptrdiff_t*)val);

	wchar_t num[32], prefix[4];
	int sign;
	unsigned j, len;
	unsigned plen;		/* additional precision length */
	wchar_t *buf;

	/* 1. Make 'number' string */

	memset(num, 0, sizeof(num));
	sign = 0;

	switch (unit->mod) {
	case MOD_NONE:
		if (unit->spec == SPEC_INT)
			sign = ((int)v < 0);
		if (sign)
			v = -((int)v);
		int2str(unit, v, num, sizeof(num)/sizeof(wchar_t));
		break;
	case MOD_INT8:
		if (unit->spec == SPEC_INT)
			sign = ((signed char)v8 < 0);
		if (sign)
			v8 = -((signed char)v8);
		int2str(unit, v8, num, sizeof(num)/sizeof(wchar_t));
		break;
	case MOD_INT16:
		if (unit->spec == SPEC_INT)
			sign = ((signed short)v16 < 0);
		if (sign)
			v16 = -((signed short)v16);
		int2str(unit, v16, num, sizeof(num)/sizeof(wchar_t));
		break;
	case MOD_L:
		if (unit->spec == SPEC_INT)
			sign = ((signed long)vl < 0);
		if (sign)
			vl = -((signed long)vl);
		int2str(unit, vl, num, sizeof(num)/sizeof(wchar_t));
		break;
	case MOD_LL:
		if (unit->spec == SPEC_INT)
			sign = ((signed long long)vll < 0);
		if (sign)
			vll = -((signed long long)vll);
		int2str(unit, vll, num, sizeof(num)/sizeof(wchar_t));
		break;
	case MOD_INTMAX:
		if (unit->spec == SPEC_INT)
			sign = ((intmax_t)vintmax < 0);
		if (sign)
			vintmax = -((intmax_t)vintmax);
		int2str(unit, vintmax, num, sizeof(num)/sizeof(wchar_t));
		break;
	case MOD_SIZE:
		/* size_t expected to be unsigned*/
		sign = 0;
		int2str(unit, vsize, num, sizeof(num)/sizeof(wchar_t));
		break;
	case MOD_PTRDIFF:
		if (unit->spec == SPEC_INT)
			sign = (vptrdiff < 0);
		if (sign)
			vptrdiff = -vptrdiff;
		int2str(unit, vptrdiff, num, sizeof(num)/sizeof(wchar_t));
		break;
	}
	len = wcslen(num);

	plen = 0;
        if (unit->precision.valid) {
		/* minimum number of digits to apper */
		if (len < unit->precision.value)
			plen = unit->precision.value - len;
	}

	/* 2. Make prefix string */

	memset(prefix, 0, sizeof(prefix));

	/* Flags FLAG_SIGN and FLAG_SPACE_PREFIX are mutual exclusive.
	 * See parse_unit()
	 */
	if (unit->flags & FLAG_SIGN) {
		if (sign)
			prefix[0] = L'-';
		else
			prefix[0] = L'+';
	}
	if (unit->flags & FLAG_SPACE_PREFIX) {
		if (sign)
			prefix[0] = L'-';
		else
			prefix[0] = L' ';
	}
	if (unit->spec == SPEC_INT) {
		if (sign)
			prefix[0] = L'-';
	}

	if (unit->flags & FLAG_ALTERNATIVE) {
		if (unit->spec == SPEC_OCT)
			wcscat(prefix, L"0");
		if (unit->spec == SPEC_HEX)
			wcscat(prefix, L"0x");
		if (unit->spec == SPEC_HEX_UPPERCASE)
			wcscat(prefix, L"0X");
	}

	/* 3. Make final string with needed width, padding and '0' filling */

	if (!unit->width.valid || (wcslen(prefix) + plen + len >= unit->width.value)) {
		buf = calloc(1, (wcslen(prefix) + plen + len + 1) * sizeof(wchar_t));
		if (buf == NULL)
			return NULL;
		wcscpy(buf, prefix);
		wmemset(buf + wcslen(prefix), L'0', plen);
		wcscat(buf, num);
		return buf;
	}

	buf = calloc(1, (unit->width.value + 1) * sizeof(wchar_t));
	if (buf == NULL)
		return NULL;

	/* Flags FLAG_ZERO and FLAG_JUSTIFY are mutually exclusive.
	 * See parse_unit()
	 */
	if (unit->flags & FLAG_ZERO) {
		wcscpy(buf, prefix);
		wmemset(buf + wcslen(prefix), L'0', unit->width.value - (wcslen(prefix) + len));
		wcscat(buf, num);
	} else {
		wmemset(buf, L' ', unit->width.value);
		if (unit->flags & FLAG_JUSTIFY)
			j = 0;
		else
			j = unit->width.value - (wcslen(prefix) + plen + len);
		wmemcpy(buf + j, prefix, wcslen(prefix));
		wmemset(buf + j + wcslen(prefix), L'0', plen);
		wmemcpy(buf + j + wcslen(prefix) + plen, num, len);
	}

	return buf;
}

static wchar_t* as_string(const struct unit *unit, const char *val)
{
	static const char NULL_MSG[] = "<NULL>";

	size_t l;		/* string length */
	size_t sz;		/* buffer size */
	size_t i, j;
	const char *v;
	wchar_t *buf, *p;
	mbstate_t ps;
	size_t r;

	/* width     - minimum width
	 * precision - max width
	 * flag[-]   - left/right justification
	 */

	if (val == NULL)
		v = NULL_MSG;
	else
		v = val;

	for (l = 0; v[l] != '\0'; l++)
		if (unit->precision.valid && (l == unit->precision.value))
			break;

	sz = l;
	if (unit->width.valid) {
		if (sz < unit->width.value)
			sz = unit->width.value;
	}

	buf = calloc(1, (sz + 1) * sizeof(wchar_t));
	if (buf == NULL)
		return NULL;

	p = buf;
	i = 0;
	if ((unit->width.valid) && !(unit->flags & FLAG_JUSTIFY)) {
		for (; i < sz - l; i++)
			*p++ = L' ';
	}

	for (j = 0; j < l; ) {
		r = mbrtowc(p, v + j, MB_CUR_MAX, &ps);
		if ((r == (size_t)-1) || (r == (size_t)-2)) {
			free(buf);
			return NULL;
		}
		j += r;
		p++;
	}
	i += l;
	for (; i < sz; i++)
		*p++ = L' ';

	*p = L'\0';
	return buf;
}

static wchar_t* as_wstring(const struct unit *unit, const wchar_t *val)
{
	static const wchar_t NULL_MSG[] = L"<NULL>";

	size_t l;		/* string length */
	size_t sz;		/* buffer size */
	size_t i, j;
	const wchar_t *v;
	wchar_t *buf, *p;

	/* width     - minimum width
	 * precision - max width
	 * flag[-]   - left/right justification
	 */

	if (val == NULL)
		v = NULL_MSG;
	else
		v = val;

	for (l = 0; v[l] != L'\0'; l++)
		if (unit->precision.valid && (l == unit->precision.value))
			break;

	sz = l;
	if (unit->width.valid) {
		if (sz < unit->width.value)
			sz = unit->width.value;
	}

	buf = calloc(1, (sz + 1) * sizeof(wchar_t));
	if (buf == NULL)
		return NULL;

	p = buf;
	i = 0;
	if ((unit->width.valid) && !(unit->flags & FLAG_JUSTIFY)) {
		for (; i < sz - l; i++)
			*p++ = L' ';
	}
	for (j = 0; j < l; j++) {
		*p = v[j];
		p++;
	}
	i += l;
	for (; i < sz; i++)
		*p++ = L' ';

	*p = L'\0';
	return buf;

}

static wchar_t* as_char(const struct unit *unit, char val)
{
	char buf[2];
	struct unit u;

	u = *unit;
	u.precision.valid = 0;

	buf[0] = val;
	buf[1] = '\0';
	return as_string(&u, buf);
}

static wchar_t* as_wchar(const struct unit *unit, wchar_t val)
{
	wchar_t buf[2];
	struct unit u;

	u = *unit;
	u.precision.valid = 0;

	buf[0] = val;
	buf[1] = L'\0';
	return as_wstring(&u, buf);
}

static wchar_t* as_ptr(const struct unit *unit, const void *val)
{
	struct unit u = {
		.flags = unit->flags | FLAG_ALTERNATIVE,
		.width = unit->width,
		.precision = { .valid = 0 },
		.mod = MOD_INTMAX,
		.spec = SPEC_HEX
	};
	uintmax_t v = (uintmax_t)val;

	return as_int(&u, &v);
}

int vswprintf(wchar_t *str, size_t size, const wchar_t *format, va_list ap)
{
	const wchar_t *p;	/* current position in /format/ argument */
	int c;			/* number of chars in current format unit */
	size_t l;		/* number of bytes written */
	struct unit u;
	wchar_t *buf;		/* value of argument, converted according to
				 * current format unit
				 */
	wchar_t *w;		/* current position in /str/ argument or
				 * NULL if we shouldn't write anything
				 */

	unsigned char v8;
	unsigned short v16;
	unsigned v;
	unsigned long vl;
	unsigned long long vll;
	uintmax_t vintmax;
	size_t vsize;
	ptrdiff_t vptrdiff;
	void *vptr;
	char symb;
	wchar_t wsymb;

	if ((str == NULL) && (size != 0)) {
		errno = EINVAL;
		return -1;
	}
	if (format == NULL) {
		errno = EINVAL;
		return -1;
	}

	if (size == 0)
		w = NULL;
	else
		w = str;

	l = 0;
	p = format;
	buf = NULL;

	while (*p != L'\0') {
		if (*p != L'%') {
			l++;
			if (w != NULL) {
				*w = *p;
				w++;
				size--;
				if (size == 0)
					break;
			}
			p++;
			continue;
		}
		if (*(p + 1) == L'%') {
			/* %% */
			l++;
			if (w != NULL) {
				*w = L'%';
				w++;
				size--;
				if (size == 0)
					break;
			}
			p += 2;
			continue;
		}
		c = parse_unit(p, &u);
		if (c == -1) {
			l++;
			if (w != NULL) {
				*w = *p;
				w++;
				size--;
				if (size == 0)
					break;
			}
			p++;
			continue;
		}

		/* Read argument */
		switch (u.spec) {
		case SPEC_INT:
		case SPEC_OCT:
		case SPEC_UINT:
		case SPEC_HEX:
		case SPEC_HEX_UPPERCASE:
			switch (u.mod) {
			case MOD_INT8:
				v8 = (unsigned char)(va_arg(ap, int) & 0xff);
				buf = as_int(&u, &v8);
				break;
			case MOD_INT16:
				v16 = (unsigned short)(va_arg(ap, int) & 0xffff);
				buf = as_int(&u, &v16);
				break;
			case MOD_L:
				vl = va_arg(ap, unsigned long);
				buf = as_int(&u, &vl);
				break;
			case MOD_LL:
				vll = va_arg(ap, unsigned long long);
				buf = as_int(&u, &vll);
				break;
			case MOD_INTMAX:
				vintmax = va_arg(ap, uintmax_t);
				buf = as_int(&u, &vintmax);
				break;
			case MOD_SIZE:
				vsize = va_arg(ap, size_t);
				buf = as_int(&u, &vsize);
				break;
			case MOD_PTRDIFF:
				vptrdiff = va_arg(ap, ptrdiff_t);
				buf = as_int(&u, &vptrdiff);
				break;
			default:
				v = va_arg(ap, unsigned);
				buf = as_int(&u, &v);
			}
			break;
		case SPEC_CHAR:
			if (u.mod != MOD_L) {
				symb = (char)(va_arg(ap, int) & 0xff);
				buf = as_char(&u, symb);
			} else {
				wsymb = (wchar_t)(va_arg(ap, wint_t));
				buf = as_wchar(&u, wsymb);
			}
			break;
		case SPEC_STRING:
			vptr = va_arg(ap, void*);
			if (u.mod != MOD_L)
				buf = as_string(&u, vptr);
			else
				buf = as_wstring(&u, vptr);
			break;
		case SPEC_PTR:
			vptr = va_arg(ap, void*);
			buf = as_ptr(&u, vptr);
			break;
		}

		if (buf == NULL)
			return -1;

		if (w == NULL)
			l += wcslen(buf);
		else {
			vsize = (size > wcslen(buf)) ? wcslen(buf) : size;
			l += vsize;
			wcsncpy(w, buf, vsize);
			w += vsize;
			size -= vsize;
			if (size == 0)
				break;
		}
		p += c;

		free(buf);
		buf = NULL;
	}

	if (buf != NULL)
		free(buf);

	if (w != NULL) {
		if (size == 0)
			*(w - 1) = L'\0';
		else
			*w = L'\0';
	}

	return l;
}
