/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 *
 * LibC: handling environment variables
 */

#include <stdint.h>
#include <limits.h>
#include <sys/types.h>
#include <bitwise.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include <ns/ns.h>
#include <ipc/rpc.h>
#include <ipc/iov.h>
#include <ipc/channel.h>

/* see general/libc/ns/ns.c */
extern chid_t nx_ch;

/* POSIX variable */
char* environ[ARG_MAX + 1] = { NULL };

/* Return 0 on success, errcode otherwise */
static int __updenv(int envid, const char *envval);
static int __setenv(int envid, const char *envname, const char *envval);
static int __unsetenv(int envid);
int __getenviron(void);

char *getenv(const char *name)
{
  size_t i;

  if (name == NULL) {
    errno = EINVAL;
    return NULL;
  }
  if (*name == '\0')
    return NULL;
  if (strchr(name, '=') != NULL) {
    errno = EINVAL;
    return NULL;
  }

  for (i = 0; environ[i] != NULL; i++) {
    if (strncmp(environ[i], name, strlen(name)) == 0) {
      if (environ[i][strlen(name)] == '=')
        break;
    }
  }
  if (environ[i] == NULL)
    return NULL;

  return environ[i] + strlen(name) + 1;
}


int setenv(const char *envname, const char *envval, int overwrite)
{
  int i, err;
  size_t l;
  char *buf;

  if ((envname == NULL) || (*envname == '\0') ||
          (strchr(envname, '=') != NULL)) {
    errno = EINVAL;
    return -1;
  }

  l = strlen(envname) + 1;
  if (envval != NULL)
    l += strlen(envval);
  l++;
  buf = malloc(l);
  if (buf == NULL)
    return -1;
  strcpy(buf, envname);
  strcat(buf, "=");
  if (envval != NULL)
    strcat(buf, envval);
  buf[l - 1] = '\0';

  /* check for already defined env variable */
  for (i = 0; environ[i] != NULL; i++) {
    if (strncmp(environ[i], envname, strlen(envname)) == 0) {
      if (environ[i][strlen(envname)] == '=')
        break;
    }
  }
  if (environ[i] != NULL) {
    if (!overwrite) {
      free(buf);
      return 0;
    }
    err = __updenv(i, envval);
  } else {
    /* no variable found, check whether we have space for one more */
    if (i == ARG_MAX) {
      free(buf);
      errno = ENOMEM;
      return -1;
    }
    err = __setenv(i, envname, envval);
  }

  if (err != 0) {
    free(buf);
    errno = err;
    return -1;
  }

  if (environ[i] != NULL)
    free(environ[i]);
  environ[i] = buf;

  return 0;
}

int unsetenv(const char *name)
{
  int i, err;

  if ((name == NULL) || (*name == '\0') ||
          (strchr(name, '=') != NULL)) {
    errno = EINVAL;
    return -1;
  }

  for (i = 0; environ[i] != NULL; i++) {
    if (strncmp(environ[i], name, strlen(name)) == 0) {
      if (environ[i][strlen(name)] == '=')
        break;
    }
  }
  if (environ[i] == NULL)
    return 0;

  err = __unsetenv(i);
  if (err != 0) {
    errno = err;
    return -1;
  }

  free(environ[i]);
  for (; environ[i + 1] != NULL; i++)
    environ[i] = environ[i + 1];
  environ[i] = NULL;

  return ns_set_environ(getpid(), environ);
}

int putenv(char *string)
{
  char *p;
  int namelen, i, err;

  if ((string == NULL) || (*string == '\0')) {
    errno = EINVAL;
    return -1;
  }
  p = strchr(string, '=');
  if (p == NULL) {
    errno = EINVAL;
    return -1;
  }
  namelen = p - string + 1;	/* Length of '<name>=' string */

  /* If such var is registered, replace it. */
  for (i = 0; environ[i] != NULL; i++)
    if (strncmp(environ[i], string, namelen) == 0)
      break;
  if (environ[i] != NULL) {
    err = __updenv(i, p + 1);
  } else {
    /* no variable found, check whether we have space for one more */
    if (i == ARG_MAX) {
      errno = ENOMEM;
      return -1;
    }
    *p = '\0';
    err = __setenv(i, string, p + 1);
    *p = '=';
  }

  if (err != 0) {
    errno = err;
    return -1;
  }

  if (environ[i] != NULL)
    free(environ[i]);
  environ[i] = string;

  return 0;
}

static int __setenv(int envid, const char *envname, const char *envval)
{
  uni_rpc_t rpc;
  char *buf, *p;
  iovec_t send[2], recv;
  int err;

  rpc.rpc_call = RPC_SETENV;
  rpc.arg0 = envid;
  rpc.arg1 = strlen(envname) + 1;
  if (envval != NULL)
    rpc.arg1 += strlen(envval);
  rpc.arg1++;

  /* buf = "<envname> \0 <envval> \0" */
  buf = malloc(rpc.arg1);
  if (buf == NULL)
    return errno;
  p = buf;
  strcpy(p, envname);
  p += strlen(envname);
  *p = '\0';
  p++;
  if (envval != NULL) {
    strcpy(p, envval);
    p += strlen(envval);
  }
  *p = '\0';

  send[0].iov_base = &rpc;
  send[0].iov_len = sizeof(uni_rpc_t);
  send[1].iov_base = buf;
  send[1].iov_len = rpc.arg1;
  recv.iov_base = &rpc;
  recv.iov_len = sizeof(uni_rpc_t);

  err = ipc_iov_iov_send(nx_ch, send, 2, &recv, 1);
  free(buf);

  if (err < 0)
    return -err;
  return rpc.rpc_call;
}

static int __updenv(int envid, const char *envval)
{
  const char *nul = "\0";
  uni_rpc_t rpc;
  iovec_t send[2], recv;
  int err;

  rpc.rpc_call = RPC_UPDENV;
  rpc.arg0 = envid;
  if (envval == NULL)
    rpc.arg1 = 1;
  else
    rpc.arg1 = strlen(envval) + 1;

  send[0].iov_base = &rpc;
  send[0].iov_len = sizeof(uni_rpc_t);
  if (envval == NULL)
    send[1].iov_base = (char*)nul;
  else
    send[1].iov_base = (char*)envval;
  send[1].iov_len = rpc.arg1;

  recv.iov_base = &rpc;
  recv.iov_len = sizeof(uni_rpc_t);

  err = ipc_iov_iov_send(nx_ch, send, 2, &recv, 1);

  if (err < 0)
    return -err;
  return rpc.rpc_call;
}

static int __unsetenv(int envid)
{
  /* TODO: add RPC_UNSETENV handling to nameserver */

  return ENOSYS;
}

int __getenviron(void)
{
  char *buf, *p;
  uni_rpc_t rpc;
  iovec_t send, recv[2];
  int err;
  size_t i;
  char *envname, *envval;

  buf = malloc(NX_BUFSIZE);
  if (buf == NULL)
    return errno;
  memset(buf, 0, NX_BUFSIZE);

  rpc.rpc_call = RPC_GETENVIRON;

  send.iov_base = &rpc;
  send.iov_len = sizeof(uni_rpc_t);
  recv[0].iov_base = &rpc;
  recv[0].iov_len = sizeof(uni_rpc_t);
  recv[1].iov_base = buf;
  recv[1].iov_len = NX_BUFSIZE;

  err = ipc_iov_iov_send(nx_ch, &send, 1, recv, 2);
  if (err < 0) {
    err = -err;
    goto done;
  }
  if (rpc.rpc_call != 0) {
    err = rpc.rpc_call;
    goto done;
  }

  /* erase current environment */
  for (i = 0; environ[i] != NULL; i++) {
      free(environ[i]);
      environ[i] = NULL;
  }

  /* fill-in environment with new values */
  p = buf;
  for (i = 0; i < rpc.size/2; i++) {
    if (i == ARG_MAX)
      break;

    envname = p;
    p += strlen(envname) + 1;
    envval = p;
    p += strlen(envval) + 1;

    environ[i] = calloc(1, strlen(envname) + strlen(envval) + 2);
    if (environ[i] == NULL)
      goto done;
    strcpy(environ[i], envname);
    strcat(environ[i], "=");
    strcat(environ[i], envval);
  }

  err = 0;

done:
  free(buf);
  return err;
}

int ns_set_environ(pid_t pid, char *envp[])
{
  size_t l, i;
  char *buf, *p;
  uni_rpc_t rpc;
  iovec_t send[2], recv;
  int err;

  if (envp == NULL)
    l = 0;
  else {
    for (l = 0; envp[l] != NULL; l++)
      ;
    if (l > ARG_MAX) {
      errno = EINVAL;
      return -1;
    }

    buf = malloc(NX_BUFSIZE);
    if (buf == NULL)
      return -1;
    memset(buf, 0, NX_BUFSIZE);

    p = buf;
    for (i = 0; i < l; i++) {
      strcpy(p, envp[i]);
      p = strchr(p, '=');
      *p++ = '\0';
      p += strlen(p) + 1;
    }
  }

  rpc.rpc_call = RPC_SETENVIRON;
  rpc.arg0 = pid;
  rpc.arg6 = l;

  send[0].iov_base = &rpc;
  send[0].iov_len = sizeof(uni_rpc_t);
  recv.iov_base = &rpc;
  recv.iov_len = sizeof(uni_rpc_t);

  if (l == 0)
    err = ipc_iov_iov_send(nx_ch, send, 1, &recv, 1);
  else {
    send[1].iov_base = buf;
    send[1].iov_len = p - buf;
    err = ipc_iov_iov_send(nx_ch, send, 2, &recv, 1);
    free(buf);
  }

  if (err < 0) {
    errno = -err;
    return -1;
  }
  if (rpc.rpc_call != 0) {
    errno = rpc.rpc_call;
    return -1;
  }

  return 0;
}
