/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Michael Tsymbalyuk <sentinel_san@jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 *
 * LibC:
 * lib/stdlib.c: standart C functions
 *
 */

#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

static char *__ntoa(int num,char *buf,size_t buflen,unsigned int radix)
{
  char table10[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' };
  char table16[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
                    'A', 'B', 'C', 'D', 'E', 'F' };
  char *table;
  char *sbuf = buf;

  switch( radix ) {
    case 10:
    case 8:
      table = table10;
      break;
    case 16:
      table = table16;
      break;
    default:
      table = NULL;
      break;
  }

  if( table != NULL && buflen > 0 ) {
    if( num < 0 && radix == 10 ) {
      *buf++ = '-';
      buflen--;
    }
    while( buflen > 0 ) {
      if( num < radix ) {
        *buf++ = table[num];
        break;
      } else {
        *buf++ = table[num % radix];
        num /= radix;
      }
      buflen--;
    }
    if(buflen>0)       *buf = '\0';
    
    /* Now reverse number. */
    buf--;
    while( sbuf < buf ) {
      char c = *sbuf;
      *sbuf = *buf;
      *buf = c;
      sbuf++;
      buf--;
    }
  }

  return buf;
}

char *otoa(int octav,char *buf,size_t buflen)
{
  return __ntoa(octav,buf,buflen,8);
}

char *dtoa(int dec,char *buf,size_t buflen)
{
  return __ntoa(dec,buf,buflen,10);
}

char *htoa(int hex,char *buf,size_t buflen)
{
  return __ntoa(hex,buf,buflen,16);
}

void qsort(void *base,size_t nmemb,size_t size,
	   int(*comp)(const void *,const void *))
{
  size_t wgap,i,j,k;
  char tmp;
  
  if ((nmemb > 1) && (size > 0)) {
    //    assert(nmemb <= ((size_t)(-1)) / size); /* check for overflow */
    wgap=0;
    do {
      wgap = 3 * wgap + 1;
    } while (wgap < (nmemb-1)/3);
    wgap *= size; 
    nmemb *= size;  
    do {
      i = wgap;
      do {
	j = i;
	do {
	  register char *a;
	  register char *b;
	  
	  j -= wgap;
	  a = j + ((char *)base);
	  b = a + wgap;
	  if ((*comp)(a, b) <= 0) {
	    break;
	  }
	  k = size;
	  do {
	    tmp = *a;
	    *a++ = *b;
	    *b++ = tmp;
	  } while (--k);
	} while (j >= wgap);
	i += size;
      } while (i < nmemb);
      wgap = (wgap - size)/3;
    } while (wgap);
  }

  return;
}

unsigned long strtoul(const char *str, char **endptr, int base)
{
  unsigned long rc = 0, v;
  unsigned long nrc;
  const char *start;
  int neg = 0;
  int consumed = 0;

  if ((str == NULL) || (*str == '\0')) {
    errno = EINVAL;
    return 0;
  }

  start = str;

  if((*str == '+')) {
    str++;
  } else if((*str == '-')) {
    neg = 1;
    str++;
  }
  
  if (base == 0) {
    if ((*str >= '1') && (*str <= '9'))
      base = 10;
    else if (*str == '0') {
      base = 8;
      str++;
      if ((*str == 'x') || (*str == 'X')) {
        base = 16;
        str++;
      }
    }
    else {
      errno = EINVAL;
      str = start;
      goto error;
    }
  }

  if ((base < 2) || (base > 36)) {
    errno = EINVAL;
    str = start;
    rc = 0;
    goto error;
  }

  do {
    if((*str >= '0') && (*str <= '9')) {
      v = *str - '0';
    } else if((*str >= 'A') && (*str <= 'Z')) {
      v = *str - 'A' + 10;
    } else if((*str >= 'a') && (*str <= 'z')) {
      v = *str - 'a' + 10;
    } else {
      goto out;
    }

    if(v >= base) {
      goto out;
    }

    consumed++;

    nrc = rc * base + v;
    
    if(nrc < rc) {
      errno = ERANGE;
      rc = ULONG_MAX;
      goto error;
    } else {
      rc = nrc;
    }
    
    str++;
  } while (*str != 0);

  out:
  if(!consumed) {
    errno = EINVAL;
    str = start;
    goto error;
  }
  
  if(neg) {
    rc = ~rc;
  }

  error:
  if (endptr != NULL) {
    *endptr = (char*)str;
  }
  return rc;
}

int atoi(const char *str)
{
  return (int) strtoul(str, NULL, 10);
}

int abs(int j)
{
  return (j >= 0) ? j : -j;
}

long int labs(long int j)
{
  return (j >= 0) ? j : -j;
}

long long int llabs(long long int j)
{
  return (j >= 0) ? j : -j;
}

unsigned long long strtoull(const char *str, char **endptr, int base)
{
  unsigned long long rc = 0, v;
  unsigned long long nrc;
  const char *start;
  int neg = 0;
  int consumed = 0;

  if ((str == NULL) || (*str == '\0')) {
    errno = EINVAL;
    return 0;
  }

  start = str;

  if((*str == '+')) {
    str++;
  } else if((*str == '-')) {
    neg = 1;
    str++;
  }

  if (base == 0) {
    if ((*str >= '1') && (*str <= '9'))
      base = 10;
    else if (*str == '0') {
      base = 8;
      str++;
      if ((*str == 'x') || (*str == 'X')) {
        base = 16;
        str++;
      }
    }
    else {
      errno = EINVAL;
      str = start;
      goto error;
    }
  }

  if ((base < 2) || (base > 36)) {
    errno = EINVAL;
    str = start;
    rc = 0;
    goto error;
  }

  do {
    if((*str >= '0') && (*str <= '9')) {
      v = *str - '0';
    } else if((*str >= 'A') && (*str <= 'Z')) {
      v = *str - 'A' + 10;
    } else if((*str >= 'a') && (*str <= 'z')) {
      v = *str - 'a' + 10;
    } else {
      goto out;
    }

    if(v >= base) {
      goto out;
    }

    consumed++;

    nrc = rc * base + v;

    if(nrc < rc) {
      errno = ERANGE;
      rc = ULONG_MAX;
      goto error;
    } else {
      rc = nrc;
    }

    str++;
  } while (*str != 0);

  out:
  if(!consumed) {
    errno = EINVAL;
    str = start;
    goto error;
  }

  if(neg) {
    rc = ~rc;
  }

  error:
  if (endptr != NULL) {
    *endptr = (char*)str;
  }
  return rc;
}

long atol(const char *str)
{
  return (long) strtoul(str, NULL, 10);
}

long long atoll(const char *str)
{
  return (long long) strtoull(str, NULL, 10);
}
