/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 *
 */

#include <sys/types.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>

#include "__stream.h"

static LIST_DEFINE(streams);

static unsigned char stdin_buf[BUFSIZ];
static struct __stream stream_in = {
  .fd = STDIN_FILENO,
  .eof = 0,
  .error = 0,
  .ofs = 0,
  .in = {
    .buf = stdin_buf,
    .limit = 0,
    .pos = 0,
    .dirty = 0
   },
  .out = {
    .buf = NULL,
    .limit = 0,
    .pos = 0,
    .dirty = 0
  }
};
FILE *stdin = &stream_in;

static unsigned char stdout_buf[BUFSIZ];
static struct __stream stream_out = {
  .fd = STDOUT_FILENO,
  .eof = 0,
  .error = 0,
  .ofs = 0,
  .in = {
    .buf = NULL,
    .limit = 0,
    .pos = 0,
    .dirty = 0
   },
  .out = {
    .buf = stdout_buf,
    .limit = BUFSIZ,
    .pos = 0,
    .dirty = 0
  }
};
FILE *stdout = &stream_out;

static unsigned char stderr_buf[BUFSIZ];
static struct __stream stream_err = {
  .fd = STDERR_FILENO,
  .eof = 0,
  .error = 0,
  .ofs = 0,
  .in = {
    .buf = NULL,
    .limit = 0,
    .pos = 0,
    .dirty = 0
   },
  .out = {
    .buf = stderr_buf,
    .limit = BUFSIZ,
    .pos = 0,
    .dirty = 0
  }
};
FILE *stderr = &stream_err;

struct __stream* create_stream(void)
{
  struct __stream *s;

  s = malloc(sizeof(struct __stream));
  if (s == NULL)
    return NULL;

  s->fd = -1;
  s->eof = s->error = 0;
  s->ofs = 0;

  s->in.buf = NULL;
  s->in.limit = s->in.pos = 0;
  s->in.dirty = 0;

  s->out.buf = NULL;
  s->out.limit = s->out.pos = 0;
  s->out.dirty = 0;

  list_init_node(&(s->node));
  list_add2tail(&streams, &(s->node));

  return s;
}

void destroy_stream(struct __stream *s)
{
  if (s == NULL)
    return;
  if ((s == stdin) || (s == stdout) || (s == stderr)) {
    s->eof = s->error = 0;
    s->ofs = 0;
    if (s->in.buf != NULL)
      memset(s->in.buf, 0, BUFSIZ);
    s->in.limit = s->in.pos = 0;
    s->in.dirty = 0;
    if (s->out.buf != NULL)
      memset(s->out.buf, 0, BUFSIZ);
    s->out.limit = s->out.pos = 0;
    s->out.dirty = 0;
    return;
  }

  if (s->in.buf != NULL)
    free(s->in.buf);
  if (s->out.buf != NULL)
    free(s->out.buf);

  list_del(&(s->node));

  memset(s, 0, sizeof(struct __stream));
  free(s);
}

int fill_stream(struct __stream *s)
{
  int err;

  if (s == NULL)
    return 0;
  if (s->in.buf == NULL) {
    s->out.limit = BUFSIZ;
    return 0;
  }

  memset(s->in.buf, 0, BUFSIZ);
  s->in.pos = 0;
  s->in.dirty = 0;

  err = 0;
  s->error = 0;
  s->eof = 0;

  s->in.limit = pread(s->fd, s->in.buf, BUFSIZ, s->ofs);

  if ((ssize_t)s->in.limit > 0) {
    if (s->out.buf != NULL) {
      memcpy(s->out.buf, s->in.buf, BUFSIZ);
      s->out.limit = s->in.limit;
      s->out.pos = s->in.pos;
      s->out.dirty = 0;
    }
  }
  if (s->in.limit == 0) {
    s->eof = 1;
    s->out.limit = s->in.limit;
    s->out.pos = s->in.pos;
    s->out.dirty = 0;
  }
  if ((ssize_t)s->in.limit < 0) {
    err = errno;
    s->error = 1;
    s->in.limit = s->out.limit = 0;
  }

  errno = err;
  return (err == 0) ? 0 : -1;
}

int flush_all_streams(void)
{
  int err = 0;
  struct __stream *s;

  list_for_each_entry(&streams, s, node) {
    if (fflush(s) == EOF)
      err = errno;
  }
  fflush(stdin);
  fflush(stdout);
  fflush(stderr);

  return (err == 0) ? 0 : EOF;
}
