/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@jarios.org>
 *
 */

#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#include <string.h>

#include "__stream.h"

static int __parse_mode(const char *mode, int *oflags, int *omode)
{
  if ((mode == NULL) || (*mode == '\0'))
    return 0;

  *oflags = 0;
  switch (mode[0]) {
  case 'r':
    *oflags = O_RDONLY;
    break;
  case 'w':
    *oflags = O_WRONLY | O_TRUNC | O_CREAT;
    break;
  case 'a':
    *oflags = O_WRONLY | O_APPEND | O_CREAT;
    break;
  default:
    return 0;
  }
  switch (mode[1]) {
  case '+':
    *oflags |= O_RDWR;
    break;
  case 'b':
  case '\0':
    break;
  default:
    return 0;
  }
  if (mode[1] != '\0') {
    switch (mode[2]) {
    case '+':
      *oflags |= O_RDWR;
      break;
    case 'b':
    case '\0':
      break;
    default:
      return 0;
    }
    if ((mode[2] != '\0') && (mode[3] != '\0')) {
      return 0;
    }
  }

  *omode = 0;
  if ((*oflags & O_CREAT) != 0) {
    /* TODO: use umask */
    *omode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
  }

  return 1;
}

FILE *fopen(const char *filename, const char *mode)
{
  int oflags, omode;
  int err;
  struct __stream *s;

  if ((filename == NULL) || (*filename == '\0')) {
    errno = EINVAL;
    return NULL;
  }

  if (!__parse_mode(mode, &oflags, &omode)) {
    errno = EINVAL;
    return NULL;
  }

  s = create_stream();
  if (s == NULL)
    return NULL;

  if ((oflags & O_RDONLY) != 0) {
    s->in.buf = malloc(BUFSIZ);
    if (s->in.buf == NULL)
      return NULL;
  }
  if ((oflags & O_WRONLY) != 0) {
    s->out.buf = malloc(BUFSIZ);
    if (s->out.buf == NULL)
      return NULL;
  }

  s->fd = open(filename, oflags, omode);
  if (s->fd == -1) {
    err = errno;
    goto error;
  }

  s->ofs = lseek(s->fd, 0, SEEK_CUR);
  if (s->ofs == (off_t)-1) {
    err = errno;
    goto error;
  }

  return s;

error:
  if (s->fd != -1)
    close(s->fd);
  destroy_stream(s);
  errno = err;
  return NULL;
}

FILE *fdopen(int fildes, const char *mode)
{
  int oflags, omode;
  int err;
  struct __stream *s;

  if ((fildes < 0) || (fildes >= OPEN_MAX)) {
    errno = EINVAL;
    return NULL;
  }

  if (!__parse_mode(mode, &oflags, &omode)) {
    errno = EINVAL;
    return NULL;
  }

  s = create_stream();
  if (s == NULL)
    return NULL;

  if ((oflags & O_RDONLY) != 0) {
    s->in.buf = malloc(BUFSIZ);
    if (s->in.buf == NULL)
      return NULL;
  }
  if ((oflags & O_WRONLY) != 0) {
    s->out.buf = malloc(BUFSIZ);
    if (s->out.buf == NULL)
      return NULL;
  }

  s->fd = fildes;

  s->ofs = lseek(s->fd, 0, SEEK_CUR);
  if (s->ofs == (off_t)-1) {
    err = errno;
    goto error;
  }

  return s;

error:
  destroy_stream(s);
  errno = err;
  return NULL;
}

FILE* freopen(const char *filename, const char *mode, FILE *stream)
{
  int oflags, omode;
  int err;
  struct __stream *s;

  /* TODO: support case when filename is NULL */

  if ((filename == NULL) || (*filename == '\0')) {
    errno = EINVAL;
    return NULL;
  }

  if (!__parse_mode(mode, &oflags, &omode)) {
    errno = EINVAL;
    return NULL;
  }

  fflush(stream);
  close(stream->fd);

  /* Clear the stream */
  stream->fd = -1;
  s->eof = s->error = 0;
  s->ofs = 0;
  s->in.limit = s->in.pos = 0;
  s->in.dirty = 0;
  s->out.limit = s->out.pos = 0;
  s->out.dirty = 0;
  if (s->in.buf != NULL)
    free(s->in.buf);
  if (s->out.buf != NULL)
    free(s->out.buf);

  if ((oflags & O_RDONLY) != 0) {
    s->in.buf = malloc(BUFSIZ);
    if (s->in.buf == NULL)
      return NULL;
  }
  if ((oflags & O_WRONLY) != 0) {
    s->out.buf = malloc(BUFSIZ);
    if (s->out.buf == NULL)
      return NULL;
  }

  s->fd = open(filename, oflags, omode);
  if (s->fd == -1) {
    err = errno;
    goto error;
  }

  s->ofs = lseek(s->fd, 0, SEEK_CUR);
  if (s->ofs == (off_t)-1) {
    err = errno;
    goto error;
  }

  return s;

error:
  if (s->fd != -1)
    close(s->fd);
  destroy_stream(s);
  errno = err;
  return NULL;
}
