/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2005,2008 MadTirra <madtirra@jarios.org> 
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 *
 * LibOS/lib/devio.c:
 *      hardware server related functions
 * 
 */

#include <sys/types.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>

#include <ipc/port.h>
#include <ipc/channel.h>
#include <ns/ns.h>
#include <ipc/rpc.h>
#include <devio/devio.h>
#include <ipc/iov.h>
#include <sys/ioctl.h>

#define LNGMSGSIZE  1024

static chid_t hwchan;
static int hw_init=0;

static bool __init_vecs(iovec_t *vec, uni_rpc_t *rpc, const char *str1, const char *str2)
{
  int len1, len2 = 1;

  len1 = strlen(str1);
  if (str2 != NULL) {
    len2 = strlen(str2);
  }
  if (len1 + len2 > LNGMSGSIZE - sizeof(uni_rpc_t) - 2)
    return false;

  vec[0].iov_base = rpc;
  vec[0].iov_len = sizeof(uni_rpc_t);
  vec[1].iov_base = (void*)str1;
  vec[1].iov_len = len1 + 1;
  if (str2 != NULL) {
    vec[2].iov_base = (void*)str2;
    vec[2].iov_len = len2 + 1;
  }

  return true;
}

static int __init_hwops(void)
{
  pid_t hwxpid;

  if(hw_init>0)
    return -1;

  hwxpid=ns_gethwxpid();
  if(!hwxpid)
    return -1;

  hwchan=ipc_channel_open(hwxpid,0,IPC_BLOCKED_ACCESS);
  if(hwchan<0)
    return -1;

  hw_init=32;

  return 0;
}

int hw_register_driver(const char *devn,pid_t pid,ulong_t cmd_id,
           ulong_t input_id,ulong_t output_id,
           int cap,char *driver_name,devid_t id)
{
  uni_rpc_t rpc;
  int r=0;
  iovec_t vec[3];

  /* construct message */
  if (!__init_vecs(vec, &rpc, devn, (const char*)driver_name))
    return -ENAMETOOLONG;

  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  /* initiate rpc */
  rpc.rpc_call=RPC_REGISTER;
  rpc.arg0=pid;
  rpc.arg1=cmd_id;
  rpc.arg2=input_id;
  rpc.arg3=output_id;
  rpc.arg4=cap;
  rpc.arg5=id;

  r=ipc_iov_send(hwchan, vec, 3, &rpc, sizeof(rpc));
  if (r < 0)
    return r;

  return -rpc.rpc_call;
}

int hw_unregister_device(const char *devn)
{
  int r=0;
  uni_rpc_t rpc;
  iovec_t vec[2];

  if (!__init_vecs(vec, &rpc, devn, NULL))
    return -ENAMETOOLONG;

  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  /* initiate rpc */
  rpc.rpc_call=RPC_UNREGISTER;
  rpc.arg0=0;  rpc.arg1=0;
  rpc.arg2=0;  rpc.arg3=0;

  r = ipc_iov_send(hwchan,vec,2,&rpc,sizeof(rpc));
  if(r < 0)
    return r;

  return -rpc.rpc_call; 
}

int hw_create(const char *devn)
{
  int r=0;
  uni_rpc_t rpc;
  iovec_t vec[2];

  if (!__init_vecs(vec, &rpc, devn, NULL))
    return -ENAMETOOLONG;

  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  /* initiate rpc */
  rpc.rpc_call=RPC_CREATE;
  rpc.arg0=0;  rpc.arg1=0;
  rpc.arg2=0;  rpc.arg3=0;

  r = ipc_iov_send(hwchan,vec,2,&rpc,sizeof(rpc));
  if(r < 0)
    return r;

  return -rpc.rpc_call;
}

int hw_unlink(const char *devn)
{
int r=0;
  uni_rpc_t rpc;
  iovec_t vec[2];

  if (!__init_vecs(vec, &rpc, devn, NULL))
    return -ENAMETOOLONG;

  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  /* initiate rpc */
  rpc.rpc_call=RPC_UNLINK;
  rpc.arg0=0;  rpc.arg1=0;
  rpc.arg2=0;  rpc.arg3=0;

  r = ipc_iov_send(hwchan,vec,2,&rpc,sizeof(rpc));
  if(r < 0)
    return r;

  return -rpc.rpc_call;
}

int hw_mkdir(const char *devn)
{
  int r=0;
  uni_rpc_t rpc;
  iovec_t vec[2];

  if (!__init_vecs(vec, &rpc, devn, NULL))
    return -ENAMETOOLONG;
  
  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  /* initiate rpc */
  rpc.rpc_call=RPC_CREATEDIR;
  rpc.arg0=0;  rpc.arg1=0;
  rpc.arg2=0;  rpc.arg3=0;

  r = ipc_iov_send(hwchan,vec,2,&rpc,sizeof(rpc));
  if (r < 0)
    return r;

  return -rpc.rpc_call;
}

int hw_opendir(const char *devn)
{
  int r=0;
  uni_rpc_t rpc;
  iovec_t vec[2];

  if (!__init_vecs(vec, &rpc, devn, NULL))
    return -ENAMETOOLONG;

  if(!hw_init) 
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  /* initiate rpc */
  rpc.rpc_call=RPC_OPENDIR;
  rpc.arg0=0;  rpc.arg1=0;
  rpc.arg2=0;  rpc.arg3=0;

  r = ipc_iov_send(hwchan,vec,2,&rpc,sizeof(rpc));
  if(r < 0)
    return r;

  if(rpc.rpc_call!=0) {
    return -rpc.rpc_call;
  } 

  return rpc.arg0;
}

int hw_dio_open(const char *devn,int flags,hw_direct_io_t *superio)
{
  int r=0;
  uni_rpc_t rpc;
  iovec_t vec[2];
  
  if (!__init_vecs(vec, &rpc, devn, NULL))
    return -ENAMETOOLONG;
  
  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  /* initiate rpc */
  rpc.rpc_call=RPC_OPENDIO;
  rpc.arg1=flags;

  r=ipc_iov_send(hwchan,vec,2,&rpc,sizeof(rpc));
  if(r < 0)
    return r;

  if(rpc.rpc_call!=0) 
    return -rpc.rpc_call;
  
  superio->pid=rpc.arg0;
  superio->cmd=rpc.arg1;
  superio->in=rpc.arg2;
  superio->out=rpc.arg3;
  superio->devno=rpc.arg4;
  /* initialize */
  if(superio->in!=superio->out) {
    superio->chi=ipc_channel_open(superio->pid,superio->in,IPC_BLOCKED_ACCESS);
    superio->cho=ipc_channel_open(superio->pid,superio->out,IPC_BLOCKED_ACCESS);
  } else { 
    superio->chi=superio->cho=
      ipc_channel_open(superio->pid,superio->out,IPC_BLOCKED_ACCESS);
  }
  superio->chc=ipc_channel_open(superio->pid,superio->cmd,IPC_BLOCKED_ACCESS);
  superio->offset=0;
  superio->devn=strdup(devn);
  if(!superio->devn)
    return -ENOMEM;

  return 0;
}


size_t hw_dio_read(hw_direct_io_t *superio,uni_rpc_t *rpc,void *buf,size_t size)
{
  int r=0;
  uni_rpc_t *rpc1;
  iovec_t ivec[2], ovec;
  uni_rpc_t rpc2;

  rpc1 = (rpc == NULL) ? &rpc2 : rpc;
  if(!rpc) memset(rpc1, 0, sizeof(*rpc1));

  rpc1->rpc_call=RPC_READ;
  rpc1->size=size;
  rpc1->offset=superio->offset;
  rpc1->arg2=superio->devno;
  
  ovec.iov_base = rpc1;
  ovec.iov_len = sizeof(uni_rpc_t);
  ivec[0].iov_base = rpc1;
  ivec[0].iov_len = sizeof(uni_rpc_t);
  ivec[1].iov_base = buf;
  ivec[1].iov_len = size;
  r = ipc_iov_iov_send(superio->cho,&ovec,1,ivec,2);
  if(r<0) {
    return r;
  }
  if(rpc1->rpc_call!=0) {
    return -rpc1->rpc_call;
  }

  return rpc1->size;
}

int hw_dio_ioctl(hw_direct_io_t *superio, uni_rpc_t *rpc, void *buf, uint32_t buflen, int dir)
{
  int r=0, n1=1, n2=1;
  iovec_t ivec[2], ovec[2];
  chid_t chan;

  rpc->rpc_call=RPC_IOCTL;
  rpc->arg2=superio->devno;
  rpc->size = buflen;

  if (buf == NULL) {
    r=ipc_channel_msgsnd(superio->chi,sizeof(*rpc),(void *)rpc,
                         sizeof(*rpc),(void *)rpc);
  } else {
    ivec[0].iov_base = ovec[0].iov_base = rpc;
    ivec[0].iov_len = ovec[0].iov_len = sizeof(*rpc);
    if (dir == IOCTL_OUTBUF) {
      ovec[1].iov_base = buf;
      ovec[1].iov_len = buflen;
      chan = superio->chi;
      n1++;
    } else {
      ivec[1].iov_base = buf;
      ivec[1].iov_len = buflen;
      chan = superio->cho;
      n2++;  
    }

    r=ipc_iov_iov_send(chan, ovec, n1, ivec, n2);
  }

  if(r<0) {
    return r;
  }
  
  return -rpc->rpc_call;
}


int hw_dio_close(hw_direct_io_t *superio)
{
#if 0
  int r=0;
  uni_rpc_t rpc;
  iovec_t vec[2];

  if (!__init_vecs(vec, &rpc, superio->devn, NULL))
    return -ENAMETOOLONG;
  
  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  /* initiate rpc */
  rpc.rpc_call = RPC_CLOSEDIO;

  r=ipc_iov_send(hwchan,vec,2,&rpc,sizeof(rpc));
  if(r < 0)
    return r;

  if(rpc.rpc_call!=0) 
    return -rpc.rpc_call;
  
  ipc_channel_close(superio->chc);
  ipc_channel_close(superio->chi);
  ipc_channel_close(superio->cho);
  free(superio->devn);
#endif

  return 0;
}

size_t hw_dio_write(hw_direct_io_t *superio,uni_rpc_t *rpc,const void *buf,size_t size)
{
  int r=0;
  uni_rpc_t *rpc1;
  iovec_t ovec[2];
  uni_rpc_t rpc2;

  rpc1 = ((rpc == NULL) ? &rpc2 : rpc);

  if(!rpc) memset(rpc1, 0, sizeof(*rpc));

  rpc1->rpc_call=RPC_WRITEDIO;
  rpc1->size=size;
  rpc1->offset=superio->offset;
  rpc1->arg2=superio->devno;

  ovec[0].iov_base = rpc1;
  ovec[0].iov_len = sizeof(uni_rpc_t);
  ovec[1].iov_base = (void *)buf;
  ovec[1].iov_len = size;
  r = ipc_iov_send(superio->chi,ovec,2,rpc1,sizeof(*rpc1));

  if(r<0) {
    return r;
  }
  if(rpc1->rpc_call!=0) {
    return -rpc1->rpc_call;
  }

  return rpc1->size;
}

/* events */
int hw_register_event(const char *devn,ulong_t listen_port)
{
  int r=0;
  uni_rpc_t rpc;
  iovec_t vec[2];

  if (!__init_vecs(vec, &rpc, devn, NULL))
    return -ENAMETOOLONG;

  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;
 
  rpc.rpc_call=RPC_REGEVENTS;
  rpc.arg1=listen_port;

  r=ipc_iov_send(hwchan,vec,2,&rpc,sizeof(rpc));
  if(r < 0)
    return r;

  return -rpc.rpc_call;
}

int hw_unregister_event(const char *devn)
{
  int r=0;
  uni_rpc_t rpc;
  iovec_t vec[2];

  if (!__init_vecs(vec, &rpc, devn, NULL))
    return -ENAMETOOLONG;

  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  rpc.rpc_call=RPC_UNREGEVENTS;

  r=ipc_iov_send(hwchan,vec,2,&rpc,sizeof(rpc));
  if(r < 0)
    return r;

  return -rpc.rpc_call;
}

int hw_readdir(int fd, hw_dirent_t *dentry)
{
  int r=0;
  uni_rpc_t rpc;
  iovec_t ovec, ivec[2];

  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  /* initiate rpc */
  rpc.rpc_call=RPC_READDIR;
  rpc.arg0=fd;  rpc.arg1=0;
  rpc.arg2=0;  rpc.arg3=0;
  
  /* construct message */
  ovec.iov_base = &rpc;
  ovec.iov_len = sizeof(rpc);
  ivec[0].iov_base = &rpc;
  ivec[0].iov_len = sizeof(rpc);
  ivec[1].iov_base = dentry;
  ivec[1].iov_len = sizeof(hw_dirent_t);

  r=ipc_iov_iov_send(hwchan, &ovec, 1, ivec, 2);
  if(r<0)
    return r;

  if (!rpc.rpc_call || rpc.rpc_call == HWEOD)
    return rpc.rpc_call;

  return -rpc.rpc_call;
}

int hw_closedir(int fd) 
{
  /* TO DO [dg]: implement later */
  return 0;
}

int hw_open(const char *devn)
{
#if 0
  uni_rpc_t rpc;
  int r=0;
  char *msg;

  if(!hw_init)
    r=__init_hwops();
  if(r!=0)
    return -ESRCH;

  /* construct message */
  rpc.rpc_call=RPC_OPEN;
  rpc.arg0=*pid;
  rpc.arg1=ioids[0];
  rpc.arg2=ioids[1];
  rpc.arg3=ioids[2];
  memset(hwport->repl_msg,'\0',LNGMSGSIZE);
  msg=hwport->repl_msg+sizeof(rpc);
  memcpy(hwport->repl_msg,&rpc,sizeof(rpc));
  memcpy(msg,devn,strlen(devn));

  ipc_port_send(hwport,sizeof(rpc)+strlen(devn)+1);

  memcpy(&rpc,hwport->repl_msg,sizeof(rpc));
  if(rpc.rpc_call!=0)
    return -rpc.rpc_call;
  *pid=rpc.arg0;
  ioids[0]=rpc.arg1;
  ioids[1]=rpc.arg2;
  ioids[2]=rpc.arg3;
#endif
  return 0;
}

