/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * (c) Copyright 2006-2009 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2009 Stanislav M. Ivankin <grepz@jarios.org>
 *
 * Fslinker userspace library functions.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <errno.h>
#include <limits.h>

#include <ns/ns.h>
#include <ipc/rpc.h>

#include <sys/fslinker.h>

#define FILL_FAIL(ret)					\
  do {  errno = (ret); goto __fill_fail; } while (0)

#define TRY_FREE(ptr)				\
  if ((ptr) != NULL) {				\
    free(ptr);					\
    (ptr) = NULL;				\
  }

static inline void __fill_req_msg(void **msg, const void *data, size_t sz);

int fslink_send_request(const fs_link_entry_t *entry, fs_link_req_t req)
{
  int ret;
  size_t rpcmsg_len, msg_len;
  size_t opts_len, link_s_len, link_p_len;
  pid_t fsl_pid;
  chid_t fsl_chan;
  void *msg, *mptr;
  uni_rpc_t rpc;

  ret = 0;

  /* Both pathes and type must be set */
  if ((!entry->mountp.plen || entry->mountp.plen > PATH_MAX) ||
      (!entry->mounts.plen || entry->mounts.plen > PATH_MAX) ||
      (entry->opts_n > FS_OPTS_MAX))
    return EINVAL;

  fsl_pid = ns_getfslpid();
  if (fsl_pid == -1)
    return ESRCH;

  fsl_chan = ipc_channel_open(fsl_pid, 0, IPC_BLOCKED_ACCESS);
  if (fsl_chan == -1)
    return EINVAL;

  switch (req) {
  case FSLINK_REQ_MOUNT:
    rpc.rpc_call = RPC_MOUNT;
    break;
  case FSLINK_REQ_UMOUNT:
    rpc.rpc_call = RPC_UMOUNT;
    break;
  default:
    return EINVAL;
  }

  rpc.arg0 = entry->fstype_len;

  /* Link point */
  rpc.arg1 = entry->mountp.plen;
  rpc.arg2 = entry->mountp.p_num;
  link_p_len = entry->mountp.p_num*sizeof(int);
  /* Link subject */
  rpc.arg3 = entry->mounts.plen;
  rpc.arg4 = entry->mounts.p_num;
  link_s_len = entry->mounts.p_num*sizeof(int);
  /* Options */
  rpc.arg5 = entry->optstr_len;
  rpc.arg6 = entry->opts_n;
  opts_len = entry->opts_n*sizeof(int);

  msg_len = rpc.arg0 + rpc.arg1 + link_p_len +
    rpc.arg3 + link_s_len + rpc.arg5 + opts_len;
  rpcmsg_len = sizeof(uni_rpc_t) + msg_len;

  msg = malloc(rpcmsg_len);
  if (!msg)
    return ENOMEM;

  mptr = msg;
  __fill_req_msg(&mptr, &rpc, sizeof(uni_rpc_t));
  __fill_req_msg(&mptr, entry->fstype, rpc.arg0);
  __fill_req_msg(&mptr, entry->mountp.path, rpc.arg1);
  __fill_req_msg(&mptr, entry->mountp.pp, link_p_len);
  __fill_req_msg(&mptr, entry->mounts.path, rpc.arg3);
  __fill_req_msg(&mptr, entry->mounts.pp, link_s_len);
  __fill_req_msg(&mptr, entry->options, rpc.arg5);
  __fill_req_msg(&mptr, entry->opt_len, opts_len);

  ret = ipc_channel_msgsnd(fsl_chan, rpcmsg_len, msg, sizeof(uni_rpc_t), &rpc);
  if (ret < 0) {
    free(msg);
    return ret;
  }

  ret = rpc.rpc_call;

  free(msg);

  return ret;
}

fs_link_entry_t *fslink_fill_struct(const char *fstype,
				    const char *lsubj, const char *lpoint,
				    const char *optstr)
{
  int ret;
  fs_link_entry_t *entry;

  /* Check if someone is trying something stupid */
  if (!lpoint || !lsubj || !fstype
      || (optstr && (strlen(optstr) > LINE_MAX))) {
    errno = EINVAL;
    return NULL;
  }

  entry = malloc(sizeof(*entry));
  if (!entry) FILL_FAIL(ENOMEM);

  memset(entry, 0, sizeof(*entry));

  if (fstype) {
    entry->fstype_len = strlen(fstype);
    entry->fstype = malloc(entry->fstype_len + 1);
    strcpy(entry->fstype, fstype);
    if (!entry->fstype) FILL_FAIL(ENOMEM);
  }

  if (optstr) {
    ret = fslink_alloc_opts(optstr, entry);
    if (ret)
      FILL_FAIL(ret);
  }

  ret = fslink_pp_fill(lpoint, &entry->mountp);
  if (ret)
    FILL_FAIL(ret);

  ret = fslink_pp_fill(lsubj, &entry->mounts);
  if (ret)
    FILL_FAIL(ret);

  return entry;

 __fill_fail:
  fslink_free_entry(entry);
  return NULL;
}

int fslink_pp_fill(const char *path, fs_link_pp_t *link_pp)
{
  int ret;

  ret = fslink_pp_alloc(path, link_pp);
  if (ret)
    return ret;

  link_pp->plen = strlen(path);
  link_pp->path = malloc(link_pp->plen + 1);
  if (!link_pp->path) {
    free(link_pp->pp);
    return ENOMEM;
  }

  strncpy(link_pp->path, path, link_pp->plen);

  return 0;
}

int fslink_alloc_opts(const char *optstr, fs_link_entry_t *entry)
{
  int optnum, optlen, i;
  const char *ptr;
  char *opt_ptr;

  fslink_count_opts(optstr, &optnum, &optlen);
  if (optnum > FS_OPTS_MAX)
    return E2BIG;
  /* opts_n is already zero */
  else if (optnum == 0)
    return 0;

  entry->options    = malloc(optlen);
  if (!entry->options)
    return ENOMEM;

  entry->opt_len = malloc(sizeof(int)*optnum);
  if (!entry->opt_len) {
    free(entry->options);
    return ENOMEM;
  }

  entry->optstr_len = optlen;
  entry->opts_n     = optnum;

  ptr = optstr;
  opt_ptr = entry->options;
  for (i=0;;) {
    const char *ptr_buf = ptr;
    optlen = 0;
    while(*ptr && *ptr != ',') {
      ptr ++; optlen ++;
    }
    if (optlen) {
      entry->opt_len[i] = optlen;
      strncpy(opt_ptr, ptr_buf, optlen);
      opt_ptr += optlen;
      i ++;
    }
    while (*ptr == ',') ptr ++;
    if (*ptr == '\0') break;
  }

  return 0;
}

void fslink_count_opts(const char *optstr, int *optnum, int *optlen)
{
  const char *ptr;

  *optnum = *optlen = 0;
  ptr = optstr;

  for (;;) {
    while(*ptr && *ptr != ',') {
      ptr ++; (*optlen) ++;
    }
    (*optnum) ++;
    while (*ptr == ',') ptr ++;
    if (*ptr == '\0') break;
  }

  return;
}

int fslink_pp_alloc(const char *path, fs_link_pp_t *entry_pp)
{
  /* TODO: Make sure max file len and max depth are respected */

  int n, i;
  int *pp, *pp_ptr;
  const char *path_ptr;

  n = 0;
  for (path_ptr = path ; *path_ptr; path_ptr ++)
    if (*path_ptr == '/')
      ++ n;

  if (n == 0)
    return EINVAL;

  pp = malloc(n * sizeof(int));
  if (!pp)
    return ENOMEM;

  for (path_ptr = path, pp_ptr = pp, i = 0;
       *path_ptr; i++, path_ptr ++)
    if (*path_ptr == '/') {
      *pp_ptr = i;
      ++ pp_ptr;
    }

  entry_pp->p_num = n;
  entry_pp->pp    = pp;

  return 0;
}

void fslink_free_entry(fs_link_entry_t *entry)
{
  entry->fstype_len = 0;
  TRY_FREE(entry->fstype);

  fslink_free_pp(&entry->mounts);
  fslink_free_pp(&entry->mountp);

  if (entry->opts_n) {
    TRY_FREE(entry->options);
    TRY_FREE(entry->opt_len);
  }

  TRY_FREE(entry);
}

void fslink_free_pp(fs_link_pp_t *link_pp)
{
  TRY_FREE(link_pp->pp);
  TRY_FREE(link_pp->path);
  link_pp->p_num = 0;
}


static inline void __fill_req_msg(void **msg, const void *data, size_t sz)
{
  *msg = (void *)((char *)memcpy(*msg, data, sz) + sz);
};
