/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008,2009 MadTirra <madtirra@jarios.org>
 * (c) Copyright 2009 Dan Kruchinin <dk@jarios.org>
 *
 * LibOS:
 * lib/mmap.c: mmap related functions
 *
 * AK: 8-apr-2009 added files mapping protocol <madtirra@jarios.org>
 *                added memory object related funcs
 *
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/syscall.h>
#include <ipc/channel.h>
#include <ipc/iov.h>
#include <ipc/rpc.h>
#include <errno.h>
#include <alloc/imcache.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>

/* local include for fds */
#include "../jarifs/__fildesc.h"

void *mmap(void *addr, size_t length, int prot, int flags, int fd, off_t offset)
{
  uni_rpc_t rpc;
  int ufsid,r;
  chid_t ep_ch;
  int prot_flags[2];
  iovec_t sndvec[2];
  
  if(fd==NOFD) {
    return mk_mmap(addr,length,prot,flags,NOFD,offset);
  } else { /* we're trying to mmap file, shared memory ... */
    /* check file desc for validity */
    if(valid_fd(fd)) {
      ep_ch=tfds[fd].chid;
      ufsid=tfds[fd].ufsid;
      /* ok prepare to send */
      rpc.rpc_call=RPC_MMAP;
      rpc.size=length;
      rpc.offset=offset;
      rpc.arg0=ufsid;
      /* prot and flags */
      prot_flags[0]=prot;
      prot_flags[1]=flags;
      /* setup vecs */
      sndvec[0].iov_base=&rpc;
      sndvec[0].iov_len=sizeof(rpc);
      sndvec[1].iov_base=prot_flags;
      sndvec[1].iov_len=sizeof(int)*2;
      r=ipc_iov_send(ep_ch,sndvec,2,&rpc,sizeof(rpc));
      if(r<0) { /* check IPC */
        errno=-r;
        return MAP_FAILED;
      }
      if(rpc.rpc_call!=0) { /* error handling */
        errno=rpc.rpc_call;
        return MAP_FAILED;
      }
      /* ok, let's try to mmap */
      return mk_mmap(addr,length,prot,flags,(int)rpc.size,offset);
    } else {
      errno = EBADF;
      return MAP_FAILED;
    }
  }

  return MAP_FAILED;
}

int munmap(void *addr, size_t length)
{
  int ret = __syscall3(SC_MUNMAP, (syscall_arg_t)0, (syscall_arg_t)addr, (syscall_arg_t)length);
  if (ret)
    errno = -ret;

  return ret;
}

void *mmap_pid(pid_t victim,void *addr,size_t length,int prot,int flags,int fd,off_t offset)
{
  unsigned long ret;
  struct mmap_args margs;

  margs.addr=(uintptr_t)addr;
  margs.size=length;
  margs.prot=prot;
  margs.flags=flags;
  margs.offset=offset;

  ret=__syscall3(SC_MMAP,victim,fd,(syscall_arg_t)&margs);
  if(!PAGE_ALIGN(ret)) {
    errno=-((int)ret);
    return MAP_FAILED;
  }

  return (void *)ret; 
}

int grant_pages(void *vaddr_from, size_t length, pid_t victim_pid, uintptr_t victim_addr)
{
  int ret;

  ret = __syscall4(SC_GRANT_PAGES, (uintptr_t)vaddr_from, length, victim_pid, victim_addr);
  if (ret) {
    errno = -ret;
  }

  return ret;
}

/* low level mmap - to microkernel , fd is a memory object id */
void *mk_mmap(void *addr, size_t length, int prot, int flags, int memobj_id, off_t offset)
{
  unsigned long ret;
  struct mmap_args margs;  

  margs.addr = (uintptr_t)addr;
  margs.size = length;
  margs.prot = prot;
  margs.flags = flags;
  margs.offset = offset;
  ret = __syscall3(SC_MMAP, 0, memobj_id, (syscall_arg_t)&margs);

  if(!PAGE_ALIGN(ret)) {
    errno=-((int)ret);
    return MAP_FAILED;
  }

  return (void *)ret;
}

#define ptr_align(s,a)    ((((uintptr_t)s)+((a)-1)) & ~((a)-1))

/* memory object ops */
int memobj_create_wbackend(prtid_t serve,size_t size,int lifetype,
                           int flags,long private)
{
  struct memobj_info mo_info;
  size_t nsize;
  int r;

  /* check */
  if(!size) {
    errno=EINVAL;
    return -1;
  }

  /* init */
  nsize=ptr_align(size,PAGE_SIZE);
  memset(&mo_info,0,sizeof(mo_info));
  mo_info.size=nsize;
  mo_info.nature=MMO_NTR_PROXY;
  mo_info.backend_port=serve;
  mo_info.lifetype=lifetype;
  mo_info.flags=flags;
  mo_info.private=private;
  /* call microkernel */
  r=__syscall1(SC_MEMOBJ_CREATE,(uintptr_t)&mo_info);
  if(r<0) {
    errno=-r;
    return -1;
  }

  return mo_info.id;
}

