/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 * 
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 *
 * event locked queue implementation
 */

#include <queue.h>
#include <errno.h>
#include <stdlib.h>

void evl_queue_add_range(evl_queue_t *q, list_node_t *first, list_node_t *last)
{
  bool empty;

  pthread_mutex_lock(&q->mutex);
  empty = list_is_empty(&q->head);
  list_add_range(first, last, list_node_last(&q->head), list_head(&q->head));
  pthread_mutex_unlock(&q->mutex);
  if (empty)
    event_signal(&q->event); 
}

void evl_queue_add_cond(evl_queue_t *q, list_node_t *node, bool (*fn) (list_node_t*))
{
  list_node_t *n;
  bool empty;

  pthread_mutex_lock(&q->mutex);
  empty = list_is_empty(&q->head);
  if (fn != NULL) {
    list_for_each(&q->head, n) {
      if (fn(n))
        break;
    }
  } else {
    node = list_node_first(&q->head);
  }
  list_add(n, node);
  pthread_mutex_unlock(&q->mutex);
  if (empty)
    event_signal(&q->event);
}

list_node_t *evl_queue_get_next_prep(evl_queue_t *q, void (*func)(list_node_t *node))
{
  list_node_t *n = NULL;

  while (1) {
    pthread_mutex_lock(&q->mutex);
    if (list_is_empty(&q->head)) {
      pthread_mutex_unlock(&q->mutex);
      if (event_wait(&q->event) == -EINTR)
        break;
    } else {
      n = list_node_first(&q->head);
      list_delfromhead(&q->head);
      if (func != NULL) {
        func(n);
      }
      pthread_mutex_unlock(&q->mutex);
      break;
    }
  }

  return n; 
}

int init_evl_queue(evl_queue_t *q)
{
  pthread_mutexattr_t attr = { PTHREAD_PROCESS_PRIVATE };
  
  list_init_head(&q->head);
  event_init(&q->event);

  return -pthread_mutex_init(&q->mutex, &attr);
}

void destroy_evl_queue(evl_queue_t *q)
{
  pthread_mutex_destroy(&q->mutex);
  event_destroy(&q->event);
}

int init_prio_queue(prio_queue_t *q, int nr_prios)
{
  int i;

  if (nr_prios < 1)
    return -EINVAL;

  if (init_bitmap(&q->bm, nr_prios))
    return -ENOMEM;

  q->heads = malloc(nr_prios * sizeof(list_head_t));
  if (q->heads == NULL)
    return -ENOMEM;

  q->nr_prios = nr_prios;
  for (i = 0; i < nr_prios; i++)
    list_init_head(q->heads + i);

  return 0;
}

void prio_queue_add(prio_queue_t *q, list_node_t *node, int prio)
{
  list_add2tail(q->heads + prio, node);
  bit_set_multi(&q->bm, prio);
}

bool prio_queue_is_empty(prio_queue_t *q)
{
  return (bit_find_lsf_multi(&q->bm, 0) == -1);
}

list_node_t *prio_queue_get_next_match(prio_queue_t *q, bool (*match)(list_node_t *node))
{
  int prio;
  list_node_t *node = NULL;
  list_head_t *head;
  
  prio = bit_find_msf_multi(&q->bm, q->nr_prios - 1);

  while ((prio >= 0) && (node == NULL)) {
    head = q->heads + prio;
    list_for_each(head, node) {
      if (!match || match(node))
        break;
    }
    if (node != list_head(head)) {
      list_del(node);
      if (list_is_empty(q->heads + prio)) {
        bit_clear_multi(&q->bm, prio);
      }
    } else {
      prio = bit_find_msf_multi(&q->bm, prio - 1);
      node = 0;
    }
  }

  return node;
}

void destroy_prio_queue(prio_queue_t *q)
{
  free_bitmap(&q->bm);
  free(q->heads);
}
