/*
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Copyright (C) 2006,2007,2008,2009 Jari OS Core Team <http://jarios.org>
 * Copyright (C) 2009 Dmitry Balakshin <balakshin@gmail.com>
 *
 * Based on compute_yday() and compute_wday() from linux kernel:
 *     drivers/char/efirtc.c
 * Copyright (C) 1999 Hewlett-Packard Co
 * Copyright (C) 1999 Stephane Eranian <eranian@hpl.hp.com>
 *
 * LibOS:
 * lib/conv.c: Conversions between time_t, struct tm and char*
 *
 */

#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <time.h>

#define is_leap(year) \
  ((year) % 4 == 0 && ((year) % 100 != 0 || (year) % 400 == 0))

static const int mod_yday[2][13] = {
  /* Normal years */
  { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 },
  { 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 }
};

static int compute_yday(struct tm *timeptr)
{
  return mod_yday[is_leap(timeptr->tm_year + 1900)][timeptr->tm_mon] +
      timeptr->tm_mday - 1;
}

static int compute_wday(struct tm *timeptr)
{
  int y;
  int ndays = 0;

  for (y = 0; y < timeptr->tm_year; y++)
    ndays += 365 + (is_leap(y + 1900) ? 1 : 0);
  ndays += compute_yday(timeptr);

  /* 1 = 1900-1-1 was a Monday */
  return (ndays + 1) % 7;
}

/* TODO: %.Ns format in snprintf is broken, thus '\0' endings here */

static const char wday_name[7][4] = {
  "Sun\0", "Mon\0", "Tue\0", "Wed\0", "Thu\0", "Fri\0", "Sat\0"
};

static const char mon_name[12][4] = {
  "Jan\0", "Feb\0", "Mar\0", "Apr\0", "May\0", "Jun\0",
  "Jul\0", "Aug\0", "Sep\0", "Oct\0", "Nov\0", "Dec\0"
};

static const int mon_days[12] = {
  31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

char *asctime(const struct tm *timeptr)
{
  static char str[26];

  sprintf(str, "%.3s %.3s%3d %.2d:%.2d:%.2d %d\n",
    wday_name[timeptr->tm_wday],
    mon_name[timeptr->tm_mon],
    timeptr->tm_mday, timeptr->tm_hour,
    timeptr->tm_min, timeptr->tm_sec,
    1900 + timeptr->tm_year);
  return str;
}

char *ctime(const time_t *clock)
{
  return asctime(localtime(clock));
}

struct tm *localtime(const time_t *timer)
{
  static struct tm time;
  int i;
  time_t sec;

  if (timer == NULL) {
    errno = EINVAL;
    return NULL;
  }
  sec = *timer;

  /* year */
  for (i = 0; ; i++) {
    if (sec < 365 * 86400)
      break;
    sec -= 365 * 86400;
    if (i % 4 == 2) {
      if (sec < 86400)
        break;
      sec -= 86400;
    }
  }
  time.tm_year = i;

  /* month */
  for (i = 0; i < 12; i++) {
    if (sec < mon_days[i] * 86400)
      break;
    sec -= mon_days[i] * 86400;
    if ((i == 1) && (time.tm_year % 4 == 2)) {
      if (sec < 86400)
        break;
      sec -= 86400;
    }
  }
  time.tm_mon = i;

  time.tm_mday = sec / 86400;
  sec -= time.tm_mday * 86400;
  time.tm_mday++;

  time.tm_hour = sec / 3600;
  sec -= time.tm_hour * 3600;

  time.tm_min = sec / 60;
  sec -= time.tm_min * 60;

  time.tm_sec = sec;

  time.tm_wday = compute_wday(&time);
  time.tm_yday = compute_yday(&time);
  time.tm_isdst = 0;

  if (sec > 60) {
    errno = EOVERFLOW;
    return NULL;
  }

  return &time;
}

time_t mktime(struct tm *timeptr)
{
  time_t ret;
  int i;

  if (timeptr == NULL)
    return (time_t)-1;

  ret = timeptr->tm_sec + timeptr->tm_min * 60 +
    timeptr->tm_hour * 3600 + (timeptr->tm_mday - 1) * 86400;

  /* month */
  for (i = 0; i < timeptr->tm_mon; i++) {
    ret += mon_days[i] * 86400;
    if ((i == 1) && (timeptr->tm_year % 4 == 2))
      ret += 86400;
  }

  /* year */
  for (i = 0; i < timeptr->tm_year; i++) {
    ret += 365 * 86400;
    if (i % 4 == 2)
      ret += 86400;
  }

  return ret;
}

double difftime(time_t time1, time_t time0)
{
  return (double)(time1 - time0);
}
