/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2008 Jari OS core team <http://jarios.org>
 * (c) Copyright 2008 Paul Graphov <graphov@jarios.org>
 *
 */

#include <sys/types.h>
#include <wchar.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>

/*
  Gentle introduction to unicode is available at
  http://www.joelonsoftware.com/articles/Unicode.html
 */


/* Used if state wasn't supplied */
static mbstate_t int_state;

/* Initialize mbstate */
int mbsinit(mbstate_t *ps)
{
  if(!ps) return -1;
  if(ps->count == 0) return 1;
  memset(ps, 0, sizeof(*ps));
  return 0;
}

size_t mbrlen(const char *s, size_t n, mbstate_t *ps)
{
  return mbrtowc (NULL, s, n, NULL);
}

size_t mbrtowc(wchar_t *pwc, const char *s, size_t n, mbstate_t *ps)
{
  uint32_t in, cnt = 0;
  
  mbstate_t *state = ps?:&int_state;
  
  // nothing to read
  if(n <= 0) return (size_t)-2;

  // null source
  if(!s) {
    if(state->count == 0) {
      state->value = 0;
      return 0;
    } else {
      errno = EILSEQ;
      return (size_t)-1;
    }
  }

  // read first byte
  if(state->count == 0) {
    in = ((unsigned char*)s)[cnt++];
    if(in < 0x80) {
      // One-byte character
      state->value = in;
      state->count = 0;
    } else if(in >= 0xC2 && in <= 0xE0) {
      //Two
      state->value = in & 0x1F;
      state->count = 1;
    } else if((in & 0xF0) == 0xE0) {
      //Three
      state->value = in & 0xF;
      state->count = 2;
    } else if((in & 0xF8) == 0xF0) {
      //Four
      state->value = in & 0x7;
      state->count = 3;
    } else if((in & 0xFC) == 0xF8) {
      //Five
      state->value = in & 0x3;
      state->count = 4;
    } else if((in & 0xFE) == 0xFC) {
      //Six
      state->value = in & 1;
      state->count = 5;
    } else {
      //Malformed character
      //TODO: maybe something special is needed
      errno = EILSEQ;
      return (size_t)-1;
    }
  }
  
  while(state->count>0 && cnt<n) {
    //read non-head parts of utf8 character
    in = ((unsigned char*)s)[cnt++];
    if((in & 0xC0) != 0x80) { // Invalid non-head code.
      errno = EILSEQ;
      state->count = 0;
      state->value = 0;
      return (size_t)-1;
    }
    state->count--;
    state->value <<= 6;
    state->value |= in & 0x3f;
  }
  if(state->count>0) {
    return (size_t)-2;
  }
  if(pwc) *pwc = state->value;
  state->count = 0;
  if(state->value)return cnt;
  return 0;
}

size_t mbsrtowcs(wchar_t *dest, const char **src, size_t len, mbstate_t *ps)
{
  int res,cnt = 0;
  while(cnt < len) {
    res = mbrtowc(&dest[cnt], *src, 6, ps);
    if(res == 0) { // end of line
      *src = NULL;
      return cnt;
    } else if(res > 0) {
      *src += res;
      cnt++;
    } else {
      // only invalid sequence can be reason
      // This is utf8-only way as we have no shift sequences
      errno = EILSEQ;
      return (size_t)-1;
    }
  }
  return ++cnt;
}

size_t mbstowcs(wchar_t *dest, const char *src, size_t n)
{
  return mbsrtowcs(dest, &src, n, NULL);
}
