/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 MadTirra <madtirra@jarios.org>
 * (c) Copyright 2008 Dmitry Balakshin <balakshin@jarios.org>
 *
 * LibC:
 * lib/string.c: C strings extra ops
 *
 */

#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>

char *strdup(const char *s)
{
  int len=0;
  char *dest;

  if(!s)
    return NULL;

  len=strlen(s)+1;
  dest=malloc(strlen(s)+1);
  if(!dest)
    return NULL;
 
  dest=strcpy(dest,s);

  return dest;
}

int strcasecmp(const char *s1, const char *s2)
{
  for (; (*s1 != '\0') && (*s2 != '\0'); s1++, s2++) {
    if (toupper(*s1) != toupper(*s2))
      return (*s1 - *s2);
  }

  return (int)(*s1) - (int)(*s2);
}

int strncasecmp(const char *s1, const char *s2, size_t n)
{
  for (; n > 0; n--, s1++, s2++) {
    if (toupper(*s1) != toupper(*s2))
      return (*s1 - *s2);
  }

  return 0;
}

/* this is a GNU extension function */
char *stpcpy(char *dest,const char *src)
{
  strcpy(dest,src);

  while ((*dest++ = *src++))
    ;

  return --dest;
}

char *strsep (char **stringp, const char *delim)
{
  char *begin, *end;

  begin = *stringp;
  if (begin == NULL)
    return NULL;

  /* A frequent case is when the delimiter string contains only one
     character.  Here we don't need to call the expensive `strpbrk'
     function and instead work using `strchr'.  */
  if (delim[0] == '\0' || delim[1] == '\0') {
    char ch = delim[0];

    if (ch == '\0')
      end = NULL;
    else {
      if (*begin == ch)
	end = begin;
      else if (*begin == '\0')
	end = NULL;
      else
	end = strchr (begin + 1, ch);
    }
  } else
    /* Find the end of the token.  */
    end = strpbrk (begin, delim);

  if (end) {
    /* Terminate the token and set *STRINGP past NUL character.  */
    *end++ = '\0';
    *stringp = end;
  } else
    /* No more delimiters; this is the last token.  */
    *stringp = NULL;

  return begin;
}

void *memchr(const void *s,int c,size_t n)
{
  register char *r=(char *)s;

  while(n) {
    if(*r==(char)c) return (void *)r;
    ++r;
    --n;
  }

  return NULL;
}

void *memrchr(const void *s,int c,size_t n)
{
  register unsigned char *r=((unsigned char *)s + (size_t)n);

  while(n) {
    if(*--r==(char)c) return (void *)r;
    --n;
  }

  return NULL;
}

/* this is a GNU extension function */
void* mempcpy(void *dst, const void *src, size_t n)
{
  void *p;

  p = memcpy(dst, src, n);
  if (p == NULL)
    return NULL;
  return ((unsigned char*)dst) + n;
}

size_t strcspn(const char *s1, const char *s2)
{
  size_t n;

  if (s1 == NULL)
    return 0;
  if (s2 == NULL)
    return strlen(s1);

  for (n = 0; s1[n] != '\0'; n++) {
    if (strchr(s2, s1[n]) != NULL)
      break;
  }

  return n;
}

int strcoll(const char *s1, const char *s2)
{
  /* TODO: make it locale sensitive */
  return strcmp(s1, s2);
}

size_t strspn(const char *s1, const char *s2)
{
  size_t n;

  if (s1 == NULL)
    return 0;
  if (s2 == NULL)
    return strlen(s1);

  for (n = 0; s1[n] != '\0'; n++) {
    if (strchr(s2, s1[n]) == NULL)
      break;
  }

  return n;
}

/* this is a GNU extension function */
char* strndup(const char *s, size_t n)
{
  size_t l;
  char *dst;

  if (s == NULL)
    return NULL;

  l = strlen(s);
  if (l > n)
    l = n;
  l++;

  dst = calloc(1, l);
  if (dst == NULL)
    return NULL;

  if (strncpy(dst, s, n) == NULL) {
    free(dst);
    return NULL;
  }

  return dst;
}
