/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Michael Tsymbalyuk <mtzaurus@gmail.com>
 *
 * libpthread/mutex.c: functions related to POSIX mutexes.
 *
 */

#include <sys/types.h>
#include <sys/syscall.h>
#include <pthread.h>
#include <errno.h>
#include <bitwise.h>

int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
  if( !mutex || mutex->__id == __SO_UNINITIALIZED_ID ) {
    return EINVAL;
  }
  return POSIX_ERR(SYNC_DESTROY(mutex));
}

int pthread_mutexattr_getpshared(const pthread_mutexattr_t *attr,int *pshared)
{
  if( !attr || !pshared ) {
    return EINVAL;
  }
  *pshared=attr->__type;
  return 0;
}

int pthread_mutexattr_setpshared(pthread_mutexattr_t *attr,int pshared)
{
  if( !attr ||
      (pshared != PTHREAD_PROCESS_SHARED && pshared != PTHREAD_PROCESS_PRIVATE) ) {
    return EINVAL;
  }

  attr->__type=pshared;
  return 0;
}

int pthread_mutex_init(pthread_mutex_t *mutex,const pthread_mutexattr_t *attr)
{
  status_t r;

  if( !mutex ) {
    return EINVAL;
  }

  r=__syscall4(SC_SYNC_CREATE,__SO_MUTEX,
               (syscall_arg_t)mutex,(syscall_arg_t)attr,0);
  return POSIX_ERR(r);
}

int pthread_mutex_lock(pthread_mutex_t *mutex)
{
  status_t r;

  if( !mutex ) {
    return EINVAL;
  }

  if( mutex->__id == __SO_UNINITIALIZED_ID ) {
    r=__syscall4(SC_SYNC_CREATE,__SO_MUTEX,
                 (syscall_arg_t)mutex,(syscall_arg_t)&mutex->__attrs,1);
  } else {
    r=MUTEX_CTRL(mutex,__SYNC_CMD_MUTEX_LOCK,0);
  }
  return POSIX_ERR(r);
}

int pthread_mutex_unlock(pthread_mutex_t *mutex)
{
  int r;

  if( !mutex || mutex->__id == __SO_UNINITIALIZED_ID ) {
    return EINVAL;
  }

  r = MUTEX_CTRL(mutex,__SYNC_CMD_MUTEX_UNLOCK,0);
  
  return POSIX_ERR(r);
}

int pthread_mutex_trylock(pthread_mutex_t *mutex)
{
  status_t r;

  if( !mutex ) {
    return EINVAL;
  }

  if( mutex->__id == __SO_UNINITIALIZED_ID ) {
    r=__syscall4(SC_SYNC_CREATE,__SO_MUTEX,
                 (syscall_arg_t)mutex,(syscall_arg_t)&mutex->__attrs,1);
  } else {  
    r=MUTEX_CTRL(mutex,__SYNC_CMD_MUTEX_TRYLOCK,0);
  }
  return POSIX_ERR(r);
}
