/*
 * This library is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU Lesser General Public License as 
 * published by the Free Software Foundation; either version 2.1 of 
 * the License.
 *
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public 
 * License along with this library; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 * (c) Copyright 2006,2007,2008 Jari OS Core Team <http://jarios.org>
 * (c) Copyright 2008 Michael Tsymbalyuk <mtzaurus@gmail.com>
 * (c) Copyright 2008 Dmitry Gromada <gromada@jarios.org>
 *
 * LibOS:
 * lib/ievent.c: Implementation of 'interrupt event', a machanism
 *               for delivering hardware interrupts to userspace.
 *
 */

#include <stdbool.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <unistd.h>
#include <arch/atomic.h>
#include <string.h>
#include <time.h>

#include <lock/lock.h>
#include <devio/devio.h>
#include <devio/irq.h>
#include <sys/ievent.h>

#define LOCK_EVENT(e) sem_lock(&e->lock)
#define UNLOCK_EVENT(e) sem_unlock(&e->lock)
#define INIT_LOCK(e) e->lock=0

static int __list_comparator(const void *a,const void *b)
{
  return a-b;
}

status_t interrupt_event_create(interrupt_event_t **e,irq_t *irqs,
                                uint16_t num_irqs)
{
  interrupt_event_t *event;
  status_t r;
  ulong_t i;

  *e=NULL;
  if(!e || !irqs || !num_irqs) {
    return -EINVAL;
  }

  event=(interrupt_event_t*)malloc(sizeof(interrupt_event_t));
  if(!event) {
    return -ENOMEM;
  }

  INIT_LOCK(event);
  r=-ENOMEM;

  /* Allocate IRQ array for interacting with kernel. */
  event->irq_array=mmap(NULL,PAGE_SIZE,PROT_READ|PROT_WRITE,MAP_ANON|MAP_POPULATE|MAP_PRIVATE,NOFD,0);
  if(event->irq_array == MAP_FAILED) {
    goto free_event;
  }

  /* Create low-level IRQ array. */
  r=__syscall4(SC_CREATE_IRQ_ARRAY,(ulong_t)irqs,num_irqs,
               (ulong_t)event->irq_array,0);
  if(r<0) {
    goto free_array;
  }
  event->array_id=r;

  event->num_irqs=num_irqs;
  r=-ENOMEM;

  /* Allocate array of lists for every IRQ line being watched. */
  event->handlers=(usrtc_t**)malloc(sizeof(usrtc_t*)*num_irqs);
  if(!event->handlers) {
    goto free_array;
  }

  event->irqs=(irq_t*)malloc(sizeof(irq_t)*num_irqs);
  if(!event->irqs) {
    goto free_handlers;
  }

  memcpy(event->irqs,irqs,sizeof(irq_t)*num_irqs);

  for(i=0;i<num_irqs;i++) {
    event->handlers[i]=(usrtc_t*)malloc(sizeof(usrtc_t));

    if(!event->handlers[i]) {
      while(i) {
        i--;
        free(event->handlers[i]);
      }
      goto free_irqs;
    }

    usrtc_init(event->handlers[i],USRTC_LIST,USRTC_COUNT_T_MAX,
               __list_comparator);
  }

  event->magic=IEV_UP;
  *e=event;
  return 0;
free_irqs:
  free(event->irqs);
free_handlers:
  free(event->handlers);
free_array:
  /* TODO: [mt] unmap page and free IRQ array here. */
free_event:
  free(event);
  return r;
}

status_t interrupt_event_destroy(interrupt_event_t *e)
{
  e->magic=IEV_DOWN;
  /* 
   * TODO: Free all resources properly,
   * notify the irq controller that this thread will
   * not more listen the irq
   */
  return 0;
}

void __process_pending_events(interrupt_event_t *e)
{
  ulong_t i;
  irq_event_mask_t mask=atomic_swap_long(&e->irq_array->ev_mask,0);
  irq_counter_t *counters=e->irq_array->counters;

  if( !mask ) {
    return;
  }

  for(i=0;mask && i<e->num_irqs;i++,mask>>=1) {
    if( (mask & 0x1) && counters[i] ) {
      usrtc_t *head=e->handlers[i];
      usrtc_node_t *node=usrtc_first(head);

      while(node) {
        irq_client_t *client=usrtc_node_getdata(node);

        if( client->handler(client->data) ) {
          counters[i]--;
        }
        node=usrtc_next(head,node);
      }
    }
  }
}

status_t interrupt_event_wait(interrupt_event_t *e)
{
  status_t r;

  if( !e || e->magic != IEV_UP) {
    return -EINVAL;
  }

  /* Flush any events arrived during handlers invocation.
   * This step gives us performance advantage.
   */
  __process_pending_events(e);

  r=__syscall1(SC_WAIT_ON_IRQ_ARRAY,e->array_id);
  if( r>=0 ) {
    /* Process new events. */
    __process_pending_events(e);
  }

  return r;
}

void *interrupt_event_set_handler(interrupt_event_t *e,irq_t irq,
                                  irq_handler_t old_handler,
                                  irq_handler_t handler,void *priv)
{
  int i,idx;
  void *prev_priv;

  if(!e || !handler || e->magic != IEV_UP) {
    return NULL;
  }

  idx=-1;
  prev_priv=NULL;

  LOCK_EVENT(e);

  for(i=0;i<e->num_irqs;i++) {
    if( e->irqs[i] == irq ) {
      idx=i;
      break;
    }
  }

  if( idx != -1 ) {
    usrtc_t *head=e->handlers[idx];
    usrtc_node_t *node=usrtc_first(head);

    while( node ) {
      irq_client_t *c=usrtc_node_getdata(node);

      if( c->handler == old_handler ) {
        c->handler=handler;
        prev_priv=c->data;
        c->data=priv;
	break;
      }
      node=usrtc_next(head,node);
    }
  }

  UNLOCK_EVENT(e);

  return prev_priv;
}

status_t interrupt_event_add_handler(interrupt_event_t *e,irq_t irq,
                                     irq_handler_t handler,void *priv)
{
  int i,idx;
  usrtc_node_t *node;
  irq_client_t *client;
  status_t  r;

  if(!e || !handler || e->magic != IEV_UP) {
    return -EINVAL;
  }

  LOCK_EVENT(e);
  idx=-1;
  for(i=0;i<e->num_irqs;i++) {
    if(e->irqs[i] == irq) {
      idx=i;
      break;
    }
  }
  UNLOCK_EVENT(e);

  if(idx==-1) {
    r=-EINVAL;
    goto out;
  }

  r=-ENOMEM;
  node=(usrtc_node_t*)malloc(sizeof(usrtc_node_t));
  if(!node) {
    goto out;
  }

  client=(irq_client_t*)malloc(sizeof(irq_client_t));
  if(!client) {
    goto free_node;
  }

  client->handler=handler;
  client->data=priv;

  LOCK_EVENT(e);
  usrtc_node_setdata(node,client);
  usrtc_insert(e->handlers[idx],node,client);
  UNLOCK_EVENT(e);

  /* Fixme [dg]: replace by hw_add_irq_listener */
  r = hw_mask_irq(irq, 0);

  return 0;
free_node:
  free(node);
out:
  return r;
}
